Package: httr2
Title: Perform HTTP Requests and Process the Responses
Version: 1.2.2
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person("Maximilian", "Girlich", role = "ctb")
  )
Description: Tools for creating and modifying HTTP requests, then
    performing them and processing the results. 'httr2' is a modern
    re-imagining of 'httr' that uses a pipe-based interface and solves
    more of the problems that API wrapping packages face.
License: MIT + file LICENSE
URL: https://httr2.r-lib.org, https://github.com/r-lib/httr2
BugReports: https://github.com/r-lib/httr2/issues
Depends: R (>= 4.1)
Imports: cli (>= 3.0.0), curl (>= 6.4.0), glue, lifecycle, magrittr,
        openssl, R6, rappdirs, rlang (>= 1.1.0), vctrs (>= 0.6.3),
        withr
Suggests: askpass, bench, clipr, covr, docopt, httpuv, jose, jsonlite,
        knitr, later (>= 1.4.0), nanonext, otel (>= 0.2.0), otelsdk (>=
        0.2.0), paws.common (>= 0.8.0), promises, rmarkdown, testthat
        (>= 3.1.8), tibble, webfakes (>= 1.4.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: resp-stream, req-perform
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2025-12-05 17:46:06 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit Software, PBC [cph, fnd],
  Maximilian Girlich [ctb]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2025-12-08 09:50:02 UTC
Built: R 4.6.0; ; 2026-04-26 05:17:04 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",httr2_headers)
S3method("[",httr2_headers)
S3method("[[",httr2_headers)
S3method(close,httr2_response)
S3method(format,httr2_redacted_sentinel)
S3method(print,httr2_cmd)
S3method(print,httr2_headers)
S3method(print,httr2_json)
S3method(print,httr2_oauth_client)
S3method(print,httr2_obfuscated)
S3method(print,httr2_redacted_sentinel)
S3method(print,httr2_request)
S3method(print,httr2_response)
S3method(print,httr2_token)
S3method(print,httr2_url)
S3method(str,httr2_headers)
S3method(str,httr2_obfuscated)
S3method(str,httr2_redacted_sentinel)
export("%>%")
export(StreamingBody)
export(curl_help)
export(curl_translate)
export(example_github_client)
export(example_url)
export(is_online)
export(iterate_with_cursor)
export(iterate_with_link_url)
export(iterate_with_offset)
export(jwt_claim)
export(jwt_encode_hmac)
export(jwt_encode_sig)
export(last_request)
export(last_request_json)
export(last_response)
export(last_response_json)
export(local_mocked_responses)
export(local_verbosity)
export(new_response)
export(oauth_cache_clear)
export(oauth_cache_path)
export(oauth_client)
export(oauth_client_req_auth)
export(oauth_client_req_auth_body)
export(oauth_client_req_auth_header)
export(oauth_client_req_auth_jwt_sig)
export(oauth_flow_auth_code)
export(oauth_flow_auth_code_listen)
export(oauth_flow_auth_code_parse)
export(oauth_flow_auth_code_pkce)
export(oauth_flow_auth_code_url)
export(oauth_flow_bearer_jwt)
export(oauth_flow_client_credentials)
export(oauth_flow_device)
export(oauth_flow_password)
export(oauth_flow_refresh)
export(oauth_flow_token_exchange)
export(oauth_redirect_uri)
export(oauth_token)
export(oauth_token_cached)
export(obfuscate)
export(obfuscated)
export(req_auth_aws_v4)
export(req_auth_basic)
export(req_auth_bearer_token)
export(req_body_file)
export(req_body_form)
export(req_body_json)
export(req_body_json_modify)
export(req_body_multipart)
export(req_body_raw)
export(req_cache)
export(req_cookie_preserve)
export(req_cookies_set)
export(req_dry_run)
export(req_error)
export(req_get_body)
export(req_get_body_type)
export(req_get_headers)
export(req_get_method)
export(req_get_url)
export(req_headers)
export(req_headers_redacted)
export(req_method)
export(req_oauth)
export(req_oauth_auth_code)
export(req_oauth_bearer_jwt)
export(req_oauth_client_credentials)
export(req_oauth_device)
export(req_oauth_password)
export(req_oauth_refresh)
export(req_oauth_token_exchange)
export(req_options)
export(req_perform)
export(req_perform_connection)
export(req_perform_iterative)
export(req_perform_parallel)
export(req_perform_promise)
export(req_perform_sequential)
export(req_perform_stream)
export(req_progress)
export(req_proxy)
export(req_retry)
export(req_template)
export(req_throttle)
export(req_timeout)
export(req_url)
export(req_url_path)
export(req_url_path_append)
export(req_url_query)
export(req_url_relative)
export(req_user_agent)
export(req_verbose)
export(request)
export(resp_body_html)
export(resp_body_json)
export(resp_body_raw)
export(resp_body_string)
export(resp_body_xml)
export(resp_check_content_type)
export(resp_check_status)
export(resp_content_type)
export(resp_date)
export(resp_encoding)
export(resp_has_body)
export(resp_header)
export(resp_header_exists)
export(resp_headers)
export(resp_is_error)
export(resp_link_url)
export(resp_raw)
export(resp_request)
export(resp_retry_after)
export(resp_status)
export(resp_status_desc)
export(resp_stream_aws)
export(resp_stream_is_complete)
export(resp_stream_lines)
export(resp_stream_raw)
export(resp_stream_sse)
export(resp_timing)
export(resp_url)
export(resp_url_path)
export(resp_url_queries)
export(resp_url_query)
export(response)
export(response_json)
export(resps_data)
export(resps_failures)
export(resps_ok)
export(resps_requests)
export(resps_successes)
export(secret_decrypt)
export(secret_decrypt_file)
export(secret_encrypt)
export(secret_encrypt_file)
export(secret_has_key)
export(secret_make_key)
export(secret_read_rds)
export(secret_write_rds)
export(signal_total_pages)
export(throttle_status)
export(url_build)
export(url_modify)
export(url_modify_query)
export(url_modify_relative)
export(url_parse)
export(url_query_build)
export(url_query_parse)
export(with_mocked_responses)
export(with_verbosity)
import(R6)
import(rlang)
importFrom(glue,glue)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")

<!-- README.md is generated from README.Rmd. Please edit that file -->

# httr2 <a href="https://httr2.r-lib.org"><img src="man/figures/logo.png" align="right" height="138" alt="httr2 website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/httr2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/httr2/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/httr2/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/httr2?branch=main)
<!-- badges: end -->

httr2 (pronounced “hitter2”) is a comprehensive HTTP client that
provides a modern, pipeable API for working with web APIs. It builds on
top of [{curl}](https://jeroen.r-universe.dev/curl) to provide features
like explicit request objects, built-in rate limiting & retry tooling,
comprehensive OAuth support, and secure handling of secrets and
credentials.

## Installation

You can install httr2 from CRAN with:

``` r
install.packages("httr2")
```

## Usage

To use httr2, start by creating a **request**:

``` r
library(httr2)

req <- request("https://r-project.org")
req
#> <httr2_request>
#> GET https://r-project.org
#> Body: empty
```

You can tailor this request with the `req_` family of functions:

``` r
# Add custom headers
req |> req_headers("Accept" = "application/json")
#> <httr2_request>
#> GET https://r-project.org
#> Headers:
#> * Accept: "application/json"
#> Body: empty

# Add a body, turning it into a POST
req |> req_body_json(list(x = 1, y = 2))
#> <httr2_request>
#> POST https://r-project.org
#> Body: JSON data

# Modify the path in the url
req |> req_url_path(path = "path/to/my/file")
#> <httr2_request>
#> GET https://r-project.org/path/to/my/file
#> Body: empty

# Automatically retry if the request fails
req |> req_retry(max_tries = 5)
#> <httr2_request>
#> GET https://r-project.org
#> Body: empty
#> Policies:
#> * retry_max_tries        : 5
#> * retry_on_failure       : FALSE
#> * retry_failure_threshold: Inf
#> * retry_failure_timeout  : 30
#> * retry_realm            : "r-project.org"

# Change the HTTP method
req |> req_method("PATCH")
#> <httr2_request>
#> PATCH https://r-project.org
#> Body: empty
```

And see exactly what httr2 will send to the server with `req_dry_run()`:

``` r
req |> req_dry_run()
#> GET / HTTP/1.1
#> accept: */*
#> accept-encoding: deflate, gzip
#> host: r-project.org
#> user-agent: httr2/1.2.1.9000 r-curl/7.0.0 libcurl/8.14.1
```

Use `req_perform()` to perform the request, retrieving a **response**:

``` r
resp <- req_perform(req)
resp
#> <httr2_response>
#> GET https://www.r-project.org/
#> Status: 200 OK
#> Content-Type: text/html
#> Body: In memory (6238 bytes)
```

The `resp_` functions help you extract various useful components of the
response:

``` r
resp |> resp_content_type()
#> [1] "text/html"
resp |> resp_status_desc()
#> [1] "OK"
resp |> resp_body_html()
#> {html_document}
#> <html lang="en">
#> [1] <head>\n<meta http-equiv="Content-Type" content="text/html; charset=UTF-8 ...
#> [2] <body>\n    <div class="container page">\n      <div class="row">\n       ...
```

## Major differences to httr

- You can now create and modify a request without performing it. This
  means that there’s now a single function to perform the request and
  fetch the result: `req_perform()`. `req_perform()` replaces
  `httr::GET()`, `httr::POST()`, `httr::DELETE()`, and more.

- HTTP errors are automatically converted into R errors. Use
  `req_error()` to override the defaults (which turn all 4xx and 5xx
  responses into errors) or to add additional details to the error
  message.

- You can automatically retry if the request fails or encounters a
  transient HTTP error (e.g. a 429 rate limit request). `req_retry()`
  defines the maximum number of retries, which errors are transient, and
  how long to wait between tries.

- OAuth support has been totally overhauled to directly support many
  more flows and to make it much easier to both customise the built-in
  flows and to create your own.

- You can manage secrets (often needed for testing) with
  `secret_encrypt()` and friends. You can obfuscate mildly confidential
  data with `obfuscate()`, preventing it from being scraped from
  published code.

- You can automatically cache all cacheable results with `req_cache()`.
  Relatively few API responses are cacheable, but when they are it
  typically makes a big difference.

## Acknowledgements

httr2 wouldn’t be possible without [curl](https://cran.dev/curl/),
[openssl](https://cran.dev/openssl/),
[jsonlite](https://cran.dev/jsonlite/), and
[jose](https://github.com/r-lib/jose/), which are all maintained by
[Jeroen Ooms](https://github.com/jeroen). A big thanks also go to [Jenny
Bryan](https://jennybryan.org) and [Craig
Citro](https://www.craigcitro.org) who have given me much useful
feedback on both the design of the internals and the user facing API.

---
output: github_document
---

<!-- README.md is generated from README.Rmd. Please edit that file -->

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "man/figures/README-",
  out.width = "100%"
)
```

# httr2 <a href="https://httr2.r-lib.org"><img src="man/figures/logo.png" align="right" height="138" alt="httr2 website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/httr2/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/httr2/actions)
[![Codecov test coverage](https://codecov.io/gh/r-lib/httr2/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/httr2?branch=main)
<!-- badges: end -->

httr2 (pronounced hitter2) is a ground-up rewrite of [httr](https://httr.r-lib.org) that provides a pipeable API with an explicit request object that solves more problems felt by packages that wrap APIs (e.g. built-in rate-limiting, retries, OAuth, secure secrets, and more).

## Installation

You can install httr2 from CRAN with:

``` r
install.packages("httr2")
```

## Usage

To use httr2, start by creating a **request**:

```{r}
library(httr2)

req <- request("https://r-project.org")
req
```

You can tailor this request with the `req_` family of functions:

```{r}
# Add custom headers
req |> req_headers("Accept" = "application/json")

# Add a body, turning it into a POST
req |> req_body_json(list(x = 1, y = 2))

# Automatically retry if the request fails
req |> req_retry(max_tries = 5)

# Change the HTTP method
req |> req_method("PATCH")
```

And see exactly what httr2 will send to the server with `req_dry_run()`:

```{r}
req |> req_dry_run()
```

Use `req_perform()` to perform the request, retrieving a **response**:

```{r}
resp <- req_perform(req)
resp
```

The `resp_` functions help you extract various useful components of the response:

```{r}
resp |> resp_content_type()
resp |> resp_status_desc()
resp |> resp_body_html()
```

## Major differences to httr

-   You can now create and modify a request without performing it.
    This means that there's now a single function to perform the request and fetch the result: `req_perform()`.
    `req_perform()` replaces `httr::GET()`, `httr::POST()`, `httr::DELETE()`, and more.
  
-   HTTP errors are automatically converted into R errors.
    Use `req_error()` to override the defaults (which turn all 4xx and 5xx responses into errors) or to add additional details to the error message.

-   You can automatically retry if the request fails or encounters a transient HTTP error (e.g. a 429 rate limit request).
    `req_retry()` defines the maximum number of retries, which errors are transient, and how long to wait between tries.

-   OAuth support has been totally overhauled to directly support many more flows and to make it much easier to both customise the built-in flows and to create your own.

-   You can manage secrets (often needed for testing) with `secret_encrypt()` and friends.
    You can obfuscate mildly confidential data with `obfuscate()`, preventing it from being scraped from published code.

-   You can automatically cache all cacheable results with `req_cache()`.
    Relatively few API responses are cacheable, but when they are it typically makes a big difference.

## Acknowledgements

httr2 wouldn't be possible without [curl](https://jeroen.cran.dev/curl/), [openssl](https://github.com/jeroen/openssl/), [jsonlite](https://jeroen.cran.dev/jsonlite/), and [jose](https://github.com/r-lib/jose/), which are all maintained by [Jeroen Ooms](https://github.com/jeroen).
A big thanks also go to [Jenny Bryan](https://jennybryan.org) and [Craig Citro](https://research.google/people/CraigCitro/) who have given me much useful feedback on both the design of the internals and the user facing API.

#' Perform a list of requests in parallel
#'
#' @description
#' This variation on [req_perform_sequential()] performs multiple requests in
#' parallel. Exercise caution when using this function; it's easy to pummel a
#' server with many simultaneous requests. Only use it with hosts designed to
#' serve many files at once, which are typically web servers, not API servers.
#'
#' `req_perform_parallel()` has a few limitations:
#'
#' * Will not retrieve a new OAuth token if it expires part way through
#'   the requests.
#' * Does not perform throttling with [req_throttle()].
#' * Does not attempt retries as described by [req_retry()].
#' * Only consults the cache set by [req_cache()] before/after all requests.
#'
#' If any of these limitations are problematic for your use case, we recommend
#' [req_perform_sequential()] instead.
#'
#' @inherit req_perform_sequential params return
#' @param pool Optionally, a curl pool made by [curl::new_pool()]. Supply
#'   this if you want to override the defaults for total concurrent connections
#'   (100) or concurrent connections per host (6).
#' @export
#' @examples
#' # Requesting these 4 pages one at a time would take 2 seconds:
#' request_base <- request(example_url())
#' reqs <- list(
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5")
#' )
#' # But it's much faster if you request in parallel
#' system.time(resps <- req_perform_parallel(reqs))
#'
#' # req_perform_parallel() will fail on error
#' reqs <- list(
#'   request_base |> req_url_path("/status/200"),
#'   request_base |> req_url_path("/status/400"),
#'   request("FAILURE")
#' )
#' try(resps <- req_perform_parallel(reqs))
#'
#' # but can use on_error to capture all successful results
#' resps <- req_perform_parallel(reqs, on_error = "continue")
#'
#' # Inspect the successful responses
#' resps |> resps_successes()
#'
#' # And the failed responses
#' resps |> resps_failures() |> resps_requests()
req_perform_parallel <- function(reqs,
                                 paths = NULL,
                                 pool = NULL,
                                 on_error = c("stop", "return", "continue"),
                                 progress = TRUE) {
  check_paths(paths, reqs)
  on_error <- arg_match(on_error)

  progress <- create_progress_bar(
    total = length(reqs),
    name = "Iterating",
    config = progress
  )

  perfs <- vector("list", length(reqs))
  for (i in seq_along(reqs)) {
    perfs[[i]] <- Performance$new(
      req = reqs[[i]],
      path = paths[[i]],
      progress = progress,
      error_call = environment()
    )
    perfs[[i]]$submit(pool)
  }

  pool_run(pool, perfs, on_error = on_error)
  progress$done()

  map(perfs, ~ .$resp)
}


#' Perform a list of requests in parallel
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' Please use [req_perform_parallel()] instead, and note:
#'
#' * `cancel_on_error = FALSE` is now `on_error = "continue"`
#' * `cancel_on_error = TRUE` is now `on_error = "return"`
#'
#' @export
#' @param cancel_on_error Should all pending requests be cancelled when you
#'   hit an error? Set this to `TRUE` to stop all requests as soon as you
#'   hit an error. Responses that were never performed be `NULL` in the result.
#' @inheritParams req_perform_parallel
#' @keywords internal
multi_req_perform <- function(reqs,
                              paths = NULL,
                              pool = NULL,
                              cancel_on_error = FALSE) {
  lifecycle::deprecate_warn(
    "1.0.0",
    "multi_req_perform()",
    "req_perform_parallel()"
  )
  check_bool(cancel_on_error)

  req_perform_parallel(
    reqs = reqs,
    paths = paths,
    pool = pool,
    on_error = if (cancel_on_error) "continue" else "return"
  )
}

pool_run <- function(pool, perfs, on_error = "continue") {
  on.exit(pool_cancel(pool, perfs), add = TRUE)

  # The done and fail callbacks for curl::multi_add() are designed to always
  # succeed. If the request actually failed, they raise a `httr_fail`
  # signal (not error) that wraps the error. Here we catch that error and
  # handle it based on the value of `on_error`
  httr2_fail <- switch(on_error,
    stop =     function(cnd) cnd_signal(cnd$error),
    continue = function(cnd) zap(),
    return =   function(cnd) NULL
  )

  try_fetch(
    repeat({
      run <- curl::multi_run(0.1, pool = pool, poll = TRUE)
      if (run$pending == 0) {
        break
      }
    }),
    interrupt = function(cnd) NULL,
    httr2_fail = httr2_fail
  )

  invisible()
}

# Wrap up all components of request -> response in a single object
Performance <- R6Class("Performance", public = list(
  req = NULL,
  path = NULL,

  handle = NULL,
  resp = NULL,
  pool = NULL,
  error_call = NULL,
  progress = NULL,

  initialize = function(req, path = NULL, progress = NULL, error_call = NULL) {
    self$req <- req
    self$path <- path
    self$progress <- progress
    self$error_call <- error_call

    req <- auth_oauth_sign(req)
    req <- cache_pre_fetch(req)
    if (is_response(req)) {
      self$resp <- req
    } else {
      self$handle <- req_handle(req)
      curl::handle_setopt(self$handle, url = req$url)
    }
  },

  submit = function(pool = NULL) {
    if (!is.null(self$resp)) {
      # cached
      return()
    }

    self$pool <- pool
    curl::multi_add(
      handle = self$handle,
      pool = self$pool,
      data = self$path,
      done = self$succeed,
      fail = self$fail
    )
    invisible(self)
  },

  succeed = function(res) {
    self$progress$update()

    body <- if (is.null(self$path)) res$content else new_path(self$path)
    resp <- new_response(
      method = req_method_get(self$req),
      url = res$url,
      status_code = res$status_code,
      headers = as_headers(res$headers),
      body = body,
      request = self$req
    )
    resp <- cache_post_fetch(self$reqs, resp, path = self$paths)

    self$resp <- tryCatch(
      resp_check_status(resp, error_call = self$error_call),
      error = identity
    )
    if (is_error(self$resp)) {
      signal("", error = self$resp, class = "httr2_fail")
    }
  },

  fail = function(msg) {
    self$progress$update()

    self$resp <- error_cnd(
      "httr2_failure",
      message = msg,
      request = self$req,
      call = self$error_call
    )
    signal("", error = self$resp, class = "httr2_fail")
  },

  cancel = function() {
    # No handle if response was cached
    if (!is.null(self$handle)) {
      curl::multi_cancel(self$handle)
    }
  }
))

pool_cancel <- function(pool, perfs) {
  walk(perfs, ~ .x$cancel())
  curl::multi_run(pool = pool)
}

test_that("request and paths must match", {
  req <- request("http://example.com")
  expect_snapshot(req_perform_parallel(req, letters), error = TRUE)
})

test_that("requests happen in parallel", {
  # GHA MacOS builder seems to be very slow
  skip_if(
    isTRUE(as.logical(Sys.getenv("CI", "false"))) &&
    Sys.info()[["sysname"]] == "Darwin"
  )

  reqs <- list2(
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
  )
  time <- system.time(req_perform_parallel(reqs))
  expect_lt(time[[3]], 1)
})

test_that("can download files", {
  reqs <- list(request_test("/json"), request_test("/html"))
  paths <- c(withr::local_tempfile(), withr::local_tempfile())
  resps <- req_perform_parallel(reqs, paths)

  expect_equal(resps[[1]]$body, new_path(paths[[1]]))
  expect_equal(resps[[2]]$body, new_path(paths[[2]]))

  # And check that something was downloaded
  expect_gt(file.size(paths[[1]]), 0)
  expect_gt(file.size(paths[[2]]), 0)
})

test_that("immutable objects retrieved from cache", {
  req <- request("http://example.com") %>% req_cache(tempfile())
  resp <- response(200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  expect_condition(
    resps <- req_perform_parallel(list(req)),
    class = "httr2_cache_cached"
  )
  expect_equal(resps[[1]], resp)
})

test_that("errors by default", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request("INVALID")
  )
  expect_snapshot(error = TRUE, {
    req_perform_parallel(reqs[1])
    req_perform_parallel(reqs[2])
  })
})

test_that("both curl and HTTP errors become errors on continue", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request("INVALID"),
  )
  out <- req_perform_parallel(reqs, on_error = "continue")
  expect_s3_class(out[[1]], "httr2_http_404")
  expect_s3_class(out[[2]], "httr2_failure")

  # and contain the responses
  expect_equal(out[[1]]$request, reqs[[1]])
  expect_equal(out[[2]]$request, reqs[[2]])
})

test_that("errors can cancel outstanding requests", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request_test("/delay/:secs", secs = 2),
  )
  out <- req_perform_parallel(reqs, on_error = "return")
  expect_s3_class(out[[1]], "httr2_http_404")
  expect_null(out[[2]])
})

test_that("multi_req_perform is deprecated", {
  expect_snapshot(multi_req_perform(list()))
})

#' Retrieve most recent request/response
#'
#' @description
#' `last_request()` and `last_response()` retrieve the most recent request
#' made by httr2 and the response it received, to facilitate debugging problems
#' _after_ they occur.
#'
#' `last_request_json()` and `last_response_json()` return the JSON bodies of
#' the most recent request and response. They will error if not JSON.
#'
#' @returns `last_request()` and `last_response()` return an HTTP
#'   [request] or [response] respectively. If no request has been made,
#'   `last_request()` will return `NULL`; if no request has been made
#'   or the last request was unsuccessful, `last_response()` will return
#'   `NULL`.
#'
#'   `last_request_json()` and `last_response_json()` always return a string.
#'   They will error if `last_request()` or `last_response()` are `NULL` or
#'   don't have JSON bodies.
#' @export
#' @examples
#' . <- request("http://httr2.r-lib.org") |> req_perform()
#' last_request()
#' last_response()
#'
#' . <- request(example_url("/post")) |>
#'   req_body_json(list(a = 1, b = 2)) |>
#'   req_perform()
#' last_request_json()
#' last_request_json(pretty = FALSE)
#' last_response_json()
#' last_response_json(pretty = FALSE)
last_response <- function() {
  the$last_response
}

#' @export
#' @rdname last_response
last_request <- function() {
  the$last_request
}

#' @export
#' @rdname last_response
last_request_json <- function(pretty = TRUE) {
  req <- last_request()
  if (is.null(req)) {
    cli::cli_abort("No request has been made yet.")
  }
  if (req_get_body_type(req) != "json") {
    cli::cli_abort("Last request doesn't have a JSON body.")
  }
  httr2_json(req_body_render(req), pretty = pretty)
}

#' @param pretty Should the JSON be pretty-printed?
#' @export
#' @rdname last_response
last_response_json <- function(pretty = TRUE) {
  resp <- last_response()
  if (is.null(resp)) {
    cli::cli_abort("No request has been made successfully yet.")
  }
  if (!identical(resp_content_type(resp), "application/json")) {
    cli::cli_abort("Last response doesn't have a JSON body.")
  }
  httr2_json(resp_body_string(resp), pretty = pretty)
}

httr2_json <- function(x, pretty = TRUE) {
  check_string(x)
  structure(x, pretty = pretty, class = "httr2_json")
}
#' @export
print.httr2_json <- function(x, ...) {
  if (attr(x, "pretty")) {
    cat(pretty_json(x))
  } else {
    cat(x)
  }

  cat("\n")
  invisible(x)
}

test_that("can retrieve last request and response", {
  req <- request_test()
  resp <- req_perform(req)

  expect_equal(last_request(), req)
  expect_equal(last_response(), resp)
})

test_that("last response is NULL if it fails", {
  req <- request("")
  try(req_perform(req), silent = TRUE)

  expect_equal(last_request(), req)
  expect_equal(last_response(), NULL)
})

test_that("returns NULL if no last request/response", {
  the$last_request <- NULL
  the$last_response <- NULL

  expect_equal(last_request(), NULL)
  expect_equal(last_response(), NULL)
})

# JSON -----------------------------------------------------------------------

test_that("can get json response", {
  req <- local_app_request(function(req, res) {
    res$set_status(200L)$send_json(text = '{"x":1}')
  })
  req_perform(req)

  expect_snapshot({
    last_response_json()
    last_response_json(pretty = FALSE)
  })
})

test_that("can get json request", {
  request(example_url("/post")) |>
    req_body_json(list(x = 1)) |>
    req_perform()

  expect_snapshot({
    last_request_json()
    last_request_json(pretty = FALSE)
  })
})

test_that("useful errors if not json request/response", {
  req_perform(request(example_url("/xml")))

  expect_snapshot(error = TRUE, {
    last_request_json()
    last_response_json()
  })
})

test_that("useful errors if no last request/response", {
  the$last_request <- NULL
  the$last_response <- NULL

  expect_snapshot(error = TRUE, {
    last_request_json()
    last_response_json()
  })
})

test_that("req_perform_stream() has been soft deprecated", {
  req <- request_test("/stream-bytes/1024")
  expect_snapshot(. <- req_perform_stream(req, \(x) NULL))
})

# req_perform_stream() --------------------------------------------------------

test_that("returns stream body; sets last request & response", {
  withr::local_options(lifecycle_verbosity = "quiet")

  req <- request_test("/stream-bytes/1024")
  resp <- req_perform_stream(req, function(x) NULL)
  expect_s3_class(resp, "httr2_response")
  expect_false(resp_has_body(resp))

  expect_equal(last_request(), req_policies(req, connection = TRUE))
  expect_equal(last_response(), resp)
})

test_that("HTTP errors become R errors", {
  withr::local_options(lifecycle_verbosity = "quiet")

  req <- request_test("/status/404")
  expect_error(
    req_perform_stream(req, function(x) TRUE),
    class = "httr2_http_404"
  )

  resp <- last_response()
  expect_s3_class(resp, "httr2_response")
  expect_equal(resp$status_code, 404)
})

test_that("can override error handling", {
  withr::local_options(lifecycle_verbosity = "quiet")

  req <- request_test("/base64/:value", value = "YWJj") |>
    req_error(is_error = function(resp) TRUE)

  expect_error(
    req |> req_perform_stream(function(x) NULL),
    class = "httr2_http_200"
  )

  resp <- last_response()
  expect_s3_class(resp, "httr2_response")
  # This also allows us to check that the body is set correctly
  # since that httpbin error responses have empty bodies
  expect_equal(resp_body_string(resp), "abc")
})

test_that("can buffer to lines", {
  withr::local_options(lifecycle_verbosity = "quiet")

  lines <- character()
  accumulate_lines <- function(x) {
    lines <<- c(lines, strsplit(rawToChar(x), "\n", fixed = TRUE)[[1]])
    TRUE
  }

  # Each line is 225 bytes, should should be split into ~2 pieces
  resp <- request_test("/stream/10") |>
    req_perform_stream(accumulate_lines, buffer_kb = 0.1, round = "line")
  expect_equal(length(lines), 10)

  valid_json <- map_lgl(lines, jsonlite::validate)
  expect_equal(valid_json, rep(TRUE, 10))
})

test_that("can supply custom rounding", {
  withr::local_options(lifecycle_verbosity = "quiet")

  out <- list()
  accumulate <- function(x) {
    out <<- c(out, list(x))
    TRUE
  }

  resp <- request_test("/stream-bytes/1024") |>
    req_perform_stream(
      accumulate,
      buffer_kb = 0.1,
      round = function(bytes) if (length(bytes) > 100) 100 else integer()
    )
  expect_equal(lengths(out), c(rep(100, 10), 24))
})

test_that("eventually terminates even if never rounded", {
  withr::local_options(lifecycle_verbosity = "quiet")

  out <- raw()
  accumulate <- function(x) {
    out <<- c(out, x)
    TRUE
  }

  resp <- request_test("/stream-bytes/1024") |>
    req_perform_stream(
      accumulate,
      buffer_kb = 0.1,
      round = function(bytes) integer()
    )
  expect_equal(length(out), 1024)
})


test_that("req_perform_stream checks its inputs", {
  withr::local_options(lifecycle_verbosity = "quiet")

  req <- request_test("/stream-bytes/1024")
  callback <- function(x) NULL

  expect_snapshot(error = TRUE, {
    req_perform_stream(1)
    req_perform_stream(req, 1)
    req_perform_stream(req, callback, timeout_sec = -1)
    req_perform_stream(req, callback, buffer_kb = "x")
  })
})

test_that("as_round_function checks its inputs", {
  expect_snapshot(error = TRUE, {
    as_round_function(1)
    as_round_function("bytes")
    as_round_function(function(x) 1)
  })
})

#' Perform requests iteratively, generating new requests from previous responses
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `req_perform_iterative()` iteratively generates and performs requests,
#' using a callback function, `next_req`, to define the next request based on
#' the current request and response. You will probably want to pair it with an
#' [iteration helper][iterate_with_offset] and use a
#' [multi-response handler][resps_successes] to process the result.
#'
#' # `next_req()`
#'
#' The key piece that makes `req_perform_iterative()` work is the `next_req()`
#' argument. For most common cases, you can use one of the canned helpers,
#' like [iterate_with_offset()]. If, however, the API you're wrapping uses a
#' different pagination system, you'll need to write your own. This section
#' gives some advice.
#'
#' Generally, your function needs to inspect the response, extract some data
#' from it, then use that to modify the previous request. For example, imagine
#' that the response returns a cursor, which needs to be added to the body of
#' the request. The simplest version of this function might look like this:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   cursor <- resp_body_json(resp)$next_cursor
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' There's one problem here: if there are no more pages to return, then
#' `cursor` will be `NULL`, but `req_body_json_modify()` will still generate
#' a meaningful request. So we need to handle this specifically by
#' returning `NULL`:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   cursor <- resp_body_json(resp)$next_cursor
#'   if (is.null(cursor))
#'     return(NULL)
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' A value of `NULL` lets `req_perform_iterative()` know there are no more
#' pages remaining.
#'
#' There's one last feature you might want to add to your iterator: if you
#' know the total number of pages, then it's nice to let
#' `req_perform_iterative()` know so it can adjust the progress bar.
#' (This will only ever decrease the number of pages, not increase it.)
#' You can signal the total number of pages by calling [signal_total_pages()],
#' like this:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   body <- resp_body_json(resp)
#'   cursor <- body$next_cursor
#'   if (is.null(cursor))
#'     return(NULL)
#'
#'   signal_total_pages(body$pages)
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' @inheritParams req_perform_sequential
#' @param req The first [request] to perform.
#' @param next_req A function that takes the previous response (`resp`) and
#'   request (`req`) and returns a [request] for the next page or `NULL` if
#'   the iteration should terminate. See below for more details.
#' @param max_reqs The maximum number of requests to perform. Use `Inf` to
#'   perform all requests until `next_req()` returns `NULL`.
#' @param on_error What should happen if a request fails?
#'
#'   * `"stop"`, the default: stop iterating with an error.
#'   * `"return"`: stop iterating, returning all the successful responses so
#'     far, as well as an error object for the failed request.
#' @param path Optionally, path to save the body of request. This should be
#'   a glue string that uses `{i}` to distinguish different requests.
#'   Useful for large responses because it avoids storing the response in
#'   memory.
#' @return
#' A list, at most length `max_reqs`, containing [response]s and possibly one
#' error object, if `on_error` is `"return"` and one of the requests errors.
#' If present, the error object will always be the last element in the list.
#'
#' Only httr2 errors are captured; see [req_error()] for more details.
#' @export
#' @examples
#' req <- request(example_url()) |>
#'   req_url_path("/iris") |>
#'   req_throttle(10) |>
#'   req_url_query(limit = 5)
#'
#' resps <- req_perform_iterative(req, iterate_with_offset("page_index"))
#'
#' resps |> resps_data(function(resp) {
#'   data <- resp_body_json(resp)$data
#'   data.frame(
#'     Sepal.Length = sapply(data, `[[`, "Sepal.Length"),
#'     Sepal.Width = sapply(data, `[[`, "Sepal.Width"),
#'     Petal.Length = sapply(data, `[[`, "Petal.Length"),
#'     Petal.Width = sapply(data, `[[`, "Petal.Width"),
#'     Species = sapply(data, `[[`, "Species")
#'   )
#' })
req_perform_iterative <- function(req,
                                  next_req,
                                  path = NULL,
                                  max_reqs = 20,
                                  on_error = c("stop", "return"),
                                  progress = TRUE) {
  check_request(req)
  check_function2(next_req, args = c("resp", "req"))
  check_number_whole(max_reqs, allow_infinite = TRUE, min = 1)
  check_string(path, allow_empty = FALSE, allow_null = TRUE)
  on_error <- arg_match(on_error)

  get_path <- function(i) {
    if (is.null(path)) {
      NULL
    } else {
      glue::glue(path)
    }
  }

  progress <- create_progress_bar(
    total = max_reqs,
    name = "Iterating",
    config = progress
  )

  resps <- vector("list", length = if (is.finite(max_reqs)) max_reqs else 100)
  i <- 1L

  tryCatch({
    repeat {
      httr2_error <- switch(on_error,
        stop = function(cnd) zap(),
        return = function(cnd) cnd
      )
      resp <- try_fetch(
        req_perform(req, path = get_path(i)),
        httr2_error = httr2_error
      )
      resps[[i]] <- resp

      if (on_error == "return" && is_error(resp)) {
        break
      }

      progress$update()

      withCallingHandlers(
        {
          req <- next_req(resp = resp, req = req)
        },
        httr2_total_pages = function(cnd) {
          # Allow next_req() to shrink the number of pages remaining
          # Most important in max_req = Inf case
          if (cnd$n < max_reqs) {
            max_reqs <<- cnd$n
            progress$update(total = max_reqs, inc = 0)
          }
        }
      )

      if (is.null(req) || i >= max_reqs) {
        break
      }
      check_request(req, arg = "next_req()")

      i <- i + 1L
      if (i > length(resps)) {
        signal("", class = "httr2:::doubled")
        length(resps) <- length(resps) * 2
      }
    }
  }, interrupt = function(cnd) {
    # interrupt might occur after i was incremented
    if (is.null(resps[[i]])) {
      i <<- i - 1
    }
    cli::cli_alert_warning(
      "Terminating iteration; returning {i} response{?s}."
    )
  })
  progress$done()

  if (i < length(resps)) {
    resps <- resps[seq_len(i)]
  }

  resps
}

#' Is your computer currently online?
#'
#' This function uses some cheap heuristics to determine if your computer is
#' currently online. It's a simple wrapper around [curl::has_internet()]
#' exported from httr2 for convenience.
#'
#' @export
#' @examples
#' is_online()
is_online <- function() {
  curl::has_internet()
}

#' Set arbitrary curl options in request
#'
#' `req_options()` is for expert use only; it allows you to directly set
#' libcurl options to access features that are otherwise not available in
#' httr2.
#'
#' @inheritParams req_headers
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> Name-value pairs. The name
#'   should be a valid curl option, as found in [curl::curl_options()].
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # req_options() allows you to access curl options that are not otherwise
#' # exposed by httr2. For example, in very special cases you may need to
#' # turn off SSL verification. This is generally a bad idea so httr2 doesn't
#' # provide a convenient wrapper, but if you really know what you're doing
#' # you can still access this libcurl option:
#' req <- request("https://example.com") |>
#'   req_options(ssl_verifypeer = 0)
req_options <- function(.req, ...) {
  check_request(.req)

  .req$options <- modify_list(.req$options, ...)
  .req
}

#' Set user-agent for a request
#'
#' This overrides the default user-agent set by httr2 which includes the
#' version numbers of httr2, the curl package, and libcurl.
#'
#' @inheritParams req_perform
#' @param string String to be sent in the `User-Agent` header. If `NULL`,
#'   will user default.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # Default user-agent:
#' request("http://example.com") |> req_dry_run()
#'
#' request("http://example.com") |> req_user_agent("MyString") |> req_dry_run()
#'
#' # If you're wrapping in an API in a package, it's polite to set the
#' # user agent to identify your package.
#' request("http://example.com") |>
#'   req_user_agent("MyPackage (http://mypackage.com)") |>
#'   req_dry_run()
req_user_agent <- function(req, string = NULL) {
  check_request(req)

  if (is.null(string)) {
    string <- env_cache(the, "user_agent", default_user_agent())
  } else {
    check_string(string)
  }

  req_options(req, useragent = string)
}

default_user_agent <- function() {
  versions <- c(
    httr2 = as.character(utils::packageVersion("httr2")),
    `r-curl` = as.character(utils::packageVersion("curl")),
    libcurl = curl_system_version()
  )
  paste0(names(versions), "/", versions, collapse = " ")
}

req_has_user_agent <- function(req) {
  has_name(req$options, "useragent")
}

curl_system_version <- function() curl::curl_version()$version

#' Set time limit for a request
#'
#' An error will be thrown if the request does not complete in the time limit.
#'
#' @inheritParams req_perform
#' @param seconds Maximum number of seconds to wait
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # Give up after at most 10 seconds
#' request("http://example.com") |> req_timeout(10)
req_timeout <- function(req, seconds) {
  check_request(req)
  check_number_decimal(seconds)
  if (seconds < 0.001) {
    cli::cli_abort("{.arg seconds} must be >1 ms.")
  }

  req_options(
    req,
    timeout_ms = seconds * 1000,
    # reset value set by curl
    # https://github.com/jeroen/curl/blob/1bcf1ab3/src/handle.c#L159
    connecttimeout = 0
  )
}


#' Use a proxy for a request
#'
#' @inheritParams req_perform
#' @param url,port Location of proxy.
#' @param username,password Login details for proxy, if needed.
#' @param auth Type of HTTP authentication to use. Should be one of the
#'   following: `basic`, `digest`, `digest_ie`, `gssnegotiate`, `ntlm`, `any`.
#' @examples
#' # Proxy from https://www.proxynova.com/proxy-server-list/
#' \dontrun{
#' request("http://hadley.nz") |>
#'   req_proxy("20.116.130.70", 3128) |>
#'   req_perform()
#' }
#' @export
req_proxy <- function(
  req,
  url,
  port = NULL,
  username = NULL,
  password = NULL,
  auth = "basic"
) {
  if (!is.null(username) || !is.null(password)) {
    proxyuserpwd <- paste0(username, ":", password)
  } else {
    proxyuserpwd <- NULL
  }

  check_number_whole(port, allow_null = TRUE)

  req_options(
    req,
    proxy = url,
    proxyport = port,
    proxyuserpwd = proxyuserpwd,
    proxyauth = auth_flags(auth)
  )
}

auth_flags <- function(x = "basic") {
  constants <- c(
    basic = 1,
    digest = 2,
    gssnegotiate = 4,
    ntlm = 8,
    digest_ie = 16,
    any = -17
  )
  idx <- arg_match0(
    x,
    names(constants),
    arg_nm = "auth",
    error_call = caller_env()
  )
  constants[[idx]]
}

test_that("can add and remove options", {
  req <- request("http://example.com")
  req <- req |> req_options(x = 1)
  expect_equal(req$options, list(x = 1))
  req <- req |> req_options(x = NULL)
  expect_equal(req$options, list())
})

test_that("can add header called req", {
  req <- request("http://example.com")
  req <- req |> req_options(req = 1)
  expect_equal(req$options, list(req = 1))
})

test_that("default user agent includes httr2 + libcurl versions", {
  req <- request("http://example.com") |> req_user_agent()

  expect_match(req$options$useragent, "httr2")
  expect_match(req$options$useragent, "libcurl")
})

test_that("can override default user agent", {
  req <- request("http://example.com") |> req_user_agent("abc")
  expect_equal(req$options$useragent, "abc")
})

test_that("default user agent works with dev curl", {
  # non-R-ish library version for curl, #416
  local_mocked_bindings(curl_system_version = function(...) "8.4.0-DEV")

  expect_match(default_user_agent(), "libcurl/8.4.0-DEV")
})

test_that("can set timeout", {
  req <- request_test("/delay/:secs", secs = 1) |> req_timeout(0.1)
  expect_error(req_perform(req), "timed out")
})

test_that("validates inputs", {
  expect_snapshot(error = TRUE, {
    request_test() |> req_timeout("x")
    request_test() |> req_timeout(0)
  })
})

test_that("req_proxy gives helpful errors", {
  req <- request_test("/get")
  expect_snapshot(error = TRUE, {
    req |> req_proxy(port = "abc")
    req |> req_proxy("abc", auth = "bsc")
  })
})

test_that("auth_flags gives correct constant", {
  expect_equal(auth_flags("digest"), 2)
  expect_equal(auth_flags("ntlm"), 8)
  expect_equal(auth_flags("any"), -17)
})

testthat::set_state_inspector(function() {
  list(
    connections = getAllConnections(),
    the_throttle = as.list(the$throttle)
  )
})

expect_redacted <- function(req, expected) {
  expect_equal(which_redacted(req$headers), expected)
}

# The default user agent includes httr2/curl/libcurl versions, which vary, so
# mock it to a fixed value. Mock req_user_agent() rather than
# default_user_agent() since the latter is cached in an internal environment.
local_mocked_user_agent <- function(env = caller_env()) {
  local_mocked_bindings(
    req_user_agent = function(req, string = NULL) {
      req_options(req, useragent = string %||% "httr2")
    },
    .env = env
  )
}

#' Find the request responsible for a response
#'
#' To make debugging easier, httr2 includes the request that was used to
#' generate every response. You can use this function to access it.
#'
#' @inheritParams resp_header
#' @export
#' @examples
#' req <- request(example_url())
#' resp <- req_perform(req)
#' resp_request(resp)
resp_request <- function(resp) {
  check_response(resp)

  resp$request
}

redacted_sentinel <- function() {
  structure(list(NULL), class = "httr2_redacted_sentinel")
}
#' @export
print.httr2_redacted_sentinel <- function(x, ...) {
  cat(format(x), "\n", sep = "")
  invisible(x)
}
#' @export
format.httr2_redacted_sentinel <- function(x, ...) {
  unclass(cli::col_grey("<REDACTED>"))
}
#' @export
str.httr2_redacted_sentinel <- function(object, ...) {
  cat(" ", cli::col_grey("<REDACTED>"), "\n", sep = "")
}

list_redact <- function(x, names) {
  x[names(x) %in% names] <- list(redacted_sentinel())
  x
}

is_redacted_sentinel <- function(x) {
  inherits(x, "httr2_redacted_sentinel")
}

test_that("req_dry_run() returns useful data", {
  resp <- request("http://example.com") |>
    req_dry_run(quiet = TRUE, testing_headers = FALSE)
  expect_equal(resp$method, "GET")
  expect_equal(resp$path, "/")
  expect_match(resp$headers$`user-agent`, "libcurl")
})

test_that("body is shown", {
  req <- request("http://example.com")

  # can display UTF-8 characters
  req_utf8 <- req_body_raw(req, "Cenário", type = "text/plain")
  expect_snapshot(req_dry_run(req_utf8))

  # json is prettified by default
  req_json <- req_body_raw(req, '{"x":1,"y":true}', type = "application/json")
  expect_snapshot(req_dry_run(req_json))
  expect_snapshot(req_dry_run(req_json, pretty_json = FALSE))

  # doesn't show binary data
  req_binary <- req_body_raw(req, charToRaw("Cenário"))
  expect_snapshot(req_dry_run(req_binary))
})

test_that("authorization headers are redacted", {
  req <- request("http://example.com") |> req_auth_basic("user", "password")
  expect_snapshot(out <- req_dry_run(req))
  expect_equal(out$headers$authorization, redacted_sentinel())

  # unless specifically requested
  req <- request("http://example.com") |> req_auth_basic("user", "password")
  expect_snapshot(out <- req_dry_run(req, redact_headers = FALSE))
  expect_equal(out$headers$authorization, "Basic dXNlcjpwYXNzd29yZA==")
})

test_that("doen't add space to urls (#567)", {
  req <- request("https://example.com/test:1:2")
  expect_output(req_dry_run(req), "test:1:2")
})

test_that("sentinel displays nicely", {
  x <- redacted_sentinel()

  expect_snapshot({
    x
    format(x)
    str(x)
  })
})

test_that("can redact named components of a list", {
  x <- list(a = 1, b = 2, c = 3)
  expect_equal(list_redact(x, "a"), list(a = redacted_sentinel(), b = 2, c = 3))
})

#' Perform request asynchronously using the promises package
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This variation on [req_perform()] returns a [promises::promise()] object immediately
#' and then performs the request in the background, returning program control before the request
#' is finished. See the
#' [promises package documentation](https://rstudio.github.io/promises/articles/promises_01_motivation.html)
#' for more details on how to work with the resulting promise object.
#'
#' If using together with [later::with_temp_loop()] or other private event loops,
#' a new curl pool made by [curl::new_pool()] should be created for requests made
#' within the loop to ensure that only these requests are being polled by the loop.
#'
#' Like with [req_perform_parallel()], exercise caution when using this function;
#' it's easy to pummel a server with many simultaneous requests. Also, not all servers
#' can handle more than 1 request at a time, so the responses may still return
#' sequentially.
#'
#' `req_perform_promise()` also has similar limitations to the
#' [req_perform_parallel()] function, it:
#'
#' * Will not retrieve a new OAuth token if it expires after the promised request
#'   is created but before it is actually requested.
#' * Does not perform throttling with [req_throttle()].
#' * Does not attempt retries as described by [req_retry()].
#' * Only consults the cache set by [req_cache()] when the request is promised.
#'
#' @inheritParams req_perform
#' @inheritParams req_perform_parallel
#' @param pool A pool created by [curl::new_pool()].
#' @return a [promises::promise()] object which resolves to a [response] if
#' successful or rejects on the same errors thrown by [req_perform()].
#' @export
#'
#' @examples
#' \dontrun{
#' library(promises)
#' request_base <- request(example_url()) |> req_url_path_append("delay")
#'
#' p <- request_base |> req_url_path_append(2) |> req_perform_promise()
#'
#' # A promise object, not particularly useful on its own
#' p
#'
#' # Use promise chaining functions to access results
#' p %...>%
#'   resp_body_json() %...>%
#'   print()
#'
#'
#' # Can run two requests at the same time
#' p1 <- request_base |> req_url_path_append(2) |> req_perform_promise()
#' p2 <- request_base |> req_url_path_append(1) |> req_perform_promise()
#'
#' p1 %...>%
#'   resp_url_path %...>%
#'   paste0(., " finished") %...>%
#'   print()
#'
#' p2 %...>%
#'   resp_url_path %...>%
#'   paste0(., " finished") %...>%
#'   print()
#'
#' # See the [promises package documentation](https://rstudio.github.io/promises/)
#' # for more information on working with promises
#' }
req_perform_promise <- function(
  req,
  path = NULL,
  pool = NULL,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL)
) {
  check_installed(c("promises", "later"))

  check_request(req)
  check_string(path, allow_null = TRUE)
  verbosity <- verbosity %||% httr2_verbosity()
  mock <- as_mock_function(mock)

  if (missing(pool)) {
    if (!identical(later::current_loop(), later::global_loop())) {
      cli::cli_abort(c(
        "Must supply {.arg pool} when calling {.code later::with_temp_loop()}.",
        i = "Do you need {.code pool = curl::new_pool()}?"
      ))
    }
  } else {
    if (!is.null(pool) && !inherits(pool, "curl_multi")) {
      stop_input_type(pool, "a {curl} pool", allow_null = TRUE)
    }
  }
  # verbosity checked by req_verbosity
  req <- req_verbosity(req, verbosity)

  promises::promise(function(resolve, reject) {
    pooled_req <- pooled_request(
      req = req,
      path = path,
      on_success = function(resp) resolve(resp),
      on_failure = function(error) reject(error),
      on_error = function(error) reject(error),
      mock = mock
    )
    pooled_req$submit(pool)
    ensure_pool_poller(pool)
  })
}

ensure_pool_poller <- function(pool) {
  monitor <- pool_poller_monitor(pool)
  if (monitor$already_going()) {
    return()
  }

  poll_pool <- function(ready) {
    withCallingHandlers(
      {
        status <- curl::multi_run(0, pool = pool)
        if (status$pending > 0) {
          fds <- curl::multi_fdset(pool = pool)
          later::later_fd(
            func = poll_pool,
            readfds = fds$reads,
            writefds = fds$writes,
            exceptfds = fds$exceptions,
            timeout = fds$timeout
          )
        } else {
          monitor$ending()
        }
      },
      error = function(cnd) {
        monitor$ending()
      }
    )
  }

  monitor$starting()
  poll_pool()
}

pool_poller_monitor <- function(pool) {
  pool_address <- obj_address(pool)
  list(
    already_going = function() {
      env_get(the$pool_pollers, pool_address, default = FALSE)
    },
    starting = function() env_poke(the$pool_pollers, pool_address, TRUE),
    ending = function() env_unbind(the$pool_pollers, pool_address)
  )
}

# Standalone file: do not edit by hand
# Source: <https://github.com/r-lib/rlang/blob/main/R/standalone-obj-type.R>
# ----------------------------------------------------------------------
#
# ---
# repo: r-lib/rlang
# file: standalone-obj-type.R
# last-updated: 2023-05-01
# license: https://unlicense.org
# imports: rlang (>= 1.1.0)
# ---
#
# ## Changelog
#
# 2023-05-01:
# - `obj_type_friendly()` now only displays the first class of S3 objects.
#
# 2023-03-30:
# - `stop_input_type()` now handles `I()` input literally in `arg`.
#
# 2022-10-04:
# - `obj_type_friendly(value = TRUE)` now shows numeric scalars
#   literally.
# - `stop_friendly_type()` now takes `show_value`, passed to
#   `obj_type_friendly()` as the `value` argument.
#
# 2022-10-03:
# - Added `allow_na` and `allow_null` arguments.
# - `NULL` is now backticked.
# - Better friendly type for infinities and `NaN`.
#
# 2022-09-16:
# - Unprefixed usage of rlang functions with `rlang::` to
#   avoid onLoad issues when called from rlang (#1482).
#
# 2022-08-11:
# - Prefixed usage of rlang functions with `rlang::`.
#
# 2022-06-22:
# - `friendly_type_of()` is now `obj_type_friendly()`.
# - Added `obj_type_oo()`.
#
# 2021-12-20:
# - Added support for scalar values and empty vectors.
# - Added `stop_input_type()`
#
# 2021-06-30:
# - Added support for missing arguments.
#
# 2021-04-19:
# - Added support for matrices and arrays (#141).
# - Added documentation.
# - Added changelog.
#
# nocov start

#' Return English-friendly type
#' @param x Any R object.
#' @param value Whether to describe the value of `x`. Special values
#'   like `NA` or `""` are always described.
#' @param length Whether to mention the length of vectors and lists.
#' @return A string describing the type. Starts with an indefinite
#'   article, e.g. "an integer vector".
#' @noRd
obj_type_friendly <- function(x, value = TRUE) {
  if (is_missing(x)) {
    return("absent")
  }

  if (is.object(x)) {
    if (inherits(x, "quosure")) {
      type <- "quosure"
    } else {
      type <- class(x)[[1L]]
    }
    return(sprintf("a <%s> object", type))
  }

  if (!is_vector(x)) {
    return(.rlang_as_friendly_type(typeof(x)))
  }

  n_dim <- length(dim(x))

  if (!n_dim) {
    if (!is_list(x) && length(x) == 1) {
      if (is_na(x)) {
        return(switch(
          typeof(x),
          logical = "`NA`",
          integer = "an integer `NA`",
          double =
            if (is.nan(x)) {
              "`NaN`"
            } else {
              "a numeric `NA`"
            },
          complex = "a complex `NA`",
          character = "a character `NA`",
          .rlang_stop_unexpected_typeof(x)
        ))
      }

      show_infinites <- function(x) {
        if (x > 0) {
          "`Inf`"
        } else {
          "`-Inf`"
        }
      }
      str_encode <- function(x, width = 30, ...) {
        if (nchar(x) > width) {
          x <- substr(x, 1, width - 3)
          x <- paste0(x, "...")
        }
        encodeString(x, ...)
      }

      if (value) {
        if (is.numeric(x) && is.infinite(x)) {
          return(show_infinites(x))
        }

        if (is.numeric(x) || is.complex(x)) {
          number <- as.character(round(x, 2))
          what <- if (is.complex(x)) "the complex number" else "the number"
          return(paste(what, number))
        }

        return(switch(
          typeof(x),
          logical = if (x) "`TRUE`" else "`FALSE`",
          character = {
            what <- if (nzchar(x)) "the string" else "the empty string"
            paste(what, str_encode(x, quote = "\""))
          },
          raw = paste("the raw value", as.character(x)),
          .rlang_stop_unexpected_typeof(x)
        ))
      }

      return(switch(
        typeof(x),
        logical = "a logical value",
        integer = "an integer",
        double = if (is.infinite(x)) show_infinites(x) else "a number",
        complex = "a complex number",
        character = if (nzchar(x)) "a string" else "\"\"",
        raw = "a raw value",
        .rlang_stop_unexpected_typeof(x)
      ))
    }

    if (length(x) == 0) {
      return(switch(
        typeof(x),
        logical = "an empty logical vector",
        integer = "an empty integer vector",
        double = "an empty numeric vector",
        complex = "an empty complex vector",
        character = "an empty character vector",
        raw = "an empty raw vector",
        list = "an empty list",
        .rlang_stop_unexpected_typeof(x)
      ))
    }
  }

  vec_type_friendly(x)
}

vec_type_friendly <- function(x, length = FALSE) {
  if (!is_vector(x)) {
    abort("`x` must be a vector.")
  }
  type <- typeof(x)
  n_dim <- length(dim(x))

  add_length <- function(type) {
    if (length && !n_dim) {
      paste0(type, sprintf(" of length %s", length(x)))
    } else {
      type
    }
  }

  if (type == "list") {
    if (n_dim < 2) {
      return(add_length("a list"))
    } else if (is.data.frame(x)) {
      return("a data frame")
    } else if (n_dim == 2) {
      return("a list matrix")
    } else {
      return("a list array")
    }
  }

  type <- switch(
    type,
    logical = "a logical %s",
    integer = "an integer %s",
    numeric = ,
    double = "a double %s",
    complex = "a complex %s",
    character = "a character %s",
    raw = "a raw %s",
    type = paste0("a ", type, " %s")
  )

  if (n_dim < 2) {
    kind <- "vector"
  } else if (n_dim == 2) {
    kind <- "matrix"
  } else {
    kind <- "array"
  }
  out <- sprintf(type, kind)

  if (n_dim >= 2) {
    out
  } else {
    add_length(out)
  }
}

.rlang_as_friendly_type <- function(type) {
  switch(
    type,

    list = "a list",

    NULL = "`NULL`",
    environment = "an environment",
    externalptr = "a pointer",
    weakref = "a weak reference",
    S4 = "an S4 object",

    name = ,
    symbol = "a symbol",
    language = "a call",
    pairlist = "a pairlist node",
    expression = "an expression vector",

    char = "an internal string",
    promise = "an internal promise",
    ... = "an internal dots object",
    any = "an internal `any` object",
    bytecode = "an internal bytecode object",

    primitive = ,
    builtin = ,
    special = "a primitive function",
    closure = "a function",

    type
  )
}

.rlang_stop_unexpected_typeof <- function(x, call = caller_env()) {
  abort(
    sprintf("Unexpected type <%s>.", typeof(x)),
    call = call
  )
}

#' Return OO type
#' @param x Any R object.
#' @return One of `"bare"` (for non-OO objects), `"S3"`, `"S4"`,
#'   `"R6"`, or `"R7"`.
#' @noRd
obj_type_oo <- function(x) {
  if (!is.object(x)) {
    return("bare")
  }

  class <- inherits(x, c("R6", "R7_object"), which = TRUE)

  if (class[[1]]) {
    "R6"
  } else if (class[[2]]) {
    "R7"
  } else if (isS4(x)) {
    "S4"
  } else {
    "S3"
  }
}

#' @param x The object type which does not conform to `what`. Its
#'   `obj_type_friendly()` is taken and mentioned in the error message.
#' @param what The friendly expected type as a string. Can be a
#'   character vector of expected types, in which case the error
#'   message mentions all of them in an "or" enumeration.
#' @param show_value Passed to `value` argument of `obj_type_friendly()`.
#' @param ... Arguments passed to [abort()].
#' @inheritParams args_error_context
#' @noRd
stop_input_type <- function(x,
                            what,
                            ...,
                            allow_na = FALSE,
                            allow_null = FALSE,
                            show_value = TRUE,
                            arg = caller_arg(x),
                            call = caller_env()) {
  # From standalone-cli.R
  cli <- env_get_list(
    nms = c("format_arg", "format_code"),
    last = topenv(),
    default = function(x) sprintf("`%s`", x),
    inherit = TRUE
  )

  if (allow_na) {
    what <- c(what, cli$format_code("NA"))
  }
  if (allow_null) {
    what <- c(what, cli$format_code("NULL"))
  }
  if (length(what)) {
    what <- oxford_comma(what)
  }
  if (inherits(arg, "AsIs")) {
    format_arg <- identity
  } else {
    format_arg <- cli$format_arg
  }

  message <- sprintf(
    "%s must be %s, not %s.",
    format_arg(arg),
    what,
    obj_type_friendly(x, value = show_value)
  )

  abort(message, ..., call = call, arg = arg)
}

oxford_comma <- function(chr, sep = ", ", final = "or") {
  n <- length(chr)

  if (n < 2) {
    return(chr)
  }

  head <- chr[seq_len(n - 1)]
  last <- chr[n]

  head <- paste(head, collapse = sep)

  # Write a or b. But a, b, or c.
  if (n > 2) {
    paste0(head, sep, final, " ", last)
  } else {
    paste0(head, " ", final, " ", last)
  }
}

# nocov end

#' Extract body from response
#'
#' @description
#' * `resp_body_raw()` returns the raw bytes.
#' * `resp_body_string()` returns a UTF-8 string.
#' * `resp_body_json()` returns parsed JSON.
#' * `resp_body_html()` returns parsed HTML.
#' * `resp_body_xml()` returns parsed XML.
#' * `resp_has_body()` returns `TRUE` if the response has a body.
#'
#' `resp_body_json()` and `resp_body_xml()` check that the content-type header
#' is correct; if the server returns an incorrect type you can suppress the
#' check with `check_type = FALSE`. These two functions also cache the parsed
#' object so the second and subsequent calls are low-cost.
#'
#' @inheritParams resp_headers
#' @returns
#' * `resp_body_raw()` returns a raw vector.
#' * `resp_body_string()` returns a string.
#' * `resp_body_json()` returns NULL, an atomic vector, or list.
#' * `resp_body_html()` and `resp_body_xml()` return an `xml2::xml_document`
#' @export
#' @examples
#' resp <- request("https://httr2.r-lib.org") |> req_perform()
#' resp
#'
#' resp |> resp_has_body()
#' resp |> resp_body_raw()
#' resp |> resp_body_string()
#'
#' if (requireNamespace("xml2", quietly = TRUE)) {
#'   resp |> resp_body_html()
#' }
resp_body_raw <- function(resp) {
  check_response(resp)

  if (!resp_has_body(resp)) {
    cli::cli_abort("Can't retrieve empty body.")
  }

  switch(
    resp_body_type(resp),
    disk = readBin(resp$body, "raw", file.size(resp$body)),
    memory = resp$body,
    stream = resp$body$read_all()
  )
}

#' @rdname resp_body_raw
#' @export
resp_has_body <- function(resp) {
  check_response(resp)

  switch(
    resp_body_type(resp),
    disk = file.size(resp$body) > 0,
    memory = length(resp$body) > 0,
    stream = resp$body$is_open()
  )
}

resp_body_type <- function(resp) {
  if (is_path(resp$body)) {
    "disk"
  } else if (inherits(resp$body, "StreamingBody")) {
    "stream"
  } else {
    "memory"
  }
}

#' @param encoding Character encoding of the body text. If not specified,
#'   will use the encoding specified by the content-type, falling back to
#'   UTF-8 with a warning if it cannot be found. The resulting string is
#'   always re-encoded to UTF-8.
#' @rdname resp_body_raw
#' @export
resp_body_string <- function(resp, encoding = NULL) {
  check_response(resp)
  encoding <- encoding %||% resp_encoding(resp)

  body <- resp_body_raw(resp)
  iconv(readBin(body, character()), from = encoding, to = "UTF-8")
}

#' @param check_type Check that response has expected content type? Set to
#'   `FALSE` to suppress the automated check
#' @param simplifyVector Should JSON arrays containing only primitives (i.e.
#'   booleans, numbers, and strings) be caused to atomic vectors?
#' @param ... Other arguments passed on to [jsonlite::fromJSON()] and
#'   [xml2::read_xml()] respectively.
#' @rdname resp_body_raw
#' @export
resp_body_json <- function(
  resp,
  check_type = TRUE,
  simplifyVector = FALSE,
  ...
) {
  check_response(resp)
  check_installed("jsonlite")

  key <- body_cache_key("json", simplifyVector = simplifyVector, ...)
  if (env_has(resp$cache, key)) {
    return(resp$cache[[key]])
  }

  resp_check_content_type(
    resp,
    valid_types = "application/json",
    valid_suffix = "json",
    check_type = check_type
  )

  text <- resp_body_string(resp, "UTF-8")
  resp$cache[[key]] <- jsonlite::fromJSON(
    text,
    simplifyVector = simplifyVector,
    ...
  )
  resp$cache[[key]]
}

#' @rdname resp_body_raw
#' @export
resp_body_html <- function(resp, check_type = TRUE, ...) {
  check_response(resp)
  check_installed("xml2")
  resp_check_content_type(
    resp,
    valid_types = c("text/html", "application/xhtml+xml"),
    check_type = check_type
  )

  body <- resp_body_raw(resp)
  xml2::read_html(body, ...)
}

#' @rdname resp_body_raw
#' @export
resp_body_xml <- function(resp, check_type = TRUE, ...) {
  check_response(resp)
  check_installed("xml2")

  key <- body_cache_key("xml", ...)
  if (env_has(resp$cache, key)) {
    return(resp$cache[[key]])
  }

  resp_check_content_type(
    resp,
    valid_types = c("application/xml", "text/xml"),
    valid_suffix = "xml",
    check_type = check_type
  )

  body <- resp_body_raw(resp)
  resp$cache[[key]] <- xml2::read_xml(body, ...)
  resp$cache[[key]]
}

body_cache_key <- function(prefix, ...) {
  key <- hash(list(...))
  paste0(prefix, "-", substr(key, 1, 10))
}

test_that("read body from disk/memory", {
  resp1 <- request_test("base64/:value", value = "SGk=") |> req_perform()
  expect_true(resp_has_body(resp1))
  expect_equal(resp_body_raw(resp1), charToRaw("Hi"))
  expect_equal(resp_body_string(resp1), "Hi")

  resp2 <- request_test("base64/:value", value = "SGk=") |>
    req_perform(tempfile())
  expect_true(resp_has_body(resp2))
  expect_equal(resp_body_string(resp2), "Hi")
})

test_that("empty body generates error", {
  resp1 <- request_test("HEAD /get") |> req_perform()
  expect_false(resp_has_body(resp1))
  expect_snapshot(resp_body_raw(resp1), error = TRUE)

  resp2 <- request_test("HEAD /get") |> req_perform(tempfile())
  expect_false(resp_has_body(resp2))
  expect_snapshot(resp_body_raw(resp2), error = TRUE)
})

test_that("can retrieve parsed body", {
  resp <- request_test("/json") |> req_perform()
  expect_type(resp_body_json(resp), "list")

  resp <- request_test("/html") |> req_perform()
  expect_s3_class(resp_body_html(resp), "xml_document")

  resp <- request_test("/xml") |> req_perform()
  expect_s3_class(resp_body_xml(resp), "xml_document")
})

test_that("can retrieve parsed body when saved to a file", {
  path <- withr::local_tempfile()
  resp <- request_test("/json") |> req_perform(path)
  expect_type(resp_body_json(resp), "list")

  resp <- request_test("/html") |> req_perform(path)
  expect_s3_class(resp_body_html(resp), "xml_document")

  resp <- request_test("/xml") |> req_perform(path)
  expect_s3_class(resp_body_xml(resp), "xml_document")
})

test_that("resp_body_json stores parsed result", {
  resp <- request_test("/json") |> req_perform()
  json1 <- resp_body_json(resp)
  # check it's saved
  expect_length(resp$cache, 1)

  # check it's not recomputed
  json2 <- resp_body_json(resp)
  expect_true(is_reference(json2, json1))

  # check the arguments matter
  json3 <- resp_body_json(resp, simplifyVector = TRUE)
  expect_false(is_reference(json3, json1))
  expect_length(resp$cache, 2)
})

test_that("resp_body_xml stores parsed result", {
  resp <- request_test("/xml") |> req_perform()
  xml1 <- resp_body_xml(resp)
  # check it's saved
  expect_length(resp$cache, 1)

  # check it's not recomputed
  xml2 <- resp_body_xml(resp)
  expect_true(is_reference(xml2, xml1))
})

test_that("check argument types before caching", {
  expect_snapshot(error = TRUE, {
    resp_body_json(1)
    resp_body_xml(1)
  })
})

test_that("content types are checked", {
  expect_snapshot(error = TRUE, {
    request_test("/xml") |> req_perform() |> resp_body_json()
    request_test("/json") |> req_perform() |> resp_body_xml()
  })

  resp <- request_test("/json") |> req_perform()
  resp$headers$`Content-Type` <- "application/xml"
  expect_error(resp_body_json(resp))
  expect_no_error(resp_body_json(resp, check_type = FALSE))
})

test_that("validates inputs", {
  expect_snapshot(error = TRUE, {
    req_perform_connection(1)
    req_perform_connection(request_test(), 1)
  })
})

test_that("correctly prepares request", {
  req <- request_test("/post") |> req_method("POST")
  expect_no_error(resp <- req_perform_connection(req))
  close(resp)
})

test_that("can read all data from a connection", {
  resp <- request_test("/stream-bytes/2048") |> req_perform_connection()
  withr::defer(close(resp))

  out <- resp_body_raw(resp)
  expect_length(out, 2048)
  expect_false(resp_has_body(resp))
})

test_that("reads body on error", {
  req <- local_app_request(function(req, res) {
    res$set_status(404L)$send_json(list(status = 404), auto_unbox = TRUE)
  })

  expect_error(req_perform_connection(req), class = "httr2_http_404")
  resp <- last_response()
  expect_equal(resp_body_json(resp), list(status = 404))
})

test_that("can retry a transient error", {
  req <- local_app_request(function(req, res) {
    if (res$app$locals$i == 1) {
      res$set_status(429)$set_header("retry-after", 0)$send_json(
        list(status = "waiting"),
        auto_unbox = TRUE
      )
    } else {
      res$send_json(list(status = "done"), auto_unbox = TRUE)
    }
  })
  req <- req_retry(req, max_tries = 2)

  cnd <- expect_condition(
    resp <- req_perform_connection(req),
    class = "httr2_retry"
  )
  expect_s3_class(cnd, "httr2_retry")
  expect_equal(cnd$tries, 1)
  expect_equal(cnd$delay, 0)

  expect_equal(last_response(), resp)
  expect_equal(resp_body_json(resp), list(status = "done"))
})

test_that("curl errors become errors", {
  local_mocked_bindings(open = function(...) abort("Failed to connect"))

  req <- request("http://127.0.0.1")
  expect_snapshot(req_perform_connection(req), error = TRUE)
  expect_error(req_perform_connection(req), class = "httr2_failure")

  # and captures request
  cnd <- catch_cnd(req_perform_connection(req), classes = "error")
  expect_equal(cnd$request, req_policies(req, connection = TRUE))

  # But last_response() is NULL
  expect_null(last_response())
})

test_that("correclty reports curl error with retries (#817)", {
  local_mocked_bindings(open = function(...) abort("Failed to connect"))

  req <- request("http://127.0.0.1") |>
    req_retry(max_tries = 2, retry_on_failure = TRUE, backoff = \(i) 0)

  expect_snapshot(req_perform_connection(req), error = TRUE)
})

test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    expect_equal(req$policies$connection, TRUE)
    if (req$url == "https://ok") {
      conn <- rawConnection(charToRaw("a\nb\n"))
      response(body = StreamingBody$new(conn))
    } else {
      response(404)
    }
  })

  resp <- req_perform_connection(req_200)
  expect_equal(resp_stream_lines(resp), "a")
  expect_equal(resp_stream_lines(resp), "b")
  close(resp)

  expect_error(req_perform_connection(req_404), class = "httr2_http_404")
})


# StreamingBody --------------------------------------------------------------

test_that("validates its input", {
  expect_snapshot(error = TRUE, {
    StreamingBody$new(1)
  })
})

test_that("can access fdset", {
  skip_on_cran()

  con <- curl::curl(example_url('/drip'), open = 'rb')
  on.exit(close(con))

  body <- StreamingBody$new(con)
  expect_length(body$get_fdset()$reads, 1)
})

# otel -----------------------------------------------------------------------

test_that("tracing works as expected", {
  skip_if_not_installed("otelsdk")
  skip_on_os("windows")

  spans <- with_otel_record({
    # A request with no URL (which shouldn't create a span).
    try(req_perform_connection(request("")), silent = TRUE)

    # A regular request.
    resp <- req_perform_connection(request_test("/headers"))

    # Verify that context propagation works as expected.
    expect_true("traceparent" %in% names(resp_body_json(resp)[["headers"]]))
    close(resp)

    # A request with an HTTP error.
    try(
      req_perform_connection(request_test("/status/:status", status = 404)),
      silent = TRUE
    )

    # A request with basic credentials that we should redact.
    parsed <- url_parse(example_url())
    parsed$username <- "user"
    parsed$password <- "secret"
    resp <- req_perform_connection(request(url_build(parsed)))
    close(resp)

    # A request with a curl error.
    with_mocked_bindings(
      try(req_perform_connection(request("http://127.0.0.1")), silent = TRUE),
      curl_fetch = function(...) abort("Failed to connect")
    )

    # A request that triggers retries, generating three individual spans.
    request_test("/status/:status", status = 429) |>
      req_retry(max_tries = 3, backoff = ~0) |>
      req_perform_connection() |>
      try(silent = TRUE)
  })[["traces"]]

  expect_length(spans, 7L)

  # Validate the span for regular requests.
  expect_equal(spans[[1]]$status, "ok")
  expect_named(
    spans[[1]]$attributes,
    c(
      "http.response.status_code",
      "user_agent.original",
      "url.full",
      "server.address",
      "server.port",
      "http.request.method"
    ),
    ignore.order = TRUE
  )
  expect_equal(spans[[1]]$attributes$http.request.method, "GET")
  expect_equal(spans[[1]]$attributes$http.response.status_code, 200L)
  expect_equal(spans[[1]]$attributes$server.address, "127.0.0.1")
  expect_match(spans[[1]]$attributes$user_agent.original, "^httr2/")

  # And for requests with HTTP errors.
  expect_equal(spans[[2]]$status, "error")
  expect_equal(spans[[2]]$description, "Not Found")
  expect_equal(spans[[2]]$attributes$http.response.status_code, 404L)
  expect_equal(spans[[2]]$attributes$error.type, "404")

  # And for spans with redacted credentials.
  expect_match(
    spans[[3]]$attributes$url.full,
    regexp = "http://REDACTED:REDACTED@127.0.0.1:[0-9]+/",
  )

  # And for spans with curl errors.
  expect_equal(spans[[4]]$status, "error")
  expect_equal(spans[[4]]$attributes$error.type, "simpleError")

  # We should have attached the curl error as an event.
  expect_length(spans[[4]]$events, 1L)
  expect_equal(spans[[4]]$events[[1]]$name, "exception")

  # For spans with retries, we expect the parent context to be the same for
  # each span. It is important that they not be children of one another.
  expect_equal(spans[[5]]$parent, spans[[6]]$parent)
  expect_equal(spans[[6]]$parent, spans[[7]]$parent)
})

# Helpers for building AWS event-stream messages in tests.

# Decode a hex string to raw. Only used to express opaque reference vectors
# captured from another implementation; build events with aws_event() instead.
hex_to_raw <- function(x) {
  x <- gsub("(\\s|\n)+", "", x)
  pairs <- substring(x, seq(1, nchar(x), by = 2), seq(2, nchar(x), by = 2))
  as.raw(strtoi(pairs, 16L))
}

# A big-endian integer in `size` raw bytes. Handles values beyond
# .Machine$integer.max (unlike writeBin()) and encodes negatives as two's
# complement, so it serves both unsigned lengths and signed header values.
aws_uint <- function(x, size) {
  x <- as.numeric(x)
  out <- raw(size)
  for (i in seq_len(size)) {
    out[[size - i + 1L]] <- as.raw(x %% 256)
    x <- x %/% 256
  }
  out
}

# CRC32 of `bytes` as 4 big-endian raw bytes, matching the AWS event-stream
# framing.
aws_crc <- function(bytes) {
  hex <- digest::digest(bytes, algo = "crc32", serialize = FALSE)
  as.raw(strtoi(substring(hex, c(1L, 3L, 5L, 7L), c(2L, 4L, 6L, 8L)), 16L))
}

# A single header: a `name`, a `type` (matching the AWS event-stream spec), and
# a `value` to encode.
aws_header <- function(name, type, value = NULL) {
  body <- switch(
    type,
    true = list(tag = 0L, bytes = raw()),
    false = list(tag = 1L, bytes = raw()),
    byte = list(tag = 2L, bytes = aws_uint(value, 1L)),
    short = list(tag = 3L, bytes = aws_uint(value, 2L)),
    integer = list(tag = 4L, bytes = aws_uint(value, 4L)),
    long = list(tag = 5L, bytes = aws_uint(value, 8L)),
    bytes = list(tag = 6L, bytes = c(aws_uint(length(value), 2L), value)),
    string = list(
      tag = 7L,
      bytes = c(aws_uint(nchar(value), 2L), charToRaw(value))
    ),
    timestamp = list(tag = 8L, bytes = aws_uint(value, 8L)),
    uuid = list(tag = 9L, bytes = value),
    unknown = list(tag = 255L, bytes = raw()),
    cli::cli_abort("Unknown header type {.val {type}}.")
  )
  c(aws_uint(nchar(name), 1L), charToRaw(name), as.raw(body$tag), body$bytes)
}

# A complete event wrapping raw `headers` and a raw `body` in the AWS
# event-stream framing, computing the prelude/header lengths and both CRCs.
aws_event <- function(headers = raw(), body = raw()) {
  total <- 12L + length(headers) + length(body) + 4L
  prelude <- c(aws_uint(total, 4L), aws_uint(length(headers), 4L))
  prelude <- c(prelude, aws_crc(prelude)) # prelude CRC over the first 8 bytes
  message <- c(prelude, headers, body)
  c(message, aws_crc(message)) # message CRC over everything before it
}

test_that("can parse empty object", {
  expect_equal(
    parse_aws_event(aws_event()),
    list(headers = list(), body = "")
  )
})


test_that("can return various types of header", {
  headers <- function(...) parse_aws_event(aws_event(aws_header(...)))$headers

  expect_equal(headers("foo", "false"), list(foo = FALSE))
  expect_equal(headers("foo", "true"), list(foo = TRUE))
  expect_equal(headers("foo", "bytes", as.raw(1:5)), list(foo = as.raw(1:5)))
  expect_equal(headers("foo", "string", "bar"), list(foo = "bar"))

  # byte, short, and integer are signed (two's complement)
  expect_equal(headers("foo", "byte", 127), list(foo = 127))
  expect_equal(headers("foo", "byte", -1), list(foo = -1))
  expect_equal(headers("foo", "short", -2), list(foo = -2))
  expect_equal(headers("foo", "integer", -3), list(foo = -3))

  # long and timestamp are 64-bit integers, returned as bit64::integer64 (see
  # the reference test below for a non-trivial long value)
  expect_equal(
    headers("foo", "timestamp", 0),
    list(foo = structure(0, class = "integer64"))
  )

  # UUID is returned as a hex string
  uuid <- as.raw(1:16)
  expect_equal(headers("foo", "uuid", uuid), list(foo = raw_to_hex(uuid)))
})


test_that("unknown header triggers error", {
  expect_snapshot(
    parse_aws_event(aws_event(aws_header("foo", "unknown"))),
    error = TRUE
  )
})

test_that("parse_aws_event() checks the prelude length", {
  expect_snapshot(parse_aws_event(as.raw(1:10)), error = TRUE)
})

test_that("can read aws events one at a time", {
  # Two empty-object events back to back.
  event <- aws_event()
  req <- local_app_request(function(req, res) {
    res$send_chunk(event)
    res$send_chunk(event)
  })
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  expect_equal(resp_stream_aws(resp), list(headers = list(), body = ""))
  expect_equal(resp_stream_aws(resp), list(headers = list(), body = ""))
  expect_equal(resp_stream_aws(resp), NULL)
})

test_that("max_size counts the buffered event bytes, approximately", {
  data <- aws_event()
  resp1 <- local_streaming_response(data)

  expect_equal(
    resp_stream_aws(resp1, max_size = length(data)),
    list(headers = list(), body = "")
  )

  resp2 <- local_streaming_response(data)
  expect_error(
    resp_stream_aws(resp2, max_size = length(data) - 2L),
    class = "httr2_streaming_error"
  )
})

test_that("verbosity = 3 shows aws events", {
  # A single event with a "foo: bar" header and empty body.
  event <- aws_event(aws_header("foo", "string", "bar"))
  req <- local_app_request(function(req, res) {
    res$send_chunk(event)
  })

  expect_output(resp <- req_perform_connection(req, verbosity = 3))
  withr::defer(close(resp))
  expect_snapshot(
    . <- resp_stream_aws(resp),
    transform = transform_verbose_response
  )
})

test_that("find_aws_event_boundaries splits a buffer into complete events", {
  event <- aws_event()

  expect_equal(find_aws_event_boundaries(event), 17)
  expect_equal(find_aws_event_boundaries(c(event, event)), c(17, 33))
  expect_equal(
    find_aws_event_boundaries(c(event, event, event)),
    c(17, 33, 49)
  )
})

test_that("find_aws_event_boundaries ignores incomplete trailing events", {
  event <- aws_event()

  # Nothing to split
  expect_equal(find_aws_event_boundaries(raw()), double())
  # Fewer than 16 bytes can't be a complete event
  expect_equal(find_aws_event_boundaries(event[1:10]), double())
  # Trailing partial event is excluded
  expect_equal(find_aws_event_boundaries(c(event, event[1:8])), 17)
  # Event claiming more bytes than are available is excluded
  big <- event
  big[1:4] <- aws_uint(256, 4L)
  expect_equal(find_aws_event_boundaries(c(event, big)), 17)
})

test_that("json content type automatically parsed", {
  event <- aws_event(
    c(
      aws_header(":event-type", "string", "SubscribeToShardEvent"),
      aws_header(":content-type", "string", "application/json"),
      aws_header(":message-type", "string", "event")
    ),
    body = charToRaw('{"records": []}')
  )
  parsed <- parse_aws_event(event)
  expect_equal(parsed$body, list(records = list()))
})

# aws_event() ------------------------------------------------------------------

test_that("aws_event() produces spec-valid bytes, including CRCs", {
  # Using values from a reference implementation at
  # https://github.com/lifion/lifion-aws-event-stream
  expect_equal(aws_event(), hex_to_raw("000000100000000005c248eb7d98c8ff"))
})

test_that("aws_event() agrees with the reference implementation", {
  # Messages captured from the lifion JS reference implementation:
  # https://github.com/lifion/lifion-aws-event-stream. These vectors encode a
  # non-spec header length (always 1), so their bytes differ from aws_event()'s,
  # but a reference message and the equivalent aws_event() must decode the same.
  agrees <- function(reference, header) {
    expect_equal(
      parse_aws_event(hex_to_raw(reference)),
      parse_aws_event(aws_event(header))
    )
  }

  agrees(
    "0000001500000001ba25f70d03666f6f013aa3e0d6",
    aws_header("foo", "false")
  )
  agrees(
    "0000001500000001ba25f70d03666f6f004da4d040",
    aws_header("foo", "true")
  )
  agrees(
    "0000001600000001fd858ddd03666f6f02ffa44bfd93",
    aws_header("foo", "byte", -1) # 0xff
  )
  agrees(
    "0000001700000001c0e5a46d03666f6f03fffff3b59291",
    aws_header("foo", "short", -1) # 0xffff
  )
  agrees(
    "00000019000000017fd51a0c03666f6f04ffffffff853b65dd",
    aws_header("foo", "integer", -1) # 0xffffffff
  )
  agrees(
    "0000001d000000018a55bccc03666f6f050000ffffffffffff6b03c255",
    aws_header("foo", "long", 281474976710655) # 0x0000ffffffffffff
  )
  agrees(
    "0000001a00000001387560dc03666f6f0700036261725bb3cecf",
    aws_header("foo", "string", "bar")
  )
  agrees(
    "0000001c00000001b735957c03666f6f0600050102030405cdda4038",
    aws_header("foo", "bytes", as.raw(1:5))
  )
  agrees(
    "00000025000000011b044f8b03666f6f093bfdac5cfe6c402983bfc1de7819f5316056148a",
    aws_header("foo", "uuid", hex_to_raw("3bfdac5cfe6c402983bfc1de7819f531"))
  )
})

#' Extract timing data
#'
#' The underlying curl library measures how long different components of the
#' request take to complete. This function retrieves that information.
#'
#' @inheritParams resp_header
#' @returns Named numeric vector of timing information.
#'  The names of the elements in this vector correspond to the names used
#'  in [libcurl's `curl_easy_getinfo()` API][curl docs].
#'  The most useful component is likely `"total"` (corresponding to
#'  `CURLINFO_TOTAL_TIME`), the overall time in seconds to complete the
#'  request including any redirects followed.
#'
#'  [curl docs]: https://curl.se/libcurl/c/curl_easy_getinfo.html
#' @export
#' @examples
#' req <- request(example_url())
#' resp <- req_perform(req)
#' resp_timing(resp)
resp_timing <- function(resp) {
  check_response(resp)

  resp$timing
}

#' Control handling of HTTP errors
#'
#' `req_perform()` will automatically convert HTTP errors (i.e. any 4xx or 5xx
#' status code) into R errors. Use `req_error()` to either override the
#' defaults, or extract additional information from the response that would
#' be useful to expose to the user.
#'
#' # Error handling
#'
#' `req_perform()` is designed to succeed if and only if you get a valid HTTP
#' response. There are two ways a request can fail:
#'
#' * The HTTP request might fail, for example if the connection is dropped
#'   or the server doesn't exist. This type of error will have class
#'   `c("httr2_failure", "httr2_error")`.
#'
#' * The HTTP request might succeed, but return an HTTP status code that
#'   represents an error, e.g. a `404 Not Found` if the specified resource is
#'   not found. This type of error will have (e.g.) class
#'   `c("httr2_http_404", "httr2_http", "httr2_error")`.
#'
#' These error classes are designed to be used in conjunction with R's
#' condition handling tools (<https://adv-r.hadley.nz/conditions.html>).
#' For example, if you want to return a default value when the server returns
#' a 404, use `tryCatch()`:
#'
#' ```
#' tryCatch(
#'   req |> req_perform() |> resp_body_json(),
#'   httr2_http_404 = function(cnd) NULL
#' )
#' ```
#'
#' Or if you want to re-throw the error with some additional context, use
#' `withCallingHandlers()`, e.g.:
#'
#' ```R
#' withCallingHandlers(
#'   req |> req_perform() |> resp_body_json(),
#'   httr2_http_404 = function(cnd) {
#'     rlang::abort("Couldn't find user", parent = cnd)
#'   }
#' )
#' ```
#'
#' Learn more about error chaining at [rlang::topic-error-chaining].
#'
#' @seealso [req_retry()] to control when errors are automatically retried.
#' @inheritParams req_perform
#' @param is_error A predicate function that takes a single argument (the
#'   response) and returns `TRUE` or `FALSE` indicating whether or not an
#'   R error should be signalled.
#' @param body A callback function that takes a single argument (the response)
#'   and returns a character vector of additional information to include in the
#'   body of the error. This vector is passed along to the `message` argument
#'   of [rlang::abort()] so you can use any formatting that it supports.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # Performing this request usually generates an error because httr2
#' # converts HTTP errors into R errors:
#' req <- request(example_url()) |>
#'   req_url_path("/status/404")
#' try(req |> req_perform())
#' # You can still retrieve it with last_response()
#' last_response()
#'
#' # But you might want to suppress this behaviour:
#' resp <- req |>
#'   req_error(is_error = \(resp) FALSE) |>
#'   req_perform()
#' resp
#'
#' # Or perhaps you're working with a server that routinely uses the
#' # wrong HTTP error codes only 500s are really errors
#' request("http://example.com") |>
#'   req_error(is_error = \(resp) resp_status(resp) == 500)
#'
#' # Most typically you'll use req_error() to add additional information
#' # extracted from the response body (or sometimes header):
#' error_body <- function(resp) {
#'   resp_body_json(resp)$error
#' }
#' request("http://example.com") |>
#'   req_error(body = error_body)
#' # Learn more in https://httr2.r-lib.org/articles/wrapping-apis.html
req_error <- function(req, is_error = NULL, body = NULL) {
  check_request(req)

  req_policies(
    req,
    error_is_error = as_callback(is_error, 1, "is_error"),
    error_body = as_callback(body, 1, "body")
  )
}

error_is_error <- function(req, resp) {
  req_policy_call(req, "error_is_error", list(resp), default = resp_is_error)
}

error_body <- function(req, resp, call = caller_env()) {
  try_fetch(
    req_policy_call(req, "error_body", list(resp), default = NULL),
    error = function(cnd) {
      cli::cli_abort(
        "Failed to parse error body with method defined in {.fn req_error}.",
        parent = cnd,
        call = call
      )
    }
  )
}

capture_curl_error <- function(code, call = caller_env()) {
  tryCatch(
    {
      code
      NULL
    },
    error = function(err) curl_cnd(err, call = call)
  )
}

curl_cnd <- function(err, call = caller_env()) {
  error_cnd(
    message = "Failed to perform HTTP request.",
    class = c("httr2_failure", "httr2_error"),
    parent = err,
    call = call
  )
}

test_that("can customise what statuses are errors", {
  req <- request_test()
  expect_equal(error_is_error(req, response(404)), TRUE)
  expect_equal(error_is_error(req, response(200)), FALSE)

  req <- req |> req_error(is_error = \(resp) !resp_is_error(resp))
  expect_equal(error_is_error(req, response(404)), FALSE)
  expect_equal(error_is_error(req, response(200)), TRUE)
})

test_that("can customise error info", {
  req <- request_test("/404")
  expect_equal(error_body(req, response(404)), NULL)

  req <- req |> req_error(body = \(resp) "Hi!")
  expect_equal(error_body(req, response(404)), "Hi!")

  expect_snapshot(req_perform(req), error = TRUE)
})

test_that("long custom body is wrapped", {
  withr::local_options(cli.condition_width = 60)

  body <- paste0(
    "Ad aliquip et occaecat consequat eiusmod enim Lorem incididunt laboris ",
    "deserunt. Consectetur magna ea ad quis dolore. Deserunt elit elit dolore ",
    "magna fugiat ipsum id eu nostrud voluptate Lorem ad id anim. Cupidatat ",
    "nulla ipsum irure nisi sunt ipsum commodo eu sint eiusmod consectetur."
  )
  req <- request("http://google.com") |>
    req_error(is_error = \(resp) TRUE, body = \(resp) body)
  expect_snapshot(req_perform(req), error = TRUE)
})

test_that("failing callback still generates useful body", {
  req <- request_test() |> req_error(body = \(resp) abort("This is an error!"))
  expect_snapshot_error(error_body(req, response(404)))

  out <- expect_snapshot(error = TRUE, {
    req <- request_test("/status/404")
    req <- req |> req_error(body = \(resp) resp_body_json(resp)$error)
    req |> req_perform()
  })
})

#' Automatically cache requests
#'
#' @description
#' Use `req_cache()` to automatically cache HTTP requests. Most API requests
#' are not cacheable, but static files often are.
#'
#' `req_cache()` caches responses to GET requests that have status code 200 and
#' at least one of the standard caching headers (e.g. `Expires`,
#' `Etag`, `Last-Modified`, `Cache-Control`), unless caching has been expressly
#' prohibited with `Cache-Control: no-store`. Typically, a request will still
#' be sent to the server to check that the cached value is still up-to-date,
#' but it will not need to re-download the body value.
#'
#' To learn more about HTTP caching, I recommend the MDN article
#' [HTTP caching](https://developer.mozilla.org/en-US/docs/Web/HTTP/Caching).
#'
#' @inheritParams req_perform
#' @param path Path to cache directory. Will be created automatically if it
#'   does not exist.
#'
#'   For quick and easy caching within a session, you can use `tempfile()`.
#'   To cache requests within a package, you can use something like
#'   `file.path(tools::R_user_dir("pkgdown", "cache"), "httr2")`.
#'
#'   httr2 doesn't provide helpers to manage the cache, but if you want to
#'   empty it, you can use something like
#'   `unlink(dir(cache_path, full.names = TRUE))`.
#' @param use_on_error If the request errors, and there's a cache response,
#'   should `req_perform()` return that instead of generating an error?
#' @param debug When `TRUE` will emit useful messages telling you about
#'   cache hits and misses. This can be helpful to understand whether or
#'   not caching is actually doing anything for your use case.
#' @param max_n,max_age,max_size Automatically prune the cache by specifying
#'   one or more of:
#'
#'   * `max_age`: to delete files older than this number of seconds.
#'   * `max_n`: to delete files (from oldest to newest) to preserve at
#'      most this many files.
#'   * `max_size`: to delete files (from oldest to newest) to preserve at
#'      most this many bytes.
#'
#'   The cache pruning is performed at most once per minute.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # GitHub uses HTTP caching for all raw files.
#' url <- paste0(
#'   "https://raw.githubusercontent.com/allisonhorst/palmerpenguins/",
#'   "master/inst/extdata/penguins.csv"
#' )
#' # Here I set debug = TRUE so you can see what's happening
#' req <- request(url) |> req_cache(tempdir(), debug = TRUE)
#'
#' # First request downloads the data
#' resp <- req |> req_perform()
#'
#' # Second request retrieves it from the cache
#' resp <- req |> req_perform()
req_cache <- function(
  req,
  path,
  use_on_error = FALSE,
  debug = getOption("httr2_cache_debug", FALSE),
  max_age = Inf,
  max_n = Inf,
  max_size = 1024^3
) {
  check_number_whole(max_age, min = 0, allow_infinite = TRUE)
  check_number_whole(max_n, min = 0, allow_infinite = TRUE)
  check_number_decimal(max_size, min = 1, allow_infinite = TRUE)

  dir.create(path, showWarnings = FALSE, recursive = TRUE)
  req_policies(
    req,
    cache_path = path,
    cache_use_on_error = use_on_error,
    cache_debug = debug,
    cache_max = list(age = max_age, n = max_n, size = max_size)
  )
}

# Do I need to worry about hash collisions?
# No - even if the user stores a billion urls, the probably of a collision
# is ~ 1e-20: https://preshing.com/20110504/hash-collision-probabilities/
req_cache_path <- function(req, ext = ".rds") {
  file.path(req$policies$cache_path, paste0(hash(req$url), ext))
}
cache_use_on_error <- function(req) {
  req$policies$cache_use_on_error %||% FALSE
}
cache_debug <- function(req) {
  req$policies$cache_debug %||% FALSE
}

# Cache management --------------------------------------------------------

cache_active <- function(req) {
  req_policy_exists(req, "cache_path")
}

cache_get <- function(req) {
  # This check should be redudant but we keep it in for safety
  if (!cache_active(req)) {
    return(req)
  }

  path <- req_cache_path(req)
  if (!file.exists(path)) {
    return(NULL)
  }

  tryCatch(
    {
      rds <- readRDS(path)
      # Update file time if read successfully
      Sys.setFileTime(path, Sys.time())
      rds
    },
    error = function(e) NULL
  )
}

cache_set <- function(req, resp) {
  signal("", "httr2_cache_save")

  if (resp_body_type(resp) == "disk") {
    body_path <- req_cache_path(req, ".body")
    file.copy(resp$body, body_path, overwrite = TRUE)
    resp$body <- new_path(body_path)
  }

  saveRDS(resp, req_cache_path(req, ".rds"))
  invisible()
}

cache_prune_if_needed <- function(req, threshold = 60, debug = FALSE) {
  path <- req$policies$cache_path

  last_prune <- the$cache_throttle[[path]]
  if (is.null(last_prune) || last_prune + threshold <= Sys.time()) {
    if (debug) {
      cli::cli_text("Pruning cache")
    }
    cache_prune(path, max = req$policies$cache_max, debug = debug)
    the$cache_throttle[[path]] <- Sys.time()

    invisible(TRUE)
  } else {
    invisible(FALSE)
  }
}

# Adapted from
# https://github.com/r-lib/cachem/blob/main/R/cache-disk.R#L396-L467
cache_prune <- function(path, max, debug = TRUE) {
  info <- cache_info(path)

  info <- cache_prune_files(
    info,
    info$mtime + max$age < Sys.time(),
    "too old",
    debug
  )
  info <- cache_prune_files(
    info,
    seq_len(nrow(info)) > max$n,
    "too numerous",
    debug
  )
  info <- cache_prune_files(
    info,
    cumsum(info$size) > max$size,
    "too big",
    debug
  )

  invisible()
}

cache_info <- function(path, pattern = "\\.rds$") {
  filenames <- dir(path, pattern, full.names = TRUE)
  info <- file.info(filenames, extra_cols = FALSE)
  info <- info[info$isdir == FALSE, ]
  info$name <- rownames(info)
  rownames(info) <- NULL
  info[order(info$mtime, decreasing = TRUE), c("name", "size", "mtime")]
}

cache_prune_files <- function(info, to_remove, why, debug = TRUE) {
  if (any(to_remove)) {
    if (debug) {
      cli::cli_text("Deleted {sum(to_remove)} file{?s} that {?is/are} {why}")
    }

    file.remove(info$name[to_remove])
    info[!to_remove, ]
  } else {
    info
  }
}

# Hooks for req_perform -----------------------------------------------------

# Can return request or response
cache_pre_fetch <- function(req, path = NULL) {
  if (!cache_active(req)) {
    return(req)
  }

  # Only GET requests should be retrieved from cache. It's not sufficient to
  # only save GET requests, because the method is not part of the cache key
  if (req_get_method(req) != "GET") {
    return(req)
  }

  debug <- cache_debug(req)
  cache_prune_if_needed(req, debug = debug)

  cached_resp <- cache_get(req)
  if (is.null(cached_resp)) {
    return(req)
  }
  if (debug) {
    cli::cli_text("Found url in cache {.val {hash(req$url)}}")
  }

  info <- resp_cache_info(cached_resp)
  if (!is.na(info$expires) && info$expires >= Sys.time()) {
    signal("", "httr2_cache_cached")
    if (debug) {
      cli::cli_text("Cached value is fresh; using response from cache")
    }

    resp <- cached_resp
    resp$body <- cache_body(cached_resp, path)
    resp
  } else {
    if (debug) {
      cli::cli_text("Cached value is stale; checking for updates")
    }
    req_headers(
      req,
      `If-Modified-Since` = info$last_modified,
      `If-None-Match` = info$etag
    )
  }
}

# Always returns response
cache_post_fetch <- function(req, resp, path = NULL) {
  if (!cache_active(req)) {
    return(resp)
  }

  debug <- cache_debug(req)
  cached_resp <- cache_get(req)

  if (is_error(resp)) {
    if (cache_use_on_error(req) && !is.null(cached_resp)) {
      if (debug) {
        cli::cli_text("Request errored; retrieving response from cache")
      }
      cached_resp
    } else {
      resp
    }
  } else if (resp_status(resp) == 304 && !is.null(cached_resp)) {
    signal("", "httr2_cache_not_modified")
    if (debug) {
      cli::cli_text("Cached value still ok; retrieving body from cache")
    }

    # Combine headers
    resp$headers <- cache_headers(cached_resp, resp)
    # Replace body with cached result
    resp$body <- cache_body(cached_resp, path)

    # Re-cache, so we get any new headers
    cache_set(req, resp)
    resp
  } else if (resp_is_cacheable(resp)) {
    if (debug) {
      cli::cli_text("Saving response to cache {.val {hash(req$url)}}")
    }

    cache_set(req, resp)
    resp
  } else {
    resp
  }
}

cache_body <- function(cached_resp, path = NULL) {
  check_response(cached_resp)

  body <- cached_resp$body
  if (is.null(path)) {
    return(body)
  }

  switch(
    resp_body_type(cached_resp),
    disk = file.copy(body, path, overwrite = TRUE),
    memory = writeBin(body, path),
    stream = cli::cli_abort("Invalid body type", .internal = TRUE)
  )

  new_path(path)
}

# https://www.rfc-editor.org/rfc/rfc7232#section-4.1
cache_headers <- function(cached_resp, resp) {
  check_response(cached_resp)

  headers <- modify_list(
    cached_resp$headers,
    !!!resp$headers,
    .ignore_case = TRUE
  )
  as_headers(headers)
}

# Caching headers ---------------------------------------------------------

resp_is_cacheable <- function(resp, control = NULL) {
  if (resp$method != "GET") {
    return(FALSE)
  }

  if (resp_status(resp) != 200L) {
    return(FALSE)
  }

  if (resp_body_type(resp) == "stream") {
    return(FALSE)
  }

  control <- control %||% resp_cache_control(resp)
  if ("no-store" %in% control$flags) {
    return(FALSE)
  }
  if (has_name(control, "max-age")) {
    return(TRUE)
  }

  if (!any(resp_header_exists(resp, c("Etag", "Last-Modified", "Expires")))) {
    return(FALSE)
  }

  TRUE
}

resp_cache_info <- function(resp, control = NULL) {
  list(
    expires = resp_cache_expires(resp, control),
    last_modified = resp_header(resp, "Last-Modified"),
    etag = resp_header(resp, "Etag")
  )
}

resp_cache_expires <- function(resp, control = NULL) {
  control <- control %||% resp_cache_control(resp)

  # Prefer max-age parameter if it exists, otherwise use Expires header
  if (has_name(control, "max-age") && resp_header_exists(resp, "Date")) {
    resp_date(resp) + as.integer(control[["max-age"]])
  } else if (resp_header_exists(resp, "Expires")) {
    parse_http_date(resp_header(resp, "Expires"))
  } else {
    NA
  }
}

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
resp_cache_control <- function(resp) {
  x <- resp_header(resp, "Cache-Control")
  if (is.null(x)) {
    return(NULL)
  }

  pieces <- strsplit(x, ",", fixed = TRUE)[[1]]
  pieces <- gsub("^\\s+|\\s+$", "", pieces)
  pieces <- tolower(pieces)

  is_value <- grepl("=", pieces, fixed = TRUE)
  flags <- pieces[!is_value]

  keyvalues <- strsplit(pieces[is_value], "\\s*=\\s*")
  keys <- c(rep("flags", length(flags)), lapply(keyvalues, "[[", 1))
  values <- c(flags, lapply(keyvalues, "[[", 2))

  stats::setNames(values, keys)
}

test_that("successful request returns expected response", {
  req <- request_test()
  resp <- req_perform(req)

  expect_s3_class(resp, "httr2_response")
  expect_equal(resp$method, "GET")
  expect_equal(resp$url, example_url("/get"))
  expect_equal(resp$status_code, 200)
  expect_s3_class(resp$headers, "httr2_headers")
  expect_type(resp$body, "raw")
  expect_equal(resp$request, req)

  expect_type(resp$timing, "double")
  expect_true(all(resp$timing >= 0))
})

test_that("request updates last_response()", {
  req200 <- request_test()
  req404 <- request_test("/404")

  resp <- req_perform(req200)
  expect_equal(last_response(), resp)
  expect_equal(last_request(), req200)

  # even if it errors
  try(req_perform(req404), silent = TRUE)
  expect_equal(last_response()$status_code, 404)
  expect_equal(last_request(), req404)
})

test_that("curl errors become errors", {
  local_mocked_bindings(curl_fetch = function(...) abort("Failed to connect"))

  req <- request("http://127.0.0.1")
  expect_snapshot(req_perform(req), error = TRUE)
  expect_error(req_perform(req), class = "httr2_failure")

  # and captures request
  cnd <- catch_cnd(req_perform(req), classes = "error")
  expect_equal(cnd$request, req)

  # But last_response() is NULL
  expect_null(last_response())
})

test_that("http errors become errors", {
  req <- request_test("/status/:status", status = 404)
  expect_error(req_perform(req), class = "httr2_http_404")
  expect_snapshot(req_perform(req), error = TRUE)

  # and captures request
  cnd <- catch_cnd(req_perform(req), classes = "error")
  expect_equal(cnd$request, req)

  # including transient errors
  req <- request_test("/status/:status", status = 429)
  expect_snapshot(req_perform(req), error = TRUE)

  req_perform(req) |>
    expect_error(class = "httr2_http_429") |>
    expect_no_condition(class = "httr2_sleep")

  # non-standard status codes don't get descriptions
  req <- request_test("/status/:status", status = 599)
  expect_snapshot(req_perform(req), error = TRUE)
})

test_that("can force successful HTTP statuses to error", {
  req <- request_test("/status/:status", status = 200) |>
    req_error(is_error = function(resp) TRUE)

  expect_error(req_perform(req), class = "httr2_http_200")
})

test_that("persistent HTTP errors only get single attempt", {
  req <- request_test("/status/:status", status = 404) |>
    req_retry(max_tries = 5)

  cnd <- req_perform(req) |>
    expect_error(class = "httr2_http_404") |>
    catch_cnd("httr2_fetch")
  expect_equal(cnd$n, 1)
})

test_that("don't retry curl errors by default", {
  req <- request("") |> req_retry(max_tries = 2, failure_realm = "x")
  expect_error(req_perform(req), class = "httr2_failure")

  # But can opt-in to it
  req <- request("") |>
    req_retry(max_tries = 2, retry_on_failure = TRUE, failure_realm = "x")
  cnd <- catch_cnd(req_perform(req), "httr2_retry")
  expect_equal(cnd$tries, 1)
})

test_that("can retry a transient error", {
  req <- local_app_request(function(req, res) {
    if (res$app$locals$i == 1) {
      res$set_status(429)$set_header("retry-after", 0)$send_json(list(
        status = "waiting"
      ))
    } else {
      res$send_json(list(status = "done"))
    }
  })
  req <- req_retry(req, max_tries = 2)

  cnd <- catch_cnd(resp <- req_perform(req), "httr2_retry")
  expect_s3_class(cnd, "httr2_retry")
  expect_equal(cnd$tries, 1)
  expect_equal(cnd$delay, 0)
})


test_that("repeated transient errors still fail", {
  req <- request_test("/status/:status", status = 429) |>
    req_retry(max_tries = 3, backoff = \(i) 0)

  cnd <- req_perform(req) |>
    expect_error(class = "httr2_http_429") |>
    catch_cnd("httr2_fetch")
  expect_equal(cnd$n, 3)
})

test_that("can download 0 byte file", {
  path <- withr::local_tempfile()
  resps <- req_perform(request_test("/bytes/0"), path = path)

  expect_equal(file.size(path[[1]]), 0)
})

test_that("can cache requests with etags", {
  req <- request_test("/etag/:etag", etag = "abc") |> req_cache(tempfile())

  resp1 <- req_perform(req)
  expect_condition(
    expect_condition(
      resp2 <- req_perform(req),
      class = "httr2_cache_not_modified"
    ),
    class = "httr2_cache_save"
  )
})

test_that("can cache requests with paths (cache-control)", {
  req <- request(example_url("/cache/2")) |>
    req_cache(withr::local_tempfile())

  path1 <- withr::local_tempfile()
  expect_condition(
    resp1 <- req |> req_perform(path = path1),
    class = "httr2_cache_save"
  )
  expect_equal(resp1$body[[1]], path1)

  path2 <- withr::local_tempfile()
  expect_condition(
    resp2 <- req |> req_perform(path = path2),
    class = "httr2_cache_cached"
  )
  expect_equal(resp2$body[[1]], path2)

  # Wait until cache expires
  cached_resp <- cache_get(req)
  info <- resp_cache_info(cached_resp)
  Sys.sleep(max(as.double(info$expires - Sys.time()), 0))

  path3 <- withr::local_tempfile()
  expect_condition(
    resp3 <- req |> req_perform(path = path3),
    class = "httr2_cache_save"
  )
  expect_equal(resp3$body[[1]], path3)
})

test_that("can cache requests with paths (if-modified-since)", {
  req <- request(example_url("/cache")) |>
    req_cache(tempfile())

  path1 <- tempfile()
  expect_condition(
    resp1 <- req |> req_perform(path = path1),
    class = "httr2_cache_save"
  )
  expect_equal(resp1$body[[1]], path1)

  path2 <- tempfile()
  expect_condition(
    expect_condition(
      resp2 <- req |> req_perform(path = path2),
      class = "httr2_cache_not_modified"
    ),
    class = "httr2_cache_save"
  )
  expect_equal(resp2$body[[1]], path2)
})

test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    if (req$url == "https://ok") {
      response()
    } else {
      response(404)
    }
  })

  expect_equal(req_perform(req_200), response())
  expect_error(req_perform(req_404), class = "httr2_http_404")
})

test_that("checks input types", {
  req <- request_test()
  expect_snapshot(error = TRUE, {
    req_perform(req, path = 1)
    req_perform(req, verbosity = 1.5)
    req_perform(req, mock = 7)
  })
})

# otel -----------------------------------------------------------------------

test_that("tracing works as expected", {
  skip_if_not_installed("otelsdk")

  spans <- with_otel_record({
    # A request with no URL (which shouldn't create a span).
    try(req_perform(request("")), silent = TRUE)

    # A regular request.
    resp <- req_perform(request_test("/headers"))

    # Verify that context propagation works as expected.
    expect_true("traceparent" %in% names(resp_body_json(resp)[["headers"]]))

    # A request with an HTTP error.
    try(
      req_perform(request_test("/status/:status", status = 404)),
      silent = TRUE
    )

    # A request with basic credentials that we should redact.
    parsed <- url_parse(example_url())
    parsed$username <- "user"
    parsed$password <- "secret"
    req_perform(request(url_build(parsed)))

    # A request with a curl error.
    with_mocked_bindings(
      try(req_perform(request("http://127.0.0.1")), silent = TRUE),
      curl_fetch = function(...) abort("Failed to connect")
    )

    # A request that triggers retries, generating three individual spans.
    request_test("/status/:status", status = 429) |>
      req_retry(max_tries = 3, backoff = ~0) |>
      req_perform() |>
      try(silent = TRUE)
  })[["traces"]]

  expect_length(spans, 7L)

  # Validate the span for regular requests.
  expect_equal(spans[[1]]$status, "ok")
  expect_named(
    spans[[1]]$attributes,
    c(
      "http.response.status_code",
      "user_agent.original",
      "url.full",
      "server.address",
      "server.port",
      "http.request.method"
    ),
    ignore.order = TRUE
  )
  expect_equal(spans[[1]]$attributes$http.request.method, "GET")
  expect_equal(spans[[1]]$attributes$http.response.status_code, 200L)
  expect_equal(spans[[1]]$attributes$server.address, "127.0.0.1")
  expect_match(spans[[1]]$attributes$user_agent.original, "^httr2/")

  # And for requests with HTTP errors.
  expect_equal(spans[[2]]$status, "error")
  expect_equal(spans[[2]]$description, "Not Found")
  expect_equal(spans[[2]]$attributes$http.response.status_code, 404L)
  expect_equal(spans[[2]]$attributes$error.type, "404")

  # And for spans with redacted credentials.
  expect_match(
    spans[[3]]$attributes$url.full,
    regexp = "http://REDACTED:REDACTED@127.0.0.1:[0-9]+/",
  )

  # And for spans with curl errors.
  expect_equal(spans[[4]]$status, "error")
  expect_equal(spans[[4]]$attributes$error.type, "rlang_error")

  # We should have attached the curl error as an event.
  expect_length(spans[[4]]$events, 1L)
  expect_equal(spans[[4]]$events[[1]]$name, "exception")

  # For spans with retries, we expect the parent context to be the same for
  # each span. (In this case, there is no parent span, so it should be empty.)
  # It is important that they not be children of one another.
  expect_equal(spans[[5]]$parent, spans[[6]]$parent)
  expect_equal(spans[[6]]$parent, spans[[7]]$parent)

  # Verify that we set resend counts correctly.
  expect_equal(spans[[6]]$attributes$http.request.resend_count, 2L)
  expect_equal(spans[[7]]$attributes$http.request.resend_count, 3L)
})

test_that("checks its inputs", {
  req <- request_test("/status/:status", status = 200)

  expect_snapshot(error = TRUE, {
    req_perform_promise(1)
    req_perform_promise(req, path = 1)
    req_perform_promise(req, pool = "INVALID")
    req_perform_promise(req, verbosity = "INVALID")
  })
})

test_that("returns a promise that resolves", {
  p1 <- req_perform_promise(request_test("/delay/:secs", secs = 0.25))
  p2 <- req_perform_promise(request_test("/delay/:secs", secs = 0.25))
  expect_s3_class(p1, "promise")
  expect_s3_class(p2, "promise")
  p1_value <- extract_promise(p1)
  expect_equal(resp_status(p1_value), 200)
  p2_value <- extract_promise(p2)
  expect_equal(resp_status(p2_value), 200)
})

test_that("correctly prepares request", {
  req <- request_test("/post") |> req_method("POST")
  prom <- req_perform_promise(req)
  expect_no_error(extract_promise(prom))
})

test_that("correctly prepares request", {
  req <- request_test("/get")
  expect_snapshot(
    . <- extract_promise(req_perform_promise(req, verbosity = 1)),
    transform = function(x) {
      gsub(
        "(Date|Host|User-Agent|ETag|Content-Length|Accept-Encoding): .*",
        "\\1: <variable>",
        x
      )
    }
  )
})

test_that("can promise to download files", {
  req <- request_test("/json")
  path <- withr::local_tempfile()
  p <- req_perform_promise(req, path)
  expect_s3_class(p, "promise")
  p_value <- extract_promise(p)
  expect_equal(p_value$body, new_path(path))

  # And check that something was downloaded
  expect_gt(file.size(path), 0)
})

test_that("promises can retrieve from cache", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(
    200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  p <- req_perform_promise(req)
  expect_s3_class(p, "promise")
  p_value <- extract_promise(p)
  expect_equal(p_value, resp)
})

test_that("both curl and HTTP errors in promises are rejected", {
  expect_error(
    extract_promise(
      req_perform_promise(request_test("/status/:status", status = 404))
    ),
    class = "httr2_http_404"
  )
  expect_error(
    extract_promise(
      req_perform_promise(request("INVALID"))
    ),
    class = "httr2_failure"
  )
})

test_that("req_perform_promise doesn't leave behind poller", {
  skip_if_not(
    later::loop_empty(),
    "later::global_loop not empty when test started"
  )
  p <- req_perform_promise(request_test("/delay/:secs", secs = 0.25))
  # Before promise is resolved, there should be an operation in our later loop
  expect_false(later::loop_empty())
  p_value <- extract_promise(p)
  # But now that that our promise is resolved, we shouldn't still be polling the pool
  expect_true(later::loop_empty())
})


test_that("req_perform_promise can use non-default pool", {
  custom_pool <- curl::new_pool()
  p1 <- req_perform_promise(request_test("/delay/:secs", secs = 0.25))
  p2 <- req_perform_promise(
    request_test("/delay/:secs", secs = 0.25),
    pool = custom_pool
  )
  expect_equal(length(curl::multi_list(custom_pool)), 1)
  p1_value <- extract_promise(p1)
  expect_equal(resp_status(p1_value), 200)
  p2_value <- extract_promise(p2)
  expect_equal(resp_status(p2_value), 200)
})

test_that("req_perform_promise uses the default loop", {
  # The main reason for temp loops is to allow an asynchronous operation to be
  # created, waited on, and resolved/rejected inside of a synchronous function,
  # all without affecting any asynchronous operations that existed before the
  # temp loop was created.

  # This can't proceed within the temp loop
  p1 <- req_perform_promise(request_test("/delay/:secs", secs = 0.25))

  later::with_temp_loop({
    # You can create an async response with explicit pool=NULL, but it can't
    # proceed until the temp loop is over
    p2 <- req_perform_promise(request_test("/get"), pool = NULL)

    # You can create an async response with explicit pool=pool, and it can
    # proceed as long as that pool was first used inside of the temp loop
    p3 <- req_perform_promise(request_test("/get"), pool = curl::new_pool())

    # You can't create an async response in the temp loop without explicitly
    # specifying a pool
    expect_snapshot(
      p4 <- req_perform_promise(request_test("/get")),
      error = TRUE
    )

    # Like I said, you can create this, but it won't work until we get back
    # outside the temp loop
    expect_null(extract_promise(p2, timeout = 1))

    # This works fine inside the temp loop, because its pool was first used
    # inside
    expect_equal(resp_status(extract_promise(p3, timeout = 1)), 200)
  })

  # These work fine now that we're back outside the temp loop
  expect_equal(resp_status(extract_promise(p1, timeout = 1)), 200)
  expect_equal(resp_status(extract_promise(p2, timeout = 1)), 200)
})

test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    if (req$url == "https://ok") {
      response()
    } else {
      response(404)
    }
  })
  expect_equal(
    extract_promise(req_perform_promise(req_200)),
    response()
  )
  expect_error(
    extract_promise(req_perform_promise(req_404)),
    class = "httr2_http_404"
  )
})

# otel -----------------------------------------------------------------------

test_that("tracing works as expected", {
  skip_if_not_installed("otelsdk")
  skip_on_os("windows")

  spans <- with_otel_record({
    # A request with no URL (which shouldn't create a span).
    try(req_perform_promise(request("")), silent = TRUE)

    # A request with an HTTP error.
    p <- req_perform_promise(request_test("/status/:status", status = 404))
    try(extract_promise(p), silent = TRUE)

    # A request with a curl error.
    with_mocked_bindings(
      {
        p <- req_perform_promise(request("http://127.0.0.1"))
        try(extract_promise(p), silent = TRUE)
      },
      curl_fetch = function(...) abort("Failed to connect")
    )

    # A request with no parent context.
    p <- req_perform_promise(request_test("/headers"))

    # A request nested inside a parent span.
    parent <- otel::start_span("parent", tracer = "test")
    otel::with_active_span(parent, {
      child <- req_perform_promise(request_test())
      # Resolve the earlier request with a different session; this should not
      # affect its parent.
      resp <- extract_promise(p)
    })
    extract_promise(child)
    parent$end()

    # Verify that context propagation works as expected.
    expect_true(
      "traceparent" %in% names(resp_body_json(resp)[["headers"]])
    )
  })[["traces"]]

  expect_length(spans, 5L)

  # And for requests with HTTP errors.
  expect_equal(spans[[1]]$status, "error")
  expect_equal(spans[[1]]$description, "Not Found")
  expect_equal(spans[[1]]$attributes$http.response.status_code, 404L)
  expect_equal(spans[[1]]$attributes$error.type, "404")

  # And for spans with curl errors.
  expect_equal(spans[[2]]$status, "error")
  expect_equal(spans[[2]]$attributes$error.type, "curl_error_couldnt_connect")

  # We should have attached the curl error as an event.
  expect_length(spans[[2]]$events, 1L)
  expect_equal(spans[[2]]$events[[1]]$name, "exception")

  # Verify that the spans for requests resolved later still have the parent
  # context in which they were submitted.
  expect_equal(spans[[3]]$parent, spans[[5]]$parent)
  expect_equal(spans[[4]]$parent, spans[[5]]$span_id)
})

test_that("nothing happens if cache not enabled", {
  req <- request("http://example.com")
  expect_equal(cache_pre_fetch(req), req)

  resp <- response()
  expect_equal(cache_post_fetch(req, resp), resp)
})

test_that("never retrieves POST request from cache", {
  req <- request("http://example.com") |>
    req_method("POST") |>
    req_cache(tempfile())

  # Fake an equivalent GET request in the cache
  resp <- response(
    200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  expect_equal(cache_pre_fetch(req), req)
})

test_that("immutable objects retrieved directly from cache", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(
    200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  expect_equal(cache_pre_fetch(req), resp)
})

test_that("cached cache header added to request", {
  req <- request("http://example.com") |> req_cache(tempfile())
  # If not cached, request returned as is
  req2 <- cache_pre_fetch(req)
  expect_equal(req2, req)

  resp <- response(
    200,
    headers = c('Etag: "abc"', "Last-Modified: Wed, 01 Jan 2020 00:00:00 GMT"),
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  # After caching adds caching headers
  req3 <- cache_pre_fetch(req)
  expect_equal(
    req3$headers$`If-Modified-Since`,
    "Wed, 01 Jan 2020 00:00:00 GMT"
  )
  expect_equal(req3$headers$`If-None-Match`, '"abc"')
})

test_that("error can use cached value", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(200, body = charToRaw("OK"))
  cache_set(req, resp)

  expect_equal(cache_post_fetch(req, error_cnd()), error_cnd())

  req$policies$cache_use_on_error <- TRUE
  expect_equal(cache_post_fetch(req, error_cnd()), resp)
})

test_that("304 retains headers but gets cached body", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(200, headers = "X: 1", body = charToRaw("OK"))
  cache_set(req, resp)

  cached <- cache_post_fetch(req, response(304, headers = "X: 2"))
  expect_equal(cached$headers$x, "2")
  expect_equal(cached$body, resp$body)

  cached <- cache_post_fetch(req, response(304, headers = "X: 3"))
  expect_equal(cached$headers$x, "3")
  expect_equal(cached$body, resp$body)
})

test_that("automatically adds to cache", {
  req <- request("http://example.com") |> req_cache(tempfile())
  expect_true(is.null(cache_get(req)))

  resp <- response(200, headers = 'Etag: "abc"', body = charToRaw("OK"))
  cached <- cache_post_fetch(req, resp)
  expect_false(is.null(cache_get(req)))
  expect_equal(cache_get(req), resp)
})

test_that("cache emits useful debugging info", {
  req <- request("http://example.com") |> req_cache(tempfile(), debug = TRUE)
  resp <- response(
    200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )

  expect_snapshot({
    "Immutable"
    invisible(cache_pre_fetch(req))
    invisible(cache_post_fetch(req, resp))
    invisible(cache_pre_fetch(req))
  })

  req <- request("http://example.com") |>
    req_cache(tempfile(), debug = TRUE, use_on_error = TRUE)
  resp <- response(200, headers = "X: 1", body = charToRaw("OK"))
  cache_set(req, resp)
  expect_snapshot({
    "freshness check"
    invisible(cache_pre_fetch(req))
    invisible(cache_post_fetch(req, response(304)))
    invisible(cache_post_fetch(req, error_cnd()))
  })
})

# cache -------------------------------------------------------------------

test_that("can get and set from cache", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(
    200,
    headers = list(
      Etag = "ABC",
      `content-type` = "application/json",
      other = "header"
    ),
    body = charToRaw(jsonlite::toJSON(list(a = jsonlite::unbox(1))))
  )

  cached_resp <- response(
    304,
    headers = list(
      Etag = "DEF",
      other = "new"
    )
  )

  expect_true(is.null(cache_get(req)))
  cache_set(req, resp)
  expect_false(is.null(cache_get(req)))

  resp_from_cache <- cache_get(req)
  expect_equal(resp_from_cache, resp)

  # Uses new headers if available, otherwise cached headers
  out_headers <- cache_headers(resp_from_cache, cached_resp)
  expect_equal(out_headers$`content-type`, "application/json")
  expect_equal(out_headers$Etag, "DEF")
  expect_equal(out_headers$other, "new")

  # If path is null can leave resp as is
  expect_equal(cache_body(resp_from_cache, NULL), resp$body)
  expect_equal(resp_body_json(resp_from_cache), list(a = 1L))
  # If path is set, need to save to path
  path <- tempfile()
  body <- cache_body(resp_from_cache, path)
  expect_equal(body, new_path(path))
  expect_equal(readLines(path, warn = FALSE), rawToChar(resp$body))
})

test_that("handles responses with files", {
  req <- request("http://example.com") |> req_cache(tempfile())

  path <- local_write_lines("Hi there")
  resp <- response(200, headers = "Etag: ABC", body = new_path(path))
  cache_set(req, resp)

  # File should be copied in cache directory, and response body updated
  body_path <- req_cache_path(req, ".body")
  expect_equal(readLines(body_path), "Hi there")

  resp_from_cache <- cache_get(req)
  expect_equal(resp_from_cache$body, new_path(body_path))

  # If path is null, just leave body as is, since req_body() already
  # papers over the differences
  expect_equal(cache_body(resp_from_cache, NULL), new_path(body_path))

  # If path is not null, copy to desired location, and update body
  path2 <- tempfile()
  body <- cache_body(resp_from_cache, path2)
  expect_equal(readLines(body), "Hi there")
  expect_equal(body, new_path(path2))
})

test_that("corrupt files are ignored", {
  cache_dir <- withr::local_tempdir()
  req <- request("http://example.com") |> req_cache(cache_dir)

  writeLines(letters, req_cache_path(req))
  expect_true(is.null(cache_get(req)))

  saveRDS(1:10, req_cache_path(req))
  expect_false(is.null(cache_get(req)))
})

# pruning -----------------------------------------------------------------

test_that("pruning is throttled", {
  path <- withr::local_tempdir()
  req <- req_cache(request_test(), path = path)

  expect_true(cache_prune_if_needed(req))
  expect_false(cache_prune_if_needed(req))
  expect_true(cache_prune_if_needed(req, threshold = 0))

  the$cache_throttle[[path]] <- Sys.time() - 61
  expect_true(cache_prune_if_needed(req, threshold = 60))
})

test_that("can prune by number", {
  path <- withr::local_tempdir()
  file.create(file.path(path, c("a.rds", "b.rds", "c.rds")))
  Sys.sleep(0.1)
  file.create(file.path(path, c("d.rds")))

  cache_prune(path, list(n = 4, age = Inf, size = Inf), debug = TRUE)
  expect_equal(dir(path), c("a.rds", "b.rds", "c.rds", "d.rds"))

  expect_snapshot(
    cache_prune(path, list(n = 1, age = Inf, size = Inf), debug = TRUE)
  )
  expect_equal(dir(path), c("d.rds"))
})

test_that("can prune by age", {
  path <- withr::local_tempdir()
  file.create(file.path(path, c("a.rds", "b.rds")))
  Sys.setFileTime(file.path(path, "a.rds"), Sys.time() - 60)

  cache_prune(path, list(n = Inf, age = 120, size = Inf), debug = TRUE)
  expect_equal(dir(path), c("a.rds", "b.rds"))

  expect_snapshot({
    cache_prune(path, list(n = Inf, age = 30, size = Inf), debug = TRUE)
  })
  expect_equal(dir(path), "b.rds")
})

test_that("can prune by size", {
  path <- withr::local_tempdir()
  writeChar(paste0(letters, collapse = ""), file.path(path, "a.rds"))
  writeChar(paste0(letters, collapse = ""), file.path(path, "b.rds"))
  Sys.sleep(0.1)
  writeChar(paste0(letters, collapse = ""), file.path(path, "c.rds"))

  cache_prune(path, list(n = Inf, age = Inf, size = 200), debug = TRUE)
  expect_equal(dir(path), c("a.rds", "b.rds", "c.rds"))

  expect_snapshot({
    cache_prune(path, list(n = Inf, age = Inf, size = 50), debug = TRUE)
  })
  expect_equal(dir(path), "c.rds")
})

# headers -----------------------------------------------------------------

test_that("correctly determines if response is cacheable", {
  is_cacheable <- function(...) {
    resp_is_cacheable(response(...))
  }

  expect_equal(is_cacheable(200, headers = "Expires: ABC"), TRUE)
  expect_equal(is_cacheable(200, headers = "Cache-Control: max-age=10"), TRUE)
  expect_equal(is_cacheable(200, headers = "Etag: ABC"), TRUE)
  expect_equal(
    is_cacheable(200, headers = c("Etag: ABC", "Cache-Control: no-store")),
    FALSE
  )
  expect_equal(is_cacheable(200), FALSE)
  expect_equal(is_cacheable(404), FALSE)
  expect_equal(is_cacheable(method = "POST"), FALSE)
})

test_that("can extract cache info with correct types", {
  resp <- response(
    headers = c(
      "Expires: Wed, 01 Jan 2020 00:00:00 GMT",
      "Last-Modified: Wed, 01 Jan 2010 00:00:00 GMT",
      "Etag: \"abc\""
    )
  )
  info <- resp_cache_info(resp)

  expect_equal(info$expires, local_time("2020-01-01"))
  # passed as is back to server, so leave as string
  expect_equal(info$last_modified, "Wed, 01 Jan 2010 00:00:00 GMT")
  # quotes are part of the etag string
  expect_equal(info$etag, '"abc"')
})

test_that("can extract various expiry values", {
  # Prefer Date + max-age
  resp1 <- response(
    headers = c(
      "Date: Wed, 01 Jan 2020 00:00:00 GMT",
      "Cache-Control: max-age=3600",
      "Expiry: Wed, 01 Jan 2020 00:00:00 GMT"
    )
  )
  expect_equal(resp_cache_expires(resp1), local_time("2020-01-01 01:00"))

  # Fall back to Expires
  resp2 <- response(
    headers = c(
      "Expires: Wed, 01 Jan 2020 00:00:00 GMT"
    )
  )
  expect_equal(resp_cache_expires(resp2), local_time("2020-01-01 00:00"))

  # Returns NA if no expiry
  resp2 <- response()
  expect_equal(resp_cache_expires(resp2), NA)
})

#' Secret management
#'
#' @description
#' httr2 provides a handful of functions designed for working with confidential
#' data. These are useful because testing packages that use httr2 often
#' requires some confidential data that needs to be available for testing,
#' but should not be available to package users.
#'
#' * `secret_encrypt()` and `secret_decrypt()` work with individual strings
#' * `secret_encrypt_file()` encrypts a file in place and
#'   `secret_decrypt_file()` decrypts a file in a temporary location.
#' * `secret_write_rds()` and `secret_read_rds()` work with `.rds` files
#' * `secret_make_key()` generates a random string to use as a key.
#' * `secret_has_key()` returns `TRUE` if the key is available; you can
#'   use it in examples and vignettes that you want to evaluate on your CI,
#'   but not for CRAN/package users.
#'
#' These all look for the key in an environment variable. When used inside of
#' testthat, they will automatically [testthat::skip()] the test if the env var
#' isn't found. (Outside of testthat, they'll error if the env var isn't
#' found.)
#'
#' # Basic workflow
#'
#' 1.  Use `secret_make_key()` to generate a password. Make this available
#'     as an env var (e.g. `{MYPACKAGE}_KEY`) by adding a line to your
#'     `.Renviron`.
#'
#' 2.  Encrypt strings with `secret_encrypt()`, files with
#'     `secret_encrypt_file()`, and other data with `secret_write_rds()`,
#'     setting `key = "{MYPACKAGE}_KEY"`.
#'
#' 3.  In your tests, decrypt the data with `secret_decrypt()`,
#'     `secret_decrypt_file()`, or `secret_read_rds()` to match how you encrypt
#'     it.
#'
#' 4.  If you push this code to your CI server, it will already "work" because
#'     all functions automatically skip tests when your `{MYPACKAGE}_KEY`
#'     env var isn't set. To make the tests actually run, you'll need to set
#'     the env var using whatever tool your CI system provides for setting
#'     env vars. Make sure to carefully inspect the test output to check that
#'     the skips have actually gone away.
#'
#' @name secrets
#' @returns
#' * `secret_decrypt()` and `secret_encrypt()` return strings.
#' * `secret_decrypt_file()` returns a path to a temporary file;
#'   `secret_encrypt_file()` encrypts the file in place.
#' * `secret_write_rds()` returns `x` invisibly; `secret_read_rds()`
#'   returns the saved object.
#' * `secret_make_key()` returns a string with class `AsIs`.
#' * `secret_has_key()` returns `TRUE` or `FALSE`.
#' @aliases NULL
#' @examples
#' key <- secret_make_key()
#'
#' path <- tempfile()
#' secret_write_rds(mtcars, path, key = key)
#' secret_read_rds(path, key)
#'
#' # While you can manage the key explicitly in a variable, it's much
#' # easier to store in an environment variable. In real life, you should
#' # NEVER use `Sys.setenv()` to create this env var because you will
#' # also store the secret in your `.Rhistory`. Instead add it to your
#' # .Renviron using `usethis::edit_r_environ()` or similar.
#' Sys.setenv("MY_KEY" = key)
#'
#' x <- secret_encrypt("This is a secret", "MY_KEY")
#' x
#' secret_decrypt(x, "MY_KEY")
NULL

#' @export
#' @rdname secrets
secret_make_key <- function() {
  I(base64_url_rand(16))
}

#' @export
#' @rdname secrets
#' @param x Object to encrypt. Must be a string for `secret_encrypt()`.
#' @param key Encryption key; this is the password that allows you to "lock"
#'   and "unlock" the secret. The easiest way to specify this is as the
#'   name of an environment variable. Alternatively, if you already have
#'   a base64url encoded string, you can wrap it in `I()`, or you can pass
#'   the raw vector in directly.
secret_encrypt <- function(x, key) {
  check_string(x)

  enc <- secret_encrypt_raw(charToRaw(x), key)
  base64_url_encode(enc)
}
#' @export
#' @rdname secrets
#' @param encrypted String to decrypt
secret_decrypt <- function(encrypted, key) {
  check_string(encrypted)

  enc <- base64_url_decode(encrypted)
  dec <- secret_decrypt_raw(enc, key)

  rawToChar(dec)
}

#' @export
#' @rdname secrets
secret_write_rds <- function(x, path, key) {
  x <- serialize(x, NULL, version = 2)
  x_cmp <- memCompress(x, "bzip2")
  enc <- secret_encrypt_raw(x_cmp, key)
  writeBin(enc, path)

  invisible(x)
}
#' @export
#' @rdname secrets
#' @param path Path to file to encrypted file to read or write. For
#'  `secret_write_rds()` and `secret_read_rds()` this should be an `.rds`
#'  file.
secret_read_rds <- function(path, key) {
  enc <- readBin(path, "raw", file.size(path))
  dec_cmp <- secret_decrypt_raw(enc, key)
  dec <- memDecompress(dec_cmp, "bzip2")

  unserialize(dec)
}

#' @export
#' @param envir The decrypted file will be automatically deleted when
#'   this environment exits. You should only need to set this argument if you
#'   want to pass the unencrypted file to another function.
#' @rdname secrets
secret_decrypt_file <- function(path, key, envir = parent.frame()) {
  enc <- readBin(path, "raw", file.size(path))
  dec <- secret_decrypt_raw(enc, key = key)

  path <- tempfile()
  withr::defer(unlink(path), envir)
  writeBin(dec, path)
  Sys.chmod(path, 400)
  path
}

#' @export
#' @rdname secrets
secret_encrypt_file <- function(path, key) {
  dec <- readBin(path, "raw", file.info(path)$size)
  enc <- secret_encrypt_raw(dec, key = key)

  writeBin(enc, path)
  invisible(path)
}

#' @export
#' @rdname secrets
secret_has_key <- function(key) {
  check_string(key)
  key <- Sys.getenv(key)
  !identical(key, "")
}

secret_get_key <- function(envvar, call = caller_env()) {
  key <- Sys.getenv(envvar)

  if (identical(key, "")) {
    msg <- glue("Can't find envvar {envvar}")
    if (is_testing()) {
      testthat::skip(msg)
    } else {
      abort(msg, call = call)
    }
  }

  base64_url_decode(key)
}


#' Obfuscate mildly secret information
#'
#' @description
#' Use `obfuscate("value")` to generate a call to `obfuscated()`, which will
#' unobfuscate the value at the last possible moment. Obfuscated values only
#' work in limited locations:
#'
#' * The `secret` argument to [oauth_client()]
#' * Elements of the `data` argument to [req_body_form()], `req_body_json()`,
#'   and `req_body_multipart()`.
#'
#' Working together this pair of functions provides a way to obfuscate mildly
#' confidential information, like OAuth client secrets. The secret can not be
#' revealed from your inspecting source code, but a skilled R programmer could
#' figure it out with some effort. The main goal is to protect against scraping;
#' there's no way for an automated tool to grab your obfuscated secrets.
#'
#' @param x A string to `obfuscate`, or mark as `obfuscated`.
#' @returns `obfuscate()` prints the `obfuscated()` call to include in your
#'   code. `obfuscated()` returns an S3 class marking the string as obfuscated
#'   so it can be unobfuscated when needed.
#' @export
#' @examples
#' obfuscate("good morning")
#'
#' # Every time you obfuscate you'll get a different value because it
#' # includes 16 bytes of random data which protects against certain types of
#' # brute force attack
#' obfuscate("good morning")
obfuscate <- function(x) {
  check_string(x)

  enc <- secret_encrypt(x, obfuscate_key())
  glue('obfuscated("{enc}")')
}
attr(obfuscate, "srcref") <- "function(x) {}"

#' @export
#' @rdname obfuscate
obfuscated <- function(x) {
  structure(x, class = "httr2_obfuscated")
}

#' @export
str.httr2_obfuscated <- function(object, ...) {
  cat(" ", glue('obfuscated("{object}")\n'), sep = "")
}

#' @export
print.httr2_obfuscated <- function(x, ...) {
  cat(glue('obfuscated("{x}")\n'))
  invisible(x)
}

unobfuscate <- function(x, handle = c("reveal", "redact", "remove")) {
  handle <- arg_match(handle)
  unobfuscate_rec(x, handle = handle)
}
unobfuscate_rec <- function(x, handle = "reveal") {
  if (is_obfuscated(x)) {
    switch(
      handle,
      reveal = secret_decrypt(x, obfuscate_key()),
      redact = "<REDACTED>",
      remove = NULL
    )
  } else if (is.list(x)) {
    x[] <- lapply(x, unobfuscate, handle = handle)
    x
  } else {
    x
  }
}
attr(unobfuscate, "srcref") <- "function(x) {}"

is_obfuscated <- function(x) {
  inherits(x, "httr2_obfuscated")
}

# Helpers -----------------------------------------------------------------

secret_encrypt_raw <- function(dec, key, error_call = caller_env()) {
  key <- as_key(key, error_call = error_call)

  enc <- openssl::aes_ctr_encrypt(dec, key)
  c(attr(enc, "iv"), enc)
}

secret_decrypt_raw <- function(enc, key, error_call = caller_env()) {
  key <- as_key(key, error_call = error_call)

  iv <- enc[1:16]
  value <- enc[-(1:16)]

  openssl::aes_ctr_decrypt(value, key, iv = iv)
}

as_key <- function(x, error_call = caller_env()) {
  if (inherits(x, "AsIs") && is_string(x)) {
    base64_url_decode(x)
  } else if (is.raw(x)) {
    x
  } else if (is_string(x)) {
    secret_get_key(x, call = error_call)
  } else {
    cli::cli_abort(
      paste0(
        "{.arg key} must be a raw vector containing the key, ",
        "a string giving the name of an env var, ",
        "or a string wrapped in {.fn I} that contains the base64url encoded key."
      ),
      call = error_call
    )
  }
}

# 1. Create service account
# 2. Add key and download json
# 3. json <- jsonlite::read_json(path)
# 4. secret_write_rds(json, "tests/testthat/test-oauth-flow-jwt-google.rds", "HTTR2_KEY")

test_that("can generate token and use it automatically", {
  secrets <- secret_read_rds(
    test_path("test-oauth-flow-jwt-google.rds"),
    "HTTR2_KEY"
  )

  client <- oauth_client(
    id = secrets$client_id,
    key = secrets$private_key,
    token_url = secrets$token_uri,
    auth = "body"
  )
  claim <- list(
    iss = secrets$client_email,
    scope = "https://www.googleapis.com/auth/userinfo.email",
    aud = "https://oauth2.googleapis.com/token"
  )

  # Can generate token
  token <- oauth_flow_bearer_jwt(client, claim)
  expect_s3_class(token, "httr2_token")

  # Can use it in request
  resp <- request("https://openidconnect.googleapis.com/v1/userinfo") |>
    req_oauth_bearer_jwt(client, claim) |>
    req_perform() |>
    resp_body_json()

  expect_type(resp, "list")
  expect_equal(resp$email_verified, TRUE)
})

test_that("validates inputs", {
  client1 <- oauth_client("test", "http://example.com")
  expect_snapshot(oauth_flow_bearer_jwt(client1), error = TRUE)

  client2 <- oauth_client(
    "test",
    "http://example.com",
    key = "abc",
    auth_params = list(claim = "123")
  )
  expect_snapshot(oauth_flow_bearer_jwt(client2, claim = NULL), error = TRUE)
})

test_that("encryption and decryption of string is symmetric", {
  key <- secret_make_key()

  x <- "Testing 1...2...3..."
  enc <- secret_encrypt(x, key)
  dec <- secret_decrypt(enc, key)
  expect_equal(dec, x)
})

test_that("encryption and decryption of object is symmetric", {
  key <- secret_make_key()
  path <- withr::local_tempfile()

  x1 <- list(1:10, letters)
  secret_write_rds(x1, path, key)
  x2 <- secret_read_rds(path, key)
  expect_equal(x1, x2)
})

test_that("encryption and decryption of file is symmetric", {
  key <- secret_make_key()
  path <- withr::local_tempfile(lines = letters)

  secret_encrypt_file(path, key)

  local({
    path_dec <<- secret_decrypt_file(path, key)
    expect_equal(readLines(path_dec, warn = FALSE), letters)
  })
  expect_false(file.exists(path_dec))
})

test_that("can unobfuscate obfuscated string", {
  x <- obfuscated("qw6Ua_n2LR_xzuk2uqp2dhb5OaE")
  expect_equal(unobfuscate(x), "test")
})

test_that("obfuscated strings are hidden", {
  expect_snapshot({
    x <- obfuscated("abcdef")
    x
    str(x)
  })
})

test_that("unobfuscate operates recursively", {
  expect_equal(unobfuscate(NULL), NULL)
  expect_equal(unobfuscate("x"), "x")
  expect_equal(
    unobfuscate(list(list(obfuscated("qw6Ua_n2LR_xzuk2uqp2dhb5OaE")))),
    list(list("test"))
  )
})

test_that("unobfuscated can control behaviour", {
  x <- list(obfuscated("JKWA-5KOJpjZcuwVYjILoayq4A"))
  expect_equal(unobfuscate(x, "reveal"), list("abc"))
  expect_equal(unobfuscate(x, "redact"), list("<REDACTED>"))
  expect_equal(unobfuscate(x, "remove"), list(NULL))
})

test_that("secret_has_key returns FALSE/TRUE", {
  withr::local_envvar(ENVVAR_THAT_DOES_EXIST = "1")
  expect_equal(secret_has_key("ENVVAR_THAT_DOESNT_EXIST"), FALSE)
  expect_equal(secret_has_key("ENVVAR_THAT_DOES_EXIST"), TRUE)
})


test_that("can coerce to a key", {
  expect_equal(as_key(I("YWJj")), charToRaw("abc"))
  expect_equal(as_key(as.raw(c(1, 2, 3))), as.raw(c(1, 2, 3)))

  withr::local_envvar(KEY = "YWJj", TESTTHAT = "false")
  expect_equal(as_key("KEY"), charToRaw("abc"))

  expect_snapshot(error = TRUE, {
    as_key("ENVVAR_THAT_DOESNT_EXIST")
    as_key(1)
  })
})

#' @export
#' @rdname resp_stream_raw
#' @param lines The maximum number of lines to return at once.
#' @param warn `r lifecycle::badge("deprecated")` `resp_stream_lines()` no longer
#'   warns when the connection ends without a final EOL, so this argument is
#'   ignored.
#' @order 2
resp_stream_lines <- function(
  resp,
  lines = 1,
  max_size = Inf,
  warn = deprecated()
) {
  splitter <- init_streaming_response(resp, LineSplitter)
  check_number_whole(lines, min = 0, allow_infinite = TRUE)
  check_number_whole(max_size, min = 1, allow_infinite = TRUE)
  if (lifecycle::is_present(warn) && !isFALSE(warn)) {
    lifecycle::deprecate_warn("1.2.3", "resp_stream_lines(warn)")
  }

  if (lines == 0) {
    return(character())
  }

  encoding <- env_cache(resp$cache, "stream_encoding", resp_encoding(resp))
  blocks <- stream_pull(resp, lines, splitter, max_size)
  lines_read <- stream_parse_lines(blocks, encoding)
  if (resp_stream_show_body(resp)) {
    log_stream(lines_read)
  }
  lines_read
}

# Splits a stream into lines terminated by LF (and hence CRLF)
LineSplitter <- R6::R6Class(
  "LineSplitter",
  inherit = StreamSplitter,
  public = list(
    name = "resp_stream_lines()",
    find_boundaries = function(buffer) {
      grepRaw(as.raw(0x0A), buffer, fixed = TRUE, all = TRUE) + 1L
    },
    # At end of stream, a trailing line without a terminator is still a line.
    finish = function(remainder) {
      if (length(remainder) == 0L) list() else list(remainder)
    }
  )
)

# Decode raw line blocks (each a line plus its trailing LF or CRLF) into a
# character vector in `encoding`, dropping the terminators.
stream_parse_lines <- function(blocks, encoding) {
  text <- vapply(blocks, rawToChar, character(1))
  Encoding(text) <- "bytes"
  text <- iconv(text, encoding, "UTF-8")
  sub("\r?\n$", "", text)
}

test_that("decodes the response encoding and joins LF and CRLF lines", {
  # Lines are decoded with the response encoding (here Shift_JIS) and split on
  # both LF and CRLF, including a CRLF straddling two reads ("split crlf\r" then
  # "\n"); a final line without a terminator is returned at EOF.
  req <- local_app_request(function(req, res) {
    res$set_header("Content-Type", "text/plain; charset=Shift_JIS")
    res$send_chunk(as.raw(c(0x82, 0xA0, 0x0A)))
    res$send_chunk("crlf\r\n")
    res$send_chunk("lf\n")
    res$send_chunk("half line/")
    res$send_chunk("other half\n")
    res$send_chunk("split crlf\r")
    res$send_chunk("\nanother line\n")
    res$send_chunk("eof without line ending")
  })
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  expected_values <- c(
    "\u3042",
    "crlf",
    "lf",
    "half line/other half",
    "split crlf",
    "another line"
  )

  for (expected in expected_values) {
    rlang::inject(expect_equal(resp_stream_lines(resp), !!expected))
  }
  expect_equal(resp_stream_lines(resp), "eof without line ending")
})

test_that("requesting zero lines returns an empty vector", {
  req <- local_app_request(function(req, res) res$send_chunk("a\n"))
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))
  expect_equal(resp_stream_lines(resp, 0), character())
})

test_that("max_size is approximate, counting the buffered delimiter bytes", {
  # A line whose content is exactly max_size is returned: we read one byte past
  # max_size, enough to capture a single LF terminator.
  resp <- local_streaming_response(c(rep(as.raw(0x61), 10), as.raw(0x0A)))
  expect_equal(resp_stream_lines(resp, max_size = 10), strrep("a", 10))

  # A two-byte CRLF tips the buffer one byte over, so the same line needs a
  # slightly larger max_size.
  crlf <- c(rep(as.raw(0x61), 10), as.raw(c(0x0D, 0x0A)))
  resp <- local_streaming_response(crlf)
  expect_error(
    resp_stream_lines(resp, max_size = 10),
    class = "httr2_streaming_error"
  )
  expect_equal(resp_stream_lines(resp, max_size = 11), strrep("a", 10))
})

test_that("max_size counts an incomplete delimiter at EOF as content", {
  resp <- local_streaming_response(c(rep(as.raw(0x61), 10), as.raw(0x0D)))

  expect_error(
    resp_stream_lines(resp, max_size = 10),
    class = "httr2_streaming_error"
  )
})

test_that("resp_stream_lines(warn) is deprecated unless FALSE", {
  req <- local_app_request(function(req, res) res$send_chunk("a\n"))
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  # warn = FALSE already requested silence, so it's accepted quietly.
  expect_no_warning(. <- resp_stream_lines(resp, warn = FALSE))
  # Any other value is deprecated.
  expect_snapshot(. <- resp_stream_lines(resp, warn = TRUE))
})

test_that("LineSplitter flushes a trailing line as a raw block", {
  s <- LineSplitter$new()
  # Nothing buffered: nothing to flush.
  expect_equal(s$finish(raw()), list())
  # Trailing bytes are emitted as a final block.
  expect_equal(s$finish(charToRaw("tail")), list(charToRaw("tail")))
  # A bare CR is ordinary content, even at end-of-stream.
  expect_equal(s$finish(charToRaw("tail\r")), list(charToRaw("tail\r")))
})

test_that("LineSplitter boundaries fall one past each LF", {
  # The block-splitting itself lives in stream_pull(); here we just check where
  # lines end. We only split on LF, so CRLF and bare CR need no special casing.
  s <- LineSplitter$new()

  expect_equal(s$find_boundaries(charToRaw("a\nb\r\nc\n")), c(3L, 6L, 8L))
  # blank lines produce adjacent boundaries
  expect_equal(s$find_boundaries(charToRaw("a\n\nb\n")), c(3L, 4L, 6L))
  # a bare CR (not followed by LF) is not a terminator
  expect_equal(s$find_boundaries(charToRaw("a\rbcd")), integer())
})

test_that("stream_parse_lines() decodes blocks and strips LF/CRLF terminators", {
  blocks <- list(
    charToRaw("a\n"),
    charToRaw("b\r\n"),
    charToRaw("\n"),
    charToRaw("tail")
  )
  expect_equal(stream_parse_lines(blocks, "UTF-8"), c("a", "b", "", "tail"))

  # honors the encoding, and a bare CR stays as content
  expect_equal(
    stream_parse_lines(list(as.raw(c(0x82, 0xA0, 0x0A))), "Shift_JIS"),
    "あ"
  )
  expect_equal(stream_parse_lines(list(charToRaw("a\rb\n")), "UTF-8"), "a\rb")

  # no blocks gives an empty vector
  expect_equal(stream_parse_lines(list(), "UTF-8"), character())
})

#' Perform a dry run
#'
#' This shows you exactly what httr2 will send to the server, without
#' actually sending anything. It requires the httpuv package because it
#' works by sending the real HTTP request to a local webserver, thanks to
#' the magic of [curl::curl_echo()].
#'
#' ## Limitations
#'
#' * The HTTP version is always `HTTP/1.1` (since you can't determine what it
#'   will actually be without connecting to the real server).
#'
#' @inheritParams req_verbose
#' @param quiet If `TRUE` doesn't print anything.
#' @param testing_headers If `TRUE`, removes headers that httr2 would otherwise
#'   be automatically added, which are likely to change across test runs. This
#'   currently includes:
#'
#'   * The default `User-Agent`, which varies based on libcurl, curl, and
#'     httr2 versions.
#'   * The `Host`` header, which is often set to a testing server.
#'   * The `Content-Length` header, which will often vary by platform because
#'     of varying newline encodings. (And is also not correct if you have
#'     `pretty_json = TRUE`.)
#'   * The `Accept-Encoding` header, which varies based on how libcurl was
#'     built.
#' @param pretty_json If `TRUE`, automatically prettify JSON bodies.
#' @returns Invisibly, a list containing information about the request,
#'   including `method`, `path`, and `headers`.
#' @export
#' @examples
#' # httr2 adds default User-Agent, Accept, and Accept-Encoding headers
#' request("http://example.com") |> req_dry_run()
#'
#' # the Authorization header is automatically redacted to avoid leaking
#' # credentials on the console
#' req <- request("http://example.com") |> req_auth_basic("user", "password")
#' req |> req_dry_run()
#'
#' # if you need to see it, use redact_headers = FALSE
#' req |> req_dry_run(redact_headers = FALSE)
req_dry_run <- function(
  req,
  quiet = FALSE,
  redact_headers = TRUE,
  testing_headers = is_testing(),
  pretty_json = getOption("httr2_pretty_json", TRUE)
) {
  check_request(req)
  check_bool(quiet)
  check_bool(redact_headers)
  check_bool(testing_headers)
  check_installed("httpuv")

  if (testing_headers) {
    if (!req_has_user_agent(req)) {
      req <- req_headers(req, `user-agent` = "")
    }
    req <- req_headers(req, `accept-encoding` = "")
  }

  req <- req_prepare(req)
  handle <- req_handle(req)
  handle_preflight(req, handle)
  resp <- curl::curl_echo(handle, progress = FALSE)
  headers <- new_headers(
    as.list(resp$headers),
    redact = which_redacted(req$headers),
    lifespan = current_env()
  )

  if (!quiet) {
    cli::cat_line(resp$method, " ", resp$path, " HTTP/1.1")

    if (testing_headers) {
      # curl::curl_echo() overrides
      headers$host <- NULL
      headers$`content-length` <- NULL
    }
    show_headers(headers, redact = redact_headers)
    cli::cat_line()
    show_body(resp$body, headers$`content-type`, pretty_json = pretty_json)
  }

  invisible(list(
    method = resp$method,
    path = resp$path,
    body = resp$body,
    headers = headers_flatten(headers, redact = redact_headers)
  ))
}

show_body <- function(body, content_type, prefix = "", pretty_json = FALSE) {
  if (!is.raw(body)) {
    return(invisible())
  }

  if (is_text_type(content_type)) {
    body <- rawToChar(body)
    Encoding(body) <- "UTF-8"

    if (pretty_json && content_type == "application/json") {
      body <- pretty_json(body)
    }

    body <- gsub("\n", paste0("\n", prefix), body)
    cli::cat_line(prefix, body)
  } else {
    cli::cat_line(prefix, "<", length(body), " bytes>")
  }

  invisible()
}

#' Temporarily set verbosity for all requests
#'
#' @description
#' `with_verbosity()` and `local_verbosity()` are useful for debugging httr2
#' code buried deep inside another package, because they allow you to change
#' the verbosity even when you don't have access to the request.
#'
#' Both functions work by temporarily setting the `httr2_verbosity` option. You
#' can also control verbosity by setting the `HTTR2_VERBOSITY` environment
#' variable. This has lower precedence than the option, but can be more easily
#' changed outside of R.
#'
#' @inheritParams req_perform
#' @param code Code to execture
#' @returns `with_verbosity()` returns the result of evaluating `code`.
#'   `local_verbosity()` is called for its side-effect and invisibly returns
#'   the previous value of the option.
#' @export
#' @examples
#' fun <- function() {
#'   request("https://httr2.r-lib.org") |> req_perform()
#' }
#' with_verbosity(fun())
#'
#' fun <- function() {
#'   local_verbosity(2)
#'   # someotherpackage::fun()
#' }
with_verbosity <- function(code, verbosity = 1) {
  withr::local_options(httr2_verbosity = verbosity)
  code
}

#' @export
#' @rdname with_verbosity
#' @inheritParams local_mocked_responses
local_verbosity <- function(verbosity, env = caller_env()) {
  withr::local_options(httr2_verbosity = verbosity, .local_envir = env)
}

httr2_verbosity <- function() {
  x <- getOption("httr2_verbosity")
  if (!is.null(x)) {
    return(x)
  }

  x <- Sys.getenv("HTTR2_VERBOSITY")
  if (nzchar(x)) {
    return(as.integer(x))
  }

  # Hackish fallback for httr::with_verbose
  old <- getOption("httr_config")
  if (!is.null(old$options$debugfunction)) {
    1
  } else {
    0
  }
}

test_that("respects httr2 verbosity option", {
  expect_equal(with_verbosity(httr2_verbosity()), 1)

  local({
    local_verbosity(2)
    expect_equal(httr2_verbosity(), 2)
  })
})

test_that("can set httr2 verbosity with env var", {
  withr::local_envvar(HTTR2_VERBOSITY = "1")
  expect_equal(httr2_verbosity(), 1)

  # but option has higher precedence
  withr::local_options(httr2_verbosity = 2)
  expect_equal(httr2_verbosity(), 2)
})

test_that("respects httr verbose config", {
  expect_equal(httr2_verbosity(), 0)

  # Simulate effect of httr::with_verbose(httr2_verbosity())
  config <- list(options = list(debugfunction = identity))
  withr::local_options(httr_config = config)
  expect_equal(httr2_verbosity(), 1)
})

#' OAuth with username and password
#'
#' @description
#' This function implements the OAuth **resource owner password flow**, as
#' defined by `r rfc(6749, 4.3)`. It allows the user to supply their password
#' once, exchanging it for an access token that can be cached locally.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>
#'
#' @export
#' @family OAuth flows
#' @inheritParams req_oauth_auth_code
#' @inheritParams req_auth_basic
#' @returns `req_oauth_password()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_password()` returns an [oauth_token].
#' @examples
#' req_auth <- function(req) {
#'   req_oauth_password(req,
#'     client = oauth_client("example", "https://example.com/get_token"),
#'     username = "username"
#'   )
#' }
#' if (interactive()) {
#'   request("https://example.com") |>
#'     req_auth()
#' }
req_oauth_password <- function(
  req,
  client,
  username,
  password = NULL,
  scope = NULL,
  token_params = list(),
  cache_disk = FALSE,
  cache_key = username
) {
  params <- list(
    client = client,
    username = username,
    password = password,
    scope = scope,
    token_params = token_params
  )
  cache <- cache_choose(client, cache_disk = cache_disk, cache_key = cache_key)
  req_oauth(req, "oauth_flow_password", params, cache = cache)
}

#' @export
#' @rdname req_oauth_password
oauth_flow_password <- function(
  client,
  username,
  password = NULL,
  scope = NULL,
  token_params = list()
) {
  oauth_flow_check(
    "resource owner password credentials",
    client,
    interactive = is.null(password)
  )
  check_string(username)

  oauth_client_get_token(
    client,
    grant_type = "password",
    username = username,
    password = check_password(password),
    scope = scope,
    !!!token_params
  )
}

check_password <- function(password, call = caller_env()) {
  if (is.null(password)) {
    check_installed("askpass", call = call)
    password <- askpass::askpass()
  }
  check_string(password, call = call)
  password
}

#' Rate limit a request by automatically adding a delay
#'
#' @description
#' Use `req_throttle()` to ensure that repeated calls to [req_perform()] never
#' exceed a specified rate.
#'
#' Throttling is implemented using a "token bucket", which steadily fills up to
#' a maximum of `capacity` tokens over `fill_time_s`. Each time you make a
#' request, it takes a token out of the bucket, and if the bucket is empty,
#' the request will wait until the bucket refills. This ensures that you never
#' make more than `capacity` requests in `fill_time_s`, but you can make
#' requests more quickly if the bucket is full. For example, if you have
#' `capacity = 10` and `fill_time_s = 60`, you can make 10 requests
#' without waiting, but the next request will wait 60 seconds. This gives the
#' same average throttling rate as the previous approach, but gives you much
#' better performance if you're only making a small number of requests.
#'
#' Some APIs enforce multiple rate limits simultaneously, e.g. no more than
#' 4 requests per second *and* no more than 200 requests per hour. You can
#' handle this by supplying a vector to `capacity` and `fill_time_s`: this
#' creates one token bucket per limit, and each request must satisfy all of
#' them. This lets you make quick bursts of requests while still respecting
#' longer term limits.
#'
#' @inheritParams req_perform
#' @param capacity The size of the bucket, i.e. the maximum number of
#'   tokens that can accumulate. To enforce multiple rate limits at once,
#'   supply a vector of capacities (one per limit); `capacity` and
#'   `fill_time_s` are recycled to a common length.
#' @param rate For backwards compatibility, you can still specify the `rate`,
#'   which is converted to `capacity` by multiplying by `fill_time_s`.
#'   However, we recommend using `capacity` and `fill_time_s` as it gives more
#'   control.
#' @param fill_time_s Time in seconds to fill the capacity. Defaults to 60s.
#' @param realm A string that uniquely identifies the throttle pool to use
#'   (throttling limits always apply *per pool*). If not supplied, defaults
#'   to the hostname of the request.
#' @returns A modified HTTP [request].
#' @seealso [req_retry()] for another way of handling rate-limited APIs.
#' @export
#' @examples
#' # Ensure we never send more than 30 requests a minute
#' req <- request(example_url()) |>
#'   req_throttle(capacity = 30, fill_time_s = 60)
#'
#' resp <- req_perform(req)
#' throttle_status()
#' resp <- req_perform(req)
#' throttle_status()
#' \dontshow{httr2:::throttle_reset()}
#'
#' # Enforce multiple limits at once: no more than 10 requests every 1s
#' # and no more than 100 requests every 60s
#' req <- request(example_url()) |>
#'   req_throttle(capacity = c(10, 100), fill_time_s = c(1, 60))
#' resp <- req_perform(req)
#' throttle_status()
#' \dontshow{httr2:::throttle_reset()}
req_throttle <- function(req, rate, capacity, fill_time_s = 60, realm = NULL) {
  check_request(req)
  check_exclusive(rate, capacity)
  check_throttle_number(fill_time_s)
  check_string(realm, allow_null = TRUE)

  if (missing(capacity)) {
    check_throttle_number(rate)
    args <- vctrs::vec_recycle_common(rate = rate, fill_time_s = fill_time_s)
    capacity <- args$rate * args$fill_time_s
  } else {
    check_throttle_number(capacity, whole = TRUE)
    args <- vctrs::vec_recycle_common(
      capacity = capacity,
      fill_time_s = fill_time_s
    )
    capacity <- args$capacity
  }
  fill_rate <- capacity / args$fill_time_s

  realm <- realm %||% url_parse(req$url)$hostname
  if (!throttle_exists(realm, capacity, fill_rate)) {
    the$throttle[[realm]] <- map2(
      capacity,
      fill_rate,
      function(capacity, fill_rate) TokenBucket$new(capacity, fill_rate)
    )
  }
  req_policies(req, throttle_realm = realm)
}

# Validates a non-negative numeric vector of throttle parameters.
check_throttle_number <- function(
  x,
  whole = FALSE,
  arg = caller_arg(x),
  call = caller_env()
) {
  if (missing(x) || !is.numeric(x) || length(x) == 0 || anyNA(x)) {
    stop_input_type(x, "a numeric vector", arg = arg, call = call)
  }
  if (any(x < 0)) {
    cli::cli_abort(
      "Every element of {.arg {arg}} must be 0 or greater.",
      call = call
    )
  }
  if (whole && any(x != trunc(x))) {
    cli::cli_abort(
      "Every element of {.arg {arg}} must be a whole number.",
      call = call
    )
  }
  invisible()
}

#' Display internal throttle status
#'
#' Sometimes useful for debugging.
#'
#' @return A data frame with one row per token bucket and four columns:
#'   * The `realm`.
#'   * The `capacity` of the bucket.
#'   * Number of `tokens` remaining in the bucket.
#'   * Time `to_wait` in seconds for next token.
#' @export
#' @keywords internal
throttle_status <- function() {
  # Trigger refill before displaying status
  walk(the$throttle, function(buckets) walk(buckets, function(b) b$refill()))

  realm <- character()
  capacity <- tokens <- to_wait <- double()
  # Just for debugging so focus on simplicity rather than efficiency
  for (r in sort(env_names(the$throttle))) {
    buckets <- the$throttle[[r]]
    realm <- c(realm, rep(r, length(buckets)))
    capacity <- c(capacity, map_dbl(buckets, function(b) b$capacity))
    tokens <- c(tokens, floor(map_dbl(buckets, function(b) b$tokens)))
    to_wait <- c(to_wait, map_dbl(buckets, function(b) b$token_wait_time()))
  }

  data.frame(
    realm = realm,
    capacity = capacity,
    tokens = tokens,
    to_wait = to_wait,
    row.names = NULL,
    check.names = FALSE
  )
}

throttle_reset <- function(realm = NULL) {
  if (is.null(realm)) {
    the$throttle <- new_environment()
  } else {
    env_unbind(the$throttle, realm)
  }

  invisible()
}

throttle_exists <- function(realm, capacity, fill_rate) {
  if (!env_has(the$throttle, realm)) {
    return(FALSE)
  }

  buckets <- the$throttle[[realm]]
  length(buckets) == length(capacity) &&
    all(map_dbl(buckets, \(b) b$capacity) == capacity) &&
    all(map_dbl(buckets, \(b) b$fill_rate) == fill_rate)
}

throttle_delay <- function(req) {
  if (!req_policy_exists(req, "throttle_realm")) {
    0
  } else {
    # Each request takes a token from every bucket, but only needs to wait
    # for the slowest one to refill.
    buckets <- the$throttle[[req$policies$throttle_realm]]
    max(map_dbl(buckets, \(b) b$take_token()))
  }
}
throttle_deadline <- function(req) {
  unix_time() + throttle_delay(req)
}
throttle_return_token <- function(req) {
  buckets <- the$throttle[[req$policies$throttle_realm]]
  walk(buckets, \(b) b$return_token())
}

TokenBucket <- R6::R6Class(
  "TokenBucket",
  public = list(
    capacity = NULL,
    fill_rate = NULL,

    last_fill = NULL,
    tokens = NULL,

    initialize = function(capacity, fill_rate) {
      self$capacity <- capacity
      self$tokens <- capacity
      self$fill_rate <- fill_rate

      self$last_fill <- unix_time()
    },

    refill = function() {
      now <- unix_time()
      # Ensure if we call rapidly we don't accumulate FP errors
      if (now - self$last_fill < 1e-6) {
        return(self$tokens)
      }
      new_tokens <- (now - self$last_fill) * self$fill_rate

      self$tokens <- min(self$capacity, self$tokens + new_tokens)
      self$last_fill <- now

      self$tokens
    },

    token_wait_time = function() {
      if (self$tokens >= 1) {
        0
      } else {
        (1 - self$tokens) / self$fill_rate
      }
    },

    # Returns the number of seconds that you need to wait to get it
    # Might cause tokens to drop below 0 temporarily so if you don't end up
    # waiting this long, you need to return the token
    take_token = function() {
      self$refill()
      wait <- self$token_wait_time()
      self$tokens <- self$tokens - 1
      wait
    },

    return_token = function() {
      self$tokens <- min(self$tokens + 1, self$capacity)
    }
  )
)

test_that("throttling affects request performance", {
  skip_on_cran()
  on.exit(throttle_reset())
  local_mocked_bindings(unix_time = function() 0)

  req <- request_test() |> req_throttle(capacity = 4, fill_time_s = 1)
  . <- replicate(4, req_perform(req))

  local_mocked_bindings(unix_time = function() 0.1)
  expect_snapshot(time <- system.time(req_perform(req))[[3]])
  expect_gte(time, 1 / 4 - 0.1 - 1e-6)
})

test_that("first request isn't throttled", {
  on.exit(throttle_reset())

  mock_time <- 0
  local_mocked_bindings(unix_time = function() mock_time)

  req <- request_test() |> req_throttle(rate = 1, fill_time_s = 1)
  expect_equal(throttle_delay(req), 0)

  mock_time <- 0.1
  expect_equal(throttle_delay(req), 0.9)

  mock_time <- 1.5
  expect_equal(throttle_delay(req), 0.5)
})

test_that("throttle only reset if parameters change", {
  on.exit(throttle_reset())

  mock_time <- 0
  local_mocked_bindings(unix_time = function() mock_time)

  req <- request_test() |> req_throttle(capacity = 1, fill_time_s = 1)
  expect_equal(throttle_delay(req), 0)

  mock_time <- 0.5
  expect_equal(throttle_delay(req), 0.5)
  mock_time <- 1

  req <- request_test() |> req_throttle(capacity = 1, fill_time_s = 1)
  expect_equal(throttle_delay(req), 1)

  # reset: capacity changed
  req <- request_test() |> req_throttle(capacity = 2, fill_time_s = 1)
  expect_equal(throttle_delay(req), 0)

  # reset: fill time changed
  req <- request_test() |> req_throttle(capacity = 2, fill_time_s = 2)
  expect_equal(throttle_delay(req), 0)
})


test_that("realm defaults to hostname but can be overridden", {
  on.exit(throttle_reset())

  expect_named(the$throttle, character())

  request_test() |> req_throttle(100 / 1)
  expect_named(the$throttle, "127.0.0.1")

  throttle_reset()
  request_test() |> req_throttle(100 / 1, realm = "custom")
  expect_named(the$throttle, "custom")
})

# token bucket ----------------------------------------------------------------

test_that("token bucket respects capacity limits", {
  mock_time <- 0
  local_mocked_bindings(unix_time = function() mock_time)

  bucket <- TokenBucket$new(capacity = 2, fill_rate = 2)
  expect_equal(bucket$take_token(), 0)
  expect_equal(bucket$tokens, 1)
  expect_equal(bucket$take_token(), 0)
  expect_equal(bucket$tokens, 0)

  expect_equal(bucket$take_token(), 0.5)
  mock_time <- 0.5
  expect_equal(bucket$take_token(), 0.5)
})

test_that("token bucket handles fractions correctly", {
  mock_time <- 0
  local_mocked_bindings(unix_time = function() mock_time)

  bucket <- TokenBucket$new(capacity = 2, fill_rate = 2)
  bucket$tokens <- 0
  expect_equal(bucket$take_token(), 0.5)
  expect_equal(bucket$tokens, -1)
  mock_time <- 0.5
  expect_equal(bucket$refill(), 0)

  bucket$last_fill <- 0
  bucket$tokens <- 0
  mock_time <- 0.4
  expect_equal(bucket$refill(), 0.80)
  expect_equal(bucket$take_token(), 0.1)
  mock_time <- mock_time + 0.1
  expect_equal(bucket$refill(), 0)
})

test_that("never returns negative time", {
  mock_time <- 0
  local_mocked_bindings(unix_time = function() mock_time)

  bucket <- TokenBucket$new(capacity = 5, fill_rate = 1)
  mock_time <- 5

  # first five are free
  replicate(5, expect_equal(bucket$take_token(), 0))

  # we get another 5 after 5 seconds
  mock_time <- mock_time + 5
  replicate(5, expect_equal(bucket$take_token(), 0))

  # if we only wait a second, we only get one
  mock_time <- mock_time + 1
  expect_equal(bucket$take_token(), 0)
  expect_equal(bucket$take_token(), 1)
  expect_equal(bucket$take_token(), 2)
})

#' Tools for working with lists of responses
#'
#' @description
#' These functions provide a basic toolkit for working with lists of
#' responses and possibly errors, as returned by [req_perform_parallel()],
#' [req_perform_sequential()], and [req_perform_iterative()].
#'
#' * `resps_successes()` returns a list of successful responses.
#' * `resps_failures()` returns a list of failed responses (i.e. errors).
#' * `resps_ok()` returns a logical vector indicating which requests were
#'   successful.
#' * `resps_requests()` returns the list of requests that corresponds to
#'   each response.
#' * `resps_data()` returns all the data in a single vector or data frame.
#'
#' @export
#' @param resps A list of responses (possibly including errors).
#' @param resp_data A function that takes a response (`resp`) and
#'   returns the data found inside that response as a vector or data frame.
#'
#'   NB: If you're using [resp_body_raw()], you're likely to want to wrap its
#'   output in `list()` to avoid combining all the bodies into a single raw
#'   vector, e.g. `resps |> resps_data(\(resp) list(resp_body_raw(resp)))`.
#'
#' @examples
#' reqs <- list(
#'   request(example_url()) |> req_url_path("/ip"),
#'   request(example_url()) |> req_url_path("/user-agent"),
#'   request(example_url()) |> req_template("/status/:status", status = 404),
#'   request("INVALID")
#' )
#' resps <- req_perform_parallel(reqs, on_error = "continue")
#'
#' # find successful responses
#' resps |> resps_successes()
#'
#' # collect all their data
#' resps |>
#'   resps_successes() |>
#'   resps_data(\(resp) resp_body_json(resp))
#'
#' # find requests corresponding to failure responses
#' resps |>
#'   resps_failures() |>
#'   resps_requests()
resps_successes <- function(resps) {
  resps[resps_ok(resps)]
}

#' @export
#' @rdname resps_successes
resps_failures <- function(resps) {
  resps[!resps_ok(resps)]
}

#' @export
#' @rdname resps_successes
resps_ok <- function(resps) {
  vapply(resps, is_response, logical(1))
}

#' @export
#' @rdname resps_successes
resps_requests <- function(resps) {
  lapply(resps, function(x) x$request)
}

#' @export
#' @rdname resps_successes
resps_data <- function(resps, resp_data) {
  check_installed("vctrs")
  check_function2(resp_data, "resp")

  vctrs::list_unchop(lapply(resps, resp_data))
}

test_that("basic helpers work", {
  reqs <- list(
    request_test("/status/:status", status = 200),
    request_test("/status/:status", status = 404),
    request("INVALID")
  )
  resps <- req_perform_parallel(reqs, on_error = "continue")

  expect_equal(resps_successes(resps), resps[1])
  expect_equal(resps_failures(resps), resps[2:3])
  expect_equal(resps_requests(resps), reqs)
})

test_that("can extract all data", {
  resps <- list(
    response_json(body = list(data = 1)),
    response_json(body = list(data = 2)),
    response_json(body = list(data = 3))
  )

  expect_equal(
    resps_data(resps, function(resp) resp_body_json(resp)$data),
    1:3
  )
})

req_auth_sign <- function(req, fun, params, cache) {
  req_policies(
    req,
    auth_sign = list(
      fun = fun,
      params = params,
      cache = cache
    )
  )
}
auth_sign <- function(req) {
  if (!req_policy_exists(req, "auth_sign")) {
    return(req)
  }

  exec(
    req$policies$auth_sign$fun,
    req = req,
    cache = req$policies$auth_sign$cache,
    !!!req$policies$auth_sign$params
  )
}

req_auth_clear_cache <- function(req) {
  cache <- req$policies$auth_sign$cache
  if (!is.null(cache)) {
    cache$clear()
    TRUE
  } else {
    FALSE
  }
}

request_test <- function(template = "/get", ...) {
  req <- request(example_url())
  req <- req_template(req, template, ..., .env = caller_env())
  req
}

#' Code for examples
#'
#' @description
#' `example_url()` runs a simple websever using the webfakes package with the
#' following endpoints:
#'
#' * all the ones from the [webfakes::httpbin_app()]
#' * `/iris`: paginate through the iris dataset. It has the query parameters
#'   `page` and `limit` to control the pagination.
#'
#' `example_github_client()` is an OAuth client for GitHub.
#'
#' @keywords internal
#' @export
example_url <- function(path = "/") {
  check_installed("webfakes")
  if (is_testing() && !interactive()) {
    testthat::skip_on_covr()
  }
  env_cache(the, "test_app", example_app())
  the$test_app$url(path)
}

example_app <- function() {
  app <- webfakes::httpbin_app()
  # paginated iris endpoint
  app$get("/iris", function(req, res) {
    page <- req$query$page %||% 1L
    page <- as.integer(page)

    page_size <- req$query$limit %||% 20L
    page_size <- as.integer(page_size)

    n <- nrow(datasets::iris)
    start <- (page - 1L) * page_size + 1L
    end <- start + page_size - 1L
    ids <- rlang::seq2(start, end)
    data <- vctrs::vec_slice(datasets::iris, intersect(ids, seq_len(n)))

    res$set_status(200L)$send_json(
      object = list(data = data, count = n, pages = ceiling(n / page_size)),
      auto_unbox = TRUE,
      pretty = TRUE
    )
  })

  webfakes::new_app_process(
    app,
    opts = webfakes::server_opts(num_threads = 20, enable_keep_alive = TRUE)
  )
}

#' @export
#' @rdname example_url
example_github_client <- function() {
  # <https://github.com/settings/applications/1636322>
  oauth_client(
    id = "28acfec0674bb3da9f38",
    secret = obfuscated(paste0(
      "J9iiGmyelHltyxqrHXW41ZZPZamyUNxSX1_uKnv",
      "PeinhhxET_7FfUs2X0LLKotXY2bpgOMoHRCo"
    )),
    token_url = "https://github.com/login/oauth/access_token",
    name = "hadley-oauth-test"
  )
}

test_that("can perform multiple requests", {
  req <- request(example_url()) %>%
    req_url_path("/iris") %>%
    req_url_query(limit = 5)

  resps <- req_perform_iterative(
    req,
    next_req = iterate_with_offset("page_index"),
    max_reqs = 4
  )

  expect_length(resps, 4)
  expect_equal(resp_url(resps[[4]]), paste0(example_url(), "iris?limit=5&page_index=4"))
})

test_that("can save results to disk", {
  req <- request(example_url()) %>%
    req_url_path("/iris") %>%
    req_url_query(limit = 5)

  dir <- withr::local_tempdir()

  resps <- req_perform_iterative(
    req,
    next_req = iterate_with_offset("page_index"),
    max_reqs = 2,
    path = paste0(dir, "/file-{i}")
  )

  expect_equal(resps[[1]]$body, new_path(file.path(dir, "file-1")))
  expect_equal(resps[[2]]$body, new_path(file.path(dir, "file-2")))
})

test_that("user temination still returns data", {
  req <- request(example_url()) %>%
    req_url_path("/iris") %>%
    req_url_query(limit = 5)
  next_req <- function(resp, req) interrupt()

  expect_snapshot(
    resps <- req_perform_iterative(req, next_req = next_req)
  )
  expect_length(resps, 1)
})


test_that("can retrieve all pages", {
  req <- request(example_url()) %>%
    req_url_path("/iris") %>%
    req_url_query(limit = 1)

  i <- 1
  next_req <- function(resp, req) {
    i <<- i + 1
    if (i <= 120) {
      req %>% req_url_query(page_index = 1)
    }
  }
  expect_condition(
    resps <- req_perform_iterative(req, next_req = next_req, max_reqs = Inf),
    class = "httr2:::doubled"
  )
  expect_length(resps, 120)
})

test_that("can choose to return on failure", {
  iterator <- function(resp, req) {
    request_test("/status/:status", status = 404)
  }
  expect_error(
    req_perform_iterative(request_test(), iterator),
    class = "httr2_http_404"
  )

  out <- req_perform_iterative(request_test(), iterator, on_error = "return")
  expect_length(out, 2)
  expect_s3_class(out[[1]], "httr2_response")
  expect_s3_class(out[[2]], "httr2_http_404")
})

test_that("checks its inputs", {
  req <- request_test()
  expect_snapshot(error = TRUE,{
    req_perform_iterative(1)
    req_perform_iterative(req, function(x, y) x + y)
    req_perform_iterative(req, function(resp, req) {}, path = 1)
    req_perform_iterative(req, function(resp, req) {}, max_reqs = -1)
    req_perform_iterative(req, function(resp, req) {}, progress = -1)
  })
})

test_that("can read/write cookies", {
  cookie_path <- withr::local_tempfile()

  set_cookie <- function(req, name, value) {
    request_test("/cookies/set/:name/:value", name = name, value = value) |>
      req_cookie_preserve(cookie_path) |>
      req_perform()
  }
  set_cookie(req, "x", "a")
  set_cookie(req, "y", "b")
  set_cookie(req, "z", "c")

  expect_snapshot(readLines(cookie_path)[-(1:4)])

  json <- request_test("/cookies") |>
    req_cookie_preserve(cookie_path) |>
    req_perform() |>
    resp_body_json()
  expect_mapequal(json$cookies, list(x = "a", y = "b", z = "c"))
})

test_that("can set cookies", {
  resp <- request(example_url("/cookies")) |>
    req_cookies_set(a = 1, b = 1) |>
    req_perform()

  expect_equal(resp_body_json(resp), list(cookies = list(a = "1", b = "1")))
})

test_that("cookie values are usually escaped", {
  resp <- request(example_url("/cookies")) |>
    req_cookies_set(a = I("%20"), b = "%") |>
    req_perform()

  expect_equal(resp_body_json(resp), list(cookies = list(a = "%20", b = "%25")))
})

test_that("checks its inputs", {
  req <- request("http://example.com")
  expect_snapshot(error = TRUE, {
    req_perform_sequential(req)
    req_perform_sequential(list(req), letters)
  })
})

test_that("can download files", {
  reqs <- list(request_test("/json"), request_test("/html"))
  paths <- c(withr::local_tempfile(), withr::local_tempfile())
  resps <- req_perform_sequential(reqs, paths)

  expect_equal(resps[[1]]$body, new_path(paths[[1]]))
  expect_equal(resps[[2]]$body, new_path(paths[[2]]))

  # And check that something was downloaded
  expect_gt(file.size(paths[[1]]), 0)
  expect_gt(file.size(paths[[2]]), 0)
})

test_that("on_error = 'return' returns error", {
  reqs <- list2(
    request_test("/status/:status", status = 200),
    request_test("/status/:status", status = 200),
    request_test("/status/:status", status = 404),
    request_test("/status/:status", status = 200)
  )
  out <- req_perform_sequential(reqs, on_error = "return")
  expect_length(out, 4)
  expect_s3_class(out[[3]], "httr2_http_404")
  expect_equal(out[[4]], NULL)
})


test_that("on_error = 'continue' captures both error types", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request("INVALID"),
  )
  out <- req_perform_sequential(reqs, on_error = "continue")
  expect_s3_class(out[[1]], "httr2_http_404")
  expect_s3_class(out[[2]], "httr2_failure")
})

test_that("on_error = 'return' returns error", {
  reqs <- list2(
    request_test("/status/:status", status = 200),
    request_test("/status/:status", status = 200),
    request_test("/status/:status", status = 404),
    request_test("/status/:status", status = 200)
  )
  out <- req_perform_sequential(reqs, on_error = "return")
  expect_length(out, 4)
  expect_s3_class(out[[3]], "httr2_http_404")
  expect_equal(out[[4]], NULL)
})

test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    if (req$url == "https://ok") {
      response()
    } else {
      response(404)
    }
  })

  resps <- req_perform_sequential(list(req_200, req_404), on_error = "continue")
  expect_equal(resps[[1]], response())
  expect_s3_class(resps[[2]], "httr2_http_404")
})

test_that("can perform multiple requests", {
  req <- request(example_url("/iris")) |>
    req_url_query(limit = 5)

  resps <- req_perform_iterative(
    req,
    next_req = iterate_with_offset("page_index"),
    max_reqs = 4
  )

  expect_length(resps, 4)
  expect_equal(
    resp_url(resps[[4]]),
    paste0(example_url(), "iris?limit=5&page_index=4")
  )
})

test_that("can save results to disk", {
  req <- request(example_url("/iris")) |>
    req_url_query(limit = 5)

  dir <- withr::local_tempdir()

  resps <- req_perform_iterative(
    req,
    next_req = iterate_with_offset("page_index"),
    max_reqs = 2,
    path = paste0(dir, "/file-{i}")
  )

  expect_equal(resps[[1]]$body, new_path(file.path(dir, "file-1")))
  expect_equal(resps[[2]]$body, new_path(file.path(dir, "file-2")))
})

test_that("user temination still returns data", {
  req <- request(example_url("/iris")) |>
    req_url_query(limit = 5)
  next_req <- function(resp, req) interrupt()

  expect_snapshot(
    resps <- req_perform_iterative(req, next_req = next_req)
  )
  expect_length(resps, 1)
})


test_that("can retrieve all pages", {
  req <- request(example_url("/iris")) |>
    req_url_query(limit = 1)

  i <- 1
  next_req <- function(resp, req) {
    i <<- i + 1
    if (i <= 120) {
      req |> req_url_query(page_index = 1)
    }
  }
  expect_condition(
    resps <- req_perform_iterative(req, next_req = next_req, max_reqs = Inf),
    class = "httr2:::doubled"
  )
  expect_length(resps, 120)
})

test_that("can choose to return on failure", {
  iterator <- function(resp, req) {
    request_test("/status/:status", status = 404)
  }
  expect_error(
    req_perform_iterative(request_test(), iterator),
    class = "httr2_http_404"
  )

  out <- req_perform_iterative(request_test(), iterator, on_error = "return")
  expect_length(out, 2)
  expect_s3_class(out[[1]], "httr2_response")
  expect_s3_class(out[[2]], "httr2_http_404")
})


test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    if (req$url == "https://ok") {
      response()
    } else {
      response(404)
    }
  })

  resps <- req_perform_iterative(
    req_200,
    \(resp, req) req_404,
    on_error = "return"
  )
  expect_equal(resps[[1]], response())
  expect_s3_class(resps[[2]], "httr2_http_404")
})


test_that("checks its inputs", {
  req <- request_test()
  expect_snapshot(error = TRUE, {
    req_perform_iterative(1)
    req_perform_iterative(req, function(x, y) x + y)
    req_perform_iterative(
      req,
      function(resp, req) {},
      path = 1
    )
    req_perform_iterative(
      req,
      function(resp, req) {},
      max_reqs = -1
    )
    req_perform_iterative(
      req,
      function(resp, req) {},
      progress = -1
    )
  })
})

#' Get URL/components from the response
#'
#' * `resp_url()` returns the complete url.
#' * `resp_url_path()` returns the path component.
#' * `resp_url_query()` returns a single query component.
#' * `resp_url_queries()` returns the query component as a named list.
#'
#' @inheritParams resp_header
#' @export
#' @examples
#' resp <- request(example_url()) |>
#'   req_url_path("/get") |>
#'   req_url_query(hello = "world") |>
#'   req_perform()
#'
#' resp |> resp_url()
#' resp |> resp_url_path()
#' resp |> resp_url_queries()
#' resp |> resp_url_query("hello")
resp_url <- function(resp) {
  check_response(resp)

  resp$url
}

#' @export
#' @rdname resp_url
resp_url_path <- function(resp) {
  check_response(resp)

  url_parse(resp$url)$path
}

#' @export
#' @rdname resp_url
#' @param name Query parameter name.
#' @param default Default value to use if query parameter doesn't exist.
resp_url_query <- function(resp, name, default = NULL) {
  check_response(resp)

  resp_url_queries(resp)[[name]] %||% default
}

#' @export
#' @rdname resp_url
resp_url_queries <- function(resp) {
  check_response(resp)

  url_parse(resp$url)$query
}

test_that("can extract url components from a response", {
  req <- request_test("/get") |> req_url_query(a = "1", b = "2")
  resp <- req_perform(req)

  expect_equal(resp_url(resp), example_url("/get?a=1&b=2"))
  expect_equal(resp_url_path(resp), "/get")
  expect_equal(resp_url_queries(resp), list(a = "1", b = "2"))

  expect_equal(resp_url_query(resp, "a"), "1")
  expect_equal(resp_url_query(resp, "c"), NULL)
  expect_equal(resp_url_query(resp, "c", "x"), "x")
})

#' Show extra output when request is performed
#'
#' @description
#' `req_verbose()` uses the following prefixes to distinguish between
#' different components of the HTTP requests and responses:
#'
#' * `* ` informative curl messages
#' * `->` request headers
#' * `>>` request body
#' * `<-` response headers
#' * `<<` response body
#'
#' @inheritParams req_perform
#' @param header_req,header_resp Show request/response headers?
#' @param body_req,body_resp Should request/response bodies? When the response
#'   body is compressed, this will show the number of bytes received in
#'   each "chunk".
#' @param info Show informational text from curl? This is mainly useful
#'   for debugging https and auth problems, so is disabled by default.
#' @param redact_headers Redact confidential data in the headers? Currently
#'   redacts the contents of the Authorization header to prevent you from
#'   accidentally leaking credentials when debugging/reprexing.
#' @seealso [req_perform()] which exposes a limited subset of these options
#'   through the `verbosity` argument and [with_verbosity()] which allows you
#'   to control the verbosity of requests deeper within the call stack.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # Use `req_verbose()` to see the headers that are sent back and forth when
#' # making a request
#' resp <- request("https://httr2.r-lib.org") |>
#'   req_verbose() |>
#'   req_perform()
#'
#' # Or use one of the convenient shortcuts:
#' resp <- request("https://httr2.r-lib.org") |>
#'   req_perform(verbosity = 1)
req_verbose <- function(
  req,
  header_req = TRUE,
  header_resp = TRUE,
  body_req = FALSE,
  body_resp = FALSE,
  info = FALSE,
  redact_headers = TRUE
) {
  check_request(req)

  # force all arguments
  list(header_req, header_resp, body_req, body_resp, info, redact_headers)

  debug <- function(type, msg) {
    # Set in req_prepare()
    headers <- req$state$headers

    if (info && type == 0) {
      verbose_info("*  ", msg)
    } else if (header_resp && type == 1) {
      verbose_header("<- ", msg)
    } else if (header_req && type == 2) {
      to_redact <- which_redacted(headers)
      verbose_header("-> ", msg, redact_headers, to_redact = to_redact)
    } else if (body_resp && type == 3) {
      # handled in handle_resp()
    } else if (body_req && type == 4) {
      verbose_body(">> ", msg, headers$`content-type`)
    }
  }
  req <- req_options(req, debugfunction = debug, verbose = TRUE)
  req <- req_policies(req, show_body = body_resp)
  req
}

# helpers -----------------------------------------------------------------

verbose_info <- function(prefix, x) {
  x <- readBin(x, character())
  lines <- unlist(strsplit(x, "\r?\n", useBytes = TRUE))

  cli::cat_line(prefix, lines)
}

verbose_header <- function(prefix, x, redact = TRUE, to_redact = NULL) {
  x <- readBin(x, character())
  lines <- unlist(strsplit(x, "\r?\n", useBytes = TRUE))

  for (line in lines) {
    if (grepl("^[-a-zA-z0-9]+:", line)) {
      headers <- as_headers(line, to_redact, lifespan = current_env())
      header <- headers_flatten(headers, redact)
      cli::cat_line(
        prefix,
        cli::style_bold(names(header)),
        ": ",
        format(header[[1]])
      )
    } else {
      cli::cat_line(prefix, line)
    }
  }
}

verbose_body <- function(prefix, x, content_type) {
  show_body(
    x,
    content_type,
    prefix = prefix,
    pretty_json = getOption("httr2_pretty_json", TRUE)
  )
}

# Testing helpers -------------------------------------------------------------

req_verbose_test <- function(req) {
  # Reset all headers that otherwise might vary
  req <- req_headers(
    req,
    `Accept-Encoding` = "",
    Accept = "",
    Host = "http://example.com",
    `User-Agent` = ""
  )
  req <- req_options(req, forbid_reuse = TRUE)
  req
}

transform_verbose_response <- function(lines) {
  lines <- gsub(example_url(), "<webfakes>/", lines, fixed = TRUE)
  lines <- lines[!grepl("^<- (Date|ETag|Content-Length):", lines)]
  lines <- lines[!grepl("\\*  Closing connection", lines)]
  lines
}

test_that("can request verbose record of request", {
  skip_on_ci()
  skip_on_cran()

  req <- local_app_request(method = "post", function(req, res) {
    res$send_json(list(x = 1), auto_unbox = TRUE)
  })
  req <- req |>
    req_body_raw("This is some text", "text/plain") |>
    req_verbose_test()

  # Snapshot test of response
  verbose_resp <- req |>
    req_verbose(header_resp = TRUE, body_resp = TRUE, header_req = FALSE)
  expect_snapshot(
    . <- req_perform(verbose_resp),
    transform = transform_verbose_response
  )

  # Snapshot test of request
  verbose_req <- req |>
    req_verbose(header_req = TRUE, body_req = TRUE, header_resp = FALSE)
  expect_snapshot(. <- req_perform(verbose_req))

  # Lightweight test for everything else
  verbose_info <- req |>
    req_verbose(info = TRUE, header_req = FALSE, header_resp = FALSE)
  expect_output(. <- req_perform(verbose_info))
})

test_that("redacts headers as needed", {
  skip_on_ci()
  skip_on_cran()

  req <- request(example_url()) |>
    req_verbose_test() |>
    req_verbose(header_req = TRUE, header_resp = FALSE) |>
    req_headers_redacted(Authorization = "abc")

  expect_snapshot(. <- req_perform(req))
})

test_that("can display compressed bodies", {
  skip_on_ci()
  skip_on_cran()

  req <- request(example_url()) |>
    req_url_path("gzip") |>
    req_verbose_test() |>
    req_verbose(header_req = FALSE, header_resp = TRUE, body_resp = TRUE)

  expect_snapshot(. <- req_perform(req), transform = transform_verbose_response)
})

test_that("response json is automatically prettified", {
  skip_on_ci()
  skip_on_cran()

  req <- local_app_request(function(req, res) {
    res$set_header("Content-Type", "application/json")
    res$send('{"foo":"bar","baz":[1,2,3]}')
  })

  req <- req |>
    req_verbose_test() |>
    req_verbose(body_resp = TRUE, header_resp = FALSE, header_req = FALSE)
  expect_snapshot(. <- req_perform(req))

  # Unless we opt-out
  local_options(httr2_pretty_json = FALSE)
  expect_snapshot(. <- req_perform(req))
})

test_that("request json is automatically prettified", {
  skip_on_ci()
  skip_on_cran()

  req <- request(example_url("/post")) |>
    req_verbose_test() |>
    req_body_json(list(foo = "bar", baz = c(1, 2, 3))) |>
    req_verbose(body_req = TRUE, header_resp = FALSE, header_req = FALSE)
  expect_snapshot(. <- req_perform(req))

  # Unless we opt-out
  local_options(httr2_pretty_json = FALSE)
  expect_snapshot(. <- req_perform(req))
})

local_app_request <- function(fun, method = "get", frame = parent.frame()) {
  # sometimes fails on CRAN and we don't need the hassle
  skip_on_cran()
  # Works interactively (useful for manaul coverage checking)
  # but not in separate process
  if (!interactive()) {
    skip_on_covr()
  }

  app <- webfakes::new_app()

  app$locals$i <- 0
  app$use(function(req, res) {
    app$locals$i <- app$locals$i + 1
    "next"
  })

  app[[method]]("/test", fun)
  app$locals$sync_rep <- sync_rep
  server <- webfakes::local_app_process(app, .local_envir = frame)

  req <- request(server$url("/test"))
  req <- req_error(req, body = function(resp) {
    if (resp_has_body(resp)) resp_body_string(resp)
  })
  req
}

#' Perform a request and return a streaming connection
#'
#' @description
#' Use `req_perform_connection()` to perform a request if you want to stream the
#' response body. A response returned by `req_perform_connection()` includes a
#' connection as the body. You can then use [resp_stream_raw()],
#' [resp_stream_lines()], or [resp_stream_sse()] to retrieve data a chunk at a
#' time. Always finish up by closing the connection by calling
#' `close(response)`.
#'
#' This is an alternative interface to [req_perform_stream()] that returns a
#' [connection][base::connections] that you can use to pull the data, rather
#' than providing callbacks that the data is pushed to. This is useful if you
#' want to do other work in between handling inputs from the stream.
#'
#' @inheritParams req_perform
#' @param blocking When retrieving data, should the connection block and wait
#'   for the desired information or immediately return what it has (possibly
#'   nothing)?
#' @param verbosity How much information to print? This is a wrapper
#'   around [req_verbose()] that uses an integer to control verbosity:
#'
#'   * `0`: no output
#'   * `1`: show headers
#'   * `2`: show headers and bodies as they're streamed
#'   * `3`: show headers, bodies, curl status messages, raw SSEs, and stream
#'     buffer management
#'
#'   Use [with_verbosity()] to control the verbosity of requests that
#'   you can't affect directly.
#' @export
#' @examples
#' req <- request(example_url()) |>
#'   req_url_path("/stream-bytes/32768")
#' resp <- req_perform_connection(req)
#'
#' length(resp_stream_raw(resp, kb = 16))
#' length(resp_stream_raw(resp, kb = 16))
#' # When the stream has no more data, you'll get an empty result:
#' length(resp_stream_raw(resp, kb = 16))
#'
#' # Always close the response when you're done
#' close(resp)
#'
#' # You can loop until complete with resp_stream_is_complete()
#' resp <- req_perform_connection(req)
#' while (!resp_stream_is_complete(resp)) {
#'   print(length(resp_stream_raw(resp, kb = 12)))
#' }
#' close(resp)
req_perform_connection <- function(
  req,
  blocking = TRUE,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL)
) {
  check_request(req)
  check_bool(blocking)
  req <- req_verbosity_connection(req, verbosity %||% httr2_verbosity())
  req <- req_policies(req, connection = TRUE)

  if (!is.null(mock)) {
    mock <- as_function(mock)
    mock_resp <- mock(req)
    if (!is.null(mock_resp)) {
      return(handle_resp(req, mock_resp))
    }
  }

  req_prep <- req_prepare(req)
  handle <- req_handle(req_prep)
  the$last_request <- req
  the$last_response <- NULL

  tries <- 0
  delay <- 0
  max_tries <- retry_max_tries(req)
  deadline <- Sys.time() + retry_max_seconds(req)
  resp <- NULL
  while (tries < max_tries && Sys.time() < deadline) {
    retry_check_breaker(req, tries)
    sys_sleep(delay, "for retry backoff")

    if (is_response(resp)) {
      close(resp)
    }
    resp <- req_perform_connection1(
      req,
      req_prep,
      handle,
      blocking = blocking,
      resend_count = tries + 1L
    )

    if (retry_is_transient(req, resp)) {
      tries <- tries + 1

      delay <- retry_after(req, resp, tries)
      signal(class = "httr2_retry", tries = tries, delay = delay)
    } else {
      break
    }
  }
  req_completed(req_prep)

  if (!is_error(resp) && error_is_error(req, resp)) {
    # Read full body if there's an error
    resp$body <- resp$body$read_all()
    the$last_response <- resp
  }
  handle_resp(req, resp)

  resp
}

# Like req_verbosity() but we want to print the streaming body when it's
# requested not when curl actually receives it
req_verbosity_connection <- function(
  req,
  verbosity,
  error_call = caller_env()
) {
  if (!is_integerish(verbosity, n = 1) || verbosity < 0 || verbosity > 3) {
    cli::cli_abort("{.arg verbosity} must 0, 1, 2, or 3.", call = error_call)
  }

  req <- switch(
    verbosity + 1,
    req,
    req_verbose(req),
    req_verbose(req, body_req = TRUE),
    req_verbose(req, body_req = TRUE, info = TRUE)
  )
  if (verbosity > 1) {
    req <- req_policies(
      req,
      show_streaming_body = verbosity >= 2,
      show_streaming_buffer = verbosity >= 3
    )
  }
  req
}

req_perform_connection1 <- function(
  req,
  req_prep,
  handle,
  blocking = TRUE,
  resend_count = 0
) {
  the$last_request <- req
  the$last_response <- NULL
  signal(class = "httr2_perform_connection")
  # Note: we need to do this before we call handle_preflight() so that request
  # signing works correctly with the added headers.
  req_prep <- req_with_span(req_prep, resend_count = resend_count)
  handle_preflight(req_prep, handle)

  err <- capture_curl_error({
    conn <- curl::curl(req$url, handle = handle)
    # Must open the stream in order to initiate the connection
    suppressWarnings(open(conn, "rbf", blocking = blocking))
    body <- StreamingBody$new(conn)
  })
  if (is_error(err)) {
    req_record_span_status(req, err)
    close(conn)
    return(err)
  }

  curl_data <- curl::handle_data(handle)
  resp <- create_response(req, curl_data, body)
  req_record_span_status(req, resp)
  resp
}

# Make open mockable
open <- NULL

#' `StreamingBody` class
#'
#' This R6 class is used to represent the body of a streaming response.
#' When using this in mocked responses, you can either create a new instance
#' using your own connection or use a subclass for some other representation.
#' In either case, you will pass to the `body` argument of [new_response()].
#'
#' @export
StreamingBody <- R6::R6Class(
  "StreamingBody",
  public = list(
    #' @description Create a new object
    #' @param conn A connection, that is open and ready for reading.
    #'   `StreamingBody` will take care of closing it.`
    initialize = function(conn) {
      if (!inherits(conn, "connection")) {
        stop_input_type(conn, "a connection", call = caller_env())
      }
      private$conn <- conn
    },

    #' @description Read `n` bytes into a raw vector.
    #' @param n Number of bytes to read
    read = function(n) {
      readBin(private$conn, "raw", n)
    },

    #' @description Read all bytes and close the connection.
    #' @param buffer Buffer size, in bytes.
    read_all = function(buffer = 32 * 1024) {
      bytes <- raw()
      repeat {
        new <- self$read(buffer)
        if (length(new) == 0) {
          break
        }
        bytes <- c(bytes, new)
      }

      self$close()
      if (length(bytes) == 0) {
        NULL
      } else {
        bytes
      }
    },

    #' @description Is the connection still open?
    is_open = function() {
      isValid(private$conn)
    },

    #' @description Is the connection complete? (i.e. is there data remaining
    #'   to be read?)
    is_complete = function() {
      !isIncomplete(private$conn)
    },

    #' @description Get the active file descriptions and timeout from the
    #'   handle. Wrapper around [curl::multi_fdset()].
    get_fdset = function() {
      curl::multi_fdset(private$conn)
    },

    #' @description Close the connection
    close = function() {
      if (self$is_open()) {
        close(private$conn)
      }
    }
  ),
  private = list(
    conn = NULL
  )
)

as_headers <- function(
  x,
  redact = character(),
  lifespan,
  error_call = caller_env()
) {
  if (is.list(x)) {
    new_headers(
      x,
      redact = redact,
      lifespan = lifespan,
      error_call = error_call
    )
  } else if (is.character(x) || is.raw(x)) {
    parsed <- curl::parse_headers(x)
    valid <- parsed[grepl(":", parsed, fixed = TRUE)]
    halves <- parse_in_half(valid, ":")

    headers <- set_names(trimws(halves$right), halves$left)
    new_headers(
      as.list(headers),
      redact = redact,
      lifespan = lifespan,
      error_call = error_call
    )
  } else {
    cli::cli_abort(
      "{.arg headers} must be a list, character vector, or raw.",
      call = error_call
    )
  }
}

new_headers <- function(
  x,
  redact = character(),
  lifespan,
  error_call = caller_env()
) {
  if (!is_list(x)) {
    cli::cli_abort("{.arg x} must be a list.", call = error_call)
  }
  if (length(x) > 0 && !is_named(x)) {
    cli::cli_abort("All elements of {.arg x} must be named.", call = error_call)
  }
  for (i in seq_along(x)) {
    if (!is.atomic(x[[i]]) && !is_weakref(x[[i]])) {
      cli::cli_abort(
        c(
          "Each element of {.arg x} must be an atomic vector or a weakref.",
          i = "{.arg x[[{i}]]} is {obj_type_friendly(x[[i]])}."
        ),
        call = error_call
      )
    }
  }

  needs_redact <- !is_redacted(x) & (tolower(names(x)) %in% tolower(redact))
  x[needs_redact] <- lapply(x[needs_redact], \(x) new_weakref(lifespan, x))

  structure(x, class = "httr2_headers")
}

#' @export
print.httr2_headers <- function(x, ..., redact = TRUE) {
  cli::cat_line(cli::format_inline("{.cls {class(x)}}"))
  show_headers(x, redact = redact)
  invisible(x)
}

show_headers <- function(x, redact = TRUE) {
  if (length(x) > 0) {
    vals <- lapply(headers_flatten(x, redact), format)
    cli::cat_line(cli::style_bold(names(x)), ": ", vals)
  }
}

#' @export
str.httr2_headers <- function(object, ..., no.list = FALSE) {
  object <- unclass(headers_flatten(object))
  cat(" <httr2_headers>\n")
  utils::str(object, ..., no.list = TRUE)
}

#' @export
`[.httr2_headers` <- function(x, i, ...) {
  if (is.character(i)) {
    i <- match(tolower(i), tolower(names(x)))
  }

  new_headers(NextMethod())
}

#' @export
`[[.httr2_headers` <- function(x, i) {
  if (is.character(i)) {
    i <- match(tolower(i), tolower(names(x)))
  }
  NextMethod()
}

#' @export
"$.httr2_headers" <- function(x, name) {
  i <- match(tolower(name), tolower(names(x)))
  x[[i]]
}

is_redacted <- function(x) {
  map_lgl(x, is_weakref)
}
is_redacted_empty <- function(x) {
  map_lgl(x, \(x) is_weakref(x) && is.null(wref_value(x)))
}

which_redacted <- function(x) {
  names(x)[is_redacted(x)]
}

# Flatten headers object into a list suitable either for display (if redacted)
# or passing to curl (if not redacted).
headers_flatten <- function(x, redact = TRUE) {
  is_redacted <- is_redacted(x)

  out <- vector("list", length(x))
  names(out) <- names(x)

  # https://datatracker.ietf.org/doc/html/rfc7230#section-3.2.2
  out[!is_redacted] <- lapply(x[!is_redacted], paste, collapse = ",")

  if (redact) {
    out[is_redacted] <- list(redacted_sentinel())
  } else {
    out[is_redacted] <- lapply(x[is_redacted], wref_value)
    # also need to ensure redacted values are simple strings
    out[is_redacted] <- lapply(out[is_redacted], function(x) {
      if (!is.null(x)) paste(x, collapse = ",")
    })
    # need to strip serialized weakrefs that now yield NULL
    out <- compact(out)
  }

  out
}

test_that("as_headers parses character vector", {
  headers <- as_headers(c("x:1", "y:2", "a", "b:"))
  expect_equal(headers, new_headers(list(x = "1", y = "2", b = "")))
})

test_that("as_headers coerces list", {
  expect_equal(as_headers(list(x = 1)), new_headers(list(x = 1)))
})

test_that("as_headers errors on invalid types", {
  expect_snapshot(error = TRUE, as_headers(1))
})

test_that("has nice print method", {
  expect_snapshot({
    as_headers(c("X:1", "Y: 2", "Z:"))
    as_headers(list())
  })
})

test_that("print and str redact headers", {
  x <- new_headers(list(x = 1, y = 2), redact = "x", lifespan = current_env())
  expect_snapshot({
    print(x)
    str(x)
  })
})

test_that("subsetting is case insensitive", {
  x <- new_headers(list(x = 1))
  expect_equal(x$X, 1)
  expect_equal(x[["X"]], 1)
  expect_equal(x["X"], new_headers(list(x = 1)))
})

test_that("new_headers checks inputs", {
  expect_snapshot(error = TRUE, {
    new_headers(1)
    new_headers(list(1))
    new_headers(list(x = mean))
  })
})

test_that("can flatten repeated inputs", {
  expect_equal(headers_flatten(list()), list())
  expect_equal(headers_flatten(list(x = 1)), list(x = "1"))
  expect_equal(headers_flatten(list(x = 1:2)), list(x = "1,2"))
})

# redaction -------------------------------------------------------------------

test_that("redacted values can't be serialized", {
  path <- withr::local_tempfile()

  headers <- new_headers(
    list(a = "x", b = "y"),
    redact = "a",
    lifespan = current_env()
  )
  saveRDS(headers, path)

  loaded <- readRDS(path)
  expect_equal(headers_flatten(loaded), list(a = redacted_sentinel(), b = "y"))
  expect_equal(headers_flatten(loaded, FALSE), list(b = "y"))
})

test_that("serialized redacted value doesn't cause curl errors", {
  path <- withr::local_tempfile()

  req <- request_test()
  req <- req_auth_basic(req, "user", "password")
  saveRDS(req, path)

  req <- readRDS(path)
  expect_no_error(req_perform(req))
})

test_that("can unredacted values", {
  x <- new_headers(list(x = "x"), redact = "x", lifespan = current_env())
  expect_equal(headers_flatten(x, redact = TRUE), list(x = redacted_sentinel()))
  expect_equal(headers_flatten(x, redact = FALSE), list(x = "x"))
})

test_that("headers can't get double redacted", {
  x1 <- new_headers(list(x = "x"), redact = "x", lifespan = current_env())
  x2 <- new_headers(x1, redact = "x", lifespan = current_env())

  expect_equal(headers_flatten(x2, FALSE), list(x = "x"))
})

test_that("redaction is case-insensitive", {
  headers <- new_headers(
    list("AUTHORIZATION" = "SECRET"),
    redact = "authorization",
    lifespan = current_env()
  )
  redacted <- headers_flatten(headers)
  expect_named(redacted, "AUTHORIZATION")
  expect_true(is_redacted_sentinel(redacted$AUTHORIZATION))
})

bullets_with_header <- function(header, x) {
  if (length(x) == 0) {
    return()
  }

  cli::cat_line(cli::format_inline("{.strong {header}}"))
  bullets(x)
}

bullets <- function(x) {
  as_simple <- function(x) {
    if (is.atomic(x) && length(x) == 1) {
      if (is.character(x)) {
        paste0('"', x, '"')
      } else {
        format(x)
      }
    } else {
      if (is_redacted_sentinel(x)) {
        format(x)
      } else {
        paste0("<", class(x)[[1L]], ">")
      }
    }
  }
  vals <- map_chr(x, as_simple)
  names <- format(names(x))
  names <- gsub(" ", "\u00a0", names, fixed = TRUE)

  for (i in seq_along(x)) {
    cli::cat_line(cli::format_inline("* {.field {names[[i]]}}: {vals[[i]]}"))
  }
}

modify_list <- function(
  .x,
  ...,
  .ignore_case = FALSE,
  error_call = caller_env()
) {
  dots <- list2(...)
  if (length(dots) == 0) {
    return(.x)
  }

  if (!is_named(dots)) {
    cli::cli_abort(
      "All components of {.arg ...} must be named.",
      call = error_call
    )
  }

  if (.ignore_case) {
    out <- .x[!tolower(names(.x)) %in% tolower(names(dots))]
  } else {
    out <- .x[!names(.x) %in% names(dots)]
  }

  out <- c(out, compact(dots))

  if (length(out) == 0) {
    names(out) <- NULL
  }

  out
}


sys_sleep <- function(seconds, task, fps = 10, progress = NULL) {
  check_number_decimal(seconds)
  check_string(task)
  check_number_decimal(fps)
  progress <- progress %||% getOption("httr2_progress", !is_testing())
  check_bool(progress, allow_null = TRUE)

  if (seconds == 0) {
    return(invisible())
  }

  if (!progress) {
    cli::cli_alert("Waiting {round(seconds, 2)}s {task}")
    Sys.sleep(seconds)
    return(invisible())
  }

  start <- cur_time()
  signal("", class = "httr2_sleep", seconds = seconds)

  cli::cli_progress_bar(
    format = "Waiting {ceiling(seconds)}s {task} {cli::pb_bar}",
    total = seconds * fps
  )

  while (
    {
      left <- start + seconds - cur_time()
      left > 0
    }
  ) {
    Sys.sleep(min(1 / fps, left))
    cli::cli_progress_update(set = (seconds - left) * fps)
  }
  cli::cli_progress_done()

  invisible()
}

# allow mocking
Sys.sleep <- NULL

cur_time <- function() proc.time()[[3]]

is_error <- function(x) inherits(x, "error")

unix_time <- function() as.integer(Sys.time())

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

# https://datatracker.ietf.org/doc/html/rfc7636#appendix-A
base64_url_encode <- function(x) {
  x <- openssl::base64_encode(x)
  x <- gsub("=+$", "", x)
  x <- gsub("+", "-", x, fixed = TRUE)
  x <- gsub("/", "_", x, fixed = TRUE)
  x
}

base64_url_decode <- function(x) {
  mod4 <- nchar(x) %% 4
  if (mod4 > 0) {
    x <- paste0(x, strrep("=", 4 - mod4))
  }

  x <- gsub("_", "/", x, fixed = TRUE)
  x <- gsub("-", "+", x, fixed = TRUE)
  # x <- gsub("=+$", "", x)
  openssl::base64_decode(x)
}

base64_url_rand <- function(bytes = 32) {
  base64_url_encode(openssl::rand_bytes(bytes))
}

local_time <- function(x, tz = "UTC") {
  out <- as.POSIXct(x, tz = tz)
  attr(out, "tzone") <- NULL
  out
}

http_date <- function(x = Sys.time()) {
  withr::local_locale(LC_TIME = "C")
  strftime(x, "%a, %d %b %Y %H:%M:%S", tz = "UTC", usetz = TRUE)
}

parse_http_date <- function(x) {
  check_string(x)

  withr::local_locale(LC_TIME = "C")

  # https://datatracker.ietf.org/doc/html/rfc7231#section-7.1.1.1
  out <- as.POSIXct(strptime(x, "%a, %d %b %Y %H:%M:%S", tz = "UTC"))
  attr(out, "tzone") <- NULL
  out
}

touch <- function(path, time = Sys.time()) {
  if (!file.exists(path)) {
    file.create(path)
  }
  Sys.setFileTime(path, time)
}

local_write_lines <- function(..., .env = caller_env()) {
  path <- withr::local_tempfile(.local_envir = .env)
  writeLines(c(...), path)
  path
}

check_function2 <- function(
  x,
  ...,
  args = NULL,
  allow_null = FALSE,
  arg = caller_arg(x),
  call = caller_env()
) {
  check_function(
    x = x,
    allow_null = allow_null,
    arg = arg,
    call = call
  )

  if (!is.null(x)) {
    .check_function_args(
      f = x,
      expected_args = args,
      arg = arg,
      call = call
    )
  }
}

# Basically copied from rlang. Can be removed when https://github.com/r-lib/rlang/pull/1652
# is merged
.check_function_args <- function(f, expected_args, arg, call) {
  if (is_null(expected_args)) {
    return(invisible(NULL))
  }

  actual_args <- fn_fmls_names(f) %||% character()
  missing_args <- setdiff(expected_args, actual_args)
  if (is_empty(missing_args)) {
    return(invisible(NULL))
  }

  n_expected_args <- length(expected_args)
  n_actual_args <- length(actual_args)

  if (n_actual_args == 0) {
    arg_info <- "instead it has no arguments"
  } else {
    arg_info <- paste0("it currently has {.arg {actual_args}}")
  }

  cli::cli_abort(
    paste0(
      "{.arg {arg}} must have the {cli::qty(n_expected_args)}argument{?s} {.arg {expected_args}}; ",
      arg_info,
      "."
    ),
    call = call,
    arg = arg
  )
}

# This is inspired by the C interface of `cli_progress_bar()` which has just
# 2 arguments: `total` and `config`
create_progress_bar <- function(
  progress,
  total,
  name = "iterating",
  format = NULL,
  envir = caller_env(),
  frame = caller_env()
) {
  if (is_false(progress)) {
    return(list(
      update = function(...) {},
      done = function() {}
    ))
  }

  if (is.null(progress) || isTRUE(progress)) {
    args <- list()
    args$name <- name
    args$format <- format
  } else if (is_scalar_character(progress)) {
    args <- list(name = progress)
  } else if (is.list(progress)) {
    args <- progress
    args$name <- args$name %||% name
    args$format <- args$format %||% format
  } else {
    stop_input_type(
      progress,
      what = c("a bool", "a string", "a list"),
      call = frame
    )
  }

  args$total <- total
  args$.envir <- envir
  args$.auto_close <- FALSE

  id <- exec(cli::cli_progress_bar, !!!args)
  withr::defer(cli::cli_progress_done(id = id), envir = frame)

  list(
    update = function(...) {
      cli::cli_progress_update(..., id = id, .envir = envir)
    },
    done = function() cli::cli_progress_done(id = id)
  )
}

imap <- function(.x, .f, ...) {
  map2(.x, names(.x), .f, ...)
}

# Slices the vector using the only sane semantics: start inclusive, end
# exclusive.
#
# * Allows start == end, which means return no elements.
# * Allows start == length(vector) + 1, which means return no elements.
# * Allows zero-length vectors.
#
# Otherwise, slice() is quite strict about what it allows start/end to be: no
# negatives, no reversed order.
slice <- function(vector, start = 1, end = length(vector) + 1) {
  stopifnot(start > 0)
  stopifnot(start <= length(vector) + 1)
  stopifnot(end > 0)
  stopifnot(end <= length(vector) + 1)
  stopifnot(end >= start)

  if (start == end) {
    vector[FALSE] # Return an empty vector of the same type
  } else {
    vector[start:(end - 1)]
  }
}

is_named_list <- function(x) {
  is_list(x) && (is_named(x) || length(x) == 0)
}

pretty_json <- function(x) {
  tryCatch(
    gsub("\n$", "", jsonlite::prettify(x, indent = 2)),
    error = function(e) x
  )
}

log_stream <- function(..., prefix = "<< ") {
  out <- gsub("\n", paste0("\n", prefix), paste0(prefix, ..., collapse = ""))
  cli::cat_line(out)
}

paste_c <- function(..., collapse = "") {
  paste0(c(...), collapse = collapse)
}

# Give user the get-out-of-jail-free card if interrupt-capturing function
# is wrapped inside a loop
check_repeated_interrupt <- function() {
  if (as.double(Sys.time()) - the$last_interrupt < 1) {
    cli::cli_alert_warning("Interrupting")
    interrupt()
  }
  the$last_interrupt <- as.double(Sys.time())
}

test_that("modify list adds, removes, and overrides", {
  x <- list(x = 1)
  expect_equal(modify_list(x), x)
  expect_equal(modify_list(x, x = NULL), list())
  expect_equal(modify_list(x, x = 2), list(x = 2))
  expect_equal(modify_list(x, y = 3), list(x = 1, y = 3))
  expect_equal(modify_list(NULL, x = 2), list(x = 2))

  expect_snapshot(modify_list(x, a = 1, 2), error = TRUE)
})

test_that("replacement affects all components with name", {
  x <- list(a = 1, a = 2)
  expect_equal(modify_list(x, a = NULL), list())
  expect_equal(modify_list(x, a = 3), list(a = 3))
  expect_equal(modify_list(x, a = 3, a = 4), list(a = 3, a = 4))
})

test_that("progress bar suppressed in tests", {
  expect_snapshot(sys_sleep(0.1, "in test"))
})


test_that("has a working slice", {
  x <- letters[1:5]
  expect_identical(slice(x), x)
  expect_identical(slice(x, 1, length(x) + 1), x)

  # start is inclusive, end is exclusive
  expect_identical(slice(x, 1, length(x)), head(x, -1))
  # zero-length slices are fine
  expect_identical(slice(x, 1, 1), character())
  # starting off the end is fine
  expect_identical(slice(x, length(x) + 1), character())
  expect_identical(slice(x, length(x) + 1, length(x) + 1), character())
  # slicing zero-length is fine
  expect_identical(slice(character()), character())

  # out of bounds
  expect_error(slice(x, 0, 1))
  expect_error(slice(x, length(x) + 2))
  expect_error(slice(x, end = length(x) + 2))
  # end too small relative to start
  expect_error(slice(x, 2, 1))
})

#' Read a streaming body a chunk at a time
#'
#' @description
#' * `resp_stream_raw()` retrieves bytes (`raw` vectors).
#' * `resp_stream_lines()` retrieves lines of text (`character` vectors).
#' * `resp_stream_sse()` retrieves a single [server-sent
#'   event](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events).
#' * `resp_stream_aws()` retrieves a single event from an AWS stream
#'   (i.e. mime type `application/vnd.amazon.eventstream``).
#'
#' Use `resp_stream_is_complete()` to determine if there is further data
#' waiting on the stream.
#'
#' @returns
#' * `resp_stream_raw()`: a raw vector.
#' * `resp_stream_lines()`: a character vector.
#' * `resp_stream_sse()`: a list with components `type`, `data`, and `id`.
#'   `type`, `data`, and `id` are always strings; `data` and `id` may be empty
#'   strings.
#' * `resp_stream_aws()`: a list with components `headers` and `body`.
#'   `body` will be automatically parsed if the event contents a `:content-type`
#'   header with `application/json`.
#'
#' `resp_stream_sse()` and `resp_stream_aws()` will return `NULL` to signal that
#' the end of the stream has been reached or, if in nonblocking mode, that
#' no event is currently available.
#' @export
#' @param resp,con A streaming [response] created by [req_perform_connection()].
#' @param kb How many kilobytes (1024 bytes) of data to read.
#' @order 1
#' @examples
#' req <- request(example_url()) |>
#'   req_template("GET /stream/:n", n = 5)
#'
#' con <- req |> req_perform_connection()
#' while (!resp_stream_is_complete(con)) {
#'   lines <- con |> resp_stream_lines(2)
#'   cat(length(lines), " lines received\n", sep = "")
#' }
#' close(con)
#'
#' # You can also see what's happening by setting verbosity
#' con <- req |> req_perform_connection(verbosity = 2)
#' while (!resp_stream_is_complete(con)) {
#'   lines <- con |> resp_stream_lines(2)
#' }
#' close(con)
resp_stream_raw <- function(resp, kb = 32) {
  init_streaming_response(resp, RawSplitter)
  check_number_decimal(kb, min = 0, allow_infinite = FALSE)

  out <- resp$body$read(kb * 1024)
  if (resp_stream_show_body(resp)) {
    log_stream("Streamed ", length(out), " bytes")
    cli::cat_line()
  }
  out
}

#' @export
#' @rdname resp_stream_raw
#' @order 6
resp_stream_is_complete <- function(resp) {
  init_streaming_response(resp)
  !stream_has_buffered(resp) && resp$body$is_complete()
}

# Is there any data still buffered in the response that hasn't been returned to
# the user yet? (Either raw bytes or decoded-but-unserved blocks.)
stream_has_buffered <- function(resp) {
  cache <- resp$cache
  if (length(cache$stream_buffer) > 0) {
    return(TRUE)
  }
  cache$stream_pos <= length(cache$stream_queue)
}

#' @export
#' @param ... Not used; included for compatibility with generic.
#' @rdname resp_stream_raw
#' @order 5
close.httr2_response <- function(con, ...) {
  check_response(con)

  if (inherits(con$body, "StreamingBody")) {
    con$body$close()
  }

  invisible()
}

# Streaming engine ------------------------------------------------------------

# How many bytes to read from the connection at a time when streaming.
stream_chunk_bytes <- 64L * 1024L

# Reads from a streaming response and uses `splitter` to divide the byte stream
# into blocks, returning up to `n` of them in a list. The byte-level details -
# where blocks begin and end, how trailing bytes are handled at the end of the
# stream, and how many bytes to read at a time - all live in the `splitter`
# (a `StreamSplitter`), so that this loop can be shared by lines, server-sent
# events, and AWS events.
#
# A whole chunk is split at once and the resulting blocks are held in a queue
# served by an index pointer, so that reading a few blocks at a time doesn't
# repeatedly rescan and recopy the buffered bytes.
stream_pull <- function(resp, n, splitter, max_size) {
  cache <- resp$cache
  queue <- cache$stream_queue
  pos <- cache$stream_pos

  # Accumulate served slices in a list and flatten once at the end, so serving
  # many blocks (e.g. `lines = Inf`) doesn't repeatedly recopy a growing vector.
  serve <- list()
  n_out <- 0L
  repeat {
    # Serve whatever is already queued.
    available <- length(queue) - pos + 1L
    if (available > 0L) {
      take <- min(available, n - n_out)
      serve[[length(serve) + 1L]] <- queue[pos:(pos + take - 1L)]
      pos <- pos + take
      n_out <- n_out + take
    }
    if (n_out >= n) {
      break
    }

    # The queue is exhausted; combine any buffered bytes with a fresh read and
    # split the next batch of blocks. We always reparse the buffered bytes (not
    # just freshly read ones), because data may have been buffered by an earlier
    # call that didn't find a complete block.

    stream_buffer <- cache$stream_buffer
    if (length(stream_buffer) > max_size) {
      stop_stream_size(max_size)
    }
    # Read up to one byte past the size limit - the extra byte lets us detect a
    # buffer that has overflowed - but never more than one chunk at a time.
    if (is.finite(max_size)) {
      read_size <- min(
        stream_chunk_bytes,
        max(max_size - length(stream_buffer) + 1L, 1L)
      )
    } else {
      read_size <- stream_chunk_bytes
    }
    chunk <- resp$body$read(read_size)
    buffer <- c(stream_buffer, chunk)
    if (length(buffer) == 0L) {
      break
    }

    if (length(chunk) > 0L && resp_stream_show_buffer(resp)) {
      log_stream(cli::rule("Buffer"), prefix = "*  ")
      log_stream(
        "Received chunk: ",
        paste(as.character(chunk), collapse = " "),
        prefix = "*  "
      )
    }

    # Divide the buffer into complete blocks at the splitter's boundaries.
    splits <- splitter$find_boundaries(buffer)
    if (length(splits) > 0L) {
      starts <- c(1L, splits[-length(splits)])
      queue <- lapply(seq_along(splits), function(i) {
        buffer[starts[i]:(splits[i] - 1L)]
      })
      pos <- 1L
      # Trailing bytes after the last boundary stay buffered for the next read.
      cache$stream_buffer <- buffer[seq2(
        splits[[length(splits)]],
        length(buffer)
      )]
      # Checkpoint the freshly parsed queue before serving from it: if a later
      # read trips the size limit, the errored call is retried and these blocks
      # are served again rather than lost (their bytes are already consumed).
      cache$stream_queue <- queue
      cache$stream_pos <- pos
      next
    }

    # No complete block in the buffer; keep it buffered for the next read.
    cache$stream_buffer <- buffer
    if (length(chunk) == 0L) {
      if (resp$body$is_complete()) {
        # The stream has ended; let the splitter flush any trailing bytes.
        final <- splitter$finish(buffer)
        if (length(final) > 0L) {
          serve[[length(serve) + 1L]] <- final
        }
        cache$stream_buffer <- raw()
      }
      # Either EOF, or no data currently available (non-blocking).
      break
    }
    # We read new bytes but still don't have a complete block; loop to read more.
  }

  # Drop the queue once it's been fully served, so its blocks can be freed.
  if (pos > length(queue)) {
    queue <- list()
    pos <- 1L
  }
  cache$stream_queue <- queue
  cache$stream_pos <- pos

  unlist(serve, recursive = FALSE, use.names = FALSE)
}

# A `StreamSplitter` describes a stream format for `stream_pull()`, which does
# the reading, splitting, and queueing. Subclasses (lines, server-sent events,
# AWS events) supply a `name` plus `find_boundaries()`, which takes a raw vector
# and returns an integer vector of split points: the position one past the end
# of each complete block.
StreamSplitter <- R6::R6Class(
  "StreamSplitter",
  public = list(
    # The full name of the public reader (e.g. "resp_stream_lines()") this
    # splitter backs. Used to detect mixing readers on one response.
    name = NULL,
    # nocov start: abstract, always overridden by a subclass.
    find_boundaries = function(buffer) {
      cli::cli_abort("Not implemented.", .internal = TRUE)
    },
    # nocov end
    # Emit any final blocks once the stream has ended with `remainder` bytes
    # left over after the last complete block. The default discards a trailing
    # partial block; line splitting overrides this to keep it.
    finish = function(remainder) {
      if (length(remainder) != 0L) {
        cli::cli_warn("Premature end of input; ignoring final partial chunk")
      }
      list()
    }
  )
)

# resp_stream_raw() reads bytes straight off the connection, so it never goes
# through stream_pull() and needs no splitting behavior. It's duck-typed rather
# than a StreamSplitter subclass: init_streaming_response() only needs a `name`
# (for reader tracking).
RawSplitter <- R6::R6Class(
  "RawSplitter",
  public = list(name = "resp_stream_raw()")
)

stop_stream_size <- function(max_size, call = caller_env()) {
  cli::cli_abort(
    "Streaming read exceeded size limit of {max_size}",
    class = "httr2_streaming_error",
    call = call
  )
}

# Helpers ----------------------------------------------------

# Validate a streaming response, initialize its cache, and (when a `splitter`
# generator is supplied) construct and cache the splitter. The cached splitter
# both drives the reads and records which reader is in use, so attempting a
# second, different reader on the same response errors. Returns the splitter,
# so callers can write `splitter <- init_streaming_response(resp, SseSplitter)`.
init_streaming_response <- function(
  resp,
  splitter = NULL,
  arg = caller_arg(resp),
  call = caller_env()
) {
  check_response(resp, arg = arg, call = call)

  if (resp_body_type(resp) != "stream") {
    stop_input_type(
      resp,
      "a streaming HTTP response object",
      allow_null = FALSE,
      arg = arg,
      call = call
    )
  }

  if (!resp$body$is_open()) {
    cli::cli_abort("{.arg {arg}} has already been closed.", call = call)
  }

  # Initialize the streaming cache so the read loop and stream_has_buffered()
  # can read these fields without `%||%` guards. Only fills missing fields, so
  # it preserves any bytes a caller has already pushed back.
  cache <- resp$cache
  cache$stream_buffer <- cache$stream_buffer %||% raw()
  cache$stream_queue <- cache$stream_queue %||% list()
  cache$stream_pos <- cache$stream_pos %||% 1L

  if (is.null(splitter)) {
    return(invisible(NULL))
  }

  # Construct the splitter once per response and cache it.
  cached <- cache$stream_splitter
  if (is.null(cached)) {
    cache$stream_splitter <- splitter$new()
    return(invisible(cache$stream_splitter))
  }
  if (!inherits(cached, splitter$classname)) {
    used <- splitter$new()$name
    cli::cli_abort(
      "Can't use {used} after {cached$name} on the same response.",
      class = "httr2_streaming_error",
      call = call
    )
  }
  invisible(cached)
}

resp_stream_show_body <- function(resp) {
  resp$request$policies$show_streaming_body %||% FALSE
}
resp_stream_show_buffer <- function(resp) {
  resp$request$policies$show_streaming_buffer %||% FALSE
}

test_that("can stream bytes from a connection", {
  resp <- request_test("/stream-bytes/2048") |> req_perform_connection()
  withr::defer(close(resp))

  expect_s3_class(resp, "httr2_response")
  expect_true(resp_has_body(resp))

  out <- resp_stream_raw(resp, 1)
  expect_length(out, 1024)

  out <- resp_stream_raw(resp, 1)
  expect_length(out, 1024)

  out <- resp_stream_raw(resp, 1)
  expect_length(out, 0)
})

test_that("can determine if a stream is complete (blocking)", {
  resp <- request_test("/stream-bytes/2048") |> req_perform_connection()
  withr::defer(close(resp))

  expect_false(resp_stream_is_complete(resp))
  expect_length(resp_stream_raw(resp, kb = 2), 2048)
  expect_length(resp_stream_raw(resp, kb = 1), 0)
  expect_true(resp_stream_is_complete(resp))
})

test_that("can determine if a stream is complete (non-blocking)", {
  resp <- request_test("/stream-bytes/2048") |>
    req_perform_connection(blocking = FALSE)
  withr::defer(close(resp))

  expect_false(resp_stream_is_complete(resp))
  expect_length(resp_stream_raw(resp, kb = 2), 2048)
  expect_length(resp_stream_raw(resp, kb = 1), 0)
  expect_true(resp_stream_is_complete(resp))
})

test_that("can't read from a closed connection", {
  resp <- request_test("/stream-bytes/1024") |> req_perform_connection()
  close(resp)

  expect_false(resp_has_body(resp))
  expect_snapshot(resp_stream_raw(resp, 1), error = TRUE)

  # and no error if we try to close it again
  expect_no_error(close(resp))
})

test_that("streaming functions require a streaming response", {
  expect_snapshot(resp_stream_raw(response()), error = TRUE)
})

test_that("resp_stream_raw() validates kb", {
  resp <- local_streaming_response(charToRaw("abc"))

  expect_length(resp_stream_raw(resp, kb = 1 / 1024), 1)
  expect_snapshot(resp_stream_raw(resp, kb = -1), error = TRUE)
  expect_snapshot(resp_stream_raw(resp, kb = Inf), error = TRUE)
})

test_that("resp_stream_is_complete() requires an open streaming response", {
  expect_snapshot(resp_stream_is_complete(response()), error = TRUE)

  resp <- local_streaming_response(charToRaw("abc"))
  close(resp)
  expect_snapshot(resp_stream_is_complete(resp), error = TRUE)
})

test_that("streaming responses use only one reader", {
  resp <- local_streaming_response(charToRaw("a\n"))

  expect_equal(resp_stream_lines(resp), "a")
  expect_snapshot(resp_stream_raw(resp), error = TRUE)
})

test_that("StreamSplitter$finish() discards a trailing partial block with a warning", {
  # The block-splitting and read sizing live in stream_pull() and are covered by
  # its integration tests; finish() is the one behavior that stays on the class.
  s <- SseSplitter$new()
  # Nothing buffered: nothing to flush, no warning.
  expect_equal(s$finish(raw()), list())
  # A trailing partial block is dropped with a warning.
  expect_snapshot(out <- s$finish(charToRaw("b")))
  expect_equal(out, list())
})

# stream_pull() drives every format (lines, sse, aws); these tests exercise its
# format-independent behavior through resp_stream_lines() as a convenient
# vehicle. Format-specific splitting and parsing are tested in the per-format
# files.

test_that("stream_pull() buffers incomplete blocks across reads (non-blocking)", {
  sync <- sync_req("pull")
  req <- local_app_request(function(req, res) {
    sync <- req$app$locals$sync_rep("pull")

    res$send_chunk("This is a ")
    sync(res$send_chunk("complete sentence.\n"))
  })

  resp <- req_perform_connection(req, blocking = FALSE)
  withr::defer(close(resp))
  wait_for_http_data(resp)

  # An incomplete block is held in stream_buffer and nothing is served yet.
  expect_equal(resp_stream_lines(resp), character())
  expect_equal(resp$cache$stream_buffer, charToRaw("This is a "))
  # Buffered bytes mean the stream isn't complete, even between blocks.
  expect_false(resp_stream_is_complete(resp))

  expect_equal(resp_stream_lines(resp), character())

  sync(resp)
  expect_equal(resp_stream_lines(resp), "This is a complete sentence.")
})

test_that("stream_pull() serves queued blocks from a single read", {
  req <- local_app_request(function(req, res) {
    res$send_chunk(paste0(letters[1:5], "\n", collapse = ""))
  })

  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  # All five blocks are split from one read; `n` limits how many are served and
  # the rest stay queued, so is_complete() is FALSE while the queue is non-empty.
  expect_equal(resp_stream_lines(resp, 3), c("a", "b", "c"))
  expect_false(resp_stream_is_complete(resp))
  expect_equal(resp_stream_lines(resp, 3), c("d", "e"))
  expect_equal(resp_stream_lines(resp, 3), character())
  expect_true(resp_stream_is_complete(resp))
})

test_that("stream_pull() serves every block when n is Inf", {
  req <- local_app_request(function(req, res) {
    res$send_chunk(paste0(letters[1:5], "\n", collapse = ""))
  })

  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  # n = Inf keeps serving past the queue and reads on until the stream ends.
  expect_equal(resp_stream_lines(resp, Inf), letters[1:5])
  expect_equal(resp_stream_lines(resp, Inf), character())
  expect_true(resp_stream_is_complete(resp))
})

test_that("stream_pull() enforces max_size (blocking and non-blocking)", {
  req <- local_app_request(function(req, res) {
    res$send_chunk(paste(rep_len("0", 1000), collapse = ""))
  })

  resp1 <- req_perform_connection(req, blocking = FALSE)
  withr::defer(close(resp1))
  wait_for_http_data(resp1)
  expect_error(
    resp_stream_lines(resp1, max_size = 999),
    class = "httr2_streaming_error"
  )

  resp2 <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp2))
  expect_error(
    resp_stream_lines(resp2, max_size = 999),
    class = "httr2_streaming_error"
  )
})

test_that("stream_pull() keeps size errors reproducible on retry", {
  resp <- local_streaming_response(c(rep(as.raw(0x61), 11), as.raw(0x0A)))

  expect_error(
    resp_stream_lines(resp, max_size = 10),
    class = "httr2_streaming_error"
  )
  expect_error(
    resp_stream_lines(resp, max_size = 10),
    class = "httr2_streaming_error"
  )
})

test_that("stream_pull() re-serves queued blocks after a failed call", {
  # "a" splits off cleanly, but the long second line trips max_size, so the
  # call errors *after* queueing "a". The queue is checkpointed, so retrying
  # (here with a larger limit) still yields "a" rather than dropping it.
  resp <- local_streaming_response(c(charToRaw("a\n"), rep(as.raw(0x30), 13)))

  expect_error(
    resp_stream_lines(resp, lines = 2, max_size = 10),
    class = "httr2_streaming_error"
  )
  expect_equal(
    resp_stream_lines(resp, lines = 2, max_size = 100),
    c("a", strrep("0", 13))
  )
})

test_that("stream_pull() flushes a trailing block at end of stream", {
  req <- local_app_request(function(req, res) {
    # "b" has no trailing line ending, so it can't be served until EOF.
    res$send_chunk("a\nb")
  })
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  expect_equal(resp_stream_lines(resp, 1), "a")
  # At end of stream the splitter flushes its buffered remainder as a block.
  expect_equal(resp_stream_lines(resp, 1), "b")
  expect_equal(resp_stream_lines(resp, 1), character())
})

test_that("verbosity = 3 logs the buffered chunk", {
  req <- local_app_request(function(req, res) {
    res$send_chunk("line 1\n")
    res$send_chunk("line 2\n")
  })

  expect_output(
    con <- req_perform_connection(req, blocking = TRUE, verbosity = 3)
  )
  on.exit(close(con))
  expect_snapshot(
    {
      while (!resp_stream_is_complete(con)) {
        resp_stream_lines(con, 1)
      }
    },
    transform = transform_verbose_response
  )
})

test_that("verbosity = 2 streams request bodies", {
  req <- local_app_request(function(req, res) {
    res$send_chunk("line 1\n")
    res$send_chunk("line 2\n")
  })

  stream_all <- function(req, fun, ...) {
    con <- req_perform_connection(req, blocking = TRUE, verbosity = 2)
    withr::defer(close(con))
    while (!resp_stream_is_complete(con)) {
      fun(con, ...)
    }
  }
  expect_snapshot(
    {
      stream_all(req, resp_stream_lines, 1)
      stream_all(req, resp_stream_raw, 5 / 1024)
    },
    transform = function(lines) lines[!grepl("^(<-|->)", lines)]
  )
})

#' OAuth with authorization code
#'
#' @description
#' Authenticate using the OAuth **authorization code flow**, as defined
#' by `r rfc(6749, 4.1)`.
#'
#' This flow is the most commonly used OAuth flow where the user
#' opens a page in their browser, approves the access, and then returns to R.
#' When possible, it redirects the browser back to a temporary local webserver
#' to capture the authorization code. When this is not possible (e.g., when
#' running on a hosted platform like RStudio Server), provide a custom
#' `redirect_uri` and httr2 will prompt the user to enter the code manually.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>, and more about the motivations
#' behind this flow in
#' <https://stack-auth.com/blog/oauth-from-first-principles>.
#'
#' # Security considerations
#'
#' The authorization code flow is used for both web applications and native
#' applications (which are equivalent to R packages). `r rfc(8252)` spells out
#' important considerations for native apps. Most importantly there's no way
#' for native apps to keep secrets from their users. This means that the
#' server should either not require a `client_secret` (i.e. it should be a
#' public client and not a confidential client) or ensure that possession of
#' the `client_secret` doesn't grant any significant privileges.
#'
#' Only modern APIs from major providers (like Azure and Google) explicitly
#' support native apps. However, in most cases, even for older APIs, possessing
#' the `client_secret` provides limited ability to perform harmful actions.
#' Therefore, our general principle is that it's acceptable to include it in an
#' R package, as long as it's mildly obfuscated to protect against credential
#' scraping attacks (which aim to acquire large numbers of client secrets by
#' scanning public sites like GitHub). The goal is to ensure that obtaining your
#' client credentials is more work than just creating a new client.
#'
#' @export
#' @family OAuth flows
#' @seealso [oauth_flow_auth_code_url()] for the components necessary to
#'   write your own auth code flow, if the API you are wrapping does not adhere
#'   closely to the standard.
#' @inheritParams req_perform
#' @param client An [oauth_client()].
#' @param auth_url Authorization url; you'll need to discover this by reading
#'   the documentation.
#' @param scope Scopes to be requested from the resource owner.
#' @param pkce Use "Proof Key for Code Exchange"? This adds an extra layer of
#'   security and should always be used if supported by the server.
#' @param auth_params A list containing additional parameters passed to
#'   [oauth_flow_auth_code_url()].
#' @param token_params List containing additional parameters passed to the
#'   `token_url`.
#' @param redirect_uri URL to redirect back to after authorization is complete.
#'   Often this must be registered with the API in advance.
#'
#'   httr2 supports three forms of redirect. Firstly, you can use a `localhost`
#'   url (the default), where httr2 will set up a temporary webserver to listen
#'   for the OAuth redirect. In this case, httr2 will automatically append a
#'   random port. If you need to set it to a fixed port because the API requires
#'   it, then specify it with (e.g.) `"http://localhost:1011"`. This technique
#'   works well when you are working on your own computer.
#'
#'   Secondly, you can provide a URL to a website that uses Javascript to
#'   give the user a code to copy and paste back into the R session (see
#'   <https://tidyverse.org/google-callback/> and
#'   <https://github.com/r-lib/gargle/blob/main/inst/pseudo-oob/google-callback/index.html>
#'   for examples). This is less convenient (because it requires more
#'   user interaction) but also works in hosted environments like RStudio
#'   Server.
#'
#'   Finally, hosted platforms might set the `HTTR2_OAUTH_REDIRECT_URL` and
#'   `HTTR2_OAUTH_CODE_SOURCE_URL` environment variables. In this case, httr2
#'   will use `HTTR2_OAUTH_REDIRECT_URL` for redirects by default, and poll the
#'   `HTTR2_OAUTH_CODE_SOURCE_URL` endpoint with the state parameter until it
#'   receives a code in the response (or encounters an error). This delegates
#'   completion of the authorization flow to the hosted platform.
#' @param cache_disk Should the access token be cached on disk? This reduces
#'   the number of times that you need to re-authenticate at the cost of
#'   storing access credentials on disk.
#'
#'   Learn more in <https://httr2.r-lib.org/articles/oauth.html>.
#' @param cache_key If you want to cache multiple tokens per app, use this
#'   key to disambiguate them.
#' @returns `req_oauth_auth_code()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_auth_code()` returns an [oauth_token].
#' @examples
#' req_auth_github <- function(req) {
#'   req_oauth_auth_code(
#'     req,
#'     client = example_github_client(),
#'     auth_url = "https://github.com/login/oauth/authorize"
#'   )
#' }
#'
#' request("https://api.github.com/user") |>
#'   req_auth_github()
req_oauth_auth_code <- function(
  req,
  client,
  auth_url,
  scope = NULL,
  pkce = TRUE,
  auth_params = list(),
  token_params = list(),
  redirect_uri = oauth_redirect_uri(),
  cache_disk = FALSE,
  cache_key = NULL
) {
  redirect <- normalize_redirect_uri(redirect_uri = redirect_uri)

  params <- list(
    client = client,
    auth_url = auth_url,
    scope = scope,
    pkce = pkce,
    auth_params = auth_params,
    token_params = token_params,
    redirect_uri = redirect$uri
  )

  cache <- cache_choose(client, cache_disk, cache_key)
  req_oauth(req, "oauth_flow_auth_code", params, cache = cache)
}

#' @export
#' @rdname req_oauth_auth_code
oauth_flow_auth_code <- function(
  client,
  auth_url,
  scope = NULL,
  pkce = TRUE,
  auth_params = list(),
  token_params = list(),
  redirect_uri = oauth_redirect_uri()
) {
  oauth_flow_check("authorization code", client, interactive = TRUE)

  redirect <- normalize_redirect_uri(redirect_uri = redirect_uri)

  if (pkce) {
    code <- oauth_flow_auth_code_pkce()
    auth_params$code_challenge <- code$challenge
    auth_params$code_challenge_method <- code$method
    token_params$code_verifier <- code$verifier
  }

  state <- base64_url_rand(32)

  # Redirect user to authorisation url.
  user_url <- oauth_flow_auth_code_url(
    client,
    auth_url = auth_url,
    redirect_uri = redirect$uri,
    scope = scope,
    state = state,
    auth_params = auth_params
  )
  utils::browseURL(user_url)

  if (redirect$can_fetch_code) {
    # Wait a bit to give the user a chance to click through the authorisation
    # process.
    if (!is_testing()) {
      sys_sleep(2, "for browser-based authentication", progress = FALSE)
    }

    code <- oauth_flow_auth_code_fetch(state)
  } else if (redirect$localhost) {
    # Listen on localhost for the result
    result <- oauth_flow_auth_code_listen(redirect$uri)
    code <- oauth_flow_auth_code_parse(result, state)
  } else {
    # Allow the user to retrieve the token out of band manually and enter it
    # into the console. This is what {gargle} terms the "pseudo out-of-band"
    # flow.
    code <- oauth_flow_auth_code_read(state)
  }

  # Get access/refresh token from authorisation code
  # https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.3
  oauth_client_get_token(
    client,
    grant_type = "authorization_code",
    code = code,
    redirect_uri = redirect_uri,
    !!!token_params
  )
}

normalize_redirect_uri <- function(redirect_uri, error_call = caller_env()) {
  old <- parsed <- url_parse(redirect_uri)
  localhost <- parsed$hostname %in% c("localhost", "127.0.0.1")

  if (localhost) {
    check_installed("httpuv", "for desktop OAuth", call = NULL)
    if (is_hosted_session()) {
      cli::cli_abort(
        "Can't use localhost {.arg redirect_uri} in a hosted environment.",
        call = error_call
      )
    }

    if (is.null(parsed$port)) {
      parsed$port <- httpuv::randomPort()
    }
  }

  list(
    uri = if (identical(old, parsed)) redirect_uri else url_build(parsed),
    localhost = localhost,
    can_fetch_code = can_fetch_oauth_code(redirect_uri)
  )
}


#' Default redirect url for OAuth
#'
#' The default redirect uri used by [req_oauth_auth_code()]. Defaults to
#' `http://localhost` unless the `HTTR2_OAUTH_REDIRECT_URL` envvar is set.
#'
#' @export
oauth_redirect_uri <- function() {
  Sys.getenv("HTTR2_OAUTH_REDIRECT_URL", "http://localhost")
}

# Authorisation request: make a url that the user navigates to
# https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.1

#' OAuth authorization code components
#'
#' @description
#' These low-level functions can be used to assemble a custom flow for
#' APIs that are further from the spec:
#'
#' * `oauth_flow_auth_code_url()` generates the url that should be opened in a
#'   browser.
#' * `oauth_flow_auth_code_listen()` starts a temporary local webserver that
#'   listens for the response from the resource server.
#' * `oauth_flow_auth_code_parse()` parses the query parameters returned from
#'   the server redirect, verifying that the `state` is correct, and returning
#'   the authorisation code.
#' * `oauth_flow_auth_code_pkce()` generates code verifier, method, and challenge
#'   components as needed for PKCE, as defined in `r rfc(7636)`.
#'
#' @export
#' @keywords internal
#' @param state Random state generated by `oauth_flow_auth_code()`. Used to
#'   verify that we're working with an authentication request that we created.
#'   (This is an unlikely threat for R packages since the webserver that
#'   listens for authorization responses is transient.)
oauth_flow_auth_code_url <- function(
  client,
  auth_url,
  redirect_uri = NULL,
  scope = NULL,
  state = NULL,
  auth_params = list()
) {
  url <- url_parse(auth_url)
  url$query <- modify_list(
    url$query,
    response_type = "code",
    client_id = client$id,
    redirect_uri = redirect_uri,
    scope = scope,
    state = state,
    !!!auth_params
  )
  url_build(url)
}

#' @export
#' @rdname oauth_flow_auth_code_url
oauth_flow_auth_code_listen <- function(
  redirect_uri = "http://localhost:1410"
) {
  parsed <- url_parse(redirect_uri)
  port <- as.integer(parsed$port)
  path <- parsed$path %||% "/"

  complete <- FALSE
  info <- NULL
  listen <- function(env) {
    if (!identical(env$PATH_INFO, path)) {
      return(list(
        status = 404L,
        headers = list("Content-Type" = "text/plain"),
        body = "Not found"
      ))
    }

    query <- env$QUERY_STRING
    if (!is.character(query) || identical(query, "")) {
      complete <<- TRUE
    } else {
      complete <<- TRUE
      info <<- parse_form_urlencoded(query)
    }

    list(
      status = 200L,
      headers = list("Content-Type" = "text/plain"),
      body = "Authentication complete. Please close this page and return to R."
    )
  }
  server <- httpuv::startServer("127.0.0.1", port, list(call = listen))
  withr::defer(httpuv::stopServer(server))

  # TODO: make this a progress bar
  inform("Waiting for authentication in browser...")
  inform("Press Esc/Ctrl + C to abort")
  while (!complete) {
    httpuv::service()
  }
  httpuv::service() # send data back to client

  if (is.null(info)) {
    cli::cli_abort("Authentication failed; invalid url from server.")
  }

  info
}

# application/x-www-form-urlencoded defined in
# https://www.w3.org/TR/html401/interact/forms.html#h-17.13.4.1
# Spaces are first replaced by +
parse_form_urlencoded <- function(query) {
  query <- url_query_parse(query)
  query[] <- gsub("+", " ", query, fixed = TRUE)
  query
}

# Authorisation response: get query params back from redirect
# https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.2
#' @export
#' @rdname oauth_flow_auth_code_url
#' @param query List of query parameters returned by `oauth_flow_auth_code_listen()`.
oauth_flow_auth_code_parse <- function(query, state) {
  if (has_name(query, "error")) {
    # https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.2.1
    # Never see problems with redirect_uri
    oauth_flow_abort(query$error, query$error_description, query$error_uri)
  }

  if (query$state != state) {
    cli::cli_abort("Authentication failure: state does not match.")
  }

  query$code
}

#' @export
#' @rdname oauth_flow_auth_code_url
oauth_flow_auth_code_pkce <- function() {
  # https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
  #
  # It is RECOMMENDED that the output of a suitable random number generator
  # be used to create a 32-octet sequence.  The octet sequence is then
  # base64url-encoded to produce a 43-octet URL safe string to use as the
  # code verifier.
  verifier <- base64_url_rand(32)

  list(
    verifier = verifier,
    method = "S256",
    challenge = base64_url_encode(openssl::sha256(charToRaw(verifier)))
  )
}

# Try to determine whether we can redirect the user's browser to a server on
# localhost, which isn't possible if we are running on a hosted platform.
#
# Currently this detects RStudio Server, Posit Workbench, and Google Colab. It
# is based on the strategy pioneered by the {gargle} package.
is_hosted_session <- function() {
  if (nzchar(Sys.getenv("COLAB_RELEASE_TAG"))) {
    return(TRUE)
  }
  # If RStudio Server or Posit Workbench is running locally (which is possible,
  # though unusual), it's not acting as a hosted environment.
  Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server" &&
    !grepl("localhost", Sys.getenv("RSTUDIO_HTTP_REFERER"), fixed = TRUE)
}

oauth_flow_auth_code_read <- function(state) {
  code <- trimws(readline("Enter authorization code or URL: "))

  if (is_string_url(code)) {
    # minimal setup where user copy & pastes a URL
    parsed <- url_parse(code)

    code <- parsed$query$code
    new_state <- parsed$query$state
  } else if (is_base64_json(code)) {
    # {gargle} style, where the user copy & pastes a base64-encoded JSON
    # object with both the code and state. This is used on
    # https://www.tidyverse.org/google-callback/
    json <- jsonlite::fromJSON(rawToChar(openssl::base64_decode(code)))

    code <- json$code
    new_state <- json$state
  } else {
    # Full manual approach, where the code and state are entered
    # independently.

    new_state <- trimws(readline("Enter state parameter: "))
  }

  if (!identical(state, new_state)) {
    abort("Authentication failure: state does not match")
  }

  code
}

is_string_url <- function(x) grepl("^https?://", x)

is_base64_json <- function(x) {
  tryCatch(
    {
      jsonlite::fromJSON(rawToChar(openssl::base64_decode(x)))
      TRUE
    },
    error = function(err) FALSE
  )
}


# Determine whether we can fetch the OAuth authorization code from an external
# source without user interaction.
can_fetch_oauth_code <- function(redirect_url) {
  nchar(Sys.getenv("HTTR2_OAUTH_CODE_SOURCE_URL")) &&
    Sys.getenv("HTTR2_OAUTH_REDIRECT_URL") == redirect_url
}

# Fetch the authorization code from an external source that is serving as a
# redirect URL. This assumes a very simple API that takes the state parameter in
# the query string and returns a JSON object with a `code` key.
oauth_flow_auth_code_fetch <- function(state) {
  req <- request(Sys.getenv("HTTR2_OAUTH_CODE_SOURCE_URL"))
  req <- req_url_query(req, state = state)
  req <- req_retry(
    req,
    max_seconds = 60,
    # The endpoint may temporarily return a 404 when no code is found for a
    # given state because the user hasn't finished clicking through yet.
    is_transient = \(resp) resp_status(resp) %in% c(404, 429, 503)
  )
  resp <- req_perform(req)
  body <- resp_body_json(resp)
  body$code
}

# Make base::readline() mockable
readline <- NULL

test_that("desktop style can't run in hosted environment", {
  client <- oauth_client("abc", "http://example.com")

  withr::local_options(rlang_interactive = TRUE)
  withr::local_envvar("RSTUDIO_PROGRAM_MODE" = "server")
  expect_snapshot(
    oauth_flow_auth_code(client, "http://localhost"),
    error = TRUE
  )
})

test_that("so-called 'hosted' sessions are detected correctly", {
  withr::with_envvar(c("RSTUDIO_PROGRAM_MODE" = "server"), {
    expect_true(is_hosted_session())
  })
  # Emulate running outside RStudio Server if we happen to be running our tests
  # under it.
  withr::with_envvar(c("RSTUDIO_PROGRAM_MODE" = NA), {
    expect_false(is_hosted_session())
  })
})

test_that("URL embedding authorisation code and state can be input manually", {
  local_mocked_bindings(
    readline = function(prompt = "") "https://x.com?code=code&state=state"
  )
  expect_equal(oauth_flow_auth_code_read("state"), "code")
  expect_error(oauth_flow_auth_code_read("invalid"), "state does not match")
})

test_that("JSON-encoded authorisation codes can be input manually", {
  input <- list(state = "state", code = "code")
  encoded <- openssl::base64_encode(jsonlite::toJSON(input))
  local_mocked_bindings(
    readline = function(prompt = "") encoded
  )
  expect_equal(oauth_flow_auth_code_read("state"), "code")
  expect_error(oauth_flow_auth_code_read("invalid"), "state does not match")
})

test_that("bare authorisation codes can be input manually", {
  state <- base64_url_rand(32)
  sent_code <- FALSE
  local_mocked_bindings(
    readline = function(prompt = "") {
      if (sent_code) {
        state
      } else {
        sent_code <<- TRUE
        "zyx987"
      }
    }
  )
  expect_equal(oauth_flow_auth_code_read(state), "zyx987")
  expect_error(oauth_flow_auth_code_read("invalid"), "state does not match")
})

# normalize_redirect_uri --------------------------------------------------

test_that("adds port to localhost url", {
  # Allow tests to run when is_hosted_session() is TRUE.
  local_mocked_bindings(is_hosted_session = function() FALSE)

  redirect <- normalize_redirect_uri("http://localhost")
  expect_false(is.null(url_parse(redirect$uri)$port))

  redirect <- normalize_redirect_uri("http://127.0.0.1")
  expect_false(is.null(url_parse(redirect$uri)$port))
})

test_that("urls left as is if not changes needed", {
  # Allow tests to run when is_hosted_session() is TRUE.
  local_mocked_bindings(is_hosted_session = function() FALSE)

  original_uri <- "http://localhost:8080"
  normalized_uri <- normalize_redirect_uri(original_uri)
  expect_equal(normalized_uri$uri, original_uri)
})

# ouath_flow_auth_code_parse ----------------------------------------------

test_that("forwards oauth error", {
  query1 <- query2 <- list(error = "123", error_description = "A bad error")
  query2$error_uri <- "http://example.com"
  query3 <- list(state = "def")

  expect_snapshot(error = TRUE, {
    oauth_flow_auth_code_parse(query1, "abc")
    oauth_flow_auth_code_parse(query2, "abc")
    oauth_flow_auth_code_parse(query3, "abc")
  })
})

# can_fetch_auth_code -----------------------------------------------------

test_that("external auth code sources are detected correctly", {
  # False by default.
  expect_false(can_fetch_oauth_code("http://localhost:8080/redirect"))

  # Only true in the presence of certain environment variables.
  env <- c(
    "HTTR2_OAUTH_CODE_SOURCE_URL" = "http://localhost:8080/code",
    "HTTR2_OAUTH_REDIRECT_URL" = "http://localhost:8080/redirect"
  )
  withr::with_envvar(env, {
    expect_true(can_fetch_oauth_code("http://localhost:8080/redirect"))

    # Non-matching redirect URLs should not count as external sources, either.
    expect_false(can_fetch_oauth_code("http://localhost:9090/redirect"))
  })
})

# oauth_flow_auth_code_fetch ----------------------------------------------

test_that("auth codes can be retrieved from an external source", {
  skip_on_cran()
  local_mocked_bindings(sys_sleep = function(...) {})

  req <- local_app_request(function(req, res) {
    # Error on first, and then respond on second
    authorized <- res$app$locals$authorized %||% FALSE
    if (!authorized) {
      res$app$locals$authorized <- TRUE
      res$set_status(404L)$set_type("text/plain")$send("Not found")
    } else {
      res$set_status(200L)$send_json(text = '{"code":"abc123"}')
    }
  })

  withr::local_envvar("HTTR2_OAUTH_CODE_SOURCE_URL" = req$url)
  expect_equal(oauth_flow_auth_code_fetch("ignored"), "abc123")
})

#' Check the content type of a response
#'
#' A different content type than expected often leads to an error in parsing
#' the response body. This function checks that the content type of the response
#' is as expected and fails otherwise.
#'
#' @param valid_types A character vector of valid MIME types. Should only
#'   be specified with `type/subtype`.
#' @param valid_suffix A string given an "structured media type" suffix.
#' @param check_type Should the type actually be checked? Provided as a
#'   convenience for when using this function inside `resp_body_*` helpers.
#' @inheritParams resp_headers
#' @inheritParams rlang::args_error_context
#' @return Called for its side-effect; erroring if the response does not
#'   have the expected content type.
#' @export
#' @examples
#' resp <- response(headers = list(`content-type` = "application/json"))
#' resp_check_content_type(resp, "application/json")
#' try(resp_check_content_type(resp, "application/xml"))
#'
#' # `types` can also specify multiple valid types
#' resp_check_content_type(resp, c("application/xml", "application/json"))
resp_check_content_type <- function(
  resp,
  valid_types = NULL,
  valid_suffix = NULL,
  check_type = TRUE,
  call = caller_env()
) {
  check_response(resp)
  check_character(valid_types, allow_null = TRUE)
  check_string(valid_suffix, allow_null = TRUE)
  check_bool(check_type, allow_na = TRUE)

  if (isFALSE(check_type)) {
    return(invisible())
  }

  check_content_type(
    resp_content_type(resp),
    valid_types = valid_types,
    valid_suffix = valid_suffix,
    inform_check_type = !is.na(check_type),
    call = call
  )
  invisible()
}

parse_content_type <- function(x) {
  # Create regex with {rex} package
  #
  # ```
  # library(rex)
  # # see https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types for the
  # # possible types
  # types <- c("application", "audio", "font", "example", "image", "message", "model", "multipart", "text", "video")
  # regex <- rex(
  #   start,
  #   capture(regex(paste0(types, collapse = "|")), name = "type"),
  #   "/",
  #   capture(
  #     maybe(or("vnd", "prs", "x"), "."),
  #     one_or_more(none_of("+;")),
  #     name = "subtype"
  #   ),
  #   maybe("+", capture(one_or_more(none_of(";")), name = "suffix")),
  #   maybe(";", capture(one_or_more(any), name = "parameters")),
  #   end
  # )
  # unclass(regex)
  # ```
  stopifnot(length(x) == 1)
  regex <- "^(?<type>application|audio|font|example|image|message|model|multipart|text|video)/(?<subtype>(?:(?:vnd|prs|x)\\.)?(?:[^+;])+)(?:\\+(?<suffix>(?:[^;])+))?(?:;(?<parameters>(?:.)+))?$"
  if (!grepl(regex, x, perl = TRUE)) {
    out <- list(
      type = "",
      subtype = "",
      suffix = ""
    )
    return(out)
  }

  match_object <- regexec(regex, x, perl = TRUE)
  match <- regmatches(x, match_object)[[1]]
  list(
    type = match[[2]],
    subtype = match[[3]],
    suffix = if (match[[4]] != "") match[[4]] else ""
  )
}

check_content_type <- function(
  content_type,
  valid_types = NULL,
  valid_suffix = NULL,
  inform_check_type = FALSE,
  call = caller_env()
) {
  parsed <- parse_content_type(content_type)
  base_type <- paste0(parsed$type, "/", parsed$subtype)

  if (is.null(valid_types) || base_type %in% valid_types) {
    return()
  }
  if (!is.null(valid_suffix) && parsed$suffix == valid_suffix) {
    return()
  }

  msg <- "Expecting type {.or {.str {valid_types}}}"
  if (!is.null(valid_suffix)) {
    msg <- paste0(msg, " or suffix {.str {valid_suffix}}.")
  }

  cli::cli_abort(
    c("Unexpected content type {.str {content_type}}.", "*" = msg),
    i = if (inform_check_type) "Override check with `check_type = FALSE`.",
    call = call
  )
}


is_text_type <- function(content_type) {
  if (is.null(content_type)) {
    return(FALSE)
  }

  parsed <- parse_content_type(content_type)
  if (parsed$type == "text") {
    return(TRUE)
  }

  special_cases <- c(
    "application/xml",
    "application/x-www-form-urlencoded",
    "application/json",
    "application/ld+json",
    "multipart/form-data"
  )
  base_type <- paste0(parsed$type, "/", parsed$subtype)
  if (base_type %in% special_cases) {
    return(TRUE)
  }

  FALSE
}

test_that("valid arguments are optional", {
  resp <- response_json()
  expect_no_error(resp_check_content_type(resp))
})

test_that("can check type of response", {
  resp1 <- response(headers = c("Content-type: application/json"))
  resp2 <- response(headers = c("Content-type: xxxxx"))

  expect_no_error(
    resp_check_content_type(resp1, "application/json")
  )
  expect_no_error(
    resp_check_content_type(resp1, "application/xml", check_type = FALSE)
  )
  expect_snapshot(error = TRUE, {
    resp_check_content_type(resp1, "application/xml")
    resp_check_content_type(resp2, "application/xml")
  })
})

test_that("useful error even if no content type", {
  resp <- response()
  expect_snapshot(
    resp_check_content_type(resp, "application/xml"),
    error = TRUE
  )
})

test_that("can parse content type", {
  expect_equal(
    parse_content_type("application/json"),
    list(type = "application", subtype = "json", suffix = "")
  )

  # can parse suffix
  expect_equal(
    parse_content_type("text/html+xml"),
    list(type = "text", subtype = "html", suffix = "xml")
  )

  # parameters don't matter
  expect_equal(
    parse_content_type("text/html+xml;charset=UTF-8"),
    list(type = "text", subtype = "html", suffix = "xml")
  )
})

test_that("invalid type returns empty strings", {
  expect_equal(
    parse_content_type(""),
    list(type = "", subtype = "", suffix = "")
  )
})

test_that("check_content_type() can consult suffixes", {
  expect_no_error(check_content_type("application/json", "application/json"))
  expect_snapshot(
    check_content_type("application/json", "application/xml"),
    error = TRUE
  )

  # works with suffixes
  expect_no_error(check_content_type(
    "application/test+json",
    "application/json",
    "json"
  ))
  expect_snapshot(
    check_content_type("application/test+json", "application/xml", "xml"),
    error = TRUE
  )

  # can use multiple valid types
  expect_no_error(
    check_content_type(
      "application/test+json",
      c("text/html", "application/json"),
      "json"
    )
  )
  expect_snapshot(
    check_content_type("application/xml", c("text/html", "application/json")),
    error = TRUE
  )
})

test_that("can detect text types", {
  expect_true(is_text_type("text/html"))
  expect_true(is_text_type("application/json"))
  expect_false(is_text_type("image/png"))
})

#' Create a HTTP response for testing
#'
#' @description
#' `response()` creates a generic response; `response_json()` creates a
#' response with a JSON body, automatically adding the correct `Content-Type`
#' header.
#'
#' Generally, you should not need to call these function directly; you'll
#' get a real HTTP response by calling [req_perform()] and friends. These
#' function is provided primarily for use in tests; if you are creating
#' responses for mocked requests, use the lower-level [new_response()].
#'
#' @inherit new_response params return
#' @param body The response body. For `response_json()`, a R data structure that
#'   will be serialized to JSON.
#' @export
#' @examples
#' response()
#' response(404, method = "POST")
#' response(headers = c("Content-Type: text/html", "Content-Length: 300"))
response <- function(
  status_code = 200,
  url = "https://example.com",
  method = "GET",
  headers = list(),
  body = raw(),
  timing = NULL
) {
  check_number_whole(status_code, min = 100, max = 700)
  check_string(url)
  check_string(method)

  headers <- as_headers(headers)
  # ensure we always have a date field
  if (!"date" %in% tolower(names(headers))) {
    headers$Date <- "Wed, 01 Jan 2020 00:00:00 UTC"
  }

  new_response(
    method = method,
    url = url,
    status_code = as.integer(status_code),
    headers = headers,
    body = body,
    timing = timing
  )
}

#' @export
#' @rdname response
response_json <- function(
  status_code = 200,
  url = "https://example.com",
  method = "GET",
  headers = list(),
  body = list()
) {
  headers <- as_headers(headers)
  headers$`Content-Type` <- "application/json"

  body <- charToRaw(jsonlite::toJSON(body, auto_unbox = TRUE))

  new_response(
    method = method,
    url = url,
    status_code = as.integer(status_code),
    headers = headers,
    body = body
  )
}

#' Create a HTTP response
#'
#' This is the constructor function for the `httr2_response` S3 class. It is
#' useful primarily for mocking.
#'
#' @param method HTTP method used to retrieve the response.
#' @param url URL response came from; might not be the same as the URL in
#'   the request if there were any redirects.
#' @param status_code HTTP status code. Must be a single integer.
#' @param headers HTTP headers. Can be supplied as a raw or character vector
#'   which will be parsed using the standard rules, or a named list.
#' @param body The body of the response. Can be a raw vector, a `<httr2_path>`,
#'   or a [StreamingBody].
#' @param timing A named numeric vector giving the time taken by various
#'   components.
#' @param request The [request] used to generate this response.
#' @param error_call Environment (on call stack) used in error messages.
#' @returns An HTTP response: an S3 list with class `httr2_response`.
#' @export
new_response <- function(
  method,
  url,
  status_code,
  headers,
  body,
  timing = NULL,
  request = NULL,
  error_call = caller_env()
) {
  check_string(method, call = error_call)
  check_string(url, call = error_call)
  check_number_whole(status_code, call = error_call)
  check_request(request, allow_null = TRUE, call = error_call)
  if (!is.null(timing) && !is_bare_numeric(timing)) {
    stop_input_type(
      timing,
      "a numeric vector",
      allow_null = TRUE,
      call = error_call
    )
  }
  headers <- as_headers(headers, error_call = error_call)
  if (!is.raw(body) && !is_path(body) && !inherits(body, "StreamingBody")) {
    stop_input_type(
      body,
      c("a raw vector", "a path", "a <StreamingBody>"),
      call = error_call
    )
  }

  structure(
    list(
      method = method,
      url = url,
      status_code = status_code,
      headers = headers,
      body = body,
      timing = timing,
      request = request,
      cache = new_environment()
    ),
    class = "httr2_response"
  )
}

create_response <- function(req, curl_data, body) {
  the$last_response <- new_response(
    method = curl_data$method,
    url = curl_data$url,
    status_code = curl_data$status_code,
    headers = as_headers(curl_data$headers),
    body = body,
    timing = curl_data$times,
    request = req
  )

  the$last_response
}


#' @export
print.httr2_response <- function(x, ...) {
  cli::cat_line(cli::format_inline("{.cls {class(x)}}"))
  cli::cat_line(cli::format_inline("{.strong {x$method}} {x$url}"))
  cli::cat_line(cli::format_inline(
    "{.field Status}: {x$status_code} {resp_status_desc(x)}"
  ))
  if (resp_header_exists(x, "Content-Type")) {
    cli::cat_line(cli::format_inline(
      "{.field Content-Type}: {resp_content_type(x)}"
    ))
  }

  body <- x$body
  if (!resp_has_body(x)) {
    cli::cat_line(cli::format_inline("{.field Body}: None"))
  } else {
    switch(
      resp_body_type(x),
      disk = cli::cat_line(cli::format_inline(
        "{.field Body}: On disk {.path {body}} ({file.size(body)} bytes)"
      )),
      memory = cli::cat_line(cli::format_inline(
        "{.field Body}: In memory ({length(body)} bytes)"
      )),
      stream = cli::cat_line(cli::format_inline(
        "{.field Body}: Streaming connection"
      ))
    )
  }

  invisible(x)
}

#' Show the raw response
#'
#' This function reconstructs the HTTP message that httr2 received from the
#' server. It's unlikely to be exactly byte-for-byte identical (because most
#' servers compress at least the body, and HTTP/2 can also compress the
#' headers), but it conveys the same information.
#'
#' @inheritParams resp_headers
#' @returns `resp` (invisibly).
#' @export
#' @examples
#' resp <- request(example_url()) |>
#'   req_url_path("/json") |>
#'   req_perform()
#' resp |> resp_raw()
resp_raw <- function(resp) {
  cli::cat_line("HTTP/1.1 ", resp$status_code, " ", resp_status_desc(resp))
  cli::cat_line(cli::style_bold(names(resp$headers)), ": ", resp$headers)
  cli::cat_line()
  if (!is.null(resp$body)) {
    cli::cat_line(resp_body_string(resp))
  }

  invisible(resp)
}

is_response <- function(x) {
  inherits(x, "httr2_response")
}

check_response <- function(resp, arg = caller_arg(resp), call = caller_env()) {
  if (!missing(resp) && is_response(resp)) {
    return(invisible(NULL))
  }

  stop_input_type(
    resp,
    "an HTTP response object",
    allow_null = FALSE,
    arg = arg,
    call = call
  )
}

test_that("response has basic print method", {
  file.create("path-empty")
  writeBin("sample content", "path-content")

  withr::defer(unlink(c("path-empty", "path-content")))
  streaming <- StreamingBody$new(file())
  withr::defer(streaming$close())

  expect_snapshot({
    response(200)
    response(200, headers = "Content-Type: text/html")
    response(200, body = charToRaw("abcdef"))
    response(200, body = new_path("path-empty"))
    response(200, body = new_path("path-content"))
    response(200, body = streaming)
  })
})

test_that("response adds date if not provided by server", {
  resp <- response(headers = "Test: 1")
  expect_named(resp_headers(resp), c("Test", "Date"))
})

test_that("check_response produces helpful error", {
  expect_snapshot(check_response(1), error = TRUE)
})

test_that("new_response() checks its inputs", {
  expect_snapshot(error = TRUE, {
    new_response(1)
    new_response("GET", 1)
    new_response("GET", "http://x.com", "x")
    new_response("GET", "http://x.com", 200, 1)
    new_response("GET", "http://x.com", 200, list(), 1)
    new_response("GET", "http://x.com", 200, list(), raw(), "x")
    new_response("GET", "http://x.com", 200, list(), raw(), c(x = 1), 1)
  })
})

# oauth_flow_fetch --------------------------------------------------------

test_that("turns oauth errors to R errors", {
  req <- request("http://example.com")
  local_mocked_bindings(req_perform = function(...) {
    response_json(400L, body = list(error = "1", error_description = "abc"))
  })

  expect_snapshot(oauth_flow_fetch(req, "test"), error = TRUE)
})

# oauth_flow_parse --------------------------------------------------------

test_that("userful errors if response isn't parseable", {
  resp1 <- response(headers = list(`content-type` = "text/plain"))
  resp2 <- response_json(body = list())

  expect_snapshot(error = TRUE, {
    oauth_flow_parse(resp1, "test")
    oauth_flow_parse(resp2, "test")
  })
})

test_that("can inspect the original response if response isn't parseable", {
  resp1 <- response(headers = list(`content-type` = "text/plain"))
  resp2 <- response_json(body = list())

  tryCatch(
    oauth_flow_parse(resp1, "test"),
    httr2_oauth_parse = function(cnd) {
      expect_equal(cnd$resp, resp1)
    }
  )
  tryCatch(
    oauth_flow_parse(resp2, "test"),
    httr2_oauth_parse = function(cnd) {
      expect_equal(cnd$resp, resp2)
    }
  )
})

test_that("returns body if known good structure", {
  resp <- response_json(body = list(access_token = "10"))
  expect_equal(oauth_flow_parse(resp, "test"), list(access_token = "10"))

  resp <- response_json(body = list(device_code = "10"))
  expect_equal(oauth_flow_parse(resp, "test"), list(device_code = "10"))

  resp <- response_json(403L, body = list(error = "10"))
  expect_snapshot(oauth_flow_parse(resp, "test"), error = TRUE)
})

test_that("converts expires_in to numeric", {
  resp <- response_json(
    200L,
    body = list(access_token = "10", expires_in = "20")
  )
  body <- oauth_flow_parse(resp, "test")
  expect_equal(body$expires_in, 20)
})

test_that("get some useful output from WWW-Authenticate header", {
  resp <- response(
    401,
    headers = 'WWW-Authenticate: Bearer realm="example",error="invalid_token",error_description="The access token expired"'
  )
  expect_snapshot_error(resp_check_status(resp))

  resp <- response(
    403,
    headers = 'WWW-Authenticate: Bearer realm="https://accounts.google.com/", error="insufficient_scope", scope="https://www.googleapis.com/auth/iam https://www.googleapis.com/auth/cloud-platform"'
  )
  expect_snapshot_error(resp_check_status(resp))
})

parse_media <- function(x) {
  # https://datatracker.ietf.org/doc/html/rfc2616#section-3.7
  pieces <- parse_delim(x, ";")

  if (is_empty(pieces)) {
    list(type = NA_character_)
  } else {
    params <- parse_name_equals_value(pieces[-1])
    c(list(type = pieces[[1]]), params)
  }
}

parse_www_authenticate <- function(x) {
  # Seems little general support for multiple schemes in one header
  # https://stackoverflow.com/questions/10239970

  pieces <- parse_in_half(x, " ")
  params <- parse_name_equals_value(parse_delim(pieces$right, ","))

  c(list(scheme = pieces$left), params)
}

parse_link <- function(x) {
  links <- parse_delim(x, ",")

  parse_one <- function(x) {
    pieces <- parse_delim(x, ";")

    url <- gsub("^<|>$", "", pieces[[1]])
    params <- parse_name_equals_value(pieces[-1])
    c(list(url = url), params)
  }

  lapply(links, parse_one)
}

# Helpers -----------------------------------------------------------------

parse_delim <- function(x, delim, quote = "\"", ...) {
  # Use scan to deal with quoted strings. It loses the quotes, but it's
  # ok because the field name can't be a quoted string so there's no ambiguity
  # about who the = belongs to.
  scan(
    text = x,
    what = character(),
    sep = delim,
    quote = quote,
    quiet = TRUE,
    strip.white = TRUE,
    ...
  )
}

parse_name_equals_value <- function(x) {
  halves <- parse_in_half(x, "=")
  set_names(halves$right, halves$left)
}

parse_in_half <- function(x, char = "=") {
  match <- regexpr(char, x, fixed = TRUE)
  match_loc <- as.vector(match)
  match_len <- attr(match, "match.length")

  left_start <- 1
  left_end <- match_loc - 1
  right_start <- match_loc + match_len
  right_end <- nchar(x)

  no_match <- match_loc == -1
  left_end[no_match] <- right_end[no_match]
  right_start[no_match] <- 0
  right_end[no_match] <- 0

  list(
    left = substr(x, left_start, left_end),
    right = substr(x, right_start, right_end)
  )
}

parse_match <- function(x, pattern) {
  match_loc <- regexpr(pattern, x, perl = TRUE)
  cap_start <- attr(match_loc, "capture.start")
  cap_len <- attr(match_loc, "capture.length")
  cap_end <- cap_start + cap_len - 1
  cap_end[cap_end == -1] <- 0
  pieces <- as.list(substring(x, cap_start, cap_end))
  pieces[pieces == ""] <- list(NULL)
  pieces
}

test_that("can parse media type", {
  # no params
  expect_equal(parse_media("text/plain"), list(type = "text/plain"))

  # single param
  expect_equal(
    parse_media("text/plain; charset=utf-8"),
    list(type = "text/plain", charset = "utf-8")
  )

  # single param with quotes
  expect_equal(
    parse_media("text/plain; charset=\"utf-8\""),
    list(type = "text/plain", charset = "utf-8")
  )

  # quoted param containing ;
  expect_equal(
    parse_media("text/plain; charset=\";\""),
    list(type = "text/plain", charset = ";")
  )

  expect_equal(parse_media(""), list(type = NA_character_))
})

test_that("can parse authenticate header", {
  header <- paste0(
    'Bearer realm="example",',
    'error="invalid_token",',
    'error_description="The access token expired"'
  )
  out <- parse_www_authenticate(header)
  expect_equal(out$scheme, "Bearer")
  expect_equal(out$realm, "example")
  expect_equal(out$error_description, "The access token expired")
})

test_that("can parse links", {
  header <- paste0(
    '<https://example.com/1>; rel="next",',
    '<https://example.com/2>; rel="last"'
  )
  expect_equal(
    parse_link(header),
    list(
      list(url = "https://example.com/1", rel = "next"),
      list(url = "https://example.com/2", rel = "last")
    )
  )
})

# Helpers -----------------------------------------------------------------

test_that("parse_in_half handles common cases", {
  parsed <- parse_in_half(c("a=b", "c=d", "e", "=f", "g=", "h=i=j"), "=")
  expect_equal(parsed$left, c("a", "c", "e", "", "g", "h"))
  expect_equal(parsed$right, c("b", "d", "", "f", "", "i=j"))
})

test_that("parse_in_half handles problematic inputs", {
  expect_equal(
    parse_in_half(character(0), "="),
    list(left = character(0), right = character(0))
  )
  expect_equal(
    parse_in_half("", "="),
    list(left = "", right = "")
  )
  expect_equal(
    parse_in_half(NA, "="),
    list(left = NA_character_, right = NA_character_)
  )
})

test_that("parse_in_half always returns two pieces", {
  expect_equal(parse_in_half("a", " "), list(left = "a", right = ""))
  expect_equal(parse_in_half("a b", " "), list(left = "a", right = "b"))
  expect_equal(parse_in_half("a b c", " "), list(left = "a", right = "b c"))
})

test_that("parse_name_equals_value handles empty values", {
  expect_equal(parse_name_equals_value("a"), c(a = ""))
})

test_that("parse_match converts missing matches to NULL", {
  expect_equal(
    parse_match("abbbd", "(a)(b+)(c*)(d)"),
    list("a", "bbb", NULL, "d")
  )
})

#' Perform requests iteratively, generating new requests from previous responses
#'
#' @description
#' `req_perform_iterative()` iteratively generates and performs requests,
#' using a callback function, `next_req`, to define the next request based on
#' the current request and response. You will probably want to pair it with an
#' [iteration helper][iterate_with_offset] and use a
#' [multi-response handler][resps_successes] to process the result.
#'
#' # `next_req()`
#'
#' The key piece that makes `req_perform_iterative()` work is the `next_req()`
#' argument. For most common cases, you can use one of the canned helpers,
#' like [iterate_with_offset()]. If, however, the API you're wrapping uses a
#' different pagination system, you'll need to write your own. This section
#' gives some advice.
#'
#' Generally, your function needs to inspect the response, extract some data
#' from it, then use that to modify the previous request. For example, imagine
#' that the response returns a cursor, which needs to be added to the body of
#' the request. The simplest version of this function might look like this:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   cursor <- resp_body_json(resp)$next_cursor
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' There's one problem here: if there are no more pages to return, then
#' `cursor` will be `NULL`, but `req_body_json_modify()` will still generate
#' a meaningful request. So we need to handle this specifically by
#' returning `NULL`:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   cursor <- resp_body_json(resp)$next_cursor
#'   if (is.null(cursor))
#'     return(NULL)
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' A value of `NULL` lets `req_perform_iterative()` know there are no more
#' pages remaining.
#'
#' There's one last feature you might want to add to your iterator: if you
#' know the total number of pages, then it's nice to let
#' `req_perform_iterative()` know so it can adjust the progress bar.
#' (This will only ever decrease the number of pages, not increase it.)
#' You can signal the total number of pages by calling [signal_total_pages()],
#' like this:
#'
#' ```R
#' next_req <- function(resp, req) {
#'   body <- resp_body_json(resp)
#'   cursor <- body$next_cursor
#'   if (is.null(cursor))
#'     return(NULL)
#'
#'   signal_total_pages(body$pages)
#'   req |> req_body_json_modify(cursor = cursor)
#' }
#' ```
#'
#' @inheritParams req_perform_sequential
#' @param req The first [request] to perform.
#' @param next_req A function that takes the previous response (`resp`) and
#'   request (`req`) and returns a [request] for the next page or `NULL` if
#'   the iteration should terminate. See below for more details.
#' @param max_reqs The maximum number of requests to perform. Use `Inf` to
#'   perform all requests until `next_req()` returns `NULL`.
#' @param on_error What should happen if a request fails?
#'
#'   * `"stop"`, the default: stop iterating with an error.
#'   * `"return"`: stop iterating, returning all the successful responses so
#'     far, as well as an error object for the failed request.
#' @param path Optionally, path to save the body of request. This should be
#'   a glue string that uses `{i}` to distinguish different requests.
#'   Useful for large responses because it avoids storing the response in
#'   memory.
#' @return
#' A list, at most length `max_reqs`, containing [response]s and possibly one
#' error object, if `on_error` is `"return"` and one of the requests errors.
#' If present, the error object will always be the last element in the list.
#'
#' Only httr2 errors are captured; see [req_error()] for more details.
#' @export
#' @examples
#' req <- request(example_url()) |>
#'   req_url_path("/iris") |>
#'   req_throttle(10) |>
#'   req_url_query(limit = 5)
#'
#' resps <- req_perform_iterative(req, iterate_with_offset("page_index"))
#'
#' data <- resps |> resps_data(function(resp) {
#'   data <- resp_body_json(resp)$data
#'   data.frame(
#'     Sepal.Length = sapply(data, `[[`, "Sepal.Length"),
#'     Sepal.Width = sapply(data, `[[`, "Sepal.Width"),
#'     Petal.Length = sapply(data, `[[`, "Petal.Length"),
#'     Petal.Width = sapply(data, `[[`, "Petal.Width"),
#'     Species = sapply(data, `[[`, "Species")
#'   )
#' })
#' str(data)
req_perform_iterative <- function(
  req,
  next_req,
  path = NULL,
  max_reqs = 20,
  on_error = c("stop", "return"),
  mock = getOption("httr2_mock", NULL),
  progress = TRUE
) {
  check_request(req)
  check_function2(next_req, args = c("resp", "req"))
  check_number_whole(max_reqs, allow_infinite = TRUE, min = 1)
  check_string(path, allow_empty = FALSE, allow_null = TRUE)
  on_error <- arg_match(on_error)
  mock <- as_mock_function(mock, error_call)

  get_path <- function(i) {
    if (is.null(path)) {
      NULL
    } else {
      glue::glue(path)
    }
  }

  progress <- create_progress_bar(progress, max_reqs)

  resps <- vector("list", length = if (is.finite(max_reqs)) max_reqs else 100)
  i <- 1L

  tryCatch(
    repeat {
      httr2_error <- switch(
        on_error,
        stop = function(cnd) zap(),
        return = function(cnd) cnd
      )
      resp <- try_fetch(
        req_perform(req, path = get_path(i), mock = mock),
        httr2_error = httr2_error
      )
      resps[[i]] <- resp

      if (on_error == "return" && is_error(resp)) {
        break
      }

      progress$update()

      withCallingHandlers(
        {
          req <- next_req(resp = resp, req = req)
        },
        httr2_total_pages = function(cnd) {
          # Allow next_req() to shrink the number of pages remaining
          # Most important in max_req = Inf case
          if (cnd$n < max_reqs) {
            max_reqs <<- cnd$n
            progress$update(total = max_reqs, inc = 0)
          }
        }
      )

      if (is.null(req) || i >= max_reqs) {
        break
      }
      check_request(req, arg = "next_req()")

      i <- i + 1L
      if (i > length(resps)) {
        signal("", class = "httr2:::doubled")
        length(resps) <- length(resps) * 2
      }
    },
    interrupt = function(cnd) {
      check_repeated_interrupt()

      # interrupt might occur after i was incremented
      if (is.null(resps[[i]])) {
        i <<- i - 1
      }
      cli::cli_alert_warning(
        "Terminating iteration; returning {i} response{?s}."
      )
    }
  )
  progress$done()

  if (i < length(resps)) {
    resps <- resps[seq_len(i)]
  }

  resps
}

pooled_request <- function(
  req,
  path = NULL,
  on_success = NULL,
  on_failure = NULL,
  on_error = NULL,
  mock = NULL,
  error_call = caller_env()
) {
  check_request(req)
  check_string(path, allow_null = TRUE)
  check_function2(on_success, args = "resp", allow_null = TRUE)
  check_function2(on_failure, args = "error", allow_null = TRUE)
  check_function2(on_error, args = "error", allow_null = TRUE)

  PooledRequest$new(
    req = req,
    path = path,
    error_call = error_call,
    on_success = on_success,
    on_failure = on_failure,
    on_error = on_error,
    mock = mock
  )
}

# Wrap up all components of request -> response in a single object
PooledRequest <- R6Class(
  "Performance",
  public = list(
    req = NULL,
    resp = NULL,

    initialize = function(
      req,
      path = NULL,
      error_call = NULL,
      on_success = NULL,
      on_failure = NULL,
      on_error = NULL,
      mock = NULL
    ) {
      self$req <- req
      private$path <- path
      private$error_call <- error_call
      private$on_success <- on_success
      private$on_failure <- on_failure
      private$on_error <- on_error
      private$mock <- mock
    },

    submit = function(pool) {
      if (!is.null(private$mock)) {
        mock_resp <- private$mock(self$req)
        if (!is.null(mock_resp)) {
          private$handle_response(mock_resp, self$req)
          return()
        }
      }

      req <- cache_pre_fetch(self$req, private$path)
      if (is_response(req)) {
        private$handle_response(req, self$req)
        return()
      }

      private$req_prep <- req_prepare(req)
      private$handle <- req_handle(private$req_prep)
      # Note: we need to do this before we call handle_preflight() so that
      # request signing works correctly with the added headers.
      #
      # TODO: Support resend_count.
      private$req_prep <- req_with_span(
        private$req_prep,
        # Pooled request spans should not become the active span; we want
        # subsequent requests to be siblings rather than parents.
        activate = FALSE
      )
      handle_preflight(private$req_prep, private$handle)

      curl::multi_add(
        handle = private$handle,
        pool = pool,
        data = private$path,
        done = private$succeed,
        fail = private$fail
      )

      invisible(self)
    },

    cancel = function() {
      # No handle if response was cached
      if (!is.null(private$handle)) {
        curl::multi_cancel(private$handle)
      }
      req_record_span_status(private$req_prep)
    }
  ),
  private = list(
    path = NULL,
    error_call = NULL,
    pool = NULL,

    req_prep = NULL,
    handle = NULL,

    on_success = NULL,
    on_failure = NULL,
    on_error = NULL,
    mock = NULL,

    # curl success could be httr2 success or httr2 failure
    succeed = function(curl_data) {
      private$handle <- NULL
      req_completed(private$req_prep)

      if (is.null(private$path)) {
        body <- curl_data$content
      } else {
        # Only needed with curl::multi_run()
        if (!file.exists(private$path)) {
          file.create(private$path)
        }
        body <- new_path(private$path)
      }

      resp <- create_response(self$req, curl_data, body)
      req_record_span_status(private$req_prep, resp)
      resp <- cache_post_fetch(self$req, resp, path = private$path)
      private$handle_response(resp, self$req)
    },

    handle_response = function(resp, req) {
      if (error_is_error(req, resp)) {
        cnd <- resp_failure_cnd(req, resp, error_call = private$error_call)
        private$on_failure(cnd)
      } else {
        private$on_success(resp)
      }
    },

    # curl failure = httr2 error
    fail = function(msg) {
      private$handle <- NULL
      req_completed(private$req_prep)

      error_class <- setdiff(class(msg), "character")
      curl_error <- error_cnd(message = msg, class = error_class, call = NULL)
      error <- curl_cnd(curl_error, call = private$error_call)
      error$request <- self$req
      req_record_span_status(private$req_prep, error)
      private$on_error(error)
    }
  )
)

req_policies <- function(.req, ..., error_call = caller_env()) {
  check_request(.req, call = error_call)
  .req$policies <- modify_list(.req$policies, ..., error_call = error_call)
  .req
}

req_policy_exists <- function(req, name) {
  has_name(req$policies, name)
}

req_policy_call <- function(req, name, args, default) {
  if (req_policy_exists(req, name)) {
    exec(req$policies[[name]], !!!args)
  } else {
    if (is_function(default)) {
      exec(default, !!!args)
    } else {
      default
    }
  }
}

as_callback <- function(x, n, name, error_call = caller_env()) {
  if (is.null(x)) {
    return(x)
  }

  x <- as_function(x)
  if (!inherits(x, "rlang_lambda_function") && length(formals(x)) != n) {
    cli::cli_abort(
      "Callback {.fn name} must have {n} argument{?s}",
      call = error_call
    )
  }
  x
}

#' Temporarily mock requests
#'
#' Mocking allows you to selectively and temporarily replace the response
#' you would typically receive from a request with your own code. These
#' functions are low-level and we don't recommend using them directly.
#' Instead use package that uses these functions under the hood, like
#' \pkg{httptest2} or \pkg{vcr}.
#'
#' @param mock A function, a list, or `NULL`.
#'
#'   * `NULL` disables mocking and returns httr2 to regular operation.
#'
#'   * A list of responses will be returned in sequence. After all responses
#'     have been used up, will return 503 server errors.
#'
#'   * For maximum flexibility, you can supply a function that that takes a
#'     single argument, `req`, and returns either `NULL` (if it doesn't want to
#'     handle the request) or a [response] (if it does).
#'
#' @param code Code to execute in the temporary environment.
#' @param env Environment to use for scoping changes.
#' @returns `with_mocked_responses()` returns the result of evaluating `code`.
#' @export
#' @examples
#' # This function should perform a response against google.com:
#' google <- function() {
#'   request("http://google.com") |>
#'     req_perform()
#' }
#'
#' # But I can use a mock to instead return my own made up response:
#' my_mock <- function(req) {
#'   response(status_code = 403)
#' }
#' try(with_mocked_responses(my_mock, google()))
with_mocked_responses <- function(mock, code) {
  mock <- as_mock_function(mock)
  withr::with_options(list(httr2_mock = mock), code)
}
#' @export
#' @rdname with_mocked_responses
local_mocked_responses <- function(mock, env = caller_env()) {
  mock <- as_mock_function(mock)
  withr::local_options(httr2_mock = mock, .local_envir = env)
}

as_mock_function <- function(mock, error_call = caller_env()) {
  if (is.null(mock)) {
    mock
  } else if (is.function(mock)) {
    check_function2(mock, args = "req", call = error_call)
    mock
  } else if (is_formula(mock)) {
    mock <- as_function(mock, call = error_call)
  } else if (is_bare_list(mock)) {
    mocked_response_sequence(!!!mock)
  } else {
    stop_input_type(mock, c("function", "list", "NULL"), call = error_call)
  }
}

mocked_response_sequence <- function(...) {
  responses <- list2(...)

  n <- length(responses)
  i <- 0
  function(req) {
    if (i >= n) {
      response(503)
    } else {
      i <<- i + 1
      responses[[i]]
    }
  }
}

test_that("can override requests with local_ or with_", {
  req <- request("https://google.com")
  resp <- response()

  with_mocked_responses(\(req) resp, {
    expect_equal(req_perform(req), resp)
  })

  local_mocked_responses(\(req) resp)
  expect_equal(req_perform(req), resp)
})

test_that("mocked_response_sequence returns responses then errors", {
  local_mocked_responses(list(
    response(200),
    response(201)
  ))

  req <- request("https://google.com")
  expect_equal(req_perform(req), response(200))
  expect_equal(req_perform(req), response(201))
  expect_error(req_perform(req), class = "httr2_http_503")
})

test_that("validates inputs", {
  expect_snapshot(error = TRUE, {
    local_mocked_responses(function(foo) {})
    local_mocked_responses(10)
  })
})

#' Authenticate request with HTTP basic authentication
#'
#' This sets the Authorization header. See details at
#' <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Authorization>.
#'
#' @inheritParams req_perform
#' @param username User name.
#' @param password Password. You should avoid entering the password directly
#'   when calling this function as it will be captured by `.Rhistory`. Instead,
#'   leave it unset and the default behaviour will prompt you for it
#'   interactively.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' req <- request("http://example.com") |> req_auth_basic("hadley", "SECRET")
#' req
#' req |> req_dry_run()
#'
#' # httr2 does its best to redact the Authorization header so that you don't
#' # accidentally reveal confidential data. Use `redact_headers` to reveal it:
#' print(req, redact_headers = FALSE)
#' req |> req_dry_run(redact_headers = FALSE)
#'
#' # We do this because the authorization header is not encrypted and the
#' # so password can easily be discovered:
#' rawToChar(jsonlite::base64_dec("aGFkbGV5OlNFQ1JFVA=="))
req_auth_basic <- function(req, username, password = NULL) {
  check_request(req)
  check_string(username)
  password <- check_password(password)

  username_password <- openssl::base64_encode(paste0(username, ":", password))
  req_headers(req, Authorization = paste0("Basic ", username_password))
}

#' Authenticate request with bearer token
#'
#' A bearer token gives the bearer access to confidential resources
#' (so you should keep them secure like you would with a user name and
#' password). They are usually produced by some large authentication scheme
#' (like the various OAuth 2.0 flows), but you are sometimes given then
#' directly.
#'
#' @seealso See `r rfc(6750)` for more details about bearer token usage
#'   with OAuth 2.0.
#' @inheritParams req_perform
#' @param token A bearer token
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' req <- request("http://example.com") |> req_auth_bearer_token("sdaljsdf093lkfs")
#' req
#'
#' # httr2 does its best to redact the Authorization header so that you don't
#' # accidentally reveal confidential data. Use `redact_headers` to reveal it:
#' print(req, redact_headers = FALSE)
req_auth_bearer_token <- function(req, token) {
  check_request(req)
  check_string(token)
  req_headers(req, Authorization = paste("Bearer", token))
}

test_that("can send username/password", {
  user <- "u"
  password <- "p"
  req1 <- request_test("/basic-auth/:user/:password")
  req2 <- req1 |> req_auth_basic(user, password)
  expect_redacted(req2, "Authorization")

  expect_error(req_perform(req1), class = "httr2_http_401")
  expect_error(req_perform(req2), NA)
})

test_that("can send bearer token", {
  req <- req_auth_bearer_token(request_test(), "abc")
  expect_equal(
    headers_flatten(req$headers, redact = FALSE),
    list(Authorization = "Bearer abc")
  )
})

#' OAuth token exchange
#'
#' @description
#' Authenticate by exchanging one security token for another, as defined by
#' `r rfc(8693, 2)`. It is typically used for advanced authorization flows that
#' involve "delegation" or "impersonation" semantics, such as when a client
#' accesses a resource on behalf of another party, or when a client's identity
#' is federated from another provider.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>.
#'
#' @export
#' @family OAuth flows
#' @inheritParams req_perform
#' @inheritParams req_oauth_auth_code
#' @param subject_token The security token to exchange. This is usually an
#'   OpenID Connect ID token or a SAML2 assertion.
#' @param subject_token_type A URI that describes the type of the security
#'   token. Usually one of the options in `r rfc(8693, 3)`.
#' @param resource The URI that identifies the resource that the client is
#'   trying to access, if applicable.
#' @param audience The logical name that identifies the resource that the client
#'   is trying to access, if applicable. Usually one of `resource` or `audience`
#'   must be supplied.
#' @param requested_token_type An optional URI that describes the type of the
#'   security token being requested. Usually one of the options in
#'   `r rfc(8693, 3)`.
#' @param actor_token An optional security token that represents the client,
#'   rather than the identity behind the subject token.
#' @param actor_token_type When `actor_token` is not `NULL`, this must be the
#'   URI that describes the type of the security token being requested. Usually
#'   one of the options in `r rfc(8693, 3)`.
#' @returns `req_oauth_token_exchange()` returns a modified HTTP [request] that
#'   will exchange one security token for another; `oauth_flow_token_exchange()`
#'   returns the resulting [oauth_token] directly.
#'
#' @examples
#' # List Google Cloud storage buckets using an OIDC token obtained
#' # from e.g. Microsoft Entra ID or Okta and federated to Google. (A real
#' # project ID and workforce pool would be required for this in practice.)
#' #
#' # See: https://cloud.google.com/iam/docs/workforce-obtaining-short-lived-credentials
#' oidc_token <- "an ID token from Okta"
#' request("https://storage.googleapis.com/storage/v1/b?project=123456") |>
#'   req_oauth_token_exchange(
#'     client = oauth_client("gcp", "https://sts.googleapis.com/v1/token"),
#'     subject_token = oidc_token,
#'     subject_token_type = "urn:ietf:params:oauth:token-type:id_token",
#'     scope = "https://www.googleapis.com/auth/cloud-platform",
#'     requested_token_type = "urn:ietf:params:oauth:token-type:access_token",
#'     audience = "//iam.googleapis.com/locations/global/workforcePools/123/providers/456",
#'     token_params = list(
#'       options = '{"userProject":"123456"}'
#'     )
#'   )
req_oauth_token_exchange <- function(
  req,
  client,
  subject_token,
  subject_token_type,
  resource = NULL,
  audience = NULL,
  scope = NULL,
  requested_token_type = NULL,
  actor_token = NULL,
  actor_token_type = NULL,
  token_params = list()
) {
  params <- list(
    client = client,
    subject_token = subject_token,
    subject_token_type = subject_token_type,
    resource = resource,
    audience = audience,
    scope = scope,
    requested_token_type = requested_token_type,
    actor_token = actor_token,
    actor_token_type = actor_token_type,
    token_params = token_params
  )
  cache <- cache_mem(client, NULL)
  req_oauth(req, "oauth_flow_token_exchange", params, cache = cache)
}

#' @export
#' @rdname req_oauth_token_exchange
oauth_flow_token_exchange <- function(
  client,
  subject_token,
  subject_token_type,
  resource = NULL,
  audience = NULL,
  scope = NULL,
  requested_token_type = NULL,
  actor_token = NULL,
  actor_token_type = NULL,
  token_params = list()
) {
  oauth_client_get_token(
    client,
    grant_type = "urn:ietf:params:oauth:grant-type:token-exchange",
    subject_token = subject_token,
    subject_token_type = subject_token_type,
    resource = resource,
    audience = audience,
    scope = scope,
    requested_token_type = requested_token_type,
    actor_token = actor_token,
    actor_token_type = actor_token_type,
    !!!token_params
  )
}

#' OAuth with a refresh token
#'
#' @description
#' Authenticate using a **refresh token**, following the process described in
#' `r rfc(6749, 6)`.
#'
#' This technique is primarily useful for testing: you can manually retrieve
#' a OAuth token using another OAuth flow (e.g. with [oauth_flow_auth_code()]),
#' extract the refresh token from the result, and then save in an environment
#' variable for use in automated tests.
#'
#' When requesting an access token, the server may also return a new refresh
#' token. If this happens, `oauth_flow_refresh()` will warn, and you'll have
#' retrieve a new update refresh token and update the stored value. If you find
#' this happening a lot, it's a sign that you should be using a different flow
#' in your automated tests.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>.
#'
#' @inheritParams req_oauth_auth_code
#' @param refresh_token A refresh token. This is equivalent to a password
#'   so shouldn't be typed into the console or stored in a script. Instead,
#'   we recommend placing in an environment variable; the default behaviour
#'   is to look in `HTTR2_REFRESH_TOKEN`.
#' @returns `req_oauth_refresh()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_refresh()` returns an [oauth_token].
#' @family OAuth flows
#' @export
#' @examples
#' client <- oauth_client("example", "https://example.com/get_token")
#' req <- request("https://example.com")
#' req |> req_oauth_refresh(client)
req_oauth_refresh <- function(
  req,
  client,
  refresh_token = Sys.getenv("HTTR2_REFRESH_TOKEN"),
  scope = NULL,
  token_params = list()
) {
  params <- list(
    client = client,
    refresh_token = refresh_token,
    scope = scope,
    token_params = token_params
  )
  cache <- cache_mem(client, refresh_token)

  req_oauth(req, "oauth_flow_refresh", params, cache = cache)
}

#' @export
#' @rdname req_oauth_refresh
oauth_flow_refresh <- function(
  client,
  refresh_token = Sys.getenv("HTTR2_REFRESH_TOKEN"),
  scope = NULL,
  token_params = list()
) {
  oauth_flow_check("refresh", client)
  token <- token_refresh(
    client,
    refresh_token = refresh_token,
    scope = scope,
    token_params = token_params
  )

  # Should generally do this automatically, but in this workflow the token will
  # often be stored in an env var or similar
  if (!is.null(token$refresh_token) && token$refresh_token != refresh_token) {
    warn("Refresh token has changed! Please update stored value")
  }

  token
}

test_that("cache considers refresh_token", {
  client <- oauth_client("example", "https://example.com/get_token")
  req <- request("https://example.com")

  # create 2 requests with different refresh token
  req1 <- req |>
    req_oauth_refresh(client, refresh_token = "rt1")
  req2 <- req |>
    req_oauth_refresh(client, refresh_token = "rt2")

  # cache must be empty
  expect_equal(req1$policies$auth_sign$cache$get(), NULL)
  expect_equal(req2$policies$auth_sign$cache$get(), NULL)

  # simulate that we made a request and got back a token
  token <- oauth_token(
    access_token = "a",
    token_type = "bearer",
    expires_in = NULL,
    refresh_token = "rt1",
    .date = Sys.time()
  )
  # ... that is now cached
  req1$policies$auth_sign$cache$set(token)

  # req1 cache must be filled, but req2 cache still be empty
  expect_equal(req1$policies$auth_sign$cache$get(), token)
  expect_equal(req2$policies$auth_sign$cache$get(), NULL)
})

test_that("warns if refresh token changes", {
  client <- oauth_client("example", "https://example.com/get_token")
  local_mocked_bindings(
    token_refresh = function(...) list(refresh_token = "def")
  )

  expect_snapshot(. <- oauth_flow_refresh(client, "abc"))
})

#' Discover OAuth server metadata
#'
#' @description
#' `oauth_server_metadata()` fetches and parses an OAuth 2.0 Authorization
#' Server Metadata document (`r rfc(8414)`) or OpenID Connect Discovery
#' document, returning the endpoints advertised by an issuer.
#'
#' Use it to discover values like `authorization_endpoint`, `token_endpoint`,
#' and `device_authorization_endpoint` rather than hard-coding them. Pass the
#' result to [oauth_client()] as `metadata` and the endpoints are wired up for
#' you:
#'
#' ```r
#' metadata <- oauth_server_metadata("https://accounts.google.com")
#' client <- oauth_client("id", metadata = metadata, secret = "...")
#' token <- oauth_flow_auth_code(client)
#' ```
#'
#' As a security measure, the `issuer` reported in the returned document is
#' validated against the requested `issuer` (`r rfc(8414, "3.3")`); a mismatch
#' is an error. This check is skipped when `url` is supplied without `issuer`.
#'
#' @param issuer The issuer URL, e.g. `"https://accounts.google.com"`. The
#'   metadata URL is derived from it according to `type`.
#' @param type Which well-known suffix to use when `url` is not supplied:
#'
#'   * `"openid"` (the default) appends `/.well-known/openid-configuration`,
#'     the form served by essentially every major provider. Despite the name,
#'     it is a superset that also advertises the OAuth endpoints, so it is the
#'     better default even for plain OAuth.
#'   * `"oauth"` inserts `/.well-known/oauth-authorization-server` between the
#'     origin and any path, as defined in `r rfc(8414)`. Use this for the few
#'     providers that serve only the OAuth document.
#' @param url Optionally, the full metadata document URL. Use this as an escape
#'   hatch for providers that follow neither well-known convention. When
#'   supplied, `issuer` is only used for validation and can be omitted.
#' @returns An S3 list with class `httr2_oauth_server_metadata` containing the
#'   full parsed metadata document. Endpoints that the provider does not
#'   advertise are simply absent.
#' @export
#' @family OAuth flows
#' @examplesIf is_online()
#' oauth_server_metadata("https://accounts.google.com")
oauth_server_metadata <- function(
  issuer,
  type = c("openid", "oauth"),
  url = NULL
) {
  type <- arg_match(type)
  check_string(url, allow_null = TRUE)

  if (missing(issuer)) {
    if (is.null(url)) {
      cli::cli_abort("Must supply either {.arg issuer} or {.arg url}.")
    }
    issuer <- NULL
  } else {
    check_string(issuer)
  }

  url <- url %||% oauth_metadata_url(issuer, type)

  metadata <- request(url) |>
    req_perform() |>
    resp_body_json(simplifyVector = TRUE, simplifyDataFrame = FALSE)

  # https://datatracker.ietf.org/doc/html/rfc8414#section-3.3
  if (!is.null(issuer) && !identical(metadata$issuer, issuer)) {
    cli::cli_abort(c(
      "Metadata {.field issuer} doesn't match the requested {.arg issuer}.",
      "*" = "Requested {.str {issuer}}.",
      "*" = "Received {.str {metadata$issuer}}."
    ))
  }

  structure(metadata, class = "httr2_oauth_server_metadata")
}

# OIDC appends the suffix to the issuer; RFC 8414 inserts it between the origin
# and any path. These differ precisely for multi-tenant, path-carrying issuers.
oauth_metadata_url <- function(issuer, type) {
  parsed <- url_parse(issuer)
  path <- sub("/$", "", parsed$path)

  parsed$path <- switch(
    type,
    openid = paste0(path, "/.well-known/openid-configuration"),
    oauth = paste0("/.well-known/oauth-authorization-server", path)
  )
  url_build(parsed)
}

oauth_metadata_check <- function(metadata, call = caller_env()) {
  if (!inherits(metadata, "httr2_oauth_server_metadata")) {
    cli::cli_abort(
      "{.arg metadata} must be created with {.fn oauth_server_metadata}.",
      call = call
    )
  }
}

#' @export
print.httr2_oauth_server_metadata <- function(x, ...) {
  cli::cat_line(cli::style_bold("<", paste(class(x), collapse = "/"), ">"))

  # Show the issuer and endpoint URLs; the remaining fields (capability arrays,
  # flags, and other metadata) are reference data, so just count them.
  x <- compact(x)
  is_url <- map_lgl(x, \(val) is_string(val) && grepl("^https?://", val))
  bullets(x[is_url])

  n_extra <- sum(!is_url)
  if (n_extra > 0) {
    cli::cat_line(cli::format_inline("* and {n_extra} more field{?s}."))
  }
  invisible(x)
}

test_that("fetches and parses metadata", {
  local_mocked_responses(list(response_json(
    body = list(
      issuer = "https://example.com",
      authorization_endpoint = "https://example.com/authorize",
      token_endpoint = "https://example.com/token",
      scopes_supported = c("openid", "email")
    )
  )))

  meta <- oauth_server_metadata("https://example.com")
  expect_s3_class(meta, "httr2_oauth_server_metadata")
  expect_equal(meta$token_endpoint, "https://example.com/token")
  # unadvertised endpoints are absent
  expect_null(meta$device_authorization_endpoint)
  # string arrays are simplified to character vectors, not lists
  expect_equal(meta$scopes_supported, c("openid", "email"))
})

test_that("builds well-known metadata URLs", {
  expect_equal(
    oauth_metadata_url("https://example.com", "openid"),
    "https://example.com/.well-known/openid-configuration"
  )
  expect_equal(
    oauth_metadata_url("https://example.com/", "openid"),
    "https://example.com/.well-known/openid-configuration"
  )
  expect_equal(
    oauth_metadata_url("https://example.com", "oauth"),
    "https://example.com/.well-known/oauth-authorization-server"
  )
})

test_that("multi-tenant issuers place the suffix differently", {
  expect_equal(
    oauth_metadata_url("https://example.com/tenant1", "openid"),
    "https://example.com/tenant1/.well-known/openid-configuration"
  )
  expect_equal(
    oauth_metadata_url("https://example.com/tenant1", "oauth"),
    "https://example.com/.well-known/oauth-authorization-server/tenant1"
  )
})

test_that("validates issuer against the returned document", {
  local_mocked_responses(list(response_json(
    body = list(
      issuer = "https://evil.com",
      token_endpoint = "https://evil.com/token"
    )
  )))

  expect_snapshot(oauth_server_metadata("https://example.com"), error = TRUE)
})

test_that("url override skips issuer validation when issuer is omitted", {
  local_mocked_responses(list(response_json(
    body = list(
      issuer = "https://example.com",
      token_endpoint = "https://example.com/token"
    )
  )))

  meta <- oauth_server_metadata(url = "https://example.com/metadata")
  expect_equal(meta$token_endpoint, "https://example.com/token")
})

test_that("requires either issuer or url", {
  expect_snapshot(oauth_server_metadata(), error = TRUE)
})

test_that("has a useful print method", {
  meta <- structure(
    list(
      issuer = "https://example.com",
      authorization_endpoint = "https://example.com/authorize",
      token_endpoint = "https://example.com/token",
      response_types_supported = c("code", "token"),
      scopes_supported = c("openid", "email"),
      authorization_response_iss_parameter_supported = TRUE
    ),
    class = "httr2_oauth_server_metadata"
  )
  expect_snapshot(meta)
})

# Standalone file: do not edit by hand
# Source: <https://github.com/r-lib/rlang/blob/main/R/standalone-purrr.R>
# ----------------------------------------------------------------------
#
# ---
# repo: r-lib/rlang
# file: standalone-purrr.R
# last-updated: 2023-02-23
# license: https://unlicense.org
# imports: rlang
# ---
#
# This file provides a minimal shim to provide a purrr-like API on top of
# base R functions. They are not drop-in replacements but allow a similar style
# of programming.
#
# ## Changelog
#
# 2023-02-23:
# * Added `list_c()`
#
# 2022-06-07:
# * `transpose()` is now more consistent with purrr when inner names
#   are not congruent (#1346).
#
# 2021-12-15:
# * `transpose()` now supports empty lists.
#
# 2021-05-21:
# * Fixed "object `x` not found" error in `imap()` (@mgirlich)
#
# 2020-04-14:
# * Removed `pluck*()` functions
# * Removed `*_cpl()` functions
# * Used `as_function()` to allow use of `~`
# * Used `.` prefix for helpers
#
# nocov start

map <- function(.x, .f, ...) {
  .f <- as_function(.f, env = global_env())
  lapply(.x, .f, ...)
}
walk <- function(.x, .f, ...) {
  map(.x, .f, ...)
  invisible(.x)
}

map_lgl <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, logical(1), ...)
}
map_int <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, integer(1), ...)
}
map_dbl <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, double(1), ...)
}
map_chr <- function(.x, .f, ...) {
  .rlang_purrr_map_mold(.x, .f, character(1), ...)
}
.rlang_purrr_map_mold <- function(.x, .f, .mold, ...) {
  .f <- as_function(.f, env = global_env())
  out <- vapply(.x, .f, .mold, ..., USE.NAMES = FALSE)
  names(out) <- names(.x)
  out
}

map2 <- function(.x, .y, .f, ...) {
  .f <- as_function(.f, env = global_env())
  out <- mapply(.f, .x, .y, MoreArgs = list(...), SIMPLIFY = FALSE)
  if (length(out) == length(.x)) {
    set_names(out, names(.x))
  } else {
    set_names(out, NULL)
  }
}
map2_lgl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "logical")
}
map2_int <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "integer")
}
map2_dbl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "double")
}
map2_chr <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "character")
}
imap <- function(.x, .f, ...) {
  map2(.x, names(.x) %||% seq_along(.x), .f, ...)
}

pmap <- function(.l, .f, ...) {
  .f <- as.function(.f)
  args <- .rlang_purrr_args_recycle(.l)
  do.call("mapply", c(
    FUN = list(quote(.f)),
    args, MoreArgs = quote(list(...)),
    SIMPLIFY = FALSE, USE.NAMES = FALSE
  ))
}
.rlang_purrr_args_recycle <- function(args) {
  lengths <- map_int(args, length)
  n <- max(lengths)

  stopifnot(all(lengths == 1L | lengths == n))
  to_recycle <- lengths == 1L
  args[to_recycle] <- map(args[to_recycle], function(x) rep.int(x, n))

  args
}

keep <- function(.x, .f, ...) {
  .x[.rlang_purrr_probe(.x, .f, ...)]
}
discard <- function(.x, .p, ...) {
  sel <- .rlang_purrr_probe(.x, .p, ...)
  .x[is.na(sel) | !sel]
}
map_if <- function(.x, .p, .f, ...) {
  matches <- .rlang_purrr_probe(.x, .p)
  .x[matches] <- map(.x[matches], .f, ...)
  .x
}
.rlang_purrr_probe <- function(.x, .p, ...) {
  if (is_logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    .p <- as_function(.p, env = global_env())
    map_lgl(.x, .p, ...)
  }
}

compact <- function(.x) {
  Filter(length, .x)
}

transpose <- function(.l) {
  if (!length(.l)) {
    return(.l)
  }

  inner_names <- names(.l[[1]])

  if (is.null(inner_names)) {
    fields <- seq_along(.l[[1]])
  } else {
    fields <- set_names(inner_names)
    .l <- map(.l, function(x) {
      if (is.null(names(x))) {
        set_names(x, inner_names)
      } else {
        x
      }
    })
  }

  # This way missing fields are subsetted as `NULL` instead of causing
  # an error
  .l <- map(.l, as.list)

  map(fields, function(i) {
    map(.l, .subset2, i)
  })
}

every <- function(.x, .p, ...) {
  .p <- as_function(.p, env = global_env())

  for (i in seq_along(.x)) {
    if (!rlang::is_true(.p(.x[[i]], ...))) return(FALSE)
  }
  TRUE
}
some <- function(.x, .p, ...) {
  .p <- as_function(.p, env = global_env())

  for (i in seq_along(.x)) {
    if (rlang::is_true(.p(.x[[i]], ...))) return(TRUE)
  }
  FALSE
}
negate <- function(.p) {
  .p <- as_function(.p, env = global_env())
  function(...) !.p(...)
}

reduce <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init)
}
reduce_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE)
}
accumulate <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init, accumulate = TRUE)
}
accumulate_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE, accumulate = TRUE)
}

detect <- function(.x, .f, ..., .right = FALSE, .p = is_true) {
  .p <- as_function(.p, env = global_env())
  .f <- as_function(.f, env = global_env())

  for (i in .rlang_purrr_index(.x, .right)) {
    if (.p(.f(.x[[i]], ...))) {
      return(.x[[i]])
    }
  }
  NULL
}
detect_index <- function(.x, .f, ..., .right = FALSE, .p = is_true) {
  .p <- as_function(.p, env = global_env())
  .f <- as_function(.f, env = global_env())

  for (i in .rlang_purrr_index(.x, .right)) {
    if (.p(.f(.x[[i]], ...))) {
      return(i)
    }
  }
  0L
}
.rlang_purrr_index <- function(x, right = FALSE) {
  idx <- seq_along(x)
  if (right) {
    idx <- rev(idx)
  }
  idx
}

list_c <- function(x) {
  inject(c(!!!x))
}

# nocov end

if (getRversion() < "4.1") {
  dir.create("man/macros", showWarnings = FALSE, recursive = TRUE)
  cat(
    paste(
      "\\renewcommand{\\examples}{\\section{Examples}{",
      "These examples are designed to work in R >= 4.1 so that we can take",
      "advantage of modern syntax like the base pipe (\\verb{|>}) and the ",
      "function shorthand (\\verb{\\(x) x + 1}). They might not work on the ",
      "version of R that you're using.",
      "\\preformatted{#1}}}",
      collapse = ""
    ),
    file = "man/macros/examples.Rd"
  )
}

#' OAuth authentication
#'
#' This is a low-level helper for automatically authenticating a request with
#' an OAuth flow, caching the access token and refreshing it where possible.
#' You should only need to use this function if you're implementing your own
#' OAuth flow.
#'
#' @inheritParams req_perform
#' @param cache An object that controls how the token is cached. This should
#'   be a list containing three functions:
#'   * `get()` retrieves the token from the cache, returning `NULL` if not
#'     cached yet.
#'   * `set()` saves the token to the cache.
#'   * `clear()` removes the token from the cache
#' @param flow An `oauth_flow_` function used to generate the access token.
#' @param flow_params Parameters for the flow. This should be a named list
#'   whose names match the argument names of `flow`.
#' @returns An [oauth_token].
#' @keywords internal
#' @export
req_oauth <- function(req, flow, flow_params, cache) {
  # Want req object to contain meaningful objects, not just a closure
  req <- req_auth_sign(
    req,
    fun = auth_oauth_sign,
    params = list(flow = flow, flow_params = flow_params),
    cache = cache
  )
  req <- req_policies(req, auth_oauth = TRUE)
  req
}

auth_oauth_sign <- function(req, cache, flow, flow_params) {
  token <- auth_oauth_token_get(
    cache = cache,
    flow = flow,
    flow_params = flow_params
  )
  req_auth_bearer_token(req, token$access_token)
}

auth_oauth_token_get <- function(cache, flow, flow_params = list()) {
  token <- cache$get()
  if (is.null(token)) {
    token <- exec(flow, !!!flow_params)
    cache$set(token)
  } else if (token_has_expired(token)) {
    cache$clear()
    if (is.null(token$refresh_token)) {
      token <- exec(flow, !!!flow_params)
    } else {
      token <- tryCatch(
        token_refresh(flow_params$client, token$refresh_token),
        httr2_oauth = function(cnd) {
          # If refresh fails, try to auth from scratch
          exec(flow, !!!flow_params)
        }
      )
    }
    cache$set(token)
  }

  token
}

#' Retrieve an OAuth token using the cache
#'
#' This function wraps around a `oauth_flow_` function to retrieve a token
#' from the cache, or to generate and cache a token if needed. Use this for
#' manual token management that still takes advantage of httr2's caching
#' system. You should only need to use this function if you're passing
#' the token
#'
#' @keywords internal
#' @inheritParams req_oauth
#' @inheritParams req_oauth_auth_code
#' @param reauth Set to `TRUE` to force re-authentication via flow, regardless
#'   of whether or not token is expired.
#' @export
#' @examples
#' \dontrun{
#' token <- oauth_token_cached(
#'   client = example_github_client(),
#'   flow = oauth_flow_auth_code,
#'   flow_params = list(
#'     auth_url = "https://github.com/login/oauth/authorize"
#'   ),
#'   cache_disk = TRUE
#' )
#' token
#' }
oauth_token_cached <- function(
  client,
  flow,
  flow_params = list(),
  cache_disk = FALSE,
  cache_key = NULL,
  reauth = FALSE
) {
  check_bool(reauth)
  cache <- cache_choose(client, cache_disk, cache_key)
  if (reauth) {
    cache$clear()
  }

  flow_params$client <- client
  auth_oauth_token_get(
    cache = cache,
    flow = flow,
    flow_params = flow_params
  )
}

resp_is_invalid_oauth_token <- function(req, resp) {
  if (!req_policy_exists(req, "auth_oauth")) {
    return(FALSE)
  }

  if (is_error(resp) || resp_status(resp) != 401) {
    return(FALSE)
  }

  auth <- resp_header(resp, "WWW-Authenticate")
  if (is.null(auth)) {
    return(FALSE)
  }

  # https://datatracker.ietf.org/doc/html/rfc6750#section-3.1
  # invalid_token:
  #   The access token provided is expired, revoked, malformed, or
  #   invalid for other reasons.  The resource SHOULD respond with
  #   the HTTP 401 (Unauthorized) status code.  The client MAY
  #   request a new access token and retry the protected resource
  #   request.
  grepl('error="invalid_token"', auth, fixed = TRUE)
}

# Caches -------------------------------------------------------------------

cache_choose <- function(client, cache_disk = FALSE, cache_key = NULL) {
  if (cache_disk) {
    cache_disk(client, cache_key)
  } else {
    cache_mem(client, cache_key)
  }
}

# Used for auth endoints that don't have a cache
cache_noop <- function() {
  list(
    get = function() {
      abort("get() was called on cache_noop")
      invisible()
    },
    set = function(token) {
      abort("set() was called on cache_noop")
      invisible()
    },
    clear = function() {}
  )
}
cache_mem <- function(client, key = NULL) {
  key <- hash(c(client$name, key))
  list(
    get = function() env_get(the$token_cache, key, default = NULL),
    set = function(token) env_poke(the$token_cache, key, token),
    clear = function() env_unbind(the$token_cache, key)
  )
}
cache_disk <- function(client, key = NULL) {
  app_path <- file.path(oauth_cache_path(), client$name)
  dir.create(app_path, showWarnings = FALSE, recursive = TRUE)

  path <- file.path(app_path, paste0(hash(key), "-token.rds.enc"))
  list(
    get = function() {
      if (file.exists(path)) secret_read_rds(path, obfuscate_key()) else NULL
    },
    set = function(token) {
      cli::cli_inform("Caching httr2 token in {.path {path}}.")
      secret_write_rds(token, path, obfuscate_key())
    },
    clear = function() if (file.exists(path)) file.remove(path)
  )
}

# Update req_oauth_auth_code() docs if change default from 30
cache_disk_prune <- function(days = 30, path = oauth_cache_path()) {
  files <- dir(
    path,
    recursive = TRUE,
    full.names = TRUE,
    pattern = "-token\\.rds$"
  )
  mtime <- file.mtime(files)

  old <- mtime < (Sys.time() - days * 86400)
  unlink(files[old])
}

#' httr2 OAuth cache location
#'
#' When opted-in to, httr2 caches OAuth tokens in this directory. By default,
#' it uses a OS-standard cache directory, but, if needed, you can override the
#' location by setting the `HTTR2_OAUTH_CACHE` env var.
#'
#' @export
oauth_cache_path <- function() {
  path <- Sys.getenv("HTTR2_OAUTH_CACHE")
  if (path != "") {
    return(path)
  }

  rappdirs::user_cache_dir("httr2")
}


#' Clear OAuth cache
#'
#' Use this function to clear cached credentials.
#'
#' @export
#' @inheritParams req_oauth_auth_code
oauth_cache_clear <- function(client, cache_disk = FALSE, cache_key = NULL) {
  cache <- cache_choose(client, cache_disk, cache_key)
  cache$clear()
  invisible()
}

test_that("invalid token test is specific", {
  req <- request("https://example.com")
  resp_invalid <- response(
    401,
    headers = 'WWW-Authenticate: Bearer realm="example", error="invalid_token", error_description="The access token expired"'
  )

  # Doesn't trigger for response if request doesn't use OAuth
  expect_false(resp_is_invalid_oauth_token(req, resp_invalid))

  req <- req_oauth(req, "", list(), NULL)
  expect_false(resp_is_invalid_oauth_token(req, response(200)))
  expect_false(resp_is_invalid_oauth_token(req, response(401)))
  expect_true(resp_is_invalid_oauth_token(req, resp_invalid))
})


# auth_oauth_token_get() --------------------------------------------------

test_that("can request and cache token if not present", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)

  token <- oauth_token("123")
  expect_equal(auth_oauth_token_get(cache, function(...) token), token)
  expect_equal(cache$get(), token)
})

test_that("can re-flow to get new token", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cache$set(oauth_token("123", expires_in = -60))

  token <- oauth_token("123")
  expect_equal(auth_oauth_token_get(cache, function(...) token), token)
  expect_equal(cache$get(), token)

  # and cache is cleared if
  cache$set(oauth_token("123", expires_in = -60))
  expect_error(auth_oauth_token_get(cache, function(...) stop("Bad!")))
  expect_equal(cache$get(), NULL)
})

test_that("can refresh to get new token", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cache$set(oauth_token("123", refresh_token = "456", expires_in = -60))

  local_mocked_bindings(
    oauth_client_get_token = function(...) oauth_token("789")
  )

  new_token <- oauth_token("789", refresh_token = "456")
  expect_equal(auth_oauth_token_get(cache, function(...) NULL), new_token)
  expect_equal(cache$get(), new_token)
})

test_that("refresh forwards token_params from the flow", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cache$set(oauth_token("123", refresh_token = "456", expires_in = -60))

  local_mocked_bindings(
    token_refresh = function(client, refresh_token, token_params = list()) {
      oauth_token("789", token_params = token_params)
    }
  )

  token <- auth_oauth_token_get(
    cache,
    function(...) NULL,
    flow_params = list(
      client = client,
      scope = "read",
      token_params = list(resource = "example")
    )
  )
  expect_equal(token$token_params, list(resource = "example"))
})

test_that("can reflow if refresh fails", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cache$set(oauth_token("123", refresh_token = "456", expires_in = -60))

  local_mocked_bindings(
    oauth_client_get_token = function(...) oauth_flow_abort("Nope")
  )

  token <- oauth_token("789")
  expect_equal(auth_oauth_token_get(cache, function(...) token), token)
  expect_equal(cache$get(), token)
})

test_that("can retrieve non-expired token from cache", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)

  token <- oauth_token("123")
  cache$set(token)
  expect_equal(auth_oauth_token_get(cache, oauth_flow_refresh), token)
})

test_that("expiry margin controls when cached tokens are refreshed", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cached <- oauth_token("cached", expires_in = 60)
  refreshed <- oauth_token("refreshed")
  cache$set(cached)

  expect_equal(
    auth_oauth_token_get(cache, function(...) refreshed, expiry_margin = 30),
    cached
  )
  expect_equal(
    auth_oauth_token_get(cache, function(...) refreshed, expiry_margin = 90),
    refreshed
  )
  expect_equal(cache$get(), refreshed)
})

test_that("expiry margin preserves refresh token behavior", {
  client <- oauth_client("test", "http://example.org/test")
  cache <- cache_mem(client)
  cache$set(
    oauth_token("cached", refresh_token = "refresh", expires_in = 60)
  )

  local_mocked_bindings(
    token_refresh = function(client, refresh_token, token_params = list()) {
      oauth_token("refreshed", used_refresh_token = refresh_token)
    }
  )

  token <- auth_oauth_token_get(
    cache,
    function(...) NULL,
    flow_params = list(client = client),
    expiry_margin = 90
  )
  expect_equal(token$access_token, "refreshed")
  expect_equal(token$used_refresh_token, "refresh")
})

test_that("req_oauth validates and stores expiry margin", {
  req <- request("https://example.com")
  req <- req_oauth(req, "", list(), NULL, expiry_margin = 40)

  expect_equal(req$policies$auth_sign$params$expiry_margin, 40)
  expect_snapshot(
    req_oauth(req, "", list(), NULL, expiry_margin = -1),
    error = TRUE
  )
})


# Cache -------------------------------------------------------------------

test_that("can store in memory", {
  client <- oauth_client(
    id = "x",
    token_url = "http://example.com",
    name = "httr2-test"
  )

  cache <- cache_mem(client, NULL)
  withr::defer(cache$clear())

  expect_equal(cache$get(), NULL)
  cache$set(1)
  expect_equal(cache$get(), 1)
  cache$clear()
  expect_equal(cache$get(), NULL)
})

test_that("can store on disk", {
  client <- oauth_client(
    id = "x",
    token_url = "http://example.com",
    name = "httr2-test"
  )

  cache <- cache_disk(client, NULL)
  withr::defer(cache$clear())

  expect_equal(cache$get(), NULL)
  expect_snapshot(
    cache$set(1),
    transform = function(x) {
      gsub(oauth_cache_path(), "<oauth-cache-path>", x, fixed = TRUE)
    }
  )
  expect_equal(cache$get(), 1)
  cache$clear()
  expect_equal(cache$get(), NULL)
})

test_that("can explicitly clear cached value", {
  client <- oauth_client(
    id = "x",
    token_url = "http://example.com",
    name = "httr2-test"
  )
  cache <- cache_mem(client, NULL)
  cache$set("abcdef")

  oauth_cache_clear(client)
  expect_equal(cache$get(), NULL)
})

test_that("can prune old files", {
  path <- withr::local_tempdir()
  touch(file.path(path, "a-token.rds.enc"), Sys.time() - 86400 * 1)
  touch(file.path(path, "b-token.rds.enc"), Sys.time() - 86400 * 2)
  cache_disk_prune(2, path)
  expect_equal(dir(path), "a-token.rds.enc")
})

test_that("prunes old files from both new and legacy locations", {
  new_path <- withr::local_tempdir()
  legacy_path <- withr::local_tempdir()
  local_mocked_bindings(
    oauth_cache_path = function() new_path,
    oauth_cache_path_legacy = function() legacy_path
  )

  touch(file.path(new_path, "a-token.rds.enc"), Sys.time() - 86400 * 1)
  touch(file.path(new_path, "b-token.rds.enc"), Sys.time() - 86400 * 2)
  touch(file.path(legacy_path, "a-token.rds.enc"), Sys.time() - 86400 * 1)
  touch(file.path(legacy_path, "b-token.rds.enc"), Sys.time() - 86400 * 2)

  cache_disk_prune(2)

  expect_equal(dir(new_path), "a-token.rds.enc")
  expect_equal(dir(legacy_path), "a-token.rds.enc")
})

test_that("oauth_cache_prune() prunes the default cache locations", {
  new_path <- withr::local_tempdir()
  legacy_path <- withr::local_tempdir()
  local_mocked_bindings(
    oauth_cache_path = function() new_path,
    oauth_cache_path_legacy = function() legacy_path
  )

  touch(file.path(new_path, "a-token.rds.enc"), Sys.time() - 86400 * 1)
  touch(file.path(new_path, "b-token.rds.enc"), Sys.time() - 86400 * 2)

  oauth_cache_prune(2)

  expect_equal(dir(new_path), "a-token.rds.enc")
})

test_that("oauth_cache_prune() validates its input", {
  expect_snapshot(oauth_cache_prune("x"), error = TRUE)
})

# cache_path --------------------------------------------------------------

test_that("can override path with env var", {
  withr::local_envvar("HTTR2_OAUTH_CACHE" = "/tmp")
  expect_equal(oauth_cache_path(), "/tmp")
})

test_that("inlined legacy path matches rappdirs", {
  path <- oauth_cache_path_legacy()
  rappdirs_path <- rappdirs::user_cache_dir("httr2")

  if (.Platform$OS.type == "windows") {
    # rappdirs uses the CSIDL API, which can return an 8.3 short form of
    # the user's home directory, while our env-var based path uses the
    # long form. Both refer to the same directory, so convert to the
    # (existing) directory's canonical short form before comparing.
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
    withr::defer(unlink(path, recursive = TRUE))
    path <- utils::shortPathName(path)
    rappdirs_path <- utils::shortPathName(rappdirs_path)
  }

  expect_equal(
    normalizePath(path, mustWork = FALSE),
    normalizePath(rappdirs_path, mustWork = FALSE)
  )
})

test_that("legacy path respects R_USER_CACHE_DIR", {
  path <- withr::local_tempdir()
  withr::local_envvar("R_USER_CACHE_DIR" = path)

  expected <- rappdirs::user_cache_dir("httr2")
  expect_equal(oauth_cache_path_legacy(), expected)
})

otel_tracer_name <- "org.r-lib.httr2"

otel_cache_tracer <- NULL
req_with_span <- NULL

# Attaches an Open Telemetry span that abides by the semantic conventions for
# HTTP clients to the request, including the associated W3C trace context
# headers.
#
# See: https://opentelemetry.io/docs/specs/semconv/http/http-spans/#http-client-span
local({
  otel_tracer <- NULL
  otel_is_tracing <- FALSE

  otel_cache_tracer <<- function() {
    if (!requireNamespace("otel", quietly = TRUE)) {
      return()
    }
    otel_tracer <<- otel::get_tracer(otel_tracer_name)
    otel_is_tracing <<- tracer_enabled(otel_tracer)
  }

  req_with_span <<- function(
    req,
    resend_count = 0,
    tracer = otel_tracer,
    activation_scope = parent.frame(),
    activate = TRUE
  ) {
    if (!otel_is_tracing) {
      return(req)
    }
    parsed <- tryCatch(url_parse(req$url), error = function(cnd) NULL)
    if (is.null(parsed)) {
      # Don't create spans for invalid URLs.
      return(req)
    }
    if (!req_has_user_agent(req)) {
      req <- req_user_agent(req)
    }
    default_port <- 443L
    if (parsed$scheme == "http") {
      default_port <- 80L
    }
    # Follow the semantic conventions and redact credentials in the URL, when
    # present.
    if (!is.null(parsed$username)) {
      parsed$username <- "REDACTED"
    }
    if (!is.null(parsed$password)) {
      parsed$password <- "REDACTED"
    }
    method <- req_get_method(req)
    # Set required (and some recommended) attributes, especially those relevant to
    # sampling at span creation time.
    attributes <- compact(list(
      "http.request.method" = method,
      "server.address" = parsed$hostname,
      "server.port" = parsed$port %||% default_port,
      "url.full" = url_build(parsed),
      "http.request.resend_count" = if (resend_count > 1) resend_count,
      "user_agent.original" = req$options$useragent
    ))
    span <- tracer$start_span(
      name = method,
      options = list(kind = "client"),
      attributes = attributes
    )
    if (activate) {
      span$activate(activation_scope, end_on_exit = TRUE)
    }
    req <- req_headers(req, !!!otel::pack_http_context())
    req$state$span <- span
    req
  }
})

tracer_enabled <- function(tracer) {
  .subset2(tracer, "is_enabled")()
}

span_recording <- function(span) {
  .subset2(span, "is_recording")()
}

with_otel_record <- function(expr) {
  on.exit(otel_cache_tracer())
  otelsdk::with_otel_record({
    otel_cache_tracer()
    expr
  })
}

req_record_span_status <- function(req, resp = NULL) {
  span <- req$state$span
  if (is.null(span) || !span_recording(span)) {
    return()
  }
  # For more accurate span timing, we end the span after the response has been
  # received, rather than at the end of the associated scope.
  on.exit(span$end())
  if (is.null(resp)) {
    return()
  }
  if (is_error(resp)) {
    span$record_exception(resp)
    span$set_status("error")
    # Surface the underlying curl error class.
    span$set_attribute("error.type", class(resp$parent)[1])
    return()
  }
  span$set_attribute("http.response.status_code", resp_status(resp))
  if (error_is_error(req, resp)) {
    desc <- resp_status_desc(resp)
    if (is.na(desc)) {
      desc <- NULL
    }
    span$set_status("error", desc)
    # The semantic conventions recommend using the status code as a string for
    # these cases.
    span$set_attribute("error.type", as.character(resp_status(resp)))
  } else {
    span$set_status("ok")
  }
}

#' Set and preserve cookies
#'
#' @description
#' Use `req_cookie_set()` to set client side cookies that are sent to the
#' server.
#'
#' By default, httr2 uses a clean slate for every request meaning that cookies
#' are not automatically preserved across requests. To preserve cookies, use
#' `req_cookie_preserve()` along with the path to cookie file that will be
#' read before and updated after each request.
#'
#' @inheritParams req_perform
#' @param path A path to a file where cookies will be read from before and updated after the request.
#' @export
#' @examples
#' # Use `req_cookies_set()` to set client-side cookies
#' request(example_url()) |>
#'   req_cookies_set(a = 1, b = 1) |>
#'   req_dry_run()
#'
#' # Use `req_cookie_preserve()` to preserve server-side cookies across requests
#' path <- tempfile()
#'
#' # Set a server-side cookie
#' request(example_url()) |>
#'   req_cookie_preserve(path) |>
#'   req_template("/cookies/set/:name/:value", name = "chocolate", value = "chip") |>
#'   req_perform() |>
#'   resp_body_json()
#'
#' # Set another sever-side cookie
#' request(example_url()) |>
#'   req_cookie_preserve(path) |>
#'   req_template("/cookies/set/:name/:value", name = "oatmeal", value = "raisin") |>
#'   req_perform() |>
#'   resp_body_json()
#'
#' # Add a client side cookie
#' request(example_url()) |>
#'   req_url_path("/cookies/set") |>
#'   req_cookie_preserve(path) |>
#'   req_cookies_set(snicker = "doodle") |>
#'   req_perform() |>
#'   resp_body_json()
#'
#' # The cookie path has a straightforward format
#' cat(readChar(path, nchars = 1e4))
req_cookie_preserve <- function(req, path) {
  check_request(req)
  check_string(path, allow_empty = FALSE)

  req_options(req, cookiejar = path, cookiefile = path)
}

#' @export
#' @rdname req_cookie_preserve
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]>
#'   Name-value pairs that define query parameters. Each value must be
#'   an atomic vector, which is automatically escaped. To opt-out of escaping,
#'   wrap strings in `I()`.
req_cookies_set <- function(req, ...) {
  check_request(req)
  req_options(req, cookie = cookies_build(list2(...)))
}

cookies_build <- function(x, error_call = caller_env()) {
  elements_build(x, "Cookies", ";", error_call = error_call)
}

#' Create an OAuth token
#'
#' Creates a S3 object of class `<httr2_token>` representing an OAuth token
#' returned from the access token endpoint.
#'
#' @param access_token The access token used to authenticate request
#' @param token_type Type of token; only `"bearer"` is currently supported.
#' @param expires_in Number of seconds until token expires.
#' @param refresh_token Optional refresh token; if supplied, this can be
#'   used to cheaply get a new access token when this one expires.
#' @param ... Additional components returned by the endpoint
#' @param .date Date the request was made; used to convert the relative
#'   `expires_in` to an absolute `expires_at`.
#' @seealso [oauth_token_cached()] to use the token cache with a specified
#'   OAuth flow.
#' @return An OAuth token: an S3 list with class `httr2_token`.
#' @export
#' @examples
#' oauth_token("abcdef")
#' oauth_token("abcdef", expires_in = 3600)
#' oauth_token("abcdef", refresh_token = "ghijkl")
oauth_token <- function(
  access_token,
  token_type = "bearer",
  expires_in = NULL,
  refresh_token = NULL,
  ...,
  .date = Sys.time()
) {
  check_string(access_token)
  check_string(token_type)
  check_number_whole(expires_in, allow_null = TRUE)
  # TODO: should tokens always store their scope?

  if (!is.null(expires_in)) {
    # Store as unix time to avoid worrying about type coercions in cache
    expires_at <- as.numeric(.date) + expires_in
  } else {
    expires_at <- NULL
  }

  structure(
    compact(list2(
      token_type = token_type,
      access_token = access_token,
      expires_at = expires_at,
      refresh_token = refresh_token,
      ...
    )),
    class = "httr2_token"
  )
}

#' @export
print.httr2_token <- function(x, ...) {
  cli::cat_line(cli::style_bold("<", paste(class(x), collapse = "/"), ">"))
  if (has_name(x, "expires_at")) {
    x$expires_at <- format(.POSIXct(x$expires_at))
  }

  redacted <- list_redact(
    compact(x),
    c("access_token", "refresh_token", "id_token")
  )
  bullets(redacted)
  invisible(x)
}

token_has_expired <- function(token, delay = 5) {
  if (is.null(token$expires_at)) {
    FALSE
  } else {
    (unix_time() + delay) > token$expires_at
  }
}

token_refresh <- function(
  client,
  refresh_token,
  scope = NULL,
  token_params = list()
) {
  out <- oauth_client_get_token(
    client,
    grant_type = "refresh_token",
    refresh_token = refresh_token,
    scope = scope,
    !!!token_params
  )
  out$refresh_token <- out[["refresh_token"]] %||% refresh_token
  out
}

test_that("new token computes expires_at", {
  withr::local_envvar(TZ = "UTC")

  time <- .POSIXct(1740000000)
  token <- oauth_token("xyz", expires_in = 10, .date = time)
  expect_s3_class(token, "httr2_token")
  expect_equal(token$expires_at, as.numeric(time + 10))
  expect_snapshot(token)
})

test_that("printing token redacts access, id and refresh token", {
  expect_snapshot({
    oauth_token(
      access_token = "secret",
      refresh_token = "secret",
      id_token = "secret"
    )
  })
})

test_that("can compute token expiry", {
  token <- oauth_token("xyz")
  expect_equal(token_has_expired(token), FALSE)

  # Respects delay
  token <- oauth_token("xyz", expires_in = 8, .date = Sys.time() - 10)
  expect_equal(token_has_expired(token), TRUE)

  token <- oauth_token("xyz", expires_in = 10, .date = Sys.time())
  expect_equal(token_has_expired(token), FALSE)
})

#' OAuth with a bearer JWT (JSON web token)
#'
#' @description
#' Authenticate using a **Bearer JWT** (JSON web token) as an authorization
#' grant to get an access token, as defined by `r rfc(7523, 2.1)`.
#' It is often used for service accounts, accounts that are used primarily in
#' automated environments.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>.
#'
#' @export
#' @family OAuth flows
#' @inheritParams req_perform
#' @inheritParams req_oauth_auth_code
#' @param claim A list of claims. If all elements of the claim set are static
#'   apart from `iat`, `nbf`, `exp`, or `jti`, provide a list and
#'   [jwt_claim()] will automatically fill in the dynamic components.
#'   If other components need to vary, you can instead provide a zero-argument
#'   callback function which should call `jwt_claim()`.
#' @param signature Function use to sign `claim`, e.g. [jwt_encode_sig()].
#' @param signature_params Additional arguments passed to `signature`, e.g.
#'   `size`, `header`.
#' @returns `req_oauth_bearer_jwt()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_bearer_jwt()` returns an [oauth_token].
#' @examples
#' req_auth <- function(req) {
#'   req_oauth_bearer_jwt(
#'     req,
#'     client = oauth_client("example", "https://example.com/get_token"),
#'     claim = jwt_claim()
#'   )
#' }
#'
#' request("https://example.com") |>
#'  req_auth()
req_oauth_bearer_jwt <- function(
  req,
  client,
  claim,
  signature = "jwt_encode_sig",
  signature_params = list(),
  scope = NULL,
  token_params = list()
) {
  params <- list(
    client = client,
    claim = claim,
    signature = signature,
    signature_params = signature_params,
    scope = scope,
    token_params = token_params
  )

  cache <- cache_mem(client, claim)
  req_oauth(req, "oauth_flow_bearer_jwt", params, cache = cache)
}

#' @export
#' @rdname req_oauth_bearer_jwt
oauth_flow_bearer_jwt <- function(
  client,
  claim,
  signature = "jwt_encode_sig",
  signature_params = list(),
  scope = NULL,
  token_params = list()
) {
  check_installed("jose")
  if (is.null(client$key)) {
    cli::cli_abort("JWT flow requires {.arg client} with a key.")
  }

  if (is_list(claim)) {
    claim <- exec("jwt_claim", !!!claim)
  } else if (is.function(claim)) {
    claim <- claim()
  } else {
    cli::cli_abort("{.arg claim} must be a list or function.")
  }

  jwt <- exec(signature, claim = claim, key = client$key, !!!signature_params)

  # https://datatracker.ietf.org/doc/html/rfc7523#section-2.1
  oauth_client_get_token(
    client,
    grant_type = "urn:ietf:params:oauth:grant-type:jwt-bearer",
    assertion = jwt,
    scope = scope,
    !!!token_params
  )
}

#' Perform multiple requests in sequence
#'
#' Given a list of requests, this function performs each in turn, returning
#' a list of responses. It's the serial equivalent of [req_perform_parallel()].
#'
#' @param reqs A list of [request]s.
#' @param paths An optional character vector of paths, if you want to download
#'   the response bodies to disk. If supplied, must be the same length as
#' `reqs`.
#' @param on_error What should happen if one of the requests fails?
#'
#'   * `stop`, the default: stop iterating with an error.
#'   * `return`: stop iterating, returning all the successful responses
#'     received so far, as well as an error object for the failed request.
#'   * `continue`: continue iterating, recording errors in the result.
#' @param progress Display a progress bar for the status of all requests? Use
#'   `TRUE` to turn on a basic progress bar, use a string to give it a name,
#'   or see [progress_bars] to customize it in other ways. Not compatible with
#'   [req_progress()], as httr2 can only display a single progress bar at a
#'   time.
#' @inheritParams req_perform
#' @return
#' A list, the same length as `reqs`, containing [response]s and possibly
#' error objects, if `on_error` is `"return"` or `"continue"` and one of the
#' responses errors. If `on_error` is `"return"` and it errors on the ith
#' request, the ith element of the result will be an error object, and the
#' remaining elements will be `NULL`. If `on_error` is `"continue"`, it will
#' be a mix of requests and error objects.
#'
#' Only httr2 errors are captured; see [req_error()] for more details.
#' @export
#' @examples
#' # One use of req_perform_sequential() is if the API allows you to request
#' # data for multiple objects, you want data for more objects than can fit
#' # in one request.
#' req <- request("https://api.restful-api.dev/objects")
#'
#' # Imagine we have 50 ids:
#' ids <- sort(sample(100, 50))
#'
#' # But the API only allows us to request 10 at time. So we first use split
#' # and some modulo arithmetic magic to generate chunks of length 10
#' chunks <- unname(split(ids, (seq_along(ids) - 1) %/% 10))
#'
#' # Then we use lapply to generate one request for each chunk:
#' reqs <- chunks |> lapply(\(idx) req |> req_url_query(id = idx, .multi = "comma"))
#'
#' # Then we can perform them all and get the results
#' \dontrun{
#' resps <- reqs |> req_perform_sequential()
#' resps_data(resps, \(resp) resp_body_json(resp))
#' }
req_perform_sequential <- function(
  reqs,
  paths = NULL,
  on_error = c("stop", "return", "continue"),
  mock = getOption("httr2_mock", NULL),
  progress = TRUE
) {
  if (!is_bare_list(reqs)) {
    stop_input_type(reqs, "a list")
  }
  check_paths(paths, reqs)
  on_error <- arg_match(on_error)
  mock <- as_mock_function(mock)

  err_catch <- on_error != "stop"
  err_return <- on_error == "return"

  progress <- create_progress_bar(progress, length(reqs))

  resps <- rep_along(reqs, list())

  tryCatch(
    {
      for (i in seq_along(reqs)) {
        check_request(reqs[[i]], arg = glue::glue("req[[{i}]]"))

        if (err_catch) {
          resps[[i]] <- tryCatch(
            req_perform(reqs[[i]], path = paths[[i]], mock = mock),
            httr2_error = function(err) err
          )
        } else {
          resps[[i]] <- req_perform(reqs[[i]], path = paths[[i]])
        }
        if (err_return && is_error(resps[[i]])) {
          break
        }
        progress$update()
      }
    },
    interrupt = function(cnd) {
      check_repeated_interrupt()

      resps <- resps[seq_len(i)]
      cli::cli_alert_warning(
        "Terminating iteration; returning {i} response{?s}."
      )
    }
  )
  progress$done()

  resps
}

check_paths <- function(paths, reqs, error_call = caller_env()) {
  if (!is.null(paths)) {
    check_character(paths)
    if (length(reqs) != length(paths)) {
      cli::cli_abort(
        "If supplied, {.arg paths} must be the same length as {.arg req}.",
        call = error_call
      )
    }
  }
}

#' Extract headers from a response
#'
#' @description
#' * `resp_headers()` retrieves a list of all headers.
#' * `resp_header()` retrieves a single header.
#' * `resp_header_exists()` checks if a header is present.
#'
#' @param resp A httr2 [response] object, created by [req_perform()].
#' @param filter A regular expression used to filter the header names.
#'   `NULL`, the default, returns all headers.
#' @return
#' * `resp_headers()` returns a list.
#' * `resp_header()` returns a string if the header exists and `NULL` otherwise.
#' * `resp_header_exists()` returns `TRUE` or `FALSE`.
#' @export
#' @examples
#' resp <- request("https://httr2.r-lib.org") |> req_perform()
#' resp |> resp_headers()
#' resp |> resp_headers("x-")
#'
#' resp |> resp_header_exists("server")
#' resp |> resp_header("server")
#' # Headers are case insensitive
#' resp |> resp_header("SERVER")
#'
#' # Returns NULL if header doesn't exist
#' resp |> resp_header("this-header-doesnt-exist")
resp_headers <- function(resp, filter = NULL) {
  check_response(resp)

  if (is.null(filter)) {
    resp$headers
  } else {
    resp$headers[grepl(
      filter,
      names(resp$headers),
      perl = TRUE,
      ignore.case = TRUE
    )]
  }
}

#' @export
#' @param header Header name (case insensitive)
#' @param default Default value to use if header doesn't exist.
#' @rdname resp_headers
resp_header <- function(resp, header, default = NULL) {
  check_response(resp)
  resp$headers[[header]] %||% default
}

#' @export
#' @rdname resp_headers
resp_header_exists <- function(resp, header) {
  check_response(resp)
  tolower(header) %in% tolower(names(resp$headers))
}

#' Extract request date from response
#'
#' All responses contain a request date in the `Date` header; if not provided
#' by the server will be automatically added by httr2.
#'
#' @export
#' @inheritParams resp_headers
#' @returns A `POSIXct` date-time.
#' @examples
#' resp <- response(headers = "Date: Wed, 01 Jan 2020 09:23:15 UTC")
#' resp |> resp_date()
#'
#' # If server doesn't add header (unusual), you get the time the request
#' # was created:
#' resp <- response()
#' resp |> resp_date()
resp_date <- function(resp) {
  parse_http_date(resp_header(resp, "Date"))
}

#' Extract response content type and encoding
#'
#' @description
#' `resp_content_type()` returns the just the type and subtype of the
#' from the `Content-Type` header. If `Content-Type` is not provided; it
#' returns `NA`. Used by [resp_body_json()], [resp_body_html()], and
#' [resp_body_xml()].
#'
#' `resp_encoding()` returns the likely character encoding of text
#' types, as parsed from the `charset` parameter of the `Content-Type`
#' header. If that header is not found, not valid, or no charset parameter
#' is found, returns `UTF-8`. Used by [resp_body_string()].
#'
#' @export
#' @returns A string. If no content type is specified `resp_content_type()`
#'   will return a character `NA`; if no encoding is specified,
#'   `resp_encoding()` will return `"UTF-8"`.
#' @inheritParams resp_headers
#' @examples
#' resp <- response(headers = "Content-type: text/html; charset=utf-8")
#' resp |> resp_content_type()
#' resp |> resp_encoding()
#'
#' # No Content-Type header
#' resp <- response()
#' resp |> resp_content_type()
#' resp |> resp_encoding()
resp_content_type <- function(resp) {
  if (resp_header_exists(resp, "content-type")) {
    parse_media(resp_header(resp, "content-type"))$type
  } else {
    NA_character_
  }
}

#' @export
#' @rdname resp_content_type
resp_encoding <- function(resp) {
  if (resp_header_exists(resp, "content-type")) {
    parse_media(resp_header(resp, "content-type"))$charset %||% "UTF-8"
  } else {
    "UTF-8"
  }
}

#' Extract wait time from a response
#'
#' Computes how many seconds you should wait before retrying a request by
#' inspecting the `Retry-After` header. It parses both forms (absolute and
#' relative) and returns the number of seconds to wait. If the heading is not
#' found, it will return `NA`.
#'
#' @export
#' @returns Scalar double giving the number of seconds to wait before retrying
#'   a request.
#' @inheritParams resp_headers
#' @examples
#' resp <- response(headers = "Retry-After: 30")
#' resp |> resp_retry_after()
#'
#' resp <- response(headers = "Retry-After: Mon, 20 Sep 2025 21:44:05 UTC")
#' resp |> resp_retry_after()
resp_retry_after <- function(resp) {
  check_response(resp)

  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Retry-After
  val <- resp_header(resp, "Retry-After")
  if (is.null(val)) {
    NA
  } else if (grepl(" ", val, fixed = TRUE)) {
    diff <- difftime(parse_http_date(val), resp_date(resp), units = "secs")
    as.numeric(diff)
  } else {
    as.numeric(val)
  }
}

#' Parse link URL from a response
#'
#' Parses URLs out of the the `Link` header as defined by `r rfc(8288)`.
#'
#' @export
#' @inheritParams resp_headers
#' @returns Either a string providing a URL, if the specified `rel` exists, or
#'   `NULL` if not.
#' @param rel The "link relation type" value for which to retrieve a URL.
#' @export
#' @examples
#' # Simulate response from GitHub code search
#' resp <- response(headers = paste0("Link: ",
#'   '<https://api.github.com/search/code?q=addClass+user%3Amozilla&page=2>; rel="next",',
#'   '<https://api.github.com/search/code?q=addClass+user%3Amozilla&page=34>; rel="last"'
#' ))
#'
#' resp_link_url(resp, "next")
#' resp_link_url(resp, "last")
#' resp_link_url(resp, "prev")
resp_link_url <- function(resp, rel) {
  if (!resp_header_exists(resp, "Link")) {
    return()
  }

  headers <- resp_headers(resp)
  link_headers <- headers[tolower(names(headers)) == "link"]

  links <- unlist(lapply(link_headers, parse_link), recursive = FALSE)
  sel <- map_lgl(links, \(x) x$rel == rel)
  if (sum(sel) != 1L) {
    return()
  }

  links[[which(sel)]]$url
}

test_that("can extract headers and check for existence", {
  resp <- response(headers = "Content-Type: application/json")
  expect_type(resp_headers(resp), "list")
  expect_equal(resp_header(resp, "Content-Type"), "application/json")
  expect_equal(resp_header_exists(resp, "Content-Type"), TRUE)
})

test_that("headers are case-insenstive", {
  resp <- response(headers = "Content-Type: application/json")
  expect_equal(resp_header(resp, "content-type"), "application/json")
  expect_equal(resp_header_exists(resp, "content-type"), TRUE)
})

test_that("has tools for non-existent headers", {
  resp <- response()

  expect_equal(resp_header(resp, "non-existent"), NULL)
  expect_equal(resp_header(resp, "non-existent", "xyz"), "xyz")
  expect_equal(resp_header_exists(resp, "non-existent"), FALSE)
})

test_that("can extract content type/encoding", {
  resp <- response(headers = "Content-Type: text/html; charset=latin1")
  expect_equal(resp_content_type(resp), "text/html")
  expect_equal(resp_encoding(resp), "latin1")
})

test_that("can parse date header", {
  resp <- response(headers = "Date: Mon, 18 Jul 2016 16:06:00 GMT")
  expect_equal(resp_date(resp), local_time('2016-07-18 16:06:06'))
})

test_that("can parse both forms of retry-after header", {
  resp_abs <- response(
    headers = c(
      "Retry-After: Mon, 18 Jul 2016 16:06:10 GMT",
      "Date: Mon, 18 Jul 2016 16:06:00 GMT"
    )
  )
  expect_equal(resp_retry_after(resp_abs), 10)

  resp_rel <- response(
    headers = "Retry-After: 20"
  )
  expect_equal(resp_retry_after(resp_rel), 20)

  resp_rel <- response()
  expect_equal(resp_retry_after(resp_rel), NA)
})

# resp_link_url() --------------------------------------------------------------

test_that("can extract specified link url", {
  resp <- response(
    headers = paste0(
      'Link: <https://example.com/1>; rel="next",',
      '<https://example.com/2>; rel="last"'
    )
  )
  expect_equal(resp_link_url(resp, "next"), "https://example.com/1")
  expect_equal(resp_link_url(resp, "last"), "https://example.com/2")

  # Falling back to NULL if not found
  expect_equal(resp_link_url(resp, "first"), NULL)
  expect_equal(resp_link_url(response(), "first"), NULL)
})

test_that("can extract from multiple link headers", {
  resp <- response(
    headers = c(
      'Link: <https://example.com/1>; rel="next"',
      'Link: <https://example.com/2>; rel="last"'
    )
  )
  expect_equal(resp_link_url(resp, "next"), "https://example.com/1")
  expect_equal(resp_link_url(resp, "last"), "https://example.com/2")
})

test_that("is case insensitive", {
  resp <- response(headers = 'LINK: <https://example.com/1>; rel="next"')
  expect_equal(resp_link_url(resp, "next"), "https://example.com/1")
})

#' Create a new HTTP request
#'
#' @description
#' There are three steps needed to perform a HTTP request with httr2:
#'
#' 1. Create a request object with `request(url)` (this function).
#' 2. Define its behaviour with `req_` functions, e.g.:
#'    * [req_headers()] to set header values.
#'    * [req_url_path()] and friends to modify the url.
#'    * [req_body_json()] and friends to add a body.
#'    * [req_auth_basic()] to perform basic HTTP authentication.
#'    * [req_oauth_auth_code()] to use the OAuth auth code flow.
#' 3. Perform the request and fetch the response with [req_perform()].
#'
#' @param base_url Base URL for request.
#' @returns An HTTP request: an S3 list with class `httr2_request`.
#' @export
#' @examples
#' request("http://r-project.org")
request <- function(base_url) {
  new_request(base_url)
}

#' @export
print.httr2_request <- function(x, ..., redact_headers = TRUE) {
  cli::cat_line(cli::format_inline("{.cls {class(x)}}"))
  method <- toupper(req_get_method(x))
  cli::cat_line(cli::format_inline("{.strong {method}} {x$url}"))

  bullets_with_header("Headers:", headers_flatten(x$headers, redact_headers))
  cli::cat_line(cli::format_inline("{.strong Body}: {req_body_info(x)}"))
  bullets_with_header("Options:", x$options)
  bullets_with_header("Policies:", x$policies)

  invisible(x)
}

new_request <- function(
  url,
  method = NULL,
  headers = list(),
  body = NULL,
  fields = list(),
  options = list(),
  policies = list(),
  error_call = caller_env()
) {
  check_string(url, call = error_call)

  structure(
    list(
      url = url,
      method = method,
      headers = headers,
      body = body,
      fields = fields,
      options = options,
      policies = policies,
      state = new_environment()
    ),
    class = "httr2_request"
  )
}

is_request <- function(x) {
  inherits(x, "httr2_request")
}

check_request <- function(
  req,
  arg = caller_arg(req),
  call = caller_env(),
  allow_null = FALSE
) {
  if (!missing(req)) {
    if (is_request(req)) {
      return(invisible(NULL))
    }

    if (allow_null && is.null(req)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    req,
    "an HTTP request object",
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

test_that("req has basic print method", {
  expect_snapshot({
    req <- request("https://example.com")
    req
    req |> req_body_raw("Test")
    req |> req_body_multipart("Test" = 1)
  })
})

test_that("printing headers works with {}", {
  expect_snapshot(req_headers(request("http://test"), x = "{z}", `{z}` = "x"))
})

test_that("individually prints repeated headers", {
  expect_snapshot(request("https://example.com") |> req_headers(A = 1:3))
})

test_that("print method obfuscates Authorization header unless requested", {
  req <- request("https://example.com") |>
    req_auth_basic("user", "SECRET")
  output <- testthat::capture_messages(print(req))
  expect_false(any(grepl("SECRET", output, fixed = TRUE)))

  output <- capture.output(print(req, redact_headers = FALSE))
  expect_true(any(grepl("Authorization: \"Basic", output, fixed = TRUE)))
  expect_false(any(grepl("REDACTED", output, fixed = TRUE)))
})

test_that("check_request() gives useful error", {
  expect_snapshot(check_request(1), error = TRUE)
})

test_that("can configure token expiry margin", {
  client <- oauth_client("test", "https://example.com/token", secret = "secret")
  default_req <- request("https://example.com") |>
    req_oauth_client_credentials(client)
  req <- request("https://example.com") |>
    req_oauth_client_credentials(client, expiry_margin = 40)

  expect_equal(default_req$policies$auth_sign$params$expiry_margin, 30)
  expect_equal(req$policies$auth_sign$params$expiry_margin, 40)
})

oauth_flow_fetch <- function(req, source, error_call = caller_env()) {
  req <- req_error(req, is_error = \(resp) FALSE)
  resp <- req_perform(req, error_call = current_call())

  oauth_flow_parse(resp, source, error_call = error_call)
}

oauth_flow_parse <- function(resp, source, error_call = caller_env()) {
  withCallingHandlers(
    body <- oauth_flow_body(resp),
    error = function(err) {
      cli::cli_abort(
        "Failed to parse response from {.arg {source}} OAuth url.",
        parent = err,
        resp = resp,
        class = "httr2_oauth_parse",
        call = error_call
      )
    }
  )

  if (has_name(body, "expires_in")) {
    body$expires_in <- as.numeric(body$expires_in)
  }

  # This is rather more flexible than what the spec requires, and should
  # hopefully be general enough to handle most token endpoints. However,
  # it would still be nice to figure out how to make user extensible,
  # especially since you might be able to give better errors.
  if (has_name(body, "access_token") || has_name(body, "device_code")) {
    body
  } else if (has_name(body, "error")) {
    oauth_flow_abort(
      body$error,
      body$error_description,
      body$error_uri,
      error_call = error_call
    )
  } else {
    cli::cli_abort(
      c(
        "Failed to parse response from {.arg {source}} OAuth url.",
        "*" = "Did not contain {.code access_token}, {.code device_code}, or {.code error} field."
      ),
      resp = resp,
      class = "httr2_oauth_parse",
      call = error_call
    )
  }
}

oauth_flow_body <- function(resp) {
  resp_body_json(resp, check_type = NA)
}

# https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.2.1
# https://datatracker.ietf.org/doc/html/rfc6749#section-4.2.2.1
# https://datatracker.ietf.org/doc/html/rfc6749#section-5.2
#
# TODO: automatically fill in description from text in RFC?
oauth_flow_abort <- function(
  error,
  description = NULL,
  uri = NULL,
  error_call = caller_env()
) {
  cli::cli_abort(
    c(
      "OAuth failure [{error}]",
      "*" = description,
      i = if (!is.null(uri)) "Learn more at {.url {uri}}."
    ),
    code = error,
    class = c(glue("httr2_oauth_{error}"), "httr2_oauth"),
    call = error_call
  )
}

#' Perform a request to get a response
#'
#' @description
#' After preparing a [request], call `req_perform()` to perform it, fetching
#' the results back to R as a [response].
#'
#' The default HTTP method is `GET` unless a body (set by [req_body_json] and
#' friends) is present, in which case it will be `POST`. You can override
#' these defaults with [req_method()].
#'
#' # Requests
#' Note that one call to `req_perform()` may perform multiple HTTP requests:
#'
#' * If the `url` is redirected with a 301, 302, 303, or 307, curl will
#'   automatically follow the `Location` header to the new location.
#'
#' * If you have configured retries with [req_retry()] and the request
#'   fails with a transient problem, `req_perform()` will try again after
#'   waiting a bit. See [req_retry()] for details.
#'
#' * If you are using OAuth, and the cached token has expired, `req_perform()`
#'   will get a new token either using the refresh token (if available)
#'   or by running the OAuth flow.
#'
#' # Progress bar
#'
#' `req_perform()` will automatically add a progress bar if it needs to wait
#' between requests for [req_throttle()] or [req_retry()]. You can turn the
#' progress bar off (and just show the total time to wait) by setting
#' `options(httr2_progress = FALSE)`.
#'
#' @param req A httr2 [request] object.
#' @param path Optionally, path to save body of the response. This is useful
#'   for large responses since it avoids storing the response in memory.
#' @param mock A mocking function. If supplied, this function is called
#'   with the request. It should return either `NULL` (if it doesn't want to
#'   handle the request) or a [response] (if it does). See
#'   [with_mocked_responses()]/`local_mocked_responses()` for more details.
#' @param verbosity How much information to print? This is a wrapper
#'   around [req_verbose()] that uses an integer to control verbosity:
#'
#'   * `0`: no output
#'   * `1`: show headers
#'   * `2`: show headers and bodies
#'   * `3`: show headers, bodies, and curl status messages.
#'
#'   Use [with_verbosity()] to control the verbosity of requests that
#'   you can't affect directly.
#' @inheritParams rlang::args_error_context
#' @returns
#'   * If the HTTP request succeeds, and the status code is ok (e.g. 200),
#'     an HTTP [response].
#'
#'   * If the HTTP request succeeds, but the status code is an error
#'     (e.g a 404), an error with class `c("httr2_http_404", "httr2_http")`.
#'     By default, all 400 and 500 status codes will be treated as an error,
#'     but you can customise this with [req_error()].
#'
#'   * If the HTTP request fails (e.g. the connection is dropped or the
#'     server doesn't exist), an error with class `"httr2_failure"`.
#' @export
#' @seealso [req_perform_parallel()] to perform multiple requests in parallel.
#'   [req_perform_iterative()] to perform multiple requests iteratively.
#' @examples
#' request("https://google.com") |>
#'   req_perform()
req_perform <- function(
  req,
  path = NULL,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL),
  error_call = current_env()
) {
  check_request(req)
  check_string(path, allow_null = TRUE)
  # verbosity checked by req_verbosity
  mock <- as_mock_function(mock, error_call)

  verbosity <- verbosity %||% httr2_verbosity()

  if (!is.null(mock)) {
    mock_resp <- mock(req)
    if (!is.null(mock_resp)) {
      return(handle_resp(req, mock_resp, error_call = error_call))
    }
  }

  req <- req_verbosity(req, verbosity)

  req <- cache_pre_fetch(req, path)
  if (is_response(req)) {
    return(req)
  }

  req_prep <- req_prepare(req)
  handle <- req_handle(req_prep)
  max_tries <- retry_max_tries(req)
  deadline <- Sys.time() + retry_max_seconds(req)

  n <- 0
  tries <- 0
  reauthed <- FALSE # only ever re-authenticate once

  sys_sleep(throttle_delay(req), "for throttling delay")

  delay <- 0
  while (tries < max_tries && Sys.time() < deadline) {
    retry_check_breaker(req, tries, error_call = error_call)
    sys_sleep(delay, "for retry backoff")
    n <- n + 1

    resp <- req_perform1(
      req,
      req_prep,
      path = path,
      handle = handle,
      resend_count = n
    )
    req_completed(req_prep)

    if (retry_is_transient(req, resp)) {
      tries <- tries + 1
      delay <- retry_after(req, resp, tries)
      signal(class = "httr2_retry", tries = tries, delay = delay)
    } else if (!reauthed && resp_is_invalid_oauth_token(req, resp)) {
      reauthed <- TRUE
      req_auth_clear_cache(req)
      req_prep <- req_prepare(req)
      handle <- req_handle(req_prep)
      delay <- 0
    } else {
      # done
      break
    }
  }
  # Used for testing
  signal(class = "httr2_fetch", n = n, tries = tries, reauth = reauthed)

  resp <- cache_post_fetch(req, resp, path = path)
  handle_resp(req, resp, error_call = error_call)
}

handle_resp <- function(req, resp, error_call = caller_env()) {
  if (resp_show_body(resp)) {
    verbose_body("<< ", resp$body, resp$headers$`content-type`)
  }

  if (is_error(resp)) {
    resp$request <- req
    resp$call <- error_call
    cnd_signal(resp)
  } else if (error_is_error(req, resp)) {
    cnd <- resp_failure_cnd(req, resp, error_call = error_call)
    cnd_signal(cnd)
  } else {
    resp
  }
}

resp_failure_cnd <- function(req, resp, error_call = caller_env()) {
  status <- resp_status(resp)
  desc <- resp_status_desc(resp)
  message <- paste0("HTTP ", status, if (!is.na(desc)) paste0(" ", desc), ".")

  info <- error_body(req, resp, error_call)

  catch_cnd(abort(
    c(message, resp_auth_message(resp), i = info),
    status = status,
    resp = resp,
    request = req,
    class = c(
      glue("httr2_http_{status}"),
      "httr2_http",
      "httr2_error",
      "rlang_error"
    ),
    use_cli_format = TRUE,
    call = error_call
  ))
}

req_perform1 <- function(
  req,
  req_prep,
  path = NULL,
  handle = NULL,
  resend_count = 0
) {
  the$last_request <- req
  the$last_response <- NULL
  signal(class = "httr2_perform")
  # Note: we need to do this before we call handle_preflight() so that request
  # signing works correctly with the added headers.
  req_prep <- req_with_span(req_prep, resend_count = resend_count)
  handle_preflight(req_prep, handle)

  err <- capture_curl_error({
    fetch <- curl_fetch(handle, req$url, path)
  })
  if (is_error(err)) {
    req_record_span_status(req, err)
    return(err)
  }

  # Ensure cookies are saved to disk now, not when request is finalised
  curl::handle_setopt(handle, cookielist = "FLUSH")
  curl::handle_setopt(handle, cookiefile = NULL, cookiejar = NULL)

  resp <- create_response(req, fetch$curl_data, fetch$body)
  req_record_span_status(req_prep, resp)
  resp
}

curl_fetch <- function(handle, url, path) {
  if (!is.null(path)) {
    curl_data <- curl::curl_fetch_disk(url, path, handle)
    body <- new_path(path)
  } else {
    curl_data <- curl::curl_fetch_memory(url, handle)
    body <- curl_data$content
  }

  list(curl_data = curl_data, body = body)
}

req_verbosity <- function(req, verbosity, error_call = caller_env()) {
  if (!is_integerish(verbosity, n = 1) || verbosity < 0 || verbosity > 3) {
    cli::cli_abort("{.arg verbosity} must 0, 1, 2, or 3.", call = error_call)
  }

  switch(
    verbosity + 1,
    req,
    req_verbose(req),
    req_verbose(req, body_req = TRUE, body_resp = TRUE),
    req_verbose(req, body_req = TRUE, body_resp = TRUE, info = TRUE)
  )
}

# Must call req_prepare(), then req_handle(), then after the request has been
# performed, req_completed() (on the prepared requests)
req_prepare <- function(req) {
  req <- req_method_apply(req)
  req <- req_body_apply(req)
  if (!req_has_user_agent(req)) {
    req <- req_user_agent(req)
  }
  req
}

req_handle <- function(req) {
  handle <- curl::new_handle()
  curl::handle_setopt(handle, url = req$url)
  curl::handle_setopt(handle, .list = req$options)
  if (length(req$fields) > 0) {
    curl::handle_setform(handle, .list = req$fields)
  }

  handle
}

# Called right before the request is sent, when the final headers are in place.
handle_preflight <- function(req, handle) {
  req <- auth_sign(req)
  curl::handle_setheaders(
    handle,
    .list = headers_flatten(req$headers, redact = FALSE)
  )
  # Save final request headers so that req_verbose() can use them
  req$state$headers <- req$headers
  invisible(handle)
}

req_completed <- function(req) {
  req_policy_call(req, "done", list(), NULL)
}

new_path <- function(x) structure(x, class = "httr2_path")
is_path <- function(x) inherits(x, "httr2_path")

resp_show_body <- function(resp) {
  resp$request$policies$show_body %||% FALSE
}

#' Add a progress bar to long downloads or uploads
#'
#' When uploading or downloading a large file, it's often useful to
#' provide a progress bar so that you know how long you have to wait.
#'
#' @inheritParams req_headers
#' @param type Type of progress to display: either number of bytes uploaded
#'   or downloaded.
#' @export
#' @examples
#' req <- request("https://r4ds.s3.us-west-2.amazonaws.com/seattle-library-checkouts.csv") |>
#'   req_progress()
#'
#' \dontrun{
#' path <- tempfile()
#' req |> req_perform(path = path)
#' }
req_progress <- function(req, type = c("down", "up")) {
  type <- arg_match(type)

  # https://curl.se/libcurl/c/CURLOPT_XFERINFOFUNCTION.html
  req_options(req, noprogress = FALSE, xferinfofunction = make_progress(type))
}

make_progress <- function(type, frame = caller_env()) {
  force(type)
  init <- FALSE

  function(down, up) {
    if (type == "down") {
      total <- down[[1]]
      now <- down[[2]]
      verb <- "Downloading"
    } else {
      total <- up[[1]]
      now <- up[[2]]
      verb <- "Uploading"
    }

    if (total == 0 && now == 0) {
      init <<- FALSE
      return(TRUE)
    }

    if (!init) {
      init <<- TRUE
      if (total == 0) {
        cli::cli_progress_bar(
          format = paste0(verb, " {cli::pb_spin}"),
          .envir = frame
        )
      } else {
        cli::cli_progress_bar(
          format = paste0(
            verb,
            " {cli::pb_percent} {cli::pb_bar} {cli::pb_eta}"
          ),
          total = total,
          .envir = frame
        )
      }
    }

    if (now < total && total > 0) {
      cli::cli_progress_update(set = now, .envir = frame)
    } else {
      cli::cli_progress_done(.envir = frame)
    }

    TRUE
  }
}

#' Modify request headers
#'
#' @description
#' `req_headers()` allows you to set the value of any header.
#'
#' `req_headers_redacted()` is a variation that adds "redacted" headers, which
#' httr2 avoids printing on the console. This is good practice for
#' authentication headers to avoid accidentally leaking them in log files.
#'
#' @param .req A [request].
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> Name-value pairs of headers
#'   and their values.
#'
#'   * Use `NULL` to reset a value to httr2's default.
#'   * Use `""` to remove a header.
#'   * Use a character vector to repeat a header.
#' @param .redact A character vector of headers to redact. The Authorization
#'   header is always redacted.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' req <- request("http://example.com")
#'
#' # Use req_headers() to add arbitrary additional headers to the request
#' req |>
#'   req_headers(MyHeader = "MyValue") |>
#'   req_dry_run()
#'
#' # Repeated use overrides the previous value:
#' req |>
#'   req_headers(MyHeader = "Old value") |>
#'   req_headers(MyHeader = "New value") |>
#'   req_dry_run()
#'
#' # Setting Accept to NULL uses curl's default:
#' req |>
#'   req_headers(Accept = NULL) |>
#'   req_dry_run()
#'
#' # Setting it to "" removes it:
#' req |>
#'   req_headers(Accept = "") |>
#'   req_dry_run()
#'
#' # If you need to repeat a header, provide a vector of values
#' # (this is rarely needed, but is important in a handful of cases)
#' req |>
#'   req_headers(HeaderName = c("Value 1", "Value 2", "Value 3")) |>
#'   req_dry_run()
#'
#' # If you have headers in a list, use !!!
#' headers <- list(HeaderOne = "one", HeaderTwo = "two")
#' req |>
#'   req_headers(!!!headers, HeaderThree = "three") |>
#'   req_dry_run()
#'
#' # Use `req_headers_redacted()`` to hide a header in the output
#' req_secret <- req |>
#'   req_headers_redacted(Secret = "this-is-private") |>
#'   req_headers(Public = "but-this-is-not")
#'
#' req_secret
#' req_secret |> req_dry_run()
req_headers <- function(.req, ..., .redact = NULL) {
  check_request(.req)
  check_character(.redact, allow_null = TRUE)
  check_header_values(...)

  headers <- modify_list(.req$headers, ..., .ignore_case = TRUE)
  redact <- c("Authorization", .redact, which_redacted(.req$headers))
  .req$headers <- new_headers(headers, redact, lifespan = .req$state)

  .req
}

#' @export
#' @rdname req_headers
req_headers_redacted <- function(.req, ...) {
  check_request(.req)

  headers <- list2(...)
  req_headers(.req, !!!headers, .redact = names(headers))
}

check_header_values <- function(..., error_call = caller_env()) {
  dots <- list2(...)

  type_ok <- map_lgl(dots, function(x) is_atomic(x) || is.null(x))
  if (any(!type_ok)) {
    cli::cli_abort(
      "All elements of {.arg ...} must be either an atomic vector or NULL.",
      call = error_call
    )
  }

  invisible()
}

#' Get request headers
#'
#' Retrieve custom headers set on the request. Use [req_dry_run()] to get all
#' headers, including those automatically generated by curl.
#'
#' @inheritParams req_perform
#' @param redacted What to do with redacted headers?
#'   * `"drop"` (the default) drops them.
#'   * `"redact"` replaces them with `<REDACTED>`.
#'   * `"reveal"` leaves them in place.
#' @returns A named list.
#' @export
#' @examples
#' req <- request("http://example.com")
#' req <- req_headers(req, a = 1L, b = 2L, .redact = "a")
#'
#' req_get_headers(req, "drop")
#' req_get_headers(req, "redact")
#' req_get_headers(req, "reveal")
req_get_headers <- function(req, redacted = c("drop", "redact", "reveal")) {
  check_request(req)
  redacted <- arg_match(redacted)

  headers <- req$headers
  headers <- headers[!is_redacted_empty(headers)]

  if (redacted == "drop") {
    headers <- headers[!is_redacted(headers)]
  } else if (redacted == "redact") {
    headers[is_redacted(headers)] <- "<REDACTED>"
  }
  headers_flatten(headers, redact = FALSE)
}

test_that("can add and remove headers", {
  req <- request("http://example.com")
  req <- req |> req_headers(x = 1)
  expect_equal(req$headers, new_headers(list(x = 1)))
  req <- req |> req_headers(x = NULL)
  expect_equal(req$headers, new_headers(list()))
})

test_that("simple vectors automatically converted to strings", {
  req <- request("http://example.com")
  req <- req |> req_headers(lgl = TRUE, int = 1L, dbl = 1.1, chr = "a")
  resp <- req_dry_run(req, quiet = TRUE)

  expect_equal(resp$headers$lgl, "TRUE")
  expect_equal(resp$headers$int, "1")
  expect_equal(resp$headers$dbl, "1.1")
  expect_equal(resp$headers$chr, "a")
})

test_that("bad inputs get clear error", {
  req <- request("http://example.com")
  expect_snapshot(error = TRUE, {
    req_headers(req, fun = mean)
    req_headers(req, 1)
  })
})

test_that("can add header called req", {
  req <- request("http://example.com")
  req <- req |> req_headers(req = 1)
  expect_equal(req$headers, new_headers(list(req = 1)))
})

test_that("can add repeated headers", {
  resp <- request_test() |>
    req_headers(a = c("a", "b")) |>
    req_dry_run(quiet = TRUE)
  # https://datatracker.ietf.org/doc/html/rfc7230#section-3.2.2
  expect_equal(resp$headers$a, "a,b")
})

test_that("replacing headers is case-insensitive", {
  req <- request("http://example.com")
  req <- req |> req_headers(A = 1)
  req <- req |> req_headers(a = 2)
  expect_equal(req$headers, new_headers(list(a = 2)))
})

# accessor ----------------------------------------------------------------

test_that("can control redaction", {
  req <- request("http://example.com")
  req <- req_headers(req, a = 1L, b = 2L, .redact = "a")

  expect_equal(req_get_headers(req, "drop"), list(b = "2"))
  expect_equal(req_get_headers(req, "redact"), list(a = "<REDACTED>", b = "2"))
  expect_equal(req_get_headers(req, "reveal"), list(a = "1", b = "2"))
})

test_that("empty redacted headers are always dropped", {
  req <- request("http://example.com")
  req <- req_headers(req, a = 1L, b = 2L, .redact = "a")
  req2 <- unserialize(serialize(req, NULL))

  expect_equal(req_get_headers(req2, "drop"), list(b = "2"))
  expect_equal(req_get_headers(req2, "redact"), list(b = "2"))
  expect_equal(req_get_headers(req2, "reveal"), list(b = "2"))
})

# redaction ---------------------------------------------------------------

test_that("can control which headers to redact", {
  req <- request("http://example.com")
  expect_redacted(req_headers(req, a = 1L, b = 2L), character())
  expect_redacted(req_headers(req, a = 1L, b = 2L, .redact = "a"), "a")
  expect_redacted(
    req_headers(req, a = 1L, b = 2L, .redact = c("a", "b")),
    c("a", "b")
  )
})

test_that("only redacts supplied headers", {
  req <- request("http://example.com")
  expect_redacted(req_headers(req, a = 1L, b = 2L, .redact = "d"), character())
})

test_that("redaction preserved across calls", {
  req <- request("http://example.com")
  req <- req_headers(req, a = 1L, .redact = "a")
  req <- req_headers(req, a = 2)
  expect_redacted(req, "a")
  expect_equal(headers_flatten(req$headers, FALSE), list(a = "2"))

  # and is reapplied regadless of case
  req <- req_headers(req, A = 3)
  expect_redacted(req, "A")
  expect_equal(headers_flatten(req$headers, FALSE), list(A = "3"))
})

test_that("req_headers_redacted redacts all headers", {
  req <- request("http://example.com")
  expect_redacted(req_headers_redacted(req, a = 1L, b = 2L), c("a", "b"))
})

test_that("is case insensitive", {
  req <- request("http://example.com")
  req <- req_headers(req, a = 1L, .redact = "A")
  expect_redacted(req, "a")
  expect_snapshot(req)

  # Test the other direction too, just to be safe
  req <- request("http://example.com")
  req <- req_headers(req, A = 1L, .redact = "a")
  expect_redacted(req, "A")
})

test_that("authorization is always redacted", {
  req <- request("http://example.com")
  expect_redacted(req_headers(req, Authorization = "X"), "Authorization")
})

test_that("checks input types", {
  req <- request("http://example.com")
  expect_snapshot(error = TRUE, {
    req_headers(req, a = 1L, b = 2L, .redact = 1L)
  })
})

#' Perform a request and handle data as it streams back
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' Please use [req_perform_connection()] instead.
#'
#' After preparing a request, call `req_perform_stream()` to perform the request
#' and handle the result with a streaming callback. This is useful for
#' streaming HTTP APIs where potentially the stream never ends.
#'
#' The `callback` will only be called if the result is successful. If you need
#' to stream an error response, you can use [req_error()] to suppress error
#' handling so that the body is streamed to you.
#'
#' @inheritParams req_perform
#' @param callback A single argument callback function. It will be called
#'   repeatedly with a raw vector whenever there is at least `buffer_kb`
#'   worth of data to process. It must return `TRUE` to continue streaming.
#' @param timeout_sec Number of seconds to process stream for.
#' @param buffer_kb Buffer size, in kilobytes.
#' @param round How should the raw vector sent to `callback` be rounded?
#'   Choose `"byte"`, `"line"`, or supply your own function that takes a
#'   raw vector of `bytes` and returns the locations of possible cut points
#'   (or `integer()` if there are none).
#' @returns An HTTP [response]. The body will be empty if the request was
#'   successful (since the `callback` function will have handled it). The body
#'   will contain the HTTP response body if the request was unsuccessful.
#' @export
#' @examples
#' # PREVIOSULY
#' show_bytes <- function(x) {
#'   cat("Got ", length(x), " bytes\n", sep = "")
#'   TRUE
#' }
#' resp <- request(example_url()) |>
#'   req_url_path("/stream-bytes/100000") |>
#'   req_perform_stream(show_bytes, buffer_kb = 32)
#'
#' # NOW
#' resp <- request(example_url()) |>
#'   req_url_path("/stream-bytes/100000") |>
#'   req_perform_connection()
#' while (!resp_stream_is_complete(resp)) {
#'   x <-  resp_stream_raw(resp, kb = 32)
#'   cat("Got ", length(x), " bytes\n", sep = "")
#' }
#' close(resp)
req_perform_stream <- function(
  req,
  callback,
  timeout_sec = Inf,
  buffer_kb = 64,
  round = c("byte", "line")
) {
  # soft deprecated because quite a few packages use it, and it was our
  # recommended approach for a while
  lifecycle::deprecate_soft(
    when = "1.2.0",
    what = "req_perform_stream()",
    with = "req_perform_connection()"
  )

  check_request(req)

  check_function(callback)
  check_number_decimal(timeout_sec, min = 0)
  check_number_decimal(buffer_kb, min = 0)
  cut_points <- as_round_function(round)

  stop_time <- Sys.time() + timeout_sec

  resp <- req_perform_connection(req)
  close <- resp$body$close
  withr::defer(close())

  continue <- TRUE
  incomplete <- TRUE
  buf <- raw()

  while (continue && !resp$body$is_complete() && Sys.time() < stop_time) {
    buf <- c(buf, resp$body$read(buffer_kb * 1024))

    if (length(buf) > 0) {
      cut <- cut_points(buf)
      n <- length(cut)
      if (n) {
        continue <- isTRUE(callback(utils::head(buf, n = cut[n])))
        buf <- utils::tail(buf, n = -cut[n])
      }
    }
  }

  # if there are leftover bytes and none of the callback()
  # returned FALSE.
  if (continue && length(buf)) {
    callback(buf)
  }

  # We're done streaming so convert to bodiless response
  resp$body <- raw()
  the$last_response <- resp
  resp
}

# Helpers ----------------------------------------------------------------------

as_round_function <- function(
  round = c("byte", "line"),
  error_call = caller_env()
) {
  if (is.function(round)) {
    check_function2(round, args = "bytes")
    round
  } else if (is.character(round)) {
    round <- arg_match(round, error_call = error_call)
    switch(
      round,
      byte = function(bytes) length(bytes),
      line = function(bytes) which(bytes == charToRaw("\n"))
    )
  } else {
    cli::cli_abort(
      '{.arg round} must be "byte", "line" or a function.',
      call = error_call
    )
  }
}

# promises package test helper
extract_promise <- function(promise, timeout = 30) {
  promise_value <- NULL
  error <- NULL
  promises::then(
    promise,
    onFulfilled = function(value) promise_value <<- value,
    onRejected = function(reason) error <<- reason
  )

  start <- Sys.time()
  while (!later::loop_empty()) {
    if (difftime(Sys.time(), start, units = "secs") > timeout) {
      stop("Waited too long")
    }
    later::run_now(0.01)
  }

  if (!is.null(error)) {
    cnd_signal(error)
  } else {
    promise_value
  }
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import R6
#' @import rlang
#' @importFrom glue glue
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

the <- new_environment()
the$throttle <- new_environment()
the$breaker <- new_environment()
the$cache_throttle <- list()
the$token_cache <- new_environment()
the$last_response <- NULL
the$last_request <- NULL
the$pool_pollers <- new_environment()
the$last_interrupt <- 0

#' Create and encode a JWT
#'
#' `jwt_claim()` is a wrapper around [jose::jwt_claim()] that creates a JWT
#' claim set with a few extra default values. `jwt_encode_sig()` and
#' `jwt_encode_hmac()` are thin wrappers around [jose::jwt_encode_sig()] and
#' [jose::jwt_encode_hmac()] that exist primarily to make specification
#' in other functions a little simpler.
#'
#' @param iss Issuer claim. Identifies the principal that issued the JWT.
#' @param sub Subject claim. Identifies the principal that is the subject of
#'   the JWT (i.e. the entity that the claims apply to).
#' @param aud Audience claim. Identifies the recipients that the JWT is
#'    intended. Each principle intended to process the JWT must be identified
#'    with a unique value.
#' @param exp Expiration claim. Identifies the expiration time on or after which
#'   the JWT MUST NOT be accepted for processing. Defaults to 5 minutes.
#' @param nbf Not before claim. Identifies the time before which the JWT
#'   MUST NOT be accepted for processing. Defaults to current time.
#' @param iat Issued at claim. Identifies the time at which the JWT was
#'   issued.  Defaults to current time.
#' @param jti JWT ID claim. Provides a unique identifier for the JWT.
#'   If omitted, uses a random 32-byte sequence encoded with base64url.
#' @param ... Any additional claims to include in the claim set.
#' @returns An S3 list with class `jwt_claim`.
#' @keywords internal
#' @export
#' @examples
#' claim <- jwt_claim()
#' str(claim)
jwt_claim <- function(
  iss = NULL,
  sub = NULL,
  aud = NULL,
  exp = unix_time() + 5L * 60L,
  nbf = unix_time(),
  iat = unix_time(),
  jti = NULL,
  ...
) {
  # https://datatracker.ietf.org/doc/html/rfc7519
  jose::jwt_claim(
    iss = iss,
    sub = sub,
    aud = aud,
    exp = exp,
    iat = iat,
    nbf = nbf,
    jti = jti %||% base64_url_rand(32),
    ...
  )
}

#' @export
#' @rdname jwt_claim
#' @param claim Claim set produced by [jwt_claim()].
#' @param key RSA or EC private key either specified as a path to a file,
#'   a connection, or a string (PEM/SSH format), or a raw vector (DER format).
#' @param size Size, in bits, of sha2 signature, i.e. 256, 384 or 512.
#'   Only for HMAC/RSA, not applicable for ECDSA keys.
#' @param header A named list giving additional fields to include in the
#'   JWT header.
jwt_encode_sig <- function(claim, key, size = 256, header = list()) {
  check_installed("jose")
  jose::jwt_encode_sig(claim, key, size = size, header = header)
}

#' @export
#' @rdname jwt_claim
#' @param secret String or raw vector with a secret passphrase.
jwt_encode_hmac <- function(claim, secret, size = 256, header = list()) {
  check_installed("jose")
  jose::jwt_encode_hmac(claim, secret, size = size, header = header)
}

rfc <- function(num, sec = NULL) {
  paste_c(
    c(
      "[",
      if (!is.null(sec)) paste0("Section ", sec, " of "),
      "RFC ",
      num,
      "]"
    ),
    c(
      "(https://datatracker.ietf.org/doc/html/rfc",
      num,
      if (!is.null(sec)) paste0("#section-", sec),
      ")"
    )
  )
}

#' Sign a request with the AWS SigV4 signing protocol
#'
#' This is a custom auth protocol implemented by AWS.
#'
#' @inheritParams req_perform
#' @param aws_access_key_id,aws_secret_access_key AWS key and secret.
#' @param aws_session_token AWS session token, if required.
#' @param aws_service,aws_region The AWS service and region to use for the
#'   request. If not supplied, will be automatically parsed from the URL
#'   hostname.
#' @export
#' @examplesIf httr2:::has_paws_credentials()
#' creds <- paws.common::locate_credentials()
#' model_id <- "anthropic.claude-3-5-sonnet-20240620-v1:0"
#' req <- request("https://bedrock-runtime.us-east-1.amazonaws.com")
#' # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html
#' req <- req_url_path_append(req, "model", model_id, "converse")
#' req <- req_body_json(req, list(
#'   messages = list(list(
#'     role = "user",
#'     content = list(list(text = "What's your name?"))
#'   ))
#' ))
#' req <- req_auth_aws_v4(
#'   req,
#'   aws_access_key_id = creds$access_key_id,
#'   aws_secret_access_key = creds$secret_access_key,
#'   aws_session_token = creds$session_token
#' )
#' resp <- req_perform_connection(req)
#' str(resp_body_json(resp))
req_auth_aws_v4 <- function(
  req,
  aws_access_key_id,
  aws_secret_access_key,
  aws_session_token = NULL,
  aws_service = NULL,
  aws_region = NULL
) {
  check_request(req)
  check_string(aws_access_key_id)
  check_string(aws_secret_access_key)
  check_string(aws_session_token, allow_null = TRUE)
  check_string(aws_service, allow_null = TRUE)
  check_string(aws_region, allow_null = TRUE)

  req_auth_sign(
    req,
    fun = auth_aws_sign,
    params = list(
      aws_access_key_id = aws_access_key_id,
      aws_secret_access_key = aws_secret_access_key,
      aws_session_token = aws_session_token,
      aws_service = aws_service,
      aws_region = aws_region
    ),
    # auth_aws_sign doesn't have anything to cache, but still need to provide
    # a cache argument because the req_auth_sign machinery expects it
    cache = cache_noop()
  )
}

auth_aws_sign <- function(
  req,
  aws_access_key_id,
  aws_secret_access_key,
  aws_session_token = NULL,
  aws_service = NULL,
  aws_region = NULL,
  cache
) {
  current_time <- Sys.time()

  type <- req_get_body_type(req)
  body <- switch(
    type,
    empty = "",
    json = req_body_render(req),
    form = req_body_render(req),
    cli::cli_abort(
      "Unsupported body type: {type}",
      call = quote(req_auth_aws_v4())
    )
  )
  body_sha256 <- openssl::sha256(body)

  # We begin by adding some necessary headers that must be added before
  # canoncalization even thought they aren't documented until later
  req <- req_aws_headers(
    req,
    current_time = current_time,
    aws_session_token = aws_session_token,
    body_sha256 = body_sha256
  )

  signature <- aws_v4_signature(
    method = req_get_method(req),
    url = url_parse(req$url),
    headers = as.list(req_get_headers(req, "reveal")),
    body_sha256 = body_sha256,
    current_time = current_time,
    aws_service = aws_service,
    aws_region = aws_region,
    aws_access_key_id = aws_access_key_id,
    aws_secret_access_key = aws_secret_access_key
  )
  req_headers(req, Authorization = signature$Authorization)
}


req_aws_headers <- function(req, current_time, aws_session_token, body_sha256) {
  RequestDateTime <- format(current_time, "%Y%m%dT%H%M%SZ", tz = "UTC")

  req_headers(
    req,
    "x-amz-date" = RequestDateTime,
    "x-amz-security-token" = aws_session_token,
    .redact = "x-amz-security-token"
  )
}

# https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html
aws_v4_signature <- function(
  method,
  url,
  headers,
  body_sha256,
  aws_access_key_id,
  aws_secret_access_key,
  current_time = Sys.time(),
  aws_service = NULL,
  aws_region = NULL
) {
  if (is.null(aws_service) || is.null(aws_region)) {
    host <- strsplit(url$hostname, ".", fixed = TRUE)[[1]]
    aws_service <- aws_service %||%
      strsplit(host[[1]], "-", fixed = TRUE)[[1]][[1]]
    aws_region <- aws_region %||% host[[2]]
  }

  # 1. Create a canonical request
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html#create-canonical-request
  HTTPMethod <- method
  CanonicalURI <- curl::curl_escape(url$path %||% "/")
  # AWS does not want / to be encoded here
  CanonicalURI <- gsub("%2F", "/", CanonicalURI, fixed = TRUE)

  if (is.null(url$query)) {
    CanonicalQueryString <- ""
  } else {
    sorted_query <- url$query[order(names(url$query))]
    CanonicalQueryString <- url_query_build(sorted_query)
  }

  headers$host <- url$hostname
  names(headers) <- tolower(names(headers))
  headers <- headers[order(names(headers))]
  headers[] <- trimws(headers)
  headers[] <- gsub(" {2,}", " ", headers)
  CanonicalHeaders <- paste0(names(headers), ":", headers, "\n", collapse = "")
  SignedHeaders <- paste0(names(headers), collapse = ";")

  CanonicalRequest <- paste_c(
    c(HTTPMethod, "\n"),
    c(CanonicalURI, "\n"),
    c(CanonicalQueryString, "\n"),
    c(CanonicalHeaders, "\n"),
    c(SignedHeaders, "\n"),
    body_sha256
  )
  # 2. Create the hash of the canonical request
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html
  HashedCanonicalRequest <- openssl::sha256(CanonicalRequest)

  # 3. Create the string to sign
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html#create-string-to-sign

  Algorithm <- "AWS4-HMAC-SHA256"
  RequestDateTime <- format(current_time, "%Y%m%dT%H%M%SZ", tz = "UTC")
  Date <- format(current_time, "%Y%m%d", tz = "UTC")
  CredentialScope <- file.path(Date, aws_region, aws_service, "aws4_request")

  string_to_sign <- paste_c(
    c(Algorithm, "\n"),
    c(RequestDateTime, "\n"),
    c(CredentialScope, "\n"),
    HashedCanonicalRequest
  )

  # 4. Derive a signing key
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html#derive-signing-key

  DateKey <- hmac_sha256(paste0("AWS4", aws_secret_access_key), Date)
  DateRegionKey <- hmac_sha256(DateKey, aws_region)
  DateRegionServiceKey <- hmac_sha256(DateRegionKey, aws_service)
  SigningKey <- hmac_sha256(DateRegionServiceKey, "aws4_request")

  # 5. Calculate signature
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html#calculate-signature

  signature <- hmac_sha256(SigningKey, string_to_sign)
  signature <- paste0(as.character(signature), collapse = "")

  # 6. Add the signature to the request
  # https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_sigv-create-signed-request.html#calculate-signature
  credential <- file.path(aws_access_key_id, CredentialScope)

  Authorization <- paste_c(
    c(Algorithm, " "),
    c("Credential=", credential, ","),
    c("SignedHeaders=", SignedHeaders, ","),
    c("Signature=", signature)
  )

  list(
    CanonicalRequest = CanonicalRequest,
    string_to_sign = string_to_sign,
    SigningKey = SigningKey,
    Authorization = Authorization
  )
}

hmac_sha256 <- function(key, value) {
  openssl::sha256(charToRaw(value), key)
}

has_paws_credentials <- function() {
  tryCatch(
    {
      paws.common::locate_credentials()
      TRUE
    },
    error = function(e) {
      FALSE
    }
  )
}

test_that("can correctly sign a request with dummy credentials", {
  req <- request("https://sts.amazonaws.com/")
  req <- req_auth_aws_v4(
    req,
    aws_access_key_id = "AKIAIOSFODNN7EXAMPLE",
    aws_secret_access_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
  )
  req <- req_body_form(
    req,
    Action = "GetCallerIdentity",
    Version = "2011-06-15"
  )
  expect_error(req_perform(req), class = "httr2_http_403")

  # And can clear non-existant cache
  expect_no_error(req_auth_clear_cache(req))
})

test_that("clear error if body type is unsupported", {
  req <- request("https://sts.amazonaws.com/")
  req <- req_auth_aws_v4(
    req,
    aws_access_key_id = "AKIAIOSFODNN7EXAMPLE",
    aws_secret_access_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
  )
  req <- req_body_multipart(req, x = "y")
  expect_snapshot(req_perform(req), error = TRUE)
})

test_that("can correctly sign a request with live credentials", {
  skip_if_not(has_paws_credentials())
  creds <- paws.common::locate_credentials()

  # https://docs.aws.amazon.com/STS/latest/APIReference/API_GetCallerIdentity.html
  req <- request("https://sts.amazonaws.com/")
  req <- req_auth_aws_v4(
    req,
    aws_access_key_id = creds$access_key_id,
    aws_secret_access_key = creds$secret_access_key,
    aws_session_token = creds$session_token,
    aws_region = creds$region
  )
  req <- req_body_form(
    req,
    Action = "GetCallerIdentity",
    Version = "2011-06-15"
  )
  expect_no_error(req_perform(req))
})

test_that('aws_v4_signature calculates correct signature', {
  req <- request("https://example.execute-api.us-east-1.amazonaws.com/v0/") |>
    req_method('POST')

  body_sha256 <- openssl::sha256(req_get_body(req) %||% "")
  current_time <- as.POSIXct(1737483742, origin = "1970-01-01", tz = "EST")

  signature <- aws_v4_signature(
    method = req_get_method(req),
    url = url_parse(req$url),
    headers = req$headers,
    body_sha256 = body_sha256,
    current_time = current_time,
    aws_service = 'execute-api',
    aws_region = 'us-east-1',
    aws_access_key_id = 'AKIAIOSFODNN7EXAMPLE',
    aws_secret_access_key = 'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'
  )
  expect_snapshot(signature)
})

test_that("signing agrees with glacier example", {
  # Example from
  # https://docs.aws.amazon.com/amazonglacier/latest/dev/amazon-glacier-signing-requests.html

  signature <- aws_v4_signature(
    method = "PUT",
    url = url_parse(
      "https://glacier.us-east-1.amazonaws.com/-/vaults/examplevault"
    ),
    headers = list(
      "x-amz-date" = "20120525T002453Z",
      "x-amz-glacier-version" = "2012-06-01"
    ),
    body_sha256 = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    current_time = as.POSIXct("2012-05-25 00:24:53", tz = "UTC"),
    aws_access_key_id = "AKIAIOSFODNN7EXAMPLE",
    aws_secret_access_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
  )

  expected <- paste0(
    "AWS4-HMAC-SHA256 ",
    "Credential=AKIAIOSFODNN7EXAMPLE/20120525/us-east-1/glacier/aws4_request,",
    "SignedHeaders=host;x-amz-date;x-amz-glacier-version,",
    "Signature=3ce5b2f2fffac9262b4da9256f8d086b4aaf42eba5f111c21681a65a127b7c2a"
  )

  expect_equal(signature$Authorization, expected)
})

test_that("validates its inputs", {
  req <- request("https://sts.amazonaws.com/")
  expect_snapshot(error = TRUE, {
    req_auth_aws_v4(1)
    req_auth_aws_v4(req, 1)
    req_auth_aws_v4(req, "", "", aws_session_token = 1)
    req_auth_aws_v4(req, "", "", aws_service = 1)
    req_auth_aws_v4(req, "", "", aws_region = 1)
  })
})

#' OAuth with client credentials
#'
#' @description
#' Authenticate using OAuth **client credentials flow**, as defined by
#' `r rfc(6749, 4.4)`. It is used to allow the client to access resources that
#' it controls directly, not on behalf of an user.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>.
#'
#' @export
#' @family OAuth flows
#' @inheritParams req_perform
#' @inheritParams req_oauth_auth_code
#' @returns `req_oauth_client_credentials()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_client_credentials()` returns an [oauth_token].
#' @examples
#' req_auth <- function(req) {
#'   req_oauth_client_credentials(
#'     req,
#'     client = oauth_client("example", "https://example.com/get_token")
#'   )
#' }
#'
#' request("https://example.com") |>
#'   req_auth()
req_oauth_client_credentials <- function(
  req,
  client,
  scope = NULL,
  token_params = list()
) {
  params <- list(
    client = client,
    scope = scope,
    token_params = token_params
  )

  cache <- cache_mem(client, NULL)
  req_oauth(req, "oauth_flow_client_credentials", params, cache = cache)
}

#' @export
#' @rdname req_oauth_client_credentials
oauth_flow_client_credentials <- function(
  client,
  scope = NULL,
  token_params = list()
) {
  oauth_flow_check("client credentials", client, is_confidential = TRUE)

  oauth_client_get_token(
    client,
    grant_type = "client_credentials",
    scope = scope,
    !!!token_params
  )
}

multi_dots <- function(
  ...,
  .multi = c("error", "comma", "pipe", "explode"),
  .space = c("percent", "form"),
  error_arg = "...",
  error_call = caller_env()
) {
  if (is.function(.multi)) {
    check_function2(.multi, call = error_call, arg = ".multi")
  } else {
    .multi <- arg_match(.multi, error_arg = ".multi", error_call = error_call)
  }
  .space <- arg_match(.space, call = error_call)
  form <- .space == "form"

  dots <- list2(...)
  if (length(dots) == 0) {
    return(list())
  }

  if (!is_named(dots)) {
    cli::cli_abort(
      "All components of {.arg {error_arg}} must be named.",
      call = error_call
    )
  }

  type_ok <- map_lgl(dots, function(x) is_atomic(x) || is.null(x))
  if (any(!type_ok)) {
    cli::cli_abort(
      "All elements of {.arg {error_arg}} must be either an atomic vector or NULL.",
      call = error_call
    )
  }

  n <- lengths(dots)
  if (any(n > 1)) {
    if (is.function(.multi)) {
      dots[n > 1] <- imap(
        dots[n > 1],
        format_query_param,
        multi = TRUE,
        form = form
      )
      dots[n > 1] <- lapply(dots[n > 1], .multi)
      dots[n > 1] <- lapply(dots[n > 1], I)
    } else if (.multi == "comma") {
      dots[n > 1] <- imap(
        dots[n > 1],
        format_query_param,
        multi = TRUE,
        form = form
      )
      dots[n > 1] <- lapply(dots[n > 1], paste0, collapse = ",")
      dots[n > 1] <- lapply(dots[n > 1], I)
    } else if (.multi == "pipe") {
      dots[n > 1] <- imap(
        dots[n > 1],
        format_query_param,
        multi = TRUE,
        form = form
      )
      dots[n > 1] <- lapply(dots[n > 1], paste0, collapse = "|")
      dots[n > 1] <- lapply(dots[n > 1], I)
    } else if (.multi == "explode") {
      dots <- explode(dots)
      n <- lengths(dots)
    } else if (.multi == "error") {
      cli::cli_abort(
        c(
          "All vector elements of {.arg {error_arg}} must be length 1.",
          i = "Use {.arg .multi} to choose a strategy for handling vectors."
        ),
        call = error_call
      )
    }
  }

  # Format other params
  dots[n == 1] <- imap(
    dots[n == 1],
    format_query_param,
    form = form,
    error_call = error_call
  )
  dots[n == 1] <- lapply(dots[n == 1], I)

  dots
}

explode <- function(x) {
  expanded <- map(x, function(x) {
    if (is.null(x)) {
      list(NULL)
    } else {
      map(seq_along(x), function(i) x[i])
    }
  })
  stats::setNames(
    unlist(expanded, recursive = FALSE, use.names = FALSE),
    rep(names(x), lengths(expanded))
  )
}

test_that("can handle multi query params", {
  expect_equal(
    multi_dots(a = 1:2, .multi = "explode"),
    list(a = I("1"), a = I("2"))
  )
  expect_equal(
    multi_dots(a = 1:2, .multi = "comma"),
    list(a = I("1,2"))
  )
  expect_equal(
    multi_dots(a = 1:2, .multi = "pipe"),
    list(a = I("1|2"))
  )
  expect_equal(
    multi_dots(a = 1:2, .multi = function(x) "X"),
    list(a = I("X"))
  )
})

test_that("can handle mixed length multi query params", {
  expect_equal(
    multi_dots(a = 1:2, b = 1, c = NULL, .multi = "explode"),
    list(a = I("1"), a = I("2"), b = I("1"), c = NULL)
  )
  expect_equal(
    multi_dots(a = 1:2, b = 1, c = NULL, .multi = "comma"),
    list(a = I("1,2"), b = I("1"), c = NULL)
  )
  expect_equal(
    multi_dots(a = 1:2, b = 1, c = NULL, .multi = "pipe"),
    list(a = I("1|2"), b = I("1"), c = NULL)
  )
  expect_equal(
    multi_dots(a = 1:2, b = 1, c = NULL, .multi = function(x) "X"),
    list(a = I("X"), b = I("1"), c = NULL)
  )
})

test_that("can opt-out of escaping for' vectors", {
  expect_equal(
    multi_dots(a = I(c(" ", " ")), .multi = "comma"),
    list(a = I(" , "))
  )
})

test_that("can handle empty dots", {
  expect_equal(multi_dots(), list())
})

test_that("preserves NULL values", {
  expect_equal(multi_dots(x = NULL), list(x = NULL))
})

test_that("preserves duplicates values", {
  expect_equal(multi_dots(x = 1, x = 2), list(x = I("1"), x = I("2")))
})

test_that("leaves already escaped values alone", {
  x <- I("1 + 2")
  expect_equal(multi_dots(x = x), list(x = x))
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    multi_dots(1)
    multi_dots(x = I(1))
    multi_dots(x = 1:2)
    multi_dots(x = mean)
  })
})

#' Translate a httr2 request to a curl command
#'
#' Convert a httr2 request object to an approximate curl command line call.
#' This is useful for debugging, for sharing a request with someone who doesn't
#' use R, or for handing off to another tool.
#'
#' @inheritParams req_perform
#' @inheritParams req_get_body
#' @return A string containing the curl command.
#' @seealso [curl_translate()] to translate in the other direction.
#' @export
#' @examples
#' # Basic GET request
#' request("https://httpbin.org/get") |>
#'   httr2_translate()
#'
#' # POST with JSON body
#' request("https://httpbin.org/post") |>
#'   req_body_json(list(name = "value")) |>
#'   httr2_translate()
#'
#' # Secrets are redacted by default, but can be revealed
#' request("https://example.com") |>
#'   req_headers_redacted(Authorization = "secret") |>
#'   httr2_translate(obfuscated = "reveal")
httr2_translate <- function(req, obfuscated = c("redact", "reveal")) {
  check_request(req)
  obfuscated <- arg_match(obfuscated)

  req <- req_prepare(req)
  req <- auth_sign(req)

  body <- curl_body(req, obfuscated)
  args <- c(
    dquote(req_get_url(req)),
    curl_method(req, has_body = !is.null(body)),
    curl_headers(req, obfuscated),
    curl_options(req),
    body
  )
  indent <- c("", rep("  ", length(args) - 1))
  backslash <- c(rep(" \\", length(args) - 1), "")
  out <- paste0("curl ", paste0(indent, args, backslash, collapse = "\n"))

  structure(out, class = "httr2_cmd")
}

curl_method <- function(req, has_body = FALSE) {
  method <- req_get_method(req)
  if (method == "GET" || (method == "POST" && has_body)) {
    NULL
  } else if (method == "HEAD") {
    "--head"
  } else {
    paste0("--request ", method)
  }
}

curl_headers <- function(req, obfuscated = c("redact", "reveal")) {
  obfuscated <- arg_match(obfuscated)

  headers <- req_get_headers(req, redacted = obfuscated)
  if (is_empty(headers)) {
    return(NULL)
  }
  paste0("--header ", dquote(paste0(names(headers), ": ", unlist(headers))))
}

curl_options <- function(req) {
  options <- req$options

  known_options <- c(
    "timeout_ms", # req_timeout()
    "connecttimeout", # req_timeout()
    "proxy", # req_proxy()
    "proxyport", # req_proxy()
    "proxyuserpwd", # req_proxy()
    "useragent", # req_user_agent()
    "followlocation",
    "verbose", # req_verbose()
    "cookiejar", # req_cookie_preserve()
    "cookiefile", # req_cookie_preserve()
    "cookie" # req_cookies_set()
  )
  # R callbacks and the like, with no command line equivalent
  ignored_options <- c(
    "debugfunction", # req_verbose()
    "xferinfofunction", # req_progress()
    "noprogress", # req_progress()
    "proxyauth", # req_proxy()
    "forbid_reuse", # req_verbose_test()
    "nobody", # req_method_apply()
    "customrequest", # req_method_apply()
    "post", # req_body_apply()
    "postfieldsize", # req_body_apply()
    "postfields", # req_body_apply()
    "readfunction", # req_body_apply()
    "seekfunction", # req_body_apply()
    "postfieldsize_large" # req_body_apply()
  )
  unknown <- setdiff(names(options), c(known_options, ignored_options))
  if (length(unknown) > 0) {
    cli::cli_warn("Can't translate option{?s} {.val {unknown}}.")
  }

  args <- lapply(names(options), function(name) {
    value <- options[[name]]
    switch(
      name,
      timeout_ms = paste0("--max-time ", value / 1000),
      connecttimeout = paste0("--connect-timeout ", value),
      proxy = paste0(
        "--proxy ",
        dquote(paste0(c(value, options$proxyport), collapse = ":"))
      ),
      proxyuserpwd = paste0("--proxy-user ", dquote(value)),
      useragent = paste0("--user-agent ", dquote(value)),
      verbose = if (value) "--verbose",
      cookiejar = paste0("--cookie-jar ", dquote(value)),
      cookiefile = paste0("--cookie ", dquote(value)),
      cookie = paste0("--cookie ", dquote(value))
    )
  })

  # httr2 follows redirects by default, but command line curl doesn't
  follow <- if (!isFALSE(options$followlocation)) "--location"

  c(follow, unlist(args))
}

curl_body <- function(req, obfuscated = c("redact", "reveal")) {
  obfuscated <- arg_match(obfuscated)

  body <- req_get_body(req, obfuscated = obfuscated)
  if (is.null(body)) {
    return(NULL)
  }
  type <- req_get_body_type(req)
  if (type == "raw") {
    cli::cli_abort("Can't translate a request with a raw body.", call = NULL)
  }

  c(
    curl_content_type(req, type),
    curl_body_data(body, type, params = req$body$params)
  )
}

# Skip if Content-Type is already set as a header
curl_content_type <- function(req, type) {
  if ("content-type" %in% tolower(names(req_get_headers(req)))) {
    return(NULL)
  }

  content_type <- req$body$content_type %||%
    switch(
      type,
      json = "application/json",
      form = "application/x-www-form-urlencoded"
    )
  if (is.null(content_type) || !nzchar(content_type)) {
    return(NULL)
  }
  paste0("--header ", dquote(paste0("Content-Type: ", content_type)))
}

curl_body_data <- function(body, type, params = list(auto_unbox = TRUE)) {
  switch(
    type,
    string = paste0("--data ", dquote(body)),
    file = paste0("--data-binary ", dquote(paste0("@", body))),
    json = paste0("--data ", dquote(exec(jsonlite::toJSON, body, !!!params))),
    form = paste0("--data ", dquote(url_query_build(body))),
    multipart = curl_body_multipart(body)
  )
}

curl_body_multipart <- function(body) {
  unlist(Map(curl_body_multipart_field, names(body), body), use.names = FALSE)
}

curl_body_multipart_field <- function(name, value) {
  if (inherits(value, "form_file")) {
    spec <- paste0(name, "=@", curl_form_quote(value$path))
    if (!is.null(value$type)) {
      spec <- paste0(spec, ";type=", value$type)
    }
    if (!is.null(value$name)) {
      spec <- paste0(spec, ";filename=", curl_form_quote(value$name))
    }
    paste0("--form ", dquote(spec))
  } else if (inherits(value, "form_data")) {
    type <- value$type
    value <- rawToChar(value$value)
    if (is.null(type)) {
      paste0("--form-string ", dquote(paste0(name, "=", value)))
    } else {
      spec <- paste0(name, "=", curl_form_quote(value), ";type=", type)
      paste0("--form ", dquote(spec))
    }
  } else {
    paste0("--form-string ", dquote(paste0(name, "=", value)))
  }
}

curl_form_quote <- function(x) {
  paste0('"', gsub('(["\\\\])', "\\\\\\1", x), '"')
}

dquote <- function(x) {
  ifelse(
    grepl("[^A-Za-z0-9._~:/@%+=,-]", x),
    paste0("'", gsub("'", "'\"'\"'", x, fixed = TRUE), "'"),
    x
  )
}

test_that("httr2_translate() works with basic GET requests", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with POST methods", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_method("POST") |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with headers", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      req_headers(
        "Accept" = "application/json",
        "User-Agent" = "httr2/1.0"
      ) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with JSON bodies", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_json(list(name = "test", value = 123)) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with form bodies", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_form(name = "test", value = "123") |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with multipart bodies", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_multipart(name = "test", value = "123") |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with string bodies", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_raw("test data", type = "text/plain") |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with file bodies", {
  local_mocked_user_agent()
  path <- tempfile()
  writeLines("test content", path)

  # normalize the path
  path <- normalizePath(path, winslash = "/")

  expect_snapshot(
    {
      request("https://hb.cran.dev/post") |>
        req_body_file(path, type = "text/plain") |>
        httr2_translate()
    },
    transform = function(x) {
      gsub(path, "<tempfile>", x, fixed = TRUE)
    }
  )
})

test_that("httr2_translate() works with custom content types", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_json(
        list(test = "data"),
        type = "application/vnd.api+json"
      ) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with options", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      req_options(verbose = TRUE, ssl_verifypeer = FALSE) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with cookies", {
  local_mocked_user_agent()
  cookie_file <- tempfile()

  # create the tempfile
  file.create(cookie_file)

  # normalize the path
  cookie_file <- normalizePath(cookie_file, winslash = "/")

  expect_snapshot(
    {
      request("https://hb.cran.dev/cookies") |>
        req_options(cookiejar = cookie_file, cookiefile = cookie_file) |>
        httr2_translate()
    },
    transform = function(x) {
      gsub(cookie_file, "<cookie-file>", x, fixed = TRUE)
    }
  )
})

test_that("httr2_translate() works with obfuscated values in headers", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      req_headers("Authorization" = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")) |>
      httr2_translate()
  })
})

test_that("httr2_translate() can reveal obfuscated values", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      req_headers_redacted(Authorization = "secret-token") |>
      httr2_translate(obfuscated = "reveal")
  })
})

test_that("httr2_translate() works with obfuscated values in JSON body", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_json(list(
        username = "test",
        password = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")
      )) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with obfuscated values in form body", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_body_form(
        username = "test",
        password = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")
      ) |>
      httr2_translate()
  })
})

test_that("httr2_translate() works with complex requests", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://api.github.com/user/repos") |>
      req_method("POST") |>
      req_headers(
        "Accept" = "application/vnd.github.v3+json",
        "Authorization" = obfuscated("ZdYJeG8zwISodg0nu4UxBhs"),
        "User-Agent" = "MyApp/1.0"
      ) |>
      req_body_json(list(
        name = "test-repo",
        description = "A test repository",
        private = TRUE
      )) |>
      httr2_translate()
  })
})

test_that("httr2_translate() puts a request with no arguments on a single line", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/get") |>
      req_options(followlocation = FALSE) |>
      httr2_translate()
  })
})

test_that("httr2_translate() validates input", {
  expect_snapshot(error = TRUE, {
    httr2_translate("not a request")
  })
})

test_that("httr2_translate() signs AWS requests", {
  req <- request("https://sts.us-east-1.amazonaws.com/") |>
    req_body_form(
      Action = "GetCallerIdentity",
      Version = "2011-06-15"
    ) |>
    req_auth_aws_v4(
      aws_access_key_id = "AKIAIOSFODNN7EXAMPLE",
      aws_secret_access_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
    )

  command <- as.character(httr2_translate(req, obfuscated = "reveal"))
  expect_match(
    command,
    "Authorization: AWS4-HMAC-SHA256 Credential=AKIAIOSFODNN7EXAMPLE/"
  )
  expect_match(
    command,
    "SignedHeaders=host;x-amz-date"
  )
  expect_match(command, "--header 'x-amz-date: [0-9]{8}T[0-9]{6}Z'")
})

test_that("httr2_translate() errors for raw bodies", {
  req <- request("https://hb.cran.dev/post") |>
    req_body_raw(as.raw(c(0x00, 0x68, 0x69, 0xff)))
  expect_snapshot(httr2_translate(req), error = TRUE)
})

test_that("an explicit Content-Type header isn't duplicated by the body", {
  local_mocked_user_agent()
  expect_snapshot({
    request("https://hb.cran.dev/post") |>
      req_headers("Content-Type" = "application/json") |>
      req_body_raw("{}") |>
      httr2_translate()
  })
})

test_that("dquote() quotes only when needed", {
  # plain values are left alone
  expect_equal(dquote("https://example.com/get"), "https://example.com/get")
  # spaces and query-string metacharacters force quoting
  expect_equal(dquote("a b"), "'a b'")
  expect_equal(
    dquote("https://example.com?a=1&b=2"),
    "'https://example.com?a=1&b=2'"
  )
})

test_that("dquote() protects shell metacharacters", {
  expect_equal(dquote('a"b'), "'a\"b'")
  expect_equal(dquote("a'b"), "'a'\"'\"'b'")
  expect_equal(dquote("a$b`c`\\d"), "'a$b`c`\\d'")
})

test_that("curl_body_data() safely quotes strings and JSON", {
  expect_equal(
    curl_body_data("a'b$HOME", "string"),
    "--data 'a'\"'\"'b$HOME'"
  )
  expect_equal(
    curl_body_data(list(value = "a'b$HOME"), "json"),
    "--data '{\"value\":\"a'\"'\"'b$HOME\"}'"
  )
})

test_that("curl_body_data() uses the request's JSON parameters", {
  expect_equal(
    curl_body_data(
      list(x = 1.23456, y = NULL),
      "json",
      params = list(auto_unbox = TRUE, digits = 2, null = "list")
    ),
    "--data '{\"x\":1.23,\"y\":{}}'"
  )
})

test_that("curl_body_data() URL encodes form data", {
  expect_equal(
    curl_body_data(list(x = "a b&c=d", y = "x+y", z = "é"), "form"),
    "--data 'x=a%20b%26c%3Dd&y=x%2By&z=%C3%A9'"
  )
})

test_that("curl_body_data() translates multipart values", {
  path <- file.path(withr::local_tempdir(), "contents")
  writeLines("contents", path)

  body <- list(
    text = "@literal;value",
    file = curl::form_file(path, type = "text/plain", name = "name.txt"),
    data = curl::form_data("a b", type = "text/plain")
  )
  expect_snapshot(
    writeLines(curl_body_data(body, "multipart")),
    # the temp file path varies and uses (escaped) \ on Windows
    transform = function(x) gsub('@"[^"]+"', '@"<tmppath>"', x)
  )
})

test_that("curl_method() only sets the method when curl can't infer it", {
  req <- request("https://example.com")

  # GET is the default, and curl infers POST from a body
  expect_null(curl_method(req))
  expect_null(curl_method(req_method(req, "POST"), has_body = TRUE))

  # HEAD has its own flag
  expect_equal(curl_method(req_method(req, "HEAD")), "--head")

  # a body-less POST and other methods need --request
  expect_equal(curl_method(req_method(req, "POST")), "--request POST")
  expect_equal(
    curl_method(req_method(req, "DELETE")),
    "--request DELETE"
  )
  # a body alone implies POST, but not PUT/DELETE/etc.
  expect_equal(
    curl_method(req_method(req, "PUT"), has_body = TRUE),
    "--request PUT"
  )
})

test_that("curl_headers() drops missing headers and reveals secrets", {
  expect_null(curl_headers(request("https://example.com")))

  req <- request("https://example.com") |>
    req_headers_redacted(Authorization = "secret")
  expect_equal(
    curl_headers(req, "redact"),
    "--header 'Authorization: <REDACTED>'"
  )
  expect_equal(
    curl_headers(req, "reveal"),
    "--header 'Authorization: secret'"
  )
})

test_that("curl_options() translates each known option", {
  req <- request("https://example.com") |>
    req_options(
      timeout_ms = 30000,
      connecttimeout = 5,
      proxy = "http://proxy.example.com",
      useragent = "agent",
      followlocation = TRUE,
      verbose = TRUE,
      cookiejar = "jar.txt",
      cookiefile = "file.txt"
    )
  expect_snapshot(cat(curl_options(req), sep = "\n"))
})

test_that("curl_options() translates options set by httr2 functions", {
  req <- request("https://example.com") |>
    req_timeout(30) |>
    req_proxy(
      "proxy.example.com",
      port = 8080,
      username = "u",
      password = "p"
    ) |>
    req_user_agent("agent") |>
    req_cookie_preserve("cookies.txt") |>
    req_cookies_set(session = "abc")
  expect_snapshot(cat(curl_options(req), sep = "\n"))
})

test_that("curl_options() follows redirects by default, unlike curl", {
  expect_equal(
    curl_options(request("https://example.com")),
    "--location"
  )

  req <- request("https://example.com") |>
    req_options(followlocation = FALSE)
  expect_null(curl_options(req))
})

test_that("curl_options() ignores options with no curl equivalent", {
  # req_verbose() also sets a debugfunction; req_progress() sets callbacks
  req <- request("https://example.com") |>
    req_options(followlocation = FALSE) |>
    req_verbose() |>
    req_progress()
  expect_no_warning(out <- curl_options(req))
  expect_equal(out, "--verbose")
})

test_that("curl_options() drops disabled flags", {
  req <- request("https://example.com") |>
    req_options(followlocation = FALSE, verbose = FALSE)
  expect_null(curl_options(req))
})

test_that("curl_options() warns about untranslatable options", {
  req <- request("https://example.com") |>
    req_options(followlocation = FALSE, ssl_verifypeer = FALSE)
  expect_snapshot(out <- curl_options(req))
  expect_null(out)
})

#' Perform a list of requests in parallel
#'
#' @description
#' This variation on [req_perform_sequential()] performs multiple requests in
#' parallel. Never use it without [req_throttle()]; otherwise it's too easy to
#' pummel a server with a very large number of simultaneous requests.
#'
#' While running, you'll get a progress bar that looks like:
#' `[working] (1 + 4) -> 5 -> 5`. The string tells you the current status of
#' the queue (e.g. working, waiting, errored) followed by (the
#' number of pending requests + pending retried requests) -> the number of
#' active requests -> the number of complete requests.
#'
#' ## Limitations
#'
#' The main limitation of `req_perform_parallel()` is that it assumes applies
#' [req_throttle()] and [req_retry()] are across all requests. This means,
#' for example, that if request 1 is throttled, but request 2 is not,
#' `req_perform_parallel()` will wait for request 1 before performing request 2.
#' This makes it most suitable for performing many parallel requests to the same
#' host, rather than a mix of different hosts. It's probably possible to remove
#' these limitation, but it's enough work that I'm unlikely to do it unless
#' I know that people would fine it useful: so please let me know!
#'
#' Additionally, it does not respect the `max_tries` argument to `req_retry()`
#' because if you have five requests in flight and the first one gets rate
#' limited, it's likely that all the others do too. This also means that
#' the circuit breaker is never triggered.
#'
#' @inherit req_perform_sequential params return
#' @param max_active Maximum number of concurrent requests.
#' @export
#' @examples
#' # Requesting these 4 pages one at a time would take 2 seconds:
#' request_base <- request(example_url()) |>
#'   req_throttle(capacity = 100, fill_time_s = 60)
#' reqs <- list(
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5"),
#'   request_base |> req_url_path("/delay/0.5")
#' )
#' # But it's much faster if you request in parallel
#' system.time(resps <- req_perform_parallel(reqs))
#'
#' # req_perform_parallel() will fail on error
#' reqs <- list(
#'   request_base |> req_url_path("/status/200"),
#'   request_base |> req_url_path("/status/400"),
#'   request("FAILURE")
#' )
#' try(resps <- req_perform_parallel(reqs))
#'
#' # but can use on_error to capture all successful results
#' resps <- req_perform_parallel(reqs, on_error = "continue")
#'
#' # Inspect the successful responses
#' resps |> resps_successes()
#'
#' # And the failed responses
#' resps |> resps_failures() |> resps_requests()
req_perform_parallel <- function(
  reqs,
  paths = NULL,
  on_error = c("stop", "return", "continue"),
  progress = TRUE,
  max_active = 10,
  mock = getOption("httr2_mock", NULL)
) {
  check_paths(paths, reqs)
  on_error <- arg_match(on_error)
  check_number_whole(max_active, min = 1)
  mock <- as_mock_function(mock, error_call)

  queue <- RequestQueue$new(
    reqs = reqs,
    paths = paths,
    max_active = max_active,
    on_error = on_error,
    progress = progress,
    mock = mock,
    frame = environment()
  )

  tryCatch(
    queue$process(),
    interrupt = function(cnd) {
      check_repeated_interrupt()

      queue$queue_status <- "errored"
      queue$process()

      n <- sum(!map_lgl(queue$resps, is.null))
      cli::cli_alert_warning(
        "Terminating iteration; returning {n} response{?s}."
      )
    }
  )
  queue$progress$done()

  if (on_error == "stop") {
    is_error <- map_lgl(queue$resps, is_error)
    if (any(is_error)) {
      i <- which(is_error)[[1]]
      the$last_response <- queue$resps[[i]]$resp %||% queue$resps[[i]]
      the$last_request <- queue$reqs[[i]]
      cnd_signal(queue$resps[[i]])
    }
  }

  queue$resps
}

RequestQueue <- R6::R6Class(
  "RequestQueue",
  public = list(
    pool = NULL,
    rate_limit_deadline = 0,
    token_deadline = 0,
    max_active = NULL,

    # Overall status for the queue
    queue_status = NULL,
    n_pending = 0,
    n_active = 0,
    n_complete = 0,
    n_retries = 0,
    on_error = "stop",
    mock = NULL,
    progress = NULL,

    # Vectorised along reqs
    reqs = list(),
    pooled_reqs = list(),
    resps = list(),
    status = character(),
    tries = integer(),

    # Requests that have failed due to OAuth expiration; used to ensure that we
    # don't retry repeatedly, but still allow all active requests to retry once
    oauth_failed = integer(),

    initialize = function(
      reqs,
      paths = NULL,
      max_active = 10,
      on_error = "stop",
      progress = FALSE,
      mock = NULL,
      frame = caller_env()
    ) {
      n <- length(reqs)

      self$progress <- create_progress_bar(
        progress,
        total = n,
        format = paste0(
          "[{self$queue_status}] ",
          "({self$n_pending} + {self$n_retries}) -> {self$n_active} -> {self$n_complete} | ",
          "{cli::pb_bar} {cli::pb_percent}"
        ),
        frame = frame
      )

      # goal is for pool to not do any queueing; i.e. the curl pool will
      # only ever contain requests that we actually want to process. Any
      # throttling is done by `req_throttle()`
      self$max_active <- max_active
      self$pool <- curl::new_pool(
        total_con = 100,
        host_con = 100,
        max_streams = 100
      )
      self$on_error <- on_error
      self$mock <- mock

      self$queue_status <- "working"
      self$n_pending <- n
      self$n_active <- 0
      self$n_complete <- 0

      self$reqs <- reqs
      self$pooled_reqs <- map(seq_along(reqs), function(i) {
        pooled_request(
          req = reqs[[i]],
          path = paths[[i]],
          on_success = function(resp) self$done_success(i, resp),
          on_failure = function(error) self$done_failure(i, error),
          on_error = function(error) self$done_error(i, error),
          mock = mock,
          error_call = frame
        )
      })
      self$resps <- vector("list", n)
      self$status <- rep("pending", n)
      self$tries <- rep(0L, n)
    },

    process = function(timeout = Inf) {
      deadline <- unix_time() + timeout

      while (unix_time() <= deadline) {
        out <- self$process1(deadline)
        if (!is.null(out)) {
          return(out)
        }
      }

      TRUE
    },

    # Exposed for testing, so we can manaully work through one step at a time
    process1 = function(deadline = Inf) {
      if (self$queue_status == "done") {
        return(FALSE)
      }

      self$progress$update(set = self$n_complete)

      if (self$queue_status == "waiting") {
        request_deadline <- max(self$token_deadline, self$rate_limit_deadline)
        if (request_deadline <= unix_time()) {
          self$queue_status <- "working"
          return()
        }

        if (self$rate_limit_deadline > self$token_deadline) {
          waiting <- "for rate limit"
        } else {
          waiting <- "for throttling"
        }
        pool_wait_for_deadline(
          self$pool,
          min(request_deadline, deadline),
          waiting
        )
        NULL
      } else if (self$queue_status == "working") {
        if (self$n_pending == 0 && self$n_active == 0) {
          self$queue_status <- "done"
        } else if (self$n_pending > 0 && self$n_active <= self$max_active) {
          if (!self$submit_next(deadline)) {
            self$queue_status <- "waiting"
          }
        } else {
          pool_wait_for_one(self$pool, deadline)
        }
        NULL
      } else if (self$queue_status == "errored") {
        # Finish out any active requests but don't add any more
        if (self$n_active > 0) {
          pool_wait_for_one(self$pool, deadline)
        } else {
          self$queue_status <- "done"
        }
        NULL
      }
    },

    submit_next = function(deadline) {
      i <- which(self$status == "pending")[[1]]

      self$token_deadline <- throttle_deadline(self$reqs[[i]])
      if (self$token_deadline > unix_time()) {
        throttle_return_token(self$reqs[[i]])
        return(FALSE)
      }

      self$set_status(i, "active")
      self$resps[i] <- list(NULL)
      self$tries[[i]] <- self$tries[[i]] + 1

      self$pooled_reqs[[i]]$submit(self$pool)
      TRUE
    },

    done_success = function(i, resp) {
      self$set_status(i, "complete")
      self$resps[[i]] <- resp

      self$oauth_failed <- NULL
    },

    done_error = function(i, error) {
      self$resps[[i]] <- error
      self$set_status(i, "complete")
      if (self$on_error != "continue") {
        self$queue_status <- "errored"
      }
    },

    done_failure = function(i, error) {
      req <- self$reqs[[i]]
      resp <- error$resp
      self$resps[[i]] <- error
      tries <- self$tries[[i]]

      if (retry_is_transient(req, resp) && self$can_retry(i)) {
        delay <- retry_after(req, resp, tries)
        self$rate_limit_deadline <- unix_time() + delay

        self$set_status(i, "pending")
        self$n_retries <- self$n_retries + 1
        self$queue_status <- "waiting"
      } else if (resp_is_invalid_oauth_token(req, resp) && self$can_reauth(i)) {
        # This isn't quite right, because if there are (e.g.) four requests in
        # the queue and the first one fails, we'll clear the cache for all four,
        # causing a token refresh more often than necessary. This shouldn't
        # affect correctness, but it does make it slower than necessary.
        self$oauth_failed <- c(self$oauth_failed, i)
        req_auth_clear_cache(self$reqs[[i]])
        self$set_status(i, "pending")
        self$n_retries <- self$n_retries + 1
      } else {
        self$set_status(i, "complete")
        if (self$on_error != "continue") {
          self$queue_status <- "errored"
        }
      }
    },

    set_status = function(i, status) {
      switch(
        self$status[[i]], # old status
        pending = self$n_pending <- self$n_pending - 1,
        active = self$n_active <- self$n_active - 1
      )
      switch(
        status, # new status
        pending = self$n_pending <- self$n_pending + 1,
        active = self$n_active <- self$n_active + 1,
        complete = self$n_complete <- self$n_complete + 1
      )

      self$status[[i]] <- status
    },

    can_retry = function(i) {
      TRUE
      # self$tries[[i]] < retry_max_tries(self$reqs[[i]])
    },
    can_reauth = function(i) {
      !i %in% self$oauth_failed
    }
  )
)

pool_wait_for_one <- function(pool, deadline) {
  timeout <- deadline - unix_time()
  pool_wait(pool, poll = TRUE, timeout = timeout)
}

pool_wait_for_deadline <- function(pool, deadline, waiting_for) {
  now <- unix_time()
  timeout <- deadline - now
  if (timeout <= 0) {
    return(TRUE)
  }

  complete <- pool_wait(pool, poll = FALSE, timeout = timeout)

  # pool might finish early; we still want to wait out the full time
  remaining <- timeout - (unix_time() - now)
  if (remaining > 2) {
    # Use a progress bar
    sys_sleep(remaining, waiting_for)
  } else if (remaining > 0) {
    Sys.sleep(remaining)
  }

  complete
}

pool_wait <- function(pool, poll, timeout) {
  signal("", class = "httr2_pool_wait", timeout = timeout)
  done <- curl::multi_run(pool = pool, poll = poll, timeout = timeout)
  (done$success + done$error) > 0 || done$pending == 0
}

test_that("request and paths must match", {
  req <- request("http://example.com")
  expect_snapshot(req_perform_parallel(req, letters), error = TRUE)
})

test_that("can perform zero requests", {
  expect_equal(req_perform_parallel(list()), list())
})

test_that("can perform a single request", {
  reqs <- list(request_test("/get"))
  resps <- req_perform_parallel(reqs)
  expect_type(resps, "list")
  expect_length(resps, 1)

  resp <- resps[[1]]
  expect_s3_class(resp, "httr2_response")
  expect_equal(resp$method, "GET")
  expect_equal(resp$url, example_url("/get"))
  expect_equal(resp$status_code, 200)
  expect_s3_class(resp$headers, "httr2_headers")
  expect_type(resp$body, "raw")
  expect_equal(resp$request, reqs[[1]])
})

test_that("requests happen in parallel", {
  skip_on_cran()

  # test works best if webfakes has ample threads and keepalive
  reqs <- list2(
    request_test("/delay/:secs", secs = 0),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
    request_test("/delay/:secs", secs = 0.25),
  )
  time <- system.time(req_perform_parallel(reqs))
  expect_lt(time[[3]], 1)
})

test_that("can perform >128 file uploads in parallel", {
  temp <- withr::local_tempfile(lines = letters)
  req <- request(example_url()) |> req_body_file(temp)
  reqs <- rep(list(req), 130)

  expect_no_error(req_perform_parallel(reqs, on_error = "continue"))
})

test_that("can download files", {
  reqs <- list(request_test("/json"), request_test("/html"))
  paths <- c(withr::local_tempfile(), withr::local_tempfile())
  resps <- req_perform_parallel(reqs, paths)

  expect_equal(resps[[1]]$body, new_path(paths[[1]]))
  expect_equal(resps[[2]]$body, new_path(paths[[2]]))

  # And check that something was downloaded
  expect_gt(file.size(paths[[1]]), 0)
  expect_gt(file.size(paths[[2]]), 0)
})

test_that("can download 0 byte file", {
  reqs <- list(request_test("/bytes/0"))
  paths <- withr::local_tempfile()
  resps <- req_perform_parallel(reqs, paths = paths)

  expect_equal(file.size(paths[[1]]), 0)
})

test_that("objects are cached", {
  temp <- withr::local_tempdir()
  req <- request_test("etag/:etag", etag = "abcd") |> req_cache(temp)

  expect_condition(
    resps1 <- req_perform_parallel(list(req)),
    class = "httr2_cache_save"
  )

  expect_condition(
    resps2 <- req_perform_parallel(list(req)),
    class = "httr2_cache_not_modified"
  )
})

test_that("immutable objects retrieved from cache", {
  req <- request("http://example.com") |> req_cache(tempfile())
  resp <- response(
    200,
    headers = "Expires: Wed, 01 Jan 3000 00:00:00 GMT",
    body = charToRaw("abc")
  )
  cache_set(req, resp)

  expect_condition(
    resps <- req_perform_parallel(list(req)),
    class = "httr2_cache_cached"
  )
  expect_equal(resps[[1]], resp)
})

test_that("errors by default", {
  req <- request_test("/status/:status", status = 404)
  err <- expect_error(req_perform_parallel(list(req)))
  expect_s3_class(err, "httr2_http_404")

  # Wraps and forwards curl errors
  req <- request("INVALID")
  err <- expect_error(req_perform_parallel(list(req)))
  expect_s3_class(err, "httr2_failure")
  expect_s3_class(err$parent, "curl_error_couldnt_resolve_host")
})

test_that("both curl and HTTP errors become errors on continue", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request("INVALID"),
  )
  out <- req_perform_parallel(reqs, on_error = "continue")
  expect_s3_class(out[[1]], "httr2_http_404")
  expect_s3_class(out[[2]], "httr2_failure")

  # and contain the responses
  expect_equal(out[[1]]$request, reqs[[1]])
  expect_equal(out[[2]]$request, reqs[[2]])
})

test_that("errors can cancel outstanding requests", {
  reqs <- list2(
    request_test("/status/:status", status = 404),
    request_test("/delay/:secs", secs = 1),
    request_test("/delay/:secs", secs = 1),
  )
  out <- req_perform_parallel(reqs, on_error = "return", max_active = 1)
  expect_s3_class(out[[1]], "httr2_http_404")
  # second request might succeed or fail depend on the timing, but the
  # third request should definitely fail
  expect_null(out[[3]])
})

test_that("req_perform_parallel resspects http_error() error override", {
  reqs <- list2(
    req_error(
      request_test("/status/:status", status = 404),
      is_error = \(resp) FALSE
    ),
    req_error(
      request_test("/status/:status", status = 500),
      is_error = \(resp) FALSE
    )
  )
  resps <- req_perform_parallel(reqs)

  expect_equal(resp_status(resps[[1]]), 404)
  expect_equal(resp_status(resps[[2]]), 500)
})

test_that("req_perform_parallel respects http_error() body message", {
  reqs <- list2(
    req_error(
      request_test("/status/:status", status = 404),
      body = \(resp) "hello"
    )
  )
  expect_snapshot(req_perform_parallel(reqs), error = TRUE)
})

test_that("requests are throttled", {
  withr::defer(throttle_reset())

  mock_time <- 0
  local_mocked_bindings(
    unix_time = function() mock_time,
    Sys.sleep = function(seconds) mock_time <<- mock_time + seconds
  )

  req <- request_test("/status/:status", status = 200)
  req <- req |> req_throttle(capacity = 1, fill_time_s = 1)
  reqs <- rep(list(req), 5)

  queue <- RequestQueue$new(reqs, progress = FALSE)
  queue$process()
  expect_equal(mock_time, 4)
})


# Tests of lower-level operation -----------------------------------------------

test_that("can retry an OAuth failure", {
  req <- local_app_request(function(req, res) {
    if (res$app$locals$i == 1) {
      res$set_status(401)$set_header(
        "WWW-Authenticate",
        'Bearer realm="example", error="invalid_token"'
      )$send_json(list(status = "failed"), auto_unbox = TRUE)
    } else {
      res$send_json(list(status = "done"), auto_unbox = TRUE)
    }
  })
  req <- req_policies(req, auth_oauth = TRUE)

  reset <- 0
  local_mocked_bindings(
    req_auth_clear_cache = function(...) reset <<- reset + 1
  )

  queue <- RequestQueue$new(list(req), progress = FALSE)
  queue$process()

  expect_equal(reset, 1)
  expect_equal(resp_body_json(queue$resps[[1]]), list(status = "done"))
})

test_that("but multiple failures causes an error", {
  req <- local_app_request(function(req, res) {
    res$set_status(401)$set_header(
      "WWW-Authenticate",
      'Bearer realm="example", error="invalid_token"'
    )$send_json(list(status = "failed"), auto_unbox = TRUE)
  })
  req <- req_policies(req, auth_oauth = TRUE)

  queue <- RequestQueue$new(list(req), progress = FALSE)
  queue$process()
  expect_s3_class(queue$resps[[1]], "httr2_http_401")
})

test_that("can retry a transient error", {
  req <- local_app_request(function(req, res) {
    if (res$app$locals$i == 1) {
      res$set_status(429)$set_header("retry-after", 2)$send_json(
        list(status = "waiting"),
        auto_unbox = TRUE
      )
    } else {
      res$send_json(list(status = "done"), auto_unbox = TRUE)
    }
  })
  req <- req_retry(req, max_tries = 2)

  mock_time <- 1
  local_mocked_bindings(
    unix_time = function() mock_time,
    Sys.sleep = function(seconds) mock_time <<- mock_time + seconds
  )

  queue <- RequestQueue$new(list(req), progress = FALSE)

  # submit the request
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "working")
  expect_equal(queue$n_active, 1)
  expect_equal(queue$n_pending, 0)
  expect_equal(queue$status[[1]], "active")

  # process the response and capture the retry
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "waiting")
  expect_equal(queue$rate_limit_deadline, mock_time + 2)
  expect_equal(queue$n_pending, 1)
  expect_s3_class(queue$resps[[1]], "httr2_http_429")
  expect_equal(resp_body_json(queue$resps[[1]]$resp), list(status = "waiting"))

  # Starting waiting
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "waiting")
  expect_equal(mock_time, 3)

  # Finishing waiting
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "working")
  expect_equal(queue$n_active, 0)
  expect_equal(queue$n_pending, 1)

  # Resubmit
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "working")
  expect_equal(queue$n_active, 1)
  expect_equal(queue$n_pending, 0)

  # Process the response
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "working")
  expect_equal(queue$n_active, 0)
  expect_equal(queue$n_pending, 0)
  expect_s3_class(queue$resps[[1]], "httr2_response")
  expect_equal(resp_body_json(queue$resps[[1]]), list(status = "done"))

  # So we're finally done
  expect_null(queue$process1())
  expect_equal(queue$queue_status, "done")
  expect_false(queue$process1())
})

test_that("throttling is limited by deadline", {
  withr::defer(throttle_reset("test"))

  mock_time <- 0
  local_mocked_bindings(
    unix_time = function() mock_time,
    Sys.sleep = function(seconds) mock_time <<- mock_time + seconds
  )

  req <- request_test("/status/:status", status = 200)
  req <- req_throttle(req, capacity = 1, fill_time_s = 1, realm = "test")
  queue <- RequestQueue$new(list(req), progress = FALSE)

  # Check time only advances by one second, and token is returned to bucket
  local_mocked_bindings(throttle_deadline = function(...) mock_time + 2)
  queue$process1(1)
  expect_equal(queue$queue_status, "waiting")
  queue$process1(1)
  expect_equal(mock_time, 1)
  expect_equal(the$throttle[["test"]]$tokens, 1)

  local_mocked_bindings(throttle_deadline = function(...) mock_time)
  queue$rate_limit_deadline <- mock_time + 2
  expect_equal(mock_time, 1)
  expect_equal(the$throttle[["test"]]$tokens, 1)
})


test_that("mocking works", {
  req_200 <- request("https://ok")
  req_404 <- request("https://notok")

  local_mocked_responses(function(req) {
    if (req$url == "https://ok") {
      response()
    } else {
      response(404)
    }
  })

  resps <- req_perform_parallel(list(req_200, req_404), on_error = "continue")
  expect_equal(resps[[1]], response())
  expect_s3_class(resps[[2]], "httr2_http_404")
})

# otel -----------------------------------------------------------------------

test_that("tracing works as expected", {
  skip_if_not_installed("otelsdk")

  spans <- with_otel_record({
    # A request with no URL (which shouldn't create a span).
    try(req_perform_parallel(list(request(""))), silent = TRUE)

    # A request with an HTTP error.
    try(
      req_perform_parallel(list(request_test("/status/:status", status = 404))),
      silent = TRUE
    )

    # A request with a curl error.
    with_mocked_bindings(
      try(req_perform(request("http://127.0.0.1")), silent = TRUE),
      curl_fetch = function(...) abort("Failed to connect")
    )

    # Three regular requests, nested inside a parent span.
    parent <- otel::start_span("parent", tracer = "test")
    otel::with_active_span(parent, {
      resp <- req_perform_parallel(list(
        request_test("/headers"),
        request_test("/headers"),
        request_test("/headers")
      ))
    })
    parent$end()

    # Verify that context propagation works as expected.
    expect_true(
      "traceparent" %in% names(resp_body_json(resp[[1]])[["headers"]])
    )
  })[["traces"]]

  expect_length(spans, 6L)

  # And for requests with HTTP errors.
  expect_equal(spans[[1]]$status, "error")
  expect_equal(spans[[1]]$description, "Not Found")
  expect_equal(spans[[1]]$attributes$http.response.status_code, 404L)
  expect_equal(spans[[1]]$attributes$error.type, "404")

  # And for spans with curl errors.
  expect_equal(spans[[2]]$status, "error")
  expect_equal(spans[[2]]$attributes$error.type, "rlang_error")

  # We should have attached the curl error as an event.
  expect_length(spans[[2]]$events, 1L)
  expect_equal(spans[[2]]$events[[1]]$name, "exception")

  # Verify that the parent span is the same for parallel requests (that is,
  # they are siblings).
  expect_equal(spans[[3]]$parent, spans[[6]]$span_id)
  expect_equal(spans[[4]]$parent, spans[[6]]$span_id)
  expect_equal(spans[[5]]$parent, spans[[6]]$span_id)
})

# Pool helpers ----------------------------------------------------------------

test_that("wait for deadline waits after pool complete", {
  pool <- curl::new_pool()
  deadline <- unix_time() + 1

  slept <- 0
  local_mocked_bindings(
    unix_time = function() 0,
    Sys.sleep = function(seconds) mock_time <<- slept <<- seconds
  )

  expect_true(pool_wait_for_deadline(pool, deadline = 1))
  expect_equal(slept, 1)
})

#' Create an OAuth client
#'
#' An OAuth app is the combination of a client, a set of endpoints
#' (i.e. urls where various requests should be sent), and an authentication
#' mechanism. A client consists of at least a `client_id`, and also often
#' a `client_secret`. You'll get these values when you create the client on
#' the API's website.
#'
#' @param id Client identifier.
#' @param token_url Url to retrieve an access token.
#' @param secret Client secret. For most apps, this is technically confidential
#'   so in principle you should avoid storing it in source code. However, many
#'   APIs require it in order to provide a user friendly authentication
#'   experience, and the risks of including it are usually low. To make things
#'   a little safer, I recommend using [obfuscate()] when recording the client
#'   secret in public code.
#' @param key Client key. As an alternative to using a `secret`, you can
#'   instead supply a confidential private key. This should never be included
#'   in a package.
#' @param auth Authentication mechanism used by the client to prove itself to
#'   the API. Can be one of three built-in methods ("body", "header", or "jwt"),
#'   or a function that will be called with arguments `req`, `client`, and
#'   the contents of `auth_params`.
#'
#'   The most common mechanism in the wild is `"body"` where the `client_id` and
#'   (optionally) `client_secret` are added to the body. `"header"` sends the
#'   `client_id` and `client_secret` in HTTP Authorization header. `"jwt_sig"`
#'   will generate a JWT, and include it in a `client_assertion` field in the
#'   body.
#'
#'   See [oauth_client_req_auth()] for more details.
#' @param auth_params Additional parameters passed to the function specified
#'   by `auth`.
#' @param name Optional name for the client. Used when generating the cache
#'   directory. If `NULL`, generated from hash of `client_id`. If you're
#'   defining a client for use in a package, I recommend that you use
#'   the package name.
#' @return An OAuth client: An S3 list with class `httr2_oauth_client`.
#' @export
#' @examples
#' oauth_client("myclient", "http://example.com/token_url", secret = "DONTLOOK")
oauth_client <- function(
  id,
  token_url,
  secret = NULL,
  key = NULL,
  auth = c("body", "header", "jwt_sig"),
  auth_params = list(),
  name = hash(id)
) {
  check_string(id)
  check_string(token_url)
  check_string(secret, allow_null = TRUE)

  if (is.character(auth)) {
    if (missing(auth)) {
      auth <- if (is.null(key)) "body" else "jwt_sig"
    }
    auth <- arg_match(auth)

    if (auth == "header" && is.null(secret)) {
      cli::cli_abort("{.code auth = 'header'} requires a {.arg secret}.")
    } else if (auth == "jwt_sig") {
      if (is.null(key)) {
        cli::cli_abort("{.code auth = 'jwt_sig'} requires a {.arg key}.")
      }
      if (!has_name(auth_params, "claim")) {
        cli::cli_abort(
          "{.code auth = 'jwt_sig'} requires a claim specification in {.arg auth_params}."
        )
      }
    }

    auth <- paste0("oauth_client_req_auth_", auth)
  } else if (!is_function(auth)) {
    cli::cli_abort("{.arg auth} must be a string or function.")
  }

  structure(
    list(
      name = name,
      id = id,
      secret = secret,
      key = key,
      token_url = token_url,
      auth = auth,
      auth_params = auth_params
    ),
    class = "httr2_oauth_client"
  )
}

#' @export
print.httr2_oauth_client <- function(x, ...) {
  cli::cat_line(cli::style_bold("<", paste(class(x), collapse = "/"), ">"))
  redacted <- list_redact(compact(x), c("secret", "key"))
  bullets(redacted)
  invisible(x)
}

#' OAuth client authentication
#'
#' @description
#' `oauth_client_req_auth()` authenticates a request using the authentication
#' strategy defined by the `auth` and `auth_param` arguments to [oauth_client()].
#' This is used to authenticate the client as part of the OAuth flow, **not**
#' to authenticate a request on behalf of a user.
#'
#' There are three built-in strategies:
#'
#' * `oauth_client_req_body()` adds the client id and (optionally) the secret
#'   to the request body, as described in `r rfc(6749, "2.3.1")`.
#'
#' * `oauth_client_req_header()` adds the client id and secret using HTTP
#'   basic authentication with the `Authorization` header, as described
#'   in `r rfc(6749, "2.3.1")`.
#'
#' * `oauth_client_jwt_rs256()` adds a client assertion to the body using a
#'   JWT signed with `jwt_sign_rs256()` using a private key, as described
#'   in `r rfc(7523, 2.2)`.
#'
#' You will generally not call these functions directly but will instead
#' specify them through the `auth` argument to [oauth_client()]. The `req` and
#' `client` parameters are automatically filled in; other parameters come from
#' the `auth_params` argument.
#' @inheritParams req_perform
#' @param client An [oauth_client].
#' @return A modified HTTP [request].
#' @export
#' @examples
#' # Show what the various forms of client authentication look like
#' req <- request("https://example.com/whoami")
#'
#' client1 <- oauth_client(
#'   id = "12345",
#'   secret = "56789",
#'   token_url = "https://example.com/oauth/access_token",
#'   name = "oauth-example",
#'   auth = "body" # the default
#' )
#' # calls oauth_client_req_auth_body()
#' req_dry_run(oauth_client_req_auth(req, client1))
#'
#' client2 <- oauth_client(
#'   id = "12345",
#'   secret = "56789",
#'   token_url = "https://example.com/oauth/access_token",
#'   name = "oauth-example",
#'   auth = "header"
#' )
#' # calls oauth_client_req_auth_header()
#' req_dry_run(oauth_client_req_auth(req, client2))
#'
#' client3 <- oauth_client(
#'   id = "12345",
#'   key = openssl::rsa_keygen(),
#'   token_url = "https://example.com/oauth/access_token",
#'   name = "oauth-example",
#'   auth = "jwt_sig",
#'   auth_params = list(claim = jwt_claim())
#' )
#' # calls oauth_client_req_auth_header_jwt_sig()
#' req_dry_run(oauth_client_req_auth(req, client3))
oauth_client_req_auth <- function(req, client) {
  exec(client$auth, req = req, client = client, !!!client$auth_params)
}

#' @export
#' @rdname oauth_client_req_auth
oauth_client_req_auth_header <- function(req, client) {
  req_auth_basic(
    req,
    username = client$id,
    password = unobfuscate(client$secret)
  )
}

#' @export
#' @rdname oauth_client_req_auth
oauth_client_req_auth_body <- function(req, client) {
  req_body_form(
    req,
    client_id = client$id,
    client_secret = unobfuscate(client$secret) # might be NULL
  )
}

#' @inheritParams jwt_claim
#' @export
#' @rdname oauth_client_req_auth
oauth_client_req_auth_jwt_sig <- function(
  req,
  client,
  claim,
  size = 256,
  header = list()
) {
  claim <- exec("jwt_claim", !!!claim)
  jwt <- jwt_encode_sig(claim, key = client$key, size = size, header = header)

  # https://datatracker.ietf.org/doc/html/rfc7523#section-2.2
  req_body_form(
    req,
    client_assertion = jwt,
    client_assertion_type = "urn:ietf:params:oauth:client-assertion-type:jwt-bearer"
  )
}

# Helpers -----------------------------------------------------------------

oauth_flow_check <- function(
  flow,
  client,
  is_confidential = FALSE,
  interactive = FALSE,
  error_call = caller_env()
) {
  if (!inherits(client, "httr2_oauth_client")) {
    cli::cli_abort(
      "{.arg client} must be an OAuth client created with {.fn oauth_client}.",
      call = error_call
    )
  }

  if (is_confidential && is.null(client$secret) && is.null(client$key)) {
    cli::cli_abort(
      c(
        "Can't use this {.arg app} with OAuth 2.0 {flow} flow.",
        i = "{.arg app} must have a confidential client (i.e. {.arg client_secret} is required)."
      ),
      call = error_call
    )
  }

  if (interactive && !is_interactive()) {
    cli::cli_abort(
      "OAuth 2.0 {flow} flow requires an interactive session",
      call = error_call
    )
  }
}

oauth_client_get_token <- function(
  client,
  grant_type,
  ...,
  error_call = caller_env()
) {
  req <- request(client$token_url)
  req <- req_body_form(req, grant_type = grant_type, ...)
  req <- oauth_client_req_auth(req, client)
  req <- req_headers(req, Accept = "application/json")

  resp <- oauth_flow_fetch(req, "client$token_url", error_call = error_call)
  exec(oauth_token, !!!resp)
}

test_that("can check app has needed pieces", {
  client <- oauth_client("id", token_url = "http://example.com")
  expect_snapshot(error = TRUE, {
    oauth_flow_check("test", NULL)
    oauth_flow_check("test", client, is_confidential = TRUE)
    oauth_flow_check("test", client, interactive = TRUE)
  })
})

test_that("checks auth types have needed args", {
  expect_snapshot(error = TRUE, {
    oauth_client("abc", "http://x.com", auth = "header")
    oauth_client("abc", "http://x.com", auth = "jwt_sig")
    oauth_client("abc", "http://x.com", auth = 123)
  })
})

test_that("client has useful print method", {
  url <- "http://example.com"

  expect_snapshot({
    oauth_client("x", url)
    oauth_client("x", url, secret = "SECRET")
    oauth_client("x", url, auth = function(...) {
      xxx
    })
  })
})

test_that("metadata fills endpoints, explicit values win", {
  metadata <- structure(
    list(
      token_endpoint = "https://example.com/token",
      authorization_endpoint = "https://example.com/auth",
      device_authorization_endpoint = "https://example.com/device"
    ),
    class = "httr2_oauth_server_metadata"
  )

  client <- oauth_client("id", metadata = metadata)
  expect_equal(client$token_url, "https://example.com/token")
  expect_equal(client$auth_url, "https://example.com/auth")
  expect_equal(client$device_auth_url, "https://example.com/device")

  client <- oauth_client(
    "id",
    token_url = "https://other.com/token",
    metadata = metadata
  )
  expect_equal(client$token_url, "https://other.com/token")
  expect_equal(client$auth_url, "https://example.com/auth")
  expect_equal(client$device_auth_url, "https://example.com/device")
})

test_that("metadata is validated and token_url must be resolvable", {
  metadata <- structure(
    list(authorization_endpoint = "https://example.com/auth"),
    class = "httr2_oauth_server_metadata"
  )
  expect_snapshot(error = TRUE, {
    oauth_client("id")
    oauth_client("id", metadata = metadata)
    oauth_client("id", token_url = "https://x.com", metadata = list())
  })
})

test_that("picks default auth", {
  expect_equal(
    oauth_client("x", "url", key = NULL)$auth,
    "oauth_client_req_auth_body"
  )
  expect_equal(
    oauth_client(
      "x",
      "url",
      key = "key",
      auth_params = list(claim = list())
    )$auth,
    "oauth_client_req_auth_jwt_sig"
  )
})


test_that("can authenticate using header or body", {
  client <- function(auth) {
    oauth_client(
      id = "id",
      secret = "secret",
      token_url = "http://example.com",
      auth = auth
    )
  }

  req <- request("http://example.com")
  req_h <- oauth_client_req_auth(req, client("header"))
  expect_equal(
    headers_flatten(req_h$headers, redact = FALSE),
    list(Authorization = "Basic aWQ6c2VjcmV0")
  )

  req_b <- oauth_client_req_auth(req, client("body"))
  expect_equal(
    req_b$body$data,
    list(client_id = I("id"), client_secret = I("secret"))
  )
})

test_that("can authenticate using a signed jwt", {
  skip_if_not_installed("jose")
  skip_if_not_installed("openssl")

  client <- oauth_client(
    "id",
    "http://example.com",
    key = openssl::rsa_keygen(),
    auth = "jwt_sig",
    auth_params = list(claim = jwt_claim())
  )
  req <- oauth_client_req_auth(request("http://example.com"), client)
  expect_named(req$body$data, c("client_assertion", "client_assertion_type"))
})

test_that("jwt_sig auth requires a claim", {
  skip_if_not_installed("openssl")

  client <- oauth_client(
    "id",
    "http://example.com",
    key = openssl::rsa_keygen(),
    auth = "jwt_sig"
  )
  expect_snapshot(
    oauth_client_req_auth(request("http://example.com"), client),
    error = TRUE
  )
})

#' Parse a URL into its component pieces
#'
#' `url_parse()` parses a URL into its component parts, powered by
#' [curl::curl_parse_url()]. The parsing algorithm follows the specifications
#' detailed in `r rfc(3986)`.
#'
#' @param url A string containing the URL to parse.
#' @param base_url Use this as a parent, if `url` is a relative URL.
#' @returns An S3 object of class `httr2_url` with the following components:
#'   `scheme`, `hostname`, `username`, `password`, `port`, `path`, `query`, and
#'   `fragment`.
#' @export
#' @family URL manipulation
#' @examples
#' url_parse("http://google.com/")
#' url_parse("http://google.com:80/")
#' url_parse("http://google.com:80/?a=1&b=2")
#' url_parse("http://username@google.com:80/path;test?a=1&b=2#40")
#'
#' # You can parse a relative URL if you also provide a base url
#' url_parse("foo", "http://google.com/bar/")
#' url_parse("..", "http://google.com/bar/")
url_parse <- function(url, base_url = NULL) {
  check_string(url)
  check_string(base_url, allow_null = TRUE)

  curl <- curl::curl_parse_url(url, baseurl = base_url)

  parsed <- list(
    scheme = curl$scheme,
    hostname = curl$host,
    username = curl$user,
    password = curl$password,
    port = curl$port,
    path = curl$path,
    query = if (length(curl$params)) as.list(curl$params),
    fragment = curl$fragment
  )
  class(parsed) <- "httr2_url"
  parsed
}

#' Modify a URL
#'
#' @description
#' Use `url_modify()` to modify any component of the URL,
#' `url_modify_relative()` to modify with a relative URL,
#' or `url_modify_query()` to modify individual query parameters.
#'
#' For `url_modify()`, components that aren't specified in the
#' function call will be left as is; components set to `NULL` will be removed,
#' and all other values will be updated. Note that removing `scheme` or
#' `hostname` will create a relative URL.
#'
#' @param url,.url A string or [parsed URL][url_parse()].
#' @param scheme The scheme, typically either `http` or `https`.
#' @param hostname The hostname, e.g., `www.google.com` or `posit.co`.
#' @param username,password Username and password to embed in the URL.
#'   Not generally recommended but needed for some legacy applications.
#' @param port An integer port number.
#' @param path The path, e.g., `/search`. Paths must start with `/`, so this
#'   will be automatically added if omitted.
#' @param query Either a query string or a named list of query components.
#' @param fragment The fragment, e.g., `#section-1`.
#' @return An object of the same type as `url`.
#' @export
#' @family URL manipulation
#' @examples
#' url_modify("http://hadley.nz", path = "about")
#' url_modify("http://hadley.nz", scheme = "https")
#' url_modify("http://hadley.nz/abc", path = "/cde")
#' url_modify("http://hadley.nz/abc", path = "")
#' url_modify("http://hadley.nz?a=1", query = "b=2")
#' url_modify("http://hadley.nz?a=1", query = list(c = 3))
#'
#' url_modify_query("http://hadley.nz?a=1&b=2", c = 3)
#' url_modify_query("http://hadley.nz?a=1&b=2", b = NULL)
#' url_modify_query("http://hadley.nz?a=1&b=2", a = 100)
#'
#' url_modify_relative("http://hadley.nz/a/b/c.html", "/d.html")
#' url_modify_relative("http://hadley.nz/a/b/c.html", "d.html")
#' url_modify_relative("http://hadley.nz/a/b/c.html", "../d.html")
url_modify <- function(
  url,
  scheme = as_is,
  hostname = as_is,
  username = as_is,
  password = as_is,
  port = as_is,
  path = as_is,
  query = as_is,
  fragment = as_is
) {
  if (!is_string(url) && !is_url(url)) {
    stop_input_type(url, "a string or parsed URL")
  }
  string_url <- is_string(url)

  check_url_component(scheme)
  check_url_component(hostname)
  check_url_component(username)
  check_url_component(password)
  if (!leave_as_is(port)) {
    check_number_whole(port, min = 1, max = 65535, allow_null = TRUE)
  }
  check_url_component(path)
  check_url_component(fragment)

  if (is_string(query)) {
    query <- url_query_parse(query)
  } else if (is_named_list(query)) {
    for (nm in names(query)) {
      check_query_param(query[[nm]], paste0("query$", nm))
    }
  } else if (!is.null(query) && !leave_as_is(query)) {
    stop_input_type(query, "a character vector, named list, or NULL")
  }

  new <- list(
    scheme = scheme,
    hostname = hostname,
    username = username,
    password = password,
    port = port,
    path = path,
    query = query,
    fragment = fragment
  )
  new <- new[!map_lgl(new, leave_as_is)]

  if (string_url) {
    new[map_lgl(new, is.null)] <- ""
    url_build(structure(c(url = url, new), class = "httr2_url"))
  } else {
    url[names(new)] <- new
    url
  }
}

as_is <- quote(as_is)
leave_as_is <- function(x) identical(x, as_is)
check_url_component <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (leave_as_is(x)) {
    return(invisible(NULL))
  }

  check_string(x, allow_null = TRUE, arg = arg, call = call)
}

#' @export
#' @rdname url_modify
#' @param relative_url A relative URL to append to the base URL.
url_modify_relative <- function(url, relative_url) {
  string_url <- is_string(url)
  if (!string_url) {
    url <- url_build(url)
  }

  new_url <- url_parse(relative_url, base_url = url)

  if (string_url) {
    url_build(new_url)
  } else {
    new_url
  }
}

#' @export
#' @rdname url_modify
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]>
#'   Name-value pairs that define query parameters. Each value must be either
#'   an atomic vector or `NULL` (which removes the corresponding parameters).
#'   If you want to opt out of escaping, wrap strings in `I()`.
#' @param .multi Controls what happens when a value is a vector:
#'
#'   * `"error"`, the default, throws an error.
#'   * `"comma"`, separates values with a `,`, e.g. `?x=1,2`.
#'   * `"pipe"`, separates values with a `|`, e.g. `?x=1|2`.
#'   * `"explode"`, turns each element into its own parameter, e.g. `?x=1&x=2`
#'
#'   If none of these options work for your needs, you can instead supply a
#'   function that takes a character vector of argument values and returns a
#'   a single string.
#' @param .space How should spaces in query params be escaped? The default,
#'   "percent", uses standard percent encoding (i.e. `%20`), but you can opt-in
#'   to "form" encoding, which uses `+` instead.
url_modify_query <- function(
  .url,
  ...,
  .multi = c("error", "comma", "pipe", "explode"),
  .space = c("percent", "form")
) {
  if (!is_string(.url) && !is_url(.url)) {
    stop_input_type(.url, "a string or parsed URL")
  }
  string_url <- is_string(.url)
  if (string_url) {
    .url <- url_parse(.url)
  }

  new_query <- multi_dots(..., .multi = .multi, .space = .space)
  if (length(new_query) > 0) {
    .url$query <- modify_list(.url$query, !!!new_query)
  }

  if (string_url) {
    url_build(.url)
  } else {
    .url
  }
}

is_url <- function(x) inherits(x, "httr2_url")

#' @export
print.httr2_url <- function(x, ...) {
  cli::cat_line(cli::format_inline("{.cls {class(x)}} {url_build(x)}"))
  if (!is.null(x$scheme)) {
    cli::cat_line(cli::format_inline("* {.field scheme}: {x$scheme}"))
  }
  if (!is.null(x$hostname)) {
    cli::cat_line(cli::format_inline("* {.field hostname}: {x$hostname}"))
  }
  if (!is.null(x$username)) {
    cli::cat_line(cli::format_inline("* {.field username}: {x$username}"))
  }
  if (!is.null(x$password)) {
    cli::cat_line(cli::format_inline("* {.field password}: {x$password}"))
  }
  if (!is.null(x$port)) {
    cli::cat_line(cli::format_inline("* {.field port}: {x$port}"))
  }
  if (!is.null(x$path)) {
    cli::cat_line(cli::format_inline("* {.field path}: {x$path}"))
  }
  if (!is.null(x$query)) {
    cli::cat_line(cli::format_inline("* {.field query}:"))
    for (i in seq_along(x$query)) {
      nm <- names(x$query)[[i]]
      val <- x$query[[i]]
      cli::cat_line(cli::format_inline("  * {.field {nm}}: {val}"))
    }
  }
  if (!is.null(x$fragment)) {
    cli::cat_line(cli::format_inline("* {.field fragment}: {x$fragment}"))
  }
  invisible(x)
}

#' Build a string from a URL object
#'
#' This is the inverse of [url_parse()], taking a parsed URL object and
#' turning it back into a string.
#'
#' @param url An URL object created by [url_parse].
#' @family URL manipulation
#' @export
url_build <- function(url) {
  if (!is_url(url)) {
    stop_input_type(url, "a parsed URL")
  }

  if (is.null(url$query)) {
    query <- NULL
  } else if (length(url$query) == 0 || identical(url$query, "")) {
    query <- ""
  } else {
    query <- I(url_query_build(url$query))
  }

  url <- curl::curl_modify_url(
    url = url$url,
    scheme = url$scheme,
    host = url$hostname,
    user = url$username,
    password = url$password,
    port = url$port,
    path = url$path,
    query = query,
    fragment = url$fragment
  )

  # Workaround https://github.com/curl/curl/issues/17977
  # curl url parser esacapes colons in paths which google seems to use
  # quite frequently. So we hack the problem away for now, restoring the
  # behaviour of httr2 1.1.2
  if (grepl("%3A", url, fixed = TRUE)) {
    path <- curl::curl_parse_url(url, decode = FALSE)$path
    path <- gsub("%3A", ":", path, fixed = TRUE)
    url <- curl::curl_modify_url(url, path = I(path))
  }
  url
}

#' Parse query parameters and/or build a string
#'
#' `url_query_parse()` parses a query string into a named list;
#' `url_query_build()` builds a query string from a named list.
#'
#' @param query A string, when parsing; a named list when building.
#' @export
#' @examples
#' str(url_query_parse("a=1&b=2"))
#'
#' url_query_build(list(x = 1, y = "z"))
#' url_query_build(list(x = 1, y = 1:2), .multi = "explode")
url_query_parse <- function(query) {
  check_string(query)

  query <- gsub("^\\?", "", query) # strip leading ?, if present
  params <- parse_name_equals_value(parse_delim(query, "&"))

  if (length(params) == 0) {
    return(NULL)
  }

  out <- as.list(curl::curl_unescape(params))
  names(out) <- curl::curl_unescape(names(params))
  out
}

#' @export
#' @rdname url_query_parse
#' @inheritParams url_modify_query
url_query_build <- function(
  query,
  .multi = c("error", "comma", "pipe", "explode")
) {
  if (!is_named_list(query)) {
    stop_input_type(query, "a named list")
  }

  query <- multi_dots(!!!query, .multi = .multi, error_arg = "query")
  elements_build(query, "Query", "&")
}

elements_build <- function(x, name, collapse, error_call = caller_env()) {
  if (!is_named_list(x)) {
    cli::cli_abort("{name} must be a named list.", call = error_call)
  }

  x <- compact(x)
  if (length(x) == 0) {
    return(NULL)
  }

  values <- map2_chr(x, names(x), format_query_param, error_call = error_call)
  names <- curl::curl_escape(names(x))

  paste0(names, "=", values, collapse = collapse)
}

format_query_param <- function(
  x,
  name,
  multi = FALSE,
  form = FALSE,
  error_call = caller_env()
) {
  check_query_param(x, name, multi = multi, error_call = error_call)

  if (inherits(x, "AsIs")) {
    unclass(x)
  } else if (is_obfuscated(x)) {
    x
  } else {
    x <- format(x, scientific = FALSE, trim = TRUE, justify = "none")
    x <- curl::curl_escape(x)
    if (form) {
      x <- gsub("%20", "+", x, fixed = TRUE)
    }
    x
  }
}
check_query_param <- function(
  x,
  name,
  multi = FALSE,
  error_call = caller_env()
) {
  if (inherits(x, "AsIs")) {
    if (multi) {
      ok <- is.character(x)
      expected <- "a character vector"
    } else {
      ok <- is.character(x) && length(x) == 1
      expected <- "a single string"
    }
    arg <- paste0("Escaped query value `", name, "`")
    x <- unclass(x)
  } else {
    if (multi) {
      ok <- is.atomic(x)
      expected <- "an atomic vector"
    } else {
      ok <- is.atomic(x) && length(x) == 1
      expected <- "a length-1 atomic vector"
    }
    arg <- paste0("Query value `", name, "`")
  }

  if (ok) {
    invisible()
  } else {
    stop_input_type(x, expected, arg = I(arg), call = error_call)
  }
}

test_that("can parse special cases", {
  url <- url_parse("file:///tmp")
  expect_equal(url$scheme, "file")
  expect_equal(url$path, "/tmp")
})

test_that("can round trip urls", {
  urls <- list(
    "http://google.com/",
    "http://google.com/path",
    "http://google.com/path?a=1&b=2",
    "http://google.com:81/path?a=1&b=2",
    "http://google.com:81/path?a=1&b=2#frag",
    "http://google.com:81/path?a=1&b=2&c=%7B1%7B2%7D3%7D#frag",
    "http://user@google.com:81/path?a=1&b=2",
    "http://user:pass@google.com:81/path?a=1&b=2"
  )

  expect_equal(map(urls, \(url) url_build(url_parse(url))), urls)
})

test_that("can parse relative urls", {
  base <- "http://example.com/a/b/c/"
  expect_equal(url_parse("d", base)$path, "/a/b/c/d")
  expect_equal(url_parse("..", base)$path, "/a/b/")

  expect_equal(url_parse("//archive.org", base)$scheme, "http")
})

test_that("can print all url details", {
  expect_snapshot(
    url_parse("http://user:pass@example.com:81/path?a=1&b=2&c={1{2}3}#frag")
  )
})

test_that("url_build validates its input", {
  expect_snapshot(url_build("abc"), error = TRUE)
})

test_that("decodes params and paths", {
  url <- url_parse("http://example.com/a%20b?q=a%20b")
  expect_equal(url$path, "/a b")
  expect_equal(url$query$q, "a b")
})

# modify url -------------------------------------------------------------

test_that("url_modify checks its inputs", {
  url <- "http://example.com"

  expect_snapshot(error = TRUE, {
    url_modify(1)
    url_modify(url, scheme = 1)
    url_modify(url, hostname = 1)
    url_modify(url, port = "x")
    url_modify(url, port = -1)
    url_modify(url, username = 1)
    url_modify(url, password = 1)
    url_modify(url, path = 1)
    url_modify(url, fragment = 1)
  })
})

test_that("no arguments is idempotent", {
  string <- "http://example.com/"
  url <- url_parse(string)

  expect_equal(url_modify(string), string)
  expect_equal(url_modify(url), url)
})

test_that("can round-trip escaped components", {
  url <- "https://example.com/a%20b"
  expect_equal(url_modify(url), url)

  url <- "https://example.com/?q=a%20b"
  expect_equal(url_modify(url), url)
})

test_that("can accept query as a string or list", {
  url <- "http://test/"

  expect_equal(url_modify(url, query = "a=1&b=2"), "http://test/?a=1&b=2")
  expect_equal(
    url_modify(url, query = list(a = 1, b = 2)),
    "http://test/?a=1&b=2"
  )

  expect_equal(url_modify(url, query = ""), "http://test/")
  expect_equal(url_modify(url, query = list()), "http://test/")
})

test_that("encodes params and paths", {
  expect_equal(
    url_modify("https://example.com", query = list(q = "a b")),
    "https://example.com/?q=a%20b"
  )
  expect_equal(
    url_modify("https://example.com", path = "a b"),
    "https://example.com/a%20b"
  )
})

test_that("colons in paths are left as is", {
  expect_equal(
    url_modify("https://example.com", path = "a:b/foo bar/"),
    "https://example.com/a:b/foo%20bar/"
  )
})

test_that("checks various query formats", {
  url <- "http://example.com"

  expect_snapshot(error = TRUE, {
    url_modify(url, query = 1)
    url_modify(url, query = list(1))
    url_modify(url, query = list(x = 1:2))
  })
})

test_that("path always starts with /", {
  expect_equal(
    url_modify("https://x.com/abc", path = "def"),
    "https://x.com/def"
  )
  expect_equal(url_modify("https://x.com/abc", path = ""), "https://x.com/")
  expect_equal(url_modify("https://x.com/abc", path = NULL), "https://x.com/")
})

test_that("only modifies specified components", {
  url <- "http://x.com/a%2Fb/"
  expect_equal(url_modify(url), url)
  expect_equal(url_modify(url, query = list(x = 1)), "http://x.com/a%2Fb/?x=1")
})

test_that("appending to path does not normalise encoding", {
  req <- request("http://x.com/a%2Fb/")
  expect_equal(
    req_url_path_append(req, "foo/bar")$url,
    "http://x.com/a%2Fb/foo/bar"
  )
  expect_equal(
    req_url_path_append(req, "foo%2Fbar")$url,
    "http://x.com/a%2Fb/foo%2Fbar"
  )
})

# relative url ------------------------------------------------------------

test_that("can set relative urls", {
  base <- "http://example.com/a/b/c/"
  expect_equal(url_modify_relative(base, "d"), "http://example.com/a/b/c/d")
  expect_equal(url_modify_relative(base, ".."), "http://example.com/a/b/")
  expect_equal(
    url_modify_relative(base, "//archive.org"),
    "http://archive.org/"
  )
})

test_that("is idempotent", {
  string <- "http://example.com/"
  url <- url_parse(string)

  expect_equal(url_modify_relative(string, "."), string)
  expect_equal(url_modify_relative(url, "."), url)
})

# modify query -------------------------------------------------------------

test_that("can add, modify, and delete query components", {
  expect_equal(
    url_modify_query("http://test/path", new = "value"),
    "http://test/path?new=value"
  )
  expect_equal(
    url_modify_query("http://test/path", new = "one", new = "two"),
    "http://test/path?new=one&new=two"
  )
  expect_equal(
    url_modify_query("http://test/path?a=old&b=old", a = "new"),
    "http://test/path?b=old&a=new"
  )
  expect_equal(
    url_modify_query("http://test/path?remove=me&keep=this", remove = NULL),
    "http://test/path?keep=this"
  )
})

test_that("can control space formatting", {
  expect_equal(
    url_modify_query("http://test/path", new = "a b"),
    "http://test/path?new=a%20b"
  )
  expect_equal(
    url_modify_query("http://test/path", new = "a b", .space = "form"),
    "http://test/path?new=a+b"
  )
})

test_that("is idempotent", {
  string <- "http://example.com/"
  url <- url_parse(string)

  expect_equal(url_modify_query(string), string)
  expect_equal(url_modify_query(url), url)
})

test_that("validates inputs", {
  url <- "http://example.com/"

  expect_snapshot(error = TRUE, {
    url_modify_query(1)
    url_modify_query(url, 1)
    url_modify_query(url, x = 1:2)
  })
})


# query -------------------------------------------------------------------

test_that("missing query values become empty strings", {
  expect_equal(url_query_parse("?q="), list(q = ""))
  expect_equal(url_query_parse("?q"), list(q = ""))
  expect_equal(url_query_parse("?a&q"), list(a = "", q = ""))
})

test_that("handles equals in values", {
  expect_equal(url_query_parse("?x==&y=="), list(x = "=", y = "="))
})

test_that("empty queries become NULL", {
  expect_equal(url_query_parse("?"), NULL)
  expect_equal(url_query_parse(""), NULL)
})

test_that("validates inputs", {
  expect_snapshot(error = TRUE, {
    url_query_build(1:3)
    url_query_build(list(x = 1:2, y = 1:3))
  })
})

# format_query_param ------------------------------------------------------

test_that("handles all atomic vectors", {
  expect_equal(format_query_param(NA, "x"), "NA")
  expect_equal(format_query_param(TRUE, "x"), "TRUE")
  expect_equal(format_query_param(1L, "x"), "1")
  expect_equal(format_query_param(1.3, "x"), "1.3")
  expect_equal(format_query_param("x", "x"), "x")
  expect_equal(format_query_param(" ", "x"), "%20")
})

test_that("doesn't add extra spaces", {
  expect_equal(
    format_query_param(c(1, 1000), "x", multi = TRUE),
    c("1", "1000")
  )
  expect_equal(
    format_query_param(c("a", "bcdef"), multi = TRUE, "x"),
    c("a", "bcdef")
  )
})

test_that("formats numbers nicely", {
  expect_equal(format_query_param(1e9, "x"), "1000000000")
})

test_that("can opt out of escaping", {
  expect_equal(format_query_param(I(","), "x"), ",")
})

test_that("can't opt out of escaping non strings", {
  expect_snapshot(format_query_param(I(1), "x"), error = TRUE)
})

#' Progress bars in httr2
#'
#' @description
#' Some of httr2's perform functions have a `progress` argument that you can use
#' to create a progress bar. `progress` can be:
#'
#' * `FALSE`, the default: does not create a progress bar.
#' * `TRUE`: creates a basic unnamed progress bar.
#' * A string: creates a basic progress bar with the given name.
#' * A named list of progress bar parameters, as described below.
#'
#' It's good practice to name your progress bars, to make it clear what
#' calculation or process they belong to. We recommend keeping the names
#' under 20 characters, so the whole progress bar fits comfortably even on
#' on narrower displays.
#'
#' ## Progress bar parameters
#'
#' * `clear`: whether to remove the progress bar from the screen after
#'   termination. Defaults to `TRUE`.
#' * `format`: format string. This overrides the default format string of
#'   the progress bar type. It must be given for the `custom` type.
#'   Format strings may contain R expressions to evaluate in braces.
#'   They support cli [pluralization][cli::pluralization], and
#'   [styling][cli::inline-markup] and they can contain special
#'   [progress variables][cli::progress-variables].
#' * `format_done`: format string for successful termination. By default
#'   the same as `format`.
#' * `format_failed`: format string for unsuccessful termination.
#'   By default the same as `format`.
#' * `name`: progress bar name. This is by default the empty string and it
#'   is displayed at the beginning of the progress bar.
#' * `type`: progress bar type. Currently supported types are:
#'   * `iterator`: the default, a for loop or a mapping function,
#'   * `tasks`: a (typically small) number of tasks,
#'   * `download`: download of one file,
#'   * `custom`: custom type, `format` must not be `NULL` for this type.
#'   The default display is different for each progress bar type.
#'
#' ## Further documentation
#'
#' purrr's progress bars are powered by cli, so see
#' [Introduction to progress bars in cli](https://cli.r-lib.org/articles/progress.html)
#' and [Advanced cli progress bars](https://cli.r-lib.org/articles/progress-advanced.html)
#' for more details.
#'
#' @keywords internal
#' @name progress_bars
NULL

#' Automatically retry a request on failure
#'
#' @description
#' `req_retry()` allows [req_perform()] to automatically retry failing
#' requests. It's particularly important for APIs with rate limiting, but can
#' also be useful when dealing with flaky servers.
#'
#' By default, `req_perform()` will retry if the response is a 429
#' ("too many requests", often used for rate limiting) or 503
#' ("service unavailable"). If the API you are wrapping has other transient
#' status codes (or conveys transience with some other property of the
#' response), you can override the default with `is_transient`. And
#' if you set `retry_on_failure = TRUE`, the request will retry
#' if either the HTTP request or HTTP response doesn't complete successfully,
#' leading to an error from curl, the lower-level library that httr2 uses to
#' perform HTTP requests. This occurs, for example, if your Wi-Fi is down.
#'
#' ## Delay
#'
#' It's a bad idea to immediately retry a request, so `req_perform()` will
#' wait a little before trying again:
#'
#' * If the response contains the `Retry-After` header, httr2 will wait the
#'   amount of time it specifies. If the API you are wrapping conveys this
#'   information with a different header (or other property of the response),
#'   you can override the default behavior with `retry_after`.
#'
#' * Otherwise, httr2 will use "truncated exponential backoff with full
#'   jitter", i.e., it will wait a random amount of time between one second and
#'   `2 ^ tries` seconds, capped at a maximum of 60 seconds. In other words, it
#'   waits `runif(1, 1, 2)` seconds after the first failure, `runif(1, 1, 4)`
#'   after the second, `runif(1, 1, 8)` after the third, and so on. If you'd
#'   prefer a different strategy, you can override the default with `backoff`.
#'
#' @inheritParams req_perform
#' @param max_tries,max_seconds Cap the maximum number of attempts
#'   (`max_tries`), the total elapsed time from the first request
#'   (`max_seconds`), or both.
#'
#'   `max_tries` is the total number of attempts made, so this should always
#'   be greater than one.
#' @param is_transient A predicate function that takes a single argument
#'   (the response) and returns `TRUE` or `FALSE` specifying whether or not
#'   the response represents a transient error.
#' @param retry_on_failure Treat low-level failures as if they are
#'   transient errors that can be retried.
#' @param backoff A function that takes a single argument (the number of failed
#'   attempts so far) and returns the number of seconds to wait.
#' @param after A function that takes a single argument (the response) and
#'   returns either a number of seconds to wait or `NA`. `NA` indicates
#'   that a precise wait time is not available and that the `backoff` strategy
#'   should be used instead.
#' @param failure_threshold,failure_timeout,failure_realm
#'   Set `failure_threshold` to activate "circuit breaking" where if a request
#'   continues to fail after `failure_threshold` times, cause the request to
#'   error until a timeout of `failure_timeout` seconds has elapsed. This
#'   timeout will persist across all requests with the same `failure_realm`
#'   (which defaults to the hostname of the request) and is intended to detect
#'   failing servers without needing to wait each time.
#' @returns A modified HTTP [request].
#' @export
#' @seealso [req_throttle()] if the API has a rate-limit but doesn't expose
#'   the limits in the response.
#' @examples
#' # google APIs assume that a 500 is also a transient error
#' request("http://google.com") |>
#'   req_retry(is_transient = \(resp) resp_status(resp) %in% c(429, 500, 503))
#'
#' # use a constant 10s delay after every failure
#' request("http://example.com") |>
#'   req_retry(backoff = \(resp) 10)
#'
#' # When rate-limited, GitHub's API returns a 403 with
#' # `X-RateLimit-Remaining: 0` and an Unix time stored in the
#' # `X-RateLimit-Reset` header. This takes a bit more work to handle:
#' github_is_transient <- function(resp) {
#'   resp_status(resp) == 403 &&
#'     identical(resp_header(resp, "X-RateLimit-Remaining"), "0")
#' }
#' github_after <- function(resp) {
#'   time <- as.numeric(resp_header(resp, "X-RateLimit-Reset"))
#'   time - unclass(Sys.time())
#' }
#' request("http://api.github.com") |>
#'   req_retry(
#'     is_transient = github_is_transient,
#'     after = github_after
#'   )
req_retry <- function(
  req,
  max_tries = NULL,
  max_seconds = NULL,
  retry_on_failure = FALSE,
  is_transient = NULL,
  backoff = NULL,
  after = NULL,
  failure_threshold = Inf,
  failure_timeout = 30,
  failure_realm = NULL
) {
  check_request(req)
  check_number_whole(max_tries, min = 1, allow_null = TRUE)
  check_number_whole(max_seconds, min = 0, allow_null = TRUE)
  check_number_whole(failure_threshold, min = 1, allow_infinite = TRUE)
  check_number_whole(failure_timeout, min = 1)

  if (is.null(max_tries) && is.null(max_seconds)) {
    max_tries <- 2
    cli::cli_inform("Setting {.code max_tries = 2}.")
  }

  check_bool(retry_on_failure)

  req_policies(
    req,
    retry_max_tries = max_tries,
    retry_max_wait = max_seconds,
    retry_on_failure = retry_on_failure,
    retry_is_transient = as_callback(is_transient, 1, "is_transient"),
    retry_backoff = as_callback(backoff, 1, "backoff"),
    retry_after = as_callback(after, 1, "after"),
    retry_failure_threshold = failure_threshold,
    retry_failure_timeout = failure_timeout,
    retry_realm = failure_realm %||% url_parse(req$url)$hostname
  )
}

retry_max_tries <- function(req) {
  has_max_wait <- !is.null(req$policies$retry_max_wait)
  req$policies$retry_max_tries %||% if (has_max_wait) Inf else 1
}

retry_max_seconds <- function(req) {
  req$policies$retry_max_wait %||% Inf
}

retry_check_breaker <- function(req, i, error_call = caller_env()) {
  realm <- req$policies$retry_realm
  if (is.null(realm)) {
    return(invisible())
  }

  now <- unix_time()
  if (env_has(the$breaker, realm)) {
    triggered <- the$breaker[[realm]]
  } else if (i > req$policies$retry_failure_threshold) {
    the$breaker[[realm]] <- triggered <- now
  } else {
    return(invisible())
  }

  remaining <- req$policies$retry_failure_timeout - (now - triggered)
  if (remaining <= 0) {
    env_unbind(the$breaker, realm)
  } else {
    cli::cli_abort(
      c(
        "Request failures have exceeded the threshold for realm {.str {realm}}.",
        i = "The server behind {.str {realm}} is likely still overloaded or down.",
        i = "Wait {remaining} seconds before retrying."
      ),
      call = error_call,
      class = "httr2_breaker"
    )
  }
}

retry_is_transient <- function(req, resp) {
  if (is_error(resp)) {
    return(req$policies$retry_on_failure %||% FALSE)
  }

  req_policy_call(
    req,
    "retry_is_transient",
    list(resp),
    default = function(resp) resp_status(resp) %in% c(429, 503)
  )
}

retry_backoff <- function(req, i) {
  req_policy_call(req, "retry_backoff", list(i), default = backoff_default)
}

retry_after <- function(req, resp, i, error_call = caller_env()) {
  if (is_error(resp)) {
    return(retry_backoff(req, i))
  }

  after <- req_policy_call(
    req,
    "retry_after",
    list(resp),
    default = resp_retry_after
  )

  # TODO: apply this idea to all callbacks
  if (!is_number_or_na(after)) {
    not <- obj_type_friendly(after)
    cli::cli_abort(
      "The {.code after} callback to {.fn req_retry} must return a single number or NA, not {not}.",
      call = error_call
    )
  }

  if (is.na(after)) {
    retry_backoff(req, i)
  } else {
    after
  }
}

is_number_or_na <- function(x) {
  (is.numeric(x) && length(x) == 1) || identical(x, NA)
}

# Helpers -----------------------------------------------------------------

# Exponential backoff with full-jitter, capped to 60s wait
# https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
backoff_default <- function(i) {
  round(min(stats::runif(1, min = 1, max = 2^i), 60), 1)
}

test_that("has useful default (with message)", {
  req <- request_test()
  expect_snapshot(req <- req_retry(req))
  expect_equal(retry_max_tries(req), 2)
  expect_equal(retry_max_seconds(req), Inf)
})

test_that("can set define maximum retries", {
  req <- request_test()
  expect_equal(retry_max_tries(req), 1)
  expect_equal(retry_max_seconds(req), Inf)

  req <- req_retry(req, max_tries = 2)
  expect_equal(retry_max_tries(req), 2)
  expect_equal(retry_max_seconds(req), Inf)

  req <- req_retry(req, max_seconds = 5)
  expect_equal(retry_max_tries(req), Inf)
  expect_equal(retry_max_seconds(req), 5)

  req <- req_retry(req, max_tries = 2, max_seconds = 5)
  expect_equal(retry_max_tries(req), 2)
  expect_equal(retry_max_seconds(req), 5)
})

test_that("can override default is_transient", {
  req <- request_test()
  expect_equal(retry_is_transient(req, response(404)), FALSE)
  expect_equal(retry_is_transient(req, response(429)), TRUE)

  req <- req_retry(
    req,
    max_tries = 2,
    is_transient = \(resp) resp_status(resp) == 404
  )
  expect_equal(retry_is_transient(req, response(404)), TRUE)
  expect_equal(retry_is_transient(req, response(429)), FALSE)
})

test_that("can override default backoff", {
  withr::local_seed(1014)

  req <- request_test()
  expect_equal(retry_backoff(req, 1), 1.1)
  expect_equal(retry_backoff(req, 5), 26.9)
  expect_equal(retry_backoff(req, 10), 60)

  req <- req_retry(req, max_tries = 2, backoff = \(i) 10)
  expect_equal(retry_backoff(req, 1), 10)
  expect_equal(retry_backoff(req, 5), 10)
  expect_equal(retry_backoff(req, 10), 10)
})

test_that("can override default retry wait", {
  resp <- response(429, headers = c("Retry-After: 10", "Wait-For: 20"))
  req <- request_test()
  expect_equal(retry_after(req, resp, 1), 10)

  req <- req_retry(
    req,
    max_tries = 2,
    after = \(resp) as.numeric(resp_header(resp, "Wait-For"))
  )
  expect_equal(retry_after(req, resp, 1), 20)
})

test_that("missing retry-after uses backoff", {
  req <- request_test()
  req <- req_retry(req, max_tries = 2, backoff = \(i) 10)

  expect_equal(retry_after(req, response(429), 1), 10)
})

test_that("useful message if `after` wrong", {
  req <- request_test() |>
    req_retry(
      is_transient = function(resp) TRUE,
      after = function(resp) resp,
      max_tries = 2
    )

  expect_snapshot(req_perform(req), error = TRUE)
})

test_that("validates its inputs", {
  req <- new_request("http://example.com")

  expect_snapshot(error = TRUE, {
    req_retry(req, max_tries = 0)
    req_retry(req, max_tries = 2, max_seconds = "x")
    req_retry(req, max_tries = 2, retry_on_failure = "x")
  })
})

test_that("is_number_or_na implemented correctly", {
  expect_equal(is_number_or_na(1), TRUE)
  expect_equal(is_number_or_na(NA_real_), TRUE)
  expect_equal(is_number_or_na(NA), TRUE)

  expect_equal(is_number_or_na(1:2), FALSE)
  expect_equal(is_number_or_na(numeric()), FALSE)
  expect_equal(is_number_or_na("x"), FALSE)
})


# circuit breaker --------------------------------------------------------

test_that("triggered after specified requests", {
  req <- request_test("/status/:status", status = 429) |>
    req_retry(
      after = function(resp) 0,
      max_tries = 10,
      failure_threshold = 1
    )

  # First attempt performs, retries, then errors
  req_perform(req) |>
    expect_condition(class = "httr2_perform") |>
    expect_condition(class = "httr2_retry") |>
    expect_error(class = "httr2_breaker")

  # Second attempt errors without performing
  req_perform(req) |>
    expect_no_condition(class = "httr2_perform") |>
    expect_error(class = "httr2_breaker")

  # Attempt on same realm errors without trying at all
  req2 <- request_test("/status/:status", status = 200) |>
    req_retry(max_tries = 2)
  req_perform(req) |>
    expect_no_condition(class = "httr2_perform") |>
    expect_error(class = "httr2_breaker")
})

#' Extract HTTP status from response
#'
#' @description
#' * `resp_status()` retrieves the numeric HTTP status code
#' * `resp_status_desc()` retrieves the brief textual description.
#' * `resp_is_error()` returns `TRUE` if the status code represents an error
#'   (i.e. a 4xx or 5xx status).
#' * `resp_check_status()` turns HTTPs errors into R errors.
#'
#' These functions are mostly for internal use because in most cases you
#' will only ever see a 200 response:
#'
#' * 1xx are handled internally by curl.
#' * 3xx redirects are automatically followed. You will only see them if you
#'   have deliberately suppressed redirects with
#'   `req |> req_options(followlocation = FALSE)`.
#' * 4xx client and 5xx server errors are automatically turned into R errors.
#'   You can stop them from being turned into R errors with [req_error()],
#'   e.g. `req |> req_error(is_error = \(resp) FALSE)`.
#'
#' @return
#' * `resp_status()` returns a scalar integer
#' * `resp_status_desc()` returns a string
#' * `resp_is_error()` returns `TRUE` or `FALSE`
#' * `resp_check_status()` invisibly returns the response if it's ok;
#'   otherwise it throws an error with class `httr2_http_{status}`.
#' @inheritParams resp_headers
#' @export
#' @examples
#' # An HTTP status code you're unlikely to see in the wild:
#' resp <- response(418)
#' resp |> resp_is_error()
#' resp |> resp_status()
#' resp |> resp_status_desc()
resp_status <- function(resp) {
  check_response(resp)
  resp$status_code
}

#' @export
#' @rdname resp_status
resp_status_desc <- function(resp) {
  check_response(resp)
  status <- resp_status(resp)
  if (status %in% names(http_statuses)) {
    http_statuses[[as.character(status)]]
  } else {
    NA_character_
  }
}

#' @export
#' @rdname resp_status
resp_is_error <- function(resp) {
  check_response(resp)
  resp_status(resp) >= 400
}

#' @export
#' @param info A character vector of additional information to include in
#'   the error message. Passed to [rlang::abort()].
#' @inheritParams rlang::args_error_context
#' @rdname resp_status
resp_check_status <- function(resp, info = NULL, error_call = caller_env()) {
  check_response(resp)
  if (!resp_is_error(resp)) {
    invisible(resp)
  } else {
    cnd <- resp_failure_cnd(resp$request, resp, error_call = error_call)
    cnd_signal(cnd)
  }
}

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
http_statuses <- c(
  "100" = "Continue",
  "101" = "Switching Protocols",
  "102" = "Processing",
  "103" = "Early Hints",
  "200" = "OK",
  "201" = "Created",
  "202" = "Accepted",
  "203" = "Non-Authoritative Information",
  "204" = "No Content",
  "205" = "Reset Content",
  "206" = "Partial Content",
  "207" = "Multi-Status",
  "208" = "Already Reported",
  "226" = "IM Used",
  "300" = "Multiple Choice",
  "301" = "Moved Permanently",
  "302" = "Found",
  "303" = "See Other",
  "304" = "Not Modified",
  "305" = "Use Proxy",
  "307" = "Temporary Redirect",
  "308" = "Permanent Redirect",
  "400" = "Bad Request",
  "401" = "Unauthorized",
  "402" = "Payment Required",
  "403" = "Forbidden",
  "404" = "Not Found",
  "405" = "Method Not Allowed",
  "406" = "Not Acceptable",
  "407" = "Proxy Authentication Required",
  "408" = "Request Timeout",
  "409" = "Conflict",
  "410" = "Gone",
  "411" = "Length Required",
  "412" = "Precondition Failed",
  "413" = "Payload Too Large",
  "414" = "URI Too Long",
  "415" = "Unsupported Media Type",
  "416" = "Range Not Satisfiable",
  "417" = "Expectation Failed",
  "418" = "I'm a teapot",
  "421" = "Misdirected Request",
  "422" = "Unprocessable Entity",
  "423" = "Locked",
  "424" = "Failed Dependency",
  "425" = "Too Early",
  "426" = "Upgrade Required",
  "428" = "Precondition Required",
  "429" = "Too Many Requests",
  "451" = "Unavailable For Legal Reasons",
  "500" = "Internal Server Error",
  "501" = "Not Implemented",
  "502" = "Bad Gateway",
  "503" = "Service Unavailable",
  "504" = "Gateway Timeout",
  "505" = "HTTP Version Not Supported",
  "506" = "Variant Also Negotiates",
  "507" = "Insufficient Storage",
  "508" = "Loop Detected",
  "510" = "Not Extended",
  "511" = "Network Authentication Required"
)


resp_auth_message <- function(resp) {
  # https://datatracker.ietf.org/doc/html/rfc6750#page-9
  www_auth <- resp_header(resp, "WWW-Authenticate")
  if (is.null(www_auth)) {
    return(NULL)
  }

  www_auth <- parse_www_authenticate(www_auth)
  if (www_auth$scheme != "Bearer") {
    return(NULL)
  }

  if (has_name(www_auth, "error")) {
    msg <- glue("OAuth error: {www_auth$error}")
    if (has_name(www_auth, "error_description")) {
      msg <- paste0(msg, " - ", www_auth$error_description)
    }
  } else {
    msg <- "OAuth error"
  }

  non_error <- www_auth[!grepl("^error|^scheme$", names(www_auth))]
  msg <- c(msg, paste0(names(non_error), ": ", non_error))
  set_names(msg, "*")
}

#' Modify request URL
#'
#' @description
#' * `req_url()` replaces the entire URL.
#' * `req_url_relative()` navigates to a relative URL.
#' * `req_url_query()` modifies individual query components.
#' * `req_url_path()` modifies just the path.
#' * `req_url_path_append()` adds to the path.
#'
#' @seealso
#' * To modify a URL without creating a request, see [url_modify()] and
#'   friends.
#' * To use a template like `GET /user/{user}`, see [req_template()].
#' @inheritParams req_perform
#' @param url A new URL; either an absolute URL for `req_url()` or a
#'   relative URL for `req_url_relative()`.
#' @param ... For `req_url_query()`: <[`dynamic-dots`][rlang::dyn-dots]>
#'   Name-value pairs that define query parameters. Each value must be either
#'   an atomic vector or `NULL` (which removes the corresponding parameters).
#'   If you want to opt out of escaping, wrap strings in `I()`.
#'
#'   For `req_url_path()` and `req_url_path_append()`: A sequence of path
#'   components that will be combined with `/`.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' # Change complete url
#' req <- request("http://example.com")
#' req |> req_url("http://google.com")
#'
#' # Use a relative url
#' req <- request("http://example.com/a/b/c")
#' req |> req_url_relative("..")
#' req |> req_url_relative("/d/e/f")
#'
#' # Change url components
#' req |>
#'   req_url_path_append("a") |>
#'   req_url_path_append("b") |>
#'   req_url_path_append("search.html") |>
#'   req_url_query(q = "the cool ice")
#'
#' # Modify individual query parameters
#' req <- request("http://example.com?a=1&b=2")
#' req |> req_url_query(a = 10)
#' req |> req_url_query(a = NULL)
#' req |> req_url_query(c = 3)
#'
#' # Use .multi to control what happens with vector parameters:
#' req |> req_url_query(id = 100:105, .multi = "comma")
#' req |> req_url_query(id = 100:105, .multi = "explode")
#'
#' # If you have query parameters in a list, use !!!
#' params <- list(a = "1", b = "2")
#' req |>
#'   req_url_query(!!!params, c = "3")
req_url <- function(req, url) {
  check_request(req)
  check_string(url)

  req$url <- url
  req
}

#' @export
#' @rdname req_url
req_url_relative <- function(req, url) {
  check_request(req)
  req_url(req, url_modify_relative(req$url, url))
}

#' @export
#' @rdname req_url
#' @inheritParams url_modify_query
req_url_query <- function(
  .req,
  ...,
  .multi = c("error", "comma", "pipe", "explode"),
  .space = c("percent", "form")
) {
  check_request(.req)
  url <- url_modify_query(.req$url, ..., .multi = .multi, .space = .space)
  req_url(.req, url)
}

#' @export
#' @rdname req_url
req_url_path <- function(req, ...) {
  check_request(req)
  path <- dots_to_path(...)

  req_url(req, url_modify(req$url, path = path))
}

#' @export
#' @rdname req_url
req_url_path_append <- function(req, ...) {
  check_request(req)
  path <- dots_to_path(...)

  # Keep verbatim url-encoding of input path
  input <- curl::curl_parse_url(req$url, decode = FALSE)$path
  path <- paste0(sub("/$", "", input), path)
  req_url(req, curl::curl_modify_url(req$url, path = I(path)))
}

#' Get request URL
#'
#' Retrieve the URL from a request.
#'
#' @inheritParams req_perform
#' @returns A character string.
#' @export
#' @examples
#' request("https://httpbin.org") |>
#'   req_url_path("/get") |>
#'   req_url_query(hello = "world") |>
#'   req_get_url()
req_get_url <- function(req) {
  check_request(req)
  req$url
}


dots_to_path <- function(...) {
  path <- paste(c(...), collapse = "/")
  # Ensure we don't add duplicate /s
  # NB: also keeps "" unchanged.
  sub("^([^/])", "/\\1", path)
}

test_that("can get and set url", {
  req1 <- request("http://example.com/")
  req2 <- req_url(req1, "http://foo.com:10")
  expect_equal(req_get_url(req2), "http://foo.com:10")
})


test_that("automatically adds /", {
  req1 <- request("http://example.com")
  req2 <- request("http://example.com/")

  expect_equal(
    req_url_path(req1, "/index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path(req1, "index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path(req2, "/index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path(req2, "index.html")$url,
    "http://example.com/index.html"
  )

  expect_equal(
    req_url_path_append(req1, "index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path_append(req1, "/index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path_append(req2, "index.html")$url,
    "http://example.com/index.html"
  )
  expect_equal(
    req_url_path_append(req2, "/index.html")$url,
    "http://example.com/index.html"
  )
})

test_that("can append multiple components", {
  req <- request("http://example.com/x")
  expect_equal(req_url_path(req, "a", "b")$url, "http://example.com/a/b")
  expect_equal(
    req_url_path_append(req, "a", "b")$url,
    "http://example.com/x/a/b"
  )
})

test_that("can handle empty path", {
  req <- request("http://example.com/x")
  expect_equal(req_url_path(req)$url, "http://example.com/")
  expect_equal(req_url_path_append(req)$url, "http://example.com/x")
  expect_equal(req_url_path(req, NULL)$url, "http://example.com/")
  expect_equal(req_url_path_append(req, NULL)$url, "http://example.com/x")

  expect_equal(req_url_path(req, "")$url, "http://example.com/")
  expect_equal(req_url_path_append(req, "")$url, "http://example.com/x")
})

test_that("can handle path vector", {
  req <- request("http://example.com/x")
  expect_equal(req_url_path(req, c("a", "b"))$url, "http://example.com/a/b")
  expect_equal(
    req_url_path_append(req, c("a", "b"))$url,
    "http://example.com/x/a/b"
  )
  expect_equal(
    req_url_path_append(req, c("a", "b"), NULL)$url,
    "http://example.com/x/a/b"
  )
})

test_that("can set query params", {
  req <- request("http://example.com/")
  expect_equal(
    req_url_query(req, a = 1, b = 2)$url,
    "http://example.com/?a=1&b=2"
  )
  expect_equal(
    req_url_query(req, a = 1, b = 2, c = NULL)$url,
    "http://example.com/?a=1&b=2"
  )
  expect_equal(
    req_url_query(req, !!!list(a = 1, b = 2))$url,
    "http://example.com/?a=1&b=2"
  )

  expect_equal(
    req_url_query(req, a = 1, a = 2)$url,
    "http://example.com/?a=1&a=2"
  )
  expect_equal(
    req_url_query(req, !!!list(a = 1, a = 2))$url,
    "http://example.com/?a=1&a=2"
  )
})

test_that("can control space handling", {
  req <- request("http://example.com/")
  expect_equal(req_url_query(req, a = " ")$url, "http://example.com/?a=%20")
  expect_equal(
    req_url_query(req, a = " ", .space = "form")$url,
    "http://example.com/?a=+"
  )

  expect_snapshot(
    req_url_query(req, a = " ", .space = "bar"),
    error = TRUE
  )
})

test_that("can handle multi query params", {
  req <- request("http://example.com/")

  req_url_query_multi <- function(multi) {
    req_url_query(req, a = 1:2, .multi = multi)$url
  }

  expect_snapshot(req_url_query_multi("error"), error = TRUE)

  expect_equal(req_url_query_multi("explode"), "http://example.com/?a=1&a=2")
  expect_equal(req_url_query_multi("comma"), "http://example.com/?a=1,2")
  expect_equal(req_url_query_multi("pipe"), "http://example.com/?a=1|2")
  expect_equal(req_url_query_multi(function(x) "X"), "http://example.com/?a=X")
})

test_that("errors are forwarded correctly", {
  req <- request("http://example.com/")
  expect_snapshot(error = TRUE, {
    req |> req_url_query(1)
    req |> req_url_query(a = I(1))
    req |> req_url_query(a = 1:2)
    req |> req_url_query(a = mean)
  })
})

test_that("empty query doesn't affect url", {
  req <- request("http://example.com/")
  expect_equal(req_url_query(req)$url, "http://example.com/")
  expect_equal(req_url_query(req, a = NULL)$url, "http://example.com/")
})

test_that("can modify query params iteratively", {
  req <- request("http://example.com/?a=1&b=2")
  expect_equal(req_url_query(req, c = 3)$url, "http://example.com/?a=1&b=2&c=3")
  expect_equal(req_url_query(req, a = 2)$url, "http://example.com/?b=2&a=2")
  expect_equal(
    req_url_query(req, a = 1, a = 2)$url,
    "http://example.com/?b=2&a=1&a=2"
  )
  expect_equal(req_url_query(req, b = NULL)$url, "http://example.com/?a=1")
})

test_that("can opt-out of query escaping", {
  req <- request("http://example.com/")
  expect_equal(req_url_query(req, a = I(","))$url, "http://example.com/?a=,")
})

test_that("can construct relative urls", {
  req <- request("http://example.com/a/b/c.html")
  expect_equal(req_url_relative(req, ".")$url, "http://example.com/a/b/")
  expect_equal(req_url_relative(req, "..")$url, "http://example.com/a/")
  expect_equal(req_url_relative(req, "/d/e/f")$url, "http://example.com/d/e/f")
})
# explode -----------------------------------------------------------------

test_that("explode handles expected inputs", {
  expect_equal(
    explode(list(a = NULL, b = 1, c = 2:3)),
    list(a = NULL, b = 1, c = 2, c = 3)
  )
})

#' Translate curl syntax to httr2
#'
#' @description
#' The curl command line tool is commonly used to demonstrate HTTP APIs and can
#' easily be generated from
#' [browser developer tools](https://everything.curl.dev/cmdline/copyas.html).
#' `curl_translate()` saves you the pain of manually translating these calls
#' by implementing a partial, but frequently used, subset of curl options.
#' Use `curl_help()` to see the supported options, and `curl_translate()`
#' to translate a curl invocation copy and pasted from elsewhere.
#'
#' Inspired by [curlconverter](https://github.com/hrbrmstr/curlconverter)
#' written by [Bob Rudis](https://rud.is/b/).
#'
#' @param cmd Call to curl. If omitted and the clipr package is installed,
#'   will be retrieved from the clipboard.
#' @param simplify_headers Remove typically unimportant headers included when
#'   copying a curl command from the browser. This includes:
#'
#'   * `sec-fetch-*`
#'   * `sec-ch-ua*`
#'   * `referer`, `pragma`, `connection`
#' @returns A string containing the translated httr2 code. If the input
#'   was copied from the clipboard, the translation will be copied back
#'   to the clipboard.
#' @export
#' @examples
#' curl_translate("curl http://example.com")
#' curl_translate("curl http://example.com -X DELETE")
#' curl_translate("curl http://example.com --header A:1 --header B:2")
#' curl_translate("curl http://example.com --verbose")
curl_translate <- function(cmd, simplify_headers = TRUE) {
  if (missing(cmd)) {
    if (is_interactive() && is_installed("clipr") && clipr::clipr_available()) {
      clip <- TRUE
      cmd <- clipr::read_clip()
      cmd <- paste0(cmd, collapse = "\n")
    } else {
      cli::cli_abort("Must supply {.arg cmd}.")
    }
  } else {
    check_string(cmd)
    clip <- FALSE
  }
  data <- curl_normalize(cmd)

  url_pieces <- httr2::url_parse(data$url)
  query <- url_pieces$query
  url_pieces$query <- NULL
  url <- url_build(url_pieces)

  steps <- glue('request("{url}")')
  steps <- add_curl_step(steps, "req_method", main_args = data$method)

  steps <- add_curl_step(steps, "req_url_query", dots = query)

  # Cookies
  cookies <- data$headers$`Cookie`
  data$headers$`Cookie` <- NULL
  if (!is.null(cookies)) {
    steps <- add_curl_step(
      steps,
      "req_cookies_set",
      dots = cookies_parse(cookies)
    )
  }

  # Content type set with data
  type <- data$headers$`Content-Type`
  if (!identical(data$data, "")) {
    data$headers$`Content-Type` <- NULL
  }

  headers <- curl_simplify_headers(data$headers, simplify_headers)
  steps <- add_curl_step(steps, "req_headers", dots = headers)

  if (!identical(data$data, "")) {
    type <- type %||% "application/x-www-form-urlencoded"
    if (type == "application/json" && idempotent_json(data$data)) {
      json <- jsonlite::parse_json(data$data)
      args <- list(data = I(deparse1(json)))
      steps <- add_curl_step(steps, "req_body_json", dots = args)
    } else {
      body <- data$data
      steps <- add_curl_step(steps, "req_body_raw", main_args = c(body, type))
    }
  }

  steps <- add_curl_step(steps, "req_auth_basic", main_args = unname(data$auth))

  perform_args <- list()
  if (data$verbose) {
    perform_args$verbosity <- 1
  }
  steps <- add_curl_step(
    steps,
    "req_perform",
    main_args = perform_args,
    keep_if_empty = TRUE
  )
  out <- paste0(steps, collapse = paste0(" |>\n  "))

  if (clip) {
    cli::cli_alert_success("Copying to clipboard:")
    clipr::write_clip(out)
  }

  out <- paste0(out, "\n")
  structure(out, class = "httr2_cmd")
}

#' @export
print.httr2_cmd <- function(x, ...) {
  cat(x)
  invisible(x)
}

#' @rdname curl_translate
#' @export
curl_help <- function() {
  cat(curl_opts)
}

curl_translate_eval <- function(cmd, env = caller_env()) {
  code <- curl_translate(cmd)
  eval(parse_expr(code), envir = env)
}

curl_normalize <- function(cmd, error_call = caller_env()) {
  args <- curl_args(cmd, error_call = error_call)

  url <- args[["--url"]] %||%
    args[["<url>"]] %||%
    cli::cli_abort("Must supply url.", call = error_call)

  if (has_name(args, "--header")) {
    headers <- as_headers(args[["--header"]])
  } else {
    headers <- as_headers(list())
  }
  if (has_name(args, "--referer")) {
    headers[["referer"]] <- args[["--referer"]]
  }
  if (has_name(args, "--user-agent")) {
    headers[["user-agent"]] <- args[["--user-agent"]]
  }

  if (has_name(args, "--user")) {
    pieces <- parse_in_half(args[["--user"]], ":")
    auth <- list(
      username = pieces$left,
      password = pieces$right
    )
  } else {
    auth <- NULL
  }

  if (has_name(args, "--request")) {
    method <- args[["--request"]]
  } else if (has_name(args, "--head")) {
    method <- "HEAD"
  } else if (has_name(args, "--get")) {
    method <- "GET"
  } else {
    method <- NULL
  }

  if (has_name(args, "--json")) {
    args <- c(args, list(`--data-raw` = args[["--json"]]))
    headers[["Content-Type"]] <- "application/json"
  }

  # https://curl.se/docs/manpage.html#-d
  # --data-ascii, --data
  #   * if first element is @, treat as path to read from, stripping CRLF
  #   * if multiple, combine with &
  # --data-raw - like data, but don't handle @ specially
  # --data-binary - like data, but don't strip CRLF
  # --data-urlencode - not supported for now
  data <- unlist(c(
    lapply(args[["--data"]], curl_data),
    lapply(args[["--data-ascii"]], curl_data),
    lapply(args[["--data-raw"]], curl_data, raw = TRUE),
    lapply(args[["--data-binary"]], curl_data, binary = TRUE)
  ))
  data <- paste0(data, collapse = "&")

  list(
    method = method,
    url = url,
    headers = headers,
    auth = auth,
    verbose = isTRUE(args[["--verbose"]]),
    data = data
  )
}

curl_simplify_headers <- function(headers, simplify_headers) {
  if (simplify_headers) {
    header_names <- tolower(names(headers))
    to_drop <- startsWith(header_names, "sec-fetch") |
      startsWith(header_names, "sec-ch-ua") |
      header_names %in% c("referer", "pragma", "connection")
    headers <- headers[!to_drop]
  }

  headers
}

curl_data <- function(x, binary = FALSE, raw = FALSE) {
  if (!raw && grepl("^@", x)) {
    path <- sub("^@", "", x)
    if (binary) {
      x <- readBin(path, "character", n = file.size(path))
    } else {
      x <- paste(readLines(path, warn = FALSE))
    }
  } else {
    x
  }
}

# Format described at <http://docopt.org>
curl_opts <- "Usage: curl [<url>] [-H <header> ...] [-d <data> ...] [options] [<url>]
      --basic                  (IGNORED)
      --compressed             (IGNORED)
      --digest                 (IGNORED)
  -d, --data <data>            HTTP POST data
      --data-raw <data>        HTTP POST data, '@' allowed
      --data-ascii <data>      HTTP POST ASCII data
      --data-binary <data>     HTTP POST binary data
      --data-urlencode <data>  HTTP POST data url encoded
      --json <data>            HTTP POST JSON
  -G, --get                    Put the post data in the URL and use GET
  -I, --head                   Show document info only
  -H, --header <header>        Pass custom header(s) to server
  -i, --include                (IGNORED)
  -k, --insecure               (IGNORED)
  -L, --location               (IGNORED)
  -m, --max-time <seconds>     Maximum time allowed for the transfer
  -u, --user <user:password>   Server user and password
  -A, --user-agent <name>      Send User-Agent STRING to server
  -#, --progress-bar           Display transfer progress as a progress bar
  -e, --referer <referer>      Referer URL
  -X, --request <command>      Specify request command to use
      --url <url>              URL to work with
  -v, --verbose                Make the operation more talkative
"

curl_args <- function(cmd, error_call = caller_env()) {
  check_installed("docopt")

  pieces <- parse_in_half(cmd, " ")
  if (pieces$left != "curl") {
    cli::cli_abort(
      "Expecting call to {.str curl} not to {.str {pieces[[1]]}}.",
      call = error_call
    )
  }
  if (grepl("'", cmd, fixed = TRUE)) {
    args <- parse_delim(pieces$right, " ", quote = "'")
  } else {
    args <- parse_delim(pieces$right, " ", quote = '"')
  }

  args <- args[args != "" & args != "\\"]

  parsed <- docopt::docopt(curl_opts, args = args, help = FALSE, strict = TRUE)

  # Drop default options
  parsed <- compact(parsed)
  is_false <- map_lgl(parsed, identical, FALSE)
  parsed <- parsed[!is_false]

  parsed
}


# Helpers -----------------------------------------------------------------

is_syntactic <- function(x) {
  x == "" | x == make.names(x)
}
quote_name <- function(x) {
  ifelse(is_syntactic(x), x, encodeString(x, quote = "`"))
}

add_curl_step <- function(
  steps,
  f,
  main_args = NULL,
  dots = NULL,
  keep_if_empty = FALSE
) {
  args <- c(main_args, dots)

  if (is_empty(args) && !keep_if_empty) {
    return(steps)
  }

  names <- quote_name(names2(args))
  string <- map_lgl(args, function(x) is.character(x) && !inherits(x, "AsIs"))
  values <- unlist(args)
  values <- ifelse(string, encode_string2(values), values)

  args_named <- ifelse(
    names == "",
    paste0(values),
    paste0(names, " = ", values)
  )
  if (is_empty(dots)) {
    args_string <- paste0(args_named, collapse = ", ")
    new_step <- paste0(f, "(", args_string, ")")
  } else {
    args_string <- paste0("    ", args_named, ",\n", collapse = "")
    new_step <- paste0(f, "(\n", args_string, "  )")
  }

  c(steps, new_step)
}

encode_string2 <- function(x) {
  supports_raw_string <- getRversion() >= "4.0.0"

  has_double_quote <- grepl('"', x, fixed = TRUE)
  has_single_quote <- grepl("'", x, fixed = TRUE)
  use_double <- !has_double_quote | has_single_quote
  out <- ifelse(
    use_double,
    encodeString(x, quote = '"'),
    encodeString(x, quote = "'")
  )
  if (supports_raw_string) {
    has_unprintable <- grepl("[^[[:cntrl:]]]", x)
    x_encoded <- encodeString(x)
    has_both_quotes <- has_double_quote & has_single_quote
    use_raw_string <- !has_unprintable & (x != x_encoded | has_both_quotes)
    out[use_raw_string] <- paste0('r"---{', x[use_raw_string], '}---"')
  }

  names(out) <- names(x)
  out
}

cookies_parse <- function(x) {
  pairs <- strsplit(x, "; ?")[[1]]
  cookies <- parse_name_equals_value(pairs)

  if (length(cookies) == 0) {
    return(NULL)
  }

  out <- as.list(curl::curl_unescape(cookies))
  names(out) <- curl::curl_unescape(names(cookies))
  out
}

idempotent_json <- function(old) {
  args <- formals(req_body_json)[c("auto_unbox", "null", "digits")]
  new <- exec(jsonlite::toJSON, jsonlite::parse_json(old), !!!args)
  jsonlite::minify(old) == jsonlite::minify(new)
}

test_that("must be call to curl", {
  expect_snapshot(error = TRUE, curl_translate("echo foo"))
})

test_that("must have cmd argument if non-interactive", {
  expect_snapshot(error = TRUE, curl_translate())
})

test_that("captures key components of call", {
  expect_equal(curl_args("curl http://x.com"), list(`<url>` = "http://x.com"))

  # Quotes are stripped
  expect_equal(curl_args("curl 'http://x.com'"), list(`<url>` = "http://x.com"))
  expect_equal(curl_args('curl "http://x.com"'), list(`<url>` = "http://x.com"))

  # Url can come before or after arguments
  expect_equal(
    curl_args("curl -H 'A: 1' 'http://example.com'"),
    curl_args("curl 'http://example.com' -H 'A: 1'")
  )
  # long name and short name are equivalent
  expect_equal(
    curl_args("curl 'http://example.com' --header 'A: 1'"),
    curl_args("curl 'http://example.com' -H 'A: 1'")
  )

  # can repeat args
  expect_equal(
    curl_args("curl 'http://example.com' -H 'A: 1' -H 'B: 2'")$`--header`,
    c("A: 1", "B: 2")
  )

  # Captures flags
  expect_equal(
    curl_args("curl 'http://example.com' --verbose")$`--verbose`,
    TRUE
  )
})

test_that("can accept multiple data arguments", {
  expect_equal(
    curl_args("curl https://example.com -d x=1 -d y=2")$`--data`,
    c("x=1", "y=2")
  )
})

test_that("can handle line breaks", {
  expect_equal(
    curl_args(
      "curl 'http://example.com' \\\n -H 'A: 1' \\\n -H 'B: 2'"
    )$`--header`,
    c("A: 1", "B: 2")
  )
})

test_that("headers are parsed", {
  expect_equal(
    curl_normalize("curl http://x.com -H 'A: 1'")$headers,
    new_headers(list(A = "1"))
  )
  expect_equal(
    curl_normalize("curl http://x.com -H 'B:'")$headers,
    new_headers(list(B = ""))
  )
})

test_that("user-agent and referer become headers", {
  expect_equal(
    curl_normalize("curl http://x.com -A test")$headers,
    as_headers(list("user-agent" = "test"))
  )

  expect_equal(
    curl_normalize("curl http://x.com -e test")$headers,
    as_headers(list("referer" = "test"))
  )
})

test_that("common headers can be removed", {
  sec_fetch_headers <- "-H 'Sec-Fetch-Dest: empty' -H 'Sec-Fetch-Mode: cors'"
  sec_ch_ua_headers <- "-H 'sec-ch-ua-mobile: ?0'"
  other_headers <- "-H 'Accept: application/vnd.api+json'"
  cmd <- paste(
    "curl http://x.com -A agent -e ref",
    sec_fetch_headers,
    sec_ch_ua_headers,
    other_headers
  )
  headers <- curl_normalize(cmd)$headers
  expect_snapshot({
    print(curl_simplify_headers(headers, simplify_headers = TRUE))
    print(curl_simplify_headers(headers, simplify_headers = FALSE))
  })
})

test_that("extract user name and password", {
  expect_equal(
    curl_normalize("curl http://x.com -u name:pass")$auth,
    list(username = "name", password = "pass")
  )
  expect_equal(
    curl_normalize("curl http://x.com -u name")$auth,
    list(username = "name", password = "")
  )
})

test_that("can override default method", {
  expect_equal(curl_normalize("curl http://x.com")$method, NULL)
  expect_equal(curl_normalize("curl http://x.com --get")$method, "GET")
  expect_equal(curl_normalize("curl http://x.com --head")$method, "HEAD")
  expect_equal(curl_normalize("curl http://x.com -X PUT")$method, "PUT")
})

test_that("prefers explicit url", {
  expect_equal(curl_normalize("curl 'http://x.com'")$url, "http://x.com")
  expect_equal(curl_normalize("curl --url 'http://x.com'")$url, "http://x.com")

  # prefers explicit
  expect_equal(
    curl_normalize("curl 'http://x.com' --url 'http://y.com'")$url,
    "http://y.com"
  )
})

test_that("can translate to httr calls", {
  skip_if(getRversion() < "4.1")

  expect_snapshot({
    curl_translate("curl http://x.com")
    curl_translate("curl http://x.com -X DELETE")
    curl_translate("curl http://x.com -H A:1")
    curl_translate("curl http://x.com -H 'A B:1'")
    curl_translate("curl http://x.com -u u:p")
    curl_translate("curl http://x.com --verbose")
  })
})

test_that("can translate query", {
  skip_if(getRversion() < "4.1")

  expect_snapshot({
    curl_translate("curl http://x.com?string=abcde&b=2")
  })
})

test_that("can translate data", {
  skip_if(getRversion() < "4.1")

  expect_snapshot({
    curl_translate("curl http://example.com --data abcdef")
    curl_translate(
      "curl http://example.com --data abcdef -H Content-Type:text/plain"
    )
  })
})

test_that("can translate ocokies", {
  skip_if(getRversion() < "4.1")

  expect_snapshot({
    curl_translate("curl 'http://test' -H 'Cookie: x=1; y=2;z=3'")
  })
})

test_that("can translate json", {
  skip_if(getRversion() < "4.1")

  expect_snapshot({
    curl_translate(
      r"--{curl http://example.com --data-raw '{"a": 1, "b": "text"}' -H Content-Type:application/json}--"
    )
    curl_translate(
      r"--{curl http://example.com --json '{"a": 1, "b": "text"}'}--"
    )
  })
})

test_that("content type stays in header if no data", {
  skip_if(getRversion() < "4.1")

  expect_snapshot(
    curl_translate("curl http://example.com -H Content-Type:text/plain")
  )
})

test_that("can evaluate simple calls", {
  request_test() # hack to start server

  resp <- curl_translate_eval(glue("curl {the$test_app$url()}/get -H A:1"))
  body <- resp_body_json(resp)
  expect_equal(body$headers$A, "1")

  resp <- curl_translate_eval(glue("curl {the$test_app$url()}/post --data A=1"))
  body <- resp_body_json(resp)
  expect_equal(body$form$A, "1")

  resp <- curl_translate_eval(glue(
    "curl {the$test_app$url()}/delete -X delete"
  ))
  body <- resp_body_json(resp)
  expect_equal(body$method, "delete")

  resp <- curl_translate_eval(glue(
    "curl {the$test_app$url()}//basic-auth/u/p -u u:p"
  ))
  body <- resp_body_json(resp)
  expect_true(body$authenticated)
})

test_that("can read from clipboard", {
  skip_on_cran()
  skip_if_not_installed("clipr")
  skip_if(getRversion() < "4.1")

  # pretend we're interactive and can use the clipboard
  withr::local_envvar(CLIPR_ALLOW = TRUE)
  rlang::local_interactive()
  skip_if_not(clipr::clipr_available())

  # restore the existing clipboard to be nice to the tester
  old_clip <- suppressWarnings(clipr::read_clip())
  if (!is.null(old_clip)) {
    withr::defer(clipr::write_clip(old_clip))
  }

  clipr::write_clip("curl 'http://example.com' \\\n -H 'A: 1' \\\n -H 'B: 2'")
  expect_snapshot({
    curl_translate()
    # also writes to clip
    writeLines(clipr::read_clip())
  })
})

test_that("encode_string2() produces simple strings", {
  # double quotes is standard
  expect_equal(encode_string2("x"), encodeString("x", quote = '"'))
  # use single quotes if double quotes but not single quotes
  expect_equal(encode_string2('x"x'), encodeString('x"x', quote = "'"))

  skip_if_not(getRversion() >= "4.0.0")
  # use raw string if single and double quotes are used
  expect_equal(encode_string2('x"\'x'), 'r"---{x"\'x}---"')

  skip_if(getRversion() < "4.1")
  cmd <- paste0(
    "curl 'http://example.com' \
  -X 'PATCH' \
  -H 'Content-Type: application/json' \
  --data-raw ",
    '{"data":{"x":1,"y":"a","nested":{"z":[1,2,3]}}}',
    "\
  --compressed"
  )
  expect_snapshot(curl_translate(cmd))
})

#' Send data in request body
#'
#' @description
#' * `req_body_file()` sends a local file.
#' * `req_body_raw()` sends a string or raw vector.
#' * `req_body_json()` sends JSON encoded data. Named components of this data
#'   can later be modified with `req_body_json_modify()`.
#' * `req_body_form()` sends form encoded data.
#' * `req_body_multipart()` creates a multi-part body.
#'
#' Adding a body to a request will automatically switch the method to POST.
#'
#' @inheritParams req_perform
#' @param type MIME content type. The default, `""`, will not emit a
#'   `Content-Type` header.  Ignored if you have set a `Content-Type` header
#'   with [req_headers()].
#' @returns A modified HTTP [request].
#' @examples
#' req <- request(example_url()) |>
#'   req_url_path("/post")
#'
#' # Most APIs expect small amounts of data in either form or json encoded:
#' req |>
#'   req_body_form(x = "A simple text string") |>
#'   req_dry_run()
#'
#' req |>
#'   req_body_json(list(x = "A simple text string")) |>
#'   req_dry_run()
#'
#' # For total control over the body, send a string or raw vector
#' req |>
#'   req_body_raw("A simple text string") |>
#'   req_dry_run()
#'
#' # There are two main ways that APIs expect entire files
#' path <- tempfile()
#' writeLines(letters[1:6], path)
#'
#' # You can send a single file as the body:
#' req |>
#'   req_body_file(path) |>
#'   req_dry_run()
#'
#' # You can send multiple files, or a mix of files and data
#' # with multipart encoding
#' req |>
#'   req_body_multipart(a = curl::form_file(path), b = "some data") |>
#'   req_dry_run()
#' @name req_body
#' @aliases NULL
NULL

#' @export
#' @rdname req_body
#' @param body A literal string or raw vector to send as body.
req_body_raw <- function(req, body, type = "") {
  check_request(req)
  check_string(type)

  if (is.raw(body)) {
    req_body(req, data = body, type = "raw", content_type = type)
  } else if (is_string(body)) {
    req_body(req, data = body, type = "string", content_type = type)
  } else {
    cli::cli_abort("{.arg body} must be a raw vector or string.")
  }
}

#' @export
#' @rdname req_body
#' @param path Path to file to upload.
req_body_file <- function(req, path, type = "") {
  check_request(req)
  check_string(path)
  if (!file.exists(path)) {
    cli::cli_abort("Can't find file {.path {path}}.")
  } else if (dir.exists(path)) {
    cli::cli_abort("{.arg path} must be a file, not a directory.")
  }
  check_string(type)

  req_body(req, data = path, type = "file", content_type = type)
}

#' @export
#' @rdname req_body
#' @param data Data to include in body.
#' @param auto_unbox Should length-1 vectors be automatically "unboxed" to
#'   JSON scalars?
#' @param digits How many digits of precision should numbers use in JSON?
#' @param null Should `NULL` be translated to JSON's null (`"null"`)
#'   or an empty list (`"list"`).
req_body_json <- function(
  req,
  data,
  auto_unbox = TRUE,
  digits = 22,
  null = "null",
  type = "application/json",
  ...
) {
  check_request(req)
  check_installed("jsonlite")
  check_string(type)
  check_content_type(type, "application/json", "json")

  params <- list2(
    auto_unbox = auto_unbox,
    digits = digits,
    null = null,
    ...
  )
  req_body(
    req,
    data = data,
    type = "json",
    content_type = type,
    params = params
  )
}

#' @export
#' @rdname req_body
req_body_json_modify <- function(req, ...) {
  check_request(req)
  if (!req_get_body_type(req) %in% c("empty", "json")) {
    cli::cli_abort("Can only be used after {.fn req_body_json}.")
  }

  req$body$data <- utils::modifyList(req$body$data %||% list(), list2(...))
  req
}

#' @export
#' @rdname req_body
#' @param ... <[`dynamic-dots`][rlang::dyn-dots]> Name-data pairs used to send
#'   data in the body.
#'
#'   * For `req_body_form()`, the values must be strings (or things easily
#'     coerced to strings). Vectors are converted to strings using the
#'     value of `.multi`.
#'   * For `req_body_multipart()` the values must be strings or objects
#'     produced by [curl::form_file()]/[curl::form_data()].
#'   * For `req_body_json_modify()`, any simple data made from atomic vectors
#'     and lists.
#'
#'   `req_body_json()` uses this argument differently; it takes additional
#'   arguments passed on to  [jsonlite::toJSON()].
#' @inheritParams req_url_query
req_body_form <- function(
  .req,
  ...,
  .multi = c("error", "comma", "pipe", "explode")
) {
  check_request(.req)

  dots <- multi_dots(..., .multi = .multi)
  data <- modify_list(.req$body$data, !!!dots)
  req_body(.req, data = data, type = "form")
}

#' @export
#' @rdname req_body
req_body_multipart <- function(.req, ...) {
  check_request(.req)

  data <- modify_list(.req$body$data, ...)
  # data must be character, raw, curl::form_file, or curl::form_data
  req_body(.req, data = data, type = "multipart")
}

# General structure -------------------------------------------------------

req_body <- function(
  req,
  data,
  type,
  content_type = NULL,
  params = list(),
  error_call = parent.frame()
) {
  # If this changes, also update docs in req_get_body_type()
  arg_match(type, c("raw", "string", "file", "json", "form", "multipart"))

  if (!is.null(req$body) && req$body$type != type) {
    cli::cli_abort(
      c(
        "Can't change body type from {req$body$type} to {type}.",
        i = "You must use only one type of `req_body_*()` per request."
      ),
      call = error_call
    )
  }

  req$body <- list(
    data = data,
    type = type,
    content_type = content_type,
    params = params
  )
  req
}

req_body_info <- function(req) {
  switch(
    req_get_body_type(req),
    empty = "empty",
    raw = glue("a {length(req$body$data)} byte raw vector"),
    string = "a string",
    file = glue("a path '{req$body$data}'"),
    json = "JSON data",
    form = "form data",
    multipart = "multipart data"
  )
}


#' Get request body
#'
#' @description
#' This pair of functions gives you sufficient information to capture the
#' body of a request, and recreate, if needed. httr2 currently supports
#' seven possible body types:
#'
#' * empty: no body.
#' * raw: created by [req_body_raw()] with a raw vector.
#' * string: created by [req_body_raw()] with a string.
#' * file: created by [req_body_file()].
#' * json: created by [req_body_json()]/[req_body_json_modify()].
#' * form: created by [req_body_form()].
#' * multipart: created by [req_body_multipart()].
#'
#' @inheritParams req_perform
#' @export
#' @examples
#' req <- request(example_url())
#' req |> req_body_raw("abc") |> req_get_body_type()
#' req |> req_body_file(system.file("DESCRIPTION")) |> req_get_body_type()
#' req |> req_body_json(list(x = 1, y = 2)) |> req_get_body_type()
#' req |> req_body_form(x = 1, y = 2) |> req_get_body_type()
#' req |> req_body_multipart(x = "x", y = "y") |> req_get_body_type()
req_get_body_type <- function(req) {
  check_request(req)
  req$body$type %||% "empty"
}

#' @export
#' @rdname req_get_body_type
#' @param obfuscated What to do with obfuscated values that can be present in
#'   JSON, form, and multipart bodies?
#'   * `"remove"` (the default) replaces them with `NULL`.
#'   * `"redact"` replaces them with `<REDACTED>`.
#'   * `"reveal"` leaves them in place.
#' @param obfuscated Form and JSON bodies can contain [obfuscated] values.
#'   This argument control what happens to them: should they be removed,
#'   redacted, or revealed.
req_get_body <- function(req, obfuscated = c("remove", "redact", "reveal")) {
  check_request(req)
  obfuscated <- arg_match(obfuscated)

  data <- req$body$data
  if (req_get_body_type(req) %in% c("json", "form", "multipart")) {
    data <- unobfuscate(data, obfuscated)
  }
  data
}

req_body_apply <- function(req) {
  req <- switch(
    req_get_body_type(req),
    empty = req,
    raw = ,
    string = req_body_apply_data(req, req$body$data),
    file = req_body_apply_stream(req, req$body$data),
    json = ,
    form = req_body_apply_data(req, req_body_render(req)),
    multipart = req_body_apply_multipart(req, req$body$data),
  )

  # Set Content-Type if not already set
  if (!is.null(req$body$content_type) && is.null(req$headers$`Content-Type`)) {
    req <- req_headers(req, `Content-Type` = req$body$content_type)
  }

  req
}

# Needed for JSON and form types since these have special representation
# in httr2 but not curl.
req_body_render <- function(req) {
  type <- req_get_body_type(req)
  switch(
    type,
    form = url_query_build(unobfuscate(req$body$data)),
    json = unclass(exec(
      jsonlite::toJSON,
      unobfuscate(req$body$data),
      !!!req$body$params
    )),
    cli::cli_abort("Unsupported type {type}", .internal = TRUE)
  )
}

req_body_apply_data <- function(req, data) {
  if (is_string(data)) {
    data <- charToRaw(enc2utf8(data))
  }
  req_options(req, post = TRUE, postfieldsize = length(data), postfields = data)
}
req_body_apply_stream <- function(req, data) {
  size <- file.info(data)$size
  # Only open connection if needed
  delayedAssign("con", file(data, "rb"))

  req <- req_policies(req, done = function() close(con))
  req <- req_options(
    req,
    post = TRUE,
    readfunction = function(nbytes, ...) readBin(con, "raw", nbytes),
    seekfunction = function(offset, ...) seek(con, where = offset),
    postfieldsize_large = size
  )
  req
}
req_body_apply_multipart <- function(req, data) {
  req$fields <- unobfuscate(req$body$data)
  req
}

test_that("can't change body type", {
  req <- request("http://example.com") |> req_body_raw(raw(1))
  expect_snapshot(req |> req_body_json(list(x = 1)), error = TRUE)
})

test_that("useful values for empty body", {
  req <- request("http://example.com")
  expect_equal(req_get_body_type(req), "empty")
  expect_equal(req_body_info(req), "empty")
})

# req_body_raw() ---------------------------------------------------------------

test_that("can send string", {
  req <- request_test("/post") |> req_body_raw("test", type = "text/plain")
  expect_equal(req_get_body_type(req), "string")
  expect_equal(req_body_info(req), "a string")

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_equal(json$headers$`Content-Type`, "text/plain")
  expect_equal(json$data, "test")
})

test_that("can send raw vector", {
  data <- charToRaw("abcdef")
  req <- request_test("/post") |> req_body_raw(data)
  expect_equal(req_get_body_type(req), "raw")
  expect_equal(req_body_info(req), "a 6 byte raw vector")

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_equal(json$headers$`Content-Type`, NULL)
  expect_equal(json$headers$`Content-Length`, "6")
})

test_that("can't send anything else", {
  req <- request_test()
  expect_snapshot(req_body_raw(req, 1), error = TRUE)
})

test_that("can override body content type", {
  req <- request_test("/post") |>
    req_body_raw('{"x":"y"}') |>
    req_headers("content-type" = "application/json")
  resp <- req_perform(req)
  headers <- resp_body_json(resp)$headers
  expect_equal(headers$`content-type`, "application/json")
  expect_equal(headers$`Content-Type`, NULL)
})

# req_body_file() --------------------------------------------------------------

test_that("can send file", {
  # curl requests in 64kb chunks so this will hopefully illustrate
  # any subtle problems
  path <- withr::local_tempfile()
  x <- strrep("x", 128 * 1024)
  writeChar(x, path, nchar(x))

  req <- request_test("/post") |> req_body_file(path, type = "text/plain")
  expect_equal(req_get_body_type(req), "file")
  expect_equal(req_body_info(req), glue::glue("a path '{path}'"))

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_equal(json$headers$`Content-Type`, "text/plain")
  expect_equal(json$data, x)
})

test_that("can send file with redirect", {
  str <- paste(letters, collapse = "")
  path <- tempfile()
  writeChar(str, path)

  resp <- request_test("/redirect-to") |>
    req_url_query(url = "/post", status_code = "307") |>
    req_body_file(path, type = "text/plain") |>
    req_perform()

  expect_equal(resp_status(resp), 200)
  expect_equal(url_parse(resp$url)$path, "/post")
  expect_equal(resp_body_json(resp)$data, str)
})

test_that("errors on invalid input", {
  expect_snapshot(error = TRUE, {
    req_body_file(request_test(), 1)
    req_body_file(request_test(), "doesntexist")
    req_body_file(request_test(), ".")
  })
})
# req_body_json() --------------------------------------------------------------

test_that("can send any type of object as json", {
  req <- request_test("/post") |> req_body_json(mtcars)
  expect_equal(req$body$data, mtcars)

  data <- list(a = "1", b = "2")
  req <- request_test("/post") |> req_body_json(data)
  expect_equal(req_get_body_type(req), "json")
  expect_equal(req_body_info(req), "JSON data")

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_equal(json$json, data)

  resp <- request_test("/post") |>
    req_body_json(letters) |>
    req_perform()
  json <- resp_body_json(resp)
  expect_equal(json$json, as.list(letters))
})

test_that("can use custom json type", {
  resp <- request_test("/post") |>
    req_body_json(mtcars, type = "application/ld+json") |>
    req_perform()

  expect_equal(
    resp_body_json(resp)$headers$`Content-Type`,
    "application/ld+json"
  )
})

test_that("non-json type errors", {
  expect_snapshot(
    req_body_json(request_test(), mtcars, type = "application/xml"),
    error = TRUE
  )
})

test_that("can modify json data", {
  req <- request_test() |>
    req_body_json(data = list(a = 1, b = 2, d = 4)) |>
    req_body_json_modify(a = 10, b = NULL, c = 3)
  expect_equal(req$body$data, list(a = 10, d = 4, c = 3))

  req <- request_test() |>
    req_body_json(data = list(a = list(b = list(c = 1, d = 2), e = 3))) |>
    req_body_json_modify(a = list(b = list(c = 101), e = 103))
  expect_equal(req$body$data, list(a = list(b = list(c = 101, d = 2), e = 103)))
})

test_that("can modify empty body", {
  req <- request_test() |>
    req_body_json_modify(a = 10, b = 20)
  expect_equal(req$body$data, list(a = 10, b = 20))
})

test_that("can't modify non-json data", {
  req <- request_test() |> req_body_raw("abc")
  expect_snapshot(req |> req_body_json_modify(a = 1), error = TRUE)
})

test_that("json is unobufcated", {
  req <- request_test() |>
    req_body_json(list(x = "x", y = obfuscated("ZdYJeG8zwISodg0nu4UxBhs"))) |>
    req_body_apply()
  expect_equal(rawToChar(req$options$postfields), '{"x":"x","y":"y"}')
})

# req_body_form() --------------------------------------------------------------

test_that("can send named elements as form", {
  data <- list(a = "1", b = "2")

  req <- request_test("/post") |>
    req_body_form(!!!data) |>
    req_body_apply()

  expect_equal(req_get_body_type(req), "form")
  expect_equal(req_body_info(req), "form data")
  expect_equal(rawToChar(req$options$postfields), "a=1&b=2")

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_equal(json$headers$`Content-Type`, "application/x-www-form-urlencoded")
  expect_equal(json$form, data)
})

test_that("can modify body data", {
  req1 <- request_test() |> req_body_form(a = 1)
  expect_equal(req1$body$data, list(a = I("1")))

  req2 <- req1 |> req_body_form(b = 2)
  expect_equal(req2$body$data, list(a = I("1"), b = I("2")))

  req3 <- req1 |> req_body_form(a = 3, a = 4)
  expect_equal(req3$body$data, list(a = I("3"), a = I("4")))
})

test_that("form data is unobufcated", {
  req <- request_test() |>
    req_body_form(x = "x", y = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")) |>
    req_body_apply()
  expect_equal(rawToChar(req$options$postfields), 'x=x&y=y')
})

# req_body_multipart() ---------------------------------------------------------

test_that("can send named elements as multipart", {
  data <- list(a = "1", b = "2")

  req <- request_test("/post") |>
    req_body_multipart(!!!data) |>
    req_body_apply()
  expect_equal(req_get_body_type(req), "multipart")
  expect_equal(req_body_info(req), "multipart data")
  expect_equal(req$fields, data)

  resp <- req_perform(req)
  json <- resp_body_json(resp)
  expect_match(json$headers$`Content-Type`, "multipart/form-data; boundary=-")
  expect_equal(json$form, list(a = "1", b = "2"))
})

test_that("can upload file with multipart", {
  skip_on_os("windows") # fails due to line ending difference

  path <- tempfile()
  writeLines("this is a test", path)

  resp <- request_test("/post") |>
    req_body_multipart(file = curl::form_file(path)) |>
    req_perform()
  json <- resp_body_json(resp)
  expect_equal(
    json$files$file$value,
    paste0(
      "data:application/octet-stream;base64,",
      openssl::base64_encode("this is a test\n")
    )
  )
})

test_that("no issues with partial name matching", {
  req <- request_test("/get") |>
    req_body_multipart(d = "some data")

  expect_named(req$body$data, "d")
})

test_that("mutlipart data is unobufcated", {
  req <- request_test() |>
    req_body_multipart(x = "x", y = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")) |>
    req_body_apply()

  expect_equal(req$fields, list(x = "x", y = "y"))
})

#' Iteration helpers
#'
#' @description
#' These functions are intended for use with the `next_req` argument to
#' [req_perform_iterative()]. Each implements iteration for a common
#' pagination pattern:
#'
#' * `iterate_with_offset()` increments a query parameter, e.g. `?page=1`,
#'   `?page=2`, or `?offset=1`, `offset=21`.
#' * `iterate_with_cursor()` updates a query parameter with the value of a
#'    cursor found somewhere in the response.
#'  * `iterate_with_link_url()` follows the url found in the `Link` header.
#'    See `resp_link_url()` for more details.
#'
#' @param param_name Name of query parameter.
#' @param start Starting value.
#' @param offset Offset for each page. The default is set to `1` so you get
#'   (e.g.) `?page=1`, `?page=2`, ... If `param_name` refers to an element
#'   index (rather than a page index) you'll want to set this to a larger number
#'   so you get (e.g.) `?items=20`, `?items=40`, ...
#' @param resp_complete A callback function that takes a response (`resp`)
#'   and returns `TRUE` if there are no further pages.
#' @param resp_pages A callback function that takes a response (`resp`) and
#'   returns the total number of pages, or `NULL` if unknown. It will only
#'   be called once.
#' @export
#' @examples
#' req <- request(example_url()) |>
#'   req_url_path("/iris") |>
#'   req_throttle(10) |>
#'   req_url_query(limit = 50)
#'
#' # If you don't know the total number of pages in advance, you can
#' # provide a `resp_complete()` callback
#' is_complete <- function(resp) {
#'   length(resp_body_json(resp)$data) == 0
#' }
#' resps <- req_perform_iterative(
#'   req,
#'   next_req = iterate_with_offset("page_index", resp_complete = is_complete),
#'   max_reqs = Inf
#' )
#'
#' \dontrun{
#' # Alternatively, if the response returns the total number of pages (or you
#' # can easily calculate it), you can use the `resp_pages()` callback which
#' # will generate a better progress bar.
#'
#' resps <- req_perform_iterative(
#'   req |> req_url_query(limit = 1),
#'   next_req = iterate_with_offset(
#'     "page_index",
#'     resp_pages = function(resp) resp_body_json(resp)$pages
#'   ),
#'   max_reqs = Inf
#' )
#' }
iterate_with_offset <- function(
  param_name,
  start = 1,
  offset = 1,
  resp_pages = NULL,
  resp_complete = NULL
) {
  check_string(param_name)
  check_number_whole(start)
  check_number_whole(offset, min = 1)
  check_function2(resp_pages, args = "resp", allow_null = TRUE)
  check_function2(resp_complete, args = "resp", allow_null = TRUE)
  resp_complete <- resp_complete %||% function(resp) FALSE

  known_total <- FALSE
  i <- start # assume already fetched first page

  function(resp, req) {
    if (!is.null(resp_pages) && !known_total) {
      n <- resp_pages(resp)
      if (!is.null(n)) {
        known_total <<- TRUE
        signal_total_pages(n)
      }
    }

    if (!isTRUE(resp_complete(resp))) {
      i <<- i + offset
      req |> req_url_query(!!param_name := i)
    }
  }
}

#' @rdname iterate_with_offset
#' @export
#' @param resp_param_value A callback function that takes a response (`resp`)
#'   and returns the next cursor value. Return `NULL` if there are no further
#'   pages.
iterate_with_cursor <- function(param_name, resp_param_value) {
  check_string(param_name)
  check_function2(resp_param_value, args = "resp")

  function(resp, req) {
    value <- resp_param_value(resp)
    if (!is.null(value)) {
      req |> req_url_query(!!param_name := value)
    }
  }
}

#' @rdname iterate_with_offset
#' @export
#' @param rel The "link relation type" to use to retrieve the next page.
iterate_with_link_url <- function(rel = "next") {
  check_string(rel)

  function(resp, req) {
    url <- resp_link_url(resp, rel)
    if (!is.null(url)) {
      req |> req_url(url)
    }
  }
}


#' Signal total number pages
#'
#' To be called within a `next_req` callback function used with
#' [req_perform_iterative()]
#'
#' @param n Total number of pages.
#' @export
#' @keywords internal
signal_total_pages <- function(n) {
  if (is.null(n)) {
    return()
  }

  check_number_whole(n, min = 1)
  signal("", class = "httr2_total_pages", n = n)
}

# iterate_with_offset -----------------------------------------------------

test_that("iterate_with_offset checks inputs", {
  expect_snapshot(error = TRUE, {
    iterate_with_offset(1)
    iterate_with_offset("x", "x")
    iterate_with_offset("x", offset = 0)
    iterate_with_offset("x", offset = "x")
    iterate_with_offset("x", resp_complete = function(x, y) x + y)
  })
})

test_that("increments param_name by offset", {
  req_1 <- request_test()
  iterator <- iterate_with_offset("page")

  req_2 <- iterator(response(), req_1)
  expect_equal(url_parse(req_2$url)$query, list(page = "2"))

  req_3 <- iterator(response(), req_2)
  expect_equal(url_parse(req_3$url)$query, list(page = "3"))
})

test_that("can compute total pages", {
  req_1 <- request_test()
  iterator <- iterate_with_offset("page", resp_pages = function(resp) {
    resp_body_json(resp)$n
  })

  expect_no_condition(
    req_2 <- iterator(response_json(), req_1),
    class = "httr2_total_pages"
  )

  expect_condition(
    req_3 <- iterator(response_json(body = list(n = 100)), req_2),
    class = "httr2_total_pages"
  )

  # Only called once
  expect_no_condition(
    req_4 <- iterator(response_json(body = list(n = 200)), req_3),
    class = "httr2_total_pages"
  )
})

test_that("can terminate early", {
  req_1 <- request_test()
  iterator <- iterate_with_offset("page", resp_complete = function(resp) {
    resp_body_json(resp)$done
  })

  req_2 <- iterator(response_json(body = list(done = FALSE)), req_1)
  expect_equal(url_parse(req_2$url)$query, list(page = "2"))

  req_3 <- iterator(response_json(body = list(done = TRUE)), req_2)
  expect_null(req_3)
})

# iterate_with_cursor -----------------------------------------------------

test_that("iterate_with_cursor", {
  expect_snapshot(error = TRUE, {
    iterate_with_cursor(1)
    iterate_with_cursor("x", function(x, y) x + y)
  })
})

test_that("updates param_name with new value", {
  req_1 <- request_test()
  iterator <- iterate_with_cursor("next_cursor", function(resp) {
    resp_body_json(resp)$cursor
  })

  req_2 <- iterator(response_json(body = list(cursor = 123)), req_1)
  expect_equal(url_parse(req_2$url)$query, list(next_cursor = "123"))

  req_3 <- iterator(response_json(body = list()), req_2)
  expect_null(req_3)
})

# iterate_with_link_url -----------------------------------------------------

test_that("iterate_with_link_url checks its inputs", {
  expect_snapshot(error = TRUE, {
    iterate_with_link_url(rel = 1)
  })
})

test_that("updates the full url", {
  req_1 <- request_test()
  iterator <- iterate_with_link_url()

  resp <- response(headers = 'Link: <https://example.com/page/2>; rel="next"')
  req_2 <- iterator(resp, req_1)
  expect_equal(req_2$url, "https://example.com/page/2")

  req_3 <- iterator(response(), req_2)
  expect_null(req_3)
})

#' Set request method/path from a template
#'
#' @description
#' Many APIs document their methods with a lightweight template mechanism
#' that looks like `GET /user/{user}` or `POST /organisation/:org`. This
#' function makes it easy to copy and paste such snippets and retrieve template
#' variables either from function arguments or the current environment.
#'
#' `req_template()` will append to the existing path so that you can set a
#' base url in the initial [request()]. This means that you'll generally want
#' to avoid multiple `req_template()` calls on the same request.
#'
#' @inheritParams req_perform
#' @param template A template string which consists of a optional HTTP method
#'   and a path containing variables labelled like either `:foo` or `{foo}`.
#' @param ... Template variables.
#' @param .env Environment in which to look for template variables not found
#'   in `...`. Expert use only.
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' httpbin <- request(example_url())
#'
#' # You can supply template parameters in `...`
#' httpbin |> req_template("GET /bytes/{n}", n = 100)
#'
#' # or you retrieve from the current environment
#' n <- 200
#' httpbin |> req_template("GET /bytes/{n}")
#'
#' # Existing path is preserved:
#' httpbin_test <- request(example_url()) |> req_url_path("/test")
#' name <- "id"
#' value <- "a3fWa"
#' httpbin_test |> req_template("GET /set/{name}/{value}")
req_template <- function(req, template, ..., .env = parent.frame()) {
  check_request(req)
  check_string(template)

  pieces <- strsplit(template, " ", fixed = TRUE)[[1]]
  if (length(pieces) == 1) {
    template <- pieces[[1]]
  } else if (length(pieces) == 2) {
    req <- req_method(req, pieces[[1]])
    template <- pieces[[2]]
  } else {
    cli::cli_abort(c(
      "Can't parse template {.arg template}.",
      i = "Should have form like 'GET /a/b/c' or 'a/b/c/'."
    ))
  }

  dots <- list2(...)
  if (length(dots) > 0 && !is_named(dots)) {
    cli::cli_abort("All elements of {.arg ...} must be named.")
  }

  path <- template_process(template, dots, .env)
  req_url_path_append(req, path)
}

template_process <- function(
  template,
  dots = list(),
  env = parent.frame(),
  error_call = caller_env()
) {
  type <- template_type(template)
  vars <- template_vars(template, type)
  vals <- map_chr(
    vars,
    template_val,
    dots = dots,
    env = env,
    error_call = error_call
  )

  for (i in seq_along(vars)) {
    pattern <- switch(
      type,
      colon = paste0(":", vars[[i]]),
      uri = paste0("{", vars[[i]], "}")
    )
    template <- gsub(pattern, vals[[i]], template, fixed = TRUE)
  }
  template
}

template_val <- function(name, dots, env, error_call = caller_env()) {
  if (has_name(dots, name)) {
    val <- dots[[name]]
  } else if (env_has(env, name, inherit = TRUE)) {
    val <- env_get(env, name, inherit = TRUE)
  } else {
    cli::cli_abort(
      "Can't find template variable {.str {name}}.",
      call = error_call
    )
  }

  if (!is.atomic(val) || length(val) != 1) {
    cli::cli_abort(
      "Template variable {.str {name}} is not a simple scalar value.",
      call = error_call
    )
  }
  as.character(val)
}

template_vars <- function(x, type) {
  if (type == "none") {
    return(character())
  }

  pattern <- switch(
    type,
    colon = ":([a-zA-Z0-9_]+)",
    uri = "\\{(\\w+?)\\}"
  )
  loc <- gregexpr(pattern, x, perl = TRUE)[[1]]
  start <- attr(loc, "capture.start")
  end <- start + attr(loc, "capture.length") - 1
  substring(x, start, end)
}

template_type <- function(x) {
  if (grepl("\\{\\w+?\\}", x)) {
    "uri"
  } else if (grepl(":", x, fixed = TRUE)) {
    "colon"
  } else {
    "none"
  }
}

test_that("can set path", {
  req <- request("http://test.com") |> req_template("/x")
  expect_equal(req$url, "http://test.com/x")
})

test_that("can set method and path", {
  req <- request("http://test.com") |> req_template("PATCH /x")
  expect_equal(req$url, "http://test.com/x")
  expect_equal(req$method, "PATCH")
})

test_that("can use args or env", {
  x <- "x"
  req <- request("http://test.com") |> req_template("/:x")
  expect_equal(req$url, "http://test.com/x")

  req <- request("http://test.com") |> req_template("/:x", x = "y")
  expect_equal(req$url, "http://test.com/y")
})

test_that("will append rather than replace path", {
  req <- request("http://test.com/x") |> req_template("PATCH /y")
  expect_equal(req$url, "http://test.com/x/y")
})

test_that("generates useful errors", {
  req <- request("http://test.com")

  expect_snapshot(error = TRUE, {
    req_template(req, 1)
    req_template(req, "x", 1)
    req_template(req, "A B C")
  })
})

# templating --------------------------------------------------------------

test_that("template_process looks in args & env", {
  a <- 1
  expect_equal(template_process(":a"), "1")
  expect_equal(template_process(":a", list(a = 2)), "2")
})

test_that("template produces useful errors", {
  expect_snapshot(error = TRUE, {
    template_process(":b")
    template_process(":b", list(b = sum))
  })
})

test_that("supports three template styles", {
  x <- "x"
  expect_equal(template_process("/:x/"), "/x/")
  expect_equal(template_process("/{x}/"), "/x/")
  expect_equal(template_process("/constant"), "/constant")
})

test_that("can use colon in uri style", {
  x <- "x"
  expect_equal(template_process("/:{x}:/"), "/:x:/")
})

#' @param max_size The maximum number of bytes to buffer while waiting for a
#'   line or event boundary; if exceeded, an error is thrown. This limit is
#'   approximate: to spot a boundary httr2 may buffer a handful of bytes beyond
#'   `max_size` (e.g. the bytes of the delimiter itself).
#' @export
#' @rdname resp_stream_raw
#' @order 3
resp_stream_sse <- function(resp, max_size = Inf) {
  splitter <- init_streaming_response(resp, SseSplitter)
  check_number_whole(max_size, min = 1, allow_infinite = TRUE)

  repeat {
    blocks <- stream_pull(resp, 1, splitter, max_size)
    if (length(blocks) == 0L) {
      return()
    }
    event_bytes <- blocks[[1L]]

    if (resp_stream_show_buffer(resp)) {
      log_stream(
        cli::rule("Raw server sent event"),
        "\n",
        rawToChar(event_bytes),
        prefix = "*  "
      )
    }

    event <- parse_event(event_bytes)
    if (!is.null(event)) break
  }

  if (resp_stream_show_body(resp)) {
    for (key in names(event)) {
      log_stream(cli::style_bold(key), ": ", pretty_json(event[[key]]))
    }
    cli::cat_line()
  }
  event
}

# Splits a server-sent event stream into events at their boundaries.
SseSplitter <- R6::R6Class(
  "SseSplitter",
  inherit = StreamSplitter,
  public = list(
    name = "resp_stream_sse()",
    find_boundaries = function(buffer) find_event_boundaries(buffer)
  )
)

# Find every event boundary in a buffer, returning a vector of split points
# (the position one past the end of each boundary). Events may be separated by
# a double LF, a double CR, or a double CRLF.
#
# Example:
#   find_event_boundaries(charToRaw("data: 1\n\nid: 12345"))
# Returns:
#   9L  (so the first event is bytes 1:8, "data: 1\n\n")
find_event_boundaries <- function(buffer) {
  nn <- grepRaw("\n\n", buffer, fixed = TRUE, all = TRUE)
  rr <- grepRaw("\r\r", buffer, fixed = TRUE, all = TRUE)
  rnrn <- grepRaw("\r\n\r\n", buffer, fixed = TRUE, all = TRUE)

  # Fast paths for the common case of a single, consistent delimiter.
  if (length(rr) == 0L && length(rnrn) == 0L) {
    return(nn + 2L)
  }
  if (length(nn) == 0L && length(rnrn) == 0L) {
    return(rr + 2L)
  }
  if (length(nn) == 0L && length(rr) == 0L) {
    return(rnrn + 4L)
  }

  # Mixed delimiters: merge candidates and walk them left to right, taking
  # non-overlapping boundaries.
  starts <- c(nn, rr, rnrn)
  ends <- c(nn + 1L, rr + 1L, rnrn + 3L)
  o <- order(starts)
  starts <- starts[o]
  ends <- ends[o]

  keep <- logical(length(starts))
  consumed <- 0L
  for (k in seq_along(starts)) {
    if (starts[k] > consumed) {
      keep[k] <- TRUE
      consumed <- ends[k]
    }
  }
  ends[keep] + 1L
}

# https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation
parse_event <- function(event_data) {
  if (is.raw(event_data)) {
    # Streams must be decoded using the UTF-8 decode algorithm.
    str_data <- rawToChar(event_data)
    Encoding(str_data) <- "UTF-8"
  } else {
    # for testing
    str_data <- event_data
  }

  # The stream must then be parsed by reading everything line by line, with a
  # U+000D CARRIAGE RETURN U+000A LINE FEED (CRLF) character pair, a single
  # U+000A LINE FEED (LF) character not preceded by a U+000D CARRIAGE RETURN
  # (CR) character, and a single U+000D CARRIAGE RETURN (CR) character not
  # followed by a U+000A LINE FEED (LF) character being the ways in
  # which a line can end.
  lines <- strsplit(str_data, "\r\n|\r|\n")[[1]]

  # When a stream is parsed, a data buffer, an event type buffer, and a
  # last event ID buffer must be associated with it. They must be initialized
  # to the empty string.
  data <- ""
  type <- ""
  last_id <- ""

  # If the line starts with a U+003A COLON character (:) - Ignore the line.
  lines <- lines[!grepl("^:", lines)]

  # If the line contains a U+003A COLON character (:)
  # * Collect the characters on the line before the first U+003A COLON
  #  character (:), and let field be that string.
  # * Collect the characters on the line after the first U+003A COLON character
  #  (:), and let value be that string. If value starts with a U+0020 SPACE
  #  character, remove it from value.
  m <- regexec("([^:]*)(: ?)?(.*)", lines)
  matches <- regmatches(lines, m)
  keys <- c("event", vapply(matches, function(x) x[2], character(1)))
  values <- c("message", vapply(matches, function(x) x[4], character(1)))

  for (i in seq_along(matches)) {
    key <- matches[[i]][2]
    value <- matches[[i]][4]

    if (key == "event") {
      # Set the event type buffer to field value.
      type <- value
    } else if (key == "data") {
      # Append the field value to the data buffer, then append a single
      # U+000A LINE FEED (LF) character to the data buffer.
      data <- paste0(data, value, "\n")
    } else if (key == "id") {
      # If the field value does not contain U+0000 NULL, then set the last
      # event ID buffer to the field value. Otherwise, ignore the field.
      last_id <- value
    }
  }

  # If the data buffer is an empty string, set the data buffer and the event
  # type buffer to the empty string and return.
  if (data == "") {
    return()
  }

  # If the data buffer's last character is a U+000A LINE FEED (LF) character,
  # then remove the last character from the data buffer.
  if (grepl("\n$", data)) {
    data <- substr(data, 1, nchar(data) - 1)
  }
  if (type == "") {
    type <- "message"
  }

  list(
    type = type,
    data = data,
    id = last_id
  )
}

test_that("can determine if incomplete data is complete", {
  req <- local_app_request(function(req, res) {
    res$send_chunk("data: 1\n\n")
    res$send_chunk("data: ")
  })

  con <- req |> req_perform_connection(blocking = TRUE)
  withr::defer(close(con))

  expect_equal(
    resp_stream_sse(con, 10),
    list(type = "message", data = "1", id = "")
  )
  expect_snapshot(expect_equal(resp_stream_sse(con), NULL))
  expect_true(resp_stream_is_complete(con))
})

test_that("can feed sse events one at a time", {
  req <- local_app_request(function(req, res) {
    for (i in 1:3) {
      res$send_chunk(sprintf("data: %s\n\n", i))
    }
  })
  resp <- req_perform_connection(req)
  withr::defer(close(resp))

  expect_equal(
    resp_stream_sse(resp),
    list(type = "message", data = "1", id = "")
  )
  expect_equal(
    resp_stream_sse(resp),
    list(type = "message", data = "2", id = "")
  )
  resp_stream_sse(resp)

  expect_equal(resp_stream_sse(resp), NULL)
})

test_that("max_size counts the buffered event bytes, delimiter included", {
  # The event with its "\r\n\r\n" delimiter is 11 bytes; max_size limits the
  # buffered bytes (delimiter included), approximately.
  data <- charToRaw("data: 1\r\n\r\n")
  resp1 <- local_streaming_response(data)

  expect_equal(
    resp_stream_sse(resp1, max_size = 10),
    list(type = "message", data = "1", id = "")
  )

  resp2 <- local_streaming_response(data)
  expect_error(
    resp_stream_sse(resp2, max_size = 9),
    class = "httr2_streaming_error"
  )
})

test_that("ignores events with no data", {
  req <- local_app_request(function(req, res) {
    res$send_chunk(": comment\n\n")
    res$send_chunk("data: 1\n\n")
  })
  resp <- req_perform_connection(req)
  withr::defer(close(resp))

  expect_equal(
    resp_stream_sse(resp),
    list(type = "message", data = "1", id = "")
  )
})

test_that("sse always interprets data as UTF-8", {
  req <- local_app_request(function(req, res) {
    res$send_chunk("data: \xE3\x81\x82\r\n\r\n")
  })

  # Data is decoded as UTF-8 regardless of the locale.
  withr::local_locale(LC_CTYPE = "C")
  resp <- req_perform_connection(req, blocking = TRUE)
  withr::defer(close(resp))

  out <- resp_stream_sse(resp)

  s <- "\xE3\x81\x82"
  Encoding(s) <- "UTF-8"
  expect_equal(out, list(type = "message", data = s, id = ""))
  expect_equal(Encoding(out$data), "UTF-8")
})

test_that("verbosity = 3 shows raw sse events", {
  req <- local_app_request(function(req, res) {
    res$send_chunk(": comment\n\n")
    res$send_chunk("data: 1\n\n")
  })

  expect_output(resp <- req_perform_connection(req, verbosity = 3))
  withr::defer(close(resp))
  expect_snapshot(
    . <- resp_stream_sse(resp),
    transform = transform_verbose_response
  )
})

test_that("has a working find_event_boundaries", {
  # Splits the buffer at the first boundary into the matched event and the
  # remaining bytes.
  boundary_test <- function(x, matched, remaining) {
    buffer <- charToRaw(x)
    splits <- find_event_boundaries(buffer)
    result <- if (length(splits) == 0) {
      NULL
    } else {
      split_at <- splits[[1]]
      list(
        matched = slice(buffer, end = split_at),
        remaining = slice(buffer, start = split_at)
      )
    }
    expect_identical(
      result,
      list(matched = charToRaw(matched), remaining = charToRaw(remaining))
    )
  }

  # Basic matches
  boundary_test("\r\r", matched = "\r\r", remaining = "")
  boundary_test("\n\n", matched = "\n\n", remaining = "")
  boundary_test("\r\n\r\n", matched = "\r\n\r\n", remaining = "")
  boundary_test("a\r\r", matched = "a\r\r", remaining = "")
  boundary_test("a\n\n", matched = "a\n\n", remaining = "")
  boundary_test("a\r\n\r\n", matched = "a\r\n\r\n", remaining = "")
  boundary_test("\r\ra", matched = "\r\r", remaining = "a")
  boundary_test("\n\na", matched = "\n\n", remaining = "a")
  boundary_test("\r\n\r\na", matched = "\r\n\r\n", remaining = "a")

  # Matches the first boundary found
  boundary_test("\r\r\r", matched = "\r\r", remaining = "\r")
  boundary_test("\r\r\r\r", matched = "\r\r", remaining = "\r\r")
  boundary_test("\n\n\r\r", matched = "\n\n", remaining = "\r\r")
  boundary_test("\r\r\n\n", matched = "\r\r", remaining = "\n\n")

  # Finds every boundary in the buffer
  expect_equal(find_event_boundaries(charToRaw("a\n\nb\n\n")), c(4L, 7L))
  expect_equal(
    find_event_boundaries(charToRaw("a\r\n\r\nb\r\n\r\n")),
    c(6L, 11L)
  )

  # Non-matches
  expect_length(find_event_boundaries(charToRaw("\n\r\n\r")), 0)
  expect_length(find_event_boundaries(charToRaw("hello\ngoodbye\n")), 0)
  expect_length(find_event_boundaries(charToRaw("")), 0)
  expect_length(find_event_boundaries(charToRaw("1")), 0)
  expect_length(find_event_boundaries(charToRaw("12")), 0)
  expect_length(find_event_boundaries(charToRaw("\r\n\r")), 0)
})

# parse_event ----------------------------------------------------------------

test_that("event with no data returns NULL", {
  expect_null(parse_event(""))
  expect_null(parse_event(":comment"))
  expect_null(parse_event("id: 1"))

  expect_equal(parse_event("data: ")$data, "")
  expect_equal(parse_event("data")$data, "")
})

test_that("examples from spec work", {
  event <- parse_event("data: YHOO\ndata: +2\ndata: 10")
  expect_equal(event$type, "message")
  expect_equal(event$data, "YHOO\n+2\n10")
})

test_that("event and id fields are parsed", {
  event <- parse_event("event: ping\ndata: x\nid: 42")
  expect_equal(event$type, "ping")
  expect_equal(event$data, "x")
  expect_equal(event$id, "42")
})

#' @export
#' @rdname resp_stream_raw
#' @order 4
resp_stream_aws <- function(resp, max_size = Inf) {
  splitter <- init_streaming_response(resp, AwsSplitter)
  check_number_whole(max_size, min = 1, allow_infinite = TRUE)

  blocks <- stream_pull(resp, 1, splitter, max_size)
  if (length(blocks) == 0L) {
    return()
  }
  event_bytes <- blocks[[1L]]

  event <- parse_aws_event(event_bytes)
  if (resp_stream_show_body(resp)) {
    # Emit header
    for (key in names(event$headers)) {
      log_stream(cli::style_bold(key), ": ", event$headers[[key]])
    }
    # Emit body
    log_stream(jsonlite::toJSON(event$body, auto_unbox = TRUE, pretty = TRUE))
    cli::cat_line()
  }
  event
}

AwsSplitter <- R6::R6Class(
  "AwsSplitter",
  inherit = StreamSplitter,
  public = list(
    name = "resp_stream_aws()",
    find_boundaries = function(buffer) find_aws_event_boundaries(buffer)
  )
)

# Find every complete AWS event in a buffer by walking the 4-byte big-endian
# length prefix at the start of each event. Returns a vector of split points
# (the position one past the end of each complete event).
find_aws_event_boundaries <- function(buffer) {
  n <- length(buffer)
  splits <- double()
  pos <- 1
  repeat {
    # No valid AWS event message is less than 16 bytes.
    if (n - pos + 1L < 16L) {
      break
    }
    # Read the first 4 bytes of the event as a big endian number.
    event_size <- parse_int(buffer[pos:(pos + 3L)])
    if (event_size > n - pos + 1L) {
      break
    }
    pos <- pos + event_size
    splits[[length(splits) + 1L]] <- pos
  }
  splits
}

# Parse a single AWS event-stream message (content type
# application/vnd.amazon.eventstream). The binary format is documented by AWS:
# * https://smithy.io/2.0/aws/amazon-eventstream.html (canonical protocol spec)
# * https://docs.aws.amazon.com/lexv2/latest/dg/event-stream-encoding.html
# Reference implementation: https://github.com/awslabs/aws-eventstream-java
#
# Key details: all integers are big-endian; the prelude (total + header lengths)
# and the whole message each end in a GZIP/zlib CRC32; header value types
# byte/short/integer/long are signed; timestamp is an int64 of epoch millis.
#
# We treat header_length as a lower bound rather than an exact count; this is
# lenient but harmless and matches some reference implementations.
parse_aws_event <- function(bytes) {
  i <- 1
  read_bytes <- function(n) {
    if (n == 0) {
      return(raw())
    }
    out <- bytes[i:(i + n - 1)]
    i <<- i + n
    out
  }

  # prelude
  total_length <- parse_int(read_bytes(4))
  if (total_length != length(bytes)) {
    cli::cli_abort(
      "AWS event metadata doesn't match supplied bytes",
      .internal = TRUE
    )
  }

  header_length <- parse_int(read_bytes(4))
  prelude_crc <- read_bytes(4)
  # TODO: use this value to check prelude lengths

  # headers
  headers <- list()
  while (i <= 12 + header_length) {
    name_length <- as.integer(read_bytes(1))
    name <- rawToChar(read_bytes(name_length))
    type <- as.integer(read_bytes(1))

    delayedAssign("length", parse_int(read_bytes(2)))
    value <- switch(
      type_enum(type),
      "TRUE" = TRUE,
      "FALSE" = FALSE,
      BYTE = parse_int(read_bytes(1), signed = TRUE),
      SHORT = parse_int(read_bytes(2), signed = TRUE),
      INTEGER = parse_int(read_bytes(4), signed = TRUE),
      LONG = parse_int64(read_bytes(8)),
      BYTE_ARRAY = read_bytes(length),
      CHARACTER = rawToChar(read_bytes(length)),
      TIMESTAMP = parse_int64(read_bytes(8)),
      UUID = raw_to_hex(read_bytes(16)),
    )
    headers[[name]] <- value
  }

  # body
  body_raw <- read_bytes(total_length - i - 4 + 1)
  crc_raw <- read_bytes(4)
  # TODO: use this value to check data

  body <- rawToChar(body_raw)
  if (identical(headers$`:content-type`, "application/json")) {
    body <- jsonlite::parse_json(body)
  }

  list(headers = headers, body = body)
}


# Helpers ----------------------------------------------------------------

parse_int <- function(x, signed = FALSE) {
  v <- sum(as.integer(x) * 256^rev(seq_along(x) - 1))
  if (signed && v >= 2^(8 * length(x) - 1)) {
    # Interpret as two's complement.
    v <- v - 2^(8 * length(x))
  }
  v
}

parse_int64 <- function(x) {
  y <- readBin(x, "double", n = 1, size = length(x), endian = "big")
  class(y) <- "integer64"
  y
}

type_enum <- function(value) {
  if (value < 0 || value > 9) {
    cli::cli_abort("Unsupported type {value}.", .internal = TRUE)
  }

  switch(
    value + 1,
    "TRUE",
    "FALSE",
    "BYTE",
    "SHORT",
    "INTEGER",
    "LONG",
    "BYTE_ARRAY",
    "CHARACTER",
    "TIMESTAMP",
    "UUID"
  )
}

raw_to_hex <- function(x) {
  paste(as.character(x), collapse = "")
}

#' OAuth with device flow
#'
#' @description
#' Authenticate using the OAuth **device flow**, as defined by `r rfc(8628)`.
#' It's designed for devices that don't have access to a web browser (if you've
#' ever authenticated an app on your TV, this is probably the flow you've used),
#' but it also works well from within R.
#'
#' Learn more about the overall OAuth authentication flow in
#' <https://httr2.r-lib.org/articles/oauth.html>.
#'
#' @export
#' @param open_browser If `TRUE` (the default in interactive sessions), the
#'   device verification URL will be opened in the user's browser. If `FALSE`,
#'   the URL is printed to the console and the user must open it themselves.
#' @inheritParams oauth_flow_password
#' @inheritParams req_oauth_auth_code
#' @returns `req_oauth_device()` returns a modified HTTP [request] that will
#'   use OAuth; `oauth_flow_device()` returns an [oauth_token].
#' @examples
#' req_auth_github <- function(req) {
#'   req_oauth_device(
#'     req,
#'     client = example_github_client(),
#'     auth_url = "https://github.com/login/device/code"
#'   )
#' }
#'
#' request("https://api.github.com/user") |>
#'   req_auth_github()
req_oauth_device <- function(
  req,
  client,
  auth_url,
  scope = NULL,
  open_browser = is_interactive(),
  auth_params = list(),
  token_params = list(),
  cache_disk = FALSE,
  cache_key = NULL
) {
  params <- list(
    client = client,
    auth_url = auth_url,
    scope = scope,
    open_browser = open_browser,
    auth_params = auth_params,
    token_params = token_params
  )
  cache <- cache_choose(client, cache_disk, cache_key)
  req_oauth(req, "oauth_flow_device", params, cache = cache)
}

#' @export
#' @rdname req_oauth_device
oauth_flow_device <- function(
  client,
  auth_url,
  pkce = FALSE,
  scope = NULL,
  open_browser = is_interactive(),
  auth_params = list(),
  token_params = list()
) {
  oauth_flow_check("device", client, interactive = open_browser)

  if (pkce) {
    code <- oauth_flow_auth_code_pkce()
    auth_params$code_challenge <- code$challenge
    auth_params$code_challenge_method <- code$method
    token_params$code_verifier <- code$verifier
  }

  request <- oauth_flow_device_request(client, auth_url, scope, auth_params)

  # User interaction
  # https://datatracker.ietf.org/doc/html/rfc8628#section-3.3
  # Azure provides a message that we might want to print?
  # Google uses verification_url instead of verification_uri
  # verification_uri_complete is optional, it would ship the user
  # code in the uri https://datatracker.ietf.org/doc/html/rfc8628#section-3.2
  url <- request$verification_uri_complete %||%
    request$verification_uri %||%
    request$verification_url

  if (open_browser) {
    cli::cli_alert(
      "Copy {.strong {request$user_code}} and paste when requested by the browser"
    )
    readline("Press <enter> to proceed:")
    utils::browseURL(url)
  } else {
    inform(glue("Visit <{url}> and enter code {request$user_code}"))
  }

  token <- oauth_flow_device_poll(client, request, token_params)
  if (is.null(token)) {
    cli::cli_abort("Expired without user confirmation; please try again.")
  }

  exec(oauth_token, !!!token)
}

# Device authorization request and response
# https://datatracker.ietf.org/doc/html/rfc8628#section-3.1
# https://datatracker.ietf.org/doc/html/rfc8628#section-3.2
oauth_flow_device_request <- function(
  client,
  auth_url,
  scope,
  auth_params,
  error_call = caller_env()
) {
  req <- request(auth_url)
  req <- req_body_form(req, scope = scope, !!!auth_params)
  req <- oauth_client_req_auth(req, client)
  req <- req_headers(req, Accept = "application/json")

  oauth_flow_fetch(req, "auth_url", error_call = error_call)
}

# Device Access Token Request
# https://datatracker.ietf.org/doc/html/rfc8628#section-3.4
oauth_flow_device_poll <- function(
  client,
  request,
  token_params,
  error_call = caller_env()
) {
  cli::cli_progress_step("Waiting for response from server", spinner = TRUE)

  delay <- request$interval %||% 5
  deadline <- Sys.time() + request$expires_in

  token <- NULL
  while (Sys.time() < deadline) {
    for (i in 1:20) {
      cli::cli_progress_update()
      Sys.sleep(delay / 20)
    }

    tryCatch(
      {
        token <- oauth_client_get_token(
          client,
          grant_type = "urn:ietf:params:oauth:grant-type:device_code",
          device_code = request$device_code,
          !!!token_params,
          error_call = error_call
        )
        break
      },
      httr2_oauth_authorization_pending = function(err) {},
      httr2_oauth_slow_down = function(err) {
        delay <<- delay + 5
      }
    )
  }
  cli::cli_progress_done()
  token
}

test_that("auth_url falls back to the client's device_auth_url, explicit wins", {
  metadata <- structure(
    list(
      token_endpoint = "https://example.com/token",
      device_authorization_endpoint = "https://example.com/device"
    ),
    class = "httr2_oauth_server_metadata"
  )
  client <- oauth_client("id", metadata = metadata)

  req <- req_oauth_device(request("https://example.com"), client)
  expect_equal(
    req$policies$auth_sign$params$flow_params$auth_url,
    "https://example.com/device"
  )

  req <- req_oauth_device(
    request("https://example.com"),
    client,
    auth_url = "https://other.com/device"
  )
  expect_equal(
    req$policies$auth_sign$params$flow_params$auth_url,
    "https://other.com/device"
  )
})

test_that("req_oauth_device() passes pkce through to the flow", {
  client <- oauth_client("id", token_url = "https://example.com/token")

  req <- req_oauth_device(
    request("https://example.com"),
    client,
    auth_url = "https://example.com/device",
    pkce = TRUE
  )
  expect_equal(req$policies$auth_sign$params$flow_params$pkce, TRUE)
})

test_that("oauth_flow_device() resolves auth_url from the client", {
  client <- oauth_client("id", token_url = "https://example.com/token")
  expect_error(
    oauth_flow_device(client, open_browser = FALSE),
    "Must supply"
  )
})

# Standalone file: do not edit by hand
# Source: <https://github.com/r-lib/rlang/blob/main/R/standalone-types-check.R>
# ----------------------------------------------------------------------
#
# ---
# repo: r-lib/rlang
# file: standalone-types-check.R
# last-updated: 2023-03-13
# license: https://unlicense.org
# dependencies: standalone-obj-type.R
# imports: rlang (>= 1.1.0)
# ---
#
# ## Changelog
#
# 2023-03-13:
# - Improved error messages of number checkers (@teunbrand)
# - Added `allow_infinite` argument to `check_number_whole()` (@mgirlich).
# - Added `check_data_frame()` (@mgirlich).
#
# 2023-03-07:
# - Added dependency on rlang (>= 1.1.0).
#
# 2023-02-15:
# - Added `check_logical()`.
#
# - `check_bool()`, `check_number_whole()`, and
#   `check_number_decimal()` are now implemented in C.
#
# - For efficiency, `check_number_whole()` and
#   `check_number_decimal()` now take a `NULL` default for `min` and
#   `max`. This makes it possible to bypass unnecessary type-checking
#   and comparisons in the default case of no bounds checks.
#
# 2022-10-07:
# - `check_number_whole()` and `_decimal()` no longer treat
#   non-numeric types such as factors or dates as numbers.  Numeric
#   types are detected with `is.numeric()`.
#
# 2022-10-04:
# - Added `check_name()` that forbids the empty string.
#   `check_string()` allows the empty string by default.
#
# 2022-09-28:
# - Removed `what` arguments.
# - Added `allow_na` and `allow_null` arguments.
# - Added `allow_decimal` and `allow_infinite` arguments.
# - Improved errors with absent arguments.
#
#
# 2022-09-16:
# - Unprefixed usage of rlang functions with `rlang::` to
#   avoid onLoad issues when called from rlang (#1482).
#
# 2022-08-11:
# - Added changelog.
#
# nocov start

# Scalars -----------------------------------------------------------------

.standalone_types_check_dot_call <- .Call

check_bool <- function(x,
                       ...,
                       allow_na = FALSE,
                       allow_null = FALSE,
                       arg = caller_arg(x),
                       call = caller_env()) {
  if (!missing(x) && .standalone_types_check_dot_call(ffi_standalone_is_bool_1.0.7, x, allow_na, allow_null)) {
    return(invisible(NULL))
  }

  stop_input_type(
    x,
    c("`TRUE`", "`FALSE`"),
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_string <- function(x,
                         ...,
                         allow_empty = TRUE,
                         allow_na = FALSE,
                         allow_null = FALSE,
                         arg = caller_arg(x),
                         call = caller_env()) {
  if (!missing(x)) {
    is_string <- .rlang_check_is_string(
      x,
      allow_empty = allow_empty,
      allow_na = allow_na,
      allow_null = allow_null
    )
    if (is_string) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a single string",
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

.rlang_check_is_string <- function(x,
                                   allow_empty,
                                   allow_na,
                                   allow_null) {
  if (is_string(x)) {
    if (allow_empty || !is_string(x, "")) {
      return(TRUE)
    }
  }

  if (allow_null && is_null(x)) {
    return(TRUE)
  }

  if (allow_na && (identical(x, NA) || identical(x, na_chr))) {
    return(TRUE)
  }

  FALSE
}

check_name <- function(x,
                       ...,
                       allow_null = FALSE,
                       arg = caller_arg(x),
                       call = caller_env()) {
  if (!missing(x)) {
    is_string <- .rlang_check_is_string(
      x,
      allow_empty = FALSE,
      allow_na = FALSE,
      allow_null = allow_null
    )
    if (is_string) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a valid name",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

IS_NUMBER_true <- 0
IS_NUMBER_false <- 1
IS_NUMBER_oob <- 2

check_number_decimal <- function(x,
                                 ...,
                                 min = NULL,
                                 max = NULL,
                                 allow_infinite = TRUE,
                                 allow_na = FALSE,
                                 allow_null = FALSE,
                                 arg = caller_arg(x),
                                 call = caller_env()) {
  if (missing(x)) {
    exit_code <- IS_NUMBER_false
  } else if (0 == (exit_code <- .standalone_types_check_dot_call(
    ffi_standalone_check_number_1.0.7,
    x,
    allow_decimal = TRUE,
    min,
    max,
    allow_infinite,
    allow_na,
    allow_null
  ))) {
    return(invisible(NULL))
  }

  .stop_not_number(
    x,
    ...,
    exit_code = exit_code,
    allow_decimal = TRUE,
    min = min,
    max = max,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_number_whole <- function(x,
                               ...,
                               min = NULL,
                               max = NULL,
                               allow_infinite = FALSE,
                               allow_na = FALSE,
                               allow_null = FALSE,
                               arg = caller_arg(x),
                               call = caller_env()) {
  if (missing(x)) {
    exit_code <- IS_NUMBER_false
  } else if (0 == (exit_code <- .standalone_types_check_dot_call(
    ffi_standalone_check_number_1.0.7,
    x,
    allow_decimal = FALSE,
    min,
    max,
    allow_infinite,
    allow_na,
    allow_null
  ))) {
    return(invisible(NULL))
  }

  .stop_not_number(
    x,
    ...,
    exit_code = exit_code,
    allow_decimal = FALSE,
    min = min,
    max = max,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

.stop_not_number <- function(x,
                             ...,
                             exit_code,
                             allow_decimal,
                             min,
                             max,
                             allow_na,
                             allow_null,
                             arg,
                             call) {
  if (allow_decimal) {
    what <- "a number"
  } else {
    what <- "a whole number"
  }

  if (exit_code == IS_NUMBER_oob) {
    min <- min %||% -Inf
    max <- max %||% Inf

    if (min > -Inf && max < Inf) {
      what <- sprintf("%s between %s and %s", what, min, max)
    } else if (x < min) {
      what <- sprintf("%s larger than or equal to %s", what, min)
    } else if (x > max) {
      what <- sprintf("%s smaller than or equal to %s", what, max)
    } else {
      abort("Unexpected state in OOB check", .internal = TRUE)
    }
  }

  stop_input_type(
    x,
    what,
    ...,
    allow_na = allow_na,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_symbol <- function(x,
                         ...,
                         allow_null = FALSE,
                         arg = caller_arg(x),
                         call = caller_env()) {
  if (!missing(x)) {
    if (is_symbol(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a symbol",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_arg <- function(x,
                      ...,
                      allow_null = FALSE,
                      arg = caller_arg(x),
                      call = caller_env()) {
  if (!missing(x)) {
    if (is_symbol(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an argument name",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_call <- function(x,
                       ...,
                       allow_null = FALSE,
                       arg = caller_arg(x),
                       call = caller_env()) {
  if (!missing(x)) {
    if (is_call(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a defused call",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_environment <- function(x,
                              ...,
                              allow_null = FALSE,
                              arg = caller_arg(x),
                              call = caller_env()) {
  if (!missing(x)) {
    if (is_environment(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an environment",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_function <- function(x,
                           ...,
                           allow_null = FALSE,
                           arg = caller_arg(x),
                           call = caller_env()) {
  if (!missing(x)) {
    if (is_function(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a function",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_closure <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is_closure(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "an R function",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_formula <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is_formula(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a formula",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}


# Vectors -----------------------------------------------------------------

check_character <- function(x,
                            ...,
                            allow_null = FALSE,
                            arg = caller_arg(x),
                            call = caller_env()) {
  if (!missing(x)) {
    if (is_character(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a character vector",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_logical <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is_logical(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a logical vector",
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_data_frame <- function(x,
                             ...,
                             allow_null = FALSE,
                             arg = caller_arg(x),
                             call = caller_env()) {
  if (!missing(x)) {
    if (is.data.frame(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    "a data frame",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

# nocov end

#' Set HTTP method in request
#'
#' Use this function to use a custom HTTP method like `HEAD`,
#' `DELETE`, `PATCH`, `UPDATE`, or `OPTIONS`. The default method is
#' `GET` for requests without a body, and `POST` for requests with a body.
#'
#' @inheritParams req_perform
#' @param method Custom HTTP method
#' @returns A modified HTTP [request].
#' @export
#' @examples
#' request(example_url()) |> req_method("PATCH")
#' request(example_url()) |> req_method("PUT")
#' request(example_url()) |> req_method("HEAD")
req_method <- function(req, method) {
  check_request(req)
  check_string(method)

  req$method <- toupper(method)
  req
}

# Used in req_handle
req_method_apply <- function(req) {
  if (is.null(req$method)) {
    return(req)
  }

  switch(
    req$method,
    HEAD = req_options(req, nobody = TRUE),
    req_options(req, customrequest = req$method)
  )
}

#' Get request method
#'
#' Defaults to `GET`, unless the request has a body, in which case it uses
#' `POST`. Either way the method can be overridden with [req_method()].
#'
#' @inheritParams req_perform
#' @export
#' @examples
#' req <- request(example_url())
#' req_get_method(req)
#' req_get_method(req |> req_body_raw("abc"))
#' req_get_method(req |> req_method("DELETE"))
#' req_get_method(req |> req_method("HEAD"))
req_get_method <- function(req) {
  # https://everything.curl.dev/libcurl-http/requests#request-method
  if (!is.null(req$method)) {
    req$method
  } else if (has_name(req$options, "nobody")) {
    "HEAD"
  } else if (!is.null(req$body)) {
    "POST"
  } else {
    "GET"
  }
}

# shim so httptest2 doesn't fail
req_method_get <- req_get_method

test_that("can override default method", {
  req <- request("http://example.com")

  expect_equal(req$method, NULL)

  req <- req_method(req, "patch")
  expect_equal(req$method, "PATCH")
})

test_that("correctly guesses default method", {
  req <- request("http://example.com")
  expect_equal(req_get_method(req), "GET")

  req <- req_body_raw(req, "abc")
  expect_equal(req_get_method(req), "POST")

  req <- req_options(req, nobody = TRUE)
  expect_equal(req_get_method(req), "HEAD")
})

sync_req <- function(name, .env = parent.frame()) {
  skip_on_cran()
  skip_if_not_installed("nanonext")

  if (missing(name) || !is.character(name)) {
    cli::cli_abort(
      "Use unique (character) name for each sync_req() / sync_rep() pair"
    )
  }
  connected <- FALSE
  cv <- nanonext::cv()
  sock <- nanonext::socket("req")
  withr::defer(close(sock), envir = .env)
  nanonext::pipe_notify(sock, cv, add = TRUE)
  nanonext::listen(sock, url = sprintf("ipc:///tmp/nanonext%s", name))

  function(resp, timeout = 1000L) {
    if (!connected) {
      nanonext::until(cv, timeout)
      connected <<- TRUE
    }
    nanonext::send(sock, 0L, mode = 2L, block = timeout)
    wait_for_http_data(resp, timeout / 1000)
  }
}

sync_rep <- function(name, .env = parent.frame()) {
  if (missing(name) || !is.character(name)) {
    cli::cli_abort(
      "Use unique (character) name for each sync_req() / sync_rep() pair"
    )
  }

  connected <- FALSE
  cv <- nanonext::cv()
  sock <- nanonext::socket("rep")
  withr::defer(close(sock), envir = .env)
  nanonext::pipe_notify(sock, cv, add = TRUE)
  nanonext::dial(sock, url = sprintf("ipc:///tmp/nanonext%s", name))

  function(expr = {}, timeout = 1000L) {
    if (!connected) {
      nanonext::until(cv, timeout)
      connected <<- TRUE
    }
    nanonext::recv(sock, mode = 8L, block = timeout)
    expr
  }
}

# Blocks (by polling the connection) until the next chunk of body data has
# arrived, or the stream is complete, so that a subsequent non-blocking
# resp_stream_lines()/resp_stream_sse() read can see it. Used both for the
# initial chunk a server sends before any sync() signal and for the chunks a
# sync() signal releases.
wait_for_http_data <- function(resp, timeout_s = 5) {
  if (resp$body$is_complete()) {
    return(invisible(TRUE))
  }

  deadline <- as.double(Sys.time()) + timeout_s
  poll_interval <- 0.01

  while (as.double(Sys.time()) < deadline) {
    chunk <- resp$body$read(256)
    if (length(chunk) > 0) {
      resp$cache$stream_buffer <- c(resp$cache$stream_buffer, chunk)
      return(invisible(TRUE))
    }

    if (resp$body$is_complete()) {
      return(invisible(TRUE))
    }

    remaining <- deadline - as.double(Sys.time())
    if (remaining > 0) {
      Sys.sleep(poll_interval)
    }
  }

  invisible(FALSE)
}

# Blocks (by polling the connection) until the stream is complete, i.e. the
# server has closed the connection. A sync() signal only guarantees the last
# chunk's *data* has arrived; the EOF that marks completion can lag behind it.
# Use this before a final read that depends on completion, such as flushing a
# trailing line that has no terminator.
wait_for_complete <- function(resp, timeout_s = 5) {
  deadline <- as.double(Sys.time()) + timeout_s

  repeat {
    chunk <- resp$body$read(256)
    if (length(chunk) > 0) {
      resp$cache$stream_buffer <- c(resp$cache$stream_buffer, chunk)
    }

    if (resp$body$is_complete()) {
      return(invisible(TRUE))
    }

    if (as.double(Sys.time()) >= deadline) {
      return(invisible(FALSE))
    }

    Sys.sleep(0.01)
  }
}

---
title: "httr2"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{httr2}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
# needs pipe & avoids error = TRUE problem in 4.3.0
run_code <- getRversion() >= "4.4.0"

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = run_code,
  purl = run_code
)
```

The goal of this document is show you the basics of httr2.
You'll learn how to create and submit HTTP requests and work with the HTTP responses that you get back.
httr2 is designed to map closely to the underlying HTTP protocol, which I'll explain as we go along.
For more details, I also recommend "[An overview of HTTP](https://developer.mozilla.org/en-US/docs/Web/HTTP/Overview)" from MDN.

```{r setup}
#| eval: true
library(httr2)
```

## Create a request

In httr2, you start by creating a request.
If you're familiar with httr, this a big change: with httr you could only submit a request, immediately receiving a response.
Having an explicit request object makes it easier to build up a complex request piece by piece and works well with the pipe.

Every request starts with a URL:

```{r}
req <- request(example_url())
req
```

Here, instead of an external website, we use a test server that's built-in to httr2 itself. That ensures that this vignette will work regardless of when or where you run it.

We can see exactly what this request will send to the server with a dry run:

```{r}
req |> req_dry_run()
```

```{r}
#| include: FALSE
#| eval: true
port <- if (run_code) paste0("`", url_parse(example_url())$port, "`") else "e.g. `1234`"
```

The first line of the request contains three important pieces of information:

-   The HTTP **method**, which is a verb that tells the server what you want to do.
    Here it's GET, the most common verb, indicating that we want to *get* a resource.
    Other verbs include POST, to create a new resource, PUT, to replace an existing resource, and DELETE, to delete a resource.

-   The **path**, which is the URL stripped of details that the server already knows, i.e. the protocol (`http` or `https`), the host (`localhost`), and the port (`r port`).

-   The version of the HTTP protocol.
    This is unimportant for our purposes because it's handled at a lower level.

The following lines specify the HTTP **headers**, a series of name-value pairs separated by `:`.
The headers in this request were automatically added by httr2, but you can override them or add your own with `req_headers()`:

```{r}
req |>
  req_headers(
    Name = "Hadley",
    `Shoe-Size` = "11",
    Accept = "application/json"
  ) |>
  req_dry_run()
```

Header names are case-insensitive, and servers will ignore headers that they don't understand.

The headers finish with a blank line which is followed by the **body**.
The requests above (like all GET requests) don't have a body, so let's add one to see what happens.
The `req_body_*()` functions provide a variety of ways to add data to the body.
Here we'll use `req_body_json()` to add some data encoded as JSON:

```{r}
req |>
  req_body_json(list(x = 1, y = "a")) |>
  req_dry_run()
```

What's changed?

-   The method has changed from GET to POST.
    POST is the standard method for sending data to a website, and is automatically used whenever you add a body.
    Use `req_method()` to use a different method.

-   There are two new headers: `Content-Type` and `Content-Length`.
    They tell the server how to interpret the body --- it's encoded as JSON and is 15 bytes long.

-   We have a body, consisting of some JSON.

Different servers want data encoded differently so httr2 provides a selection of common formats.
For example, `req_body_form()` uses the encoding used when you submit a form from a web browser:

```{r}
req |>
  req_body_form(x = "1", y = "a") |>
  req_dry_run()
```

And `req_body_multipart()` uses the multipart encoding which is particularly important when you need to send larger amounts of data or complete files:

```{r}
req |>
  req_body_multipart(x = "1", y = "a") |>
  req_dry_run()
```

If you need to send data encoded in a different form, you can use `req_body_raw()` to add the data to the body and set the `Content-Type` header.

## Perform a request and fetch the response

To actually perform a request and fetch the response back from the server, call `req_perform()`:

```{r}
req <- request(example_url()) |> req_url_path("/json")
resp <- req |> req_perform()
resp
```

You can see a simulation of what httr2 actually received with `resp_raw()`:

```{r}
resp |> resp_raw()
```

An HTTP response has a very similar structure to an HTTP request.
The first line gives the version of HTTP used, and a status code that's optionally followed by a short description.
Then we have the headers, followed by a blank line, followed by a body.
The majority of responses will have a body, unlike requests.

You can extract data from the response using the `resp_()` functions:

-   `resp_status()` returns the status code and `resp_status_desc()` returns the description:

    ```{r}
    resp |> resp_status()
    resp |> resp_status_desc()
    ```

-   You can extract all headers with `resp_headers()` or a specific header with `resp_header()`:

    ```{r}
    resp |> resp_headers()
    resp |> resp_header("Content-Length")
    ```

    Headers are case insensitive:

    ```{r}
    resp |> resp_header("ConTEnT-LeNgTH")
    ```

-   You can extract the body in various forms using the `resp_body_*()` family of functions.
    Since this response returns JSON we can use `resp_body_json()`:

    ```{r}
    resp |> resp_body_json() |> str()
    ```

Responses with status codes 4xx and 5xx are HTTP errors.
httr2 automatically turns these into R errors:

```{r, error = TRUE}
request(example_url()) |>
  req_url_path("/status/404") |>
  req_perform()

request(example_url()) |>
  req_url_path("/status/500") |>
  req_perform()
```

This is another important difference to httr, which required that you explicitly call `httr::stop_for_status()` to turn HTTP errors into R errors.
You can revert to the httr behaviour with `req_error(req, is_error = \(resp) FALSE)`.

## Control the request process

A number of `req_` functions don't directly affect the HTTP request but instead control the overall process of submitting a request and handling the response.
These include:

-   `req_cache()` sets up a cache so if repeated requests return the same results, you can avoid a trip to the server.

-   `req_throttle()` will automatically add a small delay before each request so you can avoid hammering a server with many requests.

-   `req_retry()` sets up a retry strategy so that if the request either fails or you get a transient HTTP error, it'll automatically retry after a short delay.

For more details see their documentation, as well as examples of the usage in real APIs in the [Wrapping APIs](https://httr2.r-lib.org/articles/wrapping-apis.html) article.

# happign

<details>

* Version: 0.2.0
* GitHub: https://github.com/paul-carteron/happign
* Source code: https://github.com/cran/happign
* Date/Publication: 2023-08-07 19:10:02 UTC
* Number of recursive dependencies: 120

Run `revdepcheck::cloud_details(, "happign")` for more info

</details>

## Newly broken

*   checking tests ... ERROR
    ```
      Running ‘testthat.R’
    Running the tests in ‘tests/testthat.R’ failed.
    Complete output:
      > library(testthat)
      > library(happign)
      Please make sure you have an internet connection.
      Use happign::get_last_news() to display latest geoservice news.
      > 
      > test_check("happign")
      Linking to GEOS 3.8.0, GDAL 3.0.4, PROJ 6.3.1; sf_use_s2() is TRUE
    ...
        6. │     └─rlang::eval_bare(quo_get_expr(.quo), quo_get_env(.quo))
        7. └─happign:::build_req(path, "test")
        8.   └─httr2::req_url_query(...)
        9.     └─httr2:::query_build(dots)
       10.       └─cli::cli_abort("Query must be a named list.", call = error_call)
       11.         └─rlang::abort(...)
      
      [ FAIL 1 | WARN 21 | SKIP 24 | PASS 64 ]
      Error: Test failures
      Execution halted
    ```

## In both

*   checking data for non-ASCII characters ... NOTE
    ```
      Note: found 7592 marked UTF-8 strings
    ```

## R CMD check results

0 errors | 0 warnings | 0 notes

## revdepcheck results

We checked 57 reverse dependencies, comparing R CMD check results across CRAN and dev versions of this package.

 * We saw 1 new problems with happign: this looks like user error,
   but I've supplied a patch at https://github.com/paul-carteron/happign/pull/22.
 * We failed to check 0 packages

## revdepcheck results

We checked 57 reverse dependencies, comparing R CMD check results across CRAN and dev versions of this package.

 * We saw 1 new problems
 * We failed to check 0 packages

Issues with CRAN packages are summarised below.

### New problems
(This reports the first line of each new failure)

* happign
  checking tests ... ERROR

---
title: "OAuth"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(httr2)
```

OAuth[^1] is an authorization framework design for performing work on the behalf of a user.
You've probably used it a bunch without knowing what it's called: you use it when you login to a non-Google website using your Google account, when you give your phone access to your twitter account, or when you login to a streaming app on your smart TV.

[^1]: Here I'll only talk about OAuth 2.0 which is the only version in common use today.
    OAuth 1.0 is largely only of historical interest.

You'll notice none of these common scenarios quite match to using an R package to scrape data from a web API, and that's the pleasure and pain of OAuth.
OAuth gives you the incredible power to extract data from popular web services, but it's fundamentally designed for a different use case.
This means that while httr2 can make the interactive experience pretty seamless, unattended usage (i.e. on CI or in an automated script) is usually going to include some pain.

This vignette builds on the techniques discussed in `vignette("wrapping-apis")`, so I recommend you read that vignette first.
I'll also assume that you're working in a package, since most people want to wrap up an API into a bunch of R functions, but most of the ideas will also apply if you're creating a one-off script.

## OAuth basics

OAuth is a broad framework that has many different variants, called **flows**, which makes it hard to provide sweeping generalisations, but the basic idea of OAuth is to create a hierarchy of increasingly more specific and shorter-lived credentials, so that the impact of a credential being lost is as small as possible.

The longest lived and most powerful credential is typically a user name and password.
Most people don't change their passwords regularly, and often (against all advice) reuse the same password on multiple websites.
And if you have a user name and password, you have total control over that account; you can even use it to change the password so the actual user can't log in.
That means as a programmer you never want to touch user name-password pairs because if they're lost or stolen, they give very wide access.

Avoiding that problem lead to the creation of OAuth.
The basic idea is that instead of your package asking the user to give you their user name and password, you instead ask them to log in and give your package permission to use the API on their behalf.
The API gives this permission in the form of an **access token** which is essentially a random string of numbers and letters, e.g.
`UfNlXaEog03hdRPTUPpEInEiIW01jI1WcjOB`.
An access token is very short lived, lasting maybe only days, and is bound to a specified **scope** of access.
The access token has some big advantages over a user name and password:

-   It's short-lived so if it's lost or stolen, there's only a limited amount of time where it can be abused.
-   It has limited scope, so even if stolen, it can't be used to do something particularly nefarious like changing your password or contact details.
-   It's bound to a specific application, so it can be invalidated (cancelled) without affecting any other uses.

If you have an access token you can use it to authenticate an API by passing it as a **bearer token** in the `Authorization` header, which you can do with `req_auth_bearer_token()`.
However, in most cases you will want to let httr2 manage this by calling one of the `req_oauth_` functions we'll talk about shortly.

One of the reasons that you want httr2 to manage your tokens is that because access tokens are so short lived, they're often accompanied by a **refresh token**.
A refresh token lasts for a longer amount of time and has one job: it allows you to get a new access token when the previous one expires.
You need to look after a refresh token a little more carefully than an access token.
In particular, you should never include a refresh token in a HTTP request as that's the job of the access token.

Overall this leads to a hierarchy of credentials from weakest to strongest:

-   The access token usually only lasts a couple hours and needs to be submitted with every request.
-   The refresh token lasts for days to weeks and it's designed to be stored locally so you can regenerate access tokens. (We'll talk about this more in Caching, below.)
-   The user name + password gives access to everything so your code should never touch them!

Now that you've got the basic idea of OAuth, lets talk about the details.

## Clients

The first step in working with any OAuth API is to create an **application** **client**.
It's called an application client because in the wider world OAuth is usually used with a web, phone, or tv app, but here your "app" is going to be your R package.
For this reason, httr2 calls the application client a **client**, but many APIs will call it an OAuth application.

To create a client, you'll need to first register for a developer account on the API's website.
In most cases this is easy, totally automated, and only takes a couple of minutes.
That will give you access to some sort of developer portal which you can then use to register a new OAuth app (aka a client).
This process varies from API to API (it's normal to spend some time hunting through the docs and settings), but at the end of it you'll get a client id (another random string of numbers and letters).
Sometimes the client id is all you need, and you can create an httr2 client with `oauth_client()`, e.g.:

```{r}
client <- oauth_client(
  id = "28acfec0674bb3da9f38",
  token_url = "https://github.com/login/oauth/access_token",
  name = "hadley-oauth-test-2"
)
```

The call to `oauth_client()` also includes a `name` and a `token_url`.

-   The `name` is human-facing, and should typically be the same as your package (the thing that prompted you to create the client). I have a bunch of apps that I've used for testing, so here I've used the name `hadley-oauth-test-2` to remind me which app this client corresponds to.
-   The `token_url` points to the URL that's used to obtain an access token. You'll need to find this from the documentation for the API you're wrapping; it will typically be found in the section that describes the OAuth process and will be an endpoint that returns an access token. Don't be surprised if this endpoint feels very different to the rest of the API; auth is often implemented by a third-party package with slightly different conventions to the rest of the API.

### Client secret

In most cases, however, the API will also require a client secret.
While this is called a secret, it's typically not that important to keep it a secret because of two reasons:

-   It's typically easy to create an new app on the developer website so stealing yours wouldn't save much time.
-   It's unusual for an OAuth client to be able to do anything in its own right, so stealing your secret doesn't have much benefit.

That means that unless you have paid for the app or given it private information while creating it, it's ok to embed the client in your package.
That said, httr2 provides some tooling to obfuscate your client secret so that the client secret isn't directly embedded in your source code, and hence vulnerable to scraping.

To obfuscate a string, call `obfuscate()`, then copy and paste the result into your package.
For example, if your client secret was "secret", you'd call `obfuscate()` then you'd copy and paste `obfuscated("B4Evdd5x4wl0XTWvtTpuGaw7nM7GEg")` into your client specification.

```{r}
obfuscate("secret")
```

Here's what a complete client specification for GitHub looks like, using a real app that I created specifically for this vignette:

```{r}
client <- oauth_client(
  id = "28acfec0674bb3da9f38",
  secret = obfuscated("J9iiGmyelHltyxqrHXW41ZZPZamyUNxSX1_uKnvPeinhhxET_7FfUs2X0LLKotXY2bpgOMoHRCo"),
  token_url = "https://github.com/login/oauth/access_token",
  name = "hadley-oauth-test"
)
```

You can certainly uncover my client secret if you are an experienced R programmer and are willing to spend a bit of time experimenting, but I'm pretty sure it'd be easy for you to just create your own app on GitHub.

### Packaging

I recommend wrapping client creation in a function in your package, e.g.:

```{r}
github_client <- function() {
  oauth_client(
    id = "28acfec0674bb3da9f38",
    secret = obfuscated("J9iiGmyelHltyxqrHXW41ZZPZamyUNxSX1_uKnvPeinhhxET_7FfUs2X0LLKotXY2bpgOMoHRCo"),
    token_url = "https://github.com/login/oauth/access_token",
    name = "hadley-oauth-test"
  )
}
```

You'll need this in order to run tests for your package, but you'll probably also want to use it as the default client for your users.
In some cases, it will be necessary for each user to create their own app and matching client (e.g. if rate limits are applied by app, not by user), but this is much less user friendly so you should avoid it if possible.
That said, you should always provide a way for a user to supply their own client when you bundle a default.
You can see an example of this in [the googledrive package](https://googledrive.tidyverse.org/articles/bring-your-own-app.html).

## Authorization code flow

Once you have a client you need to use it with a flow in order to get a token.
You'll need to read the docs for your API to figure out which flows it supports, but the most common is the **authorization code**[^2] flow, which works something like this:

[^2]: Confusingly, Github calls this the "web application flow", but this is not the official name.
    This is a general problem with OAuth docs; often they use different terms for the same things.
    In httr2, I've tried to stick as closely as possible to the terms used in the official RFC specifications.

1.  httr2 opens a browser using the **authorization URL** provided by the API. The URL includes parameters that identify your app and what **scope** of access you're looking for (e.g. `tweet.read`, `userinfo.write`).
2.  The user logs in using their user name and password (hopefully using a [password manager](https://www.nytimes.com/wirecutter/blog/why-you-need-a-password-manager-yes-you/)) and approves the request.
3.  The API sends an **authorization code** back to httr2 using a callback URL that was supplied in the initial request.
4.  httr2 sends the authorization code to the **token URL** to get an access token.

In httr2 this flow is implemented by a pair of functions: `oauth_flow_auth_code()` and `req_oauth_auth_code()`.
Start with `oauth_flow_auth_code()` to check that you have all the parameters correctly specified, then use `req_oauth_auth_code()` to authenticate requests.
These two steps are described in the following sections.

### Creating a token

`oauth_flow_auth_code()` is the best way to verify you've correctly specified the parameters to the client and the `auth_url`, without depending on correctly understanding any other part of the API.
For example, you could get a token to access the GitHub API (using the client defined above) with this code:

```{r, eval = FALSE}
token <- oauth_flow_auth_code(
  client = client, 
  auth_url = "https://github.com/login/oauth/authorize"
)
```

This flow can't be used inside a vignette because it's designed specifically for interactive use, but if you do run it and print out the `token`, you'll see something like this:

```{r}
#| eval: false
token
#> <httr2_token>
#> token_type: bearer
#> access_token: <REDACTED>
#> scope: ''
```

There's not much to see here because httr2 automatically redacts the access token (because that could be used to perform actions on behalf of the user).

If your call to `oauth_flow_auth_code()` succeeds then you've got everything set up correctly and you can proceed to the next step.
Otherwise, you'll get an HTTP error.
If you're very lucky, that error will be informative and will help you figure out want went wrong.
However, in most cases, you'll need to carefully double check that you've correctly copied and pasted the client id and secret, and check that you've supplied the correct authorization and token urls (`auth_url` and `token_url`).
If the docs have multiple candidates for each and you're unclear about which is which, you'll need to do some systematic experimentation.

### Authenticating a request

Initial configuration is the only time that you'll see an `httr2_token` object because you'll generally want to rely on httr2 to manage the tokens for you.
You'll do that with `req_oauth_auth_code()`.
To check that it's working correctly, I recommend finding the simplest possible API endpoint to test it with.
A good place to start is an endpoint that provides information about the "current" user, if your API provides one.

For example, the GitHub API provides a `GET` endpoint at `/user` that returns information about the current user.
If we make a request to this endpoint without authentication, we'll get an error:

```{r}
#| error: true
req <- request("https://api.github.com/user")
req |> 
  req_perform()
```

We can authenticate this request with `req_oauth_auth_code()`, using the same arguments as our previous call to `oauth_flow_auth_code()`:

```{r}
#| eval: false
req |> 
  req_oauth_auth_code(
    client = github_client(),
    auth_url = "https://github.com/login/oauth/authorize"
  ) |> 
  req_perform() |> 
  resp_body_json() |> 
  str()
```

When you run this code, you'll see something like this, but it will obviously contain information about you, not me.

```{r}
#| eval: false
#> List of 32
#>  $ login        : chr "hadley"
#>  $ id           : int 4196
#>  $ node_id      : chr "MDQ6VXNlcjQxOTY="
#>  $ avatar_url   : chr "https://avatars.githubusercontent.com/u/4196?v=4"
#>  $ gravatar_id  : chr ""
#>  $ url          : chr "https://api.github.com/users/hadley"
#>  $ html_url     : chr "https://github.com/hadley"
#>  ...
#>  $ type         : chr "User"
#>  $ site_admin   : logi FALSE
#>  $ name         : chr "Hadley Wickham"
#>  $ company      : chr "@posit-pbc"
#>  $ blog         : chr "http://hadley.nz"
#>  $ location     : chr "Houston, TX"
```

### Caching

There are two big reasons to allow httr2 to manage tokens for you.
The first is that httr2 will automatically refresh the token if it's expired.
The second is cross-session caching, as described below.

By default, the OAuth token will be cached in memory.
That means that you will only need to authenticate once in the current session, but you'll need to re-authenticate if you restart R.
In some cases, you may want to save the tokens (refresh and access) so that they can be used across sessions.
This is easy to do (just set `cache_disk = TRUE`) but you need to think through the consequences of saving the refresh token on disk.

httr2 does the best it can to save these credentials securely.
They are stored in a local cache directory (`oauth_cache_path()`) that is accessible to the current user, and they are encrypted so they will be hard for any package other than httr2 to read.
However, there's no way to prevent other R code from using httr2 to access them, so if you do choose to cache tokens, you should inform the user and give them the ability to opt-out.
httr2 automatically deletes any cached tokens that are older than 30 days whenever it's loaded.
This means that you'll need to re-auth at least once a month, but prevents tokens from hanging around on disk long after you've forgotten you created them.

You can see which clients have cached tokens by looking in the cache directory used by httr2:

```{r}
#| eval: false
dir(oauth_cache_path(), recursive = TRUE)
```

Each client gets its own subdirectory named using the client `name`, so if you turn caching on, it's particularly important to give your client a good name so that the user can easily tell which package the tokens belong to.

## Other flows

When wrapping an API, you'll need to carefully read the documentation to figure out which flows it provides.
If possible you'll want to use the authorization code flow since it generally provides the best experience, but if it's not available you'll need to carefully consider the others.
Currently, httr2 supports the following flows:

-   `req_oauth_device()` uses the "device" flow which is designed for devices like TVs that don't have an easy way to enter data.
    It also works well from the console.

-   `req_oauth_client_credentials()` and `req_oauth_bearer_jwt()` are often needed for **service accounts**, accounts that represent automated services, not people, and are often used in non-interactive environments.

-   `req_oauth_password()` exchanges a user name and password for an access token.

-   `req_oauth_refresh()` works directly with a refresh token that you already have.
    It's useful for testing and automation.

httr2 doesn't support the implicit grant flow.
This was historically important but is now [mostly deprecated](https://developer.okta.com/blog/2019/05/01/is-the-oauth-implicit-flow-dead) and was never a particularly good fit for R because it relies on a technique for returning the access token that only reliably works inside a web browser.

Regardless of which flow you use, you'll need to follow the same process as the example above: first figure out how to get a token using the `oauth_flow_` function, then actually use oauth with a request, by calling the matching `req_oauth_` function.

One additional wrinkle is that many APIs don't implement the flow in exactly the same way as the spec so httr2's built-in flows might not work at all.
If your initial attempt doesn't work, you're going to need to do some sleuthing.
This is going to be painful, but unfortunately there's no way around it.
I recommend wrapping your httr2 code in `with_verbosity()` so you can see exactly what httr2 is sending to the server.
You'll then need to carefully compare this to the API documentation and play "spot the difference".
You're very welcome to file an issue and I'll do my best to help you out.

---
title: "Wrapping APIs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Wrapping APIs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

A common use for httr2 is wrapping up a useful API and exposing it in an R package where each API endpoint (i.e. a URL with parameters) becomes an R function with documented arguments.
This vignette will show you how, starting with a very simple API that doesn't need authentication, then slowly working up in complexity.
Along the way, you'll learn about how to:

-   Expose important details from HTTP errors in R errors.

-   Handle various types of authentication.

-   Consistently throttle the rate of requests or dynamically respond to rate limiting headers sent by the server.

I assume you're familiar with the basics of building a package.
If not, you might want to read the "[The Whole Game](https://r-pkgs.org/whole-game.html)" chapter of [R packages](https://r-pkgs.org) first.

```{r setup}
library(httr2)
```

## Faker API

```{r, include = FALSE}
# Seems to return 500s from time-to-time, so avoid any problems
# by only evaluating other chunks if a simple request succeeds.
faker_status_images <- request("https://fakerapi.it/api/v1") |> 
  req_url_path_append("images") |> 
  req_error(is_error = ~ FALSE) |> 
  req_perform() |> 
  resp_status()
faker_status_persons <- request("https://fakerapi.it/api/v1") |> 
  req_url_path_append("persons") |> 
  req_error(is_error = ~ FALSE) |> 
  req_perform() |> 
  resp_status()

faker_ok <- faker_status_images < 400 && faker_status_persons < 400
```

We'll start with a very simple API, [faker API](https://fakerapi.it/en){.uri}, which provides a collection of techniques for generating fake data.
Before we start writing the sort of functions that you might put in a package, we'll perform a request just to see how the basics work:

```{r, eval = faker_ok}
# We start by creating a request that uses the base API url
req <- request("https://fakerapi.it/api/v1")
resp <- req |> 
  # Then we add on the images path
  req_url_path_append("images") |> 
  # Add query parameters _width and _quantity
  req_url_query(`_width` = 380, `_quantity` = 1) |> 
  req_perform()

# The result comes back as JSON
resp |> resp_body_json() |> str()
```

### Errors

It's always worth a little early experimentation to see if we get any useful information from errors.
The httr2 defaults get in your way here, because if you retrieve an unsuccessful HTTP response, you automatically get an error that prevents you from further inspecting the body:

```{r, error = TRUE, eval = faker_ok}
req |> 
  req_url_path_append("invalid") |> 
  req_perform()
```

However, you can access the last response (successful or not) with `last_response()`:

```{r, eval = faker_ok}
resp <- last_response()
resp |> resp_body_json()
```

It doesn't look like there's anything useful there.
Sometimes useful info is returned in the headers, so let's check:

```{r, eval = faker_ok}
resp |> resp_headers()
```

It doesn't look like we're getting any more useful information, so we can leave the `req_error()` default as is.
We'll have another go later with an API that does provide more details.

### User agent

If you're wrapping this code into a package, it's considered polite to set a user agent, so that, if your package accidentally does something horribly wrong, the developers of the website can figure out who to reach out to.
You can do this with the `req_user_agent()` function:

```{r, eval = faker_ok}
req |>
  req_user_agent("my_package_name (http://my.package.web.site)") |> 
  req_dry_run()
```

### Core request function

Once you've made a few successful requests, it's worth seeing if you can figure out the general pattern so you can wrap it up into a function that will become the core of your package.

For faker, I spent a little time with the [documentation](https://fakerapi.it/en#basic-usage) noting some commonalities:

-   Every URL is of the form `https://fakerapi.it/api/v1/{resource}`, and data is passed to the resource with query parameters.
    All parameters start with `_`.

-   Every resource has three common query parameters: `_locale`, `_quantity`, and `_seed`.

-   All endpoints return JSON data.

This led me to construct the following function:

```{r, eval = faker_ok}
faker <- function(resource, ..., quantity = 1, locale = "en_US", seed = NULL) {
  params <- list(
    ...,
    quantity = quantity,
    locale = locale,
    seed = seed
  )
  names(params) <- paste0("_", names(params))
  
  request("https://fakerapi.it/api/v1") |> 
    req_url_path_append(resource) |> 
    req_url_query(!!!params) |> 
    req_user_agent("my_package_name (http://my.package.web.site)") |> 
    req_perform() |> 
    resp_body_json()
}

str(faker("images", width = 300))
```

I've made a few important choices here:

-   I've decided to supply default values for the `quantity` and `locale` parameters.
    This makes my function easier to demo in this vignette.

-   I've used a default of `NULL` for the `seed` argument.
    `req_url_query()` will automatically drop `NULL` arguments so this means that no default value is sent to the API, but when you read the function definition you can still see that `seed` is accepted.

-   I automatically prefix all query parameters with `_` because argument names starting with `_` are hard to type in R.

-   My function generates the request, performs it, and extracts the body of the response.
    This works well for simple cases, but for more complex APIs you might want to return a request object that can be modified before being performed.

I also used one cool trick: `req_url_query()` uses [dynamic dots](https://rlang.r-lib.org/reference/dyn-dots.html), so we can use `!!!` to convert (e.g.) `` req_url_query(req, !!!list(`_quantity` = 1, `_locale` = "en_US")) `` into `` req_url_query(req, `_quantity` = 1, `_locale` = "en_US") ``.

### Wrapping individual endpoints

`faker()` is quite general --- it's a good tool for the package developer because you can read the faker documentation and translate it to a function call.
But it's not very friendly for the package user who might not know anything about web APIs.
So typically the next step in the process is to wrap up some individual endpoints with their own functions.

For example, let's take the `persons` endpoint which has three additional parameters: `gender` (male or female), `birthday_start`, and `birthday_end`.
A simple wrapper would start something like this:

```{r, eval = faker_ok}
faker_person <- function(gender = NULL, birthday_start = NULL, birthday_end = NULL, quantity = 1, locale = "en_US", seed = NULL) {
  faker(
    "persons",
    gender = gender,
    birthday_start = birthday_start,
    birthday_end = birthday_end,
    quantity = quantity,
    locale = locale,
    seed = seed
  )  
}
str(faker_person("male"))
```

We could make it more user friendly by checking the input types, and returning the result as a tibble.
I did a quick and dirty conversion using purrr; depending on your needs you could use base R code or `tidyr::hoist()`.

```{r, eval = faker_ok}
library(purrr)

faker_person <- function(gender = NULL, birthday_start = NULL, birthday_end = NULL, quantity = 1, locale = "en_US", seed = NULL) {
  if (!is.null(gender)) {
    gender <- match.arg(gender, c("male", "female"))
  }
  if (!is.null(birthday_start)) {
    if (!inherits(birthday_start, "Date")) {
      stop("`birthday_start` must be a date")
    }
    birthday_start <- format(birthday_start, "%Y-%m-%d")
  }
  if (!is.null(birthday_end)) {
    if (!inherits(birthday_end, "Date")) {
      stop("`birthday_end` must be a date")
    }
    birthday_end <- format(birthday_end, "%Y-%m-%d")
  }
  
  json <- faker(
    "persons",
    gender = gender,
    birthday_start = birthday_start,
    birthday_end = birthday_end,
    quantity = quantity,
    locale = locale,
    seed = seed
  )  
  
  tibble::tibble(
    firstname = map_chr(json$data, "firstname"),
    lastname = map_chr(json$data, "lastname"),
    email = map_chr(json$data, "email"),
    gender = map_chr(json$data, "gender")
  )
}
faker_person("male", quantity = 5)
```

The next steps would be to export and document this function; I'll leave that up to you.

## Secret management

We need to take a quick break from APIs to talk about secrets.
Secrets are important, because every API (except for very simple APIs like faker) is going to require that you identify yourself in some way, typically with an API key or a token.
And even if you plan to require your users to supply this information, you'll still need to record your own credentials in order to test your own package.

This system described below is likely to be overkill if you have one secret that only needs to be shared in a couple of places: you can just put it in your `.Renviron` and access it with `Sys.getenv()`.
But you will probably accumulate more secrets over time, and you'll need to figure out how to share them with other people and other computers, so I think spending a little time to understand this system and set up it for your package will pay off in the long term.

### Basics

httr2 provides `secret_encrypt()` and `secret_decrypt()` to scramble secrets so that you can include them in your public source code without worrying that others can read them.
There are three basic steps to this process:

1.  You create an **encryption** key with `secret_make_key()` that is used to scramble and descramble secrets using symmetric cryptography:

    ```{r}
    key <- secret_make_key()
    key
    ```

    (Note that `secret_make_key()` uses a cryptographically secure random number generator provided by OpenSSL; it is not affected by R's RNG settings, and there's no way to make it reproducible.)

2.  You scramble your secrets with `secret_encrypt()` and store the resulting text directly in the source code of your package:

    ```{r}
    secret_scrambled <- secret_encrypt("secret I need to work with an API", key)
    secret_scrambled
    ```

3.  When needed, you descramble the secret using `secret_decrypt()`:

    ```{r}
    secret_decrypt(secret_scrambled, key)
    ```

### Package keys and secrets

You can create any number of encryption keys, but I highly recommend that you create one key per package, which I'll call the **package** key.
In this section, I'll show you how to store that key so that you (and your automated tests) can use it, but no one else can.

httr2 is built around the notion that this key should live in an environment variable.
So the first step is to make your package key available on your local development machine by adding a line to your your user-level `.Renviron` (which you can easily open with `usethis::edit_r_environ()`):

```         
YOURPACKAGE_KEY=key_you_generated_with_secret_make_key
```

```{r, include = FALSE}
Sys.setenv(YOURPACKAGE_KEY = secret_make_key())
```

Now (after you restart R), you'll be able to take advantage of a special `secret_encrypt()` and `secret_decrypt()` feature: the `key` argument can be the name of an environment variable, instead of the encryption key itself.
In fact, this is most natural usage.

```{r}
secret_scrambled <- secret_encrypt("secret I need to work with an API", "YOURPACKAGE_KEY")
secret_scrambled
secret_decrypt(secret_scrambled, "YOURPACKAGE_KEY")
```

You'll also need to make the key available in your GitHub Actions (both check and pkgdown) so your automated tests can use it.
This requires two steps:

1.  Add the key to your [repository secrets](https://docs.github.com/en/actions/reference/encrypted-secrets).

2.  Share the key with the workflows that need it by adding a line to the appropriate workflow:

    ``` yaml
        env:
          YOURPACKAGE_KEY: ${{ secrets.YOURPACKAGE_KEY }}
    ```

    You can see how httr2 does it in [its GitHub workflow](https://github.com/r-lib/httr2/blob/master/.github/workflows/R-CMD-check.yaml){.uri}.

Other continuous integration platforms will offer similar ways to make a key available as a secure environment variable.

### When the package key isn't available

There are a few important cases where your code won't have access to your package key: on CRAN, on the personal machines of external contributors, and in automated checks on their PRs.
So if you want to share your package on CRAN or make it easy for others to contribute, you need to make sure that your examples, vignettes, and tests all work without error:

-   In vignettes, you can run `knitr::opts_chunk(eval = secret_has_key("YOURPACKAGE_KEY"))` so that chunks are only evaluated if your key is available.

-   In examples, you can surround code blocks that require your key with `if (httr2::secret_has_key("YOURPACKAGE_KEY")) {}`

-   You don't need to do anything in tests because when `secret_decrypt()` is run by testthat, it will automatically `skip()` the test if the key isn't available.

## NYTimes Books API

```{r include = FALSE}
has_auth <- secret_has_key("HTTR2_KEY") && 
  # Don't run in R CMD check on GHA because it hits NYTimes rate limits
  !identical(Sys.getenv("GITHUB_WORKFLOW"), "R-CMD-check")
knitr::opts_chunk$set(eval = has_auth)
```

Next we'll take a look at the NYTimes [Books API](https://developer.nytimes.com/docs/books-product/1/overview). It requires a very simple authentication with an API key that's included in every request.
When you're wrapping an API that has a key you're going to face two struggles:

-   How do you test your package without sharing your key with the whole world?

-   How do you allow your users to supply their own key, without having to pass it to every function?

So now you can understand how the following code works to get my NYTimes Book API key:

```{r}
my_key <- secret_decrypt("4Nx84VPa83dMt3X6bv0fNBlLbv3U4D1kHM76YisKEfpCarBm1UHJHARwJHCFXQSV", "HTTR2_KEY")
```

I'll start by tackling the first problem because otherwise there's no way for me to show how the API works in this vignette 😃.
We'll come back to the second at the very end of this section, because it's easiest to tackle once we have a function in place.

### Security considerations

Note that including an API key as a query parameter is relatively insecure; if an API uses this method of auth, it's typically because the key is relatively easy to create or gives relatively few privileges.
Here it only takes a couple of minutes to generate your own NYTimes API key, so there's little incentive for someone to try and steal yours.

The main problem of conveying credentials via the url is that it's easily exposed, because httr2 makes no efforts to redact confidential information stored in query parameters.
This means it's relatively easy to leak your key if you use `req_perform(verbosity = 1)`, `req_dry_run()`, or even just print the request object.
And indeed, you'll see that in the examples below --- this is bad practice for a real package, but I think it's ok here because the key doesn't allow you to do anything valuable and it makes teaching APIs so much easier.

### Basic request

Now let's perform a test request and look at the response:

```{r}
resp <- request("https://api.nytimes.com/svc/books/v3") |> 
  req_url_path_append("/reviews.json") |> 
  req_url_query(`api-key` = my_key, isbn = 9780307476463) |> 
  req_perform()
resp
```

Like most modern APIs, this one returns the results as JSON:

```{r}
resp |> 
  resp_body_json() |> 
  str()
```

Before we start wrapping this up into a function, let's consider what happens with errors.

### Error handling

What happens if there's an error?
For example, if we deliberately supply an invalid key:

```{r, error = TRUE}
resp <- request("https://api.nytimes.com/svc/books/v3") |> 
  req_url_path_append("/reviews.json") |> 
  req_url_query(`api-key` = "invalid", isbn = 9780307476463) |> 
  req_perform()
```

To see if there's any extra useful information we can again look at `last_response():`

```{r}
resp <- last_response()
resp
resp |> resp_body_json()
```

It looks like there's some useful additional info in the `faultstring`:

```{r}
resp |> resp_body_json() |> _$fault |> _$faultstring
```

To add that information to future errors we can use the `body` argument to `req_error()`.
This should be a function that takes a response and returns a character vector of additional information to include in the error.
Once we do that and re-fetch the request, we see the additional information displayed in the R error:

```{r, error = TRUE}
nytimes_error_body <- function(resp) {
  resp |> resp_body_json() |> _$fault |> _$faultstring
}

resp <- request("https://api.nytimes.com/svc/books/v3") |> 
  req_url_path_append("/reviews.json") |> 
  req_url_query(`api-key` = "invalid", isbn = 9780307476463) |> 
  req_error(body = nytimes_error_body) |> 
  req_perform()
```

### Rate limits

Another common source of errors is rate-limiting --- this is used by many servers to prevent one unruly client consuming too many resources.
The [frequently asked questions](https://developer.nytimes.com/faq#a11) page describes the rate limits for the NYT APIs:

> Yes, there are two rate limits per API: 4,000 requests per day and 10 requests per minute.
> You should sleep 6 seconds between calls to avoid hitting the per minute rate limit.
> If you need a higher rate limit, please contact us at [code\@nytimes.com](mailto:code@nytimes.com){.email}.

Many APIs return additional information about how long to wait when the rate limit is exceeded (often using the `Retry-After` header).
So I deliberately violated the rate limit by quickly making 11 requests; unfortunately while the response was a standard 429 (Too many requests), it did not include any information about how long to wait in either the response body or the headers.
That means we can't use `req_retry()`, which automatically waits the amount of time the server requests.
Instead, we'll use `req_throttle()` to ensure we don't make more than 10 requests every 60 seconds:

```{r}
req <- request("https://api.nytimes.com/svc/books/v3") |> 
  req_url_path_append("/reviews.json") |> 
  req_url_query(`api-key` = "invalid", isbn = 9780307476463) |> 
  req_throttle(10 / 60)
```

By default, `req_throttle()` shares the limit across all requests made to the host (i.e. `api.nytimes.com`).
Since the docs suggest the rate limit applies per API, you might want to use the `realm` argument to be a bit more specific:

```{r}
req <- request("https://api.nytimes.com/svc/books/v3") |> 
  req_url_path_append("/reviews.json") |> 
  req_url_query(`api-key` = "invalid", isbn = 9780307476463) |> 
  req_throttle(10 / 60, realm = "https://api.nytimes.com/svc/books")
```

### Wrapping it up

Putting together all the pieces above yields a function something like this:

```{r}
nytimes_books <- function(api_key, path, ...) {
  request("https://api.nytimes.com/svc/books/v3") |> 
    req_url_path_append("/reviews.json") |> 
    req_url_query(..., `api-key` = api_key) |> 
    req_error(body = nytimes_error_body) |> 
    req_throttle(10 / 60, realm = "https://api.nytimes.com/svc/books") |> 
    req_perform() |> 
    resp_body_json()
}

drunk <- nytimes_books(my_key, "/reviews.json", isbn = "0316453382")
drunk$results[[1]]$summary
```

To finish this up for a real package, you'd want to:

-   Add explicit arguments and check that they have the correct type.

-   Export and document the function.

-   Convert the nested list into a more user-friendly data structure (probably a data frame with one row per review).

You'd also want to provide some convenient way for the user to supply their own API key.

### User-supplied key

A good place to start is an environment variable, because environment variables are easy to set without typing anything in the console (which can get accidentally shared via your `.Rhistory`) and are easily set in automated processes.
Then you'd write a function to retrieve the API key, returning a helpful message if it's not found:

```{r}
get_api_key <- function() {
  key <- Sys.getenv("NYTIMES_KEY")
  if (identical(key, "")) {
    stop("No API key found, please supply with `api_key` argument or with NYTIMES_KEY env var")
  }
  key
}
```

Then you could modify `nytimes_books()` to use `get_api_key()` as the default value for `api_key`.
Since the argument is now optional, we can move it to end of the argument list, since it'll only be needed in exceptional circumstances.

```{r, eval = FALSE}
nytimes_books <- function(path, ..., api_key = get_api_key()) {
  ...
}
```

You can make this approach a little more user friendly by providing a helper that sets the environment variable:

```{r}
set_api_key <- function(key = NULL) {
  if (is.null(key)) {
    key <- askpass::askpass("Please enter your API key")
  }
  Sys.setenv("NYTIMES_KEY" = key)
}
```

Using askpass (or similar) here is good practice since you don't want to encourage the user to type their secret key into the console, as mentioned above.

It's a good idea to extend `get_api_key()` to automatically use your encrypted key to make it easier to write tests:

```{r}
get_api_key <- function() {
  key <- Sys.getenv("NYTIMES_KEY")
  if (!identical(key, "")) {
    return(key)
  }
  
  if (is_testing()) {
    return(testing_key())
  } else {
    stop("No API key found, please supply with `api_key` argument or with NYTIMES_KEY env var") 
  }
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

testing_key <- function() {
  secret_decrypt("4Nx84VPa83dMt3X6bv0fNBlLbv3U4D1kHM76YisKEfpCarBm1UHJHARwJHCFXQSV", "HTTR2_KEY")
}
```

## Github Gists API

Next we'll take a look at an API that can make changes on behalf of a user, not just retrieve data: [GitHub's gist API](https://docs.github.com/en/rest/reference/gists).
This uses different HTTP methods to perform different actions, like creating, updating, and deleting gists.
But before we can get to those, let's handle authentication, rate-limiting, and errors.

### Authentication

The easiest way to authenticate with a GitHub API is to use a personal access token.
A token is an alternative to a username and password.
You have one username + password per site; you can have one token per use case.
This lets each use case have a minimal set of permissions, and you can easily revoke one token without affecting any other use case.

I created a personal access token specifically for this vignette that can only access gists, and, as in the last example, stored an encrypted version in this vignette:

```{r}
token <- secret_decrypt("Guz59woxKoIO_JVtp2IzU3mFIU3ULtaUEa8xvvpYUBdVthR8jhxzc3bMZFhA9HL-ZK6YZudOI6g", "HTTR2_KEY")
```

If you want to run this vignette yourself, you'll need to create a new token in your [GitHub settings](https://github.com/settings/tokens); just make sure it includes the "gist" scope.
It's also a good idea to give every token a descriptive name, that reminds you of its motivating use case, and of where to update it when you have to re-generate it because it expired.

To authenticate a request with the token, we need to put it in the `Authorization` header with a ["token" prefix](https://docs.github.com/en/rest/overview/resources-in-the-rest-api#authentication):

```{r}
req <- request("https://api.github.com/gists") |> 
  req_headers(Authorization = paste("token", token))

req |> req_perform()
```

Because the authorization header usually contains secret information, httr2 automatically redacts it[^1]:

[^1]: Again, it's still possible to extract it with a little extra work, but httr2 tries to help you avoid revealing it by accident.
    httr2 protects you from yourself, not from someone deliberately trying to find the secret.

```{r}
req
req |> req_dry_run()
```

### Errors

Once you've got authentication working, it's always a good idea to work on errors next, since that will help you debug any failed requests.
In my experience APIs rarely do a good job of documenting their errors, so you'll often have to do a little experimentation.
To add to the pain, in large APIs different endpoints often return different amounts of information in different forms.
You'll typically need to tackle your error handling iteratively, improving your code each time you encounter a new problem.

While GitHub does [document its errors](https://docs.github.com/en/rest/overview/resources-in-the-rest-api#client-errors), I'm sufficiently distrustful that I still want to construct a deliberately malformed query and see what happens:

```{r, error = TRUE}
resp <- request("https://api.github.com/gists") |> 
  req_url_query(since = "abcdef") |> 
  req_headers(Authorization = paste("token", token)) |> 
  req_perform()
```

As documented I get a 422 "Unprocessable Entity" error.
But the response is rather different to documentation which suggests there should be a string `message` and a list of `errors`:

```{r}
resp <- last_response()
resp
resp |> resp_body_json()
```

I'll proceed anyway, writing a function that extracts the data and formats it for presentation to the user:

```{r}
gist_error_body <- function(resp) {
  body <- resp_body_json(resp)
  
  message <- body$message
  if (!is.null(body$documentation_url)) {
    message <- c(message, paste0("See docs at <", body$documentation_url, ">"))
  }
  message
}
gist_error_body(resp)
```

Now I can pass this function to the `body` argument of `req_error()` and it will be automatically included in the error when a request fails:

```{r, error = TRUE}
request("https://api.github.com/gists") |> 
  req_url_query(since = "yesterday") |> 
  req_headers(Authorization = paste("token", token)) |> 
  req_error(body = gist_error_body) |> 
  req_perform()
```

Notice that each element of the character vector produced by `gh_error_body()` becomes a bullet in the resulting error.

### Rate-limiting

While we're thinking about errors, it's useful to look at what happens if the requests are rate limited.
Luckily, GitHub consistently uses response headers to provide information about the remaining [rate limits](https://docs.github.com/en/rest/overview/resources-in-the-rest-api#rate-limiting).

```{r}
resp <- req |> req_perform() 
resp |> resp_headers("ratelimit")
```

We can teach httr2 about this so it can automatically wait for a reset if the rate limit is hit.
We need to define two functions.
The first tells us whether or not a response has a transient error, i.e. it's worth waiting and trying again.
For GitHub, when the rate limit is [exceeded](https://docs.github.com/en/rest/overview/resources-in-the-rest-api#rate-limiting), the response has a 403 status and a `X-RateLimit-Remaining: 0` header:

```{r}
gist_is_transient <- function(resp) {
  resp_status(resp) == 403 && 
    resp_header(resp, "X-RateLimit-Remaining") == "0"
}
gist_is_transient(resp)
```

Then we need a function tells how long to wait.
GitHub tells us when the rate limit resets (as number of seconds since 1970-01-01) in the `X-RateLimit-Reset` header.
To convert that to a number of seconds to wait we first convert it to a number (since HTTP headers are always strings), then subtract off the current time (in number of seconds since 1970-01-01):

```{r}
gist_after <- function(resp) {
  time <- as.numeric(resp_header(resp, "X-RateLimit-Reset"))
  time - unclass(Sys.time())
}
gist_after(resp)
```

We then pass functions to `req_retry()` so httr2 has all the information it needs to handle rate-limiting automatically:

```{r}
request("http://api.github.com") |>
  req_retry(
    is_transient = gist_is_transient,
    after = gist_after,
    max_seconds = 60
  )
```

You also need to supply either `max_tries` or `max_seconds` in order to activate `req_retry()`.

### Wrapping it all up

Let's wrap up everything we've learned so far into a single function that creates a request:

```{r}
req_gist <- function(token) {
  request("https://api.github.com/gists") |> 
    req_headers(Authorization = paste("token", token)) |> 
    req_error(body = gist_error_body) |> 
    req_retry(
      is_transient = gist_is_transient,
      after = gist_after
    )
}

# Check it works:
req_gist(token) |> 
  req_perform()
```

We'll use this as the basis to solve the next challenge: uploading a gist.

### Sending data

To [create a gist](https://docs.github.com/en/rest/reference/gists#create-a-gist) we need to change the method to `POST` and add a body that contains data encoded as JSON.
httr2 provides one function that does both of these things: `req_body_json()`:

```{r}
req <- req_gist(token) |> 
  req_body_json(list(
    description = "This is my cool gist!",
    files = list(test.R = list(content = "print('Hi!')")),
    public = FALSE
  ))
req |> req_dry_run()
```

Depending on the API you're wrapping, you might need to send data in a different way.
`req_body_form()` and `req_body_multipart()` make it easier to encode data in two other common forms.
If the API requires something different you can use `req_body_raw()`.

Typically, the API will return some useful data about the resource you've just created.
Here I'll extract the gist ID so we can use it in the next examples, culminating with deleting the gist so I don't end up with a bunch of duplicated gists 😃.

```{r}
resp <- req |> req_perform()
id <- resp |> resp_body_json() |> _$id
id
```

### Changing a gist

Actually, that description wasn't very true and I want to change it.
To do so, I need to again send JSON encoded data, but this time I need to use the `PATCH` verb.
So after adding the data to request, I use `req_method()` to override the default method:

```{r}
req <- req_gist(token) |> 
  req_url_path_append(id) |> 
  req_body_json(list(description = "This is a simple gist")) |> 
  req_method("PATCH")
req |> req_dry_run()
```

### Deleting a gist

Deleting a gist is similar, except we don't send any data, we just need to adjust the default method from `GET` to `DELETE`.

```{r}
req <- req_gist(token) |> 
  req_url_path_append(id) |> 
  req_method("DELETE")
req |> req_dry_run()
req |> req_perform()
```

# generates useful errors

    Code
      req_template(req, 1)
    Condition
      Error in `req_template()`:
      ! `template` must be a single string, not the number 1.
    Code
      req_template(req, "x", 1)
    Condition
      Error in `req_template()`:
      ! All elements of `...` must be named.
    Code
      req_template(req, "A B C")
    Condition
      Error in `req_template()`:
      ! Can't parse template `template`.
      i Should have form like 'GET /a/b/c' or 'a/b/c/'.

# template produces useful errors

    Code
      template_process(":b")
    Condition
      Error:
      ! Can't find template variable "b".
    Code
      template_process(":b", list(b = sum))
    Condition
      Error:
      ! Template variable "b" is not a simple scalar value.

# has useful default (with message)

    Code
      req <- req_retry(req)
    Message
      Setting `max_tries = 2`.

# useful message if `after` wrong

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! The `after` callback to `req_retry()` must return a single number or NA, not a <httr2_response> object.

# validates its inputs

    Code
      req_retry(req, max_tries = 0)
    Condition
      Error in `req_retry()`:
      ! `max_tries` must be a whole number larger than or equal to 1 or `NULL`, not the number 0.
    Code
      req_retry(req, max_tries = 2, max_seconds = "x")
    Condition
      Error in `req_retry()`:
      ! `max_seconds` must be a whole number or `NULL`, not the string "x".
    Code
      req_retry(req, max_tries = 2, retry_on_failure = "x")
    Condition
      Error in `req_retry()`:
      ! `retry_on_failure` must be `TRUE` or `FALSE`, not the string "x".

# can check type of response

    Code
      resp_check_content_type(resp1, "application/xml")
    Condition
      Error:
      ! Unexpected content type "application/json".
      * Expecting type "application/xml"
    Code
      resp_check_content_type(resp2, "application/xml")
    Condition
      Error:
      ! Unexpected content type "xxxxx".
      * Expecting type "application/xml"

# useful error even if no content type

    Code
      resp_check_content_type(resp, "application/xml")
    Condition
      Error:
      ! Unexpected content type "NA".
      * Expecting type "application/xml"

# check_content_type() can consult suffixes

    Code
      check_content_type("application/json", "application/xml")
    Condition
      Error:
      ! Unexpected content type "application/json".
      * Expecting type "application/xml"

---

    Code
      check_content_type("application/test+json", "application/xml", "xml")
    Condition
      Error:
      ! Unexpected content type "application/test+json".
      * Expecting type "application/xml" or suffix "xml".

---

    Code
      check_content_type("application/xml", c("text/html", "application/json"))
    Condition
      Error:
      ! Unexpected content type "application/xml".
      * Expecting type "text/html" or "application/json"

# curl errors become errors

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! Failed to perform HTTP request.
      Caused by error in `curl_fetch()`:
      ! Failed to connect

# http errors become errors

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! HTTP 404 Not Found.

---

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! HTTP 429 Too Many Requests.

---

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! HTTP 599.

# checks input types

    Code
      req_perform(req, path = 1)
    Condition
      Error in `req_perform()`:
      ! `path` must be a single string or `NULL`, not the number 1.
    Code
      req_perform(req, verbosity = 1.5)
    Condition
      Error in `req_perform()`:
      ! `verbosity` must 0, 1, 2, or 3.
    Code
      req_perform(req, mock = 7)
    Condition
      Error in `req_perform()`:
      ! `mock` must be function, list, or NULL, not the number 7.

# req_perform_stream() has been soft deprecated

    Code
      . <- req_perform_stream(req, function(x) NULL)
    Condition
      Warning:
      `req_perform_stream()` was deprecated in httr2 1.2.0.
      i Please use `req_perform_connection()` instead.

# req_perform_stream checks its inputs

    Code
      req_perform_stream(1)
    Condition
      Error in `req_perform_stream()`:
      ! `req` must be an HTTP request object, not the number 1.
    Code
      req_perform_stream(req, 1)
    Condition
      Error in `req_perform_stream()`:
      ! `callback` must be a function, not the number 1.
    Code
      req_perform_stream(req, callback, timeout_sec = -1)
    Condition
      Error in `req_perform_stream()`:
      ! `timeout_sec` must be a number larger than or equal to 0, not the number -1.
    Code
      req_perform_stream(req, callback, buffer_kb = "x")
    Condition
      Error in `req_perform_stream()`:
      ! `buffer_kb` must be a number, not the string "x".

# as_round_function checks its inputs

    Code
      as_round_function(1)
    Condition
      Error:
      ! `round` must be "byte", "line" or a function.
    Code
      as_round_function("bytes")
    Condition
      Error:
      ! `round` must be one of "byte" or "line", not "bytes".
      i Did you mean "byte"?
    Code
      as_round_function(function(x) 1)
    Condition
      Error in `as_round_function()`:
      ! `round` must have the argument `bytes`; it currently has `x`.

# can customise error info

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! HTTP 404 Not Found.
      i Hi!

# long custom body is wrapped

    Code
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! HTTP 200 OK.
      i Ad aliquip et occaecat consequat eiusmod enim Lorem
        incididunt laboris deserunt. Consectetur magna ea ad quis
        dolore. Deserunt elit elit dolore magna fugiat ipsum id
        eu nostrud voluptate Lorem ad id anim. Cupidatat nulla
        ipsum irure nisi sunt ipsum commodo eu sint eiusmod
        consectetur.

# failing callback still generates useful body

    Failed to parse error body with method defined in `req_error()`.
    Caused by error:
    ! This is an error!

---

    Code
      req <- request_test("/status/404")
      req <- req_error(req, body = function(resp) resp_body_json(resp)$error)
      req_perform(req)
    Condition
      Error in `req_perform()`:
      ! Failed to parse error body with method defined in `req_error()`.
      Caused by error in `resp_body_json()`:
      ! Unexpected content type "text/plain".
      * Expecting type "application/json" or suffix "json".

# modify list adds, removes, and overrides

    Code
      modify_list(x, a = 1, 2)
    Condition
      Error:
      ! All components of `...` must be named.

# progress bar suppressed in tests

    Code
      sys_sleep(0.1, "in test")
    Message
      > Waiting 0.1s in test

# validates inputs

    Code
      local_mocked_responses(function(foo) { })
    Condition
      Error in `local_mocked_responses()`:
      ! `mock` must have the argument `req`; it currently has `foo`.
    Code
      local_mocked_responses(10)
    Condition
      Error in `local_mocked_responses()`:
      ! `mock` must be function, list, or NULL, not the number 10.

# request and paths must match

    Code
      req_perform_parallel(req, letters)
    Condition
      Error in `req_perform_parallel()`:
      ! If supplied, `paths` must be the same length as `req`.

# errors by default

    Code
      req_perform_parallel(reqs[1])
    Condition
      Error in `req_perform_parallel()`:
      ! HTTP 404 Not Found.
    Code
      req_perform_parallel(reqs[2])
    Condition
      Error in `req_perform_parallel()`:
      ! Could not resolve host: INVALID

# multi_req_perform is deprecated

    Code
      multi_req_perform(list())
    Condition
      Warning:
      `multi_req_perform()` was deprecated in httr2 1.0.0.
      i Please use `req_perform_parallel()` instead.
    Output
      list()

# turns oauth errors to R errors

    Code
      oauth_flow_fetch(req, "test")
    Condition
      Error:
      ! OAuth failure [1]
      * abc

# userful errors if response isn't parseable

    Code
      oauth_flow_parse(resp1, "test")
    Condition
      Error:
      ! Failed to parse response from `test` OAuth url.
      Caused by error in `resp_body_json()`:
      ! Unexpected content type "text/plain".
      * Expecting type "application/json" or suffix "json".
    Code
      oauth_flow_parse(resp2, "test")
    Condition
      Error:
      ! Failed to parse response from `test` OAuth url.
      * Did not contain `access_token`, `device_code`, or `error` field.

# returns body if known good structure

    Code
      oauth_flow_parse(resp, "test")
    Condition
      Error:
      ! OAuth failure [10]

# httr2_translate() works with basic GET requests

    Code
      httr2_translate(request("https://hb.cran.dev/get"))
    Output
      curl https://hb.cran.dev/get \
        --location \
        --user-agent httr2

# httr2_translate() works with POST methods

    Code
      httr2_translate(req_method(request("https://hb.cran.dev/post"), "POST"))
    Output
      curl https://hb.cran.dev/post \
        --request POST \
        --location \
        --user-agent httr2

# httr2_translate() works with headers

    Code
      httr2_translate(req_headers(request("https://hb.cran.dev/get"), Accept = "application/json",
      `User-Agent` = "httr2/1.0"))
    Output
      curl https://hb.cran.dev/get \
        --header 'Accept: application/json' \
        --header 'User-Agent: httr2/1.0' \
        --location \
        --user-agent httr2

# httr2_translate() works with JSON bodies

    Code
      httr2_translate(req_body_json(request("https://hb.cran.dev/post"), list(name = "test",
        value = 123)))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: application/json' \
        --location \
        --user-agent httr2 \
        --data '{"name":"test","value":123}'

# httr2_translate() works with form bodies

    Code
      httr2_translate(req_body_form(request("https://hb.cran.dev/post"), name = "test",
      value = "123"))
    Output
      curl https://hb.cran.dev/post \
        --location \
        --user-agent httr2 \
        --header 'Content-Type: application/x-www-form-urlencoded' \
        --data 'name=test&value=123'

# httr2_translate() works with multipart bodies

    Code
      httr2_translate(req_body_multipart(request("https://hb.cran.dev/post"), name = "test",
      value = "123"))
    Output
      curl https://hb.cran.dev/post \
        --location \
        --user-agent httr2 \
        --form-string name=test \
        --form-string value=123

# httr2_translate() works with string bodies

    Code
      httr2_translate(req_body_raw(request("https://hb.cran.dev/post"), "test data",
      type = "text/plain"))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: text/plain' \
        --location \
        --user-agent httr2 \
        --data 'test data'

# httr2_translate() works with file bodies

    Code
      httr2_translate(req_body_file(request("https://hb.cran.dev/post"), path, type = "text/plain"))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: text/plain' \
        --location \
        --user-agent httr2 \
        --data-binary @<tempfile>

# httr2_translate() works with custom content types

    Code
      httr2_translate(req_body_json(request("https://hb.cran.dev/post"), list(test = "data"),
      type = "application/vnd.api+json"))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: application/vnd.api+json' \
        --location \
        --user-agent httr2 \
        --data '{"test":"data"}'

# httr2_translate() works with options

    Code
      httr2_translate(req_options(request("https://hb.cran.dev/get"), verbose = TRUE,
      ssl_verifypeer = FALSE))
    Condition
      Warning:
      Can't translate option "ssl_verifypeer".
    Output
      curl https://hb.cran.dev/get \
        --location \
        --verbose \
        --user-agent httr2

# httr2_translate() works with cookies

    Code
      httr2_translate(req_options(request("https://hb.cran.dev/cookies"), cookiejar = cookie_file,
      cookiefile = cookie_file))
    Output
      curl https://hb.cran.dev/cookies \
        --location \
        --cookie-jar <cookie-file> \
        --cookie <cookie-file> \
        --user-agent httr2

# httr2_translate() works with obfuscated values in headers

    Code
      httr2_translate(req_headers(request("https://hb.cran.dev/get"), Authorization = obfuscated(
        "ZdYJeG8zwISodg0nu4UxBhs")))
    Output
      curl https://hb.cran.dev/get \
        --header 'Authorization: <REDACTED>' \
        --location \
        --user-agent httr2

# httr2_translate() can reveal obfuscated values

    Code
      httr2_translate(req_headers_redacted(request("https://hb.cran.dev/get"),
      Authorization = "secret-token"), obfuscated = "reveal")
    Output
      curl https://hb.cran.dev/get \
        --header 'Authorization: secret-token' \
        --location \
        --user-agent httr2

# httr2_translate() works with obfuscated values in JSON body

    Code
      httr2_translate(req_body_json(request("https://hb.cran.dev/post"), list(
        username = "test", password = obfuscated("ZdYJeG8zwISodg0nu4UxBhs"))))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: application/json' \
        --location \
        --user-agent httr2 \
        --data '{"username":"test","password":"<REDACTED>"}'

# httr2_translate() works with obfuscated values in form body

    Code
      httr2_translate(req_body_form(request("https://hb.cran.dev/post"), username = "test",
      password = obfuscated("ZdYJeG8zwISodg0nu4UxBhs")))
    Output
      curl https://hb.cran.dev/post \
        --location \
        --user-agent httr2 \
        --header 'Content-Type: application/x-www-form-urlencoded' \
        --data 'username=test&password=%3CREDACTED%3E'

# httr2_translate() works with complex requests

    Code
      httr2_translate(req_body_json(req_headers(req_method(request(
        "https://api.github.com/user/repos"), "POST"), Accept = "application/vnd.github.v3+json",
      Authorization = obfuscated("ZdYJeG8zwISodg0nu4UxBhs"), `User-Agent` = "MyApp/1.0"),
      list(name = "test-repo", description = "A test repository", private = TRUE)))
    Output
      curl https://api.github.com/user/repos \
        --header 'Accept: application/vnd.github.v3+json' \
        --header 'Authorization: <REDACTED>' \
        --header 'User-Agent: MyApp/1.0' \
        --header 'Content-Type: application/json' \
        --location \
        --user-agent httr2 \
        --data '{"name":"test-repo","description":"A test repository","private":true}'

# httr2_translate() puts a request with no arguments on a single line

    Code
      httr2_translate(req_options(request("https://hb.cran.dev/get"), followlocation = FALSE))
    Output
      curl https://hb.cran.dev/get \
        --user-agent httr2

# httr2_translate() validates input

    Code
      httr2_translate("not a request")
    Condition
      Error in `httr2_translate()`:
      ! `req` must be an HTTP request object, not the string "not a request".

# httr2_translate() errors for raw bodies

    Code
      httr2_translate(req)
    Condition
      Error:
      ! Can't translate a request with a raw body.

# an explicit Content-Type header isn't duplicated by the body

    Code
      httr2_translate(req_body_raw(req_headers(request("https://hb.cran.dev/post"),
      `Content-Type` = "application/json"), "{}"))
    Output
      curl https://hb.cran.dev/post \
        --header 'Content-Type: application/json' \
        --location \
        --user-agent httr2 \
        --data '{}'

# curl_body_data() translates multipart values

    Code
      writeLines(curl_body_data(body, "multipart"))
    Output
      --form-string 'text=@literal;value'
      --form 'file=@"<tmppath>";type=text/plain;filename="name.txt"'
      --form 'data="a b";type=text/plain'

# curl_options() translates each known option

    Code
      cat(curl_options(req), sep = "\n")
    Output
      --location
      --max-time 30
      --connect-timeout 5
      --proxy http://proxy.example.com
      --user-agent agent
      --verbose
      --cookie-jar jar.txt
      --cookie file.txt

# curl_options() translates options set by httr2 functions

    Code
      cat(curl_options(req), sep = "\n")
    Output
      --location
      --max-time 30
      --connect-timeout 0
      --proxy proxy.example.com:8080
      --proxy-user u:p
      --user-agent agent
      --cookie-jar cookies.txt
      --cookie cookies.txt
      --cookie session=abc

# curl_options() warns about untranslatable options

    Code
      out <- curl_options(req)
    Condition
      Warning:
      Can't translate option "ssl_verifypeer".

# validates inputs

    Code
      oauth_flow_bearer_jwt(client1)
    Condition
      Error in `oauth_flow_bearer_jwt()`:
      ! JWT flow requires `client` with a key.

---

    Code
      oauth_flow_bearer_jwt(client2, claim = NULL)
    Condition
      Error in `oauth_flow_bearer_jwt()`:
      ! `claim` must be a list or function.

# must be call to curl

    Code
      curl_translate("echo foo")
    Condition
      Error in `curl_translate()`:
      ! Expecting call to "curl" not to "echo".

# must have cmd argument if non-interactive

    Code
      curl_translate()
    Condition
      Error in `curl_translate()`:
      ! Must supply `cmd`.

# common headers can be removed

    Code
      print(curl_simplify_headers(headers, simplify_headers = TRUE))
    Output
      <httr2_headers>
      Accept: application/vnd.api+json
      user-agent: agent
    Code
      print(curl_simplify_headers(headers, simplify_headers = FALSE))
    Output
      <httr2_headers>
      Sec-Fetch-Dest: empty
      Sec-Fetch-Mode: cors
      sec-ch-ua-mobile: ?0
      Accept: application/vnd.api+json
      referer: ref
      user-agent: agent

# can translate to httr calls

    Code
      curl_translate("curl http://x.com")
    Output
      request("http://x.com/") |>
        req_perform()
    Code
      curl_translate("curl http://x.com -X DELETE")
    Output
      request("http://x.com/") |>
        req_method("DELETE") |>
        req_perform()
    Code
      curl_translate("curl http://x.com -H A:1")
    Output
      request("http://x.com/") |>
        req_headers(
          A = "1",
        ) |>
        req_perform()
    Code
      curl_translate("curl http://x.com -H 'A B:1'")
    Output
      request("http://x.com/") |>
        req_headers(
          `A B` = "1",
        ) |>
        req_perform()
    Code
      curl_translate("curl http://x.com -u u:p")
    Output
      request("http://x.com/") |>
        req_auth_basic("u", "p") |>
        req_perform()
    Code
      curl_translate("curl http://x.com --verbose")
    Output
      request("http://x.com/") |>
        req_perform(verbosity = 1)

# can translate query

    Code
      curl_translate("curl http://x.com?string=abcde&b=2")
    Output
      request("http://x.com/") |>
        req_url_query(
          string = "abcde",
          b = "2",
        ) |>
        req_perform()

# can translate data

    Code
      curl_translate("curl http://example.com --data abcdef")
    Output
      request("http://example.com/") |>
        req_body_raw("abcdef", "application/x-www-form-urlencoded") |>
        req_perform()
    Code
      curl_translate(
        "curl http://example.com --data abcdef -H Content-Type:text/plain")
    Output
      request("http://example.com/") |>
        req_body_raw("abcdef", "text/plain") |>
        req_perform()

# can translate ocokies

    Code
      curl_translate("curl 'http://test' -H 'Cookie: x=1; y=2;z=3'")
    Output
      request("http://test/") |>
        req_cookies_set(
          x = "1",
          y = "2",
          z = "3",
        ) |>
        req_perform()

# can translate json

    Code
      curl_translate(
        "curl http://example.com --data-raw '{\"a\": 1, \"b\": \"text\"}' -H Content-Type:application/json")
    Output
      request("http://example.com/") |>
        req_body_json(
          data = list(a = 1L, b = "text"),
        ) |>
        req_perform()
    Code
      curl_translate("curl http://example.com --json '{\"a\": 1, \"b\": \"text\"}'")
    Output
      request("http://example.com/") |>
        req_body_json(
          data = list(a = 1L, b = "text"),
        ) |>
        req_perform()

# content type stays in header if no data

    Code
      curl_translate("curl http://example.com -H Content-Type:text/plain")
    Output
      request("http://example.com/") |>
        req_headers(
          `Content-Type` = "text/plain",
        ) |>
        req_perform()

# can read from clipboard

    Code
      curl_translate()
    Message
      v Copying to clipboard:
    Output
      request("http://example.com/") |>
        req_headers(
          A = "1",
          B = "2",
        ) |>
        req_perform()
    Code
      writeLines(clipr::read_clip())
    Output
      request("http://example.com/") |>
        req_headers(
          A = "1",
          B = "2",
        ) |>
        req_perform()

# encode_string2() produces simple strings

    Code
      curl_translate(cmd)
    Output
      request("http://example.com/") |>
        req_method("PATCH") |>
        req_body_json(
          data = list(data = list(x = 1L, y = "a", nested = list(z = list(1L, 2L, 3L)))),
        ) |>
        req_perform()

# response has basic print method

    Code
      response(200)
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Body: None
    Code
      response(200, headers = "Content-Type: text/html")
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Content-Type: text/html
      Body: None
    Code
      response(200, body = charToRaw("abcdef"))
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Body: In memory (6 bytes)
    Code
      response(200, body = new_path("path-empty"))
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Body: None
    Code
      response(200, body = new_path("path-content"))
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Body: On disk 'path-content' (15 bytes)
    Code
      response(200, body = streaming)
    Output
      <httr2_response>
      GET https://example.com
      Status: 200 OK
      Body: Streaming connection

# check_response produces helpful error

    Code
      check_response(1)
    Condition
      Error:
      ! `1` must be an HTTP response object, not the number 1.

# new_response() checks its inputs

    Code
      new_response(1)
    Condition
      Error:
      ! `method` must be a single string, not the number 1.
    Code
      new_response("GET", 1)
    Condition
      Error:
      ! `url` must be a single string, not the number 1.
    Code
      new_response("GET", "http://x.com", "x")
    Condition
      Error:
      ! `status_code` must be a whole number, not the string "x".
    Code
      new_response("GET", "http://x.com", 200, 1)
    Condition
      Error:
      ! `headers` must be a list, character vector, or raw.
    Code
      new_response("GET", "http://x.com", 200, list(), 1)
    Condition
      Error:
      ! `body` must be a raw vector, a path, or a <StreamingBody>, not the number 1.
    Code
      new_response("GET", "http://x.com", 200, list(), raw(), "x")
    Condition
      Error:
      ! `timing` must be a numeric vector or `NULL`, not the string "x".
    Code
      new_response("GET", "http://x.com", 200, list(), raw(), c(x = 1), 1)
    Condition
      Error:
      ! `request` must be an HTTP request object or `NULL`, not the number 1.

# can request verbose record of request

    Code
      . <- req_perform(verbose_resp)
    Output
      <- HTTP/1.1 200 OK
      <- Connection: close
      <- Content-Type: application/json
      <- 
      << {
      <<   "x": 1
      << }

---

    Code
      . <- req_perform(verbose_req)
    Output
      -> POST /test HTTP/1.1
      -> Host: http://example.com
      -> Content-Type: text/plain
      -> Content-Length: 17
      -> 
      >> This is some text

# redacts headers as needed

    Code
      . <- req_perform(req)
    Output
      -> GET / HTTP/1.1
      -> Host: http://example.com
      -> Authorization: <REDACTED>
      -> 

# can display compressed bodies

    Code
      . <- req_perform(req)
    Output
      <- HTTP/1.1 200 OK
      <- Content-Type: application/json
      <- Content-Encoding: gzip
      <- 
      << {
      <<   "args": {
      << 
      <<   },
      <<   "data": {
      << 
      <<   },
      <<   "files": {
      << 
      <<   },
      <<   "form": {
      << 
      <<   },
      <<   "headers": {
      <<     "Host": "http://example.com"
      <<   },
      <<   "json": {
      << 
      <<   },
      <<   "method": "get",
      <<   "path": "/gzip",
      <<   "origin": "127.0.0.1",
      <<   "url": "<webfakes>/gzip",
      <<   "gzipped": true
      << }

# response json is automatically prettified

    Code
      . <- req_perform(req)
    Output
      << {
      <<   "foo": "bar",
      <<   "baz": [
      <<     1,
      <<     2,
      <<     3
      <<   ]
      << }

---

    Code
      . <- req_perform(req)
    Output
      << {"foo":"bar","baz":[1,2,3]}

# request json is automatically prettified

    Code
      . <- req_perform(req)
    Output
      >> {
      >>   "foo": "bar",
      >>   "baz": [
      >>     1,
      >>     2,
      >>     3
      >>   ]
      >> }

---

    Code
      . <- req_perform(req)
    Output
      >> {"foo":"bar","baz":[1,2,3]}

# throttling affects request performance

    Code
      time <- system.time(req_perform(req))[[3]]
    Message
      > Waiting 0.15s for throttling delay

# req_throttle checks its inputs

    Code
      req_throttle(request_test(), capacity = "x")
    Condition
      Error in `req_throttle()`:
      ! `capacity` must be a numeric vector, not the string "x".
    Code
      req_throttle(request_test(), capacity = -1)
    Condition
      Error in `req_throttle()`:
      ! Every element of `capacity` must be 0 or greater.
    Code
      req_throttle(request_test(), capacity = 1.5)
    Condition
      Error in `req_throttle()`:
      ! Every element of `capacity` must be a whole number.
    Code
      req_throttle(request_test(), capacity = c(1, 2), fill_time_s = c(1, 2, 3))
    Condition
      Error in `req_throttle()`:
      ! Can't recycle `capacity` (size 2) to match `fill_time_s` (size 3).

# user temination still returns data

    Code
      resps <- req_perform_iterative(req, next_req = next_req)
    Message
      ! Terminating iteration; returning 1 response.

# checks its inputs

    Code
      req_perform_iterative(1)
    Condition
      Error in `req_perform_iterative()`:
      ! `req` must be an HTTP request object, not the number 1.
    Code
      req_perform_iterative(req, function(x, y) x + y)
    Condition
      Error in `req_perform_iterative()`:
      ! `next_req` must have the arguments `resp` and `req`; it currently has `x` and `y`.
    Code
      req_perform_iterative(req, function(resp, req) { }, path = 1)
    Condition
      Error in `req_perform_iterative()`:
      ! `path` must be a single string or `NULL`, not the number 1.
    Code
      req_perform_iterative(req, function(resp, req) { }, max_reqs = -1)
    Condition
      Error in `req_perform_iterative()`:
      ! `max_reqs` must be a whole number larger than or equal to 1, not the number -1.
    Code
      req_perform_iterative(req, function(resp, req) { }, progress = -1)
    Condition
      Error in `req_perform_iterative()`:
      ! `progress` must be a bool, a string, or a list, not the number -1.

# can print all url details

    Code
      url_parse("http://user:pass@example.com:81/path?a=1&b=2&c={1{2}3}#frag")
    Output
      <httr2_url> http://user:pass@example.com:81/path?a=1&b=2&c=%7B1%7B2%7D3%7D#frag
      * scheme: http
      * hostname: example.com
      * username: user
      * password: pass
      * port: 81
      * path: /path
      * query:
        * a: 1
        * b: 2
        * c: {1{2}3}
      * fragment: frag

# url_build validates its input

    Code
      url_build("abc")
    Condition
      Error in `url_build()`:
      ! `url` must be a parsed URL, not the string "abc".

# url_modify checks its inputs

    Code
      url_modify(1)
    Condition
      Error in `url_modify()`:
      ! `url` must be a string or parsed URL, not the number 1.
    Code
      url_modify(url, scheme = 1)
    Condition
      Error in `url_modify()`:
      ! `scheme` must be a single string or `NULL`, not the number 1.
    Code
      url_modify(url, hostname = 1)
    Condition
      Error in `url_modify()`:
      ! `hostname` must be a single string or `NULL`, not the number 1.
    Code
      url_modify(url, port = "x")
    Condition
      Error in `url_modify()`:
      ! `port` must be a whole number or `NULL`, not the string "x".
    Code
      url_modify(url, port = -1)
    Condition
      Error in `url_modify()`:
      ! `port` must be a whole number between 1 and 65535 or `NULL`, not the number -1.
    Code
      url_modify(url, username = 1)
    Condition
      Error in `url_modify()`:
      ! `username` must be a single string or `NULL`, not the number 1.
    Code
      url_modify(url, password = 1)
    Condition
      Error in `url_modify()`:
      ! `password` must be a single string or `NULL`, not the number 1.
    Code
      url_modify(url, path = 1)
    Condition
      Error in `url_modify()`:
      ! `path` must be a single string or `NULL`, not the number 1.
    Code
      url_modify(url, fragment = 1)
    Condition
      Error in `url_modify()`:
      ! `fragment` must be a single string or `NULL`, not the number 1.

# checks various query formats

    Code
      url_modify(url, query = 1)
    Condition
      Error in `url_modify()`:
      ! `query` must be a character vector, named list, or NULL, not the number 1.
    Code
      url_modify(url, query = list(1))
    Condition
      Error in `url_modify()`:
      ! `query` must be a character vector, named list, or NULL, not a list.
    Code
      url_modify(url, query = list(x = 1:2))
    Condition
      Error in `url_modify()`:
      ! Query value `query$x` must be a length-1 atomic vector, not an integer vector.

# validates inputs

    Code
      url_modify_query(1)
    Condition
      Error in `url_modify_query()`:
      ! `.url` must be a string or parsed URL, not the number 1.
    Code
      url_modify_query(url, 1)
    Condition
      Error in `url_modify_query()`:
      ! All components of `...` must be named.
    Code
      url_modify_query(url, x = 1:2)
    Condition
      Error in `url_modify_query()`:
      ! All vector elements of `...` must be length 1.
      i Use `.multi` to choose a strategy for handling vectors.

---

    Code
      url_query_build(1:3)
    Condition
      Error in `url_query_build()`:
      ! `query` must be a named list, not an integer vector.
    Code
      url_query_build(list(x = 1:2, y = 1:3))
    Condition
      Error in `url_query_build()`:
      ! All vector elements of `query` must be length 1.
      i Use `.multi` to choose a strategy for handling vectors.

# can't opt out of escaping non strings

    Code
      format_query_param(I(1), "x")
    Condition
      Error:
      ! Escaped query value `x` must be a single string, not the number 1.

# validates issuer against the returned document

    Code
      oauth_server_metadata("https://example.com")
    Condition
      Error in `oauth_server_metadata()`:
      ! Metadata issuer doesn't match the requested `issuer`.
      * Requested "https://example.com".
      * Received "https://evil.com".

# requires either issuer or url

    Code
      oauth_server_metadata()
    Condition
      Error in `oauth_server_metadata()`:
      ! Must supply either `issuer` or `url`.

# has a useful print method

    Code
      meta
    Output
      <httr2_oauth_server_metadata>
      * issuer                : "https://example.com"
      * authorization_endpoint: "https://example.com/authorize"
      * token_endpoint        : "https://example.com/token"
      * and 3 more fields.

# checks its inputs

    Code
      req_perform_promise(1)
    Condition
      Error in `req_perform_promise()`:
      ! `req` must be an HTTP request object, not the number 1.
    Code
      req_perform_promise(req, path = 1)
    Condition
      Error in `req_perform_promise()`:
      ! `path` must be a single string or `NULL`, not the number 1.
    Code
      req_perform_promise(req, pool = "INVALID")
    Condition
      Error in `req_perform_promise()`:
      ! `pool` must be a {curl} pool or `NULL`, not the string "INVALID".
    Code
      req_perform_promise(req, verbosity = "INVALID")
    Condition
      Error in `req_perform_promise()`:
      ! `verbosity` must 0, 1, 2, or 3.

# correctly prepares request

    Code
      . <- extract_promise(req_perform_promise(req, verbosity = 1))
    Output
      -> GET /get HTTP/1.1
      -> Host: <variable>
      -> User-Agent: <variable>
      -> Accept: */*
      -> Accept-Encoding: <variable>
      -> 
      <- HTTP/1.1 200 OK
      <- Date: <variable>
      <- Content-Type: application/json
      <- Content-Length: <variable>
      <- ETag: <variable>
      <- 

# req_perform_promise uses the default loop

    Code
      p4 <- req_perform_promise(request_test("/get"))
    Condition
      Error in `req_perform_promise()`:
      ! Must supply `pool` when calling `later::with_temp_loop()`.
      i Do you need `pool = curl::new_pool()`?

# desktop style can't run in hosted environment

    Code
      oauth_flow_auth_code(client, "http://localhost")
    Condition
      Error in `oauth_flow_auth_code()`:
      ! Can't use localhost `redirect_uri` in a hosted environment.

# forwards oauth error

    Code
      oauth_flow_auth_code_parse(query1, "abc")
    Condition
      Error in `oauth_flow_auth_code_parse()`:
      ! OAuth failure [123]
      * A bad error
    Code
      oauth_flow_auth_code_parse(query2, "abc")
    Condition
      Error in `oauth_flow_auth_code_parse()`:
      ! OAuth failure [123]
      * A bad error
      i Learn more at <http://example.com>.
    Code
      oauth_flow_auth_code_parse(query3, "abc")
    Condition
      Error in `oauth_flow_auth_code_parse()`:
      ! Authentication failure: state does not match.

# can read/write cookies

    Code
      readLines(cookie_path)[-(1:4)]
    Output
      [1] "127.0.0.1\tFALSE\t/\tFALSE\t0\tz\tc" "127.0.0.1\tFALSE\t/\tFALSE\t0\tx\ta"
      [3] "127.0.0.1\tFALSE\t/\tFALSE\t0\ty\tb"

# can get json response

    Code
      last_response_json()
    Output
      {
        "x": 1
      }
    Code
      last_response_json(pretty = FALSE)
    Output
      {"x":1}

# can get json request

    Code
      last_request_json()
    Output
      {
        "x": 1
      }
    Code
      last_request_json(pretty = FALSE)
    Output
      {"x":1}

# useful errors if not json request/response

    Code
      last_request_json()
    Condition
      Error in `last_request_json()`:
      ! Last request doesn't have a JSON body.
    Code
      last_response_json()
    Condition
      Error in `last_response_json()`:
      ! Last response doesn't have a JSON body.

# useful errors if no last request/response

    Code
      last_request_json()
    Condition
      Error in `last_request_json()`:
      ! No request has been made yet.
    Code
      last_response_json()
    Condition
      Error in `last_response_json()`:
      ! No request has been made successfully yet.

# validates inputs

    Code
      req_timeout(request_test(), "x")
    Condition
      Error in `req_timeout()`:
      ! `seconds` must be a number, not the string "x".
    Code
      req_timeout(request_test(), 0)
    Condition
      Error in `req_timeout()`:
      ! `seconds` must be >1 ms.

# req_proxy gives helpful errors

    Code
      req_proxy(req, port = "abc")
    Condition
      Error in `req_proxy()`:
      ! `port` must be a whole number or `NULL`, not the string "abc".
    Code
      req_proxy(req, "abc", auth = "bsc")
    Condition
      Error in `req_proxy()`:
      ! `auth` must be one of "basic", "digest", "gssnegotiate", "ntlm", "digest_ie", or "any", not "bsc".
      i Did you mean "basic"?

# cache emits useful debugging info

    Code
      # Immutable
      invisible(cache_pre_fetch(req))
    Message
      Pruning cache
    Code
      invisible(cache_post_fetch(req, resp))
    Message
      Saving response to cache "f3805db63ff822b4743f247cfdde10a3"
    Code
      invisible(cache_pre_fetch(req))
    Message
      Found url in cache "f3805db63ff822b4743f247cfdde10a3"
      Cached value is fresh; using response from cache

---

    Code
      # freshness check
      invisible(cache_pre_fetch(req))
    Message
      Pruning cache
      Found url in cache "f3805db63ff822b4743f247cfdde10a3"
      Cached value is stale; checking for updates
    Code
      invisible(cache_post_fetch(req, response(304)))
    Message
      Cached value still ok; retrieving body from cache
    Code
      invisible(cache_post_fetch(req, error_cnd()))
    Message
      Request errored; retrieving response from cache

# can prune by number

    Code
      cache_prune(path, list(n = 1, age = Inf, size = Inf), debug = TRUE)
    Message
      Deleted 3 files that are too numerous

# can prune by age

    Code
      cache_prune(path, list(n = Inf, age = 30, size = Inf), debug = TRUE)
    Message
      Deleted 1 file that is too old

# can prune by size

    Code
      cache_prune(path, list(n = Inf, age = Inf, size = 50), debug = TRUE)
    Message
      Deleted 2 files that are too big

# request and paths must match

    Code
      req_perform_parallel(req, letters)
    Condition
      Error in `req_perform_parallel()`:
      ! If supplied, `paths` must be the same length as `req`.

# req_perform_parallel respects http_error() body message

    Code
      req_perform_parallel(reqs)
    Condition
      Error in `req_perform_parallel()`:
      ! HTTP 404 Not Found.
      i hello

# obfuscated strings are hidden

    Code
      x <- obfuscated("abcdef")
      x
    Output
      obfuscated("abcdef")
    Code
      str(x)
    Output
       obfuscated("abcdef")

# can coerce to a key

    Code
      as_key("ENVVAR_THAT_DOESNT_EXIST")
    Condition
      Error:
      ! Can't find envvar ENVVAR_THAT_DOESNT_EXIST
    Code
      as_key(1)
    Condition
      Error:
      ! `key` must be a raw vector containing the key, a string giving the name of an env var, or a string wrapped in `I()` that contains the base64url encoded key.

# validates inputs

    Code
      req_perform_connection(1)
    Condition
      Error in `req_perform_connection()`:
      ! `req` must be an HTTP request object, not the number 1.
    Code
      req_perform_connection(request_test(), 1)
    Condition
      Error in `req_perform_connection()`:
      ! `blocking` must be `TRUE` or `FALSE`, not the number 1.

# curl errors become errors

    Code
      req_perform_connection(req)
    Condition
      Error in `req_perform_connection()`:
      ! Failed to perform HTTP request.
      Caused by error in `open()`:
      ! Failed to connect

# correclty reports curl error with retries (#817)

    Code
      req_perform_connection(req)
    Condition
      Error in `req_perform_connection()`:
      ! Failed to perform HTTP request.
      Caused by error in `open()`:
      ! Failed to connect

# validates its input

    Code
      StreamingBody$new(1)
    Condition
      Error in `StreamingBody$new()`:
      ! `conn` must be a connection, not the number 1.

# iterate_with_offset checks inputs

    Code
      iterate_with_offset(1)
    Condition
      Error in `iterate_with_offset()`:
      ! `param_name` must be a single string, not the number 1.
    Code
      iterate_with_offset("x", "x")
    Condition
      Error in `iterate_with_offset()`:
      ! `start` must be a whole number, not the string "x".
    Code
      iterate_with_offset("x", offset = 0)
    Condition
      Error in `iterate_with_offset()`:
      ! `offset` must be a whole number larger than or equal to 1, not the number 0.
    Code
      iterate_with_offset("x", offset = "x")
    Condition
      Error in `iterate_with_offset()`:
      ! `offset` must be a whole number, not the string "x".
    Code
      iterate_with_offset("x", resp_complete = function(x, y) x + y)
    Condition
      Error in `iterate_with_offset()`:
      ! `resp_complete` must have the argument `resp`; it currently has `x` and `y`.

# iterate_with_cursor

    Code
      iterate_with_cursor(1)
    Condition
      Error in `iterate_with_cursor()`:
      ! `param_name` must be a single string, not the number 1.
    Code
      iterate_with_cursor("x", function(x, y) x + y)
    Condition
      Error in `iterate_with_cursor()`:
      ! `resp_param_value` must have the argument `resp`; it currently has `x` and `y`.

# iterate_with_link_url checks its inputs

    Code
      iterate_with_link_url(rel = 1)
    Condition
      Error in `iterate_with_link_url()`:
      ! `rel` must be a single string, not the number 1.

# as_headers errors on invalid types

    Code
      as_headers(1)
    Condition
      Error:
      ! `headers` must be a list, character vector, or raw.

# has nice print method

    Code
      as_headers(c("X:1", "Y: 2", "Z:"))
    Output
      <httr2_headers>
      X: 1
      Y: 2
      Z: 
    Code
      as_headers(list())
    Output
      <httr2_headers>

# print and str redact headers

    Code
      print(x)
    Output
      <httr2_headers>
      x: <REDACTED>
      y: 2
    Code
      str(x)
    Output
       <httr2_headers>
       $ x: <REDACTED>
       $ y: chr "2"

# new_headers checks inputs

    Code
      new_headers(1)
    Condition
      Error:
      ! `x` must be a list.
    Code
      new_headers(list(1))
    Condition
      Error:
      ! All elements of `x` must be named.
    Code
      new_headers(list(x = mean))
    Condition
      Error:
      ! Each element of `x` must be an atomic vector or a weakref.
      i `x[[1]]` is a function.

# checks its inputs

    Code
      multi_dots(1)
    Condition
      Error:
      ! All components of `...` must be named.
    Code
      multi_dots(x = I(1))
    Condition
      Error:
      ! Escaped query value `x` must be a single string, not the number 1.
    Code
      multi_dots(x = 1:2)
    Condition
      Error:
      ! All vector elements of `...` must be length 1.
      i Use `.multi` to choose a strategy for handling vectors.
    Code
      multi_dots(x = mean)
    Condition
      Error:
      ! All elements of `...` must be either an atomic vector or NULL.

# unknown header triggers error

    Code
      parse_aws_event(aws_event(aws_header("foo", "unknown")))
    Condition
      Error in `type_enum()`:
      ! Unsupported type 255.
      i This is an internal error that was detected in the httr2 package.
        Please report it at <https://github.com/r-lib/httr2/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# parse_aws_event() checks the prelude length

    Code
      parse_aws_event(as.raw(1:10))
    Condition
      Error in `parse_aws_event()`:
      ! AWS event metadata doesn't match supplied bytes
      i This is an internal error that was detected in the httr2 package.
        Please report it at <https://github.com/r-lib/httr2/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# verbosity = 3 shows aws events

    Code
      . <- resp_stream_aws(resp)
    Output
      *  -- Buffer ----------------------------------------------------------------------
      *  Received chunk: 00 00 00 1a 00 00 00 0a af a7 b9 54 03 66 6f 6f 07 00 03 62 61 72 c0 c0 6d f5
      << foo: bar
      << ""
      

# empty body generates error

    Code
      resp_body_raw(resp1)
    Condition
      Error in `resp_body_raw()`:
      ! Can't retrieve empty body.

---

    Code
      resp_body_raw(resp2)
    Condition
      Error in `resp_body_raw()`:
      ! Can't retrieve empty body.

# check argument types before caching

    Code
      resp_body_json(1)
    Condition
      Error in `resp_body_json()`:
      ! `resp` must be an HTTP response object, not the number 1.
    Code
      resp_body_xml(1)
    Condition
      Error in `resp_body_xml()`:
      ! `resp` must be an HTTP response object, not the number 1.

# content types are checked

    Code
      resp_body_json(req_perform(request_test("/xml")))
    Condition
      Error in `resp_body_json()`:
      ! Unexpected content type "application/xml".
      * Expecting type "application/json" or suffix "json".
    Code
      resp_body_xml(req_perform(request_test("/json")))
    Condition
      Error in `resp_body_xml()`:
      ! Unexpected content type "application/json".
      * Expecting type "application/xml" or "text/xml" or suffix "xml".

# can't change body type

    Code
      req_body_json(req, list(x = 1))
    Condition
      Error in `req_body_json()`:
      ! Can't change body type from raw to json.
      i You must use only one type of `req_body_*()` per request.

# can't send anything else

    Code
      req_body_raw(req, 1)
    Condition
      Error in `req_body_raw()`:
      ! `body` must be a raw vector or string.

# errors on invalid input

    Code
      req_body_file(request_test(), 1)
    Condition
      Error in `req_body_file()`:
      ! `path` must be a single string, not the number 1.
    Code
      req_body_file(request_test(), "doesntexist")
    Condition
      Error in `req_body_file()`:
      ! Can't find file 'doesntexist'.
    Code
      req_body_file(request_test(), ".")
    Condition
      Error in `req_body_file()`:
      ! `path` must be a file, not a directory.

# non-json type errors

    Code
      req_body_json(request_test(), mtcars, type = "application/xml")
    Condition
      Error in `req_body_json()`:
      ! Unexpected content type "application/xml".
      * Expecting type "application/json" or suffix "json".

# can't modify non-json data

    Code
      req_body_json_modify(req, a = 1)
    Condition
      Error in `req_body_json_modify()`:
      ! Can only be used after `req_body_json()`.

# checks its inputs

    Code
      req_perform_sequential(req)
    Condition
      Error in `req_perform_sequential()`:
      ! `reqs` must be a list, not a <httr2_request> object.
    Code
      req_perform_sequential(list(req), letters)
    Condition
      Error in `req_perform_sequential()`:
      ! If supplied, `paths` must be the same length as `req`.

# can check app has needed pieces

    Code
      oauth_flow_check("test", NULL)
    Condition
      Error:
      ! `client` must be an OAuth client created with `oauth_client()`.
    Code
      oauth_flow_check("test", client, is_confidential = TRUE)
    Condition
      Error:
      ! Can't use this `app` with OAuth 2.0 test flow.
      i `app` must have a confidential client (i.e. `client_secret` is required).
    Code
      oauth_flow_check("test", client, interactive = TRUE)
    Condition
      Error:
      ! OAuth 2.0 test flow requires an interactive session

# checks auth types have needed args

    Code
      oauth_client("abc", "http://x.com", auth = "header")
    Condition
      Error in `oauth_client()`:
      ! `auth = 'header'` requires a `secret`.
    Code
      oauth_client("abc", "http://x.com", auth = "jwt_sig")
    Condition
      Error in `oauth_client()`:
      ! `auth = 'jwt_sig'` requires a `key`.
    Code
      oauth_client("abc", "http://x.com", key = "abc", auth = "jwt_sig")
    Condition
      Error in `oauth_client()`:
      ! `auth = 'jwt_sig'` requires a claim specification in `auth_params`.
    Code
      oauth_client("abc", "http://x.com", auth = 123)
    Condition
      Error in `oauth_client()`:
      ! `auth` must be a string or function.

# client has useful print method

    Code
      oauth_client("x", url)
    Output
      <httr2_oauth_client>
      * name     : "bf27508f7925b06bf28a10f3805351ab"
      * id       : "x"
      * token_url: "http://example.com"
      * auth     : "oauth_client_req_auth_body"
    Code
      oauth_client("x", url, secret = "SECRET")
    Output
      <httr2_oauth_client>
      * name     : "bf27508f7925b06bf28a10f3805351ab"
      * id       : "x"
      * secret   : <REDACTED>
      * token_url: "http://example.com"
      * auth     : "oauth_client_req_auth_body"
    Code
      oauth_client("x", url, auth = function(...) {
        xxx
      })
    Output
      <httr2_oauth_client>
      * name     : "bf27508f7925b06bf28a10f3805351ab"
      * id       : "x"
      * token_url: "http://example.com"
      * auth     : <function>

# get some useful output from WWW-Authenticate header

    HTTP 401 Unauthorized.
    * OAuth error: invalid_token - The access token expired
    * realm: example

---

    HTTP 403 Forbidden.
    * OAuth error: insufficient_scope
    * realm: https://accounts.google.com/
    * scope: https://www.googleapis.com/auth/iam https://www.googleapis.com/auth/cloud-platform

# clear error if body type is unsupported

    Code
      req_perform(req)
    Condition
      Error in `req_auth_aws_v4()`:
      ! Unsupported body type: multipart

# aws_v4_signature calculates correct signature

    Code
      signature
    Output
      $CanonicalRequest
      [1] "POST\n/v0/\n\nhost:example.execute-api.us-east-1.amazonaws.com\n\nhost\ne3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
      
      $string_to_sign
      [1] "AWS4-HMAC-SHA256\n20250121T182222Z\n20250121/us-east-1/execute-api/aws4_request\n40b845bd8e6a316382ca9f73516e236075e4af2e04ebcb5f0d8eff12a040f6a4"
      
      $SigningKey
      sha256 hmac 56:b5:cd:6c:f3:27:c6:2c:ba:96:96:4c:45:3b:10:aa:97:11:7f:3d:dc:20:c0:58:d9:c4:07:7e:07:eb:63:58 
      
      $Authorization
      [1] "AWS4-HMAC-SHA256 Credential=AKIAIOSFODNN7EXAMPLE/20250121/us-east-1/execute-api/aws4_request,SignedHeaders=host,Signature=db2ef1ec9fd9efa801b8eb6b3e754d9d2d5d46189833a947c5427dd706f9534c"
      

# validates its inputs

    Code
      req_auth_aws_v4(1)
    Condition
      Error in `req_auth_aws_v4()`:
      ! `req` must be an HTTP request object, not the number 1.
    Code
      req_auth_aws_v4(req, 1)
    Condition
      Error in `req_auth_aws_v4()`:
      ! `aws_access_key_id` must be a single string, not the number 1.
    Code
      req_auth_aws_v4(req, "", "", aws_session_token = 1)
    Condition
      Error in `req_auth_aws_v4()`:
      ! `aws_session_token` must be a single string or `NULL`, not the number 1.
    Code
      req_auth_aws_v4(req, "", "", aws_service = 1)
    Condition
      Error in `req_auth_aws_v4()`:
      ! `aws_service` must be a single string or `NULL`, not the number 1.
    Code
      req_auth_aws_v4(req, "", "", aws_region = 1)
    Condition
      Error in `req_auth_aws_v4()`:
      ! `aws_region` must be a single string or `NULL`, not the number 1.

# req has basic print method

    Code
      req <- request("https://example.com")
      req
    Output
      <httr2_request>
      GET https://example.com
      Body: empty
    Code
      req_body_raw(req, "Test")
    Output
      <httr2_request>
      POST https://example.com
      Body: a string
    Code
      req_body_multipart(req, Test = 1)
    Output
      <httr2_request>
      POST https://example.com
      Body: multipart data

# printing headers works with {}

    Code
      req_headers(request("http://test"), x = "{z}", `{z}` = "x")
    Output
      <httr2_request>
      GET http://test
      Headers:
      * x  : "{z}"
      * {z}: "x"
      Body: empty

# individually prints repeated headers

    Code
      req_headers(request("https://example.com"), A = 1:3)
    Output
      <httr2_request>
      GET https://example.com
      Headers:
      * A: "1,2,3"
      Body: empty

# check_request() gives useful error

    Code
      check_request(1)
    Condition
      Error:
      ! `1` must be an HTTP request object, not the number 1.

# bad inputs get clear error

    Code
      req_headers(req, fun = mean)
    Condition
      Error in `req_headers()`:
      ! All elements of `...` must be either an atomic vector or NULL.
    Code
      req_headers(req, 1)
    Condition
      Error in `req_headers()`:
      ! All components of `...` must be named.

# is case insensitive

    Code
      req
    Output
      <httr2_request>
      GET http://example.com
      Headers:
      * a: <REDACTED>
      Body: empty

# checks input types

    Code
      req_headers(req, a = 1L, b = 2L, .redact = 1L)
    Condition
      Error in `req_headers()`:
      ! `.redact` must be a character vector or `NULL`, not the number 1.

# req_oauth validates and stores expiry margin

    Code
      req_oauth(req, "", list(), NULL, expiry_margin = -1)
    Condition
      Error in `req_oauth()`:
      ! `expiry_margin` must be a whole number larger than or equal to 0, not the number -1.

# can store on disk

    Code
      cache$set(1)
    Message
      Caching httr2 token in '<oauth-cache-path>/httr2-test/2c0a8a99dc147d5445c3b49d035665b2-token.rds.enc'.

# oauth_cache_prune() validates its input

    Code
      oauth_cache_prune("x")
    Condition
      Error in `oauth_cache_prune()`:
      ! `max_age_days` must be a whole number, not the string "x".

# new token computes expires_at

    Code
      token
    Output
      <httr2_token>
      * token_type  : "bearer"
      * access_token: <REDACTED>
      * expires_at  : "2025-02-19 21:20:10"

# printing token redacts access, id and refresh token

    Code
      oauth_token(access_token = "secret", refresh_token = "secret", id_token = "secret")
    Output
      <httr2_token>
      * token_type   : "bearer"
      * access_token : <REDACTED>
      * refresh_token: <REDACTED>
      * id_token     : <REDACTED>

# can control space handling

    Code
      req_url_query(req, a = " ", .space = "bar")
    Condition
      Error in `multi_dots()`:
      ! `.space` must be one of "percent" or "form", not "bar".

# can handle multi query params

    Code
      req_url_query_multi("error")
    Condition
      Error in `url_modify_query()`:
      ! All vector elements of `...` must be length 1.
      i Use `.multi` to choose a strategy for handling vectors.

# errors are forwarded correctly

    Code
      req_url_query(req, 1)
    Condition
      Error in `url_modify_query()`:
      ! All components of `...` must be named.
    Code
      req_url_query(req, a = I(1))
    Condition
      Error in `url_modify_query()`:
      ! Escaped query value `a` must be a single string, not the number 1.
    Code
      req_url_query(req, a = 1:2)
    Condition
      Error in `url_modify_query()`:
      ! All vector elements of `...` must be length 1.
      i Use `.multi` to choose a strategy for handling vectors.
    Code
      req_url_query(req, a = mean)
    Condition
      Error in `url_modify_query()`:
      ! All elements of `...` must be either an atomic vector or NULL.

# resp_stream_lines(warn) is deprecated unless FALSE

    Code
      . <- resp_stream_lines(resp, warn = TRUE)
    Condition
      Warning:
      The `warn` argument of `resp_stream_lines()` is deprecated as of httr2 1.2.3.

# body is shown

    Code
      req_dry_run(req_utf8)
    Output
      POST / HTTP/1.1
      accept: */*
      content-type: text/plain
      
      Cenário

---

    Code
      req_dry_run(req_json)
    Output
      POST / HTTP/1.1
      accept: */*
      content-type: application/json
      
      {
        "x": 1,
        "y": true
      }

---

    Code
      req_dry_run(req_json, pretty_json = FALSE)
    Output
      POST / HTTP/1.1
      accept: */*
      content-type: application/json
      
      {"x":1,"y":true}

---

    Code
      req_dry_run(req_binary)
    Output
      POST / HTTP/1.1
      accept: */*
      
      <8 bytes>

# authorization headers are redacted

    Code
      out <- req_dry_run(req)
    Output
      GET / HTTP/1.1
      accept: */*
      authorization: <REDACTED>
      

---

    Code
      out <- req_dry_run(req, redact_headers = FALSE)
    Output
      GET / HTTP/1.1
      accept: */*
      authorization: Basic dXNlcjpwYXNzd29yZA==
      

# can determine if incomplete data is complete

    Code
      expect_equal(resp_stream_sse(con), NULL)
    Condition
      Warning:
      Premature end of input; ignoring final partial chunk

# verbosity = 3 shows raw sse events

    Code
      . <- resp_stream_sse(resp)
    Output
      *  -- Buffer ----------------------------------------------------------------------
      *  Received chunk: 3a 20 63 6f 6d 6d 65 6e 74 0a 0a 64 61 74 61 3a 20 31 0a 0a
      *  -- Raw server sent event -------------------------------------------------------
      *  : comment
      *  
      *  
      *  -- Raw server sent event -------------------------------------------------------
      *  data: 1
      *  
      *  
      << type: message
      << data: 1
      << id: 
      

# can't read from a closed connection

    Code
      resp_stream_raw(resp, 1)
    Condition
      Error in `resp_stream_raw()`:
      ! `resp` has already been closed.

# streaming functions require a streaming response

    Code
      resp_stream_raw(response())
    Condition
      Error in `resp_stream_raw()`:
      ! `resp` must be a streaming HTTP response object, not a <httr2_response> object.

# resp_stream_raw() validates kb

    Code
      resp_stream_raw(resp, kb = -1)
    Condition
      Error in `resp_stream_raw()`:
      ! `kb` must be a number larger than or equal to 0, not the number -1.

---

    Code
      resp_stream_raw(resp, kb = Inf)
    Condition
      Error in `resp_stream_raw()`:
      ! `kb` must be a number, not `Inf`.

# resp_stream_is_complete() requires an open streaming response

    Code
      resp_stream_is_complete(response())
    Condition
      Error in `resp_stream_is_complete()`:
      ! `resp` must be a streaming HTTP response object, not a <httr2_response> object.

---

    Code
      resp_stream_is_complete(resp)
    Condition
      Error in `resp_stream_is_complete()`:
      ! `resp` has already been closed.

# streaming responses use only one reader

    Code
      resp_stream_raw(resp)
    Condition
      Error in `resp_stream_raw()`:
      ! Can't use resp_stream_raw() after resp_stream_lines() on the same response.

# StreamSplitter$finish() discards a trailing partial block with a warning

    Code
      out <- s$finish(charToRaw("b"))
    Condition
      Warning:
      Premature end of input; ignoring final partial chunk

# verbosity = 3 logs the buffered chunk

    Code
      while (!resp_stream_is_complete(con)) {
        resp_stream_lines(con, 1)
      }
    Output
      *  -- Buffer ----------------------------------------------------------------------
      *  Received chunk: 6c 69 6e 65 20 31 0a 6c 69 6e 65 20 32 0a
      << line 1
      << line 2

# verbosity = 2 streams request bodies

    Code
      stream_all(req, resp_stream_lines, 1)
    Output
      << line 1
      << line 2
    Code
      stream_all(req, resp_stream_raw, 5 / 1024)
    Output
      << Streamed 5 bytes
      
      << Streamed 5 bytes
      
      << Streamed 4 bytes
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content-type.R
\name{resp_check_content_type}
\alias{resp_check_content_type}
\title{Check the content type of a response}
\usage{
resp_check_content_type(
  resp,
  valid_types = NULL,
  valid_suffix = NULL,
  check_type = TRUE,
  call = caller_env()
)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{valid_types}{A character vector of valid MIME types. Should only
be specified with \code{type/subtype}.}

\item{valid_suffix}{A string given an "structured media type" suffix.}

\item{check_type}{Should the type actually be checked? Provided as a
convenience for when using this function inside \verb{resp_body_*} helpers.}

\item{call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
Called for its side-effect; erroring if the response does not
have the expected content type.
}
\description{
A different content type than expected often leads to an error in parsing
the response body. This function checks that the content type of the response
is as expected and fails otherwise.
}
\examples{
resp <- response(headers = list(`content-type` = "application/json"))
resp_check_content_type(resp, "application/json")
try(resp_check_content_type(resp, "application/xml"))

# `types` can also specify multiple valid types
resp_check_content_type(resp, c("application/xml", "application/json"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/last.R
\name{last_response}
\alias{last_response}
\alias{last_request}
\alias{last_request_json}
\alias{last_response_json}
\title{Retrieve most recent request/response}
\usage{
last_response()

last_request()

last_request_json(pretty = TRUE)

last_response_json(pretty = TRUE)
}
\arguments{
\item{pretty}{Should the JSON be pretty-printed?}
}
\value{
\code{last_request()} and \code{last_response()} return an HTTP
\link{request} or \link{response} respectively. If no request has been made,
\code{last_request()} will return \code{NULL}; if no request has been made
or the last request was unsuccessful, \code{last_response()} will return
\code{NULL}.

\code{last_request_json()} and \code{last_response_json()} always return a string.
They will error if \code{last_request()} or \code{last_response()} are \code{NULL} or
don't have JSON bodies.
}
\description{
\code{last_request()} and \code{last_response()} retrieve the most recent request
made by httr2 and the response it received, to facilitate debugging problems
\emph{after} they occur.

\code{last_request_json()} and \code{last_response_json()} return the JSON bodies of
the most recent request and response. They will error if not JSON.
}
\examples{
. <- request("http://httr2.r-lib.org") |> req_perform()
last_request()
last_response()

. <- request(example_url("/post")) |>
  req_body_json(list(a = 1, b = 2)) |>
  req_perform()
last_request_json()
last_request_json(pretty = FALSE)
last_response_json()
last_response_json(pretty = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{req_oauth}
\alias{req_oauth}
\title{OAuth authentication}
\usage{
req_oauth(req, flow, flow_params, cache)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{flow}{An \code{oauth_flow_} function used to generate the access token.}

\item{flow_params}{Parameters for the flow. This should be a named list
whose names match the argument names of \code{flow}.}

\item{cache}{An object that controls how the token is cached. This should
be a list containing three functions:
\itemize{
\item \code{get()} retrieves the token from the cache, returning \code{NULL} if not
cached yet.
\item \code{set()} saves the token to the cache.
\item \code{clear()} removes the token from the cache
}}
}
\value{
An \link{oauth_token}.
}
\description{
This is a low-level helper for automatically authenticating a request with
an OAuth flow, caching the access token and refreshing it where possible.
You should only need to use this function if you're implementing your own
OAuth flow.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{url_build}
\alias{url_build}
\title{Build a string from a URL object}
\usage{
url_build(url)
}
\arguments{
\item{url}{An URL object created by \link{url_parse}.}
}
\description{
This is the inverse of \code{\link[=url_parse]{url_parse()}}, taking a parsed URL object and
turning it back into a string.
}
\seealso{
Other URL manipulation: 
\code{\link{url_modify}()},
\code{\link{url_parse}()}
}
\concept{URL manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-url.R
\name{req_get_url}
\alias{req_get_url}
\title{Get request URL}
\usage{
req_get_url(req)
}
\arguments{
\item{req}{A httr2 \link{request} object.}
}
\value{
A character string.
}
\description{
Retrieve the URL from a request.
}
\examples{
request("https://httpbin.org") |>
  req_url_path("/get") |>
  req_url_query(hello = "world") |>
  req_get_url()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-throttle.R
\name{throttle_status}
\alias{throttle_status}
\title{Display internal throttle status}
\usage{
throttle_status()
}
\value{
A data frame with three columns:
\itemize{
\item The \code{realm}.
\item Number of \code{tokens} remaining in the bucket.
\item Time \code{to_wait} in seconds for next token.
}
}
\description{
Sometimes useful for debugging.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth_cache_path}
\alias{oauth_cache_path}
\title{httr2 OAuth cache location}
\usage{
oauth_cache_path()
}
\description{
When opted-in to, httr2 caches OAuth tokens in this directory. By default,
it uses a OS-standard cache directory, but, if needed, you can override the
location by setting the \code{HTTR2_OAUTH_CACHE} env var.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform.R
\name{req_perform}
\alias{req_perform}
\title{Perform a request to get a response}
\usage{
req_perform(
  req,
  path = NULL,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL),
  error_call = current_env()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{path}{Optionally, path to save body of the response. This is useful
for large responses since it avoids storing the response in memory.}

\item{verbosity}{How much information to print? This is a wrapper
around \code{\link[=req_verbose]{req_verbose()}} that uses an integer to control verbosity:
\itemize{
\item \code{0}: no output
\item \code{1}: show headers
\item \code{2}: show headers and bodies
\item \code{3}: show headers, bodies, and curl status messages.
}

Use \code{\link[=with_verbosity]{with_verbosity()}} to control the verbosity of requests that
you can't affect directly.}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}

\item{error_call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
\itemize{
\item If the HTTP request succeeds, and the status code is ok (e.g. 200),
an HTTP \link{response}.
\item If the HTTP request succeeds, but the status code is an error
(e.g a 404), an error with class \code{c("httr2_http_404", "httr2_http")}.
By default, all 400 and 500 status codes will be treated as an error,
but you can customise this with \code{\link[=req_error]{req_error()}}.
\item If the HTTP request fails (e.g. the connection is dropped or the
server doesn't exist), an error with class \code{"httr2_failure"}.
}
}
\description{
After preparing a \link{request}, call \code{req_perform()} to perform it, fetching
the results back to R as a \link{response}.

The default HTTP method is \code{GET} unless a body (set by \link{req_body_json} and
friends) is present, in which case it will be \code{POST}. You can override
these defaults with \code{\link[=req_method]{req_method()}}.
}
\section{Requests}{
Note that one call to \code{req_perform()} may perform multiple HTTP requests:
\itemize{
\item If the \code{url} is redirected with a 301, 302, 303, or 307, curl will
automatically follow the \code{Location} header to the new location.
\item If you have configured retries with \code{\link[=req_retry]{req_retry()}} and the request
fails with a transient problem, \code{req_perform()} will try again after
waiting a bit. See \code{\link[=req_retry]{req_retry()}} for details.
\item If you are using OAuth, and the cached token has expired, \code{req_perform()}
will get a new token either using the refresh token (if available)
or by running the OAuth flow.
}
}

\section{Progress bar}{
\code{req_perform()} will automatically add a progress bar if it needs to wait
between requests for \code{\link[=req_throttle]{req_throttle()}} or \code{\link[=req_retry]{req_retry()}}. You can turn the
progress bar off (and just show the total time to wait) by setting
\code{options(httr2_progress = FALSE)}.
}

\examples{
request("https://google.com") |>
  req_perform()
}
\seealso{
\code{\link[=req_perform_parallel]{req_perform_parallel()}} to perform multiple requests in parallel.
\code{\link[=req_perform_iterative]{req_perform_iterative()}} to perform multiple requests iteratively.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-stream.R, R/resp-stream-aws.R
\name{resp_stream_raw}
\alias{resp_stream_raw}
\alias{resp_stream_lines}
\alias{resp_stream_sse}
\alias{resp_stream_aws}
\alias{close.httr2_response}
\alias{resp_stream_is_complete}
\title{Read a streaming body a chunk at a time}
\usage{
resp_stream_raw(resp, kb = 32)

resp_stream_lines(resp, lines = 1, max_size = Inf, warn = TRUE)

resp_stream_sse(resp, max_size = Inf)

resp_stream_aws(resp, max_size = Inf)

\method{close}{httr2_response}(con, ...)

resp_stream_is_complete(resp)
}
\arguments{
\item{resp, con}{A streaming \link{response} created by \code{\link[=req_perform_connection]{req_perform_connection()}}.}

\item{kb}{How many kilobytes (1024 bytes) of data to read.}

\item{lines}{The maximum number of lines to return at once.}

\item{max_size}{The maximum number of bytes to buffer; once this number of
bytes has been exceeded without a line/event boundary, an error is thrown.}

\item{warn}{Like \code{\link[=readLines]{readLines()}}: warn if the connection ends without a final
EOL.}

\item{...}{Not used; included for compatibility with generic.}
}
\value{
\itemize{
\item \code{resp_stream_raw()}: a raw vector.
\item \code{resp_stream_lines()}: a character vector.
\item \code{resp_stream_sse()}: a list with components \code{type}, \code{data}, and \code{id}.
\code{type}, \code{data}, and \code{id} are always strings; \code{data} and \code{id} may be empty
strings.
\item \code{resp_stream_aws()}: a list with components \code{headers} and \code{body}.
\code{body} will be automatically parsed if the event contents a \verb{:content-type}
header with \code{application/json}.
}

\code{resp_stream_sse()} and \code{resp_stream_aws()} will return \code{NULL} to signal that
the end of the stream has been reached or, if in nonblocking mode, that
no event is currently available.
}
\description{
\itemize{
\item \code{resp_stream_raw()} retrieves bytes (\code{raw} vectors).
\item \code{resp_stream_lines()} retrieves lines of text (\code{character} vectors).
\item \code{resp_stream_sse()} retrieves a single \href{https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events}{server-sent event}.
\item \code{resp_stream_aws()} retrieves a single event from an AWS stream
(i.e. mime type `application/vnd.amazon.eventstream``).
}

Use \code{resp_stream_is_complete()} to determine if there is further data
waiting on the stream.
}
\examples{
req <- request(example_url()) |>
  req_template("GET /stream/:n", n = 5)

con <- req |> req_perform_connection()
while (!resp_stream_is_complete(con)) {
  lines <- con |> resp_stream_lines(2)
  cat(length(lines), " lines received\n", sep = "")
}
close(con)

# You can also see what's happening by setting verbosity
con <- req |> req_perform_connection(verbosity = 2)
while (!resp_stream_is_complete(con)) {
  lines <- con |> resp_stream_lines(2)
}
close(con)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-password.R
\name{req_oauth_password}
\alias{req_oauth_password}
\alias{oauth_flow_password}
\title{OAuth with username and password}
\usage{
req_oauth_password(
  req,
  client,
  username,
  password = NULL,
  scope = NULL,
  token_params = list(),
  cache_disk = FALSE,
  cache_key = username
)

oauth_flow_password(
  client,
  username,
  password = NULL,
  scope = NULL,
  token_params = list()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{username}{User name.}

\item{password}{Password. You should avoid entering the password directly
when calling this function as it will be captured by \code{.Rhistory}. Instead,
leave it unset and the default behaviour will prompt you for it
interactively.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}

\item{cache_disk}{Should the access token be cached on disk? This reduces
the number of times that you need to re-authenticate at the cost of
storing access credentials on disk.

Learn more in \url{https://httr2.r-lib.org/articles/oauth.html}.}

\item{cache_key}{If you want to cache multiple tokens per app, use this
key to disambiguate them.}
}
\value{
\code{req_oauth_password()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_password()} returns an \link{oauth_token}.
}
\description{
This function implements the OAuth \strong{resource owner password flow}, as
defined by \href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.3}{Section 4.3 of RFC 6749}. It allows the user to supply their password
once, exchanging it for an access token that can be cached locally.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}
}
\examples{
req_auth <- function(req) {
  req_oauth_password(req,
    client = oauth_client("example", "https://example.com/get_token"),
    username = "username"
  )
}
if (interactive()) {
  request("https://example.com") |>
    req_auth()
}
}
\seealso{
Other OAuth flows: 
\code{\link{req_oauth_auth_code}()},
\code{\link{req_oauth_bearer_jwt}()},
\code{\link{req_oauth_client_credentials}()},
\code{\link{req_oauth_refresh}()},
\code{\link{req_oauth_token_exchange}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-auth-code.R
\name{oauth_flow_auth_code}
\alias{oauth_flow_auth_code}
\alias{oauth_flow_auth_code_url}
\alias{oauth_flow_auth_code_listen}
\alias{oauth_flow_auth_code_parse}
\alias{oauth_flow_auth_code_pkce}
\title{OAuth flow: authorization code}
\usage{
oauth_flow_auth_code(
  client,
  auth_url,
  scope = NULL,
  pkce = TRUE,
  auth_params = list(),
  token_params = list(),
  host_name = "localhost",
  host_ip = "127.0.0.1",
  port = httpuv::randomPort()
)

oauth_flow_auth_code_url(
  client,
  auth_url,
  redirect_uri = NULL,
  scope = NULL,
  state = NULL,
  auth_params = list()
)

oauth_flow_auth_code_listen(host_ip = "127.0.0.1", port = 1410)

oauth_flow_auth_code_parse(query, state)

oauth_flow_auth_code_pkce()
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{auth_url}{Authorization url; you'll need to discover this by reading
the documentation.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{pkce}{Use "Proof Key for Code Exchange"? This adds an extra layer of
security and should always be used if supported by the server.}

\item{auth_params}{List containing additional parameters passed to \code{oauth_flow_auth_code_url()}}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}

\item{host_name}{Host name used to generate \code{redirect_uri}}

\item{host_ip}{IP address web server will be bound to.}

\item{port}{Port to bind web server to. By default, this uses a random port.
You may need to set it to a fixed port if the API requires that the
\code{redirect_uri} specified in the client exactly matches the \code{redirect_uri}
generated by this function.}

\item{redirect_uri}{URL to which user should be redirected.}

\item{state}{Random state generated by \code{oauth_flow_auth_code()}. Used to
verify that we're working with an authentication request that we created.
(This is an unlikely threat for R packages since the webserver that
listens for authorization responses is transient.)}

\item{query}{List of query parameters returned by \code{oauth_flow_auth_code_listen()}.}
}
\value{
An \link{oauth_token}.
}
\description{
These functions implement the OAuth authorization code flow, as defined
by \href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.1}{rfc6749},
Section 4.1. This is the most commonly used OAuth flow where the user is
opens a page in their browser, approves the access, and then returns to R.

\code{oauth_flow_auth_code()} is a high-level wrapper that should
work with APIs that adhere relatively closely to the spec. The remaining
low-level functions can be used to assemble a custom flow for APIs that are
further from the spec:
\itemize{
\item \code{oauth_flow_auth_code_url()} generates the url where the user is sent.
\item \code{oauth_flow_auth_code_listen()} starts an webserver that listens for
the response from the resource server.
\item \code{oauth_flow_auth_code_parse()} parses the query parameters returned from
the server redirect, verifying that the \code{state} is correct, and returning
the authorisation code.
\item \code{oauth_flow_auth_code_pkce()} generates code verifier, method, and challenge
components as needed for PKCE, as defined in
\href{https://datatracker.ietf.org/doc/html/rfc7636}{rfc7636}.
}
}
\examples{
client <- oauth_client(
  id = "28acfec0674bb3da9f38",
  secret = obfuscated(paste0(
     "J9iiGmyelHltyxqrHXW41ZZPZamyUNxSX1_uKnv",
     "PeinhhxET_7FfUs2X0LLKotXY2bpgOMoHRCo"
  )),
  token_url = "https://github.com/login/oauth/access_token",
  name = "hadley-oauth-test"
)
if (interactive()) {
  token <- oauth_flow_auth_code(client, auth_url = "https://github.com/login/oauth/authorize")
  token
}
}
\seealso{
Other OAuth flows: 
\code{\link{oauth_flow_bearer_jwt}()},
\code{\link{oauth_flow_client_credentials}()},
\code{\link{oauth_flow_device}()},
\code{\link{oauth_flow_password}()},
\code{\link{oauth_flow_refresh}()}
}
\concept{OAuth flows}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-options.R
\name{req_options}
\alias{req_options}
\title{Set arbitrary curl options in request}
\usage{
req_options(.req, ...)
}
\arguments{
\item{.req}{A \link{request}.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> Name-value pairs. The name
should be a valid curl option, as found in \code{\link[curl:curl_options]{curl::curl_options()}}.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{req_options()} is for expert use only; it allows you to directly set
libcurl options to access features that are otherwise not available in
httr2.
}
\examples{
# req_options() allows you to access curl options that are not otherwise
# exposed by httr2. For example, in very special cases you may need to
# turn off SSL verification. This is generally a bad idea so httr2 doesn't
# provide a convenient wrapper, but if you really know what you're doing
# you can still access this libcurl option:
req <- request("https://example.com") |>
  req_options(ssl_verifypeer = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httr2-package.R
\docType{package}
\name{httr2-package}
\alias{httr2}
\alias{httr2-package}
\title{httr2: Perform HTTP Requests and Process the Responses}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Tools for creating and modifying HTTP requests, then performing them and processing the results. 'httr2' is a modern re-imagining of 'httr' that uses a pipe-based interface and solves more of the problems that API wrapping packages face.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://httr2.r-lib.org}
  \item \url{https://github.com/r-lib/httr2}
  \item Report bugs at \url{https://github.com/r-lib/httr2/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item Maximilian Girlich [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-auth-code.R
\name{req_oauth_auth_code}
\alias{req_oauth_auth_code}
\alias{oauth_flow_auth_code}
\title{OAuth with authorization code}
\usage{
req_oauth_auth_code(
  req,
  client,
  auth_url,
  scope = NULL,
  pkce = TRUE,
  auth_params = list(),
  token_params = list(),
  redirect_uri = oauth_redirect_uri(),
  cache_disk = FALSE,
  cache_key = NULL
)

oauth_flow_auth_code(
  client,
  auth_url,
  scope = NULL,
  pkce = TRUE,
  auth_params = list(),
  token_params = list(),
  redirect_uri = oauth_redirect_uri()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{auth_url}{Authorization url; you'll need to discover this by reading
the documentation.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{pkce}{Use "Proof Key for Code Exchange"? This adds an extra layer of
security and should always be used if supported by the server.}

\item{auth_params}{A list containing additional parameters passed to
\code{\link[=oauth_flow_auth_code_url]{oauth_flow_auth_code_url()}}.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}

\item{redirect_uri}{URL to redirect back to after authorization is complete.
Often this must be registered with the API in advance.

httr2 supports three forms of redirect. Firstly, you can use a \code{localhost}
url (the default), where httr2 will set up a temporary webserver to listen
for the OAuth redirect. In this case, httr2 will automatically append a
random port. If you need to set it to a fixed port because the API requires
it, then specify it with (e.g.) \code{"http://localhost:1011"}. This technique
works well when you are working on your own computer.

Secondly, you can provide a URL to a website that uses Javascript to
give the user a code to copy and paste back into the R session (see
\url{https://tidyverse.org/google-callback/} and
\url{https://github.com/r-lib/gargle/blob/main/inst/pseudo-oob/google-callback/index.html}
for examples). This is less convenient (because it requires more
user interaction) but also works in hosted environments like RStudio
Server.

Finally, hosted platforms might set the \code{HTTR2_OAUTH_REDIRECT_URL} and
\code{HTTR2_OAUTH_CODE_SOURCE_URL} environment variables. In this case, httr2
will use \code{HTTR2_OAUTH_REDIRECT_URL} for redirects by default, and poll the
\code{HTTR2_OAUTH_CODE_SOURCE_URL} endpoint with the state parameter until it
receives a code in the response (or encounters an error). This delegates
completion of the authorization flow to the hosted platform.}

\item{cache_disk}{Should the access token be cached on disk? This reduces
the number of times that you need to re-authenticate at the cost of
storing access credentials on disk.

Learn more in \url{https://httr2.r-lib.org/articles/oauth.html}.}

\item{cache_key}{If you want to cache multiple tokens per app, use this
key to disambiguate them.}
}
\value{
\code{req_oauth_auth_code()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_auth_code()} returns an \link{oauth_token}.
}
\description{
Authenticate using the OAuth \strong{authorization code flow}, as defined
by \href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.1}{Section 4.1 of RFC 6749}.

This flow is the most commonly used OAuth flow where the user
opens a page in their browser, approves the access, and then returns to R.
When possible, it redirects the browser back to a temporary local webserver
to capture the authorization code. When this is not possible (e.g., when
running on a hosted platform like RStudio Server), provide a custom
\code{redirect_uri} and httr2 will prompt the user to enter the code manually.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}, and more about the motivations
behind this flow in
\url{https://stack-auth.com/blog/oauth-from-first-principles}.
}
\section{Security considerations}{
The authorization code flow is used for both web applications and native
applications (which are equivalent to R packages). \href{https://datatracker.ietf.org/doc/html/rfc8252}{RFC 8252} spells out
important considerations for native apps. Most importantly there's no way
for native apps to keep secrets from their users. This means that the
server should either not require a \code{client_secret} (i.e. it should be a
public client and not a confidential client) or ensure that possession of
the \code{client_secret} doesn't grant any significant privileges.

Only modern APIs from major providers (like Azure and Google) explicitly
support native apps. However, in most cases, even for older APIs, possessing
the \code{client_secret} provides limited ability to perform harmful actions.
Therefore, our general principle is that it's acceptable to include it in an
R package, as long as it's mildly obfuscated to protect against credential
scraping attacks (which aim to acquire large numbers of client secrets by
scanning public sites like GitHub). The goal is to ensure that obtaining your
client credentials is more work than just creating a new client.
}

\examples{
req_auth_github <- function(req) {
  req_oauth_auth_code(
    req,
    client = example_github_client(),
    auth_url = "https://github.com/login/oauth/authorize"
  )
}

request("https://api.github.com/user") |>
  req_auth_github()
}
\seealso{
\code{\link[=oauth_flow_auth_code_url]{oauth_flow_auth_code_url()}} for the components necessary to
write your own auth code flow, if the API you are wrapping does not adhere
closely to the standard.

Other OAuth flows: 
\code{\link{req_oauth_bearer_jwt}()},
\code{\link{req_oauth_client_credentials}()},
\code{\link{req_oauth_password}()},
\code{\link{req_oauth_refresh}()},
\code{\link{req_oauth_token_exchange}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jwt.R
\name{jwt_claim}
\alias{jwt_claim}
\alias{jwt_encode_sig}
\alias{jwt_encode_hmac}
\title{Create and encode a JWT}
\usage{
jwt_claim(
  iss = NULL,
  sub = NULL,
  aud = NULL,
  exp = unix_time() + 5L * 60L,
  nbf = unix_time(),
  iat = unix_time(),
  jti = NULL,
  ...
)

jwt_encode_sig(claim, key, size = 256, header = list())

jwt_encode_hmac(claim, secret, size = 256, header = list())
}
\arguments{
\item{iss}{Issuer claim. Identifies the principal that issued the JWT.}

\item{sub}{Subject claim. Identifies the principal that is the subject of
the JWT (i.e. the entity that the claims apply to).}

\item{aud}{Audience claim. Identifies the recipients that the JWT is
intended. Each principle intended to process the JWT must be identified
with a unique value.}

\item{exp}{Expiration claim. Identifies the expiration time on or after which
the JWT MUST NOT be accepted for processing. Defaults to 5 minutes.}

\item{nbf}{Not before claim. Identifies the time before which the JWT
MUST NOT be accepted for processing. Defaults to current time.}

\item{iat}{Issued at claim. Identifies the time at which the JWT was
issued.  Defaults to current time.}

\item{jti}{JWT ID claim. Provides a unique identifier for the JWT.
If omitted, uses a random 32-byte sequence encoded with base64url.}

\item{...}{Any additional claims to include in the claim set.}

\item{claim}{Claim set produced by \code{\link[=jwt_claim]{jwt_claim()}}.}

\item{key}{RSA or EC private key either specified as a path to a file,
a connection, or a string (PEM/SSH format), or a raw vector (DER format).}

\item{size}{Size, in bits, of sha2 signature, i.e. 256, 384 or 512.
Only for HMAC/RSA, not applicable for ECDSA keys.}

\item{header}{A named list giving additional fields to include in the
JWT header.}

\item{secret}{String or raw vector with a secret passphrase.}
}
\value{
An S3 list with class \code{jwt_claim}.
}
\description{
\code{jwt_claim()} is a wrapper around \code{\link[jose:jwt_claim]{jose::jwt_claim()}} that creates a JWT
claim set with a few extra default values. \code{jwt_encode_sig()} and
\code{jwt_encode_hmac()} are thin wrappers around \code{\link[jose:jwt_encode]{jose::jwt_encode_sig()}} and
\code{\link[jose:jwt_encode]{jose::jwt_encode_hmac()}} that exist primarily to make specification
in other functions a little simpler.
}
\examples{
claim <- jwt_claim()
str(claim)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp.R
\name{response}
\alias{response}
\alias{response_json}
\title{Create a HTTP response for testing}
\usage{
response(
  status_code = 200,
  url = "https://example.com",
  method = "GET",
  headers = list(),
  body = raw(),
  timing = NULL
)

response_json(
  status_code = 200,
  url = "https://example.com",
  method = "GET",
  headers = list(),
  body = list()
)
}
\arguments{
\item{status_code}{HTTP status code. Must be a single integer.}

\item{url}{URL response came from; might not be the same as the URL in
the request if there were any redirects.}

\item{method}{HTTP method used to retrieve the response.}

\item{headers}{HTTP headers. Can be supplied as a raw or character vector
which will be parsed using the standard rules, or a named list.}

\item{body}{The response body. For \code{response_json()}, a R data structure that
will be serialized to JSON.}

\item{timing}{A named numeric vector giving the time taken by various
components.}
}
\value{
An HTTP response: an S3 list with class \code{httr2_response}.
}
\description{
\code{response()} creates a generic response; \code{response_json()} creates a
response with a JSON body, automatically adding the correct \code{Content-Type}
header.

Generally, you should not need to call these function directly; you'll
get a real HTTP response by calling \code{\link[=req_perform]{req_perform()}} and friends. These
function is provided primarily for use in tests; if you are creating
responses for mocked requests, use the lower-level \code{\link[=new_response]{new_response()}}.
}
\examples{
response()
response(404, method = "POST")
response(headers = c("Content-Type: text/html", "Content-Length: 300"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-headers.R
\name{resp_content_type}
\alias{resp_content_type}
\alias{resp_encoding}
\title{Extract response content type and encoding}
\usage{
resp_content_type(resp)

resp_encoding(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\value{
A string. If no content type is specified \code{resp_content_type()}
will return a character \code{NA}; if no encoding is specified,
\code{resp_encoding()} will return \code{"UTF-8"}.
}
\description{
\code{resp_content_type()} returns the just the type and subtype of the
from the \code{Content-Type} header. If \code{Content-Type} is not provided; it
returns \code{NA}. Used by \code{\link[=resp_body_json]{resp_body_json()}}, \code{\link[=resp_body_html]{resp_body_html()}}, and
\code{\link[=resp_body_xml]{resp_body_xml()}}.

\code{resp_encoding()} returns the likely character encoding of text
types, as parsed from the \code{charset} parameter of the \code{Content-Type}
header. If that header is not found, not valid, or no charset parameter
is found, returns \code{UTF-8}. Used by \code{\link[=resp_body_string]{resp_body_string()}}.
}
\examples{
resp <- response(headers = "Content-type: text/html; charset=utf-8")
resp |> resp_content_type()
resp |> resp_encoding()

# No Content-Type header
resp <- response()
resp |> resp_content_type()
resp |> resp_encoding()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-body.R
\name{req_get_body_type}
\alias{req_get_body_type}
\alias{req_get_body}
\title{Get request body}
\usage{
req_get_body_type(req)

req_get_body(req, obfuscated = c("remove", "redact", "reveal"))
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{obfuscated}{Form and JSON bodies can contain \link{obfuscated} values.
This argument control what happens to them: should they be removed,
redacted, or revealed.}
}
\description{
This pair of functions gives you sufficient information to capture the
body of a request, and recreate, if needed. httr2 currently supports
seven possible body types:
\itemize{
\item empty: no body.
\item raw: created by \code{\link[=req_body_raw]{req_body_raw()}} with a raw vector.
\item string: created by \code{\link[=req_body_raw]{req_body_raw()}} with a string.
\item file: created by \code{\link[=req_body_file]{req_body_file()}}.
\item json: created by \code{\link[=req_body_json]{req_body_json()}}/\code{\link[=req_body_json_modify]{req_body_json_modify()}}.
\item form: created by \code{\link[=req_body_form]{req_body_form()}}.
\item multipart: created by \code{\link[=req_body_multipart]{req_body_multipart()}}.
}
}
\examples{
req <- request(example_url())
req |> req_body_raw("abc") |> req_get_body_type()
req |> req_body_file(system.file("DESCRIPTION")) |> req_get_body_type()
req |> req_body_json(list(x = 1, y = 2)) |> req_get_body_type()
req |> req_body_form(x = 1, y = 2) |> req_get_body_type()
req |> req_body_multipart(x = "x", y = "y") |> req_get_body_type()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-url.R
\name{req_url}
\alias{req_url}
\alias{req_url_relative}
\alias{req_url_query}
\alias{req_url_path}
\alias{req_url_path_append}
\title{Modify request URL}
\usage{
req_url(req, url)

req_url_relative(req, url)

req_url_query(
  .req,
  ...,
  .multi = c("error", "comma", "pipe", "explode"),
  .space = c("percent", "form")
)

req_url_path(req, ...)

req_url_path_append(req, ...)
}
\arguments{
\item{req, .req}{A httr2 \link{request} object.}

\item{url}{A new URL; either an absolute URL for \code{req_url()} or a
relative URL for \code{req_url_relative()}.}

\item{...}{For \code{req_url_query()}: <\code{\link[rlang:dyn-dots]{dynamic-dots}}>
Name-value pairs that define query parameters. Each value must be either
an atomic vector or \code{NULL} (which removes the corresponding parameters).
If you want to opt out of escaping, wrap strings in \code{I()}.

For \code{req_url_path()} and \code{req_url_path_append()}: A sequence of path
components that will be combined with \code{/}.}

\item{.multi}{Controls what happens when a value is a vector:
\itemize{
\item \code{"error"}, the default, throws an error.
\item \code{"comma"}, separates values with a \verb{,}, e.g. \verb{?x=1,2}.
\item \code{"pipe"}, separates values with a \code{|}, e.g. \code{?x=1|2}.
\item \code{"explode"}, turns each element into its own parameter, e.g. \code{?x=1&x=2}
}

If none of these options work for your needs, you can instead supply a
function that takes a character vector of argument values and returns a
a single string.}

\item{.space}{How should spaces in query params be escaped? The default,
"percent", uses standard percent encoding (i.e. \verb{\%20}), but you can opt-in
to "form" encoding, which uses \code{+} instead.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\itemize{
\item \code{req_url()} replaces the entire URL.
\item \code{req_url_relative()} navigates to a relative URL.
\item \code{req_url_query()} modifies individual query components.
\item \code{req_url_path()} modifies just the path.
\item \code{req_url_path_append()} adds to the path.
}
}
\examples{
# Change complete url
req <- request("http://example.com")
req |> req_url("http://google.com")

# Use a relative url
req <- request("http://example.com/a/b/c")
req |> req_url_relative("..")
req |> req_url_relative("/d/e/f")

# Change url components
req |>
  req_url_path_append("a") |>
  req_url_path_append("b") |>
  req_url_path_append("search.html") |>
  req_url_query(q = "the cool ice")

# Modify individual query parameters
req <- request("http://example.com?a=1&b=2")
req |> req_url_query(a = 10)
req |> req_url_query(a = NULL)
req |> req_url_query(c = 3)

# Use .multi to control what happens with vector parameters:
req |> req_url_query(id = 100:105, .multi = "comma")
req |> req_url_query(id = 100:105, .multi = "explode")

# If you have query parameters in a list, use !!!
params <- list(a = "1", b = "2")
req |>
  req_url_query(!!!params, c = "3")
}
\seealso{
\itemize{
\item To modify a URL without creating a request, see \code{\link[=url_modify]{url_modify()}} and
friends.
\item To use a template like \code{GET /user/{user}}, see \code{\link[=req_template]{req_template()}}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-timing.R
\name{resp_timing}
\alias{resp_timing}
\title{Extract timing data}
\usage{
resp_timing(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\value{
Named numeric vector of timing information.
The names of the elements in this vector correspond to the names used
in \href{https://curl.se/libcurl/c/curl_easy_getinfo.html}{libcurl's \code{curl_easy_getinfo()} API}.
The most useful component is likely \code{"total"} (corresponding to
\code{CURLINFO_TOTAL_TIME}), the overall time in seconds to complete the
request including any redirects followed.
}
\description{
The underlying curl library measures how long different components of the
request take to complete. This function retrieves that information.
}
\examples{
req <- request(example_url())
resp <- req_perform(req)
resp_timing(resp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-options.R
\name{req_timeout}
\alias{req_timeout}
\title{Set time limit for a request}
\usage{
req_timeout(req, seconds)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{seconds}{Maximum number of seconds to wait}
}
\value{
A modified HTTP \link{request}.
}
\description{
An error will be thrown if the request does not complete in the time limit.
}
\examples{
# Give up after at most 10 seconds
request("http://example.com") |> req_timeout(10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-method.R
\name{req_get_method}
\alias{req_get_method}
\title{Get request method}
\usage{
req_get_method(req)
}
\arguments{
\item{req}{A httr2 \link{request} object.}
}
\description{
Defaults to \code{GET}, unless the request has a body, in which case it uses
\code{POST}. Either way the method can be overridden with \code{\link[=req_method]{req_method()}}.
}
\examples{
req <- request(example_url())
req_get_method(req)
req_get_method(req |> req_body_raw("abc"))
req_get_method(req |> req_method("DELETE"))
req_get_method(req |> req_method("HEAD"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-server-metadata.R
\name{oauth_server_metadata}
\alias{oauth_server_metadata}
\title{Discover OAuth server metadata}
\usage{
oauth_server_metadata(issuer, type = c("openid", "oauth"), url = NULL)
}
\arguments{
\item{issuer}{The issuer URL, e.g. \code{"https://accounts.google.com"}. The
metadata URL is derived from it according to \code{type}.}

\item{type}{Which well-known suffix to use when \code{url} is not supplied:
\itemize{
\item \code{"openid"} (the default) appends \verb{/.well-known/openid-configuration},
the form served by essentially every major provider. Despite the name,
it is a superset that also advertises the OAuth endpoints, so it is the
better default even for plain OAuth.
\item \code{"oauth"} inserts \verb{/.well-known/oauth-authorization-server} between the
origin and any path, as defined in \href{https://datatracker.ietf.org/doc/html/rfc8414}{RFC 8414}. Use this for the few
providers that serve only the OAuth document.
}}

\item{url}{Optionally, the full metadata document URL. Use this as an escape
hatch for providers that follow neither well-known convention. When
supplied, \code{issuer} is only used for validation and can be omitted.}
}
\value{
An S3 list with class \code{httr2_oauth_server_metadata} containing the
full parsed metadata document. Endpoints that the provider does not
advertise are simply absent.
}
\description{
\code{oauth_server_metadata()} fetches and parses an OAuth 2.0 Authorization
Server Metadata document (\href{https://datatracker.ietf.org/doc/html/rfc8414}{RFC 8414}) or OpenID Connect Discovery
document, returning the endpoints advertised by an issuer.

Use it to discover values like \code{authorization_endpoint}, \code{token_endpoint},
and \code{device_authorization_endpoint} rather than hard-coding them. Pass the
result to \code{\link[=oauth_client]{oauth_client()}} as \code{metadata} and the endpoints are wired up for
you:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{metadata <- oauth_server_metadata("https://accounts.google.com")
client <- oauth_client("id", metadata = metadata, secret = "...")
token <- oauth_flow_auth_code(client)
}\if{html}{\out{</div>}}

As a security measure, the \code{issuer} reported in the returned document is
validated against the requested \code{issuer} (\href{https://datatracker.ietf.org/doc/html/rfc8414#section-3.3}{Section 3.3 of RFC 8414}); a mismatch
is an error. This check is skipped when \code{url} is supplied without \code{issuer}.
}
\examples{
\dontshow{if (is_online()) withAutoprint(\{ # examplesIf}
oauth_server_metadata("https://accounts.google.com")
\dontshow{\}) # examplesIf}
}
\seealso{
Other OAuth flows:
\code{\link[=req_oauth_auth_code]{req_oauth_auth_code()}},
\code{\link[=req_oauth_bearer_jwt]{req_oauth_bearer_jwt()}},
\code{\link[=req_oauth_client_credentials]{req_oauth_client_credentials()}},
\code{\link[=req_oauth_password]{req_oauth_password()}},
\code{\link[=req_oauth_refresh]{req_oauth_refresh()}},
\code{\link[=req_oauth_token_exchange]{req_oauth_token_exchange()}}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-parallel.R
\name{multi_req_perform}
\alias{multi_req_perform}
\title{Perform a list of requests in parallel}
\usage{
multi_req_perform(
  reqs,
  paths = NULL,
  pool = deprecated(),
  cancel_on_error = FALSE
)
}
\arguments{
\item{reqs}{A list of \link{request}s.}

\item{paths}{An optional character vector of paths, if you want to download
the response bodies to disk. If supplied, must be the same length as
\code{reqs}.}

\item{pool}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}. No longer supported;
to control the maximum number of concurrent requests, set \code{max_active}.}

\item{cancel_on_error}{Should all pending requests be cancelled when you
hit an error? Set this to \code{TRUE} to stop all requests as soon as you
hit an error. Responses that were never performed be \code{NULL} in the result.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

Please use \code{\link[=req_perform_parallel]{req_perform_parallel()}} instead, and note:
\itemize{
\item \code{cancel_on_error = FALSE} is now \code{on_error = "continue"}
\item \code{cancel_on_error = TRUE} is now \code{on_error = "return"}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httr2-translate.R
\name{httr2_translate}
\alias{httr2_translate}
\title{Translate a httr2 request to a curl command}
\usage{
httr2_translate(req, obfuscated = c("redact", "reveal"))
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{obfuscated}{Form and JSON bodies can contain \link{obfuscated} values.
This argument control what happens to them: should they be removed,
redacted, or revealed.}
}
\value{
A string containing the curl command.
}
\description{
Convert a httr2 request object to an approximate curl command line call.
This is useful for debugging, for sharing a request with someone who doesn't
use R, or for handing off to another tool.
}
\examples{
# Basic GET request
request("https://httpbin.org/get") |>
  httr2_translate()

# POST with JSON body
request("https://httpbin.org/post") |>
  req_body_json(list(name = "value")) |>
  httr2_translate()

# Secrets are redacted by default, but can be revealed
request("https://example.com") |>
  req_headers_redacted(Authorization = "secret") |>
  httr2_translate(obfuscated = "reveal")
}
\seealso{
\code{\link[=curl_translate]{curl_translate()}} to translate in the other direction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/secret.R
\name{secrets}
\alias{secret_make_key}
\alias{secret_encrypt}
\alias{secret_decrypt}
\alias{secret_write_rds}
\alias{secret_read_rds}
\alias{secret_decrypt_file}
\alias{secret_encrypt_file}
\alias{secret_has_key}
\title{Secret management}
\usage{
secret_make_key()

secret_encrypt(x, key)

secret_decrypt(encrypted, key)

secret_write_rds(x, path, key)

secret_read_rds(path, key)

secret_decrypt_file(path, key, envir = parent.frame())

secret_encrypt_file(path, key)

secret_has_key(key)
}
\arguments{
\item{x}{Object to encrypt. Must be a string for \code{secret_encrypt()}.}

\item{key}{Encryption key; this is the password that allows you to "lock"
and "unlock" the secret. The easiest way to specify this is as the
name of an environment variable. Alternatively, if you already have
a base64url encoded string, you can wrap it in \code{I()}, or you can pass
the raw vector in directly.}

\item{encrypted}{String to decrypt}

\item{path}{Path to file to encrypted file to read or write. For
\code{secret_write_rds()} and \code{secret_read_rds()} this should be an \code{.rds}
file.}

\item{envir}{The decrypted file will be automatically deleted when
this environment exits. You should only need to set this argument if you
want to pass the unencrypted file to another function.}
}
\value{
\itemize{
\item \code{secret_decrypt()} and \code{secret_encrypt()} return strings.
\item \code{secret_decrypt_file()} returns a path to a temporary file;
\code{secret_encrypt_file()} encrypts the file in place.
\item \code{secret_write_rds()} returns \code{x} invisibly; \code{secret_read_rds()}
returns the saved object.
\item \code{secret_make_key()} returns a string with class \code{AsIs}.
\item \code{secret_has_key()} returns \code{TRUE} or \code{FALSE}.
}
}
\description{
httr2 provides a handful of functions designed for working with confidential
data. These are useful because testing packages that use httr2 often
requires some confidential data that needs to be available for testing,
but should not be available to package users.
\itemize{
\item \code{secret_encrypt()} and \code{secret_decrypt()} work with individual strings
\item \code{secret_encrypt_file()} encrypts a file in place and
\code{secret_decrypt_file()} decrypts a file in a temporary location.
\item \code{secret_write_rds()} and \code{secret_read_rds()} work with \code{.rds} files
\item \code{secret_make_key()} generates a random string to use as a key.
\item \code{secret_has_key()} returns \code{TRUE} if the key is available; you can
use it in examples and vignettes that you want to evaluate on your CI,
but not for CRAN/package users.
}

These all look for the key in an environment variable. When used inside of
testthat, they will automatically \code{\link[testthat:skip]{testthat::skip()}} the test if the env var
isn't found. (Outside of testthat, they'll error if the env var isn't
found.)
}
\section{Basic workflow}{
\enumerate{
\item Use \code{secret_make_key()} to generate a password. Make this available
as an env var (e.g. \verb{\{MYPACKAGE\}_KEY}) by adding a line to your
\code{.Renviron}.
\item Encrypt strings with \code{secret_encrypt()}, files with
\code{secret_encrypt_file()}, and other data with \code{secret_write_rds()},
setting \code{key = "{MYPACKAGE}_KEY"}.
\item In your tests, decrypt the data with \code{secret_decrypt()},
\code{secret_decrypt_file()}, or \code{secret_read_rds()} to match how you encrypt
it.
\item If you push this code to your CI server, it will already "work" because
all functions automatically skip tests when your \verb{\{MYPACKAGE\}_KEY}
env var isn't set. To make the tests actually run, you'll need to set
the env var using whatever tool your CI system provides for setting
env vars. Make sure to carefully inspect the test output to check that
the skips have actually gone away.
}
}

\examples{
key <- secret_make_key()

path <- tempfile()
secret_write_rds(mtcars, path, key = key)
secret_read_rds(path, key)

# While you can manage the key explicitly in a variable, it's much
# easier to store in an environment variable. In real life, you should
# NEVER use `Sys.setenv()` to create this env var because you will
# also store the secret in your `.Rhistory`. Instead add it to your
# .Renviron using `usethis::edit_r_environ()` or similar.
Sys.setenv("MY_KEY" = key)

x <- secret_encrypt("This is a secret", "MY_KEY")
x
secret_decrypt(x, "MY_KEY")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform.R
\name{req_stream}
\alias{req_stream}
\title{Perform a request, streaming data back to R}
\usage{
req_stream(req, callback, timeout_sec = Inf, buffer_kb = 64)
}
\arguments{
\item{req}{A \link{request}.}

\item{callback}{A single argument callback function. It will be called
repeatedly with a raw vector whenever there is at least \code{buffer_kb}
worth of data to process. It must return \code{TRUE} to continue streaming.}

\item{timeout_sec}{Number of seconds to processs stream for.}

\item{buffer_kb}{Buffer size, in kilobytes.}
}
\value{
An HTTP \link{response}.
}
\description{
After preparing a request, call \code{req_stream()} to perform the request
and handle the result with a streaming callback. This is useful for
streaming HTTP APIs where potentially the stream never ends.
}
\examples{
show_bytes <- function(x) {
  cat("Got ", length(x), " bytes\n", sep = "")
  TRUE
}
resp <- request("http://httpbin.org/stream-bytes/100000") \%>\%
  req_stream(show_bytes, buffer_kb = 32)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-headers.R
\name{resp_headers}
\alias{resp_headers}
\alias{resp_header}
\alias{resp_header_exists}
\title{Extract headers from a response}
\usage{
resp_headers(resp, filter = NULL)

resp_header(resp, header, default = NULL)

resp_header_exists(resp, header)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{filter}{A regular expression used to filter the header names.
\code{NULL}, the default, returns all headers.}

\item{header}{Header name (case insensitive)}

\item{default}{Default value to use if header doesn't exist.}
}
\value{
\itemize{
\item \code{resp_headers()} returns a list.
\item \code{resp_header()} returns a string if the header exists and \code{NULL} otherwise.
\item \code{resp_header_exists()} returns \code{TRUE} or \code{FALSE}.
}
}
\description{
\itemize{
\item \code{resp_headers()} retrieves a list of all headers.
\item \code{resp_header()} retrieves a single header.
\item \code{resp_header_exists()} checks if a header is present.
}
}
\examples{
resp <- request("https://httr2.r-lib.org") |> req_perform()
resp |> resp_headers()
resp |> resp_headers("x-")

resp |> resp_header_exists("server")
resp |> resp_header("server")
# Headers are case insensitive
resp |> resp_header("SERVER")

# Returns NULL if header doesn't exist
resp |> resp_header("this-header-doesnt-exist")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbosity.R
\name{with_verbosity}
\alias{with_verbosity}
\alias{local_verbosity}
\title{Temporarily set verbosity for all requests}
\usage{
with_verbosity(code, verbosity = 1)

local_verbosity(verbosity, env = caller_env())
}
\arguments{
\item{code}{Code to execture}

\item{verbosity}{How much information to print? This is a wrapper
around \code{\link[=req_verbose]{req_verbose()}} that uses an integer to control verbosity:
\itemize{
\item \code{0}: no output
\item \code{1}: show headers
\item \code{2}: show headers and bodies
\item \code{3}: show headers, bodies, and curl status messages.
}

Use \code{\link[=with_verbosity]{with_verbosity()}} to control the verbosity of requests that
you can't affect directly.}

\item{env}{Environment to use for scoping changes.}
}
\value{
\code{with_verbosity()} returns the result of evaluating \code{code}.
\code{local_verbosity()} is called for its side-effect and invisibly returns
the previous value of the option.
}
\description{
\code{with_verbosity()} and \code{local_verbosity()} are useful for debugging httr2
code buried deep inside another package, because they allow you to change
the verbosity even when you don't have access to the request.

Both functions work by temporarily setting the \code{httr2_verbosity} option. You
can also control verbosity by setting the \code{HTTR2_VERBOSITY} environment
variable. This has lower precedence than the option, but can be more easily
changed outside of R.
}
\examples{
fun <- function() {
  request("https://httr2.r-lib.org") |> req_perform()
}
with_verbosity(fun())

fun <- function() {
  local_verbosity(2)
  # someotherpackage::fun()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-throttle.R
\name{req_throttle}
\alias{req_throttle}
\title{Rate limit a request by automatically adding a delay}
\usage{
req_throttle(req, rate, capacity, fill_time_s = 60, realm = NULL)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{rate}{For backwards compatibility, you can still specify the \code{rate},
which is converted to \code{capacity} by multiplying by \code{fill_time_s}.
However, we recommend using \code{capacity} and \code{fill_time_s} as it gives more
control.}

\item{capacity}{The size of the bucket, i.e. the maximum number of
tokens that can accumulate.}

\item{fill_time_s}{Time in seconds to fill the capacity. Defaults to 60s.}

\item{realm}{A string that uniquely identifies the throttle pool to use
(throttling limits always apply \emph{per pool}). If not supplied, defaults
to the hostname of the request.}
}
\value{
A modified HTTP \link{request}.
}
\description{
Use \code{req_throttle()} to ensure that repeated calls to \code{\link[=req_perform]{req_perform()}} never
exceed a specified rate.

Throttling is implemented using a "token bucket", which steadily fills up to
a maximum of \code{capacity} tokens over \code{fill_time_s}. Each time you make a
request, it takes a token out of the bucket, and if the bucket is empty,
the request will wait until the bucket refills. This ensures that you never
make more than \code{capacity} requests in \code{fill_time_s}, but you can make
requests more quickly if the bucket is full. For example, if you have
\code{capacity = 10} and \code{fill_time_s = 60}, you can make 10 requests
without waiting, but the next request will wait 60 seconds. This gives the
same average throttling rate as the previous approach, but gives you much
better performance if you're only making a small number of requests.
}
\examples{
# Ensure we never send more than 30 requests a minute
req <- request(example_url()) |>
  req_throttle(capacity = 30, fill_time_s = 60)

resp <- req_perform(req)
throttle_status()
resp <- req_perform(req)
throttle_status()

\dontshow{httr2:::throttle_reset()}
}
\seealso{
\code{\link[=req_retry]{req_retry()}} for another way of handling rate-limited APIs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-sequential.R
\name{req_perform_sequential}
\alias{req_perform_sequential}
\title{Perform multiple requests in sequence}
\usage{
req_perform_sequential(
  reqs,
  paths = NULL,
  on_error = c("stop", "return", "continue"),
  mock = getOption("httr2_mock", NULL),
  progress = TRUE
)
}
\arguments{
\item{reqs}{A list of \link{request}s.}

\item{paths}{An optional character vector of paths, if you want to download
the response bodies to disk. If supplied, must be the same length as
\code{reqs}.}

\item{on_error}{What should happen if one of the requests fails?
\itemize{
\item \code{stop}, the default: stop iterating with an error.
\item \code{return}: stop iterating, returning all the successful responses
received so far, as well as an error object for the failed request.
\item \code{continue}: continue iterating, recording errors in the result.
}}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}

\item{progress}{Display a progress bar for the status of all requests? Use
\code{TRUE} to turn on a basic progress bar, use a string to give it a name,
or see \link{progress_bars} to customize it in other ways. Not compatible with
\code{\link[=req_progress]{req_progress()}}, as httr2 can only display a single progress bar at a
time.}
}
\value{
A list, the same length as \code{reqs}, containing \link{response}s and possibly
error objects, if \code{on_error} is \code{"return"} or \code{"continue"} and one of the
responses errors. If \code{on_error} is \code{"return"} and it errors on the ith
request, the ith element of the result will be an error object, and the
remaining elements will be \code{NULL}. If \code{on_error} is \code{"continue"}, it will
be a mix of requests and error objects.

Only httr2 errors are captured; see \code{\link[=req_error]{req_error()}} for more details.
}
\description{
Given a list of requests, this function performs each in turn, returning
a list of responses. It's the serial equivalent of \code{\link[=req_perform_parallel]{req_perform_parallel()}}.
}
\examples{
# One use of req_perform_sequential() is if the API allows you to request
# data for multiple objects, you want data for more objects than can fit
# in one request.
req <- request("https://api.restful-api.dev/objects")

# Imagine we have 50 ids:
ids <- sort(sample(100, 50))

# But the API only allows us to request 10 at time. So we first use split
# and some modulo arithmetic magic to generate chunks of length 10
chunks <- unname(split(ids, (seq_along(ids) - 1) \%/\% 10))

# Then we use lapply to generate one request for each chunk:
reqs <- chunks |> lapply(\(idx) req |> req_url_query(id = idx, .multi = "comma"))

# Then we can perform them all and get the results
\dontrun{
resps <- reqs |> req_perform_sequential()
resps_data(resps, \(resp) resp_body_json(resp))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-auth.R
\name{req_auth_basic}
\alias{req_auth_basic}
\title{Authenticate request with HTTP basic authentication}
\usage{
req_auth_basic(req, username, password = NULL)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{username}{User name.}

\item{password}{Password. You should avoid entering the password directly
when calling this function as it will be captured by \code{.Rhistory}. Instead,
leave it unset and the default behaviour will prompt you for it
interactively.}
}
\value{
A modified HTTP \link{request}.
}
\description{
This sets the Authorization header. See details at
\url{https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Authorization}.
}
\examples{
req <- request("http://example.com") |> req_auth_basic("hadley", "SECRET")
req
req |> req_dry_run()

# httr2 does its best to redact the Authorization header so that you don't
# accidentally reveal confidential data. Use `redact_headers` to reveal it:
print(req, redact_headers = FALSE)
req |> req_dry_run(redact_headers = FALSE)

# We do this because the authorization header is not encrypted and the
# so password can easily be discovered:
rawToChar(jsonlite::base64_dec("aGFkbGV5OlNFQ1JFVA=="))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-token-exchange.R
\name{req_oauth_token_exchange}
\alias{req_oauth_token_exchange}
\alias{oauth_flow_token_exchange}
\title{OAuth token exchange}
\usage{
req_oauth_token_exchange(
  req,
  client,
  subject_token,
  subject_token_type,
  resource = NULL,
  audience = NULL,
  scope = NULL,
  requested_token_type = NULL,
  actor_token = NULL,
  actor_token_type = NULL,
  token_params = list()
)

oauth_flow_token_exchange(
  client,
  subject_token,
  subject_token_type,
  resource = NULL,
  audience = NULL,
  scope = NULL,
  requested_token_type = NULL,
  actor_token = NULL,
  actor_token_type = NULL,
  token_params = list()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{subject_token}{The security token to exchange. This is usually an
OpenID Connect ID token or a SAML2 assertion.}

\item{subject_token_type}{A URI that describes the type of the security
token. Usually one of the options in \href{https://datatracker.ietf.org/doc/html/rfc8693#section-3}{Section 3 of RFC 8693}.}

\item{resource}{The URI that identifies the resource that the client is
trying to access, if applicable.}

\item{audience}{The logical name that identifies the resource that the client
is trying to access, if applicable. Usually one of \code{resource} or \code{audience}
must be supplied.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{requested_token_type}{An optional URI that describes the type of the
security token being requested. Usually one of the options in
\href{https://datatracker.ietf.org/doc/html/rfc8693#section-3}{Section 3 of RFC 8693}.}

\item{actor_token}{An optional security token that represents the client,
rather than the identity behind the subject token.}

\item{actor_token_type}{When \code{actor_token} is not \code{NULL}, this must be the
URI that describes the type of the security token being requested. Usually
one of the options in \href{https://datatracker.ietf.org/doc/html/rfc8693#section-3}{Section 3 of RFC 8693}.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
\code{req_oauth_token_exchange()} returns a modified HTTP \link{request} that
will exchange one security token for another; \code{oauth_flow_token_exchange()}
returns the resulting \link{oauth_token} directly.
}
\description{
Authenticate by exchanging one security token for another, as defined by
\href{https://datatracker.ietf.org/doc/html/rfc8693#section-2}{Section 2 of RFC 8693}. It is typically used for advanced authorization flows that
involve "delegation" or "impersonation" semantics, such as when a client
accesses a resource on behalf of another party, or when a client's identity
is federated from another provider.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}.
}
\examples{
# List Google Cloud storage buckets using an OIDC token obtained
# from e.g. Microsoft Entra ID or Okta and federated to Google. (A real
# project ID and workforce pool would be required for this in practice.)
#
# See: https://cloud.google.com/iam/docs/workforce-obtaining-short-lived-credentials
oidc_token <- "an ID token from Okta"
request("https://storage.googleapis.com/storage/v1/b?project=123456") |>
  req_oauth_token_exchange(
    client = oauth_client("gcp", "https://sts.googleapis.com/v1/token"),
    subject_token = oidc_token,
    subject_token_type = "urn:ietf:params:oauth:token-type:id_token",
    scope = "https://www.googleapis.com/auth/cloud-platform",
    requested_token_type = "urn:ietf:params:oauth:token-type:access_token",
    audience = "//iam.googleapis.com/locations/global/workforcePools/123/providers/456",
    token_params = list(
      options = '{"userProject":"123456"}'
    )
  )
}
\seealso{
Other OAuth flows: 
\code{\link{req_oauth_auth_code}()},
\code{\link{req_oauth_bearer_jwt}()},
\code{\link{req_oauth_client_credentials}()},
\code{\link{req_oauth_password}()},
\code{\link{req_oauth_refresh}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-refresh.R
\name{req_oauth_refresh}
\alias{req_oauth_refresh}
\alias{oauth_flow_refresh}
\title{OAuth with a refresh token}
\usage{
req_oauth_refresh(
  req,
  client,
  refresh_token = Sys.getenv("HTTR2_REFRESH_TOKEN"),
  scope = NULL,
  token_params = list()
)

oauth_flow_refresh(
  client,
  refresh_token = Sys.getenv("HTTR2_REFRESH_TOKEN"),
  scope = NULL,
  token_params = list()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{refresh_token}{A refresh token. This is equivalent to a password
so shouldn't be typed into the console or stored in a script. Instead,
we recommend placing in an environment variable; the default behaviour
is to look in \code{HTTR2_REFRESH_TOKEN}.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
\code{req_oauth_refresh()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_refresh()} returns an \link{oauth_token}.
}
\description{
Authenticate using a \strong{refresh token}, following the process described in
\href{https://datatracker.ietf.org/doc/html/rfc6749#section-6}{Section 6 of RFC 6749}.

This technique is primarily useful for testing: you can manually retrieve
a OAuth token using another OAuth flow (e.g. with \code{\link[=oauth_flow_auth_code]{oauth_flow_auth_code()}}),
extract the refresh token from the result, and then save in an environment
variable for use in automated tests.

When requesting an access token, the server may also return a new refresh
token. If this happens, \code{oauth_flow_refresh()} will warn, and you'll have
retrieve a new update refresh token and update the stored value. If you find
this happening a lot, it's a sign that you should be using a different flow
in your automated tests.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}.
}
\examples{
client <- oauth_client("example", "https://example.com/get_token")
req <- request("https://example.com")
req |> req_oauth_refresh(client)
}
\seealso{
Other OAuth flows: 
\code{\link{req_oauth_auth_code}()},
\code{\link{req_oauth_bearer_jwt}()},
\code{\link{req_oauth_client_credentials}()},
\code{\link{req_oauth_password}()},
\code{\link{req_oauth_token_exchange}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{url_parse}
\alias{url_parse}
\title{Parse a URL into its component pieces}
\usage{
url_parse(url, base_url = NULL)
}
\arguments{
\item{url}{A string containing the URL to parse.}

\item{base_url}{Use this as a parent, if \code{url} is a relative URL.}
}
\value{
An S3 object of class \code{httr2_url} with the following components:
\code{scheme}, \code{hostname}, \code{username}, \code{password}, \code{port}, \code{path}, \code{query}, and
\code{fragment}.
}
\description{
\code{url_parse()} parses a URL into its component parts, powered by
\code{\link[curl:curl_parse_url]{curl::curl_parse_url()}}. The parsing algorithm follows the specifications
detailed in \href{https://datatracker.ietf.org/doc/html/rfc3986}{RFC 3986}.
}
\examples{
url_parse("http://google.com/")
url_parse("http://google.com:80/")
url_parse("http://google.com:80/?a=1&b=2")
url_parse("http://username@google.com:80/path;test?a=1&b=2#40")

# You can parse a relative URL if you also provide a base url
url_parse("foo", "http://google.com/bar/")
url_parse("..", "http://google.com/bar/")
}
\seealso{
Other URL manipulation: 
\code{\link{url_build}()},
\code{\link{url_modify}()}
}
\concept{URL manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp.R
\name{resp_raw}
\alias{resp_raw}
\title{Show the raw response}
\usage{
resp_raw(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\value{
\code{resp} (invisibly).
}
\description{
This function reconstructs the HTTP message that httr2 received from the
server. It's unlikely to be exactly byte-for-byte identical (because most
servers compress at least the body, and HTTP/2 can also compress the
headers), but it conveys the same information.
}
\examples{
resp <- request(example_url()) |>
  req_url_path("/json") |>
  req_perform()
resp |> resp_raw()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{url_query_parse}
\alias{url_query_parse}
\alias{url_query_build}
\title{Parse query parameters and/or build a string}
\usage{
url_query_parse(query)

url_query_build(query, .multi = c("error", "comma", "pipe", "explode"))
}
\arguments{
\item{query}{A string, when parsing; a named list when building.}

\item{.multi}{Controls what happens when a value is a vector:
\itemize{
\item \code{"error"}, the default, throws an error.
\item \code{"comma"}, separates values with a \verb{,}, e.g. \verb{?x=1,2}.
\item \code{"pipe"}, separates values with a \code{|}, e.g. \code{?x=1|2}.
\item \code{"explode"}, turns each element into its own parameter, e.g. \code{?x=1&x=2}
}

If none of these options work for your needs, you can instead supply a
function that takes a character vector of argument values and returns a
a single string.}
}
\description{
\code{url_query_parse()} parses a query string into a named list;
\code{url_query_build()} builds a query string from a named list.
}
\examples{
str(url_query_parse("a=1&b=2"))

url_query_build(list(x = 1, y = "z"))
url_query_build(list(x = 1, y = 1:2), .multi = "explode")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-body.R
\name{resp_body_raw}
\alias{resp_body_raw}
\alias{resp_has_body}
\alias{resp_body_string}
\alias{resp_body_json}
\alias{resp_body_html}
\alias{resp_body_xml}
\title{Extract body from response}
\usage{
resp_body_raw(resp)

resp_has_body(resp)

resp_body_string(resp, encoding = NULL)

resp_body_json(resp, check_type = TRUE, simplifyVector = FALSE, ...)

resp_body_html(resp, check_type = TRUE, ...)

resp_body_xml(resp, check_type = TRUE, ...)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{encoding}{Character encoding of the body text. If not specified,
will use the encoding specified by the content-type, falling back to
UTF-8 with a warning if it cannot be found. The resulting string is
always re-encoded to UTF-8.}

\item{check_type}{Check that response has expected content type? Set to
\code{FALSE} to suppress the automated check}

\item{simplifyVector}{Should JSON arrays containing only primitives (i.e.
booleans, numbers, and strings) be caused to atomic vectors?}

\item{...}{Other arguments passed on to \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} and
\code{\link[xml2:read_xml]{xml2::read_xml()}} respectively.}
}
\value{
\itemize{
\item \code{resp_body_raw()} returns a raw vector.
\item \code{resp_body_string()} returns a string.
\item \code{resp_body_json()} returns NULL, an atomic vector, or list.
\item \code{resp_body_html()} and \code{resp_body_xml()} return an \code{xml2::xml_document}
}
}
\description{
\itemize{
\item \code{resp_body_raw()} returns the raw bytes.
\item \code{resp_body_string()} returns a UTF-8 string.
\item \code{resp_body_json()} returns parsed JSON.
\item \code{resp_body_html()} returns parsed HTML.
\item \code{resp_body_xml()} returns parsed XML.
\item \code{resp_has_body()} returns \code{TRUE} if the response has a body.
}

\code{resp_body_json()} and \code{resp_body_xml()} check that the content-type header
is correct; if the server returns an incorrect type you can suppress the
check with \code{check_type = FALSE}. These two functions also cache the parsed
object so the second and subsequent calls are low-cost.
}
\examples{
resp <- request("https://httr2.r-lib.org") |> req_perform()
resp

resp |> resp_has_body()
resp |> resp_body_raw()
resp |> resp_body_string()

if (requireNamespace("xml2", quietly = TRUE)) {
  resp |> resp_body_html()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-password.R
\name{oauth_flow_password}
\alias{oauth_flow_password}
\title{OAuth flow: user password}
\usage{
oauth_flow_password(
  client,
  username,
  password = NULL,
  scope = NULL,
  token_params = list()
)
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{username}{User name.}

\item{password}{Password. You avoid entering the password directly when
calling this function as it will be captured by \code{.Rhistory}. Instead,
leave it unset and the default behaviour will prompt you for it
interactively.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
An \link{oauth_token}.
}
\description{
This function implements the OAuth resource owner password flow, as defined
by \href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.3}{rfc6749},
Section 4.3. It allows the user to supply their password once, exchanging
it for an access token that can be cached locally.
}
\seealso{
Other OAuth flows: 
\code{\link{oauth_flow_auth_code}()},
\code{\link{oauth_flow_bearer_jwt}()},
\code{\link{oauth_flow_client_credentials}()},
\code{\link{oauth_flow_device}()},
\code{\link{oauth_flow_refresh}()}
}
\concept{OAuth flows}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iterate-helpers.R
\name{iterate_with_offset}
\alias{iterate_with_offset}
\alias{iterate_with_cursor}
\alias{iterate_with_link_url}
\title{Iteration helpers}
\usage{
iterate_with_offset(
  param_name,
  start = 1,
  offset = 1,
  resp_pages = NULL,
  resp_complete = NULL
)

iterate_with_cursor(param_name, resp_param_value)

iterate_with_link_url(rel = "next")
}
\arguments{
\item{param_name}{Name of query parameter.}

\item{start}{Starting value.}

\item{offset}{Offset for each page. The default is set to \code{1} so you get
(e.g.) \code{?page=1}, \code{?page=2}, ... If \code{param_name} refers to an element
index (rather than a page index) you'll want to set this to a larger number
so you get (e.g.) \code{?items=20}, \code{?items=40}, ...}

\item{resp_pages}{A callback function that takes a response (\code{resp}) and
returns the total number of pages, or \code{NULL} if unknown. It will only
be called once.}

\item{resp_complete}{A callback function that takes a response (\code{resp})
and returns \code{TRUE} if there are no further pages.}

\item{resp_param_value}{A callback function that takes a response (\code{resp})
and returns the next cursor value. Return \code{NULL} if there are no further
pages.}

\item{rel}{The "link relation type" to use to retrieve the next page.}
}
\description{
These functions are intended for use with the \code{next_req} argument to
\code{\link[=req_perform_iterative]{req_perform_iterative()}}. Each implements iteration for a common
pagination pattern:
\itemize{
\item \code{iterate_with_offset()} increments a query parameter, e.g. \code{?page=1},
\code{?page=2}, or \code{?offset=1}, \code{offset=21}.
\item \code{iterate_with_cursor()} updates a query parameter with the value of a
cursor found somewhere in the response.
\item \code{iterate_with_link_url()} follows the url found in the \code{Link} header.
See \code{resp_link_url()} for more details.
}
}
\examples{
req <- request(example_url()) |>
  req_url_path("/iris") |>
  req_throttle(10) |>
  req_url_query(limit = 50)

# If you don't know the total number of pages in advance, you can
# provide a `resp_complete()` callback
is_complete <- function(resp) {
  length(resp_body_json(resp)$data) == 0
}
resps <- req_perform_iterative(
  req,
  next_req = iterate_with_offset("page_index", resp_complete = is_complete),
  max_reqs = Inf
)

\dontrun{
# Alternatively, if the response returns the total number of pages (or you
# can easily calculate it), you can use the `resp_pages()` callback which
# will generate a better progress bar.

resps <- req_perform_iterative(
  req |> req_url_query(limit = 1),
  next_req = iterate_with_offset(
    "page_index",
    resp_pages = function(resp) resp_body_json(resp)$pages
  ),
  max_reqs = Inf
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-device.R
\name{oauth_flow_device}
\alias{oauth_flow_device}
\title{OAuth flow: device}
\usage{
oauth_flow_device(
  client,
  auth_url,
  pkce = FALSE,
  scope = NULL,
  auth_params = list(),
  token_params = list()
)
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{auth_url}{Authorization url; you'll need to discover this by reading
the documentation.}

\item{pkce}{Use "Proof Key for Code Exchange"? This adds an extra layer of
security and should always be used if supported by the server.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{auth_params}{List containing additional parameters passed to \code{oauth_flow_auth_code_url()}}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
An \link{oauth_token}.
}
\description{
These functions implement the OAuth device flow, as defined
by \href{https://datatracker.ietf.org/doc/html/rfc8628}{rfc8628}. It's designed
for devices that don't have access to a web browser (if you've ever
authenticated an app on your TV, this is probably the flow you've used),
but it also works well from within R.

This specification allows also some subspecifications:
\itemize{
\item \code{oauth_flow_auth_code_pkce()} is also reused here to generate code
verifier, method, and challenge components as needed for PKCE, as
defined in \href{https://datatracker.ietf.org/doc/html/rfc7636}{rfc7636}.
}
}
\seealso{
Other OAuth flows: 
\code{\link{oauth_flow_auth_code}()},
\code{\link{oauth_flow_bearer_jwt}()},
\code{\link{oauth_flow_client_credentials}()},
\code{\link{oauth_flow_password}()},
\code{\link{oauth_flow_refresh}()}
}
\concept{OAuth flows}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-connection.R
\name{StreamingBody}
\alias{StreamingBody}
\title{\code{StreamingBody} class}
\description{
This R6 class is used to represent the body of a streaming response.
When using this in mocked responses, you can either create a new instance
using your own connection or use a subclass for some other representation.
In either case, you will pass to the \code{body} argument of \code{\link[=new_response]{new_response()}}.
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-StreamingBody-initialize}{\code{StreamingBody$new()}}
    \item \href{#method-StreamingBody-read}{\code{StreamingBody$read()}}
    \item \href{#method-StreamingBody-read_all}{\code{StreamingBody$read_all()}}
    \item \href{#method-StreamingBody-is_open}{\code{StreamingBody$is_open()}}
    \item \href{#method-StreamingBody-is_complete}{\code{StreamingBody$is_complete()}}
    \item \href{#method-StreamingBody-get_fdset}{\code{StreamingBody$get_fdset()}}
    \item \href{#method-StreamingBody-close}{\code{StreamingBody$close()}}
    \item \href{#method-StreamingBody-clone}{\code{StreamingBody$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-initialize}{}}}
\subsection{\code{StreamingBody$new()}}{
  Create a new object
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$new(conn)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{conn}}{A connection, that is open and ready for reading.
\code{StreamingBody} will take care of closing it.`}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-read"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-read}{}}}
\subsection{\code{StreamingBody$read()}}{
  Read \code{n} bytes into a raw vector.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$read(n)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of bytes to read}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-read_all"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-read_all}{}}}
\subsection{\code{StreamingBody$read_all()}}{
  Read all bytes and close the connection.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$read_all(buffer = 32 * 1024)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{buffer}}{Buffer size, in bytes.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-is_open"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-is_open}{}}}
\subsection{\code{StreamingBody$is_open()}}{
  Is the connection still open?
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$is_open()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-is_complete"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-is_complete}{}}}
\subsection{\code{StreamingBody$is_complete()}}{
  Is the connection complete? (i.e. is there data remaining
to be read?)
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$is_complete()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-get_fdset"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-get_fdset}{}}}
\subsection{\code{StreamingBody$get_fdset()}}{
  Get the active file descriptions and timeout from the
handle. Wrapper around \code{\link[curl:multi_fdset]{curl::multi_fdset()}}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$get_fdset()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-close"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-close}{}}}
\subsection{\code{StreamingBody$close()}}{
  Close the connection
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$close()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-StreamingBody-clone"></a>}}
\if{latex}{\out{\hypertarget{method-StreamingBody-clone}{}}}
\subsection{\code{StreamingBody$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{StreamingBody$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req.R
\name{request}
\alias{request}
\title{Create a new HTTP request}
\usage{
request(base_url)
}
\arguments{
\item{base_url}{Base URL for request.}
}
\value{
An HTTP request: an S3 list with class \code{httr2_request}.
}
\description{
There are three steps needed to perform a HTTP request with httr2:
\enumerate{
\item Create a request object with \code{request(url)} (this function).
\item Define its behaviour with \code{req_} functions, e.g.:
\itemize{
\item \code{\link[=req_headers]{req_headers()}} to set header values.
\item \code{\link[=req_url_path]{req_url_path()}} and friends to modify the url.
\item \code{\link[=req_body_json]{req_body_json()}} and friends to add a body.
\item \code{\link[=req_auth_basic]{req_auth_basic()}} to perform basic HTTP authentication.
\item \code{\link[=req_oauth_auth_code]{req_oauth_auth_code()}} to use the OAuth auth code flow.
}
\item Perform the request and fetch the response with \code{\link[=req_perform]{req_perform()}}.
}
}
\examples{
request("http://r-project.org")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-headers.R
\name{req_headers}
\alias{req_headers}
\alias{req_headers_redacted}
\title{Modify request headers}
\usage{
req_headers(.req, ..., .redact = NULL)

req_headers_redacted(.req, ...)
}
\arguments{
\item{.req}{A \link{request}.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> Name-value pairs of headers
and their values.
\itemize{
\item Use \code{NULL} to reset a value to httr2's default.
\item Use \code{""} to remove a header.
\item Use a character vector to repeat a header.
}}

\item{.redact}{A character vector of headers to redact. The Authorization
header is always redacted.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{req_headers()} allows you to set the value of any header.

\code{req_headers_redacted()} is a variation that adds "redacted" headers, which
httr2 avoids printing on the console. This is good practice for
authentication headers to avoid accidentally leaking them in log files.
}
\examples{
req <- request("http://example.com")

# Use req_headers() to add arbitrary additional headers to the request
req |>
  req_headers(MyHeader = "MyValue") |>
  req_dry_run()

# Repeated use overrides the previous value:
req |>
  req_headers(MyHeader = "Old value") |>
  req_headers(MyHeader = "New value") |>
  req_dry_run()

# Setting Accept to NULL uses curl's default:
req |>
  req_headers(Accept = NULL) |>
  req_dry_run()

# Setting it to "" removes it:
req |>
  req_headers(Accept = "") |>
  req_dry_run()

# If you need to repeat a header, provide a vector of values
# (this is rarely needed, but is important in a handful of cases)
req |>
  req_headers(HeaderName = c("Value 1", "Value 2", "Value 3")) |>
  req_dry_run()

# If you have headers in a list, use !!!
headers <- list(HeaderOne = "one", HeaderTwo = "two")
req |>
  req_headers(!!!headers, HeaderThree = "three") |>
  req_dry_run()

# Use `req_headers_redacted()`` to hide a header in the output
req_secret <- req |>
  req_headers_redacted(Secret = "this-is-private") |>
  req_headers(Public = "but-this-is-not")

req_secret
req_secret |> req_dry_run()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-stream.R
\name{req_perform_stream}
\alias{req_perform_stream}
\title{Perform a request and handle data as it streams back}
\usage{
req_perform_stream(
  req,
  callback,
  timeout_sec = Inf,
  buffer_kb = 64,
  round = c("byte", "line")
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{callback}{A single argument callback function. It will be called
repeatedly with a raw vector whenever there is at least \code{buffer_kb}
worth of data to process. It must return \code{TRUE} to continue streaming.}

\item{timeout_sec}{Number of seconds to process stream for.}

\item{buffer_kb}{Buffer size, in kilobytes.}

\item{round}{How should the raw vector sent to \code{callback} be rounded?
Choose \code{"byte"}, \code{"line"}, or supply your own function that takes a
raw vector of \code{bytes} and returns the locations of possible cut points
(or \code{integer()} if there are none).}
}
\value{
An HTTP \link{response}. The body will be empty if the request was
successful (since the \code{callback} function will have handled it). The body
will contain the HTTP response body if the request was unsuccessful.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

Please use \code{\link[=req_perform_connection]{req_perform_connection()}} instead.

After preparing a request, call \code{req_perform_stream()} to perform the request
and handle the result with a streaming callback. This is useful for
streaming HTTP APIs where potentially the stream never ends.

The \code{callback} will only be called if the result is successful. If you need
to stream an error response, you can use \code{\link[=req_error]{req_error()}} to suppress error
handling so that the body is streamed to you.
}
\examples{
# PREVIOSULY
show_bytes <- function(x) {
  cat("Got ", length(x), " bytes\n", sep = "")
  TRUE
}
resp <- request(example_url()) |>
  req_url_path("/stream-bytes/100000") |>
  req_perform_stream(show_bytes, buffer_kb = 32)

# NOW
resp <- request(example_url()) |>
  req_url_path("/stream-bytes/100000") |>
  req_perform_connection()
while (!resp_stream_is_complete(resp)) {
  x <-  resp_stream_raw(resp, kb = 32)
  cat("Got ", length(x), " bytes\n", sep = "")
}
close(resp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/secret.R
\name{obfuscate}
\alias{obfuscate}
\alias{obfuscated}
\title{Obfuscate mildly secret information}
\usage{
obfuscate(x)

obfuscated(x)
}
\arguments{
\item{x}{A string to \code{obfuscate}, or mark as \code{obfuscated}.}
}
\value{
\code{obfuscate()} prints the \code{obfuscated()} call to include in your
code. \code{obfuscated()} returns an S3 class marking the string as obfuscated
so it can be unobfuscated when needed.
}
\description{
Use \code{obfuscate("value")} to generate a call to \code{obfuscated()}, which will
unobfuscate the value at the last possible moment. Obfuscated values only
work in limited locations:
\itemize{
\item The \code{secret} argument to \code{\link[=oauth_client]{oauth_client()}}
\item Elements of the \code{data} argument to \code{\link[=req_body_form]{req_body_form()}}, \code{req_body_json()},
and \code{req_body_multipart()}.
}

Working together this pair of functions provides a way to obfuscate mildly
confidential information, like OAuth client secrets. The secret can not be
revealed from your inspecting source code, but a skilled R programmer could
figure it out with some effort. The main goal is to protect against scraping;
there's no way for an automated tool to grab your obfuscated secrets.
}
\examples{
obfuscate("good morning")

# Every time you obfuscate you'll get a different value because it
# includes 16 bytes of random data which protects against certain types of
# brute force attack
obfuscate("good morning")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-client-credentials.R
\name{oauth_flow_client_credentials}
\alias{oauth_flow_client_credentials}
\title{OAuth flow: client credentials}
\usage{
oauth_flow_client_credentials(client, scope = NULL, token_params = list())
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
An \link{oauth_token}.
}
\description{
This function implements the OAuth client credentials flow, as defined
by \href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.4}{rfc6749},
Section 4.4. It is used to allow the client to access resources that it
controls directly, not on behalf of an user.
}
\seealso{
Other OAuth flows: 
\code{\link{oauth_flow_auth_code}()},
\code{\link{oauth_flow_bearer_jwt}()},
\code{\link{oauth_flow_device}()},
\code{\link{oauth_flow_password}()},
\code{\link{oauth_flow_refresh}()}
}
\concept{OAuth flows}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-auth.R
\name{req_auth_bearer_token}
\alias{req_auth_bearer_token}
\title{Authenticate request with bearer token}
\usage{
req_auth_bearer_token(req, token)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{token}{A bearer token}
}
\value{
A modified HTTP \link{request}.
}
\description{
A bearer token gives the bearer access to confidential resources
(so you should keep them secure like you would with a user name and
password). They are usually produced by some large authentication scheme
(like the various OAuth 2.0 flows), but you are sometimes given then
directly.
}
\examples{
req <- request("http://example.com") |> req_auth_bearer_token("sdaljsdf093lkfs")
req

# httr2 does its best to redact the Authorization header so that you don't
# accidentally reveal confidential data. Use `redact_headers` to reveal it:
print(req, redact_headers = FALSE)
}
\seealso{
See \href{https://datatracker.ietf.org/doc/html/rfc6750}{RFC 6750} for more details about bearer token usage
with OAuth 2.0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-refresh.R
\name{oauth_flow_refresh}
\alias{oauth_flow_refresh}
\title{OAuth flow: refresh token}
\usage{
oauth_flow_refresh(
  client,
  refresh_token = Sys.getenv("HTTR2_REFRESH_TOKEN"),
  scope = NULL,
  token_params = list()
)
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{refresh_token}{A refresh token. This is equivalent to a password
so shouldn't be typed into the console or stored in a script. Instead,
we recommend placing in an environment variable; the default behaviour
is to look in \code{HTTR2_REFRESH_TOKEN}.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
An \link{oauth_token}.
}
\description{
This function generates an access token from a refresh token, following
the process described in
\href{https://datatracker.ietf.org/doc/html/rfc6749#section-6}{rfc6749},
Section 6. Errors if the refresh returns a new refresh token, see
\code{\link[=req_oauth_refresh]{req_oauth_refresh()}} for details.
}
\seealso{
Other OAuth flows: 
\code{\link{oauth_flow_auth_code}()},
\code{\link{oauth_flow_bearer_jwt}()},
\code{\link{oauth_flow_client_credentials}()},
\code{\link{oauth_flow_device}()},
\code{\link{oauth_flow_password}()}
}
\concept{OAuth flows}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iterate-responses.R
\name{resps_successes}
\alias{resps_successes}
\alias{resps_failures}
\alias{resps_ok}
\alias{resps_requests}
\alias{resps_data}
\title{Tools for working with lists of responses}
\usage{
resps_successes(resps)

resps_failures(resps)

resps_ok(resps)

resps_requests(resps)

resps_data(resps, resp_data)
}
\arguments{
\item{resps}{A list of responses (possibly including errors).}

\item{resp_data}{A function that takes a response (\code{resp}) and
returns the data found inside that response as a vector or data frame.

NB: If you're using \code{\link[=resp_body_raw]{resp_body_raw()}}, you're likely to want to wrap its
output in \code{list()} to avoid combining all the bodies into a single raw
vector, e.g. \verb{resps |> resps_data(\\(resp) list(resp_body_raw(resp)))}.}
}
\description{
These functions provide a basic toolkit for working with lists of
responses and possibly errors, as returned by \code{\link[=req_perform_parallel]{req_perform_parallel()}},
\code{\link[=req_perform_sequential]{req_perform_sequential()}}, and \code{\link[=req_perform_iterative]{req_perform_iterative()}}.
\itemize{
\item \code{resps_successes()} returns a list of successful responses.
\item \code{resps_failures()} returns a list of failed responses (i.e. errors).
\item \code{resps_ok()} returns a logical vector indicating which requests were
successful.
\item \code{resps_requests()} returns the list of requests that corresponds to
each response.
\item \code{resps_data()} returns all the data in a single vector or data frame.
}
}
\examples{
reqs <- list(
  request(example_url()) |> req_url_path("/ip"),
  request(example_url()) |> req_url_path("/user-agent"),
  request(example_url()) |> req_template("/status/:status", status = 404),
  request("INVALID")
)
resps <- req_perform_parallel(reqs, on_error = "continue")

# find successful responses
resps |> resps_successes()

# collect all their data
resps |>
  resps_successes() |>
  resps_data(\(resp) resp_body_json(resp))

# find requests corresponding to failure responses
resps |>
  resps_failures() |>
  resps_requests()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth_cache_clear}
\alias{oauth_cache_clear}
\title{Clear OAuth cache}
\usage{
oauth_cache_clear(client, cache_disk = FALSE, cache_key = NULL)
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{cache_disk}{Should the access token be cached on disk? This reduces
the number of times that you need to re-authenticate at the cost of
storing access credentials on disk.

Learn more in \url{https://httr2.r-lib.org/articles/oauth.html}.}

\item{cache_key}{If you want to cache multiple tokens per app, use this
key to disambiguate them.}
}
\description{
Use this function to clear cached credentials.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-client.R
\name{oauth_client}
\alias{oauth_client}
\title{Create an OAuth client}
\usage{
oauth_client(
  id,
  token_url,
  secret = NULL,
  key = NULL,
  auth = c("body", "header", "jwt_sig"),
  auth_params = list(),
  name = hash(id)
)
}
\arguments{
\item{id}{Client identifier.}

\item{token_url}{Url to retrieve an access token.}

\item{secret}{Client secret. For most apps, this is technically confidential
so in principle you should avoid storing it in source code. However, many
APIs require it in order to provide a user friendly authentication
experience, and the risks of including it are usually low. To make things
a little safer, I recommend using \code{\link[=obfuscate]{obfuscate()}} when recording the client
secret in public code.}

\item{key}{Client key. As an alternative to using a \code{secret}, you can
instead supply a confidential private key. This should never be included
in a package.}

\item{auth}{Authentication mechanism used by the client to prove itself to
the API. Can be one of three built-in methods ("body", "header", or "jwt"),
or a function that will be called with arguments \code{req}, \code{client}, and
the contents of \code{auth_params}.

The most common mechanism in the wild is \code{"body"} where the \code{client_id} and
(optionally) \code{client_secret} are added to the body. \code{"header"} sends the
\code{client_id} and \code{client_secret} in HTTP Authorization header. \code{"jwt_sig"}
will generate a JWT, and include it in a \code{client_assertion} field in the
body.

See \code{\link[=oauth_client_req_auth]{oauth_client_req_auth()}} for more details.}

\item{auth_params}{Additional parameters passed to the function specified
by \code{auth}.}

\item{name}{Optional name for the client. Used when generating the cache
directory. If \code{NULL}, generated from hash of \code{client_id}. If you're
defining a client for use in a package, I recommend that you use
the package name.}
}
\value{
An OAuth client: An S3 list with class \code{httr2_oauth_client}.
}
\description{
An OAuth app is the combination of a client, a set of endpoints
(i.e. urls where various requests should be sent), and an authentication
mechanism. A client consists of at least a \code{client_id}, and also often
a \code{client_secret}. You'll get these values when you create the client on
the API's website.
}
\examples{
oauth_client("myclient", "http://example.com/token_url", secret = "DONTLOOK")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-cookies.R
\name{req_cookie_preserve}
\alias{req_cookie_preserve}
\alias{req_cookies_set}
\title{Set and preserve cookies}
\usage{
req_cookie_preserve(req, path)

req_cookies_set(req, ...)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{path}{A path to a file where cookies will be read from before and updated after the request.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}>
Name-value pairs that define query parameters. Each value must be
an atomic vector, which is automatically escaped. To opt-out of escaping,
wrap strings in \code{I()}.}
}
\description{
Use \code{req_cookie_set()} to set client side cookies that are sent to the
server.

By default, httr2 uses a clean slate for every request meaning that cookies
are not automatically preserved across requests. To preserve cookies, use
\code{req_cookie_preserve()} along with the path to cookie file that will be
read before and updated after each request.
}
\examples{
# Use `req_cookies_set()` to set client-side cookies
request(example_url()) |>
  req_cookies_set(a = 1, b = 1) |>
  req_dry_run()

# Use `req_cookie_preserve()` to preserve server-side cookies across requests
path <- tempfile()

# Set a server-side cookie
request(example_url()) |>
  req_cookie_preserve(path) |>
  req_template("/cookies/set/:name/:value", name = "chocolate", value = "chip") |>
  req_perform() |>
  resp_body_json()

# Set another sever-side cookie
request(example_url()) |>
  req_cookie_preserve(path) |>
  req_template("/cookies/set/:name/:value", name = "oatmeal", value = "raisin") |>
  req_perform() |>
  resp_body_json()

# Add a client side cookie
request(example_url()) |>
  req_url_path("/cookies/set") |>
  req_cookie_preserve(path) |>
  req_cookies_set(snicker = "doodle") |>
  req_perform() |>
  resp_body_json()

# The cookie path has a straightforward format
cat(readChar(path, nchars = 1e4))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-body.R
\name{req_body}
\alias{req_body_raw}
\alias{req_body_file}
\alias{req_body_json}
\alias{req_body_json_modify}
\alias{req_body_form}
\alias{req_body_multipart}
\title{Send data in request body}
\usage{
req_body_raw(req, body, type = "")

req_body_file(req, path, type = "")

req_body_json(
  req,
  data,
  auto_unbox = TRUE,
  digits = 22,
  null = "null",
  type = "application/json",
  ...
)

req_body_json_modify(req, ...)

req_body_form(.req, ..., .multi = c("error", "comma", "pipe", "explode"))

req_body_multipart(.req, ...)
}
\arguments{
\item{req, .req}{A httr2 \link{request} object.}

\item{body}{A literal string or raw vector to send as body.}

\item{type}{MIME content type. The default, \code{""}, will not emit a
\code{Content-Type} header.  Ignored if you have set a \code{Content-Type} header
with \code{\link[=req_headers]{req_headers()}}.}

\item{path}{Path to file to upload.}

\item{data}{Data to include in body.}

\item{auto_unbox}{Should length-1 vectors be automatically "unboxed" to
JSON scalars?}

\item{digits}{How many digits of precision should numbers use in JSON?}

\item{null}{Should \code{NULL} be translated to JSON's null (\code{"null"})
or an empty list (\code{"list"}).}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}> Name-data pairs used to send
data in the body.
\itemize{
\item For \code{req_body_form()}, the values must be strings (or things easily
coerced to strings). Vectors are converted to strings using the
value of \code{.multi}.
\item For \code{req_body_multipart()} the values must be strings or objects
produced by \code{\link[curl:multipart]{curl::form_file()}}/\code{\link[curl:multipart]{curl::form_data()}}.
\item For \code{req_body_json_modify()}, any simple data made from atomic vectors
and lists.
}

\code{req_body_json()} uses this argument differently; it takes additional
arguments passed on to  \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}.}

\item{.multi}{Controls what happens when a value is a vector:
\itemize{
\item \code{"error"}, the default, throws an error.
\item \code{"comma"}, separates values with a \verb{,}, e.g. \verb{?x=1,2}.
\item \code{"pipe"}, separates values with a \code{|}, e.g. \code{?x=1|2}.
\item \code{"explode"}, turns each element into its own parameter, e.g. \code{?x=1&x=2}
}

If none of these options work for your needs, you can instead supply a
function that takes a character vector of argument values and returns a
a single string.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\itemize{
\item \code{req_body_file()} sends a local file.
\item \code{req_body_raw()} sends a string or raw vector.
\item \code{req_body_json()} sends JSON encoded data. Named components of this data
can later be modified with \code{req_body_json_modify()}.
\item \code{req_body_form()} sends form encoded data.
\item \code{req_body_multipart()} creates a multi-part body.
}

Adding a body to a request will automatically switch the method to POST.
}
\examples{
req <- request(example_url()) |>
  req_url_path("/post")

# Most APIs expect small amounts of data in either form or json encoded:
req |>
  req_body_form(x = "A simple text string") |>
  req_dry_run()

req |>
  req_body_json(list(x = "A simple text string")) |>
  req_dry_run()

# For total control over the body, send a string or raw vector
req |>
  req_body_raw("A simple text string") |>
  req_dry_run()

# There are two main ways that APIs expect entire files
path <- tempfile()
writeLines(letters[1:6], path)

# You can send a single file as the body:
req |>
  req_body_file(path) |>
  req_dry_run()

# You can send multiple files, or a mix of files and data
# with multipart encoding
req |>
  req_body_multipart(a = curl::form_file(path), b = "some data") |>
  req_dry_run()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-method.R
\name{req_method}
\alias{req_method}
\title{Set HTTP method in request}
\usage{
req_method(req, method)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{method}{Custom HTTP method}
}
\value{
A modified HTTP \link{request}.
}
\description{
Use this function to use a custom HTTP method like \code{HEAD},
\code{DELETE}, \code{PATCH}, \code{UPDATE}, or \code{OPTIONS}. The default method is
\code{GET} for requests without a body, and \code{POST} for requests with a body.
}
\examples{
request(example_url()) |> req_method("PATCH")
request(example_url()) |> req_method("PUT")
request(example_url()) |> req_method("HEAD")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-client-credentials.R
\name{req_oauth_client_credentials}
\alias{req_oauth_client_credentials}
\alias{oauth_flow_client_credentials}
\title{OAuth with client credentials}
\usage{
req_oauth_client_credentials(req, client, scope = NULL, token_params = list())

oauth_flow_client_credentials(client, scope = NULL, token_params = list())
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
\code{req_oauth_client_credentials()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_client_credentials()} returns an \link{oauth_token}.
}
\description{
Authenticate using OAuth \strong{client credentials flow}, as defined by
\href{https://datatracker.ietf.org/doc/html/rfc6749#section-4.4}{Section 4.4 of RFC 6749}. It is used to allow the client to access resources that
it controls directly, not on behalf of an user.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}.
}
\examples{
req_auth <- function(req) {
  req_oauth_client_credentials(
    req,
    client = oauth_client("example", "https://example.com/get_token")
  )
}

request("https://example.com") |>
  req_auth()
}
\seealso{
Other OAuth flows: 
\code{\link{req_oauth_auth_code}()},
\code{\link{req_oauth_bearer_jwt}()},
\code{\link{req_oauth_password}()},
\code{\link{req_oauth_refresh}()},
\code{\link{req_oauth_token_exchange}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.R
\name{example_url}
\alias{example_url}
\alias{example_github_client}
\title{Code for examples}
\usage{
example_url(path = "/")

example_github_client()
}
\description{
\code{example_url()} runs a simple websever using the webfakes package with the
following endpoints:
\itemize{
\item all the ones from the \code{\link[webfakes:httpbin_app]{webfakes::httpbin_app()}}
\item \verb{/iris}: paginate through the iris dataset. It has the query parameters
\code{page} and \code{limit} to control the pagination.
}

\code{example_github_client()} is an OAuth client for GitHub.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-headers.R
\name{req_get_headers}
\alias{req_get_headers}
\title{Get request headers}
\usage{
req_get_headers(req, redacted = c("drop", "redact", "reveal"))
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{redacted}{What to do with redacted headers?
\itemize{
\item \code{"drop"} (the default) drops them.
\item \code{"redact"} replaces them with \verb{<REDACTED>}.
\item \code{"reveal"} leaves them in place.
}}
}
\value{
A named list.
}
\description{
Retrieve custom headers set on the request. Use \code{\link[=req_dry_run]{req_dry_run()}} to get all
headers, including those automatically generated by curl.
}
\examples{
req <- request("http://example.com")
req <- req_headers(req, a = 1L, b = 2L, .redact = "a")

req_get_headers(req, "drop")
req_get_headers(req, "redact")
req_get_headers(req, "reveal")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-headers.R
\name{resp_retry_after}
\alias{resp_retry_after}
\title{Extract wait time from a response}
\usage{
resp_retry_after(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\value{
Scalar double giving the number of seconds to wait before retrying
a request.
}
\description{
Computes how many seconds you should wait before retrying a request by
inspecting the \code{Retry-After} header. It parses both forms (absolute and
relative) and returns the number of seconds to wait. If the heading is not
found, it will return \code{NA}.
}
\examples{
resp <- response(headers = "Retry-After: 30")
resp |> resp_retry_after()

resp <- response(headers = "Retry-After: Mon, 20 Sep 2025 21:44:05 UTC")
resp |> resp_retry_after()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-jwt.R
\name{req_oauth_bearer_jwt}
\alias{req_oauth_bearer_jwt}
\alias{oauth_flow_bearer_jwt}
\title{OAuth with a bearer JWT (JSON web token)}
\usage{
req_oauth_bearer_jwt(
  req,
  client,
  claim,
  signature = "jwt_encode_sig",
  signature_params = list(),
  scope = NULL,
  token_params = list()
)

oauth_flow_bearer_jwt(
  client,
  claim,
  signature = "jwt_encode_sig",
  signature_params = list(),
  scope = NULL,
  token_params = list()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{claim}{A list of claims. If all elements of the claim set are static
apart from \code{iat}, \code{nbf}, \code{exp}, or \code{jti}, provide a list and
\code{\link[=jwt_claim]{jwt_claim()}} will automatically fill in the dynamic components.
If other components need to vary, you can instead provide a zero-argument
callback function which should call \code{jwt_claim()}.}

\item{signature}{Function use to sign \code{claim}, e.g. \code{\link[=jwt_encode_sig]{jwt_encode_sig()}}.}

\item{signature_params}{Additional arguments passed to \code{signature}, e.g.
\code{size}, \code{header}.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}
}
\value{
\code{req_oauth_bearer_jwt()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_bearer_jwt()} returns an \link{oauth_token}.
}
\description{
Authenticate using a \strong{Bearer JWT} (JSON web token) as an authorization
grant to get an access token, as defined by \href{https://datatracker.ietf.org/doc/html/rfc7523#section-2.1}{Section 2.1 of RFC 7523}.
It is often used for service accounts, accounts that are used primarily in
automated environments.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}.
}
\examples{
req_auth <- function(req) {
  req_oauth_bearer_jwt(
    req,
    client = oauth_client("example", "https://example.com/get_token"),
    claim = jwt_claim()
  )
}

request("https://example.com") |>
 req_auth()
}
\seealso{
Other OAuth flows: 
\code{\link{req_oauth_auth_code}()},
\code{\link{req_oauth_client_credentials}()},
\code{\link{req_oauth_password}()},
\code{\link{req_oauth_refresh}()},
\code{\link{req_oauth_token_exchange}()}
}
\concept{OAuth flows}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-headers.R
\name{resp_date}
\alias{resp_date}
\title{Extract request date from response}
\usage{
resp_date(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\value{
A \code{POSIXct} date-time.
}
\description{
All responses contain a request date in the \code{Date} header; if not provided
by the server will be automatically added by httr2.
}
\examples{
resp <- response(headers = "Date: Wed, 01 Jan 2020 09:23:15 UTC")
resp |> resp_date()

# If server doesn't add header (unusual), you get the time the request
# was created:
resp <- response()
resp |> resp_date()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-options.R
\name{req_user_agent}
\alias{req_user_agent}
\title{Set user-agent for a request}
\usage{
req_user_agent(req, string = NULL)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{string}{String to be sent in the \code{User-Agent} header. If \code{NULL},
will user default.}
}
\value{
A modified HTTP \link{request}.
}
\description{
This overrides the default user-agent set by httr2 which includes the
version numbers of httr2, the curl package, and libcurl.
}
\examples{
# Default user-agent:
request("http://example.com") |> req_dry_run()

request("http://example.com") |> req_user_agent("MyString") |> req_dry_run()

# If you're wrapping in an API in a package, it's polite to set the
# user agent to identify your package.
request("http://example.com") |>
  req_user_agent("MyPackage (http://mypackage.com)") |>
  req_dry_run()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-verbose.R
\name{req_verbose}
\alias{req_verbose}
\title{Show extra output when request is performed}
\usage{
req_verbose(
  req,
  header_req = TRUE,
  header_resp = TRUE,
  body_req = FALSE,
  body_resp = FALSE,
  info = FALSE,
  redact_headers = TRUE
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{header_req, header_resp}{Show request/response headers?}

\item{body_req, body_resp}{Should request/response bodies? When the response
body is compressed, this will show the number of bytes received in
each "chunk".}

\item{info}{Show informational text from curl? This is mainly useful
for debugging https and auth problems, so is disabled by default.}

\item{redact_headers}{Redact confidential data in the headers? Currently
redacts the contents of the Authorization header to prevent you from
accidentally leaking credentials when debugging/reprexing.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{req_verbose()} uses the following prefixes to distinguish between
different components of the HTTP requests and responses:
\itemize{
\item \verb{* } informative curl messages
\item \verb{->} request headers
\item \verb{>>} request body
\item \verb{<-} response headers
\item \verb{<<} response body
}
}
\examples{
# Use `req_verbose()` to see the headers that are sent back and forth when
# making a request
resp <- request("https://httr2.r-lib.org") |>
  req_verbose() |>
  req_perform()

# Or use one of the convenient shortcuts:
resp <- request("https://httr2.r-lib.org") |>
  req_perform(verbosity = 1)
}
\seealso{
\code{\link[=req_perform]{req_perform()}} which exposes a limited subset of these options
through the \code{verbosity} argument and \code{\link[=with_verbosity]{with_verbosity()}} which allows you
to control the verbosity of requests deeper within the call stack.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iterate-helpers.R
\name{signal_total_pages}
\alias{signal_total_pages}
\title{Signal total number pages}
\usage{
signal_total_pages(n)
}
\arguments{
\item{n}{Total number of pages.}
}
\description{
To be called within a \code{next_req} callback function used with
\code{\link[=req_perform_iterative]{req_perform_iterative()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-device.R
\name{req_oauth_device}
\alias{req_oauth_device}
\alias{oauth_flow_device}
\title{OAuth with device flow}
\usage{
req_oauth_device(
  req,
  client,
  auth_url,
  scope = NULL,
  open_browser = is_interactive(),
  auth_params = list(),
  token_params = list(),
  cache_disk = FALSE,
  cache_key = NULL
)

oauth_flow_device(
  client,
  auth_url,
  pkce = FALSE,
  scope = NULL,
  open_browser = is_interactive(),
  auth_params = list(),
  token_params = list()
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{auth_url}{Authorization url; you'll need to discover this by reading
the documentation.}

\item{scope}{Scopes to be requested from the resource owner.}

\item{open_browser}{If \code{TRUE} (the default in interactive sessions), the
device verification URL will be opened in the user's browser. If \code{FALSE},
the URL is printed to the console and the user must open it themselves.}

\item{auth_params}{A list containing additional parameters passed to
\code{\link[=oauth_flow_auth_code_url]{oauth_flow_auth_code_url()}}.}

\item{token_params}{List containing additional parameters passed to the
\code{token_url}.}

\item{cache_disk}{Should the access token be cached on disk? This reduces
the number of times that you need to re-authenticate at the cost of
storing access credentials on disk.

Learn more in \url{https://httr2.r-lib.org/articles/oauth.html}.}

\item{cache_key}{If you want to cache multiple tokens per app, use this
key to disambiguate them.}

\item{pkce}{Use "Proof Key for Code Exchange"? This adds an extra layer of
security and should always be used if supported by the server.}
}
\value{
\code{req_oauth_device()} returns a modified HTTP \link{request} that will
use OAuth; \code{oauth_flow_device()} returns an \link{oauth_token}.
}
\description{
Authenticate using the OAuth \strong{device flow}, as defined by \href{https://datatracker.ietf.org/doc/html/rfc8628}{RFC 8628}.
It's designed for devices that don't have access to a web browser (if you've
ever authenticated an app on your TV, this is probably the flow you've used),
but it also works well from within R.

Learn more about the overall OAuth authentication flow in
\url{https://httr2.r-lib.org/articles/oauth.html}.
}
\examples{
req_auth_github <- function(req) {
  req_oauth_device(
    req,
    client = example_github_client(),
    auth_url = "https://github.com/login/device/code"
  )
}

request("https://api.github.com/user") |>
  req_auth_github()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-dry-run.R
\name{req_dry_run}
\alias{req_dry_run}
\title{Perform a dry run}
\usage{
req_dry_run(
  req,
  quiet = FALSE,
  redact_headers = TRUE,
  testing_headers = is_testing(),
  pretty_json = getOption("httr2_pretty_json", TRUE)
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{quiet}{If \code{TRUE} doesn't print anything.}

\item{redact_headers}{Redact confidential data in the headers? Currently
redacts the contents of the Authorization header to prevent you from
accidentally leaking credentials when debugging/reprexing.}

\item{testing_headers}{If \code{TRUE}, removes headers that httr2 would otherwise
be automatically added, which are likely to change across test runs. This
currently includes:
\itemize{
\item The default \code{User-Agent}, which varies based on libcurl, curl, and
httr2 versions.
\item The `Host`` header, which is often set to a testing server.
\item The \code{Content-Length} header, which will often vary by platform because
of varying newline encodings. (And is also not correct if you have
\code{pretty_json = TRUE}.)
\item The \code{Accept-Encoding} header, which varies based on how libcurl was
built.
}}

\item{pretty_json}{If \code{TRUE}, automatically prettify JSON bodies.}
}
\value{
Invisibly, a list containing information about the request,
including \code{method}, \code{path}, and \code{headers}.
}
\description{
This shows you exactly what httr2 will send to the server, without
actually sending anything. It requires the httpuv package because it
works by sending the real HTTP request to a local webserver, thanks to
the magic of \code{\link[curl:curl_echo]{curl::curl_echo()}}.
}
\details{
\subsection{Limitations}{
\itemize{
\item The HTTP version is always \code{HTTP/1.1} (since you can't determine what it
will actually be without connecting to the real server).
}
}
}
\examples{
# httr2 adds default User-Agent, Accept, and Accept-Encoding headers
request("http://example.com") |> req_dry_run()

# the Authorization header is automatically redacted to avoid leaking
# credentials on the console
req <- request("http://example.com") |> req_auth_basic("user", "password")
req |> req_dry_run()

# if you need to see it, use redact_headers = FALSE
req |> req_dry_run(redact_headers = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-client.R
\name{oauth_client_req_auth}
\alias{oauth_client_req_auth}
\alias{oauth_client_req_auth_header}
\alias{oauth_client_req_auth_body}
\alias{oauth_client_req_auth_jwt_sig}
\title{OAuth client authentication}
\usage{
oauth_client_req_auth(req, client)

oauth_client_req_auth_header(req, client)

oauth_client_req_auth_body(req, client)

oauth_client_req_auth_jwt_sig(req, client, claim, size = 256, header = list())
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{client}{An \link{oauth_client}.}

\item{claim}{Claim set produced by \code{\link[=jwt_claim]{jwt_claim()}}.}

\item{size}{Size, in bits, of sha2 signature, i.e. 256, 384 or 512.
Only for HMAC/RSA, not applicable for ECDSA keys.}

\item{header}{A named list giving additional fields to include in the
JWT header.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{oauth_client_req_auth()} authenticates a request using the authentication
strategy defined by the \code{auth} and \code{auth_param} arguments to \code{\link[=oauth_client]{oauth_client()}}.
This is used to authenticate the client as part of the OAuth flow, \strong{not}
to authenticate a request on behalf of a user.

There are three built-in strategies:
\itemize{
\item \code{oauth_client_req_body()} adds the client id and (optionally) the secret
to the request body, as described in \href{https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1}{Section 2.3.1 of RFC 6749}.
\item \code{oauth_client_req_header()} adds the client id and secret using HTTP
basic authentication with the \code{Authorization} header, as described
in \href{https://datatracker.ietf.org/doc/html/rfc6749#section-2.3.1}{Section 2.3.1 of RFC 6749}.
\item \code{oauth_client_jwt_rs256()} adds a client assertion to the body using a
JWT signed with \code{jwt_sign_rs256()} using a private key, as described
in \href{https://datatracker.ietf.org/doc/html/rfc7523#section-2.2}{Section 2.2 of RFC 7523}.
}

You will generally not call these functions directly but will instead
specify them through the \code{auth} argument to \code{\link[=oauth_client]{oauth_client()}}. The \code{req} and
\code{client} parameters are automatically filled in; other parameters come from
the \code{auth_params} argument.
}
\examples{
# Show what the various forms of client authentication look like
req <- request("https://example.com/whoami")

client1 <- oauth_client(
  id = "12345",
  secret = "56789",
  token_url = "https://example.com/oauth/access_token",
  name = "oauth-example",
  auth = "body" # the default
)
# calls oauth_client_req_auth_body()
req_dry_run(oauth_client_req_auth(req, client1))

client2 <- oauth_client(
  id = "12345",
  secret = "56789",
  token_url = "https://example.com/oauth/access_token",
  name = "oauth-example",
  auth = "header"
)
# calls oauth_client_req_auth_header()
req_dry_run(oauth_client_req_auth(req, client2))

client3 <- oauth_client(
  id = "12345",
  key = openssl::rsa_keygen(),
  token_url = "https://example.com/oauth/access_token",
  name = "oauth-example",
  auth = "jwt_sig",
  auth_params = list(claim = jwt_claim())
)
# calls oauth_client_req_auth_header_jwt_sig()
req_dry_run(oauth_client_req_auth(req, client3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-auth-code.R
\name{oauth_flow_auth_code_url}
\alias{oauth_flow_auth_code_url}
\alias{oauth_flow_auth_code_listen}
\alias{oauth_flow_auth_code_parse}
\alias{oauth_flow_auth_code_pkce}
\title{OAuth authorization code components}
\usage{
oauth_flow_auth_code_url(
  client,
  auth_url,
  redirect_uri = NULL,
  scope = NULL,
  state = NULL,
  auth_params = list()
)

oauth_flow_auth_code_listen(redirect_uri = "http://localhost:1410")

oauth_flow_auth_code_parse(query, state)

oauth_flow_auth_code_pkce()
}
\arguments{
\item{state}{Random state generated by \code{oauth_flow_auth_code()}. Used to
verify that we're working with an authentication request that we created.
(This is an unlikely threat for R packages since the webserver that
listens for authorization responses is transient.)}

\item{query}{List of query parameters returned by \code{oauth_flow_auth_code_listen()}.}
}
\description{
These low-level functions can be used to assemble a custom flow for
APIs that are further from the spec:
\itemize{
\item \code{oauth_flow_auth_code_url()} generates the url that should be opened in a
browser.
\item \code{oauth_flow_auth_code_listen()} starts a temporary local webserver that
listens for the response from the resource server.
\item \code{oauth_flow_auth_code_parse()} parses the query parameters returned from
the server redirect, verifying that the \code{state} is correct, and returning
the authorisation code.
\item \code{oauth_flow_auth_code_pkce()} generates code verifier, method, and challenge
components as needed for PKCE, as defined in \href{https://datatracker.ietf.org/doc/html/rfc7636}{RFC 7636}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is-online.R
\name{is_online}
\alias{is_online}
\title{Is your computer currently online?}
\usage{
is_online()
}
\description{
This function uses some cheap heuristics to determine if your computer is
currently online. It's a simple wrapper around \code{\link[curl:nslookup]{curl::has_internet()}}
exported from httr2 for convenience.
}
\examples{
is_online()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-progress.R
\name{req_progress}
\alias{req_progress}
\title{Add a progress bar to long downloads or uploads}
\usage{
req_progress(req, type = c("down", "up"))
}
\arguments{
\item{req}{A \link{request}.}

\item{type}{Type of progress to display: either number of bytes uploaded
or downloaded.}
}
\description{
When uploading or downloading a large file, it's often useful to
provide a progress bar so that you know how long you have to wait.
}
\examples{
req <- request("https://r4ds.s3.us-west-2.amazonaws.com/seattle-library-checkouts.csv") |>
  req_progress()

\dontrun{
path <- tempfile()
req |> req_perform(path = path)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/curl.R
\name{curl_translate}
\alias{curl_translate}
\alias{curl_help}
\title{Translate curl syntax to httr2}
\usage{
curl_translate(cmd, simplify_headers = TRUE)

curl_help()
}
\arguments{
\item{cmd}{Call to curl. If omitted and the clipr package is installed,
will be retrieved from the clipboard.}

\item{simplify_headers}{Remove typically unimportant headers included when
copying a curl command from the browser. This includes:
\itemize{
\item \verb{sec-fetch-*}
\item \verb{sec-ch-ua*}
\item \code{referer}, \code{pragma}, \code{connection}
}}
}
\value{
A string containing the translated httr2 code. If the input
was copied from the clipboard, the translation will be copied back
to the clipboard.
}
\description{
The curl command line tool is commonly used to demonstrate HTTP APIs and can
easily be generated from
\href{https://everything.curl.dev/cmdline/copyas.html}{browser developer tools}.
\code{curl_translate()} saves you the pain of manually translating these calls
by implementing a partial, but frequently used, subset of curl options.
Use \code{curl_help()} to see the supported options, and \code{curl_translate()}
to translate a curl invocation copy and pasted from elsewhere.

Inspired by \href{https://github.com/hrbrmstr/curlconverter}{curlconverter}
written by \href{https://rud.is/b/}{Bob Rudis}.
}
\examples{
curl_translate("curl http://example.com")
curl_translate("curl http://example.com -X DELETE")
curl_translate("curl http://example.com --header A:1 --header B:2")
curl_translate("curl http://example.com --verbose")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-retries.R
\name{req_retry}
\alias{req_retry}
\title{Automatically retry a request on failure}
\usage{
req_retry(
  req,
  max_tries = NULL,
  max_seconds = NULL,
  retry_on_failure = FALSE,
  is_transient = NULL,
  backoff = NULL,
  after = NULL,
  failure_threshold = Inf,
  failure_timeout = 30,
  failure_realm = NULL
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{max_tries, max_seconds}{Cap the maximum number of attempts
(\code{max_tries}), the total elapsed time from the first request
(\code{max_seconds}), or both.

\code{max_tries} is the total number of attempts made, so this should always
be greater than one.}

\item{retry_on_failure}{Treat low-level failures as if they are
transient errors that can be retried.}

\item{is_transient}{A predicate function that takes a single argument
(the response) and returns \code{TRUE} or \code{FALSE} specifying whether or not
the response represents a transient error.}

\item{backoff}{A function that takes a single argument (the number of failed
attempts so far) and returns the number of seconds to wait.}

\item{after}{A function that takes a single argument (the response) and
returns either a number of seconds to wait or \code{NA}. \code{NA} indicates
that a precise wait time is not available and that the \code{backoff} strategy
should be used instead.}

\item{failure_threshold, failure_timeout, failure_realm}{Set \code{failure_threshold} to activate "circuit breaking" where if a request
continues to fail after \code{failure_threshold} times, cause the request to
error until a timeout of \code{failure_timeout} seconds has elapsed. This
timeout will persist across all requests with the same \code{failure_realm}
(which defaults to the hostname of the request) and is intended to detect
failing servers without needing to wait each time.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{req_retry()} allows \code{\link[=req_perform]{req_perform()}} to automatically retry failing
requests. It's particularly important for APIs with rate limiting, but can
also be useful when dealing with flaky servers.

By default, \code{req_perform()} will retry if the response is a 429
("too many requests", often used for rate limiting) or 503
("service unavailable"). If the API you are wrapping has other transient
status codes (or conveys transience with some other property of the
response), you can override the default with \code{is_transient}. And
if you set \code{retry_on_failure = TRUE}, the request will retry
if either the HTTP request or HTTP response doesn't complete successfully,
leading to an error from curl, the lower-level library that httr2 uses to
perform HTTP requests. This occurs, for example, if your Wi-Fi is down.
\subsection{Delay}{

It's a bad idea to immediately retry a request, so \code{req_perform()} will
wait a little before trying again:
\itemize{
\item If the response contains the \code{Retry-After} header, httr2 will wait the
amount of time it specifies. If the API you are wrapping conveys this
information with a different header (or other property of the response),
you can override the default behavior with \code{retry_after}.
\item Otherwise, httr2 will use "truncated exponential backoff with full
jitter", i.e., it will wait a random amount of time between one second and
\code{2 ^ tries} seconds, capped at a maximum of 60 seconds. In other words, it
waits \code{runif(1, 1, 2)} seconds after the first failure, \code{runif(1, 1, 4)}
after the second, \code{runif(1, 1, 8)} after the third, and so on. If you'd
prefer a different strategy, you can override the default with \code{backoff}.
}
}
}
\examples{
# google APIs assume that a 500 is also a transient error
request("http://google.com") |>
  req_retry(is_transient = \(resp) resp_status(resp) \%in\% c(429, 500, 503))

# use a constant 10s delay after every failure
request("http://example.com") |>
  req_retry(backoff = \(resp) 10)

# When rate-limited, GitHub's API returns a 403 with
# `X-RateLimit-Remaining: 0` and an Unix time stored in the
# `X-RateLimit-Reset` header. This takes a bit more work to handle:
github_is_transient <- function(resp) {
  resp_status(resp) == 403 &&
    identical(resp_header(resp, "X-RateLimit-Remaining"), "0")
}
github_after <- function(resp) {
  time <- as.numeric(resp_header(resp, "X-RateLimit-Reset"))
  time - unclass(Sys.time())
}
request("http://api.github.com") |>
  req_retry(
    is_transient = github_is_transient,
    after = github_after
  )
}
\seealso{
\code{\link[=req_throttle]{req_throttle()}} if the API has a rate-limit but doesn't expose
the limits in the response.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-template.R
\name{req_template}
\alias{req_template}
\title{Set request method/path from a template}
\usage{
req_template(req, template, ..., .env = parent.frame())
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{template}{A template string which consists of a optional HTTP method
and a path containing variables labelled like either \verb{:foo} or \code{{foo}}.}

\item{...}{Template variables.}

\item{.env}{Environment in which to look for template variables not found
in \code{...}. Expert use only.}
}
\value{
A modified HTTP \link{request}.
}
\description{
Many APIs document their methods with a lightweight template mechanism
that looks like \code{GET /user/{user}} or \verb{POST /organisation/:org}. This
function makes it easy to copy and paste such snippets and retrieve template
variables either from function arguments or the current environment.

\code{req_template()} will append to the existing path so that you can set a
base url in the initial \code{\link[=request]{request()}}. This means that you'll generally want
to avoid multiple \code{req_template()} calls on the same request.
}
\examples{
httpbin <- request(example_url())

# You can supply template parameters in `...`
httpbin |> req_template("GET /bytes/{n}", n = 100)

# or you retrieve from the current environment
n <- 200
httpbin |> req_template("GET /bytes/{n}")

# Existing path is preserved:
httpbin_test <- request(example_url()) |> req_url_path("/test")
name <- "id"
value <- "a3fWa"
httpbin_test |> req_template("GET /set/{name}/{value}")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token.R
\name{oauth_token}
\alias{oauth_token}
\title{Create an OAuth token}
\usage{
oauth_token(
  access_token,
  token_type = "bearer",
  expires_in = NULL,
  refresh_token = NULL,
  ...,
  .date = Sys.time()
)
}
\arguments{
\item{access_token}{The access token used to authenticate request}

\item{token_type}{Type of token; only \code{"bearer"} is currently supported.}

\item{expires_in}{Number of seconds until token expires.}

\item{refresh_token}{Optional refresh token; if supplied, this can be
used to cheaply get a new access token when this one expires.}

\item{...}{Additional components returned by the endpoint}

\item{.date}{Date the request was made; used to convert the relative
\code{expires_in} to an absolute \code{expires_at}.}
}
\value{
An OAuth token: an S3 list with class \code{httr2_token}.
}
\description{
Creates a S3 object of class \verb{<httr2_token>} representing an OAuth token
returned from the access token endpoint.
}
\examples{
oauth_token("abcdef")
oauth_token("abcdef", expires_in = 3600)
oauth_token("abcdef", refresh_token = "ghijkl")
}
\seealso{
\code{\link[=oauth_token_cached]{oauth_token_cached()}} to use the token cache with a specified
OAuth flow.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth_token_cached}
\alias{oauth_token_cached}
\title{Retrieve an OAuth token using the cache}
\usage{
oauth_token_cached(
  client,
  flow,
  flow_params = list(),
  cache_disk = FALSE,
  cache_key = NULL,
  reauth = FALSE
)
}
\arguments{
\item{client}{An \code{\link[=oauth_client]{oauth_client()}}.}

\item{flow}{An \code{oauth_flow_} function used to generate the access token.}

\item{flow_params}{Parameters for the flow. This should be a named list
whose names match the argument names of \code{flow}.}

\item{cache_disk}{Should the access token be cached on disk? This reduces
the number of times that you need to re-authenticate at the cost of
storing access credentials on disk.

Learn more in \url{https://httr2.r-lib.org/articles/oauth.html}.}

\item{cache_key}{If you want to cache multiple tokens per app, use this
key to disambiguate them.}

\item{reauth}{Set to \code{TRUE} to force re-authentication via flow, regardless
of whether or not token is expired.}
}
\description{
This function wraps around a \code{oauth_flow_} function to retrieve a token
from the cache, or to generate and cache a token if needed. Use this for
manual token management that still takes advantage of httr2's caching
system. You should only need to use this function if you're passing
the token
}
\examples{
\dontrun{
token <- oauth_token_cached(
  client = example_github_client(),
  flow = oauth_flow_auth_code,
  flow_params = list(
    auth_url = "https://github.com/login/oauth/authorize"
  ),
  cache_disk = TRUE
)
token
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-options.R
\name{req_proxy}
\alias{req_proxy}
\title{Use a proxy for a request}
\usage{
req_proxy(
  req,
  url,
  port = NULL,
  username = NULL,
  password = NULL,
  auth = "basic"
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{url, port}{Location of proxy.}

\item{username, password}{Login details for proxy, if needed.}

\item{auth}{Type of HTTP authentication to use. Should be one of the
following: \code{basic}, \code{digest}, \code{digest_ie}, \code{gssnegotiate}, \code{ntlm}, \code{any}.}
}
\description{
Use a proxy for a request
}
\examples{
# Proxy from https://www.proxynova.com/proxy-server-list/
\dontrun{
request("http://hadley.nz") |>
  req_proxy("20.116.130.70", 3128) |>
  req_perform()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-error.R
\name{req_error}
\alias{req_error}
\title{Control handling of HTTP errors}
\usage{
req_error(req, is_error = NULL, body = NULL)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{is_error}{A predicate function that takes a single argument (the
response) and returns \code{TRUE} or \code{FALSE} indicating whether or not an
R error should be signalled.}

\item{body}{A callback function that takes a single argument (the response)
and returns a character vector of additional information to include in the
body of the error. This vector is passed along to the \code{message} argument
of \code{\link[rlang:abort]{rlang::abort()}} so you can use any formatting that it supports.}
}
\value{
A modified HTTP \link{request}.
}
\description{
\code{req_perform()} will automatically convert HTTP errors (i.e. any 4xx or 5xx
status code) into R errors. Use \code{req_error()} to either override the
defaults, or extract additional information from the response that would
be useful to expose to the user.
}
\section{Error handling}{
\code{req_perform()} is designed to succeed if and only if you get a valid HTTP
response. There are two ways a request can fail:
\itemize{
\item The HTTP request might fail, for example if the connection is dropped
or the server doesn't exist. This type of error will have class
\code{c("httr2_failure", "httr2_error")}.
\item The HTTP request might succeed, but return an HTTP status code that
represents an error, e.g. a \verb{404 Not Found} if the specified resource is
not found. This type of error will have (e.g.) class
\code{c("httr2_http_404", "httr2_http", "httr2_error")}.
}

These error classes are designed to be used in conjunction with R's
condition handling tools (\url{https://adv-r.hadley.nz/conditions.html}).
For example, if you want to return a default value when the server returns
a 404, use \code{tryCatch()}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{tryCatch(
  req |> req_perform() |> resp_body_json(),
  httr2_http_404 = function(cnd) NULL
)
}\if{html}{\out{</div>}}

Or if you want to re-throw the error with some additional context, use
\code{withCallingHandlers()}, e.g.:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{withCallingHandlers(
  req |> req_perform() |> resp_body_json(),
  httr2_http_404 = function(cnd) \{
    rlang::abort("Couldn't find user", parent = cnd)
  \}
)
}\if{html}{\out{</div>}}

Learn more about error chaining at \link[rlang:topic-error-chaining]{rlang::topic-error-chaining}.
}

\examples{
# Performing this request usually generates an error because httr2
# converts HTTP errors into R errors:
req <- request(example_url()) |>
  req_url_path("/status/404")
try(req |> req_perform())
# You can still retrieve it with last_response()
last_response()

# But you might want to suppress this behaviour:
resp <- req |>
  req_error(is_error = \(resp) FALSE) |>
  req_perform()
resp

# Or perhaps you're working with a server that routinely uses the
# wrong HTTP error codes only 500s are really errors
request("http://example.com") |>
  req_error(is_error = \(resp) resp_status(resp) == 500)

# Most typically you'll use req_error() to add additional information
# extracted from the response body (or sometimes header):
error_body <- function(resp) {
  resp_body_json(resp)$error
}
request("http://example.com") |>
  req_error(body = error_body)
# Learn more in https://httr2.r-lib.org/articles/wrapping-apis.html
}
\seealso{
\code{\link[=req_retry]{req_retry()}} to control when errors are automatically retried.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-promise.R
\name{req_perform_promise}
\alias{req_perform_promise}
\title{Perform request asynchronously using the promises package}
\usage{
req_perform_promise(
  req,
  path = NULL,
  pool = NULL,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL)
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{path}{Optionally, path to save body of the response. This is useful
for large responses since it avoids storing the response in memory.}

\item{pool}{A pool created by \code{\link[curl:multi]{curl::new_pool()}}.}

\item{verbosity}{How much information to print? This is a wrapper
around \code{\link[=req_verbose]{req_verbose()}} that uses an integer to control verbosity:
\itemize{
\item \code{0}: no output
\item \code{1}: show headers
\item \code{2}: show headers and bodies
\item \code{3}: show headers, bodies, and curl status messages.
}

Use \code{\link[=with_verbosity]{with_verbosity()}} to control the verbosity of requests that
you can't affect directly.}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}
}
\value{
a \code{\link[promises:promise]{promises::promise()}} object which resolves to a \link{response} if
successful or rejects on the same errors thrown by \code{\link[=req_perform]{req_perform()}}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This variation on \code{\link[=req_perform]{req_perform()}} returns a \code{\link[promises:promise]{promises::promise()}} object immediately
and then performs the request in the background, returning program control before the request
is finished. See the
\href{https://rstudio.github.io/promises/articles/promises_01_motivation.html}{promises package documentation}
for more details on how to work with the resulting promise object.

If using together with \code{\link[later:create_loop]{later::with_temp_loop()}} or other private event loops,
a new curl pool made by \code{\link[curl:multi]{curl::new_pool()}} should be created for requests made
within the loop to ensure that only these requests are being polled by the loop.

Like with \code{\link[=req_perform_parallel]{req_perform_parallel()}}, exercise caution when using this function;
it's easy to pummel a server with many simultaneous requests. Also, not all servers
can handle more than 1 request at a time, so the responses may still return
sequentially.

\code{req_perform_promise()} also has similar limitations to the
\code{\link[=req_perform_parallel]{req_perform_parallel()}} function, it:
\itemize{
\item Will not retrieve a new OAuth token if it expires after the promised request
is created but before it is actually requested.
\item Does not perform throttling with \code{\link[=req_throttle]{req_throttle()}}.
\item Does not attempt retries as described by \code{\link[=req_retry]{req_retry()}}.
\item Only consults the cache set by \code{\link[=req_cache]{req_cache()}} when the request is promised.
}
}
\examples{
\dontrun{
library(promises)
request_base <- request(example_url()) |> req_url_path_append("delay")

p <- request_base |> req_url_path_append(2) |> req_perform_promise()

# A promise object, not particularly useful on its own
p

# Use promise chaining functions to access results
p \%...>\%
  resp_body_json() \%...>\%
  print()


# Can run two requests at the same time
p1 <- request_base |> req_url_path_append(2) |> req_perform_promise()
p2 <- request_base |> req_url_path_append(1) |> req_perform_promise()

p1 \%...>\%
  resp_url_path \%...>\%
  paste0(., " finished") \%...>\%
  print()

p2 \%...>\%
  resp_url_path \%...>\%
  paste0(., " finished") \%...>\%
  print()

# See the [promises package documentation](https://rstudio.github.io/promises/)
# for more information on working with promises
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-status.R
\name{resp_status}
\alias{resp_status}
\alias{resp_status_desc}
\alias{resp_is_error}
\alias{resp_check_status}
\title{Extract HTTP status from response}
\usage{
resp_status(resp)

resp_status_desc(resp)

resp_is_error(resp)

resp_check_status(resp, info = NULL, error_call = caller_env())
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{info}{A character vector of additional information to include in
the error message. Passed to \code{\link[rlang:abort]{rlang::abort()}}.}

\item{error_call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\value{
\itemize{
\item \code{resp_status()} returns a scalar integer
\item \code{resp_status_desc()} returns a string
\item \code{resp_is_error()} returns \code{TRUE} or \code{FALSE}
\item \code{resp_check_status()} invisibly returns the response if it's ok;
otherwise it throws an error with class \verb{httr2_http_\{status\}}.
}
}
\description{
\itemize{
\item \code{resp_status()} retrieves the numeric HTTP status code
\item \code{resp_status_desc()} retrieves the brief textual description.
\item \code{resp_is_error()} returns \code{TRUE} if the status code represents an error
(i.e. a 4xx or 5xx status).
\item \code{resp_check_status()} turns HTTPs errors into R errors.
}

These functions are mostly for internal use because in most cases you
will only ever see a 200 response:
\itemize{
\item 1xx are handled internally by curl.
\item 3xx redirects are automatically followed. You will only see them if you
have deliberately suppressed redirects with
\code{req |> req_options(followlocation = FALSE)}.
\item 4xx client and 5xx server errors are automatically turned into R errors.
You can stop them from being turned into R errors with \code{\link[=req_error]{req_error()}},
e.g. \verb{req |> req_error(is_error = \\(resp) FALSE)}.
}
}
\examples{
# An HTTP status code you're unlikely to see in the wild:
resp <- response(418)
resp |> resp_is_error()
resp |> resp_status()
resp |> resp_status_desc()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-headers.R
\name{resp_link_url}
\alias{resp_link_url}
\title{Parse link URL from a response}
\usage{
resp_link_url(resp, rel)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{rel}{The "link relation type" value for which to retrieve a URL.}
}
\value{
Either a string providing a URL, if the specified \code{rel} exists, or
\code{NULL} if not.
}
\description{
Parses URLs out of the the \code{Link} header as defined by \href{https://datatracker.ietf.org/doc/html/rfc8288}{RFC 8288}.
}
\examples{
# Simulate response from GitHub code search
resp <- response(headers = paste0("Link: ",
  '<https://api.github.com/search/code?q=addClass+user\%3Amozilla&page=2>; rel="next",',
  '<https://api.github.com/search/code?q=addClass+user\%3Amozilla&page=34>; rel="last"'
))

resp_link_url(resp, "next")
resp_link_url(resp, "last")
resp_link_url(resp, "prev")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{url_modify}
\alias{url_modify}
\alias{url_modify_relative}
\alias{url_modify_query}
\title{Modify a URL}
\usage{
url_modify(
  url,
  scheme = as_is,
  hostname = as_is,
  username = as_is,
  password = as_is,
  port = as_is,
  path = as_is,
  query = as_is,
  fragment = as_is
)

url_modify_relative(url, relative_url)

url_modify_query(
  .url,
  ...,
  .multi = c("error", "comma", "pipe", "explode"),
  .space = c("percent", "form")
)
}
\arguments{
\item{url, .url}{A string or \link[=url_parse]{parsed URL}.}

\item{scheme}{The scheme, typically either \code{http} or \code{https}.}

\item{hostname}{The hostname, e.g., \code{www.google.com} or \code{posit.co}.}

\item{username, password}{Username and password to embed in the URL.
Not generally recommended but needed for some legacy applications.}

\item{port}{An integer port number.}

\item{path}{The path, e.g., \verb{/search}. Paths must start with \code{/}, so this
will be automatically added if omitted.}

\item{query}{Either a query string or a named list of query components.}

\item{fragment}{The fragment, e.g., \verb{#section-1}.}

\item{relative_url}{A relative URL to append to the base URL.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}>
Name-value pairs that define query parameters. Each value must be either
an atomic vector or \code{NULL} (which removes the corresponding parameters).
If you want to opt out of escaping, wrap strings in \code{I()}.}

\item{.multi}{Controls what happens when a value is a vector:
\itemize{
\item \code{"error"}, the default, throws an error.
\item \code{"comma"}, separates values with a \verb{,}, e.g. \verb{?x=1,2}.
\item \code{"pipe"}, separates values with a \code{|}, e.g. \code{?x=1|2}.
\item \code{"explode"}, turns each element into its own parameter, e.g. \code{?x=1&x=2}
}

If none of these options work for your needs, you can instead supply a
function that takes a character vector of argument values and returns a
a single string.}

\item{.space}{How should spaces in query params be escaped? The default,
"percent", uses standard percent encoding (i.e. \verb{\%20}), but you can opt-in
to "form" encoding, which uses \code{+} instead.}
}
\value{
An object of the same type as \code{url}.
}
\description{
Use \code{url_modify()} to modify any component of the URL,
\code{url_modify_relative()} to modify with a relative URL,
or \code{url_modify_query()} to modify individual query parameters.

For \code{url_modify()}, components that aren't specified in the
function call will be left as is; components set to \code{NULL} will be removed,
and all other values will be updated. Note that removing \code{scheme} or
\code{hostname} will create a relative URL.
}
\examples{
url_modify("http://hadley.nz", path = "about")
url_modify("http://hadley.nz", scheme = "https")
url_modify("http://hadley.nz/abc", path = "/cde")
url_modify("http://hadley.nz/abc", path = "")
url_modify("http://hadley.nz?a=1", query = "b=2")
url_modify("http://hadley.nz?a=1", query = list(c = 3))

url_modify_query("http://hadley.nz?a=1&b=2", c = 3)
url_modify_query("http://hadley.nz?a=1&b=2", b = NULL)
url_modify_query("http://hadley.nz?a=1&b=2", a = 100)

url_modify_relative("http://hadley.nz/a/b/c.html", "/d.html")
url_modify_relative("http://hadley.nz/a/b/c.html", "d.html")
url_modify_relative("http://hadley.nz/a/b/c.html", "../d.html")
}
\seealso{
Other URL manipulation: 
\code{\link{url_build}()},
\code{\link{url_parse}()}
}
\concept{URL manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-url.R
\name{resp_url}
\alias{resp_url}
\alias{resp_url_path}
\alias{resp_url_query}
\alias{resp_url_queries}
\title{Get URL/components from the response}
\usage{
resp_url(resp)

resp_url_path(resp)

resp_url_query(resp, name, default = NULL)

resp_url_queries(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}

\item{name}{Query parameter name.}

\item{default}{Default value to use if query parameter doesn't exist.}
}
\description{
\itemize{
\item \code{resp_url()} returns the complete url.
\item \code{resp_url_path()} returns the path component.
\item \code{resp_url_query()} returns a single query component.
\item \code{resp_url_queries()} returns the query component as a named list.
}
}
\examples{
resp <- request(example_url()) |>
  req_url_path("/get") |>
  req_url_query(hello = "world") |>
  req_perform()

resp |> resp_url()
resp |> resp_url_path()
resp |> resp_url_queries()
resp |> resp_url_query("hello")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-connection.R
\name{req_perform_connection}
\alias{req_perform_connection}
\title{Perform a request and return a streaming connection}
\usage{
req_perform_connection(
  req,
  blocking = TRUE,
  verbosity = NULL,
  mock = getOption("httr2_mock", NULL)
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{blocking}{When retrieving data, should the connection block and wait
for the desired information or immediately return what it has (possibly
nothing)?}

\item{verbosity}{How much information to print? This is a wrapper
around \code{\link[=req_verbose]{req_verbose()}} that uses an integer to control verbosity:
\itemize{
\item \code{0}: no output
\item \code{1}: show headers
\item \code{2}: show headers and bodies as they're streamed
\item \code{3}: show headers, bodies, curl status messages, raw SSEs, and stream
buffer management
}

Use \code{\link[=with_verbosity]{with_verbosity()}} to control the verbosity of requests that
you can't affect directly.}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}
}
\description{
Use \code{req_perform_connection()} to perform a request if you want to stream the
response body. A response returned by \code{req_perform_connection()} includes a
connection as the body. You can then use \code{\link[=resp_stream_raw]{resp_stream_raw()}},
\code{\link[=resp_stream_lines]{resp_stream_lines()}}, or \code{\link[=resp_stream_sse]{resp_stream_sse()}} to retrieve data a chunk at a
time. Always finish up by closing the connection by calling
\code{close(response)}.

This is an alternative interface to \code{\link[=req_perform_stream]{req_perform_stream()}} that returns a
\link[base:connections]{connection} that you can use to pull the data, rather
than providing callbacks that the data is pushed to. This is useful if you
want to do other work in between handling inputs from the stream.
}
\examples{
req <- request(example_url()) |>
  req_url_path("/stream-bytes/32768")
resp <- req_perform_connection(req)

length(resp_stream_raw(resp, kb = 16))
length(resp_stream_raw(resp, kb = 16))
# When the stream has no more data, you'll get an empty result:
length(resp_stream_raw(resp, kb = 16))

# Always close the response when you're done
close(resp)

# You can loop until complete with resp_stream_is_complete()
resp <- req_perform_connection(req)
while (!resp_stream_is_complete(resp)) {
  print(length(resp_stream_raw(resp, kb = 12)))
}
close(resp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-flow-auth-code.R
\name{oauth_redirect_uri}
\alias{oauth_redirect_uri}
\title{Default redirect url for OAuth}
\usage{
oauth_redirect_uri()
}
\description{
The default redirect uri used by \code{\link[=req_oauth_auth_code]{req_oauth_auth_code()}}. Defaults to
\verb{http://localhost} unless the \code{HTTR2_OAUTH_REDIRECT_URL} envvar is set.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-bars.R
\name{progress_bars}
\alias{progress_bars}
\title{Progress bars in httr2}
\description{
Some of httr2's perform functions have a \code{progress} argument that you can use
to create a progress bar. \code{progress} can be:
\itemize{
\item \code{FALSE}, the default: does not create a progress bar.
\item \code{TRUE}: creates a basic unnamed progress bar.
\item A string: creates a basic progress bar with the given name.
\item A named list of progress bar parameters, as described below.
}

It's good practice to name your progress bars, to make it clear what
calculation or process they belong to. We recommend keeping the names
under 20 characters, so the whole progress bar fits comfortably even on
on narrower displays.
\subsection{Progress bar parameters}{
\itemize{
\item \code{clear}: whether to remove the progress bar from the screen after
termination. Defaults to \code{TRUE}.
\item \code{format}: format string. This overrides the default format string of
the progress bar type. It must be given for the \code{custom} type.
Format strings may contain R expressions to evaluate in braces.
They support cli \link[cli:pluralization]{pluralization}, and
\link[cli:inline-markup]{styling} and they can contain special
\link[cli:progress-variables]{progress variables}.
\item \code{format_done}: format string for successful termination. By default
the same as \code{format}.
\item \code{format_failed}: format string for unsuccessful termination.
By default the same as \code{format}.
\item \code{name}: progress bar name. This is by default the empty string and it
is displayed at the beginning of the progress bar.
\item \code{type}: progress bar type. Currently supported types are:
\itemize{
\item \code{iterator}: the default, a for loop or a mapping function,
\item \code{tasks}: a (typically small) number of tasks,
\item \code{download}: download of one file,
\item \code{custom}: custom type, \code{format} must not be \code{NULL} for this type.
The default display is different for each progress bar type.
}
}
}

\subsection{Further documentation}{

purrr's progress bars are powered by cli, so see
\href{https://cli.r-lib.org/articles/progress.html}{Introduction to progress bars in cli}
and \href{https://cli.r-lib.org/articles/progress-advanced.html}{Advanced cli progress bars}
for more details.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-iterative.R
\name{req_perform_iterative}
\alias{req_perform_iterative}
\title{Perform requests iteratively, generating new requests from previous responses}
\usage{
req_perform_iterative(
  req,
  next_req,
  path = NULL,
  max_reqs = 20,
  on_error = c("stop", "return"),
  mock = getOption("httr2_mock", NULL),
  progress = TRUE
)
}
\arguments{
\item{req}{The first \link{request} to perform.}

\item{next_req}{A function that takes the previous response (\code{resp}) and
request (\code{req}) and returns a \link{request} for the next page or \code{NULL} if
the iteration should terminate. See below for more details.}

\item{path}{Optionally, path to save the body of request. This should be
a glue string that uses \code{{i}} to distinguish different requests.
Useful for large responses because it avoids storing the response in
memory.}

\item{max_reqs}{The maximum number of requests to perform. Use \code{Inf} to
perform all requests until \code{next_req()} returns \code{NULL}.}

\item{on_error}{What should happen if a request fails?
\itemize{
\item \code{"stop"}, the default: stop iterating with an error.
\item \code{"return"}: stop iterating, returning all the successful responses so
far, as well as an error object for the failed request.
}}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}

\item{progress}{Display a progress bar for the status of all requests? Use
\code{TRUE} to turn on a basic progress bar, use a string to give it a name,
or see \link{progress_bars} to customize it in other ways. Not compatible with
\code{\link[=req_progress]{req_progress()}}, as httr2 can only display a single progress bar at a
time.}
}
\value{
A list, at most length \code{max_reqs}, containing \link{response}s and possibly one
error object, if \code{on_error} is \code{"return"} and one of the requests errors.
If present, the error object will always be the last element in the list.

Only httr2 errors are captured; see \code{\link[=req_error]{req_error()}} for more details.
}
\description{
\code{req_perform_iterative()} iteratively generates and performs requests,
using a callback function, \code{next_req}, to define the next request based on
the current request and response. You will probably want to pair it with an
\link[=iterate_with_offset]{iteration helper} and use a
\link[=resps_successes]{multi-response handler} to process the result.
}
\section{\code{next_req()}}{
The key piece that makes \code{req_perform_iterative()} work is the \code{next_req()}
argument. For most common cases, you can use one of the canned helpers,
like \code{\link[=iterate_with_offset]{iterate_with_offset()}}. If, however, the API you're wrapping uses a
different pagination system, you'll need to write your own. This section
gives some advice.

Generally, your function needs to inspect the response, extract some data
from it, then use that to modify the previous request. For example, imagine
that the response returns a cursor, which needs to be added to the body of
the request. The simplest version of this function might look like this:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{next_req <- function(resp, req) \{
  cursor <- resp_body_json(resp)$next_cursor
  req |> req_body_json_modify(cursor = cursor)
\}
}\if{html}{\out{</div>}}

There's one problem here: if there are no more pages to return, then
\code{cursor} will be \code{NULL}, but \code{req_body_json_modify()} will still generate
a meaningful request. So we need to handle this specifically by
returning \code{NULL}:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{next_req <- function(resp, req) \{
  cursor <- resp_body_json(resp)$next_cursor
  if (is.null(cursor))
    return(NULL)
  req |> req_body_json_modify(cursor = cursor)
\}
}\if{html}{\out{</div>}}

A value of \code{NULL} lets \code{req_perform_iterative()} know there are no more
pages remaining.

There's one last feature you might want to add to your iterator: if you
know the total number of pages, then it's nice to let
\code{req_perform_iterative()} know so it can adjust the progress bar.
(This will only ever decrease the number of pages, not increase it.)
You can signal the total number of pages by calling \code{\link[=signal_total_pages]{signal_total_pages()}},
like this:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{next_req <- function(resp, req) \{
  body <- resp_body_json(resp)
  cursor <- body$next_cursor
  if (is.null(cursor))
    return(NULL)

  signal_total_pages(body$pages)
  req |> req_body_json_modify(cursor = cursor)
\}
}\if{html}{\out{</div>}}
}

\examples{
req <- request(example_url()) |>
  req_url_path("/iris") |>
  req_throttle(10) |>
  req_url_query(limit = 5)

resps <- req_perform_iterative(req, iterate_with_offset("page_index"))

data <- resps |> resps_data(function(resp) {
  data <- resp_body_json(resp)$data
  data.frame(
    Sepal.Length = sapply(data, `[[`, "Sepal.Length"),
    Sepal.Width = sapply(data, `[[`, "Sepal.Width"),
    Petal.Length = sapply(data, `[[`, "Petal.Length"),
    Petal.Width = sapply(data, `[[`, "Petal.Width"),
    Species = sapply(data, `[[`, "Species")
  )
})
str(data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-auth-aws.R
\name{req_auth_aws_v4}
\alias{req_auth_aws_v4}
\title{Sign a request with the AWS SigV4 signing protocol}
\usage{
req_auth_aws_v4(
  req,
  aws_access_key_id,
  aws_secret_access_key,
  aws_session_token = NULL,
  aws_service = NULL,
  aws_region = NULL
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{aws_access_key_id, aws_secret_access_key}{AWS key and secret.}

\item{aws_session_token}{AWS session token, if required.}

\item{aws_service, aws_region}{The AWS service and region to use for the
request. If not supplied, will be automatically parsed from the URL
hostname.}
}
\description{
This is a custom auth protocol implemented by AWS.
}
\examples{
\dontshow{if (httr2:::has_paws_credentials()) withAutoprint(\{ # examplesIf}
creds <- paws.common::locate_credentials()
model_id <- "anthropic.claude-3-5-sonnet-20240620-v1:0"
req <- request("https://bedrock-runtime.us-east-1.amazonaws.com")
# https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html
req <- req_url_path_append(req, "model", model_id, "converse")
req <- req_body_json(req, list(
  messages = list(list(
    role = "user",
    content = list(list(text = "What's your name?"))
  ))
))
req <- req_auth_aws_v4(
  req,
  aws_access_key_id = creds$access_key_id,
  aws_secret_access_key = creds$secret_access_key,
  aws_session_token = creds$session_token
)
resp <- req_perform_connection(req)
str(resp_body_json(resp))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp-request.R
\name{resp_request}
\alias{resp_request}
\title{Find the request responsible for a response}
\usage{
resp_request(resp)
}
\arguments{
\item{resp}{A httr2 \link{response} object, created by \code{\link[=req_perform]{req_perform()}}.}
}
\description{
To make debugging easier, httr2 includes the request that was used to
generate every response. You can use this function to access it.
}
\examples{
req <- request(example_url())
resp <- req_perform(req)
resp_request(resp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resp.R
\name{new_response}
\alias{new_response}
\title{Create a HTTP response}
\usage{
new_response(
  method,
  url,
  status_code,
  headers,
  body,
  timing = NULL,
  request = NULL,
  error_call = caller_env()
)
}
\arguments{
\item{method}{HTTP method used to retrieve the response.}

\item{url}{URL response came from; might not be the same as the URL in
the request if there were any redirects.}

\item{status_code}{HTTP status code. Must be a single integer.}

\item{headers}{HTTP headers. Can be supplied as a raw or character vector
which will be parsed using the standard rules, or a named list.}

\item{body}{The body of the response. Can be a raw vector, a \verb{<httr2_path>},
or a \link{StreamingBody}.}

\item{timing}{A named numeric vector giving the time taken by various
components.}

\item{request}{The \link{request} used to generate this response.}

\item{error_call}{Environment (on call stack) used in error messages.}
}
\value{
An HTTP response: an S3 list with class \code{httr2_response}.
}
\description{
This is the constructor function for the \code{httr2_response} S3 class. It is
useful primarily for mocking.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-mock.R
\name{with_mocked_responses}
\alias{with_mocked_responses}
\alias{local_mocked_responses}
\title{Temporarily mock requests}
\usage{
with_mocked_responses(mock, code)

local_mocked_responses(mock, env = caller_env())
}
\arguments{
\item{mock}{A function, a list, or \code{NULL}.
\itemize{
\item \code{NULL} disables mocking and returns httr2 to regular operation.
\item A list of responses will be returned in sequence. After all responses
have been used up, will return 503 server errors.
\item For maximum flexibility, you can supply a function that that takes a
single argument, \code{req}, and returns either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does).
}}

\item{code}{Code to execute in the temporary environment.}

\item{env}{Environment to use for scoping changes.}
}
\value{
\code{with_mocked_responses()} returns the result of evaluating \code{code}.
}
\description{
Mocking allows you to selectively and temporarily replace the response
you would typically receive from a request with your own code. These
functions are low-level and we don't recommend using them directly.
Instead use package that uses these functions under the hood, like
\pkg{httptest2} or \pkg{vcr}.
}
\examples{
# This function should perform a response against google.com:
google <- function() {
  request("http://google.com") |>
    req_perform()
}

# But I can use a mock to instead return my own made up response:
my_mock <- function(req) {
  response(status_code = 403)
}
try(with_mocked_responses(my_mock, google()))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-cache.R
\name{req_cache}
\alias{req_cache}
\title{Automatically cache requests}
\usage{
req_cache(
  req,
  path,
  use_on_error = FALSE,
  debug = getOption("httr2_cache_debug", FALSE),
  max_age = Inf,
  max_n = Inf,
  max_size = 1024^3
)
}
\arguments{
\item{req}{A httr2 \link{request} object.}

\item{path}{Path to cache directory. Will be created automatically if it
does not exist.

For quick and easy caching within a session, you can use \code{tempfile()}.
To cache requests within a package, you can use something like
\code{file.path(tools::R_user_dir("pkgdown", "cache"), "httr2")}.

httr2 doesn't provide helpers to manage the cache, but if you want to
empty it, you can use something like
\code{unlink(dir(cache_path, full.names = TRUE))}.}

\item{use_on_error}{If the request errors, and there's a cache response,
should \code{req_perform()} return that instead of generating an error?}

\item{debug}{When \code{TRUE} will emit useful messages telling you about
cache hits and misses. This can be helpful to understand whether or
not caching is actually doing anything for your use case.}

\item{max_n, max_age, max_size}{Automatically prune the cache by specifying
one or more of:
\itemize{
\item \code{max_age}: to delete files older than this number of seconds.
\item \code{max_n}: to delete files (from oldest to newest) to preserve at
most this many files.
\item \code{max_size}: to delete files (from oldest to newest) to preserve at
most this many bytes.
}

The cache pruning is performed at most once per minute.}
}
\value{
A modified HTTP \link{request}.
}
\description{
Use \code{req_cache()} to automatically cache HTTP requests. Most API requests
are not cacheable, but static files often are.

\code{req_cache()} caches responses to GET requests that have status code 200 and
at least one of the standard caching headers (e.g. \code{Expires},
\code{Etag}, \code{Last-Modified}, \code{Cache-Control}), unless caching has been expressly
prohibited with \code{Cache-Control: no-store}. Typically, a request will still
be sent to the server to check that the cached value is still up-to-date,
but it will not need to re-download the body value.

To learn more about HTTP caching, I recommend the MDN article
\href{https://developer.mozilla.org/en-US/docs/Web/HTTP/Caching}{HTTP caching}.
}
\examples{
# GitHub uses HTTP caching for all raw files.
url <- paste0(
  "https://raw.githubusercontent.com/allisonhorst/palmerpenguins/",
  "master/inst/extdata/penguins.csv"
)
# Here I set debug = TRUE so you can see what's happening
req <- request(url) |> req_cache(tempdir(), debug = TRUE)

# First request downloads the data
resp <- req |> req_perform()

# Second request retrieves it from the cache
resp <- req |> req_perform()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/req-perform-parallel.R
\name{req_perform_parallel}
\alias{req_perform_parallel}
\title{Perform a list of requests in parallel}
\usage{
req_perform_parallel(
  reqs,
  paths = NULL,
  on_error = c("stop", "return", "continue"),
  progress = TRUE,
  max_active = 10,
  mock = getOption("httr2_mock", NULL)
)
}
\arguments{
\item{reqs}{A list of \link{request}s.}

\item{paths}{An optional character vector of paths, if you want to download
the response bodies to disk. If supplied, must be the same length as
\code{reqs}.}

\item{on_error}{What should happen if one of the requests fails?
\itemize{
\item \code{stop}, the default: stop iterating with an error.
\item \code{return}: stop iterating, returning all the successful responses
received so far, as well as an error object for the failed request.
\item \code{continue}: continue iterating, recording errors in the result.
}}

\item{progress}{Display a progress bar for the status of all requests? Use
\code{TRUE} to turn on a basic progress bar, use a string to give it a name,
or see \link{progress_bars} to customize it in other ways. Not compatible with
\code{\link[=req_progress]{req_progress()}}, as httr2 can only display a single progress bar at a
time.}

\item{max_active}{Maximum number of concurrent requests.}

\item{mock}{A mocking function. If supplied, this function is called
with the request. It should return either \code{NULL} (if it doesn't want to
handle the request) or a \link{response} (if it does). See
\code{\link[=with_mocked_responses]{with_mocked_responses()}}/\code{local_mocked_responses()} for more details.}
}
\value{
A list, the same length as \code{reqs}, containing \link{response}s and possibly
error objects, if \code{on_error} is \code{"return"} or \code{"continue"} and one of the
responses errors. If \code{on_error} is \code{"return"} and it errors on the ith
request, the ith element of the result will be an error object, and the
remaining elements will be \code{NULL}. If \code{on_error} is \code{"continue"}, it will
be a mix of requests and error objects.

Only httr2 errors are captured; see \code{\link[=req_error]{req_error()}} for more details.
}
\description{
This variation on \code{\link[=req_perform_sequential]{req_perform_sequential()}} performs multiple requests in
parallel. Never use it without \code{\link[=req_throttle]{req_throttle()}}; otherwise it's too easy to
pummel a server with a very large number of simultaneous requests.

While running, you'll get a progress bar that looks like:
\verb{[working] (1 + 4) -> 5 -> 5}. The string tells you the current status of
the queue (e.g. working, waiting, errored) followed by (the
number of pending requests + pending retried requests) -> the number of
active requests -> the number of complete requests.
\subsection{Limitations}{

The main limitation of \code{req_perform_parallel()} is that it assumes applies
\code{\link[=req_throttle]{req_throttle()}} and \code{\link[=req_retry]{req_retry()}} are across all requests. This means,
for example, that if request 1 is throttled, but request 2 is not,
\code{req_perform_parallel()} will wait for request 1 before performing request 2.
This makes it most suitable for performing many parallel requests to the same
host, rather than a mix of different hosts. It's probably possible to remove
these limitation, but it's enough work that I'm unlikely to do it unless
I know that people would fine it useful: so please let me know!

Additionally, it does not respect the \code{max_tries} argument to \code{req_retry()}
because if you have five requests in flight and the first one gets rate
limited, it's likely that all the others do too. This also means that
the circuit breaker is never triggered.
}
}
\examples{
# Requesting these 4 pages one at a time would take 2 seconds:
request_base <- request(example_url()) |>
  req_throttle(capacity = 100, fill_time_s = 60)
reqs <- list(
  request_base |> req_url_path("/delay/0.5"),
  request_base |> req_url_path("/delay/0.5"),
  request_base |> req_url_path("/delay/0.5"),
  request_base |> req_url_path("/delay/0.5")
)
# But it's much faster if you request in parallel
system.time(resps <- req_perform_parallel(reqs))

# req_perform_parallel() will fail on error
reqs <- list(
  request_base |> req_url_path("/status/200"),
  request_base |> req_url_path("/status/400"),
  request("FAILURE")
)
try(resps <- req_perform_parallel(reqs))

# but can use on_error to capture all successful results
resps <- req_perform_parallel(reqs, on_error = "continue")

# Inspect the successful responses
resps |> resps_successes()

# And the failed responses
resps |> resps_failures() |> resps_requests()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth_cache_prune}
\alias{oauth_cache_prune}
\title{Prune the OAuth token cache}
\usage{
oauth_cache_prune(max_age_days = 30)
}
\arguments{
\item{max_age_days}{Delete cached tokens that haven't been modified in
this many days.}
}
\description{
Deletes cached OAuth tokens (from both the current and legacy cache
directories, see \code{\link[=oauth_cache_path]{oauth_cache_path()}}) that are older than
\code{max_age_days}. This is called automatically when httr2 is loaded, so
you should only need to call it yourself if you want to prune the cache
immediately.
}
